@@ -19,6 +19,7 @@
 
 #include <gutil/strings/substitute.h>
 #include <thrift/protocol/TDebugProtocol.h>
+#include <thrift/protocol/TJSONProtocol.h>
 
 #include "catalog/catalog-util.h"
 #include "exec/read-write-util.h"
@@ -31,6 +32,7 @@
 #include "util/event-metrics.h"
 #include "util/logging-support.h"
 #include "util/metrics.h"
+#include "util/thrift-debug-util.h"
 #include "util/webserver.h"
 
 #include "common/names.h"
@@ -38,7 +40,6 @@
 using boost::bind;
 using boost::mem_fn;
 using namespace apache::thrift;
-using namespace impala;
 using namespace rapidjson;
 using namespace strings;
 
@@ -73,12 +74,42 @@ DEFINE_int64_hidden(catalog_partial_fetch_rpc_queue_timeout_s, LLONG_MAX, "Maxim
     "(in seconds) a partial catalog object fetch RPC spends in the queue waiting "
     "to run. Must be set to a value greater than zero.");
 
+DEFINE_int32(catalog_max_lock_skipped_topic_updates, 3, "Maximum number of topic "
+    "updates skipped for a table due to lock contention in catalogd after which it must"
+    "be added to the topic the update log. This limit only applies to distinct lock "
+    "operations which block the topic update thread.");
+
+DEFINE_int64(topic_update_tbl_max_wait_time_ms, 120000, "Maximum time "
+     "(in milliseconds) catalog's topic update thread will wait to acquire lock on "
+     "table. If the topic update thread cannot acquire a table lock it skips the table "
+     "from that topic update and processes the table in the next update. However to "
+     "prevent starvation it only skips the table catalog_max_lock_skipped_topic_updates "
+     "many times. After that limit is hit, topic thread block until it acquires the "
+     "table lock. A value of 0 disables the timeout based locking which means topic "
+     "update thread will always block until table lock is acquired.");
+
+DEFINE_int32(max_wait_time_for_sync_ddl_s, 0, "Maximum time (in seconds) until "
+     "which a sync ddl operation will wait for the updated tables "
+     "to be the added to the catalog topic. A value of 0 means sync ddl operation will "
+     "wait as long as necessary until the update is propogated to all the coordinators. "
+     "This flag only takes effect when topic_update_tbl_max_wait_time_ms is enabled."
+     "A value greater than 0 means catalogd will wait until that number of seconds "
+     "before throwing an error indicating that not all the "
+     "coordinators might have applied the changes caused due to the ddl.");
+
+DECLARE_string(debug_actions);
 DECLARE_string(state_store_host);
 DECLARE_int32(state_store_subscriber_port);
 DECLARE_int32(state_store_port);
 DECLARE_string(hostname);
 DECLARE_bool(compact_catalog_topic);
 
+#ifndef NDEBUG
+DECLARE_int32(stress_catalog_startup_delay_ms);
+#endif
+
+namespace impala {
+
 string CatalogServer::IMPALA_CATALOG_TOPIC = "catalog-update";
 
 const string CATALOG_SERVER_TOPIC_PROCESSING_TIMES =
@@ -123,6 +154,7 @@ class CatalogServiceThriftIf : public CatalogServiceIf {
   void ResetMetadata(TResetMetadataResponse& resp, const TResetMetadataRequest& req)
       override {
     VLOG_RPC << "ResetMetadata(): request=" << ThriftDebugString(req);
+    DebugActionNoFail(FLAGS_debug_actions, "RESET_METADATA_DELAY");
     Status status = catalog_server_->catalog()->ResetMetadata(req, &resp);
     if (!status.ok()) LOG(ERROR) << status.GetDetail();
     TStatus thrift_status;
@@ -164,7 +196,7 @@ class CatalogServiceThriftIf : public CatalogServiceIf {
     Status status = catalog_server_->catalog()->GetCatalogObject(req.object_desc,
         &resp.catalog_object);
     if (!status.ok()) LOG(ERROR) << status.GetDetail();
-    VLOG_RPC << "GetCatalogObject(): response=" << ThriftDebugString(resp);
+    VLOG_RPC << "GetCatalogObject(): response=" << ThriftDebugStringNoThrow(resp);
   }
 
   void GetPartialCatalogObject(TGetPartialCatalogObjectResponse& resp,
@@ -182,7 +214,7 @@ class CatalogServiceThriftIf : public CatalogServiceIf {
     TStatus thrift_status;
     status.ToThrift(&thrift_status);
     resp.__set_status(thrift_status);
-    VLOG_RPC << "GetPartialCatalogObject(): response=" << ThriftDebugString(resp);
+    VLOG_RPC << "GetPartialCatalogObject(): response=" << ThriftDebugStringNoThrow(resp);
   }
 
   void GetPartitionStats(TGetPartitionStatsResponse& resp,
@@ -193,7 +225,7 @@ class CatalogServiceThriftIf : public CatalogServiceIf {
     TStatus thrift_status;
     status.ToThrift(&thrift_status);
     resp.__set_status(thrift_status);
-    VLOG_RPC << "GetPartitionStats(): response=" << ThriftDebugString(resp);
+    VLOG_RPC << "GetPartitionStats(): response=" << ThriftDebugStringNoThrow(resp);
   }
 
   // Prioritizes the loading of metadata for one or more catalog objects. Currently only
@@ -210,17 +242,6 @@ class CatalogServiceThriftIf : public CatalogServiceIf {
     VLOG_RPC << "PrioritizeLoad(): response=" << ThriftDebugString(resp);
   }
 
-  void SentryAdminCheck(TSentryAdminCheckResponse& resp,
-      const TSentryAdminCheckRequest& req) override {
-    VLOG_RPC << "SentryAdminCheck(): request=" << ThriftDebugString(req);
-    Status status = catalog_server_->catalog()->SentryAdminCheck(req, &resp);
-    if (!status.ok()) LOG(ERROR) << status.GetDetail();
-    TStatus thrift_status;
-    status.ToThrift(&thrift_status);
-    resp.__set_status(thrift_status);
-    VLOG_RPC << "SentryAdminCheck(): response=" << ThriftDebugString(resp);
-  }
-
   void UpdateTableUsage(TUpdateTableUsageResponse& resp,
       const TUpdateTableUsageRequest& req) override {
     VLOG_RPC << "UpdateTableUsage(): request=" << ThriftDebugString(req);
@@ -253,6 +274,11 @@ Status CatalogServer::Start() {
 
   // This will trigger a full Catalog metadata load.
   catalog_.reset(new Catalog());
+#ifndef NDEBUG
+  if (FLAGS_stress_catalog_startup_delay_ms > 0) {
+    SleepForMs(FLAGS_stress_catalog_startup_delay_ms);
+  }
+#endif
   RETURN_IF_ERROR(Thread::Create("catalog-server", "catalog-update-gathering-thread",
       &CatalogServer::GatherCatalogUpdatesThread, this,
       &catalog_update_gathering_thread_));
@@ -621,6 +647,7 @@ void CatalogServer::CatalogObjectsUrlCallback(const Webserver::WebRequest& req,
   const auto& args = req.parsed_args;
   Webserver::ArgumentMap::const_iterator object_type_arg = args.find("object_type");
   Webserver::ArgumentMap::const_iterator object_name_arg = args.find("object_name");
+  Webserver::ArgumentMap::const_iterator json_arg = args.find("json");
   if (object_type_arg != args.end() && object_name_arg != args.end()) {
     TCatalogObjectType::type object_type =
         TCatalogObjectTypeFromName(object_type_arg->second);
@@ -630,13 +657,28 @@ void CatalogServer::CatalogObjectsUrlCallback(const Webserver::WebRequest& req,
     Status status =
         TCatalogObjectFromObjectName(object_type, object_name_arg->second, &request);
 
-    // Get the object and dump its contents.
-    TCatalogObject result;
-    if (status.ok()) status = catalog_->GetCatalogObject(request, &result);
-    if (status.ok()) {
-      Value debug_string(ThriftDebugString(result).c_str(), document->GetAllocator());
-      document->AddMember("thrift_string", debug_string, document->GetAllocator());
+    if (json_arg != args.end()) {
+      // Get the JSON string from FE since Thrift doesn't have a cpp implementation for
+      // SimpleJsonProtocol (THRIFT-2476), so we use it's java implementation.
+      // TODO: switch to use cpp implementation of SimpleJsonProtocol after THRIFT-2476
+      //  is resolved.
+      string json_str;
+      if (status.ok()) status = catalog_->GetJsonCatalogObject(request, &json_str);
+      if (status.ok()) {
+        Value debug_string(json_str.c_str(), document->GetAllocator());
+        document->AddMember("json_string", debug_string, document->GetAllocator());
+      }
     } else {
+      // Get the object and dump its contents.
+      TCatalogObject result;
+      if (status.ok()) status = catalog_->GetCatalogObject(request, &result);
+      if(status.ok()) {
+        Value debug_string(
+            ThriftDebugStringNoThrow(result).c_str(), document->GetAllocator());
+        document->AddMember("thrift_string", debug_string, document->GetAllocator());
+      }
+    }
+    if (!status.ok()) {
       Value error(status.GetDetail().c_str(), document->GetAllocator());
       document->AddMember("error", error, document->GetAllocator());
     }
@@ -675,7 +717,7 @@ void CatalogServer::OperationUsageUrlCallback(
   // Create a summary and add it to the document
   map<string, int> aggregated_operations;
   for (const auto& catalog_op : opeartion_usage.catalog_op_counters) {
-    ++aggregated_operations[catalog_op.catalog_op_name];
+    aggregated_operations[catalog_op.catalog_op_name] += catalog_op.op_counter;
   }
   Value catalog_op_summary(kArrayType);
   for (const auto& catalog_op : aggregated_operations) {
@@ -723,7 +765,7 @@ void CatalogServer::TableMetricsUrlCallback(const Webserver::WebRequest& req,
   }
 }
 
-bool CatalogServer::AddPendingTopicItem(std::string key, int64_t version,
+int CatalogServer::AddPendingTopicItem(std::string key, int64_t version,
     const uint8_t* item_data, uint32_t size, bool deleted) {
   pending_topic_updates_.emplace_back();
   TTopicItem& item = pending_topic_updates_.back();
@@ -732,7 +774,7 @@ bool CatalogServer::AddPendingTopicItem(std::string key, int64_t version,
     if (!status.ok()) {
       pending_topic_updates_.pop_back();
       LOG(ERROR) << "Error compressing topic item: " << status.GetDetail();
-      return false;
+      return -1;
     }
   } else {
     item.value.assign(reinterpret_cast<const char*>(item_data),
@@ -740,11 +782,13 @@ bool CatalogServer::AddPendingTopicItem(std::string key, int64_t version,
   }
   item.key = std::move(key);
   item.deleted = deleted;
+  // Skip logging partition items since FE will log their summary (IMPALA-10076).
+  if (item.key.find("HDFS_PARTITION") != string::npos) return item.value.size();
   VLOG(1) << "Collected " << (deleted ? "deletion: " : "update: ") << item.key
           << ", version=" << version << ", original size=" << size
           << (FLAGS_compact_catalog_topic ?
               Substitute(", compressed size=$0", item.value.size()) : string());
-  return true;
+  return item.value.size();
 }
 
 void CatalogServer::MarkServiceAsStarted() { service_started_ = true; }
@@ -759,3 +803,5 @@ void CatalogServer::HealthzHandler(
   *(data) << "Not Available";
   *response = HttpStatusCode::ServiceUnavailable;
 }
+
+}
