@@ -18,7 +18,6 @@
 #include "codegen/llvm-codegen.h"
 
 #include <fstream>
-#include <iostream>
 #include <sstream>
 #include <unordered_set>
 
@@ -58,6 +57,7 @@
 
 #include "codegen/codegen-anyval.h"
 #include "codegen/codegen-callgraph.h"
+#include "codegen/codegen-fn-ptr.h"
 #include "codegen/codegen-symbol-emitter.h"
 #include "codegen/impala-ir-data.h"
 #include "codegen/instruction-counter.h"
@@ -71,15 +71,19 @@
 #include "runtime/lib-cache.h"
 #include "runtime/mem-pool.h"
 #include "runtime/mem-tracker.h"
+#include "runtime/fragment-state.h"
 #include "runtime/runtime-state.h"
 #include "runtime/string-value.h"
 #include "runtime/timestamp-value.h"
+#include "gutil/sysinfo.h"
 #include "util/cpu-info.h"
+#include "util/debug-util.h"
 #include "util/hdfs-util.h"
 #include "util/path-builder.h"
 #include "util/runtime-profile-counters.h"
 #include "util/symbols-util.h"
 #include "util/test-info.h"
+#include "util/thread.h"
 
 #include "common/names.h"
 
@@ -108,6 +112,12 @@ DECLARE_string(local_library_dir);
 // avx512ifma,avx512pf,avx512vbmi,avx512vl,clflushopt,clwb,fma4,mwaitx.1.2,pcommit,pku,
 // prefetchwt1,sgx,sha,sse4a,tbm,xop,xsavec,xsaves. If new attrs are added to LLVM,
 // they will be disabled until added to this whitelist.
+#ifdef __aarch64__
+DEFINE_string_hidden(llvm_cpu_attr_whitelist, "crc,neon,fp-armv8,crypto",
+    "(Experimental) a comma-separated list of LLVM CPU attribute flags that are enabled "
+    "for runtime code generation. This flag is provided to enable additional LLVM CPU "
+    "attribute flags for testing.");
+#else
 DEFINE_string_hidden(llvm_cpu_attr_whitelist, "adx,aes,avx,avx2,bmi,bmi2,cmov,cx16,f16c,"
     "fma,fsgsbase,hle,invpcid,lzcnt,mmx,movbe,pclmul,popcnt,prfchw,rdrnd,rdseed,rtm,smap,"
     "sse,sse2,sse3,sse4.1,sse4.2,ssse3,xsave,xsaveopt",
@@ -115,9 +125,12 @@ DEFINE_string_hidden(llvm_cpu_attr_whitelist, "adx,aes,avx,avx2,bmi,bmi2,cmov,cx
     "for runtime code generation. The default flags are a known-good set that are "
     "routinely tested. This flag is provided to enable additional LLVM CPU attribute "
     "flags for testing.");
+#endif
+DECLARE_bool(enable_legacy_avx_support);
 
 namespace impala {
 
+const string LlvmCodeGen::ASYNC_CODEGEN_THREAD_COUNTERS_PREFIX = "CodegenCompileThread";
 bool LlvmCodeGen::llvm_initialized_ = false;
 string LlvmCodeGen::cpu_name_;
 std::unordered_set<string> LlvmCodeGen::cpu_attrs_;
@@ -194,7 +207,7 @@ Status LlvmCodeGen::InitializeLlvm(bool load_backend) {
   return Status::OK();
 }
 
-LlvmCodeGen::LlvmCodeGen(RuntimeState* state, ObjectPool* pool,
+LlvmCodeGen::LlvmCodeGen(FragmentState* state, ObjectPool* pool,
     MemTracker* parent_mem_tracker, const string& id)
   : state_(state),
     id_(id),
@@ -216,12 +229,15 @@ LlvmCodeGen::LlvmCodeGen(RuntimeState* state, ObjectPool* pool,
   ir_generation_timer_ = ADD_TIMER(profile_, "IrGenerationTime");
   optimization_timer_ = ADD_TIMER(profile_, "OptimizationTime");
   compile_timer_ = ADD_TIMER(profile_, "CompileTime");
+  main_thread_timer_ = ADD_TIMER(profile_, "MainThreadCodegenTime");
+  compile_thread_counters_ = ADD_THREAD_COUNTERS(profile_,
+      ASYNC_CODEGEN_THREAD_COUNTERS_PREFIX);
   num_functions_ = ADD_COUNTER(profile_, "NumFunctions", TUnit::UNIT);
   num_instructions_ = ADD_COUNTER(profile_, "NumInstructions", TUnit::UNIT);
   llvm_thread_counters_ = ADD_THREAD_COUNTERS(profile_, "Codegen");
 }
 
-Status LlvmCodeGen::CreateFromFile(RuntimeState* state, ObjectPool* pool,
+Status LlvmCodeGen::CreateFromFile(FragmentState* state, ObjectPool* pool,
     MemTracker* parent_mem_tracker, const string& file, const string& id,
     scoped_ptr<LlvmCodeGen>* codegen) {
   codegen->reset(new LlvmCodeGen(state, pool, parent_mem_tracker, id));
@@ -239,26 +255,42 @@ error:
   return status;
 }
 
-Status LlvmCodeGen::CreateFromMemory(RuntimeState* state, ObjectPool* pool,
+Status LlvmCodeGen::CreateFromMemory(FragmentState* state, ObjectPool* pool,
     MemTracker* parent_mem_tracker, const string& id, scoped_ptr<LlvmCodeGen>* codegen) {
   codegen->reset(new LlvmCodeGen(state, pool, parent_mem_tracker, id));
   SCOPED_TIMER((*codegen)->profile_->total_time_counter());
   SCOPED_TIMER((*codegen)->prepare_module_timer_);
   SCOPED_THREAD_COUNTER_MEASUREMENT((*codegen)->llvm_thread_counters());
 
-  // Select the appropriate IR version. We cannot use LLVM IR with SSE4.2 instructions on
-  // a machine without SSE4.2 support.
   llvm::StringRef module_ir;
   string module_name;
-  if (IsCPUFeatureEnabled(CpuInfo::SSE4_2)) {
+
+#if __x86_64__
+  // By default, Impala now requires AVX2 support, but the enable_legacy_avx_support
+  // flag can allow running on AVX machines. The minimum requirement must have already
+  // been enforced prior to this call, so this only needs to select the appropriate
+  // LLVM IR to use.
+  if (IsCPUFeatureEnabled(CpuInfo::AVX2)) {
+    // Use the default IR that supports AVX2
     module_ir = llvm::StringRef(
-        reinterpret_cast<const char*>(impala_sse_llvm_ir), impala_sse_llvm_ir_len);
-    module_name = "Impala IR with SSE 4.2 support";
+        reinterpret_cast<const char*>(impala_llvm_ir), impala_llvm_ir_len);
+    module_name = "Impala IR with AVX2 support";
+  } else if (FLAGS_enable_legacy_avx_support && IsCPUFeatureEnabled(CpuInfo::AVX)) {
+    // If there is no AVX but legacy mode is enabled, use legacy IR with AVX support
+    module_ir = llvm::StringRef(
+        reinterpret_cast<const char*>(impala_legacy_avx_llvm_ir),
+        impala_legacy_avx_llvm_ir_len);
+    module_name = "Legacy Impala IR with AVX support";
   } else {
-    module_ir = llvm::StringRef(
-        reinterpret_cast<const char*>(impala_no_sse_llvm_ir), impala_no_sse_llvm_ir_len);
-    module_name = "Impala IR with no SSE 4.2 support";
+    // This should have been enforced earlier.
+    CHECK(false) << "CPU is missing AVX/AVX2 support";
   }
+#else
+  // Non-x86_64 always use the default IR
+  module_ir = llvm::StringRef(
+      reinterpret_cast<const char*>(impala_llvm_ir), impala_llvm_ir_len);
+  module_name = "Impala IR";
+#endif
 
   unique_ptr<llvm::MemoryBuffer> module_ir_buf(
       llvm::MemoryBuffer::getMemBuffer(module_ir, "", false));
@@ -369,7 +401,7 @@ void LlvmCodeGen::StripGlobalCtorsDtors(llvm::Module* module) {
   if (destructors != NULL) destructors->eraseFromParent();
 }
 
-Status LlvmCodeGen::CreateImpalaCodegen(RuntimeState* state,
+Status LlvmCodeGen::CreateImpalaCodegen(FragmentState* state,
     MemTracker* parent_mem_tracker, const string& id,
     scoped_ptr<LlvmCodeGen>* codegen_ret) {
   DCHECK(state != nullptr);
@@ -470,6 +502,8 @@ LlvmCodeGen::~LlvmCodeGen() {
 }
 
 void LlvmCodeGen::Close() {
+  if (async_compile_thread_ != nullptr) async_compile_thread_->Join();
+
   if (memory_manager_ != nullptr) {
     mem_tracker_->Release(memory_manager_->bytes_tracked());
     memory_manager_ = nullptr;
@@ -585,13 +619,6 @@ llvm::PointerType* LlvmCodeGen::GetNamedPtrPtrType(const string& name) {
   return llvm::PointerType::get(GetNamedPtrType(name), 0);
 }
 
-// Llvm doesn't let you create a PointerValue from a c-side ptr.  Instead
-// cast it to an int and then to 'type'.
-llvm::Value* LlvmCodeGen::CastPtrToLlvmPtr(llvm::Type* type, const void* ptr) {
-  llvm::Constant* const_int = GetI64Constant((int64_t)ptr);
-  return llvm::ConstantExpr::getIntToPtr(const_int, type);
-}
-
 llvm::Constant* LlvmCodeGen::GetIntConstant(
     int num_bytes, uint64_t low_bits, uint64_t high_bits) {
   DCHECK_GE(num_bytes, 1);
@@ -601,7 +628,8 @@ llvm::Constant* LlvmCodeGen::GetIntConstant(
   return llvm::ConstantInt::get(context(), llvm::APInt(8 * num_bytes, vals));
 }
 
-llvm::Value* LlvmCodeGen::GetStringConstant(LlvmBuilder* builder, char* data, int len) {
+llvm::Value* LlvmCodeGen::GetStringConstant(
+    LlvmBuilder* builder, const char* data, int len) {
   // Create a global string with private linkage.
   llvm::Constant* const_string =
       llvm::ConstantDataArray::getString(context(), llvm::StringRef(data, len), false);
@@ -876,7 +904,12 @@ Status LlvmCodeGen::LoadFunction(const TFunction& fn, const string& symbol,
     // declaration, not a definition, since we do not create any basic blocks or
     // instructions in it.
     *llvm_fn = prototype.GeneratePrototype(nullptr, nullptr);
-
+#ifdef __aarch64__
+    if (is_decimal) {
+      // Mark first argument as sret
+      (*llvm_fn)->addAttribute(1, llvm::Attribute::StructRet);
+    }
+#endif
     // Associate the dynamically loaded function pointer with the Function* we defined.
     // This tells LLVM where the compiled function definition is located in memory.
     execution_engine_->addGlobalMapping(*llvm_fn, fn_ptr);
@@ -993,8 +1026,9 @@ int LlvmCodeGen::InlineConstFnAttrs(const FunctionContext::TypeDesc& ret_type,
     int i_val = static_cast<int>(i_arg->getSExtValue());
     DCHECK(state_ != nullptr);
     // All supported constants are currently integers.
-    call_instr->replaceAllUsesWith(GetI32Constant(
-        FunctionContextImpl::GetConstFnAttr(state_, ret_type, arg_types, t_val, i_val)));
+    call_instr->replaceAllUsesWith(GetI32Constant(FunctionContextImpl::GetConstFnAttr(
+        state_->query_options().decimal_v2, state_->query_options().utf8_mode, ret_type,
+        arg_types, t_val, i_val)));
     call_instr->eraseFromParent();
     ++replaced;
   }
@@ -1143,14 +1177,7 @@ Status LlvmCodeGen::FinalizeModule() {
     execution_engine_->finalizeObject();
   }
 
-  // Get pointers to all codegen'd functions
-  for (int i = 0; i < fns_to_jit_compile_.size(); ++i) {
-    llvm::Function* function = fns_to_jit_compile_[i].first;
-    void* jitted_function = execution_engine_->getPointerToFunction(function);
-    DCHECK(jitted_function != NULL) << "Failed to jit " << function->getName().data();
-    *fns_to_jit_compile_[i].second = jitted_function;
-  }
-
+  SetFunctionPointers();
   DestroyModule();
 
   // Track the memory consumed by the compiled code.
@@ -1164,6 +1191,33 @@ Status LlvmCodeGen::FinalizeModule() {
   return Status::OK();
 }
 
+Status LlvmCodeGen::FinalizeModuleAsync(RuntimeProfile::EventSequence* event_sequence) {
+  DCHECK(event_sequence != nullptr);
+  Status thread_start_status = Thread::Create("async-codegen", "async-codegen",
+      [this, event_sequence]() {
+        SCOPED_THREAD_COUNTER_MEASUREMENT(compile_thread_counters_);
+        VLOG(2) << "Starting async code generation.";
+
+        Status status = DebugAction(state_->query_options(),
+            "BEFORE_CODEGEN_IN_ASYNC_CODEGEN_THREAD");
+        if (status.ok()) {
+          status = this->FinalizeModule();
+        }
+
+        const std::string status_msg = status.ok() ? "OK." : status.msg().msg();
+        auto log_level = status.ok() ? 2 : 1;
+        event_sequence->MarkEvent("AsyncCodegenFinished");
+        VLOG(log_level) << "Finished async code generation with result: " << status;
+      }, &async_compile_thread_);
+
+  event_sequence->MarkEvent("AsyncCodegenStarted");
+
+  RETURN_IF_ERROR(DebugAction(state_->query_options(),
+        "AFTER_STARTING_ASYNC_CODEGEN_IN_FRAGMENT_THREAD"));
+  return thread_start_status;
+}
+
+/// TODO: In asynchronous mode, return early if the query is cancelled or finished.
 Status LlvmCodeGen::OptimizeModule() {
   SCOPED_TIMER(optimization_timer_);
 
@@ -1252,6 +1306,17 @@ Status LlvmCodeGen::OptimizeModule() {
   return Status::OK();
 }
 
+void LlvmCodeGen::SetFunctionPointers() {
+  // Get pointers to all codegen'd functions.
+  for (const std::pair<llvm::Function*, CodegenFnPtrBase*>& fn_pair
+      : fns_to_jit_compile_) {
+    llvm::Function* function = fn_pair.first;
+    void* jitted_function = execution_engine_->getPointerToFunction(function);
+    DCHECK(jitted_function != nullptr) << "Failed to jit " << function->getName().data();
+    fn_pair.second->store(jitted_function);
+  }
+}
+
 void LlvmCodeGen::DestroyModule() {
   // Clear all references to LLVM objects owned by the module.
   cross_compiled_functions_.clear();
@@ -1265,7 +1330,7 @@ void LlvmCodeGen::DestroyModule() {
   module_ = NULL;
 }
 
-void LlvmCodeGen::AddFunctionToJit(llvm::Function* fn, void** fn_ptr) {
+void LlvmCodeGen::AddFunctionToJit(llvm::Function* fn, CodegenFnPtrBase* fn_ptr) {
   DCHECK(finalized_functions_.find(fn) != finalized_functions_.end())
       << "Attempted to add a non-finalized function to Jit: " << fn->getName().str();
   llvm::Type* decimal_val_type = GetNamedType(CodegenAnyVal::LLVM_DECIMALVAL_NAME);
@@ -1303,7 +1368,7 @@ void LlvmCodeGen::AddFunctionToJit(llvm::Function* fn, void** fn_ptr) {
   AddFunctionToJitInternal(fn, fn_ptr);
 }
 
-void LlvmCodeGen::AddFunctionToJitInternal(llvm::Function* fn, void** fn_ptr) {
+void LlvmCodeGen::AddFunctionToJitInternal(llvm::Function* fn, CodegenFnPtrBase* fn_ptr) {
   DCHECK(!is_compiled_);
   fns_to_jit_compile_.push_back(make_pair(fn, fn_ptr));
 }
@@ -1312,19 +1377,15 @@ void LlvmCodeGen::CodegenDebugTrace(
     LlvmBuilder* builder, const char* str, llvm::Value* v1) {
   LOG(ERROR) << "Remove IR codegen debug traces before checking in.";
 
-  // Make a copy of str into memory owned by this object.  This is no guarantee that str is
-  // still around when the debug printf is executed.
-  debug_strings_.push_back(Substitute("LLVM Trace: $0", str));
-  str = debug_strings_.back().c_str();
+  // Call printf by embedding the string into the module and getting a pointer to it.
+  llvm::Value* const llvm_str =
+      GetStringConstant(builder, Substitute("LLVM Trace: $0", str));
 
   llvm::Function* printf = module_->getFunction("printf");
-  DCHECK(printf != NULL);
-
-  // Call printf by turning 'str' into a constant ptr value
-  llvm::Value* str_ptr = CastPtrToLlvmPtr(ptr_type_, const_cast<char*>(str));
+  DCHECK(printf != nullptr);
 
   vector<llvm::Value*> calling_args;
-  calling_args.push_back(str_ptr);
+  calling_args.push_back(llvm_str);
   if (v1 != NULL) calling_args.push_back(v1);
   builder->CreateCall(printf, calling_args);
 }
@@ -1436,10 +1497,17 @@ Status LlvmCodeGen::LoadIntrinsics() {
     llvm::Intrinsic::ID id;
     const char* error;
   } non_overloaded_intrinsics[] = {
+#ifdef __aarch64__
+      {llvm::Intrinsic::aarch64_crc32cb, "aarch64 crc32_u8"},
+      {llvm::Intrinsic::aarch64_crc32ch, "aarch64 crc32_u16"},
+      {llvm::Intrinsic::aarch64_crc32cw, "aarch64 crc32_u32"},
+      {llvm::Intrinsic::aarch64_crc32cx, "aarch64 crc32_u64"},
+#else
       {llvm::Intrinsic::x86_sse42_crc32_32_8, "sse4.2 crc32_u8"},
       {llvm::Intrinsic::x86_sse42_crc32_32_16, "sse4.2 crc32_u16"},
       {llvm::Intrinsic::x86_sse42_crc32_32_32, "sse4.2 crc32_u32"},
       {llvm::Intrinsic::x86_sse42_crc32_64_64, "sse4.2 crc32_u64"},
+#endif
   };
   const int num_intrinsics =
       sizeof(non_overloaded_intrinsics) / sizeof(non_overloaded_intrinsics[0]);
@@ -1550,7 +1618,7 @@ void LlvmCodeGen::ClearHashFns() {
 //   ret i32 %12
 // }
 llvm::Function* LlvmCodeGen::GetHashFunction(int num_bytes) {
-  if (IsCPUFeatureEnabled(CpuInfo::SSE4_2)) {
+  if (base::IsAarch64() || IsCPUFeatureEnabled(CpuInfo::SSE4_2)) {
     if (num_bytes == -1) {
       // -1 indicates variable length, just return the generic loop based
       // hash fn.
@@ -1575,25 +1643,39 @@ llvm::Function* LlvmCodeGen::GetHashFunction(int num_bytes) {
     llvm::Function* fn = prototype.GeneratePrototype(&builder, &args[0]);
     llvm::Value* data = args[0];
     llvm::Value* result = args[2];
-
+#ifdef __aarch64__
+    llvm::Function* crc8_fn = llvm_intrinsics_[llvm::Intrinsic::aarch64_crc32cb];
+    llvm::Function* crc16_fn = llvm_intrinsics_[llvm::Intrinsic::aarch64_crc32ch];
+    llvm::Function* crc32_fn = llvm_intrinsics_[llvm::Intrinsic::aarch64_crc32cw];
+    llvm::Function* crc64_fn = llvm_intrinsics_[llvm::Intrinsic::aarch64_crc32cx];
+#else
     llvm::Function* crc8_fn = llvm_intrinsics_[llvm::Intrinsic::x86_sse42_crc32_32_8];
     llvm::Function* crc16_fn = llvm_intrinsics_[llvm::Intrinsic::x86_sse42_crc32_32_16];
     llvm::Function* crc32_fn = llvm_intrinsics_[llvm::Intrinsic::x86_sse42_crc32_32_32];
     llvm::Function* crc64_fn = llvm_intrinsics_[llvm::Intrinsic::x86_sse42_crc32_64_64];
+#endif
 
     // Generate the crc instructions starting with the highest number of bytes
     if (num_bytes >= 8) {
+#ifndef __aarch64__
       llvm::Value* result_64 = builder.CreateZExt(result, i64_type());
+#endif
       llvm::Value* ptr = builder.CreateBitCast(data, i64_ptr_type());
       int i = 0;
       while (num_bytes >= 8) {
         llvm::Value* index[] = {GetI32Constant(i++)};
         llvm::Value* d = builder.CreateLoad(builder.CreateInBoundsGEP(ptr, index));
+#ifdef __aarch64__
+        result = builder.CreateCall(crc64_fn, llvm::ArrayRef<llvm::Value*>({result, d}));
+#else
         result_64 =
             builder.CreateCall(crc64_fn, llvm::ArrayRef<llvm::Value*>({result_64, d}));
+#endif
         num_bytes -= 8;
       }
+#ifndef __aarch64__
       result = builder.CreateTrunc(result_64, i32_type());
+#endif
       llvm::Value* index[] = {GetI32Constant(i * 8)};
       // Update data to past the 8-byte chunks
       data = builder.CreateInBoundsGEP(data, index);
@@ -1613,6 +1695,9 @@ llvm::Function* LlvmCodeGen::GetHashFunction(int num_bytes) {
       DCHECK_LT(num_bytes, 4);
       llvm::Value* ptr = builder.CreateBitCast(data, i16_ptr_type());
       llvm::Value* d = builder.CreateLoad(ptr);
+#ifdef __aarch64__
+      d = builder.CreateZExt(d, i32_type());
+#endif
       result = builder.CreateCall(crc16_fn, llvm::ArrayRef<llvm::Value*>({result, d}));
       llvm::Value* index[] = {GetI16Constant(2)};
       data = builder.CreateInBoundsGEP(data, index);
@@ -1622,6 +1707,9 @@ llvm::Function* LlvmCodeGen::GetHashFunction(int num_bytes) {
     if (num_bytes > 0) {
       DCHECK_EQ(num_bytes, 1);
       llvm::Value* d = builder.CreateLoad(data);
+#ifdef __aarch64__
+      d = builder.CreateZExt(d, i32_type());
+#endif
       result = builder.CreateCall(crc8_fn, llvm::ArrayRef<llvm::Value*>({result, d}));
       --num_bytes;
     }
