@@ -35,8 +35,9 @@ DEFINE_string(hostname, "", "Hostname to use for this daemon, also used as part
               "the Kerberos principal, if enabled. If not set, the system default will be"
               " used");
 
-DEFINE_int32(be_port, 22000,
-    "port on which thrift based ImpalaInternalService is exported");
+DEFINE_bool(use_resolved_hostname, false, "If true, --hostname is resolved before use, "
+              "so that the IP address will be used everywhere instead of the hostname.");
+
 DEFINE_int32(krpc_port, 27000,
     "port on which KRPC based ImpalaInternalService is exported");
 
@@ -65,7 +66,9 @@ DEFINE_bool(skip_external_kerberos_auth, false,
     "(Advanced) skip kerberos authentication for incoming external connections to "
     "this daemon, e.g. clients connecting to the HS2 interface. Only has an effect "
     "if --principal is set, i.e. Kerberos is enabled.");
-
+DEFINE_string(anonymous_user_name, "anonymous",
+    "Default username used when a client connects to an unsecured impala daemon and "
+    "does not specify a username.");
 
 static const string mem_limit_help_msg = "Limit on process memory consumption. "
     "Includes the JVM's memory consumption only if --mem_limit_includes_jvm is true. "
@@ -161,6 +164,9 @@ DEFINE_bool(thread_creation_fault_injection, false, "A fault injection option th
 DEFINE_int32(stress_catalog_init_delay_ms, 0, "A stress option that injects extra delay"
     " in milliseconds when initializing an impalad's local catalog replica. Delay <= 0"
     " inject no delay.");
+DEFINE_int32(stress_catalog_startup_delay_ms, 0, "A stress option that injects extra "
+    "delay in milliseconds during the startup of catalogd. The delay is before the "
+    "catalogd opens ports or accepts connections. Delay <= 0 injects no delay.");
 DEFINE_int32(stress_disk_read_delay_ms, 0, "A stress option that injects extra delay"
     " in milliseconds when the I/O manager is reading from disk.");
 #endif
@@ -192,6 +198,13 @@ DEFINE_int32(kudu_client_rpc_timeout_ms, default_kudu_client_rpc_timeout_ms,
     "kudu_operation_timeout_ms. This must be a positive value or it will be ignored and "
     "Kudu's default of 10s will be used. There is no way to disable timeouts.");
 
+// Timeout for connection negotiation between Kudu client in the BE and Kudu
+// servers, in milliseconds. For details on connection negotiation, see
+// https://github.com/apache/kudu/blob/master/docs/design-docs/rpc.md#negotiation
+DEFINE_int32(kudu_client_connection_negotiation_timeout_ms, 3000,
+    "(Advanced) Timeout for connection negotiation between Kudu client and "
+    "Kudu masters and tablet servers, in milliseconds");
+
 DEFINE_int64(inc_stats_size_limit_bytes, 200 * (1LL<<20), "Maximum size of "
     "incremental stats the catalog is allowed to serialize per table. "
     "This limit is set as a safety check, to prevent the JVM from "
@@ -263,6 +276,12 @@ DEFINE_int32(hms_event_polling_interval_s, 0,
     "feature and not recommended to be deployed on production systems until it is "
     "made generally available.");
 
+DEFINE_bool(enable_insert_events, true,
+    "Enables insert events in the events processor. When this configuration is set to "
+    "true Impala will generate INSERT event types which when received by other Impala "
+    "clusters can be used to automatically refresh the tables or partitions. Event "
+    "processing must be turned on for this flag to have any effect.");
+
 DEFINE_string(blacklisted_dbs, "sys,information_schema",
     "Comma separated list for blacklisted databases. Configure which databases to be "
     "skipped for loading (in startup and global INVALIDATE METADATA). Users can't access,"
@@ -280,18 +299,11 @@ DEFINE_double_hidden(invalidate_tables_fraction_on_memory_pressure, 0.1,
     "The fraction of tables to invalidate when CatalogdTableInvalidator considers the "
     "old GC generation to be almost full.");
 
-DEFINE_bool_hidden(unlock_mt_dop, false,
-    "(Experimental) If true, allow specifying mt_dop for all queries.");
-
-DEFINE_bool_hidden(mt_dop_auto_fallback, false,
-    "(Experimental) If true, fall back to non-mt_dop if mt_dop query option is set and "
-    "a query does not support it. Has no effect if --unlock_mt_dop is true.");
-
 DEFINE_bool_hidden(recursively_list_partitions, true,
     "If true, recursively list the content of partition directories.");
 
-DEFINE_bool(unlock_zorder_sort, false,
-    "(Experimental) If true, enables using ZORDER option for SORT BY.");
+DEFINE_bool(unlock_zorder_sort, true,
+    "If true, enables using ZORDER option for SORT BY.");
 
 DEFINE_string(min_privilege_set_for_show_stmts, "any",
     "Comma separated list of privileges. Any one of them is required to show a database "
@@ -326,8 +338,19 @@ DEFINE_bool_hidden(use_customized_user_groups_mapper_for_ranger, false,
     "If true, use the customized user-to-groups mapper when performing authorization via"
     " Ranger.");
 
-DEFINE_bool(enable_column_masking, true,
-    "If false, disable the column masking feature. Defaults to be true.");
+DEFINE_bool(enable_incremental_metadata_updates, true,
+    "If true, Catalog Server will send incremental table updates in partition level in "
+    "the statestore topic updates. Legacy coordinators will apply the partition updates "
+    "incrementally, i.e. reuse unchanged partition metadata. Disable this feature by "
+    "setting this to false in the Catalog Server. Then metadata of each table will be "
+    "propagated as a whole object in the statestore topic updates. Note that legacy "
+    "coordinators can apply incremental or full table updates so don't need this flag.");
+
+DEFINE_bool(enable_legacy_avx_support, false,
+    "If true, Impala relaxes its x86_64 CPU feature requirement to allow running on "
+    "machines with AVX but no AVX2. This allows running Impala on older machines "
+    "without AVX2 support. This is a legacy mode that will be removed in a "
+    "future release.");
 
 // ++========================++
 // || Startup flag graveyard ||
@@ -358,6 +381,8 @@ DEFINE_bool(enable_column_masking, true,
 REMOVED_FLAG(abfs_read_chunk_size);
 REMOVED_FLAG(adls_read_chunk_size);
 REMOVED_FLAG(authorization_policy_file);
+REMOVED_FLAG(authorization_policy_provider_class);
+REMOVED_FLAG(be_port);
 REMOVED_FLAG(be_service_threads);
 REMOVED_FLAG(cgroup_hierarchy_path);
 REMOVED_FLAG(coordinator_rpc_threads);
@@ -369,6 +394,7 @@ REMOVED_FLAG(enable_partitioned_hash_join);
 REMOVED_FLAG(enable_phj_probe_side_filtering);
 REMOVED_FLAG(enable_rm);
 REMOVED_FLAG(kerberos_reinit_interval);
+REMOVED_FLAG(ldap_manual_config);
 REMOVED_FLAG(llama_addresses);
 REMOVED_FLAG(llama_callback_port);
 REMOVED_FLAG(llama_host);
@@ -378,6 +404,7 @@ REMOVED_FLAG(llama_registration_timeout_secs);
 REMOVED_FLAG(llama_registration_wait_secs);
 REMOVED_FLAG(local_nodemanager_url);
 REMOVED_FLAG(max_free_io_buffers);
+REMOVED_FLAG(mt_dop_auto_fallback);
 REMOVED_FLAG(pull_incremental_statistics);
 REMOVED_FLAG(report_status_retry_interval_ms);
 REMOVED_FLAG(resource_broker_cnxn_attempts);
@@ -389,11 +416,14 @@ REMOVED_FLAG(rm_default_cpu_vcores);
 REMOVED_FLAG(rm_default_memory);
 REMOVED_FLAG(rpc_cnxn_attempts);
 REMOVED_FLAG(rpc_cnxn_retry_interval_ms);
+REMOVED_FLAG(sentry_catalog_polling_frequency_s);
+REMOVED_FLAG(sentry_config);
 REMOVED_FLAG(skip_lzo_version_check);
 REMOVED_FLAG(staging_cgroup);
 REMOVED_FLAG(status_report_interval);
 REMOVED_FLAG(status_report_max_retries);
 REMOVED_FLAG(suppress_unknown_disk_id_warnings);
+REMOVED_FLAG(unlock_mt_dop);
 REMOVED_FLAG(use_krpc);
 REMOVED_FLAG(use_kudu_kinit);
 REMOVED_FLAG(use_statestore);
