@@ -27,6 +27,7 @@
 #include "common/status.h"
 #include "exec/kudu-util.h"
 #include "exprs/scalar-expr-evaluator.h"
+#include "exprs/timezone_db.h"
 #include "gutil/atomicops.h"
 #include "gutil/strings/substitute.h"
 #include "rpc/authentication.h"
@@ -42,7 +43,6 @@
 #include "util/cgroup-util.h"
 #include "util/cpu-info.h"
 #include "util/debug-util.h"
-#include "util/decimal-util.h"
 #include "util/disk-info.h"
 #include "util/jni-util.h"
 #include "util/logging-support.h"
@@ -67,18 +67,16 @@ using namespace impala;
 DECLARE_bool(enable_process_lifetime_heap_profiling);
 DECLARE_string(heap_profile_dir);
 DECLARE_string(hostname);
+DECLARE_bool(use_resolved_hostname);
 // TODO: rename this to be more generic when we have a good CM release to do so.
 DECLARE_int32(logbufsecs);
 DECLARE_int32(max_log_files);
 DECLARE_int32(max_minidumps);
 DECLARE_string(redaction_rules_file);
+DECLARE_bool(redirect_stdout_stderr);
 DECLARE_string(reserved_words_version);
 DECLARE_bool(symbolize_stacktrace);
 
-DEFINE_int32(max_audit_event_log_files, 0, "Maximum number of audit event log files "
-    "to retain. The most recent audit event log files are retained. If set to 0, "
-    "all audit event log files are retained.");
-
 DEFINE_int32(memory_maintenance_sleep_time_ms, 10000, "Sleep time in milliseconds "
     "between memory maintenance iterations");
 
@@ -147,8 +145,6 @@ extern "C" { void __gcov_flush(); }
     if (impala::TestInfo::is_test()) continue;
     // Check for log rotation in every interval of the maintenance thread
     impala::CheckAndRotateLogFiles(FLAGS_max_log_files);
-    // Check for audit event log rotation in every interval of the maintenance thread
-    impala::CheckAndRotateAuditEventLogFiles(FLAGS_max_audit_event_log_files);
     // Check for minidump rotation in every interval of the maintenance thread. This is
     // necessary since an arbitrary number of minidumps can be written by sending SIGUSR1
     // to the process.
@@ -195,8 +191,7 @@ extern "C" { void __gcov_flush(); }
     CHECK(err == 0) << "sigwait(): " << GetStrErrMsg(err) << ": " << err;
     CHECK_EQ(IMPALA_SHUTDOWN_SIGNAL, signal);
     ShutdownStatusPB shutdown_status;
-    const int ONE_YEAR_IN_SECONDS = 365 * 24 * 60 * 60;
-    Status status = impala_server->StartShutdown(ONE_YEAR_IN_SECONDS, &shutdown_status);
+    Status status = impala_server->StartShutdown(-1, &shutdown_status);
     if (!status.ok()) {
       LOG(ERROR) << "Shutdown signal received but unable to initiate shutdown. Status: "
                  << status.GetDetail();
@@ -258,7 +253,7 @@ void BlockImpalaShutdownSignal() {
 }
 
 void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
-    TestInfo::Mode test_mode) {
+    TestInfo::Mode test_mode, bool external_fe) {
   srand(time(NULL));
   BlockImpalaShutdownSignal();
 
@@ -268,10 +263,6 @@ void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
   OsInfo::Init();
   TestInfo::Init(test_mode);
 
-  // Verify CPU meets the minimum requirements before calling InitGoogleLoggingSafe()
-  // which might use SSSE3 instructions (see IMPALA-160).
-  CpuInfo::VerifyCpuRequirements();
-
   // Set the default hostname. The user can override this with the hostname flag.
   ABORT_IF_ERROR(GetHostname(&FLAGS_hostname));
 
@@ -281,6 +272,16 @@ void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
 # else
   FLAGS_symbolize_stacktrace = true;
 #endif
+
+  if (external_fe) {
+    // Change defaults for flags when loaded as part of external frontend.
+    // Write logs to stderr by default (otherwise logs get written to
+    // FeSupport.INFO/ERROR).
+    FLAGS_logtostderr = true;
+    // Do not redirct stdout/stderr by default.
+    FLAGS_redirect_stdout_stderr = false;
+  }
+
   google::SetVersionString(impala::GetBuildVersion());
   google::ParseCommandLineFlags(&argc, &argv, true);
   if (!FLAGS_redaction_rules_file.empty()) {
@@ -304,9 +305,13 @@ void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
   }
   impala::InitGoogleLoggingSafe(argv[0]);
   // Breakpad needs flags and logging to initialize.
-  ABORT_IF_ERROR(RegisterMinidump(argv[0]));
+  if (!external_fe) {
+    ABORT_IF_ERROR(RegisterMinidump(argv[0]));
+  }
 #ifndef THREAD_SANITIZER
+#ifndef __aarch64__
   AtomicOps_x86CPUFeaturesInit();
+#endif
 #endif
   impala::InitThreading();
   impala::datetime_parse_util::SimpleDateFormatTokenizer::InitCtx();
@@ -355,15 +360,35 @@ void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
   LOG(INFO) << "Default AES cipher mode for spill-to-disk: "
             << EncryptionKey::ModeToString(EncryptionKey::GetSupportedDefaultMode());
 
+  // After initializing logging and printing the machine information, verify the
+  // minimal CPU requirements and exit if they are not met.
+  Status cpu_requirement_status = CpuInfo::EnforceCpuRequirements();
+  if (!cpu_requirement_status.ok()) {
+    CLEAN_EXIT_WITH_ERROR(cpu_requirement_status.GetDetail());
+  }
+
+  if (FLAGS_use_resolved_hostname) {
+    IpAddr ip_address;
+    Status status = HostnameToIpAddr(FLAGS_hostname, &ip_address);
+    if (!status.ok()) CLEAN_EXIT_WITH_ERROR(status.GetDetail());
+    LOG(INFO) << Substitute("Resolved hostname $0 to $1", FLAGS_hostname, ip_address);
+    FLAGS_hostname = ip_address;
+  }
+
   // Required for the FE's Catalog
-  ABORT_IF_ERROR(impala::LibCache::Init());
+  ABORT_IF_ERROR(impala::LibCache::Init(external_fe));
   Status fs_cache_init_status = impala::HdfsFsCache::Init();
   if (!fs_cache_init_status.ok()) CLEAN_EXIT_WITH_ERROR(fs_cache_init_status.GetDetail());
 
   if (init_jvm) {
+    if (!external_fe) {
+      JniUtil::InitLibhdfs();
+    }
     ABORT_IF_ERROR(JniUtil::Init());
     InitJvmLoggingSupport();
-    ABORT_IF_ERROR(JniUtil::InitJvmPauseMonitor());
+    if (!external_fe) {
+      ABORT_IF_ERROR(JniUtil::InitJvmPauseMonitor());
+    }
     ZipUtil::InitJvm();
   }
 
@@ -394,6 +419,12 @@ void impala::InitCommonRuntime(int argc, char** argv, bool init_jvm,
     error_msg << "Failed to register action for SIGTERM: " << GetStrErrMsg();
     CLEAN_EXIT_WITH_ERROR(error_msg.str());
   }
+
+  if (external_fe) {
+    // Explicitly load the timezone database for external FEs. Impala daemons load it
+    // through ImpaladMain
+    ABORT_IF_ERROR(TimezoneDatabase::Initialize());
+  }
 }
 
 Status impala::StartMemoryMaintenanceThread() {
@@ -431,7 +462,7 @@ extern "C" const char* __tsan_default_options() {
 #else
          "1 "
 #endif
-         "halt_on_error=0 history_size=7 allocator_may_return_null=1 "
+         "halt_on_error=1 history_size=7 allocator_may_return_null=1 "
          "suppressions=" THREAD_SANITIZER_SUPPRESSIONS;
 }
 #endif
