@@ -22,6 +22,7 @@
 #include <memory>
 #include <vector>
 
+#include "codegen/codegen-fn-ptr.h"
 #include "exec/aggregator.h"
 #include "exec/hash-table.h"
 #include "runtime/buffered-tuple-stream.h"
@@ -36,6 +37,7 @@ class AggFnEvaluator;
 class GroupingAggregator;
 class PlanNode;
 class LlvmCodeGen;
+class QueryState;
 class RowBatch;
 class RuntimeState;
 struct ScalarExprsResultsRowLayout;
@@ -119,11 +121,11 @@ class Tuple;
 class GroupingAggregatorConfig : public AggregatorConfig {
  public:
   GroupingAggregatorConfig(
-      const TAggregator& taggregator, RuntimeState* state, PlanNode* pnode, int agg_idx);
+      const TAggregator& taggregator, FragmentState* state, PlanNode* pnode, int agg_idx);
   Status Init(
-      const TAggregator& taggregator, RuntimeState* state, PlanNode* pnode) override;
+      const TAggregator& taggregator, FragmentState* state, PlanNode* pnode) override;
   void Close() override;
-  Status Codegen(RuntimeState* state) override;
+  void Codegen(FragmentState* state) override;
   ~GroupingAggregatorConfig() override {}
 
   /// Row with the intermediate tuple as its only tuple.
@@ -160,14 +162,14 @@ class GroupingAggregatorConfig : public AggregatorConfig {
   const HashTableConfig* hash_table_config_;
 
   typedef Status (*AddBatchImplFn)(
-      GroupingAggregator*, RowBatch*, TPrefetchMode::type, HashTableCtx*);
+      GroupingAggregator*, RowBatch*, TPrefetchMode::type, HashTableCtx*, bool);
   /// Jitted AddBatchImpl function pointer. Null if codegen is disabled.
-  AddBatchImplFn add_batch_impl_fn_ = nullptr;
+  CodegenFnPtr<AddBatchImplFn> add_batch_impl_fn_;
 
   typedef Status (*AddBatchStreamingImplFn)(GroupingAggregator*, int, bool,
       TPrefetchMode::type, RowBatch*, RowBatch*, HashTableCtx*, int[]);
   /// Jitted AddBatchStreamingImpl function pointer. Null if codegen is disabled.
-  AddBatchStreamingImplFn add_batch_streaming_impl_fn_ = nullptr;
+  CodegenFnPtr<AddBatchStreamingImplFn> add_batch_streaming_impl_fn_;
 
   int GetNumGroupingExprs() const override { return grouping_exprs_.size(); }
 
@@ -193,7 +195,6 @@ class GroupingAggregator : public Aggregator {
       const GroupingAggregatorConfig& config, int64_t estimated_input_cardinality);
 
   virtual Status Prepare(RuntimeState* state) override;
-  virtual void Codegen(RuntimeState* state) override;
   virtual Status Open(RuntimeState* state) override;
   virtual Status GetNext(RuntimeState* state, RowBatch* row_batch, bool* eos) override;
   virtual Status Reset(RuntimeState* state, RowBatch* row_batch) override;
@@ -214,10 +215,6 @@ class GroupingAggregator : public Aggregator {
  private:
   struct Partition;
 
-  /// TODO: Remove reference once codegen is performed before FIS creation.
-  /// Reference to the config object and only used to call Codegen().
-  const GroupingAggregatorConfig& agg_config_;
-
   /// Reference to the hash table config which is a part of the GroupingAggregatorConfig
   /// that was used to create this object. Its used to create an instance of the
   /// HashTableCtx in Prepare(). Not Owned.
@@ -282,6 +279,15 @@ class GroupingAggregator : public Aggregator {
 
   /// Allocator for hash table memory.
   std::unique_ptr<Suballocator> ht_allocator_;
+  /// Saved reservation for writing a large page to a spilled stream or writing the last
+  /// large row to a pinned partition when building/repartitioning a spilled partition.
+  /// ('max_page_len' - 'default_page_len') reservation is saved when claiming the initial
+  /// min reservation.
+  BufferPool::SubReservation large_write_page_reservation_;
+  /// Saved reservation for reading a large page from a spilled stream.
+  /// ('max_page_len' - 'default_page_len') reservation is saved when claiming the initial
+  /// min reservation.
+  BufferPool::SubReservation large_read_page_reservation_;
 
   /// MemPool used to allocate memory during Close() when creating new output tuples. The
   /// pool should not be Reset() to allow amortizing memory allocation over a series of
@@ -310,15 +316,12 @@ class GroupingAggregator : public Aggregator {
   int64_t limit_; // -1: no limit
   bool ReachedLimit() { return limit_ != -1 && num_rows_returned_ >= limit_; }
 
-  typedef Status (*AddBatchImplFn)(
-      GroupingAggregator*, RowBatch*, TPrefetchMode::type, HashTableCtx*);
   /// Jitted AddBatchImpl function pointer. Null if codegen is disabled.
-  const AddBatchImplFn& add_batch_impl_fn_;
+  const CodegenFnPtr<GroupingAggregatorConfig::AddBatchImplFn>& add_batch_impl_fn_;
 
-  typedef Status (*AddBatchStreamingImplFn)(GroupingAggregator*, int, bool,
-      TPrefetchMode::type, RowBatch*, RowBatch*, HashTableCtx*, int[PARTITION_FANOUT]);
-  /// Jitted AddBatchStreamingImpl function pointer.  Null if codegen is disabled.
-  const AddBatchStreamingImplFn& add_batch_streaming_impl_fn_ ;
+  /// Jitted AddBatchStreamingImpl function pointer. Null if codegen is disabled.
+  const CodegenFnPtr<GroupingAggregatorConfig::AddBatchStreamingImplFn>&
+      add_batch_streaming_impl_fn_;
 
   /// Total time spent resizing hash tables.
   RuntimeProfile::Counter* ht_resize_timer_ = nullptr;
@@ -427,8 +430,7 @@ class GroupingAggregator : public Aggregator {
 
     /// Initializes the hash table. 'aggregated_row_stream' must be non-NULL.
     /// Sets 'got_memory' to true if the hash table was initialised or false on OOM.
-    /// After returning, 'hash_tbl' will be non-null iff 'got_memory' is true and the
-    /// returned status is OK.
+    /// After returning, 'hash_tbl' will be non-null iff the returned status is OK.
     Status InitHashTable(bool* got_memory) WARN_UNUSED_RESULT;
 
     /// Called in case we need to serialize aggregated rows. This step effectively does
@@ -448,6 +450,8 @@ class GroupingAggregator : public Aggregator {
 
     bool is_spilled() const { return hash_tbl.get() == nullptr; }
 
+    std::string DebugString() const;
+
     GroupingAggregator* parent;
 
     /// If true, this partition is closed and there is nothing left to do.
@@ -534,12 +538,15 @@ class GroupingAggregator : public Aggregator {
   /// 'prefetch_mode' specifies the prefetching mode in use. If it's not PREFETCH_NONE,
   ///     hash table buckets will be prefetched based on the hash values computed. Note
   ///     that 'prefetch_mode' will be substituted with constants during codegen time.
-  //
+  /// 'has_more_rows' is used in building/repartitioning a spilled partition, to indicate
+  ///     whether there are more rows after the given 'batch' in the input. We may restore
+  ///     the 'large_write_page_reservation_' when adding the last row.
+  ///
   /// This function is replaced by codegen. We pass in ht_ctx_.get() as an argument for
   /// performance.
   template <bool AGGREGATED_ROWS>
   Status IR_ALWAYS_INLINE AddBatchImpl(RowBatch* batch, TPrefetchMode::type prefetch_mode,
-      HashTableCtx* ht_ctx) WARN_UNUSED_RESULT;
+      HashTableCtx* ht_ctx, bool has_more_rows) WARN_UNUSED_RESULT;
 
   /// Evaluates the rows in 'batch' starting at 'start_row_idx' and stores the results in
   /// the expression values cache in 'ht_ctx'. The number of rows evaluated depends on
@@ -555,7 +562,7 @@ class GroupingAggregator : public Aggregator {
   /// May spill partitions if not enough memory is available.
   template <bool AGGREGATED_ROWS>
   Status IR_ALWAYS_INLINE ProcessRow(
-      TupleRow* row, HashTableCtx* ht_ctx) WARN_UNUSED_RESULT;
+      TupleRow* row, HashTableCtx* ht_ctx, bool has_more_rows) WARN_UNUSED_RESULT;
 
   /// Create a new intermediate tuple in partition, initialized with row. ht_ctx is
   /// the context for the partition's hash table and hash is the precomputed hash of
@@ -566,7 +573,8 @@ class GroupingAggregator : public Aggregator {
   /// for insertion returned from HashTable::FindBuildRowBucket().
   template <bool AGGREGATED_ROWS>
   Status IR_ALWAYS_INLINE AddIntermediateTuple(Partition* partition, TupleRow* row,
-      uint32_t hash, HashTable::Iterator insert_it) WARN_UNUSED_RESULT;
+      uint32_t hash, HashTable::Iterator insert_it, bool has_more_rows)
+      WARN_UNUSED_RESULT;
 
   /// Append a row to a spilled partition. The row may be aggregated or unaggregated
   /// according to AGGREGATED_ROWS. May spill partitions if needed to append the row
@@ -577,7 +585,8 @@ class GroupingAggregator : public Aggregator {
 
   /// Reads all the rows from input_stream and process them by calling AddBatchImpl().
   template <bool AGGREGATED_ROWS>
-  Status ProcessStream(BufferedTupleStream* input_stream) WARN_UNUSED_RESULT;
+  Status ProcessStream(BufferedTupleStream* input_stream, bool has_more_streams)
+      WARN_UNUSED_RESULT;
 
   /// Get rows for the next rowbatch from the next partition. Sets 'partition_eos_' to
   /// true if all rows from all partitions have been returned or the limit is reached.
@@ -689,6 +698,39 @@ class GroupingAggregator : public Aggregator {
   /// as the partitions aggregate stream needs to be serialized and rewritten.
   /// We do not spill streaming preaggregations, so we do not need to reserve any buffers.
   int64_t MinReservation() const;
+
+  /// Try to save the extra reservation ('max_row_buffer_size' - 'spillable_buffer_size')
+  /// for a large write page to 'large_write_page_reservation_'. Do nothing if there are
+  /// not enough unused reservation. Return true if succeeds.
+  bool TrySaveLargeWritePageReservation();
+
+  /// Similar to above but for the large read page.
+  bool TrySaveLargeReadPageReservation();
+
+  /// Same as TrySaveLargeWritePageReservation() but make sure it succeeds.
+  void SaveLargeWritePageReservation();
+
+  /// Same as TrySaveLargeReadPageReservation() but make sure it succeeds.
+  void SaveLargeReadPageReservation();
+
+  /// Restore the extra reservation we saved in 'large_write_page_reservation_' for a
+  /// large write page. 'large_write_page_reservation_' must not be used.
+  void RestoreLargeWritePageReservation();
+
+  /// Similar to above but for the large read page.
+  void RestoreLargeReadPageReservation();
+
+  /// A wrapper of 'stream->AddRow()' to add 'row' to a spilled 'stream'. When it fails to
+  /// add a large row due to run out of unused reservation and fails to increase the
+  /// reservation, retry it after restoring the large write page reservation when we don't
+  /// need to save this reservation for spilling partitions. If succeeds, returns true and
+  /// save back the large write page reservation. Otherwise, returns false with a non-ok
+  /// status.
+  bool AddRowToSpilledStream(BufferedTupleStream* stream, TupleRow* __restrict__ row,
+      Status* status);
+
+  /// Gets current number of pinned partitions.
+  int GetNumPinnedPartitions();
 };
 } // namespace impala
 
