@@ -26,6 +26,7 @@
 #include "exec/hdfs-scan-node.h"
 #include "exec/read-write-util.h"
 #include "exec/scanner-context.inline.h"
+#include "runtime/fragment-state.h"
 #include "runtime/raw-value.h"
 #include "runtime/runtime-state.h"
 #include "util/codec.h"
@@ -79,7 +80,7 @@ Status HdfsAvroScanner::Open(ScannerContext* context) {
 }
 
 Status HdfsAvroScanner::Codegen(HdfsScanPlanNode* node,
-   RuntimeState* state, llvm::Function** decode_avro_data_fn) {
+   FragmentState* state, llvm::Function** decode_avro_data_fn) {
   *decode_avro_data_fn = nullptr;
   DCHECK(state->ShouldCodegen());
   DCHECK(state->codegen() != nullptr);
@@ -480,8 +481,7 @@ Status HdfsAvroScanner::InitNewRange() {
   }
 
   if (avro_header_->use_codegend_decode_avro_data) {
-    codegend_decode_avro_data_ = reinterpret_cast<DecodeAvroDataFn>(
-        scan_node_->GetCodegenFn(THdfsFileFormat::AVRO));
+    codegend_decode_avro_data_ = scan_node_->GetCodegenFn(THdfsFileFormat::AVRO);
   }
   if (codegend_decode_avro_data_ == nullptr) {
     scan_node_->IncNumScannersCodegenDisabled();
@@ -546,18 +546,18 @@ Status HdfsAvroScanner::ProcessRange(RowBatch* row_batch) {
       Tuple* tuple = tuple_;
       TupleRow* tuple_row = row_batch->GetRow(row_batch->AddRow());
       int max_tuples = row_batch->capacity() - row_batch->num_rows();
+      DCHECK_GT(num_records_in_block_, record_pos_);
       max_tuples = min<int64_t>(max_tuples, num_records_in_block_ - record_pos_);
       int num_to_commit;
       if (scan_node_->materialized_slots().empty()) {
         // No slots to materialize (e.g. count(*)), no need to decode data
         num_to_commit = WriteTemplateTuples(tuple_row, max_tuples);
-      } else if (codegend_decode_avro_data_ != nullptr) {
-        num_to_commit = codegend_decode_avro_data_(this, max_tuples,
-            row_batch->tuple_data_pool(), &data_block_, data_block_end_, tuple, tuple_row);
       } else {
-        num_to_commit = DecodeAvroData(max_tuples, row_batch->tuple_data_pool(),
-            &data_block_, data_block_end_, tuple, tuple_row);
+        num_to_commit = DecodeAvroDataCodegenOrInterpret(max_tuples,
+            row_batch->tuple_data_pool(), &data_block_,
+            data_block_end_, tuple, tuple_row);
       }
+
       RETURN_IF_ERROR(parse_status_);
       RETURN_IF_ERROR(CommitRows(num_to_commit, row_batch));
       record_pos_ += max_tuples;
@@ -1098,7 +1098,7 @@ Status HdfsAvroScanner::CodegenReadScalar(const AvroSchemaElement& element,
 }
 
 Status HdfsAvroScanner::CodegenDecodeAvroData(const HdfsScanPlanNode* node,
-    RuntimeState* state, llvm::Function** decode_avro_data_fn) {
+    FragmentState* state, llvm::Function** decode_avro_data_fn) {
   const vector<ScalarExpr*>& conjuncts = node->conjuncts_;
   LlvmCodeGen* codegen = state->codegen();
 
@@ -1140,3 +1140,10 @@ Status HdfsAvroScanner::CodegenDecodeAvroData(const HdfsScanPlanNode* node,
   }
   return Status::OK();
 }
+
+int HdfsAvroScanner::DecodeAvroDataCodegenOrInterpret(int max_tuples, MemPool* pool,
+    uint8_t** data, uint8_t* data_end, Tuple* tuple, TupleRow* tuple_row) {
+  return CallCodegendOrInterpreted<DecodeAvroDataFn>::invoke(this,
+      codegend_decode_avro_data_, &HdfsAvroScanner::DecodeAvroData, max_tuples, pool,
+      &data_block_, data_block_end_, tuple, tuple_row);
+}
