@@ -19,12 +19,15 @@
 
 #include <queue>
 
+#include "exec/exec-node.inline.h"
 #include "exec/orc-column-readers.h"
 #include "exec/scanner-context.inline.h"
+#include "exec/scratch-tuple-batch.h"
 #include "exprs/expr.h"
 #include "runtime/collection-value-builder.h"
 #include "runtime/exec-env.h"
 #include "runtime/io/request-context.h"
+#include "runtime/mem-tracker.h"
 #include "runtime/runtime-filter.inline.h"
 #include "runtime/timestamp-value.inline.h"
 #include "runtime/tuple-row.h"
@@ -113,8 +116,7 @@ void HdfsOrcScanner::ScanRangeInputStream::read(void* buf, uint64_t length,
   int cache_options = split_range->cache_options() & ~BufferOpts::USE_HDFS_CACHE;
   ScanRange* range = scanner_->scan_node_->AllocateScanRange(
       metadata_range->fs(), scanner_->filename(), length, offset, partition_id,
-      split_range->disk_id(), expected_local, split_range->is_erasure_coded(),
-      split_range->mtime(),
+      split_range->disk_id(), expected_local, split_range->mtime(),
       BufferOpts::ReadInto(reinterpret_cast<uint8_t*>(buf), length, cache_options));
 
   unique_ptr<BufferDescriptor> io_buffer;
@@ -155,8 +157,7 @@ Status HdfsOrcScanner::Open(ScannerContext* context) {
   process_footer_timer_stats_ =
       ADD_SUMMARY_STATS_TIMER(scan_node_->runtime_profile(), "OrcFooterProcessingTime");
 
-  codegend_process_scratch_batch_fn_ = reinterpret_cast<ProcessScratchBatchFn>(
-      scan_node_->GetCodegenFn(THdfsFileFormat::ORC));
+  codegend_process_scratch_batch_fn_ = scan_node_->GetCodegenFn(THdfsFileFormat::ORC);
   if (codegend_process_scratch_batch_fn_ == nullptr) {
     scan_node_->IncNumScannersCodegenDisabled();
   } else {
@@ -187,8 +188,49 @@ Status HdfsOrcScanner::Open(ScannerContext* context) {
   context_->ReleaseCompletedResources(true);
   RETURN_IF_ERROR(footer_status);
 
+  bool is_table_full_acid = scan_node_->hdfs_table()->IsTableFullAcid();
   schema_resolver_.reset(new OrcSchemaResolver(*scan_node_->hdfs_table(),
-      &reader_->getType(), filename()));
+      &reader_->getType(), filename(), is_table_full_acid));
+  bool is_file_full_acid = schema_resolver_->HasFullAcidV2Schema();
+  acid_original_file_ = is_table_full_acid && !is_file_full_acid;
+  if (is_table_full_acid) {
+    acid_write_id_range_ = valid_write_ids_.GetWriteIdRange(filename());
+    if (acid_original_file_ &&
+        acid_write_id_range_.first != acid_write_id_range_.second) {
+      return Status(Substitute("Found non-ACID file in directory that can only contain "
+          "files with full ACID schema: $0", filename()));
+    }
+  }
+  if (acid_original_file_) {
+    int32_t filename_len = strlen(filename());
+    if (filename_len >= 2 && strcmp(filename() + filename_len - 2, "_0") != 0) {
+      // It's an original file that should be included in the result.
+      // If it doesn't end with "_0" it means that it belongs to a bucket with other
+      // files. Impala rejects such files and tables.
+      // These files should only exist at table/partition root directory level.
+      // Original files in delta directories are created via the LOAD DATA
+      // statement. LOAD DATA assigns virtual bucket ids to files in non-bucketed
+      // tables, so we will have one file per (virtual) bucket (all of them having "_0"
+      // ending). For bucketed tables LOAD DATA will write ACID files. So after the first
+      // major compaction the table should never get into this state ever again.
+      return Status(Substitute("Found original file with unexpected name: $0 "
+          "Please run a major compaction on the partition/table to overcome this.",
+          filename()));
+    }
+  }
+
+  // Hive Streaming Ingestion allocates multiple write ids, hence create delta directories
+  // like delta_5_10. Then it continuously appends new stripes (and footers) to the
+  // ORC files of the delte dir. So it's possible that the file has rows that Impala is
+  // not allowed to see based on its valid write id list. In such cases we need to
+  // validate the write ids of the row batches.
+  if (is_table_full_acid && !ValidWriteIdList::IsCompacted(filename())) {
+    valid_write_ids_.InitFrom(scan_node_->hdfs_table()->ValidWriteIdList());
+    ValidWriteIdList::RangeResponse rows_valid = valid_write_ids_.IsWriteIdRangeValid(
+        acid_write_id_range_.first, acid_write_id_range_.second);
+    DCHECK_NE(rows_valid, ValidWriteIdList::NONE);
+    row_batches_need_validation_ = rows_valid == ValidWriteIdList::SOME;
+  }
 
   // Update 'row_reader_options_' based on the tuple descriptor so the ORC lib can skip
   // columns we don't need.
@@ -198,12 +240,6 @@ Status HdfsOrcScanner::Open(ScannerContext* context) {
   // blob more efficiently.
   row_reader_options_.setEnableLazyDecoding(true);
 
-  // Build 'col_id_path_map_' that maps from ORC column ids to their corresponding
-  // SchemaPath in the table. The map is used in the constructors of OrcColumnReaders
-  // where we resolve SchemaPaths of the descriptors.
-  RETURN_IF_ERROR(schema_resolver_->BuildSchemaPaths(scan_node_->num_partition_keys(),
-      &col_id_path_map_));
-
   // To create OrcColumnReaders, we need the selected orc schema. It's a subset of the
   // file schema: a tree of selected orc types and can only be got from an orc::RowReader
   // (by orc::RowReader::getSelectedType).
@@ -340,6 +376,7 @@ Status HdfsOrcScanner::ResolveColumns(const TupleDescriptor& tuple_desc,
     return Status(Substitute("Could not find nested column '$0' in file '$1'.",
         PrintPath(*scan_node_->hdfs_table(), tuple_desc.tuple_path()), filename()));
   }
+  tuple_to_col_id_.insert({&tuple_desc, node->getColumnId()});
   if (tuple_desc.byte_size() == 0) {
     // Don't need to materialize any slots but just generate an empty tuple for each
     // (collection) row. (E.g. count(*) or 'exists' on results of subquery).
@@ -385,10 +422,15 @@ Status HdfsOrcScanner::ResolveColumns(const TupleDescriptor& tuple_desc,
         *template_tuple =
             Tuple::Create(tuple_desc.byte_size(), template_tuple_pool_.get());
       }
-      (*template_tuple)->SetNull(slot_desc->null_indicator_offset());
+      if (acid_original_file_ && schema_resolver_->IsAcidColumn(slot_desc->col_path())) {
+        SetSyntheticAcidFieldForOriginalFile(slot_desc, *template_tuple);
+      } else {
+        (*template_tuple)->SetNull(slot_desc->null_indicator_offset());
+      }
       missing_field_slots_.insert(slot_desc);
       continue;
     }
+    slot_to_col_id_.insert({slot_desc, node->getColumnId()});
     if (pos_field) {
       DCHECK(pos_slot_desc == nullptr)
           << "There should only be one position slot per tuple";
@@ -414,6 +456,31 @@ Status HdfsOrcScanner::ResolveColumns(const TupleDescriptor& tuple_desc,
   return Status::OK();
 }
 
+void HdfsOrcScanner::SetSyntheticAcidFieldForOriginalFile(const SlotDescriptor* slot_desc,
+    Tuple* template_tuple) {
+  DCHECK_EQ(1, slot_desc->col_path().size());
+  int field_idx = slot_desc->col_path().front() - scan_node_->num_partition_keys();
+  switch(field_idx) {
+    case ACID_FIELD_OPERATION_INDEX:
+      *template_tuple->GetIntSlot(slot_desc->tuple_offset()) = 0;
+      break;
+    case ACID_FIELD_ORIGINAL_TRANSACTION_INDEX:
+    case ACID_FIELD_CURRENT_TRANSACTION_INDEX:
+      DCHECK_EQ(acid_write_id_range_.first, acid_write_id_range_.second);
+      *template_tuple->GetBigIntSlot(slot_desc->tuple_offset()) =
+          acid_write_id_range_.first;
+      break;
+    case ACID_FIELD_BUCKET_INDEX:
+      *template_tuple->GetBigIntSlot(slot_desc->tuple_offset()) =
+          ValidWriteIdList::GetBucketProperty(filename());
+      break;
+    case ACID_FIELD_ROWID_INDEX:
+      acid_synthetic_rowid_ = slot_desc;
+    default:
+      break;
+  }
+}
+
 /// Whether 'selected_type_ids' contains the id of any children of 'node'
 bool HasChildrenSelected(const orc::Type& node,
     const list<uint64_t>& selected_type_ids) {
@@ -430,6 +497,7 @@ Status HdfsOrcScanner::SelectColumns(const TupleDescriptor& tuple_desc) {
   RETURN_IF_ERROR(ResolveColumns(tuple_desc, &selected_nodes, &pos_slots));
 
   for (auto t : selected_nodes) selected_type_ids_.push_back(t->getColumnId());
+
   // Select columns for array positions. Due to ORC-450 we can't materialize array
   // offsets without materializing its items, so we should still select the item or any
   // sub column of the item. To be simple, we choose the max column id in the subtree
@@ -454,6 +522,17 @@ Status HdfsOrcScanner::SelectColumns(const TupleDescriptor& tuple_desc) {
     selected_nodes.push_back(array_node);
   }
 
+  // Select "CurrentTransaction" when we need to validate rows.
+  if (row_batches_need_validation_) {
+    // In case of zero-slot scans (e.g. count(*) over the table) we only select the
+    // 'currentTransaction' column.
+    if (scan_node_->IsZeroSlotTableScan()) selected_type_ids_.clear();
+    if (std::find(selected_type_ids_.begin(), selected_type_ids_.end(),
+        CURRENT_TRANSCACTION_TYPE_ID) == selected_type_ids_.end()) {
+      selected_type_ids_.push_back(CURRENT_TRANSCACTION_TYPE_ID);
+    }
+  }
+
   COUNTER_SET(num_cols_counter_, static_cast<int64_t>(selected_type_ids_.size()));
   row_reader_options_.includeTypes(selected_type_ids_);
   return Status::OK();
@@ -465,7 +544,13 @@ Status HdfsOrcScanner::ProcessSplit() {
   do {
     unique_ptr<RowBatch> batch = make_unique<RowBatch>(scan_node_->row_desc(),
         state_->batch_size(), scan_node_->mem_tracker());
+    if (scan_node_->is_partition_key_scan()) batch->limit_capacity(1);
     Status status = GetNextInternal(batch.get());
+
+    // If we are doing a partition key scan, we are done scanning the file after
+    // returning at least one row.
+    if (scan_node_->is_partition_key_scan() && batch->num_rows() > 0) eos_ = true;
+
     // Always add batch to the queue because it may contain data referenced by previously
     // appended batches.
     scan_node->AddMaterializedRowBatch(move(batch));
@@ -479,7 +564,10 @@ Status HdfsOrcScanner::ProcessSplit() {
 }
 
 Status HdfsOrcScanner::GetNextInternal(RowBatch* row_batch) {
-  if (scan_node_->IsZeroSlotTableScan()) {
+  // In case 'row_batches_need_validation_' is true, we need to look at the row
+  // batches and check their validity. In that case 'currentTransaction' is the only
+  // selected field from the file (in case of zero slot scans).
+  if (scan_node_->IsZeroSlotTableScan() && !row_batches_need_validation_) {
     uint64_t file_rows = reader_->getNumberOfRows();
     // There are no materialized slots, e.g. count(*) over the table.  We can serve
     // this query from just the file metadata.  We don't need to read the column data.
@@ -519,8 +607,7 @@ Status HdfsOrcScanner::GetNextInternal(RowBatch* row_batch) {
   // 'TransferTuples' here.
   if (!orc_root_reader_->EndOfBatch()) {
     assemble_rows_timer_.Start();
-    RETURN_IF_ERROR(TransferTuples(orc_root_reader_, row_batch,
-        !orc_root_reader_->HasCollectionChild()));
+    RETURN_IF_ERROR(TransferTuples(row_batch));
     assemble_rows_timer_.Stop();
     if (row_batch->AtCapacity()) return Status::OK();
     DCHECK(orc_root_reader_->EndOfBatch());
@@ -586,6 +673,9 @@ Status HdfsOrcScanner::NextStripe() {
   advance_stripe_ = false;
   stripe_rows_read_ = 0;
 
+  bool first_invocation = stripe_idx_ == -1;
+  int64_t skipped_rows = 0;
+
   // Loop until we have found a non-empty stripe.
   while (true) {
     // Reset the parse status for the next stripe.
@@ -615,12 +705,17 @@ Status HdfsOrcScanner::NextStripe() {
         stripe_mid_pos < split_offset + split_length)) {
       // Middle pos not in split, this stripe will be handled by a different scanner.
       // Mark if the stripe overlaps with the split.
+      if (first_invocation) skipped_rows += stripe->getNumberOfRows();
       misaligned_stripe_skipped |= CheckStripeOverlapsSplit(stripe_offset,
           stripe_offset + stripe_len, split_offset, split_offset + split_length);
       continue;
     }
 
-    // TODO: check if this stripe can be skipped by stats. e.g. IMPALA-6505
+    // TODO: check if this stripe can be skipped by stats. e.g. IMPALA-6505 In that case,
+    // set the file row index in 'orc_root_reader_' accordingly.
+    if (first_invocation && acid_synthetic_rowid_ != nullptr) {
+      orc_root_reader_->SetFileRowIndex(skipped_rows);
+    }
 
     COUNTER_ADD(num_stripes_counter_, 1);
     row_reader_options_.range(stripe->getOffset(), stripe_len);
@@ -680,8 +775,7 @@ Status HdfsOrcScanner::AssembleRows(RowBatch* row_batch) {
       num_rows_read += orc_root_batch_->numElements;
     }
 
-    RETURN_IF_ERROR(TransferTuples(orc_root_reader_, row_batch,
-       !orc_root_reader_->HasCollectionChild()));
+    RETURN_IF_ERROR(TransferTuples(row_batch));
     if (row_batch->AtCapacity()) break;
     continue_execution &= !scan_node_->ReachedLimitShared() && !context_->cancelled();
   }
@@ -693,70 +787,24 @@ Status HdfsOrcScanner::AssembleRows(RowBatch* row_batch) {
   return Status::OK();
 }
 
-Status HdfsOrcScanner::TransferTuples(OrcComplexColumnReader* coll_reader,
-    RowBatch* dst_batch, bool do_batch_read) {
-  if (!coll_reader->MaterializeTuple()) {
-    // Top-level readers that are not materializing tuples will delegate the
-    // materialization to its unique child.
-    DCHECK_EQ(coll_reader->children().size(), 1);
-    OrcColumnReader* child = coll_reader->children()[0];
-    // Only complex type readers can be top-level readers.
-    DCHECK(child->IsComplexColumnReader());
-    return TransferTuples(
-        static_cast<OrcComplexColumnReader*>(child), dst_batch, do_batch_read);
-  }
-  const TupleDescriptor* tuple_desc = scan_node_->tuple_desc();
-
-  ScalarExprEvaluator* const* conjunct_evals = conjunct_evals_->data();
-  int num_conjuncts = conjunct_evals_->size();
-
+Status HdfsOrcScanner::TransferTuples(RowBatch* dst_batch) {
   DCHECK_LT(dst_batch->num_rows(), dst_batch->capacity());
   if (tuple_ == nullptr) RETURN_IF_ERROR(AllocateTupleMem(dst_batch));
   int row_id = dst_batch->num_rows();
   int capacity = dst_batch->capacity();
-  int num_to_commit = 0;
-  TupleRow* dst_row = dst_batch->GetRow(row_id);
-  Tuple* tuple = tuple_;  // tuple_ is updated in CommitRows
 
-  // TODO(IMPALA-6506): codegen the runtime filter + conjunct evaluation loop
-  while (row_id < capacity && !coll_reader->EndOfBatch()) {
-    if (do_batch_read) {
-      DCHECK(scratch_batch_ != nullptr);
-      DCHECK(scratch_batch_->AtEnd());
-      RETURN_IF_ERROR(scratch_batch_->Reset(state_));
-      InitTupleBuffer(template_tuple_, scratch_batch_->tuple_mem,
-          scratch_batch_->capacity);
-
-      RETURN_IF_ERROR(coll_reader->TopLevelReadValueBatch(scratch_batch_.get(),
-          &scratch_batch_->aux_mem_pool));
-      int num_tuples_transferred = TransferScratchTuples(dst_batch);
-      row_id += num_tuples_transferred;
-      num_to_commit += num_tuples_transferred;
-      // Commit rows before next iteration. otherwise, they will be overwritten.
-      // This happens when more than one iteration is needed, e.g. when the capacity of
-      // scratch_batch is small (due to reading wide rows).
-      VLOG_ROW << Substitute(
-          "Transfer $0 rows from scratch batch to dst_batch ($1 rows)",
-          num_to_commit, dst_batch->num_rows());
-      RETURN_IF_ERROR(CommitRows(num_tuples_transferred, dst_batch));
-    } else {
-      if (tuple_desc->byte_size() > 0) DCHECK_LT((void*)tuple, (void*)tuple_mem_end_);
-      InitTuple(tuple_desc, template_tuple_, tuple);
-      RETURN_IF_ERROR(coll_reader->TransferTuple(tuple, dst_batch->tuple_data_pool()));
-      dst_row->SetTuple(0, tuple);
-      if (!EvalRuntimeFilters(dst_row)) continue;
-      if (ExecNode::EvalConjuncts(conjunct_evals, num_conjuncts, dst_row)) {
-        dst_row = next_row(dst_row);
-        tuple = next_tuple(tuple_desc->byte_size(), tuple);
-        ++row_id;
-        ++num_to_commit;
-      }
-    }
-  }
-  if (!do_batch_read) {
+  while (row_id < capacity && !orc_root_reader_->EndOfBatch()) {
+    DCHECK(scratch_batch_ != nullptr);
+    DCHECK(scratch_batch_->AtEnd());
+    RETURN_IF_ERROR(scratch_batch_->Reset(state_));
+    InitTupleBuffer(template_tuple_, scratch_batch_->tuple_mem, scratch_batch_->capacity);
+    RETURN_IF_ERROR(orc_root_reader_->TopLevelReadValueBatch(scratch_batch_.get(),
+        &scratch_batch_->aux_mem_pool));
+    int num_tuples_transferred = TransferScratchTuples(dst_batch);
+    row_id += num_tuples_transferred;
     VLOG_ROW << Substitute("Transfer $0 rows from scratch batch to dst_batch ($1 rows)",
-        num_to_commit, dst_batch->num_rows());
-    return CommitRows(num_to_commit, dst_batch);
+        num_tuples_transferred, dst_batch->num_rows());
+    RETURN_IF_ERROR(CommitRows(num_tuples_transferred, dst_batch));
   }
   return Status::OK();
 }
@@ -774,7 +822,7 @@ Status HdfsOrcScanner::AllocateTupleMem(RowBatch* row_batch) {
 Status HdfsOrcScanner::AssembleCollection(
     const OrcComplexColumnReader& complex_col_reader, int row_idx,
     CollectionValueBuilder* coll_value_builder) {
-  int total_tuples = complex_col_reader.GetNumTuples(row_idx);
+  int total_tuples = complex_col_reader.GetNumChildValues(row_idx);
   if (!complex_col_reader.MaterializeTuple()) {
     // 'complex_col_reader' maps to a STRUCT or collection column of STRUCTs/collections
     // and there're no need to materialize current level tuples. Delegate the
@@ -809,9 +857,8 @@ Status HdfsOrcScanner::AssembleCollection(
 
     int64_t num_rows;
     // We're assembling item tuples into an CollectionValue
-    parse_status_ =
-        GetCollectionMemory(coll_value_builder, &pool, &tuple, &row, &num_rows);
-    if (UNLIKELY(!parse_status_.ok())) break;
+    RETURN_IF_ERROR(
+        GetCollectionMemory(coll_value_builder, &pool, &tuple, &row, &num_rows));
     // 'num_rows' can be very high if we're writing to a large CollectionValue. Limit
     // the number of rows we read at one time so we don't spend too long in the
     // 'num_rows' loop below before checking for cancellation or limit reached.
