@@ -19,12 +19,15 @@
 
 #include "codegen/codegen-anyval.h"
 #include "exec/base-sequence-scanner.h"
-#include "exec/text-converter.h"
+#include "exec/exec-node.inline.h"
 #include "exec/hdfs-scan-node.h"
 #include "exec/hdfs-scan-node-mt.h"
 #include "exec/read-write-util.h"
+#include "exec/text-converter.h"
 #include "exec/text-converter.inline.h"
+#include "exprs/scalar-expr-evaluator.h"
 #include "runtime/collection-value-builder.h"
+#include "runtime/fragment-state.h"
 #include "runtime/hdfs-fs-cache.h"
 #include "runtime/runtime-filter.inline.h"
 #include "runtime/tuple-row.h"
@@ -106,10 +109,14 @@ Status HdfsScanner::Open(ScannerContext* context) {
     }
   }
 
-  // Initialize the template_tuple_.
-  template_tuple_ = scan_node_->InitTemplateTuple(
-      context_->partition_descriptor()->partition_key_value_evals(),
-      template_tuple_pool_.get(), state_);
+  // Initialize the template_tuple_, it is copied from the template tuple map in the
+  // HdfsScanNodeBase.
+  Tuple* template_tuple =
+      scan_node_->GetTemplateTupleForPartitionId(context_->partition_descriptor()->id());
+  if (template_tuple != nullptr) {
+    template_tuple_ =
+        template_tuple->DeepCopy(*scan_node_->tuple_desc(), template_tuple_pool_.get());
+  }
   template_tuple_map_[scan_node_->tuple_desc()] = template_tuple_;
 
   decompress_timer_ = ADD_TIMER(scan_node_->runtime_profile(), "DecompressionTime");
@@ -132,12 +139,15 @@ Status HdfsScanner::ProcessSplit() {
     }
     unique_ptr<RowBatch> batch = std::make_unique<RowBatch>(scan_node_->row_desc(),
         state_->batch_size(), scan_node_->mem_tracker());
+    if (scan_node_->is_partition_key_scan()) batch->limit_capacity(1);
     Status status = GetNextInternal(batch.get());
     if (batch->num_rows() > 0) returned_rows = true;
     // Always add batch to the queue if any rows were returned because it may contain
     // data referenced by previously appended batches.
     if (returned_rows) scan_node->AddMaterializedRowBatch(move(batch));
     RETURN_IF_ERROR(status);
+    // Only need to return one row per partition for partition key scans.
+    if (returned_rows && scan_node_->is_partition_key_scan()) eos_ = true;
   } while (!eos_ && !scan_node_->ReachedLimitShared());
   return Status::OK();
 }
@@ -176,7 +186,7 @@ Status HdfsScanner::InitializeWriteTuplesFn(HdfsPartitionDescriptor* partition,
     return Status::OK();
   }
 
-  write_tuples_fn_ = reinterpret_cast<WriteTuplesFn>(scan_node_->GetCodegenFn(type));
+  write_tuples_fn_ = scan_node_->GetCodegenFn(type);
   if (write_tuples_fn_ == NULL) {
     scan_node_->IncNumScannersCodegenDisabled();
     return Status::OK();
@@ -212,6 +222,15 @@ Status HdfsScanner::CommitRows(int num_rows, RowBatch* row_batch) {
   return Status::OK();
 }
 
+int HdfsScanner::WriteAlignedTuplesCodegenOrInterpret(MemPool* pool,
+    TupleRow* tuple_row_mem, FieldLocation* fields, int num_tuples,
+    int max_added_tuples, int slots_per_tuple, int row_idx_start, bool copy_strings) {
+  return CallCodegendOrInterpreted<WriteTuplesFn>::invoke(this, write_tuples_fn_,
+      &HdfsScanner::WriteAlignedTuples, pool, tuple_row_mem, fields, num_tuples,
+      max_added_tuples, scan_node_->materialized_slots().size(), row_idx_start,
+      copy_strings);
+}
+
 int HdfsScanner::WriteTemplateTuples(TupleRow* row, int num_tuples) {
   DCHECK_GE(num_tuples, 0);
   DCHECK_EQ(scan_node_->materialized_slots().size(), 0);
@@ -321,7 +340,7 @@ bool HdfsScanner::WriteCompleteTuple(MemPool* pool, FieldLocation* fields,
 //   ret i1 false
 // }
 Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
-    RuntimeState* state, llvm::Function** write_complete_tuple_fn) {
+    FragmentState* state, llvm::Function** write_complete_tuple_fn) {
   const vector<ScalarExpr*>& conjuncts = node->conjuncts_;
   LlvmCodeGen* codegen = state->codegen();
   *write_complete_tuple_fn = NULL;
@@ -332,10 +351,17 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
   for (int i = 0; i < node->materialized_slots_.size(); ++i) {
     llvm::Function* fn = nullptr;
     SlotDescriptor* slot_desc = node->materialized_slots_[i];
-    RETURN_IF_ERROR(TextConverter::CodegenWriteSlot(codegen, tuple_desc, slot_desc, &fn,
-        node->hdfs_table_->null_column_value().data(),
-        node->hdfs_table_->null_column_value().size(), true, state->strict_mode()));
-    if (i >= LlvmCodeGen::CODEGEN_INLINE_EXPRS_THRESHOLD) codegen->SetNoInline(fn);
+
+    // If the type is CHAR, WriteSlot for this slot cannot be codegen'd. To keep codegen
+    // for other things, we call the interpreted code for this slot from the codegen'd
+    // code instead of failing codegen. See IMPALA-9747.
+    if (TextConverter::SupportsCodegenWriteSlot(slot_desc->type())) {
+      RETURN_IF_ERROR(TextConverter::CodegenWriteSlot(codegen, tuple_desc, slot_desc, &fn,
+          node->hdfs_table_->null_column_value().data(),
+          node->hdfs_table_->null_column_value().size(), true,
+          state->query_options().strict_mode));
+      if (i >= LlvmCodeGen::CODEGEN_INLINE_EXPRS_THRESHOLD) codegen->SetNoInline(fn);
+    }
     slot_fns.push_back(fn);
   }
 
@@ -381,6 +407,7 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
 
   // Extract the input args
   llvm::Value* this_arg = args[0];
+  llvm::Value* mem_pool_arg = args[1];
   llvm::Value* fields_arg = args[2];
   llvm::Value* opaque_tuple_arg = args[3];
   llvm::Value* tuple_arg =
@@ -463,8 +490,21 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
 
       // Call slot parse function
       llvm::Function* slot_fn = slot_fns[slot_idx];
-      llvm::Value* slot_parsed = builder.CreateCall(
-          slot_fn, llvm::ArrayRef<llvm::Value*>({tuple_arg, data, len}));
+
+      llvm::Value* slot_parsed = nullptr;
+      if (LIKELY(slot_fn != nullptr)) {
+        slot_parsed = builder.CreateCall(
+            slot_fn, llvm::ArrayRef<llvm::Value*>({tuple_arg, data, len}));
+      } else {
+        llvm::Value* slot_idx_value = codegen->GetI32Constant(slot_idx);
+        llvm::Function* interpreted_slot_fn = codegen->GetFunction(
+            IRFunction::HDFS_SCANNER_TEXT_CONVERTER_WRITE_SLOT_INTERPRETED_IR, false);
+        DCHECK(interpreted_slot_fn != nullptr);
+
+        slot_parsed = builder.CreateCall(interpreted_slot_fn,
+            {this_arg, slot_idx_value, opaque_tuple_arg, data, len, mem_pool_arg});
+      }
+
       llvm::Value* slot_error = builder.CreateNot(slot_parsed, "slot_parse_error");
       error_in_row = builder.CreateOr(error_in_row, slot_error, "error_in_row");
       slot_error = builder.CreateZExt(slot_error, codegen->i8_type());
@@ -484,14 +524,8 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
       // tuple against that conjunct and start a new parse_block for the next conjunct
       parse_block = llvm::BasicBlock::Create(context, "parse", fn, eval_fail_block);
       llvm::Function* conjunct_fn;
-      Status status =
-          conjuncts[conjunct_idx]->GetCodegendComputeFn(codegen, false, &conjunct_fn);
-      if (!status.ok()) {
-        stringstream ss;
-        ss << "Failed to codegen conjunct: " << status.GetDetail();
-        state->LogError(ErrorMsg(TErrorCode::GENERAL, ss.str()));
-        return status;
-      }
+      RETURN_IF_ERROR(
+          conjuncts[conjunct_idx]->GetCodegendComputeFn(codegen, false, &conjunct_fn));
       if (node->materialized_slots_.size() + conjunct_idx
           >= LlvmCodeGen::CODEGEN_INLINE_EXPRS_THRESHOLD) {
         codegen->SetNoInline(conjunct_fn);
@@ -505,7 +539,8 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
 
       llvm::Value* conjunct_args[] = {eval, tuple_row_arg};
       CodegenAnyVal result = CodegenAnyVal::CreateCallWrapped(
-          codegen, &builder, TYPE_BOOLEAN, conjunct_fn, conjunct_args, "conjunct_eval");
+          codegen, &builder, ColumnType(TYPE_BOOLEAN), conjunct_fn, conjunct_args,
+          "conjunct_eval");
       builder.CreateCondBr(result.GetVal(), parse_block, eval_fail_block);
       builder.SetInsertPoint(parse_block);
     }
@@ -527,7 +562,7 @@ Status HdfsScanner::CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
 }
 
 Status HdfsScanner::CodegenWriteAlignedTuples(const HdfsScanPlanNode* node,
-    RuntimeState* state, llvm::Function* write_complete_tuple_fn,
+    FragmentState* state, llvm::Function* write_complete_tuple_fn,
     llvm::Function** write_aligned_tuples_fn) {
   LlvmCodeGen* codegen = state->codegen();
   *write_aligned_tuples_fn = NULL;
@@ -609,7 +644,7 @@ Status HdfsScanner::CodegenInitTuple(
 // }
 //
 // EvalRuntimeFilter() is the same as the cross-compiled version except EvalOneFilter()
-// is replaced with the one generated by CodegenEvalOneFilter().
+// is replaced with the one generated by FilterContext::CodegenEval().
 Status HdfsScanner::CodegenEvalRuntimeFilters(
     LlvmCodeGen* codegen, const vector<ScalarExpr*>& filter_exprs, llvm::Function** fn) {
   llvm::LLVMContext& context = codegen->context();
@@ -638,6 +673,7 @@ Status HdfsScanner::CodegenEvalRuntimeFilters(
 
     DCHECK_GT(num_filters, 0);
     for (int i = 0; i < num_filters; ++i) {
+      // Make a clone of HdfsScanner::EvalRuntimeFilter(int i, TupleRow* row)
       llvm::Function* eval_runtime_filter_fn =
           codegen->GetFunction(IRFunction::HDFS_SCANNER_EVAL_RUNTIME_FILTER, true);
       DCHECK(eval_runtime_filter_fn != nullptr);
@@ -650,17 +686,25 @@ Status HdfsScanner::CodegenEvalRuntimeFilters(
           &eval_one_filter_fn));
       DCHECK(eval_one_filter_fn != nullptr);
 
+      // "FilterContext4Eval" is FilterContext::Eval(impala::TupleRow*).
+      // Replace ctx->Eval(row) in the above cloned function with its LLVM version
+      // which is 'eval_one_filter_fn'.
       int replaced = codegen->ReplaceCallSites(eval_runtime_filter_fn, eval_one_filter_fn,
           "FilterContext4Eval");
       DCHECK_REPLACE_COUNT(replaced, 1);
 
+      // Create a call to the above LLVM optimized eval_runtime_filter_fn.
       llvm::Value* idx = codegen->GetI32Constant(i);
       llvm::Value* passed_filter = builder.CreateCall(
           eval_runtime_filter_fn, llvm::ArrayRef<llvm::Value*>({this_arg, idx, row_arg}));
 
+      // Create a new block "continue" at the end of eval_runtime_filters_fn block.
       llvm::BasicBlock* continue_block =
           llvm::BasicBlock::Create(context, "continue", eval_runtime_filters_fn);
+      // Create a condition on top of eval_runtime_filter_fn.
       builder.CreateCondBr(passed_filter, continue_block, row_rejected_block);
+      // Mark 'continue_block' the location to accumulate any additional LLVM code
+      // for the next filter (if any).
       builder.SetInsertPoint(continue_block);
     }
     builder.CreateRet(codegen->true_value());
@@ -774,7 +818,7 @@ void HdfsScanner::CheckFiltersEffectiveness() {
     double reject_ratio = stats->rejected / static_cast<double>(stats->considered);
     if (filter->AlwaysTrue() ||
         reject_ratio < FLAGS_min_filter_reject_ratio) {
-      stats->enabled = 0;
+      stats->enabled_for_row = 0;
     }
   }
 }
@@ -813,7 +857,7 @@ Status HdfsScanner::IssueFooterRanges(HdfsScanNodeBase* scan_node,
           footer_range = scan_node->AllocateScanRange(files[i]->fs,
               files[i]->filename.c_str(), footer_size, footer_start,
               split_metadata->partition_id, footer_split->disk_id(),
-              footer_split->expected_local(), files[i]->is_erasure_coded, files[i]->mtime,
+              footer_split->expected_local(), files[i]->mtime,
               BufferOpts(footer_split->cache_options()), split);
         } else {
           // If we did not find the last split, we know it is going to be a remote read.
@@ -823,8 +867,7 @@ Status HdfsScanner::IssueFooterRanges(HdfsScanNodeBase* scan_node,
           footer_range =
               scan_node->AllocateScanRange(files[i]->fs, files[i]->filename.c_str(),
                    footer_size, footer_start, split_metadata->partition_id, -1,
-                   expected_local, files[i]->is_erasure_coded, files[i]->mtime,
-                   BufferOpts(cache_options), split);
+                   expected_local, files[i]->mtime, BufferOpts(cache_options), split);
         }
         footer_ranges.push_back(footer_range);
       } else {
