@@ -19,20 +19,24 @@
 #ifndef IMPALA_EXEC_HDFS_SCANNER_H_
 #define IMPALA_EXEC_HDFS_SCANNER_H_
 
-#include <vector>
 #include <memory>
 #include <stdint.h>
+#include <type_traits>
+#include <utility>
+#include <vector>
 #include <boost/scoped_ptr.hpp>
 
 #include "codegen/impala-ir.h"
 #include "common/global-flags.h"
 #include "common/object-pool.h"
 #include "common/status.h"
+#include "exec/exec-node.inline.h"
 #include "exec/hdfs-scan-node-base.h"
 #include "exec/scanner-context.h"
 #include "runtime/io/disk-io-mgr.h"
 #include "runtime/row-batch.h"
 #include "runtime/tuple.h"
+#include "gutil/port.h"
 
 namespace impala {
 
@@ -177,6 +181,18 @@ class HdfsScanner {
     return desc.byte_size();
   }
 
+  /// Find the index of a filter in filter_ctxs_ (a vector) through id.
+  int FindFilterIndex(int filter_id) {
+    for (int i = 0; i < filter_ctxs_.size(); ++i) {
+      const RuntimeFilter* filter = filter_ctxs_[i]->filter;
+      if (filter && filter->id() == filter_id) {
+        return i;
+      }
+    }
+    DCHECK(false);
+    return -1;
+  }
+
   /// Scanner subclasses must implement these static functions as well.  Unfortunately,
   /// c++ does not allow static virtual functions.
 
@@ -301,12 +317,13 @@ class HdfsScanner {
   /// Time spent decompressing bytes.
   RuntimeProfile::Counter* decompress_timer_ = nullptr;
 
-  /// Matching typedef for WriteAlignedTuples for codegen.  Refer to comments for
-  /// that function.
+  /// Matching typedef for WriteAlignedTuples for codegen. Refer to comments for that
+  /// function.
   typedef int (*WriteTuplesFn)(HdfsScanner*, MemPool*, TupleRow*, FieldLocation*,
       int, int, int, int, bool);
-  /// Jitted write tuples function pointer.  Null if codegen is disabled.
-  WriteTuplesFn write_tuples_fn_ = nullptr;
+  /// Jitted write tuples function pointer. Null if codegen is disabled.
+  /// Function type: WriteTuplesFn.
+  const CodegenFnPtrBase* write_tuples_fn_ = nullptr;
 
   struct LocalFilterStats {
     /// Total number of rows to which each filter was applied
@@ -320,13 +337,26 @@ class HdfsScanner {
     int64_t total_possible;
 
     /// Use known-width type to act as logical boolean.  Set to 1 if corresponding filter
-    /// in filter_ctxs_ should be applied, 0 if it was ineffective and was disabled.
-    uint8_t enabled;
+    /// in filter_ctxs_ should be applied at row level, 0 if it was ineffective and was
+    /// disabled.
+    uint8_t enabled_for_row;
+
+    /// Apply the filter at page level only.
+    uint8_t enabled_for_page;
+
+    /// Apply the filter at row group level only.
+    uint8_t enabled_for_rowgroup;
 
     /// Padding to ensure structs do not straddle cache-line boundary.
-    uint8_t padding[7];
+    uint8_t padding[5];
 
-    LocalFilterStats() : considered(0), rejected(0), total_possible(0), enabled(1) { }
+    LocalFilterStats()
+      : considered(0),
+        rejected(0),
+        total_possible(0),
+        enabled_for_row(1),
+        enabled_for_page(1),
+        enabled_for_rowgroup(1) {}
   };
 
   /// Cached runtime filter contexts, one for each filter that applies to this column.
@@ -440,6 +470,51 @@ class HdfsScanner {
       int num_tuples, int max_added_tuples, int slots_per_tuple, int row_idx_start,
       bool copy_strings);
 
+  /// Convenience struct for pointers to codegen'd function pointers (CodegenFnPtrBase*).
+  /// Calls the codegen'd function if neither the outer nor the inner pointer is nullptr,
+  /// otherwise calls the provided interpreted function.
+  ///
+  /// The signatures of the condegen'd and the interpreted functions must be the same
+  /// except that the interpreted function is a member function and the codegen'd function
+  /// is a non-member function that takes the 'this' pointer explicitly as its first
+  /// argument.
+  ///
+  /// CodegendFnT: Function pointer type of the codegen'd version. This template parameter
+  /// is separated from the other template parameters because the compiler can infer the
+  /// others but this needs to be written explicitly.
+  ///
+  /// Example:
+  ///   SomeType var = CallCodegendOrInterpreted<SomeCodegendFnType>(this,
+  ///     codegend_fn_ptr_ptr, interpreted_member_fn, arg1, arg2, arg3);
+  template <class CodegendFnT>
+  struct CallCodegendOrInterpreted {
+    /// InterpretedFnT: Member function pointer type of the interpreted version.
+    /// ThisT: type of the 'this' pointer of the caller: needed because of polymorphism.
+    /// Args: argument pack for all other arguments.
+    template <class ThisT, class InterpretedFnT, class ...Args>
+    static decltype(auto) invoke(ThisT This, const CodegenFnPtrBase* atomic_codegend_fn,
+        InterpretedFnT interpreted_fn, Args&& ...args) {
+      if (atomic_codegend_fn != nullptr) {
+        // Codegen is enabled but may not be ready (if asynchronous).
+        CodegendFnT codegend_fn
+            = reinterpret_cast<CodegendFnT>(atomic_codegend_fn->load());
+        if (codegend_fn != nullptr) {
+          // Codegen is ready.
+          return codegend_fn(This, std::forward<Args>(args)...);
+        }
+      }
+
+      // Codegen is either disabled or not ready yet.
+      return (This->*interpreted_fn)(std::forward<Args>(args)...);
+    }
+  };
+
+  /// Calls the codegen'd version of WriteAlignedTuples if codegen is enabled and ready
+  /// (in case of asynchronous codegen) or the interpreted version otherwise.
+  int WriteAlignedTuplesCodegenOrInterpret(MemPool* pool, TupleRow* tuple_row_mem,
+      FieldLocation* fields, int num_tuples, int max_added_tuples, int slots_per_tuple,
+      int row_idx_start, bool copy_strings);
+
   /// Update the decompressor_ object given a compression type or codec name. Depending on
   /// the old compression type and the new one, it may close the old decompressor and/or
   /// create a new one of different type.
@@ -489,13 +564,13 @@ class HdfsScanner {
   /// to WriteCompleteTuple. Stores the resulting function in 'write_complete_tuple_fn'
   /// if codegen was successful or NULL otherwise.
   static Status CodegenWriteCompleteTuple(const HdfsScanPlanNode* node,
-      RuntimeState* state, llvm::Function** write_complete_tuple_fn);
+      FragmentState* state, llvm::Function** write_complete_tuple_fn);
 
   /// Codegen function to replace WriteAlignedTuples.  WriteAlignedTuples is cross
   /// compiled to IR.  This function loads the precompiled IR function, modifies it,
   /// and stores the resulting function in 'write_aligned_tuples_fn' if codegen was
   /// successful or NULL otherwise.
-  static Status CodegenWriteAlignedTuples(const HdfsScanPlanNode*, RuntimeState*,
+  static Status CodegenWriteAlignedTuples(const HdfsScanPlanNode*, FragmentState*,
       llvm::Function* write_tuple_fn, llvm::Function** write_aligned_tuples_fn);
 
   /// Codegen function to replace InitTuple() removing runtime constants like the tuple
@@ -552,7 +627,7 @@ class HdfsScanner {
             template_tuple, reinterpret_cast<Tuple*>(tuple_mem), tuple_byte_size);
         tuple_mem += tuple_byte_size;
       }
-    } else if (tuple_byte_size <= CACHE_LINE_SIZE) {
+    } else if (tuple_byte_size <= CACHELINE_SIZE) {
       // If each tuple fits in a cache line, it is quicker to zero the whole memory buffer
       // instead of just the null indicators. This is because we are fetching the cache
       // line anyway and zeroing a cache line is cheap (a couple of AVX2 instructions)
@@ -586,6 +661,15 @@ class HdfsScanner {
     return reinterpret_cast<TupleRow*>(mem + sizeof(Tuple*));
   }
 
+  /// This is cross-compiled to IR. Used to call the interpreted version of
+  /// TextConverter::WriteSlot from codegen'd code. For types for which
+  /// TextConverter::SupportsCodegenWriteSlot returns false (for example CHAR),
+  /// TextConverter does not support codegenning WriteSlot, but other columns may still
+  /// benefit from codegen so instead of disabling it entirely we call the interpreted
+  /// version from codegen'd code.
+  static bool TextConverterWriteSlotInterpretedIR(HdfsScanner* hdfs_scanner,
+      int slot_idx, Tuple* tuple, const char* data, int len, MemPool* pool);
+
   /// Simple wrapper around conjunct_evals_[idx]. Used in the codegen'd version of
   /// WriteCompleteTuple() because it's easier than writing IR to access
   /// conjunct_evals_.
