@@ -17,26 +17,56 @@
 
 #include "exec/hdfs-text-scanner.h"
 
+#include <string.h>
+#include <algorithm>
+#include <map>
 #include <memory>
+#include <ostream>
+#include <utility>
 
-#include "codegen/llvm-codegen.h"
+#include <boost/algorithm/string/case_conv.hpp>
+#include <boost/algorithm/string/predicate.hpp>
+
+#include "common/compiler-util.h"
+#include "common/logging.h"
 #include "exec/delimited-text-parser.h"
 #include "exec/delimited-text-parser.inline.h"
+#include "exec/exec-node.inline.h"
 #include "exec/hdfs-plugin-text-scanner.h"
+#include "exec/hdfs-scan-node-base.h"
 #include "exec/hdfs-scan-node.h"
+#include "exec/scanner-context.h"
 #include "exec/scanner-context.inline.h"
 #include "exec/text-converter.h"
 #include "exec/text-converter.inline.h"
+#include "gen-cpp/ErrorCodes_types.h"
+#include "gutil/strings/substitute.h"
+#include "runtime/descriptors.h"
+#include "runtime/fragment-state.h"
+#include "runtime/io/request-context.h"
+#include "runtime/io/request-ranges.h"
+#include "runtime/mem-pool.h"
+#include "runtime/mem-tracker.h"
 #include "runtime/row-batch.h"
 #include "runtime/runtime-state.h"
 #include "runtime/tuple-row.h"
+#include "runtime/tuple.h"
 #include "util/codec.h"
-#include "util/decompress.h"
-#include "util/cpu-info.h"
-#include "util/debug-util.h"
+#include "util/error-util.h"
+#include "util/runtime-profile-counters.h"
+#include "util/stopwatch.h"
 
 #include "common/names.h"
 
+namespace impala {
+class LlvmCodeGen;
+class ScalarExpr;
+}
+
+namespace llvm {
+class Function;
+}
+
 using boost::algorithm::ends_with;
 using boost::algorithm::to_lower;
 using namespace impala;
@@ -118,8 +148,7 @@ Status HdfsTextScanner::IssueInitialRanges(HdfsScanNodeBase* scan_node,
           ScanRange* file_range = scan_node->AllocateScanRange(files[i]->fs,
               files[i]->filename.c_str(), files[i]->file_length, 0,
               metadata->partition_id, split->disk_id(), split->expected_local(),
-              files[i]->is_erasure_coded, files[i]->mtime,
-              BufferOpts(split->cache_options()));
+              files[i]->mtime, BufferOpts(split->cache_options()));
           compressed_text_scan_ranges.push_back(file_range);
           scan_node->max_compressed_text_file_length()->Set(files[i]->file_length);
         }
@@ -614,7 +643,7 @@ Status HdfsTextScanner::FillByteBufferCompressedFile(bool* eosr) {
   // For other compressed text: attempt to read and decompress the entire file, point
   // to the decompressed buffer, and then continue normal processing.
   DCHECK(decompression_type_ != THdfsCompression::SNAPPY);
-  HdfsFileDesc* desc = scan_node_->GetFileDesc(
+  const HdfsFileDesc* desc = scan_node_->GetFileDesc(
       context_->partition_descriptor()->id(), stream_->filename());
   int64_t file_size = desc->file_length;
   DCHECK_GT(file_size, 0);
@@ -762,7 +791,7 @@ Status HdfsTextScanner::CheckForSplitDelimiter(bool* split_delimiter) {
 // Codegen for materializing parsed data into tuples.  The function WriteCompleteTuple is
 // handcrafted using the IRBuilder for the specific tuple description.  This function
 // is then injected into the cross-compiled driving function, WriteAlignedTuples().
-Status HdfsTextScanner::Codegen(HdfsScanPlanNode* node, RuntimeState* state,
+Status HdfsTextScanner::Codegen(HdfsScanPlanNode* node, FragmentState* state,
     llvm::Function** write_aligned_tuples_fn) {
   *write_aligned_tuples_fn = nullptr;
   DCHECK(state->ShouldCodegen());
@@ -857,21 +886,18 @@ int HdfsTextScanner::WriteFields(int num_fields, int num_tuples, MemPool* pool,
     int max_added_tuples = (scan_node_->limit() == -1) ?
         num_tuples :
         scan_node_->limit() - scan_node_->rows_returned_shared();
-    int tuples_returned = 0;
-    // Call jitted function if possible
+
     if (write_tuples_fn_ != nullptr) {
       // HdfsScanner::InitializeWriteTuplesFn() will skip codegen if there are string
       // slots and escape characters. TextConverter::WriteSlot() will be used instead.
       DCHECK(scan_node_->tuple_desc()->string_slots().empty() ||
           delimited_text_parser_->escape_char() == '\0');
-      tuples_returned = write_tuples_fn_(this, pool, row, fields, num_tuples,
-          max_added_tuples, scan_node_->materialized_slots().size(),
-          num_tuples_processed, copy_strings);
-    } else {
-      tuples_returned = WriteAlignedTuples(pool, row, fields, num_tuples,
-          max_added_tuples, scan_node_->materialized_slots().size(),
-          num_tuples_processed, copy_strings);
     }
+
+    int tuples_returned = WriteAlignedTuplesCodegenOrInterpret(pool, row, fields,
+        num_tuples, max_added_tuples, scan_node_->materialized_slots().size(),
+        num_tuples_processed, copy_strings);
+
     if (tuples_returned == -1) return 0;
     DCHECK_EQ(slot_idx_, 0);
 
