@@ -21,7 +21,10 @@
 #include <gutil/strings/substitute.h>
 
 #include "common/names.h"
+#include "exec/blocking-join-node.inline.h"
+#include "exec/exec-node.inline.h"
 #include "exec/exec-node-util.h"
+#include "exec/join-op.h"
 #include "exprs/scalar-expr-evaluator.h"
 #include "exprs/scalar-expr.h"
 #include "gen-cpp/PlanNodes_types.h"
@@ -36,7 +39,7 @@
 using namespace impala;
 using namespace strings;
 
-Status NestedLoopJoinPlanNode::Init(const TPlanNode& tnode, RuntimeState* state) {
+Status NestedLoopJoinPlanNode::Init(const TPlanNode& tnode, FragmentState* state) {
   RETURN_IF_ERROR(BlockingJoinPlanNode::Init(tnode, state));
   DCHECK(tnode.join_node.__isset.nested_loop_join_node);
   // join_conjunct_evals_ are evaluated in the context of rows assembled from
@@ -283,7 +286,15 @@ end:
   if (eos_) {
     *eos = true;
     probe_batch_->TransferResourceOwnership(output_batch);
-    build_batches_->TransferResourceOwnership(output_batch);
+    if (ReturnsBuildData(join_op_)) {
+      if (builder_->num_probe_threads() == 1) {
+        build_batches_->TransferResourceOwnership(output_batch);
+      } else {
+        // Resources can't be transferred to all NestedLoopJoinNodes sharing the builder,
+        // each finstance must copy out data it needs to hold onto.
+        output_batch->MarkNeedsDeepCopy();
+      }
+    }
   }
   COUNTER_SET(rows_returned_counter_, rows_returned());
   return Status::OK();
