@@ -20,7 +20,11 @@
 #include <queue>
 
 #include "runtime/collection-value-builder.h"
+#include "runtime/date-value.h"
+#include "runtime/decimal-value.h"
+#include "runtime/string-value.inline.h"
 #include "runtime/timestamp-value.inline.h"
+#include "util/mem-util.h"
 #include "common/names.h"
 
 using namespace impala;
@@ -31,6 +35,13 @@ string PrintNode(const orc::Type* node) {
   return Substitute("$0 column (ORC id=$1)", node->toString(), node->getColumnId());
 }
 
+bool OrcRowValidator::IsRowBatchValid() const {
+  if (write_ids_ == nullptr || write_ids_->numElements == 0) return true;
+
+  int64_t write_id = write_ids_->data[0];
+  return valid_write_ids_.IsWriteIdValid(write_id);
+}
+
 OrcColumnReader* OrcColumnReader::Create(const orc::Type* node,
     const SlotDescriptor* slot_desc, HdfsOrcScanner* scanner) {
   DCHECK(node != nullptr);
@@ -248,75 +259,66 @@ Status OrcDecimal16ColumnReader::ReadValue(int row_idx, Tuple* tuple, MemPool* p
   orc::Int128 orc_val = batch_->values.data()[row_idx];
 
   DCHECK_EQ(slot_desc_->type().GetByteSize(), 16);
-  int128_t val = orc_val.getHighBits();
+  __int128_t val = orc_val.getHighBits();
   val <<= 64;
   val |= orc_val.getLowBits();
   // Use memcpy to avoid gcc generating unaligned instructions like movaps
   // for int128_t. They will raise SegmentFault when addresses are not
   // aligned to 16 bytes.
-  memcpy(GetSlot(tuple), &val, sizeof(int128_t));
+  memcpy(GetSlot(tuple), &val, sizeof(__int128_t));
   return Status::OK();
 }
 
 OrcComplexColumnReader::OrcComplexColumnReader(const orc::Type* node,
     const TupleDescriptor* table_tuple_desc, HdfsOrcScanner* scanner)
-    : OrcColumnReader(node, nullptr, scanner) {
-  SchemaPath& path = scanner->col_id_path_map_[node->getColumnId()];
-  if (path == table_tuple_desc->tuple_path()) tuple_desc_ = table_tuple_desc;
+    : OrcBatchedReader(node, nullptr, scanner) {
+  uint64_t node_col_id = node->getColumnId();
+  uint64_t tuple_desc_col_id = GetColId(table_tuple_desc);
+  if (node_col_id == tuple_desc_col_id) tuple_desc_ = table_tuple_desc;
   materialize_tuple_ = (tuple_desc_ != nullptr);
   VLOG(3) << "Created top level ComplexColumnReader for " << PrintNode(node)
       << ": tuple_desc_=" << (tuple_desc_ ? tuple_desc_->DebugString() : "null");
 }
 
-bool OrcComplexColumnReader::EndOfBatch() {
+bool OrcStructReader::EndOfBatch() {
   DCHECK(slot_desc_ == nullptr
       && (tuple_desc_ == nullptr || tuple_desc_ == scanner_->scan_node_->tuple_desc()))
       << "Should be top level reader when calling EndOfBatch()";
-  if (!materialize_tuple_) {
-    // If this reader is not materializing tuples, its 'row_idx_' is invalid and the
-    // progress is tracked in the child. Delegate the judgement to the child recursively.
-    DCHECK_EQ(children_.size(), 1);
-    return static_cast<OrcComplexColumnReader*>(children_[0])->EndOfBatch();
-  }
-  DCHECK(vbatch_ == nullptr || row_idx_ <= vbatch_->numElements);
-  return vbatch_ == nullptr || row_idx_ == vbatch_->numElements;
+  DCHECK(vbatch_ == nullptr || row_idx_ <= NumElements());
+  return vbatch_ == nullptr || row_idx_ == NumElements();
 }
 
-bool OrcComplexColumnReader::HasCollectionChild() const {
-  return HasCollectionChildRecursive(this);
-}
-
-bool OrcComplexColumnReader::HasCollectionChildRecursive(
-    const OrcColumnReader* reader) const {
-  DCHECK(reader != nullptr);
-  if (reader->IsCollectionReader()) return true;
-  if (!reader->IsComplexColumnReader()) return false;
-  const OrcComplexColumnReader* complex_reader =
-      static_cast<const OrcComplexColumnReader*>(reader);
-  DCHECK(complex_reader == dynamic_cast<const OrcComplexColumnReader*>(reader));
-  for (OrcColumnReader* child : complex_reader->children_) {
-    if (HasCollectionChildRecursive(child)) return true;
-  }
-  return false;
-}
-
-inline bool PathContains(const SchemaPath& path, const SchemaPath& sub_path) {
-  return path.size() >= sub_path.size() &&
-      std::equal(sub_path.begin(), sub_path.end(), path.begin());
-}
-
-inline const SchemaPath& GetTargetColPath(const SlotDescriptor* slot_desc) {
+inline uint64_t OrcComplexColumnReader::GetTargetColId(
+    const SlotDescriptor* slot_desc) const {
   return slot_desc->type().IsCollectionType() ?
-         slot_desc->collection_item_descriptor()->tuple_path(): slot_desc->col_path();
+         GetColId(slot_desc->collection_item_descriptor()):
+         GetColId(slot_desc);
 }
 
-bool OrcStructReader::FindChild(const orc::Type& parent, const SchemaPath& child_path,
+/**
+ * Returns true if 'candidate_col_id' is under 'node' in the type hierarchy.
+ */
+inline bool IsDescendant(const orc::Type& node, uint64_t candidate_col_id) {
+  uint64_t node_col_id = node.getColumnId();
+  uint64_t node_max_col_id = node.getMaximumColumnId();
+  return node_col_id <= candidate_col_id && candidate_col_id <= node_max_col_id;
+}
+
+/**
+ * 'parent' is a complex type that might have multiple children, i.e. STRUCT or MAP.
+ * 'descendant_col_id' refers to an ORC type somewhere under 'parent' in the type
+ * hieararchy. This method selects the direct child of 'parent' that is also the ancestor
+ * of 'descendant_col_id'. Return the result in '*child' and its index in '*field'.
+ * Returns false for not found.
+ */
+bool FindChild(const orc::Type& parent, uint64_t descendant_col_id,
     const orc::Type** child, int* field) {
+  DCHECK(parent.getKind() == orc::TypeKind::STRUCT ||
+         parent.getKind() == orc::TypeKind::MAP);
   int size = parent.getSubtypeCount();
   for (int c = 0; c < size; ++c) {
     const orc::Type* node = parent.getSubtype(c);
-    const SchemaPath& node_path = scanner_->col_id_path_map_[node->getColumnId()];
-    if (PathContains(child_path, node_path)) {
+    if (node && IsDescendant(*node, descendant_col_id)) {
       *child = node;
       *field = c;
       return true;
@@ -332,7 +334,7 @@ void OrcStructReader::CreateChildForSlot(const orc::Type* curr_node,
   // Create a child reader and pass down 'slot_desc'.
   const orc::Type* child_node;
   int field;
-  if (!FindChild(*curr_node, GetTargetColPath(slot_desc), &child_node, &field)) {
+  if (!FindChild(*curr_node, GetTargetColId(slot_desc), &child_node, &field)) {
     DCHECK(false) << PrintNode(curr_node) << " has no children selected for "
         << slot_desc->DebugString();
   }
@@ -344,6 +346,9 @@ void OrcStructReader::CreateChildForSlot(const orc::Type* curr_node,
 OrcStructReader::OrcStructReader(const orc::Type* node,
     const TupleDescriptor* table_tuple_desc, HdfsOrcScanner* scanner)
     : OrcComplexColumnReader(node, table_tuple_desc, scanner) {
+  bool needs_row_validation = table_tuple_desc == scanner_->scan_node_->tuple_desc() &&
+                              node->getColumnId() == 0 &&
+                              scanner_->row_batches_need_validation_;
   if (materialize_tuple_) {
     for (SlotDescriptor* child_slot : tuple_desc_->slots()) {
       // Skip partition columns and missed columns
@@ -358,11 +363,21 @@ OrcStructReader::OrcStructReader(const orc::Type* node,
     // matches to a descendant of 'node'. Those tuples should be materialized by the
     // corresponding descendant reader. So 'node' should have exactly one selected
     // subtype: the child in the path to the target descendant.
-    DCHECK_EQ(node->getSubtypeCount(), 1);
+    DCHECK_EQ(node->getSubtypeCount(), needs_row_validation ? 2 : 1);
+    int child_index = needs_row_validation ? 1 : 0;
     OrcComplexColumnReader* child = OrcComplexColumnReader::CreateTopLevelReader(
-        node->getSubtype(0), table_tuple_desc, scanner);
+        node->getSubtype(child_index), table_tuple_desc, scanner);
     children_.push_back(child);
-    children_fields_.push_back(0);
+    children_fields_.push_back(child_index);
+  }
+  if (needs_row_validation) {
+    row_validator_.reset(new OrcRowValidator(scanner_->valid_write_ids_));
+    for (int i = 0; i < node->getSubtypeCount(); ++i) {
+      if (node->getSubtype(i)->getColumnId() == CURRENT_TRANSCACTION_TYPE_ID) {
+        current_write_id_field_index_ = i;
+        break;
+      }
+    }
   }
 }
 
@@ -378,6 +393,11 @@ OrcStructReader::OrcStructReader(const orc::Type* node,
 }
 
 Status OrcStructReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool) {
+  if (!MaterializeTuple()) {
+    DCHECK_EQ(1, children_.size());
+    OrcColumnReader* child = children_[0];
+    return child->ReadValue(row_idx, tuple, pool);
+  }
   if (IsNull(DCHECK_NOTNULL(batch_), row_idx)) {
     for (OrcColumnReader* child : children_) child->SetNullSlot(tuple);
     return Status::OK();
@@ -390,6 +410,12 @@ Status OrcStructReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool) {
 
 Status OrcStructReader::TopLevelReadValueBatch(ScratchTupleBatch* scratch_batch,
     MemPool* pool) {
+  // Validate row batch if needed.
+  if (row_validator_) DCHECK(scanner_->row_batches_need_validation_);
+  if (row_validator_ && !row_validator_->IsRowBatchValid()) {
+    row_idx_ = NumElements();
+    return Status::OK();
+  }
   // Saving the initial value of num_tuples because each child->ReadValueBatch() will
   // update it.
   int scratch_batch_idx = scratch_batch->num_tuples;
@@ -405,21 +431,48 @@ Status OrcStructReader::TopLevelReadValueBatch(ScratchTupleBatch* scratch_batch,
           item_count, orc_column_id_, scratch_batch->num_tuples));
     }
   }
-  row_idx_ += scratch_batch->num_tuples - scratch_batch_idx;
+  int num_rows_read = scratch_batch->num_tuples - scratch_batch_idx;
   if (children_.empty()) {
-    DCHECK_EQ(scratch_batch_idx, scratch_batch->num_tuples);
-    int num_to_fake_read = std::min(scratch_batch->capacity - scratch_batch->num_tuples,
-                                    (int)batch_->numElements - row_idx_);
-    scratch_batch->num_tuples += num_to_fake_read;
-    row_idx_ += num_to_fake_read;
+    // We allow empty 'children_' for original files, because we might select the
+    // synthetic 'rowid' field which is not present in original files.
+    // We also allow empty 'children_' when we need to validate row batches of a zero slot
+    // scan. In that case 'children_' is empty and only 'row_validator_' owns an ORC
+    // vector batch (the write id batch).
+    bool valid_empty_children = scanner_->acid_original_file_ ||
+         (scanner_->row_batches_need_validation_ &&
+          scanner_->scan_node_->IsZeroSlotTableScan());
+    if (!valid_empty_children) {
+      return Status(Substitute("Parse error in possibly corrupt ORC file: '$0'",
+          scanner_->filename()));
+    }
+    DCHECK_EQ(0, num_rows_read);
+    num_rows_read = std::min(scratch_batch->capacity - scratch_batch->num_tuples,
+                             NumElements() - row_idx_);
+    scratch_batch->num_tuples += num_rows_read;
   }
+  if (scanner_->acid_synthetic_rowid_ != nullptr) {
+    FillSyntheticRowId(scratch_batch, scratch_batch_idx, num_rows_read);
+  }
+  row_idx_ += num_rows_read;
   return Status::OK();
 }
 
+void OrcStructReader::FillSyntheticRowId(ScratchTupleBatch* scratch_batch,
+    int scratch_batch_idx, int num_rows) {
+    DCHECK(scanner_->acid_synthetic_rowid_ != nullptr);
+    int tuple_size = OrcColumnReader::scanner_->tuple_byte_size();
+    uint8_t* first_tuple = scratch_batch->tuple_mem + scratch_batch_idx * tuple_size;
+    int64_t* first_slot = reinterpret_cast<Tuple*>(first_tuple)->GetBigIntSlot(
+        scanner_->acid_synthetic_rowid_->tuple_offset());
+    StrideWriter<int64_t> out{first_slot, tuple_size};
+    for (int i = 0; i < num_rows; ++i) {
+      *out.Advance() = file_row_idx_++;
+    }
+}
+
 Status OrcStructReader::ReadValueBatch(int row_idx, ScratchTupleBatch* scratch_batch,
     MemPool* pool, int scratch_batch_idx) {
   for (OrcColumnReader* child : children_) {
-    DCHECK(!child->IsCollectionReader());
     RETURN_IF_ERROR(
         child->ReadValueBatch(row_idx, scratch_batch, pool, scratch_batch_idx));
   }
@@ -443,32 +496,31 @@ Status OrcStructReader::UpdateInputBatch(orc::ColumnVectorBatch* orc_batch) {
   for (int c = 0; c < size; ++c) {
     RETURN_IF_ERROR(children_[c]->UpdateInputBatch(batch_->fields[children_fields_[c]]));
   }
-  return Status::OK();
-}
-
-Status OrcStructReader::TransferTuple(Tuple* tuple, MemPool* pool) {
-  for (OrcColumnReader* child : children_) {
-    RETURN_IF_ERROR(child->ReadValue(row_idx_, tuple, pool));
+  if (row_validator_) {
+    orc::ColumnVectorBatch* write_id_batch =
+        batch_->fields[current_write_id_field_index_];
+    DCHECK_EQ(static_cast<orc::LongVectorBatch*>(write_id_batch),
+              dynamic_cast<orc::LongVectorBatch*>(write_id_batch));
+    row_validator_->UpdateTransactionBatch(
+        static_cast<orc::LongVectorBatch*>(write_id_batch));
   }
-  ++row_idx_;
   return Status::OK();
 }
 
 OrcCollectionReader::OrcCollectionReader(const orc::Type* node,
     const SlotDescriptor* slot_desc, HdfsOrcScanner* scanner)
     : OrcComplexColumnReader(node, slot_desc, scanner) {
-  const SchemaPath& path = scanner->col_id_path_map_[node->getColumnId()];
   if (slot_desc->type().IsCollectionType() &&
-      slot_desc->collection_item_descriptor()->tuple_path() == path) {
-    // This is a collection SlotDescriptor whose item TupleDescriptor matches our
-    // SchemaPath. We should materialize the slot (creating a CollectionValue) and its
+      node->getColumnId() == GetTargetColId(slot_desc)) {
+    // This is a collection SlotDescriptor whose item TupleDescriptor matches
+    // 'node'. We should materialize the slot (creating a CollectionValue) and its
     // collection tuples (see more in HdfsOrcScanner::AssembleCollection).
     tuple_desc_ = slot_desc->collection_item_descriptor();
     materialize_tuple_ = true;
   }
 }
 
-Status OrcCollectionReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool) {
+Status OrcCollectionReader::AssembleCollection(int row_idx, Tuple* tuple, MemPool* pool) {
   if (IsNull(DCHECK_NOTNULL(vbatch_), row_idx)) {
     SetNullSlot(tuple);
     return Status::OK();
@@ -480,29 +532,84 @@ Status OrcCollectionReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool)
   return scanner_->AssembleCollection(*this, row_idx, &builder);
 }
 
-Status OrcCollectionReader::TopLevelReadValueBatch(ScratchTupleBatch* scratch_batch,
-    MemPool* pool) {
-  DCHECK(false);
+int OrcListReader::NumElements() const {
+  if (DirectReader()) return batch_ != nullptr ? batch_->numElements : 0;
+  if (children_.empty()) {
+    return batch_ != nullptr ? batch_->offsets[batch_->numElements] : 0;
+  }
+  return children_[0]->NumElements();
+}
+
+Status OrcListReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool) {
+  if (DirectReader()) return AssembleCollection(row_idx, tuple, pool);
+
+  for (OrcColumnReader* child : children_) {
+    RETURN_IF_ERROR(child->ReadValue(row_idx, tuple, pool));
+  }
+  if (pos_slot_desc_ != nullptr) {
+    RETURN_IF_ERROR(SetPositionSlot(row_idx, tuple));
+  }
   return Status::OK();
 }
 
-Status OrcCollectionReader::ReadValueBatch(int row_idx,
-    ScratchTupleBatch* scratch_batch, MemPool* pool, int scratch_batch_idx) {
-  DCHECK(false);
+Status OrcMapReader::ReadValue(int row_idx, Tuple* tuple, MemPool* pool) {
+  if (DirectReader()) return AssembleCollection(row_idx, tuple, pool);
+
+  for (OrcColumnReader* child : children_) {
+    RETURN_IF_ERROR(child->ReadValue(row_idx, tuple, pool));
+  }
+  return Status::OK();
+}
+
+Status OrcListReader::SetPositionSlot(int row_idx, Tuple* tuple) {
+  DCHECK(pos_slot_desc_ != nullptr);
+
+  int64_t pos = -1;
+  DCHECK_LT(list_idx_, batch_->numElements);
+  if (list_idx_ == batch_->numElements - 1 ||
+      (batch_->offsets[list_idx_] <= row_idx && row_idx < batch_->offsets[list_idx_+1])) {
+    // We are somewhere in the current list.
+    pos = row_idx - batch_->offsets[list_idx_];
+  } else if (row_idx == batch_->offsets[list_idx_+1]) {
+    // Let's move to the next list.
+    pos = 0;
+    list_idx_ += 1;
+  }
+  else if (row_idx > batch_->offsets[list_idx_+1]) {
+    // We lagged behind. Let's find our list.
+    for (int i = list_idx_; i < batch_->numElements; ++i) {
+      if (row_idx < batch_->offsets[i+1]) {
+        pos = row_idx - batch_->offsets[i];
+        list_idx_ = i;
+        break;
+      }
+    }
+  }
+  if (pos < 0) {
+      // Oops, something went wrong. It can be caused by a corrupt file, so let's raise
+      // an error.
+      return Status(Substitute(
+          "ORC list indexes and elements are inconsistent in file $0",
+          scanner_->filename()));
+  }
+  int64_t* slot_val_ptr = reinterpret_cast<int64_t*>(tuple->GetSlot(
+      pos_slot_desc_->tuple_offset()));
+  *slot_val_ptr = pos;
   return Status::OK();
 }
 
 void OrcListReader::CreateChildForSlot(const orc::Type* node,
     const SlotDescriptor* slot_desc) {
-  int depth = scanner_->col_id_path_map_[node->getColumnId()].size();
-  const SchemaPath& target_path = GetTargetColPath(slot_desc);
-  DCHECK_GT(target_path.size(), depth);
-  int field = target_path[depth];
-  if (field == SchemaPathConstants::ARRAY_POS) {
+  uint64_t slot_col_id = GetTargetColId(slot_desc);
+  DCHECK(IsDescendant(*node, slot_col_id));
+  // We have a position slot descriptor if it refers to this LIST ORC type, but it isn't
+  // a collection slot.
+  bool is_pos_slot = slot_col_id == node->getColumnId() &&
+                     slot_desc->collection_item_descriptor() == nullptr;
+  if (is_pos_slot) {
     DCHECK(pos_slot_desc_ == nullptr) << "Should have unique pos slot";
     pos_slot_desc_ = slot_desc;
   } else {
-    DCHECK_EQ(field, SchemaPathConstants::ARRAY_ITEM);
     OrcColumnReader* child = OrcColumnReader::Create(node->getSubtype(0), slot_desc,
         scanner_);
     children_.push_back(child);
@@ -549,14 +656,11 @@ Status OrcListReader::UpdateInputBatch(orc::ColumnVectorBatch* orc_batch) {
   for (OrcColumnReader* child : children_) {
     RETURN_IF_ERROR(child->UpdateInputBatch(item_batch));
   }
-  if (batch_) {
-    row_idx_ = -1;
-    NextRow();
-  }
+  list_idx_ = 0;
   return Status::OK();
 }
 
-int OrcListReader::GetNumTuples(int row_idx) const {
+int OrcListReader::GetNumChildValues(int row_idx) const {
   if (IsNull(DCHECK_NOTNULL(batch_), row_idx)) return 0;
   DCHECK_GT(batch_->offsets.size(), row_idx + 1);
   return batch_->offsets[row_idx + 1] - batch_->offsets[row_idx];
@@ -566,20 +670,6 @@ int OrcListReader::GetChildBatchOffset(int row_idx) const {
   return batch_->offsets[row_idx];
 }
 
-Status OrcListReader::TransferTuple(Tuple* tuple, MemPool* pool) {
-  if (pos_slot_desc_) {
-    int64_t* slot_val_ptr = reinterpret_cast<int64_t*>(
-        tuple->GetSlot(pos_slot_desc_->tuple_offset()));
-    *slot_val_ptr = array_idx_;
-  }
-  for (OrcColumnReader* child : children_) {
-    RETURN_IF_ERROR(child->ReadValue(array_start_ + array_idx_, tuple, pool));
-  }
-  array_idx_++;
-  if (array_start_ + array_idx_ >= array_end_) NextRow();
-  return Status::OK();
-}
-
 Status OrcListReader::ReadChildrenValue(int row_idx, int tuple_idx, Tuple* tuple,
     MemPool* pool) const {
   DCHECK_LT(row_idx, batch_->numElements);
@@ -595,38 +685,19 @@ Status OrcListReader::ReadChildrenValue(int row_idx, int tuple_idx, Tuple* tuple
   return Status::OK();
 }
 
-void OrcListReader::NextRow() {
-  do {
-    ++row_idx_;
-    if (row_idx_ >= batch_->numElements) break;
-    array_start_ = batch_->offsets[row_idx_];
-    array_end_ = batch_->offsets[row_idx_ + 1];
-  } while (IsNull(batch_, row_idx_) || array_start_ == array_end_);
-  array_idx_ = 0;
-}
-
 void OrcMapReader::CreateChildForSlot(const orc::Type* node,
     const SlotDescriptor* slot_desc) {
-  const SchemaPath& path = scanner_->col_id_path_map_[node->getColumnId()];
-  const SchemaPath& target_path = GetTargetColPath(slot_desc);
-  int depth = path.size();
-  // The target of 'slot_desc' matches a descendant so its SchemaPath should be deeper
-  DCHECK_GT(target_path.size(), depth);
-  int field = target_path[depth];
   const orc::Type* child_type;
-  if (field == SchemaPathConstants::MAP_KEY) {
-    child_type = node->getSubtype(0);
-  } else {
-    DCHECK_EQ(field, SchemaPathConstants::MAP_VALUE);
-    child_type = node->getSubtype(1);
+  int field;
+  if (!FindChild(*node, GetTargetColId(slot_desc), &child_type, &field)) {
+    DCHECK(false) << PrintNode(node) << " has no children selected for "
+        << slot_desc->DebugString();
   }
-  DCHECK(child_type != nullptr) << Substitute(
-      "$0 matches an empty child of $1: path=$2, target_path=$3",
-      slot_desc->DebugString(), PrintNode(node), PrintNumericPath(path),
-      PrintNumericPath(target_path));
+  // Map type only has two children.
+  DCHECK_LE(field, 1);
   OrcColumnReader* child = OrcColumnReader::Create(child_type, slot_desc, scanner_);
   children_.push_back(child);
-  if (field == SchemaPathConstants::MAP_KEY) {
+  if (field == 0) {
     key_readers_.push_back(child);
   } else {
     value_readers_.push_back(child);
@@ -642,20 +713,19 @@ OrcMapReader::OrcMapReader(const orc::Type* node,
       CreateChildForSlot(node, child_slot);
     }
   } else {
-    // 'table_tuple_desc' should match to a descendant of 'node'
-    int depth = scanner->col_id_path_map_[node->getColumnId()].size();
-    DCHECK_GT(table_tuple_desc->tuple_path().size(), depth);
-    // Create a child corresponding to the subtype in the path to the descendant.
-    int field = table_tuple_desc->tuple_path()[depth];
-    DCHECK(field == SchemaPathConstants::MAP_KEY ||
-        field == SchemaPathConstants::MAP_VALUE);
-    bool key_selected = (field == SchemaPathConstants::MAP_KEY);
-    const orc::Type* child_type =
-        key_selected ? node->getSubtype(0) : node->getSubtype(1);
+    const orc::Type* child_type;
+    int field;
+    if (!FindChild(*node, GetColId(table_tuple_desc),
+        &child_type, &field)) {
+      DCHECK(false) << PrintNode(node) << " has no children selected for "
+          << table_tuple_desc->DebugString();
+    }
+    // Map type only has two children.
+    DCHECK_LE(field, 1);
     OrcComplexColumnReader* child = OrcComplexColumnReader::CreateTopLevelReader(
         child_type, table_tuple_desc, scanner);
     children_.push_back(child);
-    if (key_selected) {
+    if (field == 0) {
       key_readers_.push_back(child);
     } else {
       value_readers_.push_back(child);
@@ -690,23 +760,10 @@ Status OrcMapReader::UpdateInputBatch(orc::ColumnVectorBatch* orc_batch) {
   for (OrcColumnReader* child : value_readers_) {
     RETURN_IF_ERROR(child->UpdateInputBatch(value_batch));
   }
-  if (batch_) {
-    row_idx_ = -1;
-    NextRow();
-  }
   return Status::OK();
 }
 
-void OrcMapReader::NextRow() {
-  do {
-    ++row_idx_;
-    if (row_idx_ >= batch_->numElements) break;
-    array_offset_ = batch_->offsets[row_idx_];
-    array_end_ = batch_->offsets[row_idx_ + 1];
-  } while (IsNull(batch_, row_idx_) || array_offset_ == array_end_);
-}
-
-int OrcMapReader::GetNumTuples(int row_idx) const {
+int OrcMapReader::GetNumChildValues(int row_idx) const {
   if (IsNull(batch_, row_idx)) return 0;
   DCHECK_GT(batch_->offsets.size(), row_idx + 1);
   return batch_->offsets[row_idx + 1] - batch_->offsets[row_idx];
@@ -716,15 +773,6 @@ int OrcMapReader::GetChildBatchOffset(int row_idx) const {
   return batch_->offsets[row_idx];
 }
 
-Status OrcMapReader::TransferTuple(Tuple* tuple, MemPool* pool) {
-  for (OrcColumnReader* child : children_) {
-    RETURN_IF_ERROR(child->ReadValue(array_offset_, tuple, pool));
-  }
-  array_offset_++;
-  if (array_offset_ >= array_end_) NextRow();
-  return Status::OK();
-}
-
 Status OrcMapReader::ReadChildrenValue(int row_idx, int tuple_idx, Tuple* tuple,
     MemPool* pool) const {
   DCHECK_LT(row_idx, batch_->numElements);
@@ -734,4 +782,13 @@ Status OrcMapReader::ReadChildrenValue(int row_idx, int tuple_idx, Tuple* tuple,
   }
   return Status::OK();
 }
+
+int OrcMapReader::NumElements() const {
+  if (DirectReader()) return batch_ != nullptr ? batch_->numElements : 0;
+  if (children_.empty()) {
+    return batch_ != nullptr ? batch_->offsets[batch_->numElements] : 0;
+  }
+  return children_[0]->NumElements();
 }
+
+} // namespace impala
