@@ -25,10 +25,13 @@
 #include <gutil/strings/substitute.h>
 
 #include "codegen/codegen-anyval.h"
+#include "exec/exec-node.inline.h"
 #include "exec/hdfs-scan-node.h"
 #include "exec/parquet/parquet-collection-column-reader.h"
 #include "exec/parquet/parquet-column-readers.h"
 #include "exec/scanner-context.inline.h"
+#include "exec/scratch-tuple-batch.h"
+#include "exprs/scalar-expr-evaluator.h"
 #include "rpc/thrift-util.h"
 #include "runtime/collection-value-builder.h"
 #include "runtime/exec-env.h"
@@ -36,6 +39,8 @@
 #include "runtime/io/request-context.h"
 #include "runtime/runtime-filter.inline.h"
 #include "runtime/runtime-state.h"
+#include "runtime/scoped-buffer.h"
+#include "service/hs2-util.h"
 #include "util/dict-encoding.h"
 #include "util/pretty-printer.h"
 #include "util/scope-exit-trigger.h"
@@ -66,12 +71,6 @@ static const string ACTUAL_RESERVATION_COUNTER_NAME = "ParquetRowGroupActualRese
 Status HdfsParquetScanner::IssueInitialRanges(HdfsScanNodeBase* scan_node,
     const vector<HdfsFileDesc*>& files) {
   DCHECK(!files.empty());
-  // Add Parquet-specific counters.
-  ADD_SUMMARY_STATS_COUNTER(
-      scan_node->runtime_profile(), IDEAL_RESERVATION_COUNTER_NAME, TUnit::BYTES);
-  ADD_SUMMARY_STATS_COUNTER(
-      scan_node->runtime_profile(), ACTUAL_RESERVATION_COUNTER_NAME, TUnit::BYTES);
-
   for (HdfsFileDesc* file : files) {
     // If the file size is less than 12 bytes, it is an invalid Parquet file.
     if (file->file_length < 12) {
@@ -95,7 +94,10 @@ HdfsParquetScanner::HdfsParquetScanner(HdfsScanNodeBase* scan_node, RuntimeState
     process_footer_timer_stats_(nullptr),
     num_cols_counter_(nullptr),
     num_stats_filtered_row_groups_counter_(nullptr),
+    num_minmax_filtered_row_groups_counter_(nullptr),
+    num_rowgroups_skipped_by_unuseful_filters_counter_(nullptr),
     num_row_groups_counter_(nullptr),
+    num_minmax_filtered_pages_counter_(nullptr),
     num_scanners_with_no_reads_counter_(nullptr),
     num_dict_filtered_row_groups_counter_(nullptr),
     parquet_compressed_page_size_counter_(nullptr),
@@ -113,6 +115,11 @@ Status HdfsParquetScanner::Open(ScannerContext* context) {
   num_stats_filtered_row_groups_counter_ =
       ADD_COUNTER(scan_node_->runtime_profile(), "NumStatsFilteredRowGroups",
           TUnit::UNIT);
+  num_minmax_filtered_row_groups_counter_ =
+      ADD_COUNTER(scan_node_->runtime_profile(), "NumRuntimeFilteredRowGroups",
+          TUnit::UNIT);
+  num_rowgroups_skipped_by_unuseful_filters_counter_ = ADD_COUNTER(
+      scan_node_->runtime_profile(), "NumRowGroupsSkippedByUnusefulFilters", TUnit::UNIT);
   num_row_groups_counter_ =
       ADD_COUNTER(scan_node_->runtime_profile(), "NumRowGroups", TUnit::UNIT);
   num_row_groups_with_page_index_counter_ =
@@ -120,6 +127,8 @@ Status HdfsParquetScanner::Open(ScannerContext* context) {
           TUnit::UNIT);
   num_stats_filtered_pages_counter_ =
       ADD_COUNTER(scan_node_->runtime_profile(), "NumStatsFilteredPages", TUnit::UNIT);
+  num_minmax_filtered_pages_counter_ =
+      ADD_COUNTER(scan_node_->runtime_profile(), "NumRuntimeFilteredPages", TUnit::UNIT);
   num_pages_counter_ =
       ADD_COUNTER(scan_node_->runtime_profile(), "NumPages", TUnit::UNIT);
   num_scanners_with_no_reads_counter_ =
@@ -134,9 +143,12 @@ Status HdfsParquetScanner::Open(ScannerContext* context) {
       scan_node_->runtime_profile(), "ParquetUncompressedPageSize", TUnit::BYTES);
   process_page_index_stats_ =
       ADD_SUMMARY_STATS_TIMER(scan_node_->runtime_profile(), "PageIndexProcessingTime");
+  row_group_ideal_reservation_counter_ = ADD_SUMMARY_STATS_COUNTER(
+      scan_node_->runtime_profile(), IDEAL_RESERVATION_COUNTER_NAME, TUnit::BYTES);
+  row_group_actual_reservation_counter_ = ADD_SUMMARY_STATS_COUNTER(
+      scan_node_->runtime_profile(), ACTUAL_RESERVATION_COUNTER_NAME, TUnit::BYTES);
 
-  codegend_process_scratch_batch_fn_ = reinterpret_cast<ProcessScratchBatchFn>(
-      scan_node_->GetCodegenFn(THdfsFileFormat::PARQUET));
+  codegend_process_scratch_batch_fn_ = scan_node_->GetCodegenFn(THdfsFileFormat::PARQUET);
   if (codegend_process_scratch_batch_fn_ == nullptr) {
     scan_node_->IncNumScannersCodegenDisabled();
   } else {
@@ -345,7 +357,13 @@ Status HdfsParquetScanner::ProcessSplit() {
     }
     unique_ptr<RowBatch> batch = make_unique<RowBatch>(scan_node_->row_desc(),
         state_->batch_size(), scan_node_->mem_tracker());
+    if (scan_node_->is_partition_key_scan()) batch->limit_capacity(1);
     Status status = GetNextInternal(batch.get());
+
+    // If we are doing a partition key scan, we are done scanning the file after
+    // returning at least one row.
+    if (scan_node_->is_partition_key_scan() && batch->num_rows() > 0) eos_ = true;
+
     // Always add batch to the queue because it may contain data referenced by previously
     // appended batches.
     scan_node->AddMaterializedRowBatch(move(batch));
@@ -459,6 +477,52 @@ Status HdfsParquetScanner::GetNextInternal(RowBatch* row_batch) {
   return Status::OK();
 }
 
+Status HdfsParquetScanner::ResolveSchemaForStatFiltering(SlotDescriptor* slot_desc,
+    bool* missing_field, SchemaNode** schema_node_ptr) {
+  DCHECK(missing_field);
+  *missing_field = false;
+  // Resolve column path.
+  SchemaNode* node = nullptr;
+  bool pos_field;
+  RETURN_IF_ERROR(schema_resolver_->ResolvePath(
+      slot_desc->col_path(), &node, &pos_field, missing_field));
+
+  if (pos_field) {
+    // The planner should not send predicates with 'pos' for stats filtering to the BE.
+    // In case there is a bug, we return an error, which will abort the query.
+    stringstream err;
+    err << "Statistics not supported for pos fields: " << slot_desc->DebugString();
+    DCHECK(false) << err.str();
+    return Status(err.str());
+  }
+
+  if (schema_node_ptr) *schema_node_ptr = node;
+
+  return Status::OK();
+}
+
+ColumnStatsReader HdfsParquetScanner::CreateStatsReader(
+    const parquet::FileMetaData& file_metadata, const parquet::RowGroup& row_group,
+    SchemaNode* node, const ColumnType& col_type) {
+  DCHECK(node);
+
+  int col_idx = node->col_idx;
+  DCHECK_LT(col_idx, row_group.columns.size());
+
+  const vector<parquet::ColumnOrder>& col_orders = file_metadata.column_orders;
+  const parquet::ColumnOrder* col_order =
+      col_idx < col_orders.size() ? &col_orders[col_idx] : nullptr;
+
+  const parquet::ColumnChunk& col_chunk = row_group.columns[col_idx];
+
+  DCHECK(node->element != nullptr);
+
+  ColumnStatsReader stat_reader =
+      CreateColumnStatsReader(col_chunk, col_type, col_order, *node->element);
+
+  return stat_reader;
+}
+
 Status HdfsParquetScanner::EvaluateStatsConjuncts(
     const parquet::FileMetaData& file_metadata, const parquet::RowGroup& row_group,
     bool* skip_row_group) {
@@ -474,17 +538,16 @@ Status HdfsParquetScanner::EvaluateStatsConjuncts(
   DCHECK(min_max_tuple_ != nullptr);
   min_max_tuple_->Init(tuple_size);
 
-  DCHECK_EQ(min_max_tuple_desc->slots().size(), min_max_conjunct_evals_.size());
+  DCHECK_GE(min_max_tuple_desc->slots().size(), min_max_conjunct_evals_.size());
   for (int i = 0; i < min_max_conjunct_evals_.size(); ++i) {
+
     SlotDescriptor* slot_desc = min_max_tuple_desc->slots()[i];
     ScalarExprEvaluator* eval = min_max_conjunct_evals_[i];
 
-    // Resolve column path to determine col idx.
+    bool missing_field = false;
     SchemaNode* node = nullptr;
-    bool pos_field;
-    bool missing_field;
-    RETURN_IF_ERROR(schema_resolver_->ResolvePath(slot_desc->col_path(),
-        &node, &pos_field, &missing_field));
+
+    RETURN_IF_ERROR(ResolveSchemaForStatFiltering(slot_desc, &missing_field, &node));
 
     if (missing_field) {
       // We are selecting a column that is not in the file. We would set its slot to NULL
@@ -495,33 +558,11 @@ Status HdfsParquetScanner::EvaluateStatsConjuncts(
       break;
     }
 
-    if (pos_field) {
-      // The planner should not send predicates with 'pos' for stats filtering to the BE.
-      // In case there is a bug, we return an error, which will abort the query.
-      stringstream err;
-      err << "Statistics not supported for pos fields: " << slot_desc->DebugString();
-      DCHECK(false) << err.str();
-      return Status(err.str());
-    }
+    ColumnStatsReader stats_reader =
+        CreateStatsReader(file_metadata, row_group, node, slot_desc->type());
 
-    int col_idx = node->col_idx;
-    DCHECK_LT(col_idx, row_group.columns.size());
-
-    const vector<parquet::ColumnOrder>& col_orders = file_metadata.column_orders;
-    const parquet::ColumnOrder* col_order = col_idx < col_orders.size() ?
-        &col_orders[col_idx] : nullptr;
-
-    const parquet::ColumnChunk& col_chunk = row_group.columns[col_idx];
-    const ColumnType& col_type = slot_desc->type();
-
-    DCHECK(node->element != nullptr);
-
-    ColumnStatsReader stat_reader = CreateColumnStatsReader(col_chunk, col_type,
-        col_order, *node->element);
-
-    int64_t null_count = 0;
-    bool null_count_result = stat_reader.ReadNullCountStat(&null_count);
-    if (null_count_result && null_count == col_chunk.meta_data.num_values) {
+    bool all_nulls = false;
+    if (stats_reader.AllNulls(&all_nulls) && all_nulls) {
       *skip_row_group = true;
       break;
     }
@@ -531,12 +572,13 @@ Status HdfsParquetScanner::EvaluateStatsConjuncts(
     if (!ColumnStatsReader::GetRequiredStatsField(fn_name, &stats_field)) continue;
 
     void* slot = min_max_tuple_->GetSlot(slot_desc->tuple_offset());
-    bool stats_read = stat_reader.ReadFromThrift(stats_field, slot);
+    bool stats_read = stats_reader.ReadFromThrift(stats_field, slot);
 
     if (stats_read) {
       TupleRow row;
       row.SetTuple(0, min_max_tuple_);
-      if (!ExecNode::EvalPredicate(eval, &row)) {
+      // Accept NULL as the predicate can contain a CAST which may fail.
+      if (!eval->EvalPredicateAcceptNull(&row)) {
         *skip_row_group = true;
         break;
       }
@@ -548,14 +590,206 @@ Status HdfsParquetScanner::EvaluateStatsConjuncts(
   return Status::OK();
 }
 
+bool HdfsParquetScanner::FilterAlreadyDisabledOrOverlapWithColumnStats(
+    int filter_id, MinMaxFilter* minmax_filter, int idx, float threshold) {
+  const TRuntimeFilterDesc& filter_desc = filter_ctxs_[idx]->filter->filter_desc();
+  const TRuntimeFilterTargetDesc& target_desc = filter_desc.targets[0];
+
+  /// If the filter is always true, not enabled for row group, or covers too much area
+  /// with respect to column min and max stats, disable the use of the filter at all
+  /// levels and proceed to next predicate.
+  bool filterAlwaysTrue = false;
+  bool columnStatsRejected = false;
+  if ((filterAlwaysTrue = minmax_filter->AlwaysTrue())
+      || !filter_stats_[idx].enabled_for_rowgroup
+      || (columnStatsRejected = FilterContext::ShouldRejectFilterBasedOnColumnStats(
+              target_desc, minmax_filter, threshold))) {
+    filter_stats_[idx].enabled_for_rowgroup = false;
+    filter_stats_[idx].enabled_for_row = false;
+    filter_stats_[idx].enabled_for_page = false;
+    filter_ctxs_[idx]->stats->IncrCounters(FilterStats::ROW_GROUPS_KEY, 1, 1, 0);
+    VLOG(3) << "A filter is determined to be not useful:"
+            << " fid=" << filter_id
+            << ", enabled_for_rowgroup=" << (bool)filter_stats_[idx].enabled_for_rowgroup
+            << ", content=" << minmax_filter->DebugString()
+            << ", target column stats: low=" << PrintTColumnValue(target_desc.low_value)
+            << ", high=" << PrintTColumnValue(target_desc.high_value)
+            << ", threshold=" << threshold;
+    return true;
+  }
+  return false;
+}
+
+Status HdfsParquetScanner::EvaluateOverlapForRowGroup(
+    const parquet::FileMetaData& file_metadata, const parquet::RowGroup& row_group,
+    bool* skip_row_group) {
+  *skip_row_group = false;
+
+  if (!state_->query_options().parquet_read_statistics) return Status::OK();
+
+  const TupleDescriptor* min_max_tuple_desc = scan_node_->min_max_tuple_desc();
+  if (GetOverlapPredicateDescs().size() > 0 && !min_max_tuple_desc) {
+    stringstream err;
+    err << "min_max_tuple_desc is null.";
+    DCHECK(false) << err.str();
+    return Status(err.str());
+  }
+
+  DCHECK(min_max_tuple_ != nullptr);
+  min_max_tuple_->Init(min_max_tuple_desc->byte_size());
+
+  // The total number slots in min_max_tuple_ should be equal to or larger than
+  // the number of min/max conjuncts. The extra number slots are for the overlap
+  // predicates. # min_max_conjuncts + 2 * # overlap predicates = # min_max_slots.
+  DCHECK_GE(min_max_tuple_desc->slots().size(), min_max_conjunct_evals_.size());
+
+  TMinmaxFilteringLevel::type level = state_->query_options().minmax_filtering_level;
+  float threshold = (float)(state_->query_options().minmax_filter_threshold);
+  bool row_group_skipped_by_unuseful_filters = false;
+
+  for (auto desc: GetOverlapPredicateDescs()) {
+    int filter_id = desc.filter_id;
+    int slot_idx = desc.slot_index;
+    /// Find the index of the filter that is common in data structure
+    /// filter_ctxs_ and filter_stats_.
+    int idx = FindFilterIndex(filter_id);
+    DCHECK(idx >= 0);
+    MinMaxFilter* minmax_filter = FindMinMaxFilter(idx);
+
+    VLOG(3) << "Try to filter out a rowgroup via overlap predicate filter: "
+            << " fid=" << filter_id
+            << ", Current time since reboot(ms)=" << MonotonicMillis();
+
+    if (!minmax_filter) {
+      // The filter is not available yet.
+      filter_ctxs_[idx]->stats->IncrCounters(FilterStats::ROW_GROUPS_KEY, 1, 0, 0);
+      continue;
+    }
+
+    if (HdfsParquetScanner::FilterAlreadyDisabledOrOverlapWithColumnStats(
+            filter_id, minmax_filter, idx, threshold)) {
+      // The filter is already disabled or too close to the column min/max stats, ignore
+      // it.
+      row_group_skipped_by_unuseful_filters = true;
+      VLOG(3) << "The filter is available but either disabled or overlapping too much "
+                 "with column stats: "
+              << " fid=" << filter_id;
+      continue;
+    }
+
+    SlotDescriptor* slot_desc = min_max_tuple_desc->slots()[slot_idx];
+
+    bool missing_field = false;
+    SchemaNode* node = nullptr;
+
+    RETURN_IF_ERROR(ResolveSchemaForStatFiltering(slot_desc, &missing_field, &node));
+
+    if (missing_field) {
+      // We are selecting a column that is not in the file. We would set its slot to NULL
+      // during the scan, so any predicate would evaluate to false. Return early. NULL
+      // comparisons cannot happen here, since predicates with NULL literals are filtered
+      // in the frontend.
+      *skip_row_group = true;
+      break;
+    }
+    ColumnStatsReader stats_reader =
+        CreateStatsReader(file_metadata, row_group, node, slot_desc->type());
+
+    bool all_nulls = false;
+    if (stats_reader.AllNulls(&all_nulls) && all_nulls) {
+      *skip_row_group = true;
+      break;
+    }
+
+    void* min_slot = nullptr;
+    void* max_slot = nullptr;
+    GetMinMaxSlotsForOverlapPred(slot_idx, &min_slot, &max_slot);
+
+    bool value_read = stats_reader.ReadMinMaxFromThrift(min_slot, max_slot);
+    if (!value_read) {
+      continue;
+    }
+
+    const ColumnType& col_type = slot_desc->type();
+
+    float overlap_ratio =
+        minmax_filter->ComputeOverlapRatio(col_type, min_slot, max_slot);
+
+    /// Compute the likelyhood of this filter being effective. Threshold 'threshold' is
+    /// the upper bound for the overlapping ratio. Any filter with an overlap ratio
+    /// (>0.0) less than 'threshold' will undertake overlap check at the page and
+    /// the row level when the filtering level control allows.
+    bool worthiness = !(minmax_filter->AlwaysTrue()) && (overlap_ratio < threshold);
+
+    VLOG(3) << "The filter is available: "
+            << "fid=" << filter_id
+            << ", SchemaNode=" << node->DebugString()
+            << ", columnType=" << col_type.DebugString()
+            << ", overlap ratio=" << overlap_ratio
+            << ", threshold=" << threshold
+            << ", worthiness=" << worthiness
+            << ", enabled for page=" <<
+              ((level != TMinmaxFilteringLevel::ROW_GROUP) && worthiness)
+            << ", enabled for row=" <<
+              ((level == TMinmaxFilteringLevel::ROW) && worthiness)
+            << ", data min="
+            << RawValue::PrintValue(min_slot, col_type, col_type.scale)
+            << ", data max="
+            << RawValue::PrintValue(max_slot, col_type, col_type.scale)
+            << ", content=" << minmax_filter->DebugString();
+
+    /// If not overlapping with this particular filter, the row group can be filtered
+    /// out safely.
+    if (overlap_ratio == 0.0) {
+      VLOG(3) << "The rowgroup was filtered out.";
+      *skip_row_group = true;
+
+      /// Update the row groups stats: rejected.
+      filter_ctxs_[idx]->stats->IncrCounters(FilterStats::ROW_GROUPS_KEY, 1, 1, 1);
+      break;
+    }
+
+    /// Update two fields in the entry at index 'idx' in local filter stats so that
+    /// the info can be used when checking out the pages in the group in
+    /// FindSkipRangesForPagesWithMinMaxFilters() and rows in these pages in
+    /// HdfsScanner::EvalRuntimeFilter(). The worthiness flag is ANDed into the enabled
+    /// field so that the state in enabled will never go back to 1 once set to 0.
+    /// The filtering level control is considered too.
+    filter_stats_[idx].enabled_for_row &=
+        (level == TMinmaxFilteringLevel::ROW) && worthiness;
+    filter_stats_[idx].enabled_for_page &=
+        (level != TMinmaxFilteringLevel::ROW_GROUP) && worthiness;
+
+    /// Update the row groups stats: no rejection.
+    filter_ctxs_[idx]->stats->IncrCounters(FilterStats::ROW_GROUPS_KEY, 1, 1, 0);
+  }
+
+  if (row_group_skipped_by_unuseful_filters) {
+    COUNTER_ADD(num_rowgroups_skipped_by_unuseful_filters_counter_, 1);
+  }
+
+  return Status::OK();
+}
+
+bool HdfsParquetScanner::ShouldProcessPageIndex() {
+  if (!state_->query_options().parquet_read_page_index) return false;
+  if (!min_max_conjunct_evals_.empty()) return true;
+  for (auto desc : GetOverlapPredicateDescs()) {
+    if (IsFilterWorthyForOverlapCheck(FindFilterIndex(desc.filter_id))) {
+      return true;
+    }
+  }
+  return false;
+}
+
 Status HdfsParquetScanner::NextRowGroup() {
   const ScanRange* split_range = static_cast<ScanRangeMetadata*>(
       metadata_range_->meta_data())->original_split;
   int64_t split_offset = split_range->offset();
   int64_t split_length = split_range->len();
 
-  HdfsFileDesc* file_desc = scan_node_->GetFileDesc(
-      context_->partition_descriptor()->id(), filename());
+  const HdfsFileDesc* file_desc =
+      scan_node_->GetFileDesc(context_->partition_descriptor()->id(), filename());
 
   bool start_with_first_row_group = row_group_idx_ == -1;
   bool misaligned_row_group_skipped = false;
@@ -610,7 +844,7 @@ Status HdfsParquetScanner::NextRowGroup() {
       COUNTER_ADD(num_row_groups_with_page_index_counter_, 1);
     }
 
-    // Evaluate row group statistics.
+    // Evaluate row group statistics with stats conjuncts.
     bool skip_row_group_on_stats;
     RETURN_IF_ERROR(
         EvaluateStatsConjuncts(file_metadata_, row_group, &skip_row_group_on_stats));
@@ -619,9 +853,17 @@ Status HdfsParquetScanner::NextRowGroup() {
       continue;
     }
 
-    // Evaluate page index.
-    if (!min_max_conjunct_evals_.empty() &&
-        state_->query_options().parquet_read_page_index) {
+    // Evaluate row group statistics with min/max filters.
+    bool skip_row_group_on_minmax;
+    RETURN_IF_ERROR(
+      EvaluateOverlapForRowGroup(file_metadata_, row_group, &skip_row_group_on_minmax));
+    if (skip_row_group_on_minmax) {
+      COUNTER_ADD(num_minmax_filtered_row_groups_counter_, 1);
+      continue;
+    }
+
+    // Evaluate page index with min-max conjuncts and/or min/max overlap predicates.
+    if (ShouldProcessPageIndex()) {
       Status page_index_status = ProcessPageIndex();
       if (!page_index_status.ok()) {
         RETURN_IF_ERROR(state_->LogOrReturnError(page_index_status.msg()));
@@ -703,6 +945,19 @@ bool HdfsParquetScanner::ReadStatFromIndex(const ColumnStatsReader& stats_reader
   return false;
 }
 
+bool HdfsParquetScanner::ReadStatFromIndex(const ColumnStatsReader& stats_reader,
+    const parquet::ColumnIndex& column_index, int page_idx, bool* is_null_page,
+    void* min_slot, void* max_slot) {
+  *is_null_page = column_index.null_pages[page_idx];
+  if (*is_null_page) return false;
+  return (
+      stats_reader.ReadFromString(ColumnStatsReader::StatsField::MIN,
+          column_index.min_values[page_idx], &column_index.max_values[page_idx], min_slot)
+      && stats_reader.ReadFromString(ColumnStatsReader::StatsField::MAX,
+          column_index.max_values[page_idx], &column_index.min_values[page_idx],
+          max_slot));
+}
+
 void HdfsParquetScanner::ResetPageFiltering() {
   filter_pages_ = false;
   candidate_ranges_.clear();
@@ -724,22 +979,86 @@ Status HdfsParquetScanner::ProcessPageIndex() {
   return Status::OK();
 }
 
-Status HdfsParquetScanner::EvaluatePageIndex() {
+const vector<TOverlapPredicateDesc>& HdfsParquetScanner::GetOverlapPredicateDescs() {
+  const HdfsScanPlanNode& pnode =
+      static_cast<const HdfsScanPlanNode&>(scan_node_->plan_node());
+  return pnode.overlap_predicate_descs_;
+}
+
+void HdfsParquetScanner::GetMinMaxSlotsForOverlapPred(
+    int overlap_slot_idx, void** min_slot, void** max_slot) {
+  DCHECK(min_slot);
+  DCHECK(max_slot);
+  SlotDescriptor* min_slot_desc =
+      scan_node_->min_max_tuple_desc()->slots()[overlap_slot_idx];
+  SlotDescriptor* max_slot_desc =
+      scan_node_->min_max_tuple_desc()->slots()[overlap_slot_idx + 1];
+  *min_slot = min_max_tuple_->GetSlot(min_slot_desc->tuple_offset());
+  *max_slot = min_max_tuple_->GetSlot(max_slot_desc->tuple_offset());
+}
+
+MinMaxFilter* HdfsParquetScanner::FindMinMaxFilter(int filter_idx) {
+  if (filter_idx >= 0 && filter_idx < filter_ctxs_.size()) {
+    const RuntimeFilter* filter = filter_ctxs_[filter_idx]->filter;
+    if (filter && filter->is_min_max_filter()) {
+      return filter->get_min_max();
+    }
+  }
+  return nullptr;
+}
+
+bool HdfsParquetScanner::IsFilterWorthyForOverlapCheck(int filter_idx) {
+  if (filter_idx >= 0 && filter_idx < filter_stats_.size()) {
+    LocalFilterStats* stats = &filter_stats_[filter_idx];
+    return stats && stats->enabled_for_page;
+  }
+  DCHECK(false);
+  return true;
+}
+
+Status HdfsParquetScanner::FindSkipRangesForPagesWithMinMaxFilters(
+    vector<RowRange>* skip_ranges) {
+  DCHECK(skip_ranges);
+  const TupleDescriptor* min_max_tuple_desc = scan_node_->min_max_tuple_desc();
+  if (GetOverlapPredicateDescs().size() > 0 && !min_max_tuple_desc) {
+    stringstream err;
+    err << "min_max_tuple_desc is null.";
+    DCHECK(false) << err.str();
+    return Status(err.str());
+  }
+
+  min_max_tuple_->Init(min_max_tuple_desc->byte_size());
   parquet::RowGroup& row_group = file_metadata_.row_groups[row_group_idx_];
-  vector<RowRange> skip_ranges;
 
-  for (int i = 0; i < min_max_conjunct_evals_.size(); ++i) {
-    ScalarExprEvaluator* eval = min_max_conjunct_evals_[i];
-    SlotDescriptor* slot_desc = scan_node_->min_max_tuple_desc()->slots()[i];
+  int filtered_pages = 0;
 
-    // Resolve column path to determine col idx.
+  for (auto desc: GetOverlapPredicateDescs()) {
+    int filter_id = desc.filter_id;
+    int slot_idx = desc.slot_index;
+
+    int filter_idx = FindFilterIndex(filter_id);
+    MinMaxFilter* minmax_filter = FindMinMaxFilter(filter_idx);
+    if (!minmax_filter || !IsFilterWorthyForOverlapCheck(filter_idx)) {
+      continue;
+    }
+
+    // This is the desc for the min slot, from which we find the stats_reader
+    // through col_path.
+    SlotDescriptor* slot_desc = min_max_tuple_desc->slots()[slot_idx];
+
+    bool missing_field = false;
     SchemaNode* node = nullptr;
-    bool pos_field, missing_field;
-    RETURN_IF_ERROR(schema_resolver_->ResolvePath(slot_desc->col_path(), &node,
-        &pos_field, &missing_field));
-    if (pos_field || missing_field) continue;
+
+    RETURN_IF_ERROR(ResolveSchemaForStatFiltering(slot_desc, &missing_field, &node));
+
+    if (missing_field) {
+      continue;
+    }
 
     int col_idx = node->col_idx;
+    ColumnStatsReader stats_reader =
+        CreateStatsReader(file_metadata_, row_group, node, slot_desc->type());
+
     DCHECK_LT(col_idx, row_group.columns.size());
     const parquet::ColumnChunk& col_chunk = row_group.columns[col_idx];
     if (col_chunk.column_index_length == 0) continue;
@@ -748,11 +1067,72 @@ Status HdfsParquetScanner::EvaluatePageIndex() {
     RETURN_IF_ERROR(page_index_.DeserializeColumnIndex(col_chunk, &column_index));
 
     const ColumnType& col_type = slot_desc->type();
-    const vector<parquet::ColumnOrder>& col_orders = file_metadata_.column_orders;
-    const parquet::ColumnOrder* col_order = col_idx < col_orders.size() ?
-        &col_orders[col_idx] : nullptr;
-    ColumnStatsReader stats_reader = CreateColumnStatsReader(col_chunk, col_type,
-        col_order, *node->element);
+    void* min_slot = nullptr;
+    void* max_slot = nullptr;
+    GetMinMaxSlotsForOverlapPred(slot_idx, &min_slot, &max_slot);
+
+    const int num_of_pages = column_index.null_pages.size();
+
+    VLOG(3) << "Try to filter out pages via overlap predicate."
+            << "  fid=" << filter_id << ", columnType=" << col_type.DebugString()
+            << ", data min="
+            << RawValue::PrintValue(min_slot, col_type, col_type.scale)
+            << ", data max="
+            << RawValue::PrintValue(max_slot, col_type, col_type.scale)
+            << ", filter=" << minmax_filter->DebugString();
+
+    for (int page_idx = 0; page_idx < num_of_pages; ++page_idx) {
+      bool is_null_page = false;
+      if (!ReadStatFromIndex(
+              stats_reader, column_index, page_idx, &is_null_page, min_slot, max_slot)) {
+        continue;
+      }
+
+      if (is_null_page || !minmax_filter->EvalOverlap(col_type, min_slot, max_slot)) {
+        VLOG(3) << "Page " << page_idx << " was filtered out.";
+        BaseScalarColumnReader* scalar_reader = scalar_reader_map_[col_idx];
+        RETURN_IF_ERROR(
+            page_index_.DeserializeOffsetIndex(col_chunk, &scalar_reader->offset_index_));
+        RowRange row_range;
+        GetRowRangeForPage(row_group, scalar_reader->offset_index_, page_idx, &row_range);
+        skip_ranges->push_back(row_range);
+        filtered_pages++;
+      }
+    }
+  }
+
+  if (filtered_pages > 0) {
+    COUNTER_ADD(num_minmax_filtered_pages_counter_, filtered_pages);
+  }
+  return Status::OK();
+}
+
+Status HdfsParquetScanner::EvaluatePageIndex() {
+  parquet::RowGroup& row_group = file_metadata_.row_groups[row_group_idx_];
+  vector<RowRange> skip_ranges;
+
+  for (int i = 0; i < min_max_conjunct_evals_.size(); ++i) {
+    ScalarExprEvaluator* eval = min_max_conjunct_evals_[i];
+    SlotDescriptor* slot_desc = scan_node_->min_max_tuple_desc()->slots()[i];
+
+    bool missing_field = false;
+    SchemaNode* node = nullptr;
+
+    RETURN_IF_ERROR(ResolveSchemaForStatFiltering(slot_desc, &missing_field, &node));
+
+    if (missing_field) {
+      continue;
+    }
+    int col_idx = node->col_idx;;
+    ColumnStatsReader stats_reader =
+        CreateStatsReader(file_metadata_, row_group, node, slot_desc->type());
+
+    DCHECK_LT(col_idx, row_group.columns.size());
+    const parquet::ColumnChunk& col_chunk = row_group.columns[col_idx];
+    if (col_chunk.column_index_length == 0) continue;
+
+    parquet::ColumnIndex column_index;
+    RETURN_IF_ERROR(page_index_.DeserializeColumnIndex(col_chunk, &column_index));
 
     min_max_tuple_->Init(scan_node_->min_max_tuple_desc()->byte_size());
     void* slot = min_max_tuple_->GetSlot(slot_desc->tuple_offset());
@@ -764,12 +1144,13 @@ Status HdfsParquetScanner::EvaluatePageIndex() {
 
     for (int page_idx = 0; page_idx < num_of_pages; ++page_idx) {
       bool value_read, is_null_page;
-      value_read = ReadStatFromIndex(stats_reader, column_index, page_idx, stats_field,
-          &is_null_page, slot);
+      value_read = ReadStatFromIndex(
+          stats_reader, column_index, page_idx, stats_field, &is_null_page, slot);
       if (!is_null_page && !value_read) continue;
       TupleRow row;
       row.SetTuple(0, min_max_tuple_);
-      if (is_null_page || !ExecNode::EvalPredicate(eval, &row)) {
+      // Accept NULL as the predicate can contain a CAST which may fail.
+      if (is_null_page || !eval->EvalPredicateAcceptNull(&row)) {
         BaseScalarColumnReader* scalar_reader = scalar_reader_map_[col_idx];
         RETURN_IF_ERROR(page_index_.DeserializeOffsetIndex(col_chunk,
             &scalar_reader->offset_index_));
@@ -779,6 +1160,13 @@ Status HdfsParquetScanner::EvaluatePageIndex() {
       }
     }
   }
+
+  // On top of min/max conjuncts, apply min/max filters to filter out pages.
+  if (state_->query_options().minmax_filtering_level
+      != TMinmaxFilteringLevel::ROW_GROUP) {
+    RETURN_IF_ERROR(FindSkipRangesForPagesWithMinMaxFilters(&skip_ranges));
+  }
+
   if (skip_ranges.empty()) return Status::OK();
 
   for (BaseScalarColumnReader* scalar_reader : scalar_readers_) {
@@ -946,9 +1334,8 @@ bool HdfsParquetScanner::IsDictionaryEncoded(
   if (col_metadata.__isset.encoding_stats) {
     // Condition #1 above
     for (const parquet::PageEncodingStats& enc_stat : col_metadata.encoding_stats) {
-      if (enc_stat.page_type == parquet::PageType::DATA_PAGE &&
-          enc_stat.encoding != parquet::Encoding::PLAIN_DICTIONARY &&
-          enc_stat.count > 0) {
+      if (enc_stat.page_type == parquet::PageType::DATA_PAGE
+          && !IsDictionaryEncoding(enc_stat.encoding) && enc_stat.count > 0) {
         return false;
       }
     }
@@ -957,10 +1344,10 @@ bool HdfsParquetScanner::IsDictionaryEncoded(
     bool has_dict_encoding = false;
     bool has_nondict_encoding = false;
     for (const parquet::Encoding::type& encoding : col_metadata.encodings) {
-      if (encoding == parquet::Encoding::PLAIN_DICTIONARY) has_dict_encoding = true;
+      if (IsDictionaryEncoding(encoding)) has_dict_encoding = true;
 
       // RLE and BIT_PACKED are used for repetition/definition levels
-      if (encoding != parquet::Encoding::PLAIN_DICTIONARY &&
+      if (!IsDictionaryEncoding(encoding) &&
           encoding != parquet::Encoding::RLE &&
           encoding != parquet::Encoding::BIT_PACKED) {
         has_nondict_encoding = true;
@@ -1350,8 +1737,9 @@ Status HdfsParquetScanner::ProcessFooter() {
   uint8_t* magic_number_ptr = buffer + scan_range_len - sizeof(PARQUET_VERSION_NUMBER);
   if (memcmp(magic_number_ptr, PARQUET_VERSION_NUMBER,
              sizeof(PARQUET_VERSION_NUMBER)) != 0) {
+    // Report the ill-formatted Parquet version string in hex.
     return Status(TErrorCode::PARQUET_BAD_VERSION_NUMBER, filename(),
-        string(reinterpret_cast<char*>(magic_number_ptr), sizeof(PARQUET_VERSION_NUMBER)),
+        ReadWriteUtil::HexDump(magic_number_ptr, sizeof(PARQUET_VERSION_NUMBER)),
         scan_node_->hdfs_table()->fully_qualified_name());
   }
 
@@ -1395,7 +1783,7 @@ Status HdfsParquetScanner::ProcessFooter() {
     ScanRange* metadata_range = scan_node_->AllocateScanRange(
         metadata_range_->fs(), filename(), metadata_size, metadata_start, partition_id,
         metadata_range_->disk_id(), metadata_range_->expected_local(),
-        metadata_range_->is_erasure_coded(), metadata_range_->mtime(),
+        metadata_range_->mtime(),
         BufferOpts::ReadInto(metadata_buffer.buffer(), metadata_size, cache_options));
 
     unique_ptr<BufferDescriptor> io_buffer;
@@ -1547,8 +1935,8 @@ Status HdfsParquetScanner::CreateColumnReaders(const TupleDescriptor& tuple_desc
 Status HdfsParquetScanner::CreateCountingReader(const SchemaPath& parent_path,
     const ParquetSchemaResolver& schema_resolver, ParquetColumnReader** reader) {
   SchemaNode* parent_node;
-  bool pos_field;
   bool missing_field;
+  bool pos_field;
   RETURN_IF_ERROR(schema_resolver.ResolvePath(
       parent_path, &parent_node, &pos_field, &missing_field));
 
@@ -1668,10 +2056,8 @@ Status HdfsParquetScanner::DivideReservationBetweenColumns(
   if (ideal_reservation > context_->total_reservation()) {
     context_->TryIncreaseReservation(ideal_reservation);
   }
-  scan_node_->runtime_profile()->GetSummaryStatsCounter(ACTUAL_RESERVATION_COUNTER_NAME)->
-      UpdateCounter(context_->total_reservation());
-  scan_node_->runtime_profile()->GetSummaryStatsCounter(IDEAL_RESERVATION_COUNTER_NAME)->
-      UpdateCounter(ideal_reservation);
+  row_group_actual_reservation_counter_->UpdateCounter(context_->total_reservation());
+  row_group_ideal_reservation_counter_->UpdateCounter(ideal_reservation);
 
   vector<pair<int, int64_t>> tmp_reservations = DivideReservationBetweenColumnsHelper(
       min_buffer_size, max_buffer_size, col_range_lengths, context_->total_reservation());
@@ -1844,9 +2230,8 @@ ColumnStatsReader HdfsParquetScanner::CreateColumnStatsReader(
 ParquetTimestampDecoder HdfsParquetScanner::CreateTimestampDecoder(
     const parquet::SchemaElement& element) {
   bool timestamp_conversion_needed_for_int96_timestamps =
-      FLAGS_convert_legacy_hive_parquet_utc_timestamps &&
       file_version_.application == "parquet-mr" &&
-      state_->local_time_zone() != UTCPTR;
+      state_->time_zone_for_legacy_parquet_time_conversions() != UTCPTR;
 
   return ParquetTimestampDecoder(element, state_->local_time_zone(),
       timestamp_conversion_needed_for_int96_timestamps);
