@@ -86,8 +86,29 @@ using namespace apache::thrift;
 // the columns and run that function over row batches.
 // TODO: we need to pass in the compression from the FE/metadata
 
+static const string PARQUET_MEM_LIMIT_EXCEEDED =
+    "HdfsParquetTableWriter::$0() failed to allocate $1 bytes for $2.";
+
+DEFINE_bool_hidden(write_new_parquet_dictionary_encodings, false,
+    "(Experimental) Write parquet files with PLAIN/RLE_DICTIONARY encoding instead of "
+    "PLAIN_DICTIONARY as recommended by Parquet 2.0 standard");
+
 namespace impala {
 
+// Returns the parquet::Encoding enum value to use for plain-encoded dictionary pages.
+static parquet::Encoding::type DictPageEncoding() {
+  return FLAGS_write_new_parquet_dictionary_encodings ?
+          parquet::Encoding::PLAIN :
+          parquet::Encoding::PLAIN_DICTIONARY;
+}
+
+// Returns the parquet::Encoding enum value to use for dictionary-encoded data pages.
+static parquet::Encoding::type DataPageDictionaryEncoding() {
+  return FLAGS_write_new_parquet_dictionary_encodings ?
+          parquet::Encoding::RLE_DICTIONARY:
+          parquet::Encoding::PLAIN_DICTIONARY;
+}
+
 // Base class for column writers. This contains most of the logic except for
 // the type specific functions which are implemented in the subclasses.
 class HdfsParquetTableWriter::BaseColumnWriter {
@@ -326,8 +347,8 @@ class HdfsParquetTableWriter::BaseColumnWriter {
 
   // Size of newly created pages. Defaults to DEFAULT_DATA_PAGE_SIZE and is increased
   // when pages are not big enough. This only happens when there are enough unique values
-  // such that we switch from PLAIN_DICTIONARY to PLAIN encoding and then have very
-  // large values (i.e. greater than DEFAULT_DATA_PAGE_SIZE).
+  // such that we switch from PLAIN_DICTIONARY/RLE_DICTIONARY to PLAIN encoding and then
+  // have very large values (i.e. greater than DEFAULT_DATA_PAGE_SIZE).
   // TODO: Consider removing and only creating a single large page as necessary.
   int64_t page_size_;
 
@@ -419,11 +440,10 @@ class HdfsParquetTableWriter::ColumnWriter :
     valid_column_index_ = true;
     // Default to dictionary encoding.  If the cardinality ends up being too high,
     // it will fall back to plain.
-    current_encoding_ = parquet::Encoding::PLAIN_DICTIONARY;
-    next_page_encoding_ = parquet::Encoding::PLAIN_DICTIONARY;
-    dict_encoder_.reset(
-        new DictEncoder<T>(parent_->per_file_mem_pool_.get(), plain_encoded_value_size_,
-            parent_->parent_->mem_tracker()));
+    current_encoding_ = DataPageDictionaryEncoding();
+    next_page_encoding_ = DataPageDictionaryEncoding();
+    dict_encoder_.reset(new DictEncoder<T>(parent_->per_file_mem_pool_.get(),
+        plain_encoded_value_size_, parent_->parent_->mem_tracker()));
     dict_encoder_base_ = dict_encoder_.get();
     page_stats_.reset(
         new ColumnStats<T>(parent_->per_file_mem_pool_.get(), plain_encoded_value_size_));
@@ -436,7 +456,7 @@ class HdfsParquetTableWriter::ColumnWriter :
  protected:
   virtual bool ProcessValue(void* value, int64_t* bytes_needed) {
     T* val = CastValue(value);
-    if (current_encoding_ == parquet::Encoding::PLAIN_DICTIONARY) {
+    if (IsDictionaryEncoding(current_encoding_)) {
       if (UNLIKELY(num_values_since_dict_size_check_ >=
                    DICTIONARY_DATA_PAGE_SIZE_CHECK_PERIOD)) {
         num_values_since_dict_size_check_ = 0;
@@ -751,7 +771,7 @@ Status HdfsParquetTableWriter::BaseColumnWriter::Flush(int64_t* file_pos,
     // Write dictionary page header
     parquet::DictionaryPageHeader dict_header;
     dict_header.num_values = dict_encoder_base_->num_entries();
-    dict_header.encoding = parquet::Encoding::PLAIN_DICTIONARY;
+    dict_header.encoding = DictPageEncoding();
     ++dict_encoding_stats_[dict_header.encoding];
 
     parquet::PageHeader header;
@@ -760,8 +780,14 @@ Status HdfsParquetTableWriter::BaseColumnWriter::Flush(int64_t* file_pos,
     header.__set_dictionary_page_header(dict_header);
 
     // Write the dictionary page data, compressing it if necessary.
-    uint8_t* dict_buffer = parent_->per_file_mem_pool_->Allocate(
-        header.uncompressed_page_size);
+    uint8_t* dict_buffer =
+        parent_->per_file_mem_pool_->TryAllocate(header.uncompressed_page_size);
+    if (UNLIKELY(dict_buffer == nullptr)) {
+      string details = (Substitute(PARQUET_MEM_LIMIT_EXCEEDED, "BaseColumnWriter::Flush",
+          header.uncompressed_page_size, "dictionary page"));
+      return parent_->per_file_mem_pool_->mem_tracker()->MemLimitExceeded(
+          parent_->state_, details, header.uncompressed_page_size);
+    }
     dict_encoder_base_->WriteDict(dict_buffer);
     if (compressor_.get() != nullptr) {
       SCOPED_TIMER(parent_->parent_->compress_timer());
@@ -769,7 +795,14 @@ Status HdfsParquetTableWriter::BaseColumnWriter::Flush(int64_t* file_pos,
           compressor_->MaxOutputLen(header.uncompressed_page_size);
       DCHECK_GT(max_compressed_size, 0);
       uint8_t* compressed_data =
-          parent_->per_file_mem_pool_->Allocate(max_compressed_size);
+          parent_->per_file_mem_pool_->TryAllocate(max_compressed_size);
+      if (UNLIKELY(compressed_data == nullptr)) {
+        string details =
+            (Substitute(PARQUET_MEM_LIMIT_EXCEEDED, "BaseColumnWriter::Flush",
+                max_compressed_size, "compressed dictionary page"));
+        return parent_->per_file_mem_pool_->mem_tracker()->MemLimitExceeded(
+            parent_->state_, details, max_compressed_size);
+      }
       header.compressed_page_size = max_compressed_size;
       const Status& status =
           compressor_->ProcessBlock32(true, header.uncompressed_page_size, dict_buffer,
@@ -854,7 +887,7 @@ Status HdfsParquetTableWriter::BaseColumnWriter::FinalizeCurrentPage() {
   // around a parquet MR bug (see IMPALA-759 for more details).
   if (current_page_->num_non_null == 0) current_encoding_ = parquet::Encoding::PLAIN;
 
-  if (current_encoding_ == parquet::Encoding::PLAIN_DICTIONARY) WriteDictDataPage();
+  if (IsDictionaryEncoding(current_encoding_)) WriteDictDataPage();
 
   parquet::PageHeader& header = current_page_->header;
   header.data_page_header.encoding = current_encoding_;
@@ -921,6 +954,7 @@ Status HdfsParquetTableWriter::BaseColumnWriter::FinalizeCurrentPage() {
 
   // Update row group statistics from page statistics.
   DCHECK(row_group_stats_base_ != nullptr);
+  RETURN_IF_ERROR(row_group_stats_base_->MaterializeStringValuesToInternalBuffers());
   row_group_stats_base_->Merge(*page_stats_base_);
 
   // Add the size of the data page header
@@ -965,11 +999,27 @@ HdfsParquetTableWriter::HdfsParquetTableWriter(HdfsTableSink* parent, RuntimeSta
     file_size_limit_(0),
     reusable_col_mem_pool_(new MemPool(parent_->mem_tracker())),
     per_file_mem_pool_(new MemPool(parent_->mem_tracker())),
-    row_idx_(0) {}
+    row_idx_(0) {
+  is_iceberg_file_ = table_desc->IsIcebergTable();
+}
 
 HdfsParquetTableWriter::~HdfsParquetTableWriter() {
 }
 
+void HdfsParquetTableWriter::ConfigureTimestampType() {
+  if (is_iceberg_file_) {
+    // The Iceberg spec states that timestamps are stored as INT64 micros.
+    timestamp_type_ = TParquetTimestampType::INT64_MICROS;
+    return;
+  }
+  timestamp_type_ = state_->query_options().parquet_timestamp_type;
+}
+
+void HdfsParquetTableWriter::ConfigureStringType() {
+  string_utf8_ = is_iceberg_file_ ||
+                 state_->query_options().parquet_annotate_strings_utf8;
+}
+
 Status HdfsParquetTableWriter::Init() {
   // Initialize file metadata
   file_metadata_.version = PARQUET_CURRENT_VERSION;
@@ -1034,6 +1084,9 @@ Status HdfsParquetTableWriter::Init() {
 
   Codec::CodecInfo codec_info(codec, clevel);
 
+  ConfigureTimestampType();
+  ConfigureStringType();
+
   columns_.resize(num_cols);
   // Initialize each column structure.
   for (int i = 0; i < columns_.size(); ++i) {
@@ -1071,7 +1124,7 @@ Status HdfsParquetTableWriter::Init() {
           new ColumnWriter<double>(this, output_expr_evals_[i], codec_info, col_name);
         break;
       case TYPE_TIMESTAMP:
-        switch (state_->query_options().parquet_timestamp_type) {
+        switch (timestamp_type_) {
           case TParquetTimestampType::INT96_NANOS:
             writer =
                 new ColumnWriter<TimestampValue>(
@@ -1139,13 +1192,18 @@ Status HdfsParquetTableWriter::CreateSchema() {
   file_metadata_.schema.resize(columns_.size() + 1);
   file_metadata_.schema[0].__set_num_children(columns_.size());
   file_metadata_.schema[0].name = "schema";
+  const int num_clustering_cols = table_desc_->num_clustering_cols();
 
   for (int i = 0; i < columns_.size(); ++i) {
     parquet::SchemaElement& col_schema = file_metadata_.schema[i + 1];
     const ColumnType& col_type = output_expr_evals_[i]->root().type();
     col_schema.name = columns_[i]->column_name();
-    ParquetMetadataUtils::FillSchemaElement(col_type, state_->query_options(),
-                                            &col_schema);
+    const ColumnDescriptor& col_desc = table_desc_->col_descs()[i + num_clustering_cols];
+    DCHECK_EQ(col_desc.name(), columns_[i]->column_name());
+    const int field_id = col_desc.field_id();
+    if (field_id != -1) col_schema.__set_field_id(field_id);
+    ParquetMetadataUtils::FillSchemaElement(col_type, string_utf8_,
+                                            timestamp_type_, &col_schema);
   }
 
   return Status::OK();
@@ -1161,7 +1219,7 @@ Status HdfsParquetTableWriter::AddRowGroup() {
   for (int i = 0; i < columns_.size(); ++i) {
     parquet::ColumnMetaData metadata;
     metadata.type = ParquetMetadataUtils::ConvertInternalToParquetType(
-        columns_[i]->type().type, state_->query_options());
+        columns_[i]->type().type, timestamp_type_);
     metadata.path_in_schema.push_back(columns_[i]->column_name());
     metadata.codec = columns_[i]->GetParquetCodec();
     current_row_group_->columns[i].__set_meta_data(metadata);
@@ -1262,7 +1320,7 @@ Status HdfsParquetTableWriter::AppendRows(
     }
     ++row_idx_;
     ++row_count_;
-    ++output_->num_rows;
+    ++output_->current_file_rows;
 
     if (file_size_estimate_ > file_size_limit_) {
       // This file is full.  We need a new file.
@@ -1279,6 +1337,7 @@ Status HdfsParquetTableWriter::AppendRows(
   // Reset the row_idx_ when we exhaust the batch.  We can exit before exhausting
   // the batch if we run out of file space and will continue from the last index.
   row_idx_ = 0;
+  RETURN_IF_ERROR(state_->CheckQueryState());
   return Status::OK();
 }
 
@@ -1301,9 +1360,14 @@ Status HdfsParquetTableWriter::Finalize() {
   RETURN_IF_ERROR(WriteFileFooter());
   *stats_.mutable_parquet_stats() = parquet_dml_stats_;
   COUNTER_ADD(parent_->rows_inserted_counter(), row_count_);
+  FinalizePartitionInfo();
   return Status::OK();
 }
 
+void HdfsParquetTableWriter::FinalizePartitionInfo() {
+  output_->current_file_bytes = file_pos_;
+}
+
 void HdfsParquetTableWriter::Close() {
   // Release all accumulated memory
   for (int i = 0; i < columns_.size(); ++i) {
@@ -1463,9 +1527,12 @@ Status HdfsParquetTableWriter::WriteFileFooter() {
   RETURN_IF_ERROR(thrift_serializer_->SerializeToBuffer(
       &file_metadata_, &file_metadata_len, &buffer));
   RETURN_IF_ERROR(Write(buffer, file_metadata_len));
+  file_pos_ += file_metadata_len;
 
   // Write footer
   RETURN_IF_ERROR(Write<uint32_t>(file_metadata_len));
+  file_pos_ += sizeof(uint32_t);
   RETURN_IF_ERROR(Write(PARQUET_VERSION_NUMBER, sizeof(PARQUET_VERSION_NUMBER)));
+  file_pos_ += sizeof(PARQUET_VERSION_NUMBER);
   return Status::OK();
 }
