@@ -24,8 +24,11 @@
 #include "exec/parquet/parquet-bool-decoder.h"
 #include "exec/parquet/parquet-level-decoder.h"
 #include "exec/parquet/parquet-metadata-utils.h"
+#include "exec/scratch-tuple-batch.h"
 #include "parquet-collection-column-reader.h"
 #include "runtime/runtime-state.h"
+#include "runtime/scoped-buffer.h"
+#include "runtime/string-value.inline.h"
 #include "runtime/tuple.h"
 #include "util/debug-util.h"
 #include "util/dict-encoding.h"
@@ -33,11 +36,6 @@
 
 #include "common/names.h"
 
-// Provide a workaround for IMPALA-1658.
-DEFINE_bool(convert_legacy_hive_parquet_utc_timestamps, false,
-    "When true, TIMESTAMPs read from files written by Parquet-MR (used by Hive) will "
-    "be converted from UTC to local time. Writes are unaffected.");
-
 using namespace impala::io;
 
 using parquet::Encoding;
@@ -320,14 +318,14 @@ Status ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::InitDataPag
   DCHECK(slot_desc_ == nullptr || slot_desc_->type().type != TYPE_BOOLEAN)
       << "Bool has specialized impl";
   page_encoding_ = col_chunk_reader_.encoding();
-  if (page_encoding_ != parquet::Encoding::PLAIN_DICTIONARY
+  if (!IsDictionaryEncoding(page_encoding_)
       && page_encoding_ != parquet::Encoding::PLAIN) {
     return GetUnsupportedDecodingError();
   }
 
-  // If slot_desc_ is NULL, we don't need so decode any values so dict_decoder_ does
+  // If slot_desc_ is NULL, we don't need to decode any values so dict_decoder_ does
   // not need to be initialized.
-  if (page_encoding_ == Encoding::PLAIN_DICTIONARY && slot_desc_ != nullptr) {
+  if (IsDictionaryEncoding(page_encoding_) && slot_desc_ != nullptr) {
     if (!dict_decoder_init_) {
       return Status("File corrupt. Missing dictionary page.");
     }
@@ -365,7 +363,7 @@ bool ScalarColumnReader<InternalType, PARQUET_TYPE,
   if (bool_decoder_) {
     return bool_decoder_->SkipValues(num_values);
   }
-  if (page_encoding_ == Encoding::PLAIN_DICTIONARY) {
+  if (IsDictionaryEncoding(page_encoding_)) {
     return dict_decoder_.SkipValues(num_values);
   } else {
     DCHECK_EQ(page_encoding_, Encoding::PLAIN);
@@ -391,7 +389,7 @@ bool ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::ReadValue(
   if (MATERIALIZED) {
     if (def_level_ >= max_def_level()) {
       bool continue_execution;
-      if (page_encoding_ == Encoding::PLAIN_DICTIONARY) {
+      if (IsDictionaryEncoding(page_encoding_)) {
         continue_execution = NeedsConversionInline() ?
             ReadSlot<Encoding::PLAIN_DICTIONARY, true>(tuple) :
             ReadSlot<Encoding::PLAIN_DICTIONARY, false>(tuple);
@@ -645,7 +643,7 @@ bool ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::MaterializeVa
     int max_values, int tuple_size, uint8_t* RESTRICT tuple_mem,
     int* RESTRICT num_values) RESTRICT {
   // Dispatch to the correct templated implementation of MaterializeValueBatch().
-  if (page_encoding_ == Encoding::PLAIN_DICTIONARY) {
+  if (IsDictionaryEncoding(page_encoding_)) {
     if (NeedsConversionInline()) {
       return MaterializeValueBatch<IN_COLLECTION, Encoding::PLAIN_DICTIONARY, true>(
           max_values, tuple_size, tuple_mem, num_values);
@@ -755,7 +753,7 @@ bool ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::DecodeValue(
     uint8_t** RESTRICT data, const uint8_t* RESTRICT data_end,
     InternalType* RESTRICT val) RESTRICT {
   DCHECK_EQ(page_encoding_, ENCODING);
-  if (ENCODING == Encoding::PLAIN_DICTIONARY) {
+  if (IsDictionaryEncoding(ENCODING)) {
     if (UNLIKELY(!dict_decoder_.GetNextValue(val))) {
       SetDictDecodeError();
       return false;
@@ -805,7 +803,7 @@ bool ScalarColumnReader<bool, parquet::Type::BOOLEAN, true>::DecodeValue(
 template <typename InternalType, parquet::Type::type PARQUET_TYPE, bool MATERIALIZED>
 bool ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::DecodeValues(
     int64_t stride, int64_t count, InternalType* RESTRICT out_vals) RESTRICT {
-  if (page_encoding_ == Encoding::PLAIN_DICTIONARY) {
+  if (IsDictionaryEncoding(page_encoding_)) {
     return DecodeValues<Encoding::PLAIN_DICTIONARY>(stride, count, out_vals);
   } else {
     DCHECK_EQ(page_encoding_, Encoding::PLAIN);
@@ -817,7 +815,7 @@ template <typename InternalType, parquet::Type::type PARQUET_TYPE, bool MATERIAL
 template <Encoding::type ENCODING>
 bool ScalarColumnReader<InternalType, PARQUET_TYPE, MATERIALIZED>::DecodeValues(
     int64_t stride, int64_t count, InternalType* RESTRICT out_vals) RESTRICT {
-  if (page_encoding_ == Encoding::PLAIN_DICTIONARY) {
+  if (IsDictionaryEncoding(page_encoding_)) {
     if (UNLIKELY(!dict_decoder_.GetNextValues(out_vals, stride, count))) {
       SetDictDecodeError();
       return false;
@@ -917,7 +915,6 @@ template <>
 bool ScalarColumnReader<TimestampValue, parquet::Type::INT96, true>::ConvertSlot(
     const TimestampValue* src, void* slot) {
   // Conversion should only happen when this flag is enabled.
-  DCHECK(FLAGS_convert_legacy_hive_parquet_utc_timestamps);
   DCHECK(timestamp_decoder_.NeedsConversion());
   TimestampValue* dst_ts = reinterpret_cast<TimestampValue*>(slot);
   *dst_ts = *src;
@@ -1254,6 +1251,7 @@ Status BaseScalarColumnReader::StartPageFiltering() {
 
 bool BaseScalarColumnReader::SkipTopLevelRows(int64_t num_rows) {
   DCHECK_GE(num_buffered_values_, num_rows);
+  DCHECK_GT(num_rows, 0);
   // Fastest path: field is required and not nested.
   // So row count equals value count, and every value is stored in the page data.
   if (max_def_level() == 0 && max_rep_level() == 0) {
@@ -1289,10 +1287,17 @@ bool BaseScalarColumnReader::SkipTopLevelRows(int64_t num_rows) {
   } else {
     // 'rep_level_' being zero denotes the start of a new top-level row.
     // From the 'def_level_' we can determine the number of non-NULL values.
-    while (!(num_rows == 0 && rep_levels_.PeekLevel() == 0)) {
-      def_level_ = def_levels_.ReadLevel();
-      rep_level_ = rep_levels_.ReadLevel();
+    while (true) {
+      if (!def_levels_.CacheNextBatchIfEmpty(num_buffered_values_).ok()) return false;
+      if (!rep_levels_.CacheNextBatchIfEmpty(num_buffered_values_).ok()) return false;
+      if (num_rows == 0 && rep_levels_.PeekLevel() == 0) {
+        // No more rows to skip, and the next value belongs to a new top-level row.
+        break;
+      }
+      def_level_ = def_levels_.CacheGetNext();
+      rep_level_ = rep_levels_.CacheGetNext();
       --num_buffered_values_;
+      DCHECK_GE(num_buffered_values_, 0);
       if (def_level_ >= max_def_level()) ++num_values_to_skip;
       if (rep_level_ == 0) {
         ++current_row_;
