@@ -94,6 +94,7 @@ static bool IsEncodingSupported(parquet::Encoding::type e) {
     case parquet::Encoding::PLAIN_DICTIONARY:
     case parquet::Encoding::BIT_PACKED:
     case parquet::Encoding::RLE:
+    case parquet::Encoding::RLE_DICTIONARY:
       return true;
     default:
       return false;
@@ -151,12 +152,12 @@ void SetDecimalConvertedAndLogicalType(
 /// normalization, and older readers that do not use logical types would incorrectly
 /// interpret TIMESTAMP_MILLIS/MICROS as UTC normalized.
 /// Leaves logical type empty for int96 timestamps.
-void SetTimestampLogicalType(const TQueryOptions& query_options,
+void SetTimestampLogicalType(TParquetTimestampType::type parquet_timestamp_type,
     parquet::SchemaElement* col_schema) {
-  if(query_options.parquet_timestamp_type == TParquetTimestampType::INT96_NANOS) return;
+  if (parquet_timestamp_type == TParquetTimestampType::INT96_NANOS) return;
 
   parquet::TimeUnit time_unit;
-  switch (query_options.parquet_timestamp_type) {
+  switch (parquet_timestamp_type) {
     case TParquetTimestampType::INT64_MILLIS:
       time_unit.__set_MILLIS(parquet::MilliSeconds());
       break;
@@ -245,7 +246,7 @@ Status ParquetMetadataUtils::ValidateColumnOffsets(const string& filename,
     }
     int64_t col_len = col_chunk.meta_data.total_compressed_size;
     int64_t col_end = col_start + col_len;
-    if (col_end <= 0 || col_end > file_length) {
+    if (col_end <= 0 || col_end >= file_length) {
       return Status(Substitute("Parquet file '$0': metadata is corrupt. Column $1 has "
           "invalid column offsets (offset=$2, size=$3, file_size=$4).", filename, i,
           col_start, col_len, file_length));
@@ -338,10 +339,10 @@ Status ParquetMetadataUtils::ValidateColumn(const char* filename,
       }
 
       int expected_len = ParquetPlainEncoder::DecimalSize(slot_desc->type());
-      if (schema_element.type_length != expected_len) {
+      if (schema_element.type_length < expected_len) {
         return Status(Substitute("File '$0' column '$1' has an invalid type length. "
-            "Expecting: $2 len in file: $3", filename, schema_element.name, expected_len,
-            schema_element.type_length));
+            "Expecting: len >= $2 in file: $3", filename, schema_element.name,
+            expected_len, schema_element.type_length));
       }
     }
     if (!schema_element.__isset.scale) {
@@ -386,19 +387,19 @@ Status ParquetMetadataUtils::ValidateColumn(const char* filename,
 }
 
 parquet::Type::type ParquetMetadataUtils::ConvertInternalToParquetType(
-    PrimitiveType type, const TQueryOptions& query_options) {
+    PrimitiveType type, TParquetTimestampType::type timestamp_type) {
   DCHECK_GE(type, 0);
   DCHECK_LT(type, INTERNAL_TO_PARQUET_TYPES_SIZE);
-  if (type == TYPE_TIMESTAMP &&
-      query_options.parquet_timestamp_type != TParquetTimestampType::INT96_NANOS) {
+  if (type == TYPE_TIMESTAMP && timestamp_type != TParquetTimestampType::INT96_NANOS) {
     return parquet::Type::INT64;
   }
   return INTERNAL_TO_PARQUET_TYPES[type];
 }
 
 void ParquetMetadataUtils::FillSchemaElement(const ColumnType& col_type,
-    const TQueryOptions& query_options, parquet::SchemaElement* col_schema) {
-  col_schema->__set_type(ConvertInternalToParquetType(col_type.type, query_options));
+    bool string_utf8, TParquetTimestampType::type timestamp_type,
+    parquet::SchemaElement* col_schema) {
+  col_schema->__set_type(ConvertInternalToParquetType(col_type.type, timestamp_type));
   col_schema->__set_repetition_type(parquet::FieldRepetitionType::OPTIONAL);
 
   switch (col_type.type) {
@@ -412,7 +413,7 @@ void ParquetMetadataUtils::FillSchemaElement(const ColumnType& col_type,
     case TYPE_STRING:
       // By default STRING has no logical type, see IMPALA-5982.
       // VARCHAR and CHAR are always set to UTF8.
-      if (query_options.parquet_annotate_strings_utf8) {
+      if (string_utf8) {
         SetUtf8ConvertedAndLogicalType(col_schema);
       }
       break;
@@ -433,7 +434,7 @@ void ParquetMetadataUtils::FillSchemaElement(const ColumnType& col_type,
       SetIntLogicalType(64, col_schema);
       break;
     case TYPE_TIMESTAMP:
-      SetTimestampLogicalType(query_options, col_schema);
+      SetTimestampLogicalType(timestamp_type, col_schema);
       break;
     case TYPE_DATE:
       SetDateLogicalType(col_schema);
@@ -743,6 +744,31 @@ SchemaNode* ParquetSchemaResolver::NextSchemaNode(
         file_idx = table_idx;
       }
     }
+  } else if (fallback_schema_resolution_ ==
+        TParquetFallbackSchemaResolution::type::FIELD_ID) {
+    // Resolution by field id for Iceberg table.
+    if (next_idx == 0) {
+      // Resolve top-level table column by field id.
+      DCHECK_LT(table_idx, tbl_desc_.col_descs().size());
+      const int& field_id = tbl_desc_.col_descs()[table_idx].field_id();
+      file_idx = FindChildWithFieldId(node, field_id);
+    } else if (col_type->type == TYPE_STRUCT) {
+      // Resolve struct field by field id.
+      DCHECK_LT(table_idx, col_type->field_ids.size());
+      const int& field_id = col_type->field_ids[table_idx];
+      file_idx = FindChildWithFieldId(node, field_id);
+    } else if (col_type->type == TYPE_ARRAY) {
+      // Arrays have only one child in the file.
+      DCHECK_EQ(table_idx, SchemaPathConstants::ARRAY_ITEM);
+      file_idx = table_idx;
+    } else {
+      DCHECK_EQ(col_type->type, TYPE_MAP);
+      DCHECK(table_idx == SchemaPathConstants::MAP_KEY ||
+             table_idx == SchemaPathConstants::MAP_VALUE);
+      // At this point we've found a MAP with a matching field id. It's safe to resolve
+      // the child (key or value) by position.
+      file_idx = table_idx;
+    }
   } else {
     // Resolution by position.
     DCHECK_EQ(fallback_schema_resolution_,
@@ -781,6 +807,15 @@ int ParquetSchemaResolver::FindChildWithName(SchemaNode* node,
   return idx;
 }
 
+int ParquetSchemaResolver::FindChildWithFieldId(SchemaNode* node,
+    const int& field_id) const {
+  int idx;
+  for (idx = 0; idx < node->children.size(); ++idx) {
+    if (node->children[idx].element->field_id == field_id) break;
+  }
+  return idx;
+}
+
 // There are three types of array encodings:
 //
 // 1. One-level encoding
