@@ -24,6 +24,8 @@
 #include <vector>
 #include <boost/scoped_ptr.hpp>
 
+#include "common/atomic.h"
+#include "codegen/codegen-fn-ptr.h"
 #include "common/object-pool.h"
 #include "common/status.h"
 #include "exec/filter-context.h"
@@ -37,21 +39,25 @@
 
 namespace impala {
 
+class CyclicBarrier;
 class PhjBuilder;
+class PhjBuilderPartition;
 class RowDescriptor;
 class RuntimeState;
 class ScalarExpr;
 class ScalarExprEvaluator;
 
+/// Method signature of the codegened version of InsertBatch().
+typedef bool (*InsertBatchFn)(PhjBuilderPartition*, TPrefetchMode::type, HashTableCtx*,
+    RowBatch*, const std::vector<BufferedTupleStream::FlatRowPtr>&, Status*);
+
 /// Partitioned Hash Join Builder Config class. This has a few extra methods to be used
 /// directly by the PartitionedHashJoinPlanNode. Since it is expected to only be created
 /// and used by PartitionedHashJoinPlanNode only, the DataSinkConfig::Init() and
 /// DataSinkConfig::CreateSink() are not implemented for it.
 class PhjBuilderConfig : public JoinBuilderConfig {
  public:
-  DataSink* CreateSink(const TPlanFragmentCtx& fragment_ctx,
-    const TPlanFragmentInstanceCtx& fragment_instance_ctx,
-    RuntimeState* state) const override;
+  DataSink* CreateSink(RuntimeState* state) const override;
 
   /// Creates a PhjBuilder for embedded use within a PartitionedHashJoinNode.
   PhjBuilder* CreateSink(BufferPool::ClientHandle* buffer_pool_client,
@@ -61,15 +67,14 @@ class PhjBuilderConfig : public JoinBuilderConfig {
   /// Creates a PhjBuilderConfig for embedded use within a PartitionedHashJoinNode.
   /// Creates the object in the state's object pool. To be used only by
   /// PartitionedHashJoinPlanNode.
-  static Status CreateConfig(RuntimeState* state, int join_node_id, TJoinOp::type join_op,
-      const RowDescriptor* build_row_desc,
+  static Status CreateConfig(FragmentState* state, int join_node_id,
+      TJoinOp::type join_op, const RowDescriptor* build_row_desc,
       const std::vector<TEqJoinCondition>& eq_join_conjuncts,
       const std::vector<TRuntimeFilterDesc>& filters, uint32_t hash_seed,
       PhjBuilderConfig** sink);
 
   void Close() override;
-
-  void Codegen(RuntimeState* state, RuntimeProfile* profile);
+  void Codegen(FragmentState* state) override;
 
   ~PhjBuilderConfig() override {}
 
@@ -104,24 +109,26 @@ class PhjBuilderConfig : public JoinBuilderConfig {
   /// used for subsequent levels.
   typedef Status (*ProcessBuildBatchFn)(
       PhjBuilder*, RowBatch*, HashTableCtx*, bool build_filters, bool is_null_aware);
-  /// Jitted ProcessBuildBatch function pointers.  NULL if codegen is disabled.
-  ProcessBuildBatchFn process_build_batch_fn_ = nullptr;
-  ProcessBuildBatchFn process_build_batch_fn_level0_ = nullptr;
+  /// Jitted ProcessBuildBatch function pointers. NULL if codegen is disabled.
+  CodegenFnPtr<ProcessBuildBatchFn> process_build_batch_fn_;
+  CodegenFnPtr<ProcessBuildBatchFn> process_build_batch_fn_level0_;
 
   /// Jitted Partition::InsertBatch() function pointers. NULL if codegen is disabled.
-  void* insert_batch_fn_ = nullptr;
-  void* insert_batch_fn_level0_ = nullptr;
+
+  /// Method signature of the codegened version of Partition::InsertBatch().
+  CodegenFnPtr<InsertBatchFn> insert_batch_fn_;
+  CodegenFnPtr<InsertBatchFn> insert_batch_fn_level0_;
 
  protected:
   /// Initialization for separate sink.
   Status Init(const TDataSink& tsink, const RowDescriptor* input_row_desc,
-      RuntimeState* state) override;
+      FragmentState* state) override;
 
  private:
   /// Helper method used by CreateConfig() to initialize embedded builder.
   /// 'tsink' does not need to be initialized by the caller - all values to be used are
   /// passed in as arguments and this function fills in required fields in 'tsink'.
-  Status Init(RuntimeState* state, int join_node_id, TJoinOp::type join_op,
+  Status Init(FragmentState* state, int join_node_id, TJoinOp::type join_op,
       const RowDescriptor* build_row_desc,
       const std::vector<TEqJoinCondition>& eq_join_conjuncts,
       const std::vector<TRuntimeFilterDesc>& filters, uint32_t hash_seed,
@@ -130,7 +137,7 @@ class PhjBuilderConfig : public JoinBuilderConfig {
   /// Initializes the build and filter expressions, creates a copy of the filter
   /// descriptors that will be generated by this sink and initializes the hash table
   /// config object.
-  Status InitExprsAndFilters(RuntimeState* state,
+  Status InitExprsAndFilters(FragmentState* state,
       const std::vector<TEqJoinCondition>& eq_join_conjuncts,
       const std::vector<TRuntimeFilterDesc>& filters);
 
@@ -148,8 +155,8 @@ class PhjBuilderConfig : public JoinBuilderConfig {
 
   /// Codegen inserting rows into runtime filters. Identical signature to
   /// InsertRuntimeFilters(). Returns non-OK if codegen was not possible.
-  Status CodegenInsertRuntimeFilters(
-      LlvmCodeGen* codegen, const vector<ScalarExpr*>& filter_exprs, llvm::Function** fn);
+  Status CodegenInsertRuntimeFilters(LlvmCodeGen* codegen,
+      const std::vector<ScalarExpr*>& filter_exprs, llvm::Function** fn);
 };
 
 /// See partitioned-hash-join-node.h for explanation of the top-level algorithm and how
@@ -177,6 +184,110 @@ enum class HashJoinState {
   REPARTITIONING_PROBE,
 };
 
+/// A partition containing a subset of build rows.
+///
+/// A partition may contain two data structures: the build rows and optionally a hash
+/// table built over the build rows.  Building the hash table requires all build rows to
+/// be pinned in memory. The build rows are kept in memory if all partitions fit in
+/// memory, but can be unpinned and spilled to disk to free up memory. Reading or
+/// writing the unpinned rows requires a single read or write buffer. If the unpinned
+/// rows are not being read or written, they can be completely unpinned, requiring no
+/// buffers.
+///
+/// The build input is first partitioned by hash function level 0 into the level 0
+/// partitions. Then, if the join spills and the size of a level n partition is too
+/// large to fit in memory, the partition's rows can be repartitioned with the level
+/// n + 1 hash function into the level n + 1 partitions.
+class PhjBuilderPartition {
+ public:
+  PhjBuilderPartition(RuntimeState* state, PhjBuilder* parent, int level);
+  ~PhjBuilderPartition();
+
+  using PartitionId = int;
+
+  /// Close the partition and attach resources to 'batch' if non-NULL or free the
+  /// resources if 'batch' is NULL. Idempotent.
+  void Close(RowBatch* batch);
+
+  /// Returns the estimated byte size of the in-memory data structures for this
+  /// partition. This includes all build rows and the hash table.
+  int64_t EstimatedInMemSize() const;
+
+  /// Pins the build tuples for this partition and constructs the hash table from it.
+  /// Build rows cannot be added after calling this. If the build rows could not be
+  /// pinned or the hash table could not be built due to memory pressure, sets *built
+  /// to false and returns OK. Returns an error status if any other error is
+  /// encountered.
+  Status BuildHashTable(bool* built) WARN_UNUSED_RESULT;
+
+  /// Spills this partition, the partition's stream is unpinned with 'mode' and
+  /// its hash table is destroyed if it was built. Calling with 'mode' UNPIN_ALL
+  /// unpins all pages and frees all buffers associated with the partition so that
+  /// the partition does not use any reservation. Calling with 'mode'
+  /// UNPIN_ALL_EXCEPT_CURRENT may leave the read or write pages of the unpinned stream
+  /// pinned and therefore using reservation. If the partition was previously
+  /// spilled with mode UNPIN_ALL_EXCEPT_CURRENT, then calling Spill() again with
+  /// UNPIN_ALL may release more reservation by unpinning the read or write page
+  /// in the stream.
+  Status Spill(BufferedTupleStream::UnpinMode mode) WARN_UNUSED_RESULT;
+
+  std::string DebugString();
+
+  bool ALWAYS_INLINE IsClosed() const { return build_rows_ == nullptr; }
+  BufferedTupleStream* ALWAYS_INLINE build_rows() { return build_rows_.get(); }
+  HashTable* ALWAYS_INLINE hash_tbl() const { return hash_tbl_.get(); }
+  PartitionId id() const { return id_; }
+  bool ALWAYS_INLINE is_spilled() const { return is_spilled_; }
+  int ALWAYS_INLINE level() const { return level_; }
+  /// Return true if the partition can be spilled - is not closed and is not spilled.
+  bool CanSpill() const { return !IsClosed() && !is_spilled(); }
+  int64_t num_spilled_probe_rows() const { return num_spilled_probe_rows_.Load(); }
+
+  /// Increment the number of spilled probe rows. Thread-safe.
+  void IncrementNumSpilledProbeRows(int64_t count) {
+    num_spilled_probe_rows_.Add(count);
+  }
+
+ private:
+  /// Inserts each row in 'batch' into 'hash_tbl_' using 'ctx'. 'flat_rows' is an array
+  /// containing the rows in the hash table's tuple stream.
+  /// 'prefetch_mode' is the prefetching mode in use. If it's not PREFETCH_NONE, hash
+  /// table buckets which the rows hashes to will be prefetched. This parameter is
+  /// replaced with a constant during codegen time. This function may be replaced with
+  /// a codegen'd version. Returns true if all rows in 'batch' are successfully
+  /// inserted and false otherwise. If inserting failed, 'status' indicates why it
+  /// failed: if 'status' is ok, inserting failed because not enough reservation
+  /// was available and if 'status' is an error, inserting failed because of that error.
+  bool InsertBatch(TPrefetchMode::type prefetch_mode, HashTableCtx* ctx,
+      RowBatch* batch, const std::vector<BufferedTupleStream::FlatRowPtr>& flat_rows,
+      Status* status);
+
+  const PhjBuilder* parent_;
+
+  /// Id for this partition that is unique within the builder.
+  const PartitionId id_;
+
+  /// True if this partition is spilled.
+  bool is_spilled_;
+
+  /// How many times rows in this partition have been repartitioned. Partitions created
+  /// from the node's children's input is level 0, 1 after the first repartitioning,
+  /// etc.
+  const int level_;
+
+  /// The hash table for this partition.
+  boost::scoped_ptr<HashTable> hash_tbl_;
+
+  /// Stream of build tuples in this partition. Initially owned by this object but
+  /// transferred to the parent exec node (via the row batch) when the partition
+  /// is closed. If NULL, ownership has been transferred and the partition is closed.
+  std::unique_ptr<BufferedTupleStream> build_rows_;
+
+  /// The number of spilled probe rows associated with this partition. Updated in
+  /// DoneProbingHashPartitions().
+  AtomicInt64 num_spilled_probe_rows_{0};
+};
+
 /// The build side for the PartitionedHashJoinNode. Build-side rows are hash-partitioned
 /// into PARTITION_FANOUT partitions, with partitions spilled if the full build side
 /// does not fit in memory. Spilled partitions can be repartitioned with a different
@@ -207,8 +318,32 @@ enum class HashJoinState {
 /// be a different client.
 ///
 /// The full hash join algorithm is documented in PartitionedHashJoinNode.
+///
+/// Shared Build
+/// ------------
+/// A separate builder can be shared between multiple PartitionedHashJoinNodes. The
+/// spilling hash join algorithm mutates the state of the builder between phases, so
+/// requires synchronization between the probe threads executing PartitionedHashJoinNode
+/// that are reading that state.
+///
+/// The algorithm (specifically the HashJoinState state machine) is executed in lockstep
+/// across all probe threads with each probe thread working on the same set of partitions
+/// at the same time. A CyclicBarrier, 'probe_barrier_', is used for synchronization.
+/// At each state transition where the builder state needs to be mutated, all probe
+/// threads must arrive at the barrier before proceeding. The state transition is executed
+/// serially by a single thread before all threads proceed. All probe threads go through
+/// the same state transitions in lockstep, even if they have no work to do. E.g. if a
+/// probe thread has zero rows remaining in its spilled partitions, it still needs to
+/// wait for the other probe threads.
+///
+/// Not all join ops can be used with a shared build. For example, RIGHT_OUTER_JOIN is
+/// not supported currently, in part because it mutates the hash table during probing to
+/// track matches, but also because hash table matches would need to be broadcast across
+/// all instances within the query, not just the backend.
 class PhjBuilder : public JoinBuilder {
  public:
+  friend class PhjBuilderPartition;
+
   /// Number of initial partitions to create. Must be a power of two.
   static const int PARTITION_FANOUT = 16;
 
@@ -226,8 +361,6 @@ class PhjBuilder : public JoinBuilder {
   /// TODO: we can revisit and try harder to explicitly detect skew.
   static const int MAX_PARTITION_DEPTH = 16;
 
-  class Partition;
-
   using PartitionId = int;
 
   // Constructor for separate join build.
@@ -248,11 +381,6 @@ class PhjBuilder : public JoinBuilder {
   virtual Status FlushFinal(RuntimeState* state) override;
   virtual void Close(RuntimeState* state) override;
 
-  /// Does all codegen for the builder (if codegen is enabled).
-  /// Updates the the builder's runtime profile with info about whether any errors
-  /// occured during codegen.
-  virtual void Codegen(RuntimeState* state) override;
-
   /////////////////////////////////////////
   // The following functions are used only by PartitionedHashJoinNode.
   /////////////////////////////////////////
@@ -265,7 +393,7 @@ class PhjBuilder : public JoinBuilder {
   struct HashPartitions {
     HashPartitions() { Reset(); }
     HashPartitions(int level,
-        const std::vector<std::unique_ptr<Partition>>* hash_partitions,
+        const std::vector<std::unique_ptr<PhjBuilderPartition>>* hash_partitions,
         bool non_empty_build)
       : level(level),
         hash_partitions(hash_partitions),
@@ -284,7 +412,7 @@ class PhjBuilder : public JoinBuilder {
     // The current set of hash partitions. Always contains PARTITION_FANOUT partitions.
     // The partitions may be in-memory, spilled, or closed. Valid until
     // DoneProbingHashPartitions() is called.
-    const std::vector<std::unique_ptr<Partition>>* hash_partitions;
+    const std::vector<std::unique_ptr<PhjBuilderPartition>>* hash_partitions;
 
     // True iff the build side had at least one row in a partition.
     bool non_empty_build;
@@ -312,9 +440,13 @@ class PhjBuilder : public JoinBuilder {
   /// PARTITION_FANOUT new partitions with level input_partition->level() + 1. The
   /// previous hash partitions must have been cleared with DoneProbingHashPartitions().
   /// The new hash partitions are returned in 'new_partitions'.
-  /// TODO: IMPALA-9156: this will be a synchronization point for shared join build.
-  Status BeginSpilledProbe(BufferPool::ClientHandle* probe_client, bool* repartitioned,
-      Partition** input_partition, HashPartitions* new_partitions);
+  ///
+  /// This is a synchronization point for shared join build. The time elapsed during the
+  /// serial execution phase is attributed to the builder. All probe threads must call
+  /// this function before continuing the next phase of the hash join algorithm.
+  Status BeginSpilledProbe(BufferPool::ClientHandle* probe_client,
+      RuntimeProfile* probe_profile, bool* repartitioned,
+      PhjBuilderPartition** input_partition, HashPartitions* new_partitions);
 
   /// Called after probing of the hash partitions returned by BeginInitialProbe() or
   /// BeginSpilledProbe() (when *repartitioned is true) is complete, i.e. all of the
@@ -330,10 +462,14 @@ class PhjBuilder : public JoinBuilder {
   /// transferred back to the builder.
   ///
   /// Returns an error if an error was encountered or if the query was cancelled.
-  /// TODO: IMPALA-9156: this will be a synchronization point for shared join build.
+  ///
+  /// This is a synchronization point for shared join build. The time elapsed during the
+  /// serial execution phase is attributed to the builder. All probe threads must call
+  /// this function before continuing the next phase of the hash join algorithm.
   Status DoneProbingHashPartitions(const int64_t num_spilled_probe_rows[PARTITION_FANOUT],
-      BufferPool::ClientHandle* probe_client,
-      std::deque<std::unique_ptr<Partition>>* output_partitions, RowBatch* batch);
+      BufferPool::ClientHandle* probe_client, RuntimeProfile* probe_profile,
+      std::deque<std::unique_ptr<PhjBuilderPartition>>* output_partitions,
+      RowBatch* batch);
 
   /// Called after probing of a single spilled partition returned by
   /// BeginSpilledProbe() when *repartitioned is false.
@@ -351,21 +487,31 @@ class PhjBuilder : public JoinBuilder {
   /// transferred back to the builder.
   ///
   /// Returns an error if an error was encountered or if the query was cancelled.
-  /// TODO: IMPALA-9156: this will be a synchronization point for shared join build.
+  ///
+  /// This is a synchronization point for shared join build. The time elapsed during the
+  /// serial execution phase is attributed to the builder. All probe threads must call
+  /// this function before continuing the next phase of the hash join algorithm.
   Status DoneProbingSinglePartition(BufferPool::ClientHandle* probe_client,
-      std::deque<std::unique_ptr<Partition>>* output_partitions, RowBatch* batch);
+      RuntimeProfile* probe_profile,
+      std::deque<std::unique_ptr<PhjBuilderPartition>>* output_partitions,
+      RowBatch* batch);
 
-  /// Close the null aware partition (if there is one) and set it to NULL.
-  /// TODO: IMPALA-9176: improve the encapsulation of the null-aware partition.
-  void CloseNullAwarePartition() {
-    if (null_aware_partition_ != nullptr) {
-      // We don't need to pass in a batch because the anti-join only returns tuple data
-      // from the probe side - i.e. the RowDescriptor for PartitionedHashJoinNode does
-      // not include the build tuple.
-      null_aware_partition_->Close(nullptr);
-      null_aware_partition_.reset();
-    }
-  }
+  /// Called to begin probing of the null-aware partition, after all other partitions
+  /// have been fully processed. This should only be called if there are build rows in the
+  /// null-aware partition. This pins the null-aware build rows in memory and allows all
+  /// probe threads to access those rows in a read-only manner.
+  ///
+  /// Returns an error if an error was encountered or if the query was cancelled.
+  ///
+  /// This is a synchronization point for shared join build. All probe threads must
+  /// call this function before continuing the next phase of the hash join algorithm.
+  Status BeginNullAwareProbe();
+
+  /// Called after probing of the null-aware build partition is complete.
+  ///
+  /// This is a synchronization point for shared join build. All probe threads must
+  /// call this function before continuing the next phase of the hash join algorithm.
+  Status DoneProbingNullAwarePartition();
 
   /// True if the hash table may contain rows with one or more NULL join keys. This
   /// depends on the join type, passed in via 'join_op' and the 'is_not_distinct_from'
@@ -373,10 +519,6 @@ class PhjBuilder : public JoinBuilder {
   static bool HashTableStoresNulls(
       TJoinOp::type join_op, const std::vector<bool>& is_not_distinct_from);
 
-  /// TODO: IMPALA-9156: document thread safety for accessing this from
-  /// multiple PartitionedHashJoinNodes.
-  void AddHashTableStatsToProfile(RuntimeProfile* profile);
-
   /// Returns 'bytes' of reservation to the builder from 'probe_client'.
   /// Called by the probe side to return surplus reservation. This is usually handled by
   /// the above methods, but if an error occured during execution, the probe may still
@@ -384,122 +526,21 @@ class PhjBuilder : public JoinBuilder {
   /// Must only be called if this is a separate build.
   void ReturnReservation(BufferPool::ClientHandle* probe_client, int64_t bytes);
 
-  /// TODO: IMPALA-9156: document thread safety for accessing this from
-  /// multiple PartitionedHashJoinNodes.
+  /// Safe to call from PartitionedHashJoinNode threads during the probe phase.
   HashJoinState state() const { return state_; }
 
-  /// Accessor to allow PartitionedHashJoinNode to access null_aware_partition_.
-  /// TODO: IMPALA-9176: improve the encapsulation of the null-aware partition.
-  inline Partition* null_aware_partition() const { return null_aware_partition_.get(); }
+  /// Accessor to allow PartitionedHashJoinNode to access 'null_aware_partition_'.
+  /// Generally the PartitionedHashJoinNode should only access this partition in
+  /// a read-only manner.
+  inline PhjBuilderPartition* null_aware_partition() const {
+    return null_aware_partition_.get();
+  }
 
-  /// TODO: IMPALA-9156: document thread safety for accessing this from
-  /// multiple PartitionedHashJoinNodes.
+  /// Thread-safe.
   HashTableStatsProfile* ht_stats_profile() const { return ht_stats_profile_.get(); }
 
   std::string DebugString() const;
 
-  /// A partition containing a subset of build rows.
-  ///
-  /// A partition may contain two data structures: the build rows and optionally a hash
-  /// table built over the build rows.  Building the hash table requires all build rows to
-  /// be pinned in memory. The build rows are kept in memory if all partitions fit in
-  /// memory, but can be unpinned and spilled to disk to free up memory. Reading or
-  /// writing the unpinned rows requires a single read or write buffer. If the unpinned
-  /// rows are not being read or written, they can be completely unpinned, requiring no
-  /// buffers.
-  ///
-  /// The build input is first partitioned by hash function level 0 into the level 0
-  /// partitions. Then, if the join spills and the size of a level n partition is too
-  /// large to fit in memory, the partition's rows can be repartitioned with the level
-  /// n + 1 hash function into the level n + 1 partitions.
-  class Partition {
-   public:
-    Partition(RuntimeState* state, PhjBuilder* parent, int level);
-    ~Partition();
-
-    /// Close the partition and attach resources to 'batch' if non-NULL or free the
-    /// resources if 'batch' is NULL. Idempotent.
-    void Close(RowBatch* batch);
-
-    /// Returns the estimated byte size of the in-memory data structures for this
-    /// partition. This includes all build rows and the hash table.
-    int64_t EstimatedInMemSize() const;
-
-    /// Pins the build tuples for this partition and constructs the hash table from it.
-    /// Build rows cannot be added after calling this. If the build rows could not be
-    /// pinned or the hash table could not be built due to memory pressure, sets *built
-    /// to false and returns OK. Returns an error status if any other error is
-    /// encountered.
-    Status BuildHashTable(bool* built) WARN_UNUSED_RESULT;
-
-    /// Spills this partition, the partition's stream is unpinned with 'mode' and
-    /// its hash table is destroyed if it was built. Calling with 'mode' UNPIN_ALL
-    /// unpins all pages and frees all buffers associated with the partition so that
-    /// the partition does not use any reservation. Calling with 'mode'
-    /// UNPIN_ALL_EXCEPT_CURRENT may leave the read or write pages of the unpinned stream
-    /// pinned and therefore using reservation. If the partition was previously
-    /// spilled with mode UNPIN_ALL_EXCEPT_CURRENT, then calling Spill() again with
-    /// UNPIN_ALL may release more reservation by unpinning the read or write page
-    /// in the stream.
-    Status Spill(BufferedTupleStream::UnpinMode mode) WARN_UNUSED_RESULT;
-
-    std::string DebugString();
-
-    bool ALWAYS_INLINE IsClosed() const { return build_rows_ == nullptr; }
-    BufferedTupleStream* ALWAYS_INLINE build_rows() { return build_rows_.get(); }
-    HashTable* ALWAYS_INLINE hash_tbl() const { return hash_tbl_.get(); }
-    PartitionId id() const { return id_; }
-    bool ALWAYS_INLINE is_spilled() const { return is_spilled_; }
-    int ALWAYS_INLINE level() const { return level_; }
-    /// Return true if the partition can be spilled - is not closed and is not spilled.
-    bool CanSpill() const { return !IsClosed() && !is_spilled(); }
-    int64_t num_spilled_probe_rows() const { return num_spilled_probe_rows_; }
-    void IncrementNumSpilledProbeRows(int64_t count) { num_spilled_probe_rows_ += count; }
-
-    /// Method signature of the codegened version of InsertBatch().
-    typedef bool (*InsertBatchFn)(Partition*, TPrefetchMode::type, HashTableCtx*,
-        RowBatch*, const std::vector<BufferedTupleStream::FlatRowPtr>&, Status*);
-
-   private:
-    /// Inserts each row in 'batch' into 'hash_tbl_' using 'ctx'. 'flat_rows' is an array
-    /// containing the rows in the hash table's tuple stream.
-    /// 'prefetch_mode' is the prefetching mode in use. If it's not PREFETCH_NONE, hash
-    /// table buckets which the rows hashes to will be prefetched. This parameter is
-    /// replaced with a constant during codegen time. This function may be replaced with
-    /// a codegen'd version. Returns true if all rows in 'batch' are successfully
-    /// inserted and false otherwise. If inserting failed, 'status' indicates why it
-    /// failed: if 'status' is ok, inserting failed because not enough reservation
-    /// was available and if 'status' is an error, inserting failed because of that error.
-    bool InsertBatch(TPrefetchMode::type prefetch_mode, HashTableCtx* ctx,
-        RowBatch* batch, const std::vector<BufferedTupleStream::FlatRowPtr>& flat_rows,
-        Status* status);
-
-    const PhjBuilder* parent_;
-
-    /// Id for this partition that is unique within the builder.
-    const PartitionId id_;
-
-    /// True if this partition is spilled.
-    bool is_spilled_;
-
-    /// How many times rows in this partition have been repartitioned. Partitions created
-    /// from the node's children's input is level 0, 1 after the first repartitioning,
-    /// etc.
-    const int level_;
-
-    /// The hash table for this partition.
-    boost::scoped_ptr<HashTable> hash_tbl_;
-
-    /// Stream of build tuples in this partition. Initially owned by this object but
-    /// transferred to the parent exec node (via the row batch) when the partition
-    /// is closed. If NULL, ownership has been transferred and the partition is closed.
-    std::unique_ptr<BufferedTupleStream> build_rows_;
-
-    /// The number of spilled probe rows associated with this partition. Updated in
-    /// DoneProbingHashPartitions().
-    int64_t num_spilled_probe_rows_ = 0;
-  };
-
   /// Computes the minimum reservation required to execute the spilling partitioned
   /// hash algorithm successfully for any input size (assuming enough disk space is
   /// available for spilled rows). This includes buffers used by the build side,
@@ -536,8 +577,8 @@ class PhjBuilder : public JoinBuilder {
   /// Updates 'state_' to 'next_state', logging the transition.
   void UpdateState(HashJoinState next_state);
 
-  /// Returns the current 'state_' as a string.
-  std::string PrintState() const;
+  /// Returns the string represenvation of 'state'.
+  static std::string PrintState(HashJoinState state);
 
   /// Create and initialize a set of hash partitions for partitioning level 'level'.
   /// The previous hash partitions must have been cleared with DoneProbing().
@@ -546,14 +587,22 @@ class PhjBuilder : public JoinBuilder {
 
   /// Create a new partition and prepare it for writing. Returns an error if initializing
   /// the partition or allocating the write buffer fails.
-  Status CreateAndPreparePartition(int level, std::unique_ptr<Partition>* partition);
+  Status CreateAndPreparePartition(int level,
+      std::unique_ptr<PhjBuilderPartition>* partition);
 
   /// Reads the rows in build_batch and partitions them into hash_partitions_. If
   /// 'build_filters' is true, runtime filters are populated. 'is_null_aware' is
   /// set to true if the join type is a null aware join.
   Status ProcessBuildBatch(
-      RowBatch* build_batch, HashTableCtx* ctx, bool build_filters,
-      bool is_null_aware) WARN_UNUSED_RESULT;
+      RowBatch* build_batch, HashTableCtx* ctx, bool build_filters, bool is_null_aware);
+
+  /// Helper method for Send() that that does the actual work apart from updating the
+  /// counters. Also used by RepartitionBuildInput().
+  Status AddBatch(RowBatch* build_batch);
+
+  /// Helper method for FlushFinal() that does the actual work. Also used by
+  /// RepartitionBuildInput().
+  Status FinalizeBuild(RuntimeState* state);
 
   /// Append 'row' to 'stream'. In the common case, appending the row to the stream
   /// immediately succeeds. Otherwise this function falls back to the slower path of
@@ -577,7 +626,7 @@ class PhjBuilder : public JoinBuilder {
   /// status if we couldn't spill a partition. If 'spilled_partition' is non-NULL, set
   /// to the partition that was the one spilled.
   Status SpillPartition(BufferedTupleStream::UnpinMode mode,
-      Partition** spilled_partition = nullptr) WARN_UNUSED_RESULT;
+      PhjBuilderPartition** spilled_partition = nullptr) WARN_UNUSED_RESULT;
 
   /// Tries to build hash tables for all unspilled hash partitions. Called after
   /// FlushFinal() when all build rows have been partitioned and added to the appropriate
@@ -601,34 +650,58 @@ class PhjBuilder : public JoinBuilder {
   /// Ensures that 'probe_stream_reservation_' has enough reservation for a stream per
   /// spilled partition in 'hash_partitions_', plus for the input stream if the input
   /// is a spilled partition (determined by 'next_state' - either PARTITIONING_PROBE or
-  /// REPARTITIONING_PROBE). May spill additional partitions until it can free enough
+  /// REPARTITIONING_PROBE). If num_probe_threads_ is > 1, reserves this amount for each
+  /// probe thread. May spill additional partitions until it can free enough
   /// reservation. Returns an error if an error is encountered or if it runs out of
   /// partitions to spill.
   Status ReserveProbeBuffers(HashJoinState next_state);
 
   /// Returns the number of partitions in 'partitions' that are spilled.
   static int GetNumSpilledPartitions(
-      const std::vector<std::unique_ptr<Partition>>& partitions);
+      const std::vector<std::unique_ptr<PhjBuilderPartition>>& partitions);
 
   /// Transfer reservation for probe streams to 'probe_client'. Memory for one stream was
   /// reserved per spilled partition in FlushFinal(), plus the input stream if the input
   /// partition was spilled.
+  /// This is safe to call from multiple probe threads concurrently.
   Status TransferProbeStreamReservation(BufferPool::ClientHandle* probe_client);
 
-  /// Calculates the amount of memory to be transferred for probe streams when probing
-  /// in the given 'state'. Depends on 'hash_partitions_' and 'spillable_buffer_size_'.
+  /// Calculates the amount of memory per probe thread/join node instance to be
+  /// transferred for probe streams when probing in the given 'state'. Depends on
+  /// 'hash_partitions_', 'spilled_partitions_' and 'spillable_buffer_size_'.
   int64_t CalcProbeStreamReservation(HashJoinState state) const;
 
+  /// The serial part of BeginSpilledProbe() that is executed by a single thread.
+  Status BeginSpilledProbeSerial();
+
   /// Creates new hash partitions and repartitions 'input_partition' into PARTITION_FANOUT
   /// new partitions with level input_partition->level() + 1. The previous hash partitions
   /// must have been cleared with ClearHashPartitions(). This function reserves enough
   /// memory for a read buffer for the input probe stream and a write buffer for each
   /// spilled partition after repartitioning.
-  Status RepartitionBuildInput(Partition* input_partition) WARN_UNUSED_RESULT;
+  Status RepartitionBuildInput(PhjBuilderPartition* input_partition) WARN_UNUSED_RESULT;
 
   /// Returns the largest build row count out of the current hash partitions.
   int64_t LargestPartitionRows() const;
 
+  /// Helper for DoneProbingHashPartitions() that processes and cleans up the hash
+  /// partitions.
+  void CleanUpHashPartitions(
+      std::deque<std::unique_ptr<PhjBuilderPartition>>* output_partitions,
+      RowBatch* batch);
+
+  /// Helper for DoneProbingSinglePartition() that processes and cleans up the current
+  /// spilled partition.
+  void CleanUpSinglePartition(
+      std::deque<std::unique_ptr<PhjBuilderPartition>>* output_partitions,
+      RowBatch* batch);
+
+  /// The serial part of BeginNullAwareProbe() that is executed by a single thread.
+  Status BeginNullAwareProbeSerial();
+
+  /// Close the null aware partition (if there is one) and set it to NULL.
+  void CloseNullAwarePartition();
+
   /// Calls Close() on every Partition, deletes them, and cleans up any pointers that
   /// may reference them. If 'row_batch' if not NULL, transfers the ownership of all
   /// row-backing resources to it.
@@ -719,9 +792,6 @@ class PhjBuilder : public JoinBuilder {
   /// Level of max partition (i.e. number of repartitioning steps).
   RuntimeProfile::HighWaterMarkCounter* max_partition_level_ = nullptr;
 
-  /// Number of build rows that have been partitioned.
-  RuntimeProfile::Counter* num_build_rows_partitioned_ = nullptr;
-
   /// Number of partitions that have been spilled.
   RuntimeProfile::Counter* num_spilled_partitions_ = nullptr;
 
@@ -742,6 +812,10 @@ class PhjBuilder : public JoinBuilder {
   /// that were not spilled.
   RuntimeProfile::Counter* repartition_timer_ = nullptr;
 
+  // Barrier used to synchronize the probe-side threads at synchronization points in the
+  // partitioned hash join algorithm. Used only when 'num_probe_threads_' > 1.
+  std::unique_ptr<CyclicBarrier> probe_barrier_;
+
   /////////////////////////////////////////
   /// BEGIN: Members that must be Reset()
 
@@ -760,7 +834,7 @@ class PhjBuilder : public JoinBuilder {
   /// initialized before partitioning or re-partitioning the build input
   /// and cleared after we've finished probing the partitions.
   /// This is not used when processing a single spilled partition.
-  std::vector<std::unique_ptr<Partition>> hash_partitions_;
+  std::vector<std::unique_ptr<PhjBuilderPartition>> hash_partitions_;
 
   /// Spilled partitions that need further processing. Populated in
   /// DoneProbingHashPartitions() with the spilled hash partitions.
@@ -772,7 +846,7 @@ class PhjBuilder : public JoinBuilder {
   /// spilled_partitions_.back() is the spilled partition being processed, if one is
   /// currently being processed (i.e. between BeginSpilledProbe() and the corresponding
   /// DoneProbing*() call).
-  std::vector<std::unique_ptr<PhjBuilder::Partition>> spilled_partitions_;
+  std::vector<std::unique_ptr<PhjBuilderPartition>> spilled_partitions_;
 
   /// Partition used for null-aware joins. This partition is always processed at the end
   /// after all build and probe rows are processed. In this partition's 'build_rows_', we
@@ -780,7 +854,7 @@ class PhjBuilder : public JoinBuilder {
   /// NULL (i.e. it has a NULL on the eq join slot).
   /// NULL if the join is not null aware or we are done processing this partition.
   /// This partition starts off in memory but can be spilled.
-  std::unique_ptr<Partition> null_aware_partition_;
+  std::unique_ptr<PhjBuilderPartition> null_aware_partition_;
 
   /// Populated during the hash table building phase if any partitions spilled.
   /// Reservation for one probe stream write buffer per spilled partition is
@@ -803,12 +877,13 @@ class PhjBuilder : public JoinBuilder {
   /////////////////////////////////////////
 
   /// Jitted ProcessBuildBatch function pointers. NULL if codegen is disabled.
-  const PhjBuilderConfig::ProcessBuildBatchFn& process_build_batch_fn_;
-  const PhjBuilderConfig::ProcessBuildBatchFn& process_build_batch_fn_level0_;
+  const CodegenFnPtr<PhjBuilderConfig::ProcessBuildBatchFn>& process_build_batch_fn_;
+  const CodegenFnPtr<PhjBuilderConfig::ProcessBuildBatchFn>&
+      process_build_batch_fn_level0_;
 
   /// Jitted Partition::InsertBatch() function pointers. NULL if codegen is disabled.
-  const Partition::InsertBatchFn& insert_batch_fn_ ;
-  const Partition::InsertBatchFn& insert_batch_fn_level0_;
+  const CodegenFnPtr<InsertBatchFn>& insert_batch_fn_;
+  const CodegenFnPtr<InsertBatchFn>& insert_batch_fn_level0_;
 };
 } // namespace impala
 #endif
