@@ -25,13 +25,15 @@
 #include "exec/kudu-scan-node-mt.h"
 #include "exec/kudu-scan-node.h"
 #include "exprs/scalar-expr.h"
+#include "exprs/scalar-expr-evaluator.h"
 #include "runtime/blocking-row-batch-queue.h"
+#include "runtime/fragment-state.h"
 #include "runtime/io/disk-io-mgr.h"
-#include "runtime/query-state.h"
 #include "runtime/row-batch.h"
 #include "runtime/runtime-filter.inline.h"
 #include "runtime/runtime-state.h"
 #include "runtime/scanner-mem-limiter.h"
+#include "util/debug-util.h"
 #include "util/disk-info.h"
 #include "util/pretty-printer.h"
 #include "util/runtime-profile-counters.h"
@@ -96,7 +98,7 @@ PROFILE_DEFINE_HIGH_WATER_MARK_COUNTER(PeakScannerThreadConcurrency, STABLE_LOW,
 
 const string ScanNode::SCANNER_THREAD_COUNTERS_PREFIX = "ScannerThreads";
 
-Status ScanPlanNode::Init(const TPlanNode& tnode, RuntimeState* state) {
+Status ScanPlanNode::Init(const TPlanNode& tnode, FragmentState* state) {
   RETURN_IF_ERROR(PlanNode::Init(tnode, state));
   const TQueryOptions& query_options = state->query_options();
   for (const TRuntimeFilterDesc& filter_desc : tnode.runtime_filters) {
@@ -155,7 +157,7 @@ Status ScanNode::Prepare(RuntimeState* state) {
     filter_ctx.filter = state->filter_bank()->RegisterConsumer(filter_desc);
     // TODO: Enable stats for min-max filters when Kudu exposes info about filters
     // (KUDU-2162).
-    if (filter_ctx.filter->is_bloom_filter()) {
+    if (filter_ctx.filter->is_bloom_filter() || filter_ctx.filter->is_min_max_filter()) {
       string filter_profile_title = Substitute("Filter $0 ($1)", filter_desc.filter_id,
           PrettyPrinter::Print(filter_ctx.filter->filter_size(), TUnit::BYTES));
       RuntimeProfile* profile =
@@ -233,7 +235,8 @@ bool ScanNode::WaitForRuntimeFilters() {
     runtime_profile()->AddInfoString("Runtime filters",
         Substitute("All filters arrived. Waited $0. Maximum arrival delay: $1.",
                                          wait_time, arrival_delay));
-    VLOG(2) << "Filters arrived. Waited " << wait_time;
+    VLOG(2) << "Filters arrived. Waited " << wait_time
+            << ". Current time since reboot(ms) " << end;
     return true;
   }
 
