@@ -24,6 +24,7 @@
 #include <vector>
 #include <boost/scoped_ptr.hpp>
 
+#include "codegen/codegen-fn-ptr.h"
 #include "common/global-types.h"
 #include "common/status.h"
 #include "exprs/expr.h"
@@ -56,6 +57,7 @@ using impala_udf::DecimalVal;
 using impala_udf::DateVal;
 using impala_udf::CollectionVal;
 
+class FragmentState;
 struct LibCacheEntry;
 class LlvmCodeGen;
 class MemTracker;
@@ -119,11 +121,9 @@ struct ScalarExprsResultsRowLayout {
 /// ScalarExpr's compute functions are codegend to replace calls to the generic compute
 /// function of child expressions with the exact compute functions based on the return
 /// types of the child expressions known at runtime. Subclasses should override
-/// GetCodegendComputeFnImpl() to either generate custom IR compute functions using
-/// IRBuilder, which inlines calls to child expressions' compute functions, or simply call
-/// GetCodegendComputeFnWrapper() to generate a wrapper function to call the interpreted
-/// compute function. Note that we do not need a separate GetCodegendComputeFn() for each
-/// type.
+/// GetCodegendComputeFnImpl() to generate custom IR compute functions using IRBuilder,
+/// which inlines calls to child expressions' compute functions. Note that we do not need
+/// a separate GetCodegendComputeFn() for each type.
 ///
 /// The two main usage patterns for ScalarExpr are:
 /// * The codegen'd expressions are called from other codegen'd functions, e.g. from a
@@ -136,9 +136,6 @@ struct ScalarExprsResultsRowLayout {
 /// (e.g. by ScalarExprEvaluator::GetConstValue()) but may fail back to a slower
 /// interpreted implementation.
 ///
-/// TODO: remove GetCodegendComputeFnWrapper(), which is a hack to enable "codegen"
-/// by generating LLVM functions that actually call into the interpreted path.
-///
 class ScalarExpr : public Expr {
  public:
   /// Create a new ScalarExpr based on thrift Expr 'texpr'. The newly created ScalarExpr
@@ -146,22 +143,22 @@ class ScalarExpr : public Expr {
   /// tuple row descriptor of the input tuple row. On failure, 'expr' is set to NULL and
   /// the expr tree (if created) will be closed. Error status will be returned too.
   static Status Create(const TExpr& texpr, const RowDescriptor& row_desc,
-      RuntimeState* state, ObjectPool* pool, ScalarExpr** expr) WARN_UNUSED_RESULT;
+      FragmentState* state, ObjectPool* pool, ScalarExpr** expr) WARN_UNUSED_RESULT;
 
   /// Create a new ScalarExpr based on thrift Expr 'texpr'. The newly created ScalarExpr
   /// is stored in ObjectPool 'state->obj_pool()' and returned in 'expr'. 'row_desc' is
   /// the tuple row descriptor of the input tuple row. Returns error status on failure.
   static Status Create(const TExpr& texpr, const RowDescriptor& row_desc,
-      RuntimeState* state, ScalarExpr** expr) WARN_UNUSED_RESULT;
+      FragmentState* state, ScalarExpr** expr) WARN_UNUSED_RESULT;
 
   /// Convenience functions creating multiple ScalarExpr.
   static Status Create(const std::vector<TExpr>& texprs, const RowDescriptor& row_desc,
-      RuntimeState* state, ObjectPool* pool, std::vector<ScalarExpr*>* exprs)
-      WARN_UNUSED_RESULT;
+      FragmentState* state, ObjectPool* pool,
+      std::vector<ScalarExpr*>* exprs) WARN_UNUSED_RESULT;
 
   /// Convenience functions creating multiple ScalarExpr.
   static Status Create(const std::vector<TExpr>& texprs, const RowDescriptor& row_desc,
-      RuntimeState* state, std::vector<ScalarExpr*>* exprs) WARN_UNUSED_RESULT;
+      FragmentState* state, std::vector<ScalarExpr*>* exprs) WARN_UNUSED_RESULT;
 
   /// Returns true if this expression is a SlotRef. Overridden by SlotRef.
   virtual bool IsSlotRef() const { return false; }
@@ -308,7 +305,7 @@ class ScalarExpr : public Expr {
   /// point into the codegen'd code. Currently we assume all roots of ScalarExpr subtrees
   /// exprs are potential entry points.
   virtual Status Init(const RowDescriptor& row_desc, bool is_entry_point,
-      RuntimeState* state) WARN_UNUSED_RESULT;
+      FragmentState* state) WARN_UNUSED_RESULT;
 
   /// Initializes 'eval' for execution. If scope if FRAGMENT_LOCAL, both
   /// fragment-local and thread-local states should be initialized. If scope is
@@ -342,29 +339,11 @@ class ScalarExpr : public Expr {
   /// tree.
   llvm::Function* CreateIrFunctionPrototype(const std::string& name, LlvmCodeGen* codegen,
       llvm::Value* (*args)[2]);
-
-  /// Generates an IR compute function that calls the interpreted compute function.
-  /// It doesn't provide any performance benefit over the interpreted path. This is
-  /// useful for builtins (e.g. && and || operators) and UDF which don't generate
-  /// custom IR code but are part of a larger expr tree. The IR compute function of
-  /// the larger expr tree may still benefit from custom IR and inlining of other
-  /// sub-expressions.
-  ///
-  /// TODO: this should be removed in the long run and replaced with cross-compilation
-  /// together with constant propagation and loop unrolling.
-  Status GetCodegendComputeFnWrapper(LlvmCodeGen* codegen, llvm::Function** fn)
-      WARN_UNUSED_RESULT;
-
-  /// Helper function for GetCodegendComputeFnWrapper(). Returns the cross-compiled IR
-  /// function of the static Get*Val wrapper function for return type 'type'. Must not
-  /// be called for a type that doesn't have a wrapper function in the built-in
-  /// IR module. Never returns NULL.
-  llvm::Function* GetStaticGetValWrapper(ColumnType type, LlvmCodeGen* codegen);
-
  protected:
+
   /// Return true if we should codegen this expression node, based on query options
   /// and the properties of this ScalarExpr node.
-  bool ShouldCodegen(const RuntimeState* state) const;
+  bool ShouldCodegen(const FragmentState* state) const;
 
   /// Return true if it is possible to evaluate this expression node without codegen.
   /// The vast majority of exprs support interpretation, so default to true. Scalars
@@ -397,7 +376,9 @@ class ScalarExpr : public Expr {
   /// Non-NULL if this expr is codegen'd and the constructor of this Expr requested
   /// that this expr should be an entry point from interpreted to codegen'd code.
   /// (see class comment for explanation of usage patterns and motivation).
-  void* codegend_compute_fn_ = nullptr;
+  /// The actual types of the functions differ so we use void* and reinterpret cast before
+  /// calling the functions.
+  CodegenFnPtr<void*> codegend_compute_fn_;
 
   /// True if 'codegend_compute_fn_' is registered with LlvmCodeGen as an entry point
   /// to codegen to fill in . If this is true, then 'codegend_compute_fn_' will be set
@@ -406,14 +387,8 @@ class ScalarExpr : public Expr {
   bool added_to_jit_ = false;
 
   /// Static wrappers which call the compute function of the given ScalarExpr, passing
-  /// it the ScalarExprEvaluator and TupleRow. These are cross-compiled and called by
-  /// the IR wrapper functions generated by GetCodegendComputeFnWrapper(). The wrapper
-  /// functions avoid the need for codegen'd callers to do a virtual function call on
-  /// the ScalarExpr.
-  /// Note: the ScalarExpr subclass is known at codegen time, so codegen *could* replace
-  /// the virtual function call with a direct function call. However, it would be better
-  /// to focus on removing GetCodegendComputeFnWrapper() instead of micro-optimising this
-  /// inefficient mechanism.
+  /// it the ScalarExprEvaluator and TupleRow. These are cross-compiled and used by
+  /// GetStaticGetValWrapper.
   static BooleanVal GetBooleanValInterpreted(
       ScalarExpr*, ScalarExprEvaluator*, const TupleRow*);
   static TinyIntVal GetTinyIntValInterpreted(
