@@ -17,15 +17,16 @@
 
 #include "exprs/timestamp-functions.h"
 
+#include <ctime>
+#include <iomanip>
+
 #include <boost/date_time/compiler_config.hpp>
 #include <boost/date_time/posix_time/posix_time_types.hpp>
 #include <boost/date_time/gregorian/gregorian_types.hpp>
-#include <ctime>
 #include <gutil/strings/substitute.h>
 
 #include "exprs/anyval-util.h"
 #include "runtime/datetime-simple-date-format-parser.h"
-#include "runtime/string-value.inline.h"
 #include "runtime/timestamp-value.inline.h"
 #include "runtime/timestamp-value.h"
 #include "udf/udf.h"
@@ -64,7 +65,7 @@ StringVal TimestampFunctions::StringValFromTimestamp(FunctionContext* context,
   DateTimeFormatContext* dt_ctx = reinterpret_cast<DateTimeFormatContext*>(state);
   if (!context->IsArgConstant(1)) {
     dt_ctx->Reset(reinterpret_cast<const char*>(fmt.ptr), fmt.len);
-    if (!SimpleDateFormatTokenizer::Tokenize(dt_ctx)){
+    if (!SimpleDateFormatTokenizer::Tokenize(dt_ctx, FORMAT)){
       ReportBadFormat(context, datetime_parse_util::GENERAL_ERROR, fmt, false);
       return StringVal::null();
     }
@@ -164,11 +165,11 @@ TimestampVal TimestampFunctions::ToTimestamp(FunctionContext* context,
   void* state = context->GetFunctionState(FunctionContext::THREAD_LOCAL);
   DateTimeFormatContext* dt_ctx = reinterpret_cast<DateTimeFormatContext*>(state);
   if (!context->IsArgConstant(1)) {
-     dt_ctx->Reset(reinterpret_cast<const char*>(fmt.ptr), fmt.len);
-     if (!SimpleDateFormatTokenizer::Tokenize(dt_ctx)) {
-       ReportBadFormat(context, datetime_parse_util::GENERAL_ERROR, fmt, false);
-       return TimestampVal::null();
-     }
+    dt_ctx->Reset(reinterpret_cast<const char*>(fmt.ptr), fmt.len);
+    if (!SimpleDateFormatTokenizer::Tokenize(dt_ctx, PARSE)) {
+      ReportBadFormat(context, datetime_parse_util::GENERAL_ERROR, fmt, false);
+      return TimestampVal::null();
+    }
   }
   const TimestampValue& tv = TimestampValue::ParseSimpleDateFormat(
       reinterpret_cast<const char*>(date.ptr), date.len, *dt_ctx);
