@@ -29,7 +29,9 @@
 #include <gutil/strings/split.h>
 #include <gutil/strings/strip.h>
 #include <gutil/strings/substitute.h>
-#include <random>
+#include <algorithm>
+#include <map>
+#include <vector>
 #include <string>
 #include <vector>
 #include <thrift/Thrift.h>
@@ -43,15 +45,19 @@
 #include "kudu/rpc/sasl_common.h"
 #include "kudu/security/gssapi.h"
 #include "kudu/security/init.h"
+#include "kudu/security/openssl_util.h"
 #include "rpc/auth-provider.h"
-#include "rpc/cookie-util.h"
+#include "rpc/authentication-util.h"
 #include "rpc/thrift-server.h"
+#include "runtime/exec-env.h"
+#include "service/frontend.h"
 #include "transport/THttpServer.h"
 #include "transport/TSaslClientTransport.h"
 #include "util/auth-util.h"
 #include "util/coding-util.h"
 #include "util/debug-util.h"
 #include "util/error-util.h"
+#include "util/ldap-util.h"
 #include "util/network-util.h"
 #include "util/os-util.h"
 #include "util/promise.h"
@@ -64,7 +70,6 @@
 #include "common/names.h"
 
 using boost::algorithm::is_any_of;
-using boost::algorithm::replace_all;
 using boost::algorithm::split;
 using boost::algorithm::trim;
 using boost::mt19937;
@@ -82,41 +87,56 @@ DECLARE_string(be_principal);
 DECLARE_string(krb5_ccname);
 DECLARE_string(krb5_conf);
 DECLARE_string(krb5_debug_file);
+DECLARE_int32(external_fe_port);
 
 // Defined in kudu/security/init.cc
 DECLARE_bool(use_system_auth_to_local);
 
 DECLARE_int64(max_cookie_lifetime_s);
 
+DECLARE_int32(hs2_http_port);
+
 DEFINE_string(sasl_path, "", "Colon separated list of paths to look for SASL "
     "security library plugins.");
 DEFINE_bool(enable_ldap_auth, false,
     "If true, use LDAP authentication for client connections");
-
-DEFINE_string(ldap_uri, "", "The URI of the LDAP server to authenticate users against");
-DEFINE_bool(ldap_tls, false, "If true, use the secure TLS protocol to connect to the LDAP"
-    " server");
 DEFINE_string(ldap_ca_certificate, "", "The full path to the certificate file used to"
     " authenticate the LDAP server's certificate for SSL / TLS connections.");
-DEFINE_bool(ldap_passwords_in_clear_ok, false, "If set, will allow LDAP passwords "
-    "to be sent in the clear (without TLS/SSL) over the network.  This option should not "
-    "be used in production environments" );
-DEFINE_bool(ldap_allow_anonymous_binds, false, "(Advanced) If true, LDAP authentication "
-    "with a blank password (an 'anonymous bind') is allowed by Impala.");
-DEFINE_bool(ldap_manual_config, false, "Obsolete; Ignored");
-DEFINE_string(ldap_domain, "", "If set, Impala will try to bind to LDAP with a name of "
-    "the form <userid>@<ldap_domain>");
-DEFINE_string(ldap_baseDN, "", "If set, Impala will try to bind to LDAP with a name of "
-    "the form uid=<userid>,<ldap_baseDN>");
-DEFINE_string(ldap_bind_pattern, "", "If set, Impala will try to bind to LDAP with a name"
-     " of <ldap_bind_pattern>, but where the string #UID is replaced by the user ID. Use"
-     " to control the bind name precisely; do not set --ldap_domain or --ldap_baseDN with"
-     " this option");
+DEFINE_string(ldap_user_filter, "", "Used as filter for both simple and search "
+    "bind mechanisms. For simple bind it is a comma separated list of user names. If "
+    "specified, users must be on this list for authentication to succeed. For search "
+    "bind it is an LDAP filter that will be used during LDAP search, it can contain "
+    "'{0}' pattern which will be replaced with the user name.");
+DEFINE_string(ldap_group_filter, "", "Used as filter for both simple and search "
+    "bind mechanisms. For simple bind it is a comma separated list of groups. If "
+    "specified, users must belong to one of these groups for authentication to succeed. "
+    "For search bind it is an LDAP filter that will be used during LDAP group search, it "
+    "can contain '{0}' pattern which will be replaced with the user name and/or '{1}' "
+    "which will be replace with the user dn.");
+
 DEFINE_string(internal_principals_whitelist, "hdfs", "(Advanced) Comma-separated list of "
     " additional usernames authorized to access Impala's internal APIs. Defaults to "
     "'hdfs' which is the system user that in certain deployments must access "
     "catalog server APIs.");
 
+DEFINE_string(trusted_domain, "",
+    "If set, Impala will skip authentication for connections originating from this "
+    "domain. Currently, only connections over HTTP support this. Note: It still requires "
+    "the client to specify a username via the Basic Authorization header in the format "
+    "<username>:<password> where the password is not used and can be left blank.");
+
+DEFINE_bool(trusted_domain_use_xff_header, false,
+    "If set to true, this uses the 'X-Forwarded-For' HTML header to check for origin "
+    "while attempting to verify if the connection request originated from a trusted "
+    "domain. Only used if '--trusted_domain' is specified. Warning: Only use this if you "
+    "trust the incoming connection to have this set correctly.");
+
+DEFINE_bool_hidden(saml2_allow_without_tls_debug_only, false,
+    "When this configuration is set to true, Impala allows SAML2 authentication "
+    "on unsecure channel. This should be only enabled for development / testing.");
+
+DECLARE_string(saml2_sp_callback_url);
+
 namespace impala {
 
 // Sasl callbacks.  Why are these here?  Well, Sasl isn't that bright, and
@@ -144,10 +164,6 @@ static string APP_NAME;
 static const string KERBEROS_MECHANISM = "GSSAPI";
 static const string PLAIN_MECHANISM = "PLAIN";
 
-// Required prefixes for ldap URIs:
-static const string LDAP_URI_PREFIX = "ldap://";
-static const string LDAPS_URI_PREFIX = "ldaps://";
-
 // We implement an "auxprop" plugin for the Sasl layer in order to have a hook in which
 // to log messages about the start of authentication. This is that plugin's name.
 static const string IMPALA_AUXPROP_PLUGIN = "impala-auxprop";
@@ -194,91 +210,31 @@ static int SaslLogCallback(void* context, int level, const char* message) {
   return SASL_OK;
 }
 
-// This callback is only called when we're providing LDAP authentication. This "check
-// pass" callback is our hook to ask the real LDAP server if we're allowed to log in or
-// not. We can be thought of as a proxy for LDAP logins - the user gives their password
-// to us, and we pass it to the real LDAP server.
-//
-// Note that this method uses ldap_sasl_bind_s(), which does *not* provide any security
-// to the connection between Impala and the LDAP server. You must either set --ldap_tls,
-// or have a URI which has "ldaps://" as the scheme in order to get a secure connection.
-// Use --ldap_ca_certificate to specify the location of the certificate used to confirm
-// the authenticity of the LDAP server certificate.
-//
-// user: The username to authenticate
-// pass: The password to use
-// passlen: The length of pass
-// Return: true on success, false otherwise
-static bool LdapCheckPass(const char* user, const char* pass, unsigned passlen) {
-  if (passlen == 0 && !FLAGS_ldap_allow_anonymous_binds) {
-    // Disable anonymous binds.
-    return false;
-  }
+// Calls into the LDAP utils to check the provided user/pass.
+bool DoLdapCheck(const char* user, const char* pass, unsigned passlen) {
+  ImpalaLdap* ldap = AuthManager::GetInstance()->GetLdap();
+  bool success = ldap->LdapCheckPass(user, pass, passlen);
 
-  LDAP* ld;
-  int rc = ldap_initialize(&ld, FLAGS_ldap_uri.c_str());
-  if (rc != LDAP_SUCCESS) {
-    LOG(WARNING) << "Could not initialize connection with LDAP server ("
-                 << FLAGS_ldap_uri << "). Error: " << ldap_err2string(rc);
-    return false;
-  }
-
-  // Force the LDAP version to 3 to make sure TLS is supported.
-  int ldap_ver = 3;
-  ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &ldap_ver);
-
-  // If -ldap_tls is turned on, and the URI is ldap://, issue a STARTTLS operation.
-  // Note that we'll ignore -ldap_tls when using ldaps:// because we've already
-  // got a secure connection (and the LDAP server will reject the STARTTLS).
-  if (FLAGS_ldap_tls && (FLAGS_ldap_uri.find(LDAP_URI_PREFIX) == 0)) {
-    int tls_rc = ldap_start_tls_s(ld, NULL, NULL);
-    if (tls_rc != LDAP_SUCCESS) {
-      LOG(WARNING) << "Could not start TLS secure connection to LDAP server ("
-                   << FLAGS_ldap_uri << "). Error: " << ldap_err2string(tls_rc);
-      ldap_unbind_ext(ld, NULL, NULL);
-      return false;
+  if (success) {
+    ImpalaServer* server = ExecEnv::GetInstance()->impala_server();
+    if (server == nullptr) {
+      LOG(FATAL) << "Invalid config: SASL LDAP is only supported for client connections "
+                 << "to an impalad.";
     }
-    VLOG(2) << "Started TLS connection with LDAP server: " << FLAGS_ldap_uri;
-  }
-
-  // Map the user string into an acceptable LDAP "DN" (distinguished name)
-  string user_str = user;
-  if (!FLAGS_ldap_domain.empty()) {
-    // Append @domain if there isn't already an @ in the user string.
-    if (user_str.find("@") == string::npos) {
-      user_str = Substitute("$0@$1", user_str, FLAGS_ldap_domain);
+    // If the user is an authorized proxy user, we do not yet know the effective user as
+    // it may be set by 'impala.doas.user', in which case we defer checking LDAP filters
+    // until OpenSession(). Otherwise, we prefer to check the filters as easly as
+    // possible, so check them here.
+    if (!server->IsAuthorizedProxyUser(user)) {
+      success = ldap->LdapCheckFilters(user);
     }
-  } else if (!FLAGS_ldap_baseDN.empty()) {
-    user_str = Substitute("uid=$0,$1", user_str, FLAGS_ldap_baseDN);
-  } else if (!FLAGS_ldap_bind_pattern.empty()) {
-    user_str = FLAGS_ldap_bind_pattern;
-    replace_all(user_str, "#UID", user);
   }
 
-  // Map the password into a credentials structure
-  struct berval cred;
-  cred.bv_val = const_cast<char*>(pass);
-  cred.bv_len = passlen;
-
-  VLOG_QUERY << "Trying simple LDAP bind for: " << user_str;
-
-  rc = ldap_sasl_bind_s(ld, user_str.c_str(), LDAP_SASL_SIMPLE, &cred,
-      NULL, NULL, NULL);
-  // Free ld
-  ldap_unbind_ext(ld, NULL, NULL);
-  if (rc != LDAP_SUCCESS) {
-    LOG(WARNING) << "LDAP authentication failure for " << user_str
-                 << " : " << ldap_err2string(rc);
-    return false;
-  }
-
-  VLOG_QUERY << "LDAP bind successful";
-
-  return true;
+  return success;
 }
 
-// Wrapper around the function we use to check passwords with LDAP which converts the
-// return value to something appropriate for SASL.
+// Wrapper around the function we use to check passwords with LDAP which has the function
+// signature required to work with SASL.
 //
 // conn: The Sasl connection struct, which we ignore
 // context: Ignored; always NULL
@@ -289,7 +245,7 @@ static bool LdapCheckPass(const char* user, const char* pass, unsigned passlen)
 // Return: SASL_OK on success, SASL_FAIL otherwise
 int SaslLdapCheckPass(sasl_conn_t* conn, void* context, const char* user,
     const char* pass, unsigned passlen, struct propctx* propctx) {
-  return LdapCheckPass(user, pass, passlen) ? SASL_OK : SASL_FAIL;
+  return DoLdapCheck(user, pass, passlen) ? SASL_OK : SASL_FAIL;
 }
 
 // Sasl wants a way to ask us about some options, this function provides
@@ -522,29 +478,41 @@ bool CookieAuth(ThriftServer::ConnectionContext* connection_context,
   return false;
 }
 
+bool TrustedDomainCheck(ThriftServer::ConnectionContext* connection_context,
+    const AuthenticationHash& hash, const std::string& origin, string auth_header) {
+  if (!IsTrustedDomain(origin, FLAGS_trusted_domain)) return false;
+  string stripped_basic_auth_token;
+  StripWhiteSpace(&auth_header);
+  bool got_basic_auth =
+      TryStripPrefixString(auth_header, "Basic ", &stripped_basic_auth_token);
+  string basic_auth_token = got_basic_auth ? move(stripped_basic_auth_token) : "";
+  string username, password;
+  Status status = BasicAuthExtractCredentials(basic_auth_token, username, password);
+  if (!status.ok()) {
+    LOG(ERROR) << "Error parsing basic authentication token from: "
+               << TNetworkAddressToString(connection_context->network_address)
+               << " Error: " << status;
+    return false;
+  }
+  connection_context->username = username;
+  // Create a cookie to return.
+  connection_context->return_headers.push_back(
+      Substitute("Set-Cookie: $0", GenerateCookie(username, hash)));
+  return true;
+}
+
 bool BasicAuth(ThriftServer::ConnectionContext* connection_context,
-    const AuthenticationHash& hash, const std::string& base64) {
-  if (base64.empty()) {
+    const AuthenticationHash& hash, const string& base64) {
+  string username, password;
+  Status status = BasicAuthExtractCredentials(base64, username, password);
+  if (!status.ok()) {
+    LOG(ERROR) << "Error parsing basic authentication token from: "
+               << TNetworkAddressToString(connection_context->network_address)
+               << " Error: " << status;
     connection_context->return_headers.push_back("WWW-Authenticate: Basic");
     return false;
   }
-  string decoded;
-  if (!Base64Unescape(base64, &decoded)) {
-    LOG(ERROR) << "Failed to decode base64 auth string from: "
-               << TNetworkAddressToString(connection_context->network_address);
-    connection_context->return_headers.push_back("WWW-Authenticate: Basic");
-    return false;
-  }
-  std::size_t colon = decoded.find(':');
-  if (colon == std::string::npos) {
-    LOG(ERROR) << "Auth string must be in the form '<username>:<password>' from: "
-               << TNetworkAddressToString(connection_context->network_address);
-    connection_context->return_headers.push_back("WWW-Authenticate: Basic");
-    return false;
-  }
-  string username = decoded.substr(0, colon);
-  string password = decoded.substr(colon + 1);
-  bool ret = LdapCheckPass(username.c_str(), password.c_str(), password.length());
+  bool ret = DoLdapCheck(username.c_str(), password.c_str(), password.length());
   if (ret) {
     // Authenication was successful, so set the username on the connection.
     connection_context->username = username;
@@ -612,31 +580,147 @@ vector<string> ReturnHeaders(ThriftServer::ConnectionContext* connection_context
   return std::move(connection_context->return_headers);
 }
 
-// Takes the path component of an HTTP request and parses it. For now, we only care about
-// the 'doAs' parameter.
-bool HttpPathFn(ThriftServer::ConnectionContext* connection_context, const string& path,
-    string* err_msg) {
-  // 'path' should be of the form '/.*[?<key=value>[&<key=value>...]]'
-  vector<string> split = Split(path, delimiter::Limit("?", 1));
-  if (split.size() == 2) {
-    for (auto pair : Split(split[1], "&")) {
-      vector<string> key_value = Split(pair, delimiter::Limit("=", 1));
-      if (key_value.size() == 2 && key_value[0] == "doAs") {
-        string decoded;
-        if (!UrlDecode(key_value[1], &decoded)) {
-          *err_msg = Substitute(
-              "Could not decode 'doAs' parameter from HTTP request with path: $0", path);
-          return false;
-        } else {
-          connection_context->do_as_user = decoded;
-        }
-        break;
+// Parses a param string.
+//  params_to_check: map from the name of params that should be parsed to pointers where
+//                   the value should be written
+//  original: the full http path (used only in error message)
+//  params_string: & delimited param string to parse
+//  err_msg: write detailed message here in case of error
+bool ParseParams(std::map<string, string*>& params_to_check,
+   const string& original, const string& params_string, string* err_msg) {
+  for (auto pair : Split(params_string, "&")) {
+    vector<string> key_value = Split(pair, delimiter::Limit("=", 1));
+    if (key_value.size() == 2) {
+      auto it = params_to_check.find(key_value[0]);
+      if (it == params_to_check.end()) continue;
+      string decoded;
+      if (!UrlDecode(key_value[1], &decoded)) {
+        *err_msg = Substitute(
+            "Could not decode '$0' parameter from HTTP request with path: $1",
+                key_value[0], original);
+        return false;
+      } else {
+        *it->second = decoded;
       }
     }
   }
   return true;
 }
 
+// Takes the path component of an HTTP request and parses it.
+// The followings are inspected in the path:
+// - the 'doAs' parameter if it exists
+// - if SAML auth is enabled then the path is checked whether it matches
+//   the SP callback URL, if yes, then the whole body has to be read for,
+//   authentication purposes, not just the headers
+bool HttpPathFn(ThriftServer::ConnectionContext* connection_context,
+     const string& saml_path, const string& path, string* err_msg,
+     bool* read_whole_body) {
+  TWrappedHttpRequest* request = connection_context->request.get();
+  // 'path' should be of the form '/.*[?<key=value>[&<key=value>...]]'
+  vector<string> split = Split(path, delimiter::Limit("?", 1));
+  bool is_saml_response = !saml_path.empty() && split[0] == saml_path;
+  *read_whole_body = is_saml_response;
+
+  if (request) request->path = split[0];
+
+  if (split.size() != 2) return true;
+
+  std::map<string, string*> params_to_check;
+  params_to_check["doAs"] = &connection_context->do_as_user;
+
+  return ParseParams(params_to_check, path, split[1], err_msg);
+}
+
+TWrappedHttpResponse* GetSaml2Redirect(
+    ThriftServer::ConnectionContext* connection_context) {
+  TWrappedHttpRequest* request = connection_context->request.get();
+  TWrappedHttpResponse* response = connection_context->response.get();
+  DCHECK(request != nullptr);
+  DCHECK(response == nullptr);
+  response = new TWrappedHttpResponse();
+  connection_context->response.reset(response);
+  Status status =
+      ExecEnv::GetInstance()->frontend()->GetSaml2Redirect(*request, response);
+  if (!status.ok()) return nullptr;
+
+  return response;
+}
+
+TWrappedHttpResponse* ValidateSaml2AuthnResponse(
+    ThriftServer::ConnectionContext* connection_context,
+    const AuthenticationHash& hash) {
+  TWrappedHttpRequest* request = connection_context->request.get();
+  TWrappedHttpResponse* response = connection_context->response.get();
+  DCHECK(request != nullptr);
+  DCHECK(response == nullptr);
+  response = new TWrappedHttpResponse();
+  connection_context->response.reset(response);
+  // Parse the body.
+  std::map<string, string*> params_to_check;
+  string saml_relay_state;
+  params_to_check["RelayState"] =  &saml_relay_state;
+  string error_msg;
+  const string& content = request->content;
+  if(!ParseParams(params_to_check, content, content, &error_msg)) {
+    LOG(ERROR) << "failed to parse SAML response params: " << error_msg;
+    return nullptr;
+  }
+  StripWhiteSpace(&saml_relay_state);
+  request->params["RelayState"] = saml_relay_state;
+  // We return some html in case of auth error. TODO: Should handle other
+  // errors where no response is generated.
+  Status status =
+      ExecEnv::GetInstance()->frontend()->ValidateSaml2Response(*request, response);
+  return response;
+}
+
+bool ValidateSaml2Bearer(ThriftServer::ConnectionContext* connection_context,
+    const AuthenticationHash& hash) {
+  TWrappedHttpRequest* request = connection_context->request.get();
+  string username;
+  Status status =
+      ExecEnv::GetInstance()->frontend()->ValidateSaml2Bearer(*request, &username);
+
+  if (!status.ok()) return false;
+
+  connection_context->username = username;
+  // Create a cookie to return.
+  connection_context->return_headers.push_back(
+      Substitute("Set-Cookie: $0", GenerateCookie(username, hash)));
+  return true;
+}
+
+TWrappedHttpRequest* InitWrappedHttpRequest(
+      ThriftServer::ConnectionContext* connection_context) {
+  TWrappedHttpRequest* request = new TWrappedHttpRequest();
+  request->remote_ip = connection_context->network_address.hostname;
+  request->server_name = connection_context->server_name;
+  // Intentionally lying to the pac4j lib  in the frontend to ensure that non-secure
+  // test setups work similarly to secure production systems. SAML must use TLS if
+  // FLAGS_saml2_allow_without_tls_debug_only is not true.
+  request->secure = true;
+  connection_context->request.reset(request);
+  return request;
+}
+
+// Parses and validates FLAGS_saml2_sp_callback_url.
+// Sets saml_sp_path to the path part if successful.
+Status ParseSamlSpUrl(string* saml_sp_path) {
+  vector<string> split =
+      Split(FLAGS_saml2_sp_callback_url, delimiter::Limit("/", 3));
+  if (split.size() != 4 || (split[0] != "http:" && split[0] != "https:")) {
+    return Status(
+        Substitute("Bad saml2_sp_callback_url: $0", FLAGS_saml2_sp_callback_url));
+  }
+  // The port in the url should be the same as FLAGS_hs2_http_port in general,
+  // but this is not enforced to allow the use case when the port is mapped,
+  // e.g. external_port->http_port. FLAGS_saml2_sp_callback_url has to contain the
+  // external_port in this case.
+  if (saml_sp_path != nullptr) *saml_sp_path = "/" + split[3];
+  return Status::OK();
+}
+
 namespace {
 
 // SASL requires mutexes for thread safety, but doesn't implement
@@ -975,18 +1059,22 @@ Status SecureAuthProvider::Start() {
 
 Status SecureAuthProvider::GetServerTransportFactory(
     ThriftServer::TransportType underlying_transport_type, const std::string& server_name,
-    MetricGroup* metrics, boost::shared_ptr<TTransportFactory>* factory) {
-  DCHECK(!principal_.empty() || has_ldap_);
+    MetricGroup* metrics, std::shared_ptr<TTransportFactory>* factory) {
+  DCHECK(!principal_.empty() || has_ldap_ || has_saml_);
 
   if (underlying_transport_type == ThriftServer::HTTP) {
     bool has_kerberos = !principal_.empty();
     bool use_cookies = FLAGS_max_cookie_lifetime_s > 0;
-    factory->reset(new THttpServerTransportFactory(
-        server_name, metrics, has_ldap_, has_kerberos, use_cookies));
+    bool check_trusted_domain = !FLAGS_trusted_domain.empty();
+    factory->reset(new THttpServerTransportFactory(server_name, metrics, has_ldap_,
+        has_kerberos, use_cookies, check_trusted_domain, has_saml_));
     return Status::OK();
   }
 
   DCHECK(underlying_transport_type == ThriftServer::BINARY);
+
+  DCHECK(!principal_.empty() || has_ldap_);
+
   // This is the heart of the link between this file and thrift.  Here we
   // associate a Sasl mechanism with our callbacks.
   try {
@@ -1022,9 +1110,9 @@ Status SecureAuthProvider::GetServerTransportFactory(
 }
 
 Status SecureAuthProvider::WrapClientTransport(const string& hostname,
-    boost::shared_ptr<TTransport> raw_transport, const string& service_name,
-    boost::shared_ptr<TTransport>* wrapped_transport) {
-  boost::shared_ptr<sasl::TSasl> sasl_client;
+    std::shared_ptr<TTransport> raw_transport, const string& service_name,
+    std::shared_ptr<TTransport>* wrapped_transport) {
+  std::shared_ptr<sasl::TSasl> sasl_client;
   const map<string, string> props; // Empty; unused by thrift
   const string auth_id; // Empty; unused by thrift
 
@@ -1052,9 +1140,9 @@ Status SecureAuthProvider::WrapClientTransport(const string& hostname,
 }
 
 void SecureAuthProvider::SetupConnectionContext(
-    const boost::shared_ptr<ThriftServer::ConnectionContext>& connection_ptr,
-    ThriftServer::TransportType underlying_transport_type, TTransport* input_transport,
-    TTransport* output_transport) {
+    const shared_ptr<ThriftServer::ConnectionContext>& connection_ptr,
+    ThriftServer::TransportType underlying_transport_type,
+    TTransport* input_transport, TTransport* output_transport) {
   TSocket* socket = nullptr;
   switch (underlying_transport_type) {
     case ThriftServer::BINARY: {
@@ -1071,11 +1159,20 @@ void SecureAuthProvider::SetupConnectionContext(
       THttpServer* http_input_transport = down_cast<THttpServer*>(input_transport);
       THttpServer* http_output_transport = down_cast<THttpServer*>(output_transport);
       THttpServer::HttpCallbacks callbacks;
+
+      string saml_path;
+      if (has_saml_) {
+        Status parse_status = ParseSamlSpUrl(&saml_path);
+        DCHECK(parse_status.ok());
+      }
       callbacks.path_fn = std::bind(
-          HttpPathFn, connection_ptr.get(), std::placeholders::_1, std::placeholders::_2);
+          HttpPathFn, connection_ptr.get(), saml_path, std::placeholders::_1,
+          std::placeholders::_2, std::placeholders::_3);
       callbacks.return_headers_fn = std::bind(ReturnHeaders, connection_ptr.get());
       callbacks.cookie_auth_fn =
           std::bind(CookieAuth, connection_ptr.get(), hash_, std::placeholders::_1);
+      callbacks.trusted_domain_check_fn = std::bind(TrustedDomainCheck,
+          connection_ptr.get(), hash_, std::placeholders::_1, std::placeholders::_2);
       if (has_ldap_) {
         callbacks.basic_auth_fn =
             std::bind(BasicAuth, connection_ptr.get(), hash_, std::placeholders::_1);
@@ -1084,6 +1181,16 @@ void SecureAuthProvider::SetupConnectionContext(
         callbacks.negotiate_auth_fn = std::bind(NegotiateAuth, connection_ptr.get(),
             hash_, std::placeholders::_1, std::placeholders::_2);
       }
+      if (has_saml_) {
+        callbacks.init_wrapped_http_request_fn = std::bind(
+            InitWrappedHttpRequest, connection_ptr.get());
+        callbacks.get_saml_redirect_fn =
+            std::bind(GetSaml2Redirect, connection_ptr.get());
+        callbacks.validate_saml2_authn_response_fn =
+            std::bind(ValidateSaml2AuthnResponse, connection_ptr.get(), hash_);
+        callbacks.validate_saml2_bearer_fn =
+            std::bind(ValidateSaml2Bearer, connection_ptr.get(), hash_);
+      }
       http_input_transport->setCallbacks(callbacks);
       http_output_transport->setCallbacks(callbacks);
       socket = down_cast<TSocket*>(http_input_transport->getUnderlyingTransport().get());
@@ -1098,7 +1205,7 @@ void SecureAuthProvider::SetupConnectionContext(
 
 Status NoAuthProvider::GetServerTransportFactory(
     ThriftServer::TransportType underlying_transport_type, const std::string& server_name,
-    MetricGroup* metrics, boost::shared_ptr<TTransportFactory>* factory) {
+    MetricGroup* metrics, std::shared_ptr<TTransportFactory>* factory) {
   // No Sasl - yawn.  Here, have a regular old buffered transport.
   switch (underlying_transport_type) {
     case ThriftServer::BINARY:
@@ -1114,18 +1221,19 @@ Status NoAuthProvider::GetServerTransportFactory(
 }
 
 Status NoAuthProvider::WrapClientTransport(const string& hostname,
-    boost::shared_ptr<TTransport> raw_transport, const string& dummy_service,
-    boost::shared_ptr<TTransport>* wrapped_transport) {
+    std::shared_ptr<TTransport> raw_transport, const string& dummy_service,
+    std::shared_ptr<TTransport>* wrapped_transport) {
   // No Sasl - yawn.  Don't do any transport wrapping for clients.
   *wrapped_transport = raw_transport;
   return Status::OK();
 }
 
 void NoAuthProvider::SetupConnectionContext(
-    const boost::shared_ptr<ThriftServer::ConnectionContext>& connection_ptr,
-    ThriftServer::TransportType underlying_transport_type, TTransport* input_transport,
-    TTransport* output_transport) {
+    const std::shared_ptr<ThriftServer::ConnectionContext>& connection_ptr,
+    ThriftServer::TransportType underlying_transport_type,
+    TTransport* input_transport, TTransport* output_transport) {
   connection_ptr->username = "";
+  connection_ptr->do_as_user = "";
   TSocket* socket = nullptr;
   switch (underlying_transport_type) {
     case ThriftServer::BINARY: {
@@ -1141,7 +1249,8 @@ void NoAuthProvider::SetupConnectionContext(
       // Even though there's no security, we set up some callbacks, eg. to allow
       // impersonation over unsecured connections for testing purposes.
       callbacks.path_fn = std::bind(
-          HttpPathFn, connection_ptr.get(), std::placeholders::_1, std::placeholders::_2);
+          HttpPathFn, connection_ptr.get(), "", std::placeholders::_1,
+          std::placeholders::_2, std::placeholders::_3);
       callbacks.return_headers_fn = std::bind(ReturnHeaders, connection_ptr.get());
       http_input_transport->setCallbacks(callbacks);
       http_output_transport->setCallbacks(callbacks);
@@ -1155,60 +1264,32 @@ void NoAuthProvider::SetupConnectionContext(
       MakeNetworkAddress(socket->getPeerAddress(), socket->getPeerPort());
 }
 
-Status AuthManager::Init() {
-  ssl_socket_factory_.reset(new TSSLSocketFactory(TLSv1_0));
+AuthManager::AuthManager() {}
 
-  bool use_ldap = false;
-  const string excl_msg = "--$0 and --$1 are mutually exclusive "
-      "and should not be set together";
+AuthManager::~AuthManager() {}
+
+Status AuthManager::Init() {
+  // Tell Thrift not to initialize SSL for us, as we use Kudu's SSL initializtion.
+  TSSLSocketFactory::setManualOpenSSLInitialization(true);
+  kudu::security::InitializeOpenSSL();
+  LOG(INFO) << "Initialized " << OPENSSL_VERSION_TEXT;
+
+  // Could use any other requiered flag for SAML
+  bool use_saml = !FLAGS_saml2_sp_callback_url.empty();
+  if (use_saml) {
+    RETURN_IF_ERROR(ParseSamlSpUrl(nullptr));
+    if (!IsExternalTlsConfigured()) {
+      if (!FLAGS_saml2_allow_without_tls_debug_only) {
+        return Status("SAML SSO authentication should be only used with TLS enabled.");
+      }
+      LOG(WARNING) << "SAML SSO authentication is used without TLS.";
+    }
+  }
 
   // Get all of the flag validation out of the way
   if (FLAGS_enable_ldap_auth) {
-    use_ldap = true;
-
-    if (!FLAGS_ldap_domain.empty()) {
-      if (!FLAGS_ldap_baseDN.empty()) {
-        return Status(Substitute(excl_msg, "ldap_domain", "ldap_baseDN"));
-      }
-      if (!FLAGS_ldap_bind_pattern.empty()) {
-        return Status(Substitute(excl_msg, "ldap_domain", "ldap_bind_pattern"));
-      }
-    } else if (!FLAGS_ldap_baseDN.empty()) {
-      if (!FLAGS_ldap_bind_pattern.empty()) {
-        return Status(Substitute(excl_msg, "ldap_baseDN", "ldap_bind_pattern"));
-      }
-    }
-
-    if (FLAGS_ldap_uri.empty()) {
-      return Status("--ldap_uri must be supplied when --ldap_enable_auth is set");
-    }
-
-    if ((FLAGS_ldap_uri.find(LDAP_URI_PREFIX) != 0) &&
-        (FLAGS_ldap_uri.find(LDAPS_URI_PREFIX) != 0)) {
-      return Status(Substitute("--ldap_uri must start with either $0 or $1",
-              LDAP_URI_PREFIX, LDAPS_URI_PREFIX ));
-    }
-
-    LOG(INFO) << "Using LDAP authentication with server " << FLAGS_ldap_uri;
-
-    if (!FLAGS_ldap_tls && (FLAGS_ldap_uri.find(LDAPS_URI_PREFIX) != 0)) {
-      if (FLAGS_ldap_passwords_in_clear_ok) {
-        LOG(WARNING) << "LDAP authentication is being used, but without TLS. "
-                     << "ALL PASSWORDS WILL GO OVER THE NETWORK IN THE CLEAR.";
-      } else {
-        return Status("LDAP authentication specified, but without TLS. "
-                      "Passwords would go over the network in the clear. "
-                      "Enable TLS with --ldap_tls or use an ldaps:// URI. "
-                      "To override this is non-production environments, "
-                      "specify --ldap_passwords_in_clear_ok");
-      }
-    } else if (FLAGS_ldap_ca_certificate.empty()) {
-      LOG(WARNING) << "LDAP authentication is being used with TLS, but without "
-                   << "an --ldap_ca_certificate file, the identity of the LDAP "
-                   << "server cannot be verified.  Network communication (and "
-                   << "hence passwords) could be intercepted by a "
-                   << "man-in-the-middle attack";
-    }
+    RETURN_IF_ERROR(
+        ImpalaLdap::CreateLdap(&ldap_, FLAGS_ldap_user_filter, FLAGS_ldap_group_filter));
   }
 
   if (FLAGS_principal.empty() && !FLAGS_be_principal.empty()) {
@@ -1270,20 +1351,33 @@ Status AuthManager::Init() {
   // Set up the external auth provider as per above.  Either a "front end"
   // principal or ldap tells us to use a SecureAuthProvider, and we fill in
   // details from there.
-  if (use_ldap || external_kerberos_enabled) {
+  if (FLAGS_enable_ldap_auth || external_kerberos_enabled) {
     SecureAuthProvider* sap = NULL;
     external_auth_provider_.reset(sap = new SecureAuthProvider(false));
     if (external_kerberos_enabled) {
       RETURN_IF_ERROR(sap->InitKerberos(kerberos_external_principal));
       LOG(INFO) << "External communication is authenticated with Kerberos";
     }
-    if (use_ldap) {
+    if (FLAGS_enable_ldap_auth) {
       sap->InitLdap();
       LOG(INFO) << "External communication is authenticated with LDAP";
     }
+    if (use_saml) {
+      LOG(INFO) << "External communication can be also authenticated with SAML2 SSO";
+      sap->InitSaml();
+    }
   } else {
     external_auth_provider_.reset(new NoAuthProvider());
-    LOG(INFO) << "External communication is not authenticated";
+    LOG(INFO) << "External communication is not authenticated for binary protocols";
+    if (use_saml) {
+      SecureAuthProvider* sap = NULL;
+      external_http_auth_provider_.reset(sap = new SecureAuthProvider(false));
+      sap->InitSaml();
+      LOG(INFO) <<
+          "External communication is authenticated for hs2-http protocol with SAML2 SSO";
+    } else {
+      LOG(INFO) << "External communication is not authenticated for hs2-http protocol";
+    }
   }
 
   // Acquire a kerberos ticket and start the background renewal thread before starting
@@ -1306,6 +1400,12 @@ Status AuthManager::Init() {
   }
   RETURN_IF_ERROR(internal_auth_provider_->Start());
   RETURN_IF_ERROR(external_auth_provider_->Start());
+  if (FLAGS_external_fe_port > 0 ||
+      (TestInfo::is_test() && FLAGS_external_fe_port == 0)) {
+    external_fe_auth_provider_.reset(new NoAuthProvider());
+    LOG(INFO) << "External Frontend communication is not authenticated";
+    RETURN_IF_ERROR(external_fe_auth_provider_->Start());
+  }
   return Status::OK();
 }
 
@@ -1314,9 +1414,19 @@ AuthProvider* AuthManager::GetExternalAuthProvider() {
   return external_auth_provider_.get();
 }
 
+AuthProvider* AuthManager::GetExternalFrontendAuthProvider() {
+  DCHECK(external_fe_auth_provider_.get() != NULL);
+  return external_fe_auth_provider_.get();
+}
+
 AuthProvider* AuthManager::GetInternalAuthProvider() {
   DCHECK(internal_auth_provider_.get() != NULL);
   return internal_auth_provider_.get();
 }
 
+AuthProvider* AuthManager::GetExternalHttpAuthProvider() {
+  return external_http_auth_provider_.get() != nullptr ?
+      external_http_auth_provider_.get() : GetExternalAuthProvider();
+}
+
 }
