@@ -44,7 +44,6 @@ DECLARE_string(ssl_minimum_version);
 
 DECLARE_int32(state_store_port);
 
-DECLARE_int32(be_port);
 DECLARE_int32(beeswax_port);
 
 static string IMPALA_HOME(getenv("IMPALA_HOME"));
@@ -78,9 +77,9 @@ class DummyStatestoreService : public StatestoreServiceIf {
   }
 };
 
-boost::shared_ptr<TProcessor> MakeProcessor() {
-  boost::shared_ptr<DummyStatestoreService> service(new DummyStatestoreService());
-  return boost::shared_ptr<TProcessor>(new StatestoreServiceProcessor(service));
+std::shared_ptr<TProcessor> MakeProcessor() {
+  std::shared_ptr<DummyStatestoreService> service(new DummyStatestoreService());
+  return std::shared_ptr<TProcessor>(new StatestoreServiceProcessor(service));
 }
 
 int GetServerPort() {
@@ -274,6 +273,7 @@ TEST(SslTest, ClientBeforeServer) {
 
 TEST(SslTest, BadCiphers) {
   int port = GetServerPort();
+#ifndef __aarch64__
   {
     ThriftServer* server;
     EXPECT_OK(ThriftServerBuilder("DummyStatestore", MakeProcessor(), port)
@@ -282,14 +282,14 @@ TEST(SslTest, BadCiphers) {
                   .Build(&server));
     EXPECT_FALSE(server->Start().ok());
   }
-
+#endif
   {
     ThriftServer* server;
     EXPECT_OK(ThriftServerBuilder("DummyStatestore", MakeProcessor(), port)
                   .ssl(SERVER_CERT, PRIVATE_KEY)
                   .Build(&server));
     EXPECT_OK(server->Start());
-
+#ifndef __aarch64__
     auto s1 =
         ScopedFlagSetter<string>::Make(&FLAGS_ssl_cipher_list, "this_is_not_a_cipher");
     auto s2 =
@@ -298,6 +298,7 @@ TEST(SslTest, BadCiphers) {
     ThriftClient<StatestoreServiceClientWrapper> ssl_client(
         "localhost", port, "", nullptr, true);
     EXPECT_FALSE(ssl_client.Open().ok());
+#endif
   }
 }
 
@@ -320,12 +321,13 @@ TEST(SslTest, MismatchedCiphers) {
   // Failure to negotiate a cipher will show up when data is sent, not when socket is
   // opened.
   EXPECT_OK(ssl_client.Open());
-
+#ifndef __aarch64__
   bool send_done = false;
   TRegisterSubscriberResponse resp;
   EXPECT_THROW(ssl_client.iface()->RegisterSubscriber(
                    resp, TRegisterSubscriberRequest(), &send_done),
       TTransportException);
+#endif
 }
 
 // Test that StringToProtocol() correctly maps strings to their symbolic protocol
@@ -507,31 +509,6 @@ TEST(ConcurrencyTest, MaxConcurrentConnections) {
   EXPECT_TRUE(did_reach_max);
 }
 
-/// Test disabled because requires a high ulimit -n on build machines. Since the test does
-/// not always fail, we don't lose much coverage by disabling it until we fix the build
-/// infra issue.
-TEST(ConcurrencyTest, DISABLED_ManyConcurrentConnections) {
-  // Test that a large number of concurrent connections will all succeed and not time out
-  // waiting to be accepted. (IMPALA-4135)
-  // Note that without the fix for IMPALA-4135, this test won't always fail, depending on
-  // the hardware that it is run on.
-  int port = GetServerPort();
-  ThriftServer* server;
-  EXPECT_OK(ThriftServerBuilder("DummyServer", MakeProcessor(), port).Build(&server));
-  ASSERT_OK(server->Start());
-
-  ThreadPool<int64_t> pool(
-      "group", "test", 256, 10000, [port](int tid, const int64_t& item) {
-        using Client = ThriftClient<ImpalaInternalServiceClient>;
-        Client* client = new Client("127.0.0.1", port, "", nullptr, false);
-        Status status = client->Open();
-        ASSERT_OK(status);
-      });
-  ASSERT_OK(pool.Init());
-  for (int i = 0; i < 1024 * 16; ++i) pool.Offer(i);
-  pool.DrainAndShutdown();
-}
-
 TEST(NoPasswordPemFile, BadServerCertificate) {
   int port = GetServerPort();
   ThriftServer* server;
