@@ -18,6 +18,7 @@
 #include <boost/bind.hpp>
 #include <boost/filesystem.hpp>
 #include <boost/scoped_ptr.hpp>
+#include <boost/thread/thread.hpp>
 
 #include <limits> // for std::numeric_limits<int>::max()
 #include <set>
@@ -38,6 +39,7 @@
 #include "service/fe-support.h"
 #include "testutil/desc-tbl-builder.h"
 #include "testutil/gtest-util.h"
+#include "util/error-util.h"
 #include "util/test-info.h"
 
 #include "gen-cpp/ImpalaInternalService_types.h"
@@ -60,6 +62,8 @@ static const uint32_t PRIME = 479001599;
 
 namespace impala {
 
+constexpr int ErrorMsg::MAX_ERROR_MESSAGE_LEN;
+
 static const StringValue STRINGS[] = {
     StringValue("ABC"), StringValue("HELLO"), StringValue("123456789"),
     StringValue("FOOBAR"), StringValue("ONE"), StringValue("THREE"),
@@ -254,15 +258,31 @@ class SimpleTupleStreamTest : public testing::Test {
     }
   }
 
+  /// Read values from 'stream' into 'results' using the embedded read iterator. 'stream'
+  /// must have been prepared for reading.
   template <typename T>
   void ReadValues(BufferedTupleStream* stream, RowDescriptor* desc, vector<T>* results,
       int num_batches = -1) {
+    return ReadValues(stream, nullptr, desc, results, num_batches);
+  }
+
+  /// Read values from 'stream' into 'results'. If 'read_it' is non-NULL, reads via that
+  /// iterator. Otherwise use the embedded read iterator, in which case 'stream' must have
+  /// been prepared for reading.
+  template <typename T>
+  void ReadValues(BufferedTupleStream* stream,
+      BufferedTupleStream::ReadIterator* read_it, RowDescriptor* desc, vector<T>* results,
+      int num_batches = -1) {
     bool eos = false;
     RowBatch batch(desc, BATCH_SIZE, &tracker_);
     int batches_read = 0;
     do {
       batch.Reset();
-      EXPECT_OK(stream->GetNext(&batch, &eos));
+      if (read_it != nullptr) {
+        EXPECT_OK(stream->GetNext(read_it, &batch, &eos));
+      } else {
+        EXPECT_OK(stream->GetNext(&batch, &eos));
+      }
       ++batches_read;
       for (int i = 0; i < batch.num_rows(); ++i) {
         AppendRowTuples(batch.GetRow(i), desc, results);
@@ -531,11 +551,11 @@ class ArrayTupleStreamTest : public SimpleTupleStreamTest {
     tuple_ids.push_back(static_cast<TTupleId>(1));
     ColumnType string_array_type;
     string_array_type.type = TYPE_ARRAY;
-    string_array_type.children.push_back(TYPE_STRING);
+    string_array_type.children.push_back(ColumnType(TYPE_STRING));
 
     ColumnType int_array_type;
     int_array_type.type = TYPE_ARRAY;
-    int_array_type.children.push_back(TYPE_STRING);
+    int_array_type.children.push_back(ColumnType(TYPE_STRING));
 
     ColumnType nested_array_type;
     nested_array_type.type = TYPE_ARRAY;
@@ -837,7 +857,9 @@ void SimpleTupleStreamTest::TestAttachMemory(bool pin_stream, bool attach_on_rea
   } else {
     EXPECT_EQ(0, num_buffers_attached) << "No buffers attached during iteration.";
   }
-  if (attach_on_read || !pin_stream) EXPECT_EQ(4, num_flushes);
+  if (attach_on_read || !pin_stream) {
+    EXPECT_EQ(4, num_flushes);
+  }
   out_batch->Reset();
   stream.Close(out_batch, RowBatch::FlushMode::FLUSH_RESOURCES);
   if (attach_on_read) {
@@ -1173,7 +1195,7 @@ TEST_F(SimpleTupleStreamTest, BigStringReadWrite) {
     bool eos;
     ASSERT_OK(stream.GetNext(&read_batch, &eos));
     EXPECT_EQ(1, read_batch.num_rows());
-    EXPECT_TRUE(eos);
+    EXPECT_TRUE(eos) << i << " " << stream.DebugString();
     Tuple* tuple = read_batch.GetRow(0)->GetTuple(0);
     StringValue* str = tuple->GetStringSlot(tuple_desc->slots()[0]->tuple_offset());
     EXPECT_EQ(string_len, str->len);
@@ -1275,6 +1297,76 @@ TEST_F(SimpleTupleStreamTest, UnpinReadPage) {
   write_batch->Reset();
 }
 
+// Test that UnpinStream defer advancing the read page when all rows from the read page
+// are attached to a returned RowBatch but got not enough reservation.
+TEST_F(SimpleTupleStreamTest, DeferAdvancingReadPage) {
+  int num_rows = 1024;
+  int buffer_size = 4 * 1024;
+  // Only give 2 * buffer_size for the stream initial read and write page reservation.
+  Init(2 * buffer_size);
+
+  bool eos;
+  bool got_reservation;
+  Status status;
+  RowBatch* write_batch = CreateIntBatch(0, num_rows, false);
+
+  {
+    // Test unpinning a stream when the read page has been attached to the output batch
+    // and the output batch has NOT been reset.
+    BufferedTupleStream stream(
+        runtime_state_, int_desc_, &client_, buffer_size, buffer_size);
+    ASSERT_OK(stream.Init("SimpleTupleStreamTest::DeferAdvancingReadPage", true));
+    ASSERT_OK(stream.PrepareForReadWrite(true, &got_reservation));
+    ASSERT_TRUE(got_reservation);
+
+    // Add rows to stream.
+    for (int i = 0; i < write_batch->num_rows(); ++i) {
+      EXPECT_TRUE(stream.AddRow(write_batch->GetRow(i), &status));
+      ASSERT_OK(status);
+    }
+
+    // Read until the read page is attached to the output.
+    RowBatch read_batch(int_desc_, num_rows, &tracker_);
+    ASSERT_OK(stream.GetNext(&read_batch, &eos));
+    // If GetNext did hit the capacity of the RowBatch, then the read page should have
+    // been attached to read_batch.
+    ASSERT_TRUE(read_batch.num_rows() < num_rows);
+    ASSERT_TRUE(!eos);
+
+    // We continue adding rows into the stream without releasing the read_batch. We expect
+    // that reservation limit will be hit and stream will need to be unpinned. We also
+    // expect that, after unpinning the stream, subsequent AddRow is always successful
+    // even if we're not immediately releasing the read_batch. We insert write_batch twice
+    // to ensure that we're inserting both in pinned and unpinned mode.
+    ASSERT_TRUE(stream.is_pinned());
+    for (int j = 0; j < 2; ++j) {
+      for (int i = 0; i < write_batch->num_rows(); ++i) {
+        bool succeed = stream.AddRow(write_batch->GetRow(i), &status);
+        ASSERT_TRUE(succeed || stream.is_pinned());
+        if (!succeed) {
+          // Unpin the stream.
+          status = stream.UnpinStream(BufferedTupleStream::UNPIN_ALL_EXCEPT_CURRENT);
+          ASSERT_OK(status);
+          ASSERT_FALSE(stream.is_pinned());
+          ASSERT_EQ(stream.bytes_unpinned(), 0);
+          ASSERT_EQ(stream.pages_.size(), 2);
+          ASSERT_EQ(stream.num_pages_, 2);
+          // Retry inserting this row by decreasing the index.
+          // After stream get into unpinned mode, further inserts should be successful,
+          // even if we're not immediately cleaning up the read_batch.
+          // Stream should be able to unpin the previous write page to reclaim some
+          // memory reservation to allocate new write page.
+          --i;
+        }
+      }
+    }
+    ASSERT_FALSE(stream.is_pinned());
+    stream.Close(nullptr, RowBatch::FlushMode::NO_FLUSH_RESOURCES);
+    read_batch.Reset();
+  }
+  write_batch->Reset();
+}
+
 // Test writing to a stream (AddRow and UnpinStream), even though attached pages have not
 // been released yet.
 TEST_F(SimpleTupleStreamTest, WriteAfterReadAttached) {
@@ -1362,6 +1454,138 @@ TEST_F(SimpleTupleStreamTest, WriteAfterReadAttached) {
   write_batch->Reset();
 }
 
+/// Test multiple threads reading from a pinned stream using separate read iterators.
+TEST_F(SimpleTupleStreamTest, ConcurrentReaders) {
+  const int BUFFER_SIZE = 1024;
+  // Each tuple is an integer plus a null indicator byte.
+  const int VALS_PER_BUFFER = BUFFER_SIZE / (sizeof(int32_t) + 1);
+  const int NUM_BUFFERS = 100;
+  const int TOTAL_MEM = NUM_BUFFERS * BUFFER_SIZE;
+  Init(TOTAL_MEM);
+  BufferedTupleStream stream(
+      runtime_state_, int_desc_, &client_, BUFFER_SIZE, BUFFER_SIZE);
+  ASSERT_OK(stream.Init("ConcurrentReaders", true));
+  bool got_write_reservation;
+  ASSERT_OK(stream.PrepareForWrite(&got_write_reservation));
+  ASSERT_TRUE(got_write_reservation);
+
+  // Add rows to the stream.
+  int offset = 0;
+  const int NUM_BATCHES = NUM_BUFFERS;
+  const int ROWS_PER_BATCH = VALS_PER_BUFFER;
+  for (int i = 0; i < NUM_BATCHES; ++i) {
+    RowBatch* batch = nullptr;
+    Status status;
+    batch = CreateBatch(int_desc_, offset, ROWS_PER_BATCH, false);
+    for (int j = 0; j < batch->num_rows(); ++j) {
+      bool b = stream.AddRow(batch->GetRow(j), &status);
+      ASSERT_OK(status);
+      ASSERT_TRUE(b);
+    }
+    offset += batch->num_rows();
+    // Reset the batch to make sure the stream handles the memory correctly.
+    batch->Reset();
+  }
+  // Invalidate the write iterator explicitly so that we can read concurrently.
+  stream.DoneWriting();
+
+  const int READ_ITERS = 10; // Do multiple read passes per thread.
+  const int NUM_THREADS = 4;
+
+  // Read from the main thread with the built-in iterator and other threads with
+  // external iterators.
+  thread_group workers;
+  for (int i = 0; i < NUM_THREADS; ++i) {
+    workers.add_thread(new thread([&] () {
+      for (int j = 0; j < READ_ITERS; ++j) {
+        BufferedTupleStream::ReadIterator it;
+        ASSERT_OK(stream.PrepareForPinnedRead(&it));
+
+        // Read all the rows back
+        vector<int> results;
+        ReadValues(&stream, &it, int_desc_, &results);
+
+        // Verify result
+        VerifyResults<int>(*int_desc_, results, ROWS_PER_BATCH * NUM_BATCHES, false);
+      }
+    }));
+  }
+
+  for (int i = 0; i < READ_ITERS; ++i) {
+    bool got_read_reservation;
+    ASSERT_OK(stream.PrepareForRead(false, &got_read_reservation));
+    ASSERT_TRUE(got_read_reservation);
+
+    // Read all the rows back
+    vector<int> results;
+    ReadValues(&stream, int_desc_, &results);
+
+    // Verify result
+    VerifyResults<int>(*int_desc_, results, ROWS_PER_BATCH * NUM_BATCHES, false);
+  }
+  workers.join_all();
+  stream.Close(nullptr, RowBatch::FlushMode::NO_FLUSH_RESOURCES);
+}
+
+TEST_F(SimpleTupleStreamTest, ShortDebugString) {
+  Init(BUFFER_POOL_LIMIT);
+
+  int num_batches = 50;
+  RowDescriptor* desc = int_desc_;
+  bool gen_null = false;
+  int64_t default_page_len = 128 * sizeof(int);
+  int64_t max_page_len = default_page_len;
+  int num_rows = BATCH_SIZE;
+
+  BufferedTupleStream stream(
+      runtime_state_, desc, &client_, default_page_len, max_page_len);
+  ASSERT_OK(stream.Init("SimpleTupleStreamTest::ShortDebugString", true));
+  bool got_write_reservation;
+  ASSERT_OK(stream.PrepareForWrite(&got_write_reservation));
+  ASSERT_TRUE(got_write_reservation);
+
+  // Add rows to the stream
+  int offset = 0;
+  for (int i = 0; i < num_batches; ++i) {
+    RowBatch* batch = nullptr;
+
+    Status status;
+    batch = CreateBatch(desc, offset, num_rows, gen_null);
+    for (int j = 0; j < batch->num_rows(); ++j) {
+      bool b = stream.AddRow(batch->GetRow(j), &status);
+      ASSERT_OK(status);
+      ASSERT_TRUE(b);
+    }
+    offset += batch->num_rows();
+    // Reset the batch to make sure the stream handles the memory correctly.
+    batch->Reset();
+  }
+
+  bool got_read_reservation;
+  ASSERT_OK(stream.PrepareForRead(false, &got_read_reservation));
+  ASSERT_TRUE(got_read_reservation);
+
+  // Read all the rows back
+  vector<int> results;
+  ReadValues(&stream, desc, &results);
+
+  // Verify result
+  VerifyResults<int>(*desc, results, num_rows * num_batches, gen_null);
+
+  // Verify that stream contains more than MAX_PAGE_ITER_DEBUG pages and only subset of
+  // pages are included in DebugString().
+  DCHECK_GT(stream.num_pages_, BufferedTupleStream::MAX_PAGE_ITER_DEBUG);
+  string page_count_substr = Substitute(
+      "$0 out of $1 pages=", BufferedTupleStream::MAX_PAGE_ITER_DEBUG, stream.num_pages_);
+  string debug_string = stream.DebugString();
+  ASSERT_NE(debug_string.find(page_count_substr), string::npos)
+      << page_count_substr << " not found at BufferedTupleStream::DebugString(). "
+      << debug_string;
+  ASSERT_LE(debug_string.length(), ErrorMsg::MAX_ERROR_MESSAGE_LEN);
+
+  stream.Close(nullptr, RowBatch::FlushMode::NO_FLUSH_RESOURCES);
+}
+
 // Basic API test. No data should be going to disk.
 TEST_F(SimpleNullStreamTest, Basic) {
   Init(BUFFER_POOL_LIMIT);
