@@ -82,6 +82,8 @@
 #include "runtime/bufferpool/buffer-pool.h"
 #include "runtime/bufferpool/reservation-tracker.h"
 #include "util/condition-variable.h"
+#include "util/internal-queue.h"
+#include "util/spinlock.h"
 
 // Ensure that DCheckConsistency() function calls get removed in release builds.
 #ifndef NDEBUG
@@ -92,11 +94,16 @@
 
 namespace impala {
 
+class TmpFileGroup;
+class TmpWriteHandle;
+
 /// The internal representation of a page, which can be pinned or unpinned. See the
 /// class comment for explanation of the different page states.
 struct BufferPool::Page : public InternalList<Page>::Node {
-  Page(Client* client, int64_t len)
-    : client(client), len(len), pin_count(0), pin_in_flight(false) {}
+  // Define constructor and destructor out-of-line to avoid include of TmpWriteHandle
+  // body in header.
+  Page(Client* client, int64_t len);
+  ~Page();
 
   std::string DebugString();
 
@@ -114,12 +121,14 @@ struct BufferPool::Page : public InternalList<Page>::Node {
   int pin_count;
 
   /// True if the read I/O to pin the page was started but not completed. Only accessed
-  /// in contexts that are passed the associated PageHandle, so it cannot be accessed
-  /// by multiple threads concurrently.
-  bool pin_in_flight;
+  /// in contexts that are passed the associated PageHandle, so can only be accessed
+  /// by multiple threads concurrently via PageHandle::GetBuffer(), since other page
+  /// handle operators are not thread-safe. This is atomic so that GetBuffer() can do
+  /// optimistic checks to avoid acquiring 'buffer_lock'.
+  AtomicBool pin_in_flight;
 
   /// Non-null if there is a write in flight, the page is clean, or the page is evicted.
-  std::unique_ptr<TmpFileMgr::WriteHandle> write_handle;
+  std::unique_ptr<TmpWriteHandle> write_handle;
 
   /// Condition variable signalled when a write for this page completes. Protected by
   /// client->lock_.
@@ -173,6 +182,9 @@ class BufferPool::PageList {
   }
 
   void Iterate(boost::function<bool(Page*)> fn) { list_.Iterate(fn); }
+  void IterateFirstN(boost::function<bool(Page*)> fn, int n) {
+    list_.IterateFirstN(fn, n);
+  }
   bool Contains(Page* page) { return list_.Contains(page); }
   Page* tail() { return list_.tail(); }
   bool empty() const { return list_.empty(); }
@@ -192,7 +204,7 @@ class BufferPool::PageList {
 /// The internal state for the client.
 class BufferPool::Client {
  public:
-  Client(BufferPool* pool, TmpFileMgr::FileGroup* file_group, const string& name,
+  Client(BufferPool* pool, TmpFileGroup* file_group, const string& name,
       ReservationTracker* parent_reservation, MemTracker* mem_tracker,
       MemLimit mem_limit_mode, int64_t reservation_limit, RuntimeProfile* profile);
 
@@ -342,7 +354,7 @@ class BufferPool::Client {
 
   /// The file group that should be used for allocating scratch space. If NULL, spilling
   /// is disabled.
-  TmpFileMgr::FileGroup* const file_group_;
+  TmpFileGroup* const file_group_;
 
   /// A name identifying the client.
   const std::string name_;
@@ -364,6 +376,10 @@ class BufferPool::Client {
   /// Condition variable signalled when a write for this client completes.
   ConditionVariable write_complete_cv_;
 
+  /// Used to ensure that only one thread at a time is active in CleanPages().
+  bool cleaning_pages_ = false;
+  ConditionVariable clean_pages_done_cv_;
+
   /// All non-OK statuses returned by write operations are merged into this status.
   /// All operations that depend on pages being written to disk successfully (e.g.
   /// reading pages back from disk) must check 'write_status_' before proceeding, so
