@@ -29,10 +29,10 @@
 #include <boost/accumulators/statistics/stats.hpp>
 #include <boost/accumulators/statistics/variance.hpp>
 
+#include "gen-cpp/admission_control_service.pb.h"
 #include "gen-cpp/control_service.proxy.h"
 #include "kudu/rpc/rpc_controller.h"
 #include "runtime/coordinator.h"
-#include "scheduling/query-schedule.h"
 #include "util/error-util-internal.h"
 #include "util/progress-updater.h"
 #include "util/runtime-profile.h"
@@ -51,7 +51,6 @@ class ReportExecStatusRequestPB;
 class TUniqueId;
 class TQueryCtx;
 class ExecSummary;
-struct FInstanceExecParams;
 
 /// This class manages all aspects of the execution of all fragment instances of a
 /// single query on a particular backend. For the coordinator backend its possible to have
@@ -68,16 +67,17 @@ struct FInstanceExecParams;
 ///   subject to the restrictions above on when methods may be called.
 class Coordinator::BackendState {
  public:
-  BackendState(const QuerySchedule& schedule, const TQueryCtx& query_ctx, int state_idx,
-      TRuntimeFilterMode::type filter_mode, const BackendExecParams& exec_params);
+  BackendState(const QueryExecParams& exec_params, int state_idx,
+      TRuntimeFilterMode::type filter_mode,
+      const BackendExecParamsPB& backend_exec_params);
 
   /// The following are initialized in the constructor and always valid to call.
   int state_idx() const { return state_idx_; }
-  const BackendExecParams* exec_params() const { return backend_exec_params_; }
-  const TNetworkAddress& impalad_address() const { return host_; }
-  const TNetworkAddress& krpc_impalad_address() const { return krpc_host_; }
+  const BackendExecParamsPB& exec_params() const { return backend_exec_params_; }
+  const NetworkAddressPB& impalad_address() const { return host_; }
+  const NetworkAddressPB& krpc_impalad_address() const { return krpc_host_; }
   /// Returns true if there are no fragment instances scheduled on this backend.
-  bool IsEmptyBackend() const { return backend_exec_params_->instance_params.empty(); }
+  bool IsEmptyBackend() const { return backend_exec_params_.instance_params().empty(); }
   /// Return true if execution at this backend is done.
   bool IsDone();
 
@@ -107,13 +107,28 @@ class Coordinator::BackendState {
   /// times and from different threads.
   void WaitOnExecRpc();
 
-  /// Cancel execution at this backend if anything is running. Returns true if
-  /// cancellation was attempted, false otherwise.
+  /// Result of attempting to cancel a backend.
+  struct CancelResult {
+    // Whether we tried to cancel the backend in some fashion, e.g. attempted to send
+    // an RPC. This is for informational purpose only (logging, metrics, etc), not
+    // control flow.
+    bool cancel_attempted = false;
+
+    // True iff this changed IsDone() from false to true.
+    bool became_done = false;
+  };
+
+  /// Cancel execution at this backend if anything is running. See CancelResult
+  /// for explanation of the return value.
   ///
   /// May be called at any time after Init(). If the ExecQueryFInstances() rpc is
   /// inflight, will attempt to cancel the rpc. If ExecQueryFInstances() has already
   /// completed or cancelling it is unsuccessful, sends the Cancel() rpc.
-  bool Cancel();
+  /// If 'fire_and_forget' is true, the RPC is sent and the backend is immediately
+  /// considered done, without waiting for a final status report. If 'fire_and_forget'
+  /// is false, the backend is only considered done once the final status report is
+  /// received.
+  CancelResult Cancel(bool fire_and_forget);
 
   /////////////////////////////////////////
   /// BEGIN: Functions that should only be called after WaitOnExecRpc() has returned.
@@ -121,22 +136,29 @@ class Coordinator::BackendState {
   /// Update overall execution status, including the instances' exec status/profiles
   /// and the error log, if this backend is not already done. Updates the fragment
   /// instances' TExecStats in exec_summary (exec_summary->nodes.exec_stats) and updates
-  /// progress_update. If any instance reports an error, the overall execution status
-  /// becomes the first reported error status. Returns true iff this update changed
-  /// IsDone() from false to true, either because it was the last fragment to complete or
-  /// because it was the first error received. Adds the AuxErrorInfoPB from each
-  /// FragmentInstanceExecStatusPB in backend_exec_status to the vector
-  /// aux_error_info.
+  /// scan_range_progress with any newly-completed scan ranges.
+  ///
+  /// If any instance reports an error, the overall execution status becomes the first
+  /// reported error status. Returns true iff this update changed IsDone() from false
+  /// to true, either because it was the last fragment to complete or because it was
+  /// the first error received. Adds the AuxErrorInfoPB from each
+  /// FragmentInstanceExecStatusPB in backend_exec_status to the vector aux_error_info.
   bool ApplyExecStatusReport(const ReportExecStatusRequestPB& backend_exec_status,
       const TRuntimeProfileForest& thrift_profiles, ExecSummary* exec_summary,
       ProgressUpdater* scan_range_progress, DmlExecState* dml_exec_state,
-      std::vector<AuxErrorInfoPB>* aux_error_info);
+      std::vector<AuxErrorInfoPB>* aux_error_info,
+      const std::vector<FragmentStats*>& fragment_stats);
 
   /// Merges the incoming 'thrift_profile' into this backend state's host profile.
   void UpdateHostProfile(const TRuntimeProfileTree& thrift_profile);
 
-  /// Update completion_times, rates, and avg_profile for all fragment_stats.
-  void UpdateExecStats(const std::vector<FragmentStats*>& fragment_stats);
+  /// Update completion_times, rates, and agg_profile for instances where the stats are
+  /// not up-to-date with the latest status report received for that instance. Only
+  /// updates for completed instances when --gen_experimental_profile=false, consistent
+  /// with past behaviour, to avoid adding overhead.  If 'finalize' is true, this is the
+  /// last call to this function for the query and we must update all instances,
+  /// regardless of whether they are done or not.
+  void UpdateExecStats(const std::vector<FragmentStats*>& fragment_stats, bool finalize);
 
   /// Make a PublishFilter rpc with given params to this backend. The backend
   /// must be one of the filter targets for the filter being published.
@@ -162,7 +184,12 @@ class Coordinator::BackendState {
 
   /// Return peak memory consumption and aggregated resource usage across all fragment
   /// instances for this backend.
-  ResourceUtilization ComputeResourceUtilization();
+  ResourceUtilization GetResourceUtilization();
+
+  bool IsLocalDiskFaulty() {
+    std::lock_guard<std::mutex> l(lock_);
+    return local_disk_faulty_;
+  }
 
   /// Merge the accumulated error log into 'merged'.
   void MergeErrorLog(ErrorLogMap* merged);
@@ -207,19 +234,32 @@ class Coordinator::BackendState {
   /// Not thread-safe.
   class InstanceStats {
    public:
-    InstanceStats(const FInstanceExecParams& exec_params, FragmentStats* fragment_stats,
+    InstanceStats(const FInstanceExecParamsPB& exec_params, const TPlanFragment* fragment,
+        const NetworkAddressPB& address, FragmentStats* fragment_stats,
         ObjectPool* obj_pool);
 
-    /// Updates 'this' with exec_status and the fragment intance's thrift profile. Also
-    /// updates the fragment instance's TExecStats in exec_summary and 'progress_updater'
-    /// with the number of newly completed scan ranges. Also updates the instance's avg
-    /// profile. Caller must hold BackendState::lock_.
+    /// Updates 'this' with exec_status and, if --gen_experimental_profile=false, the
+    /// fragment instance's thrift profile. Also updates the fragment instance's
+    /// TExecStats in exec_summary. Also updates the instance's avg profile.
+    /// 'report_time_ms' should be the UnixMillis() value of when the report was received.
+    /// Caller must hold BackendState::lock_.
     void Update(const FragmentInstanceExecStatusPB& exec_status,
-        const TRuntimeProfileTree& thrift_profile, ExecSummary* exec_summary,
-        ProgressUpdater* scan_range_progress);
+        const TRuntimeProfileTree* thrift_profile, int64_t report_time_ms,
+        ExecSummary* exec_summary);
+
+    /// Update completion_times, rates, and agg_profile for this instance's
+    /// corresponding fragment in 'fragment_stats'. This may only be called
+    /// once the exec RPC is done, i.e. 'BackendState::exec_done_' is true.
+    /// This depends on the status report applied in Update(), so can be
+    /// called sometime after each Update() call to refresh the stats. It
+    /// may make sense to defer the updates in some cases because the amount of work
+    /// involved, particularly in merging RuntimeProfiles, is not trivial.
+    /// If 'finalize' is true, this is the last call to this function for the query
+    /// and we must update the instance, whether it is done or not.
+    void UpdateExecStats(const vector<FragmentStats*>& fragment_stats, bool finalize);
 
     int per_fragment_instance_idx() const {
-      return exec_params_.per_fragment_instance_idx;
+      return exec_params_.per_fragment_instance_idx();
     }
 
     /// Serializes instance stats to JSON by adding members to 'value', including its
@@ -230,8 +270,12 @@ class Coordinator::BackendState {
    private:
     friend class BackendState;
 
-    /// query lifetime
-    const FInstanceExecParams& exec_params_;
+    /// The exec params for this fragment instance. Owned by the QuerySchedulePB in
+    /// ClientRequestState.
+    const FInstanceExecParamsPB& exec_params_;
+
+    /// The corresponding fragment. Owned by the TExecRequest in QueryDriver.
+    const TPlanFragment* fragment_;
 
     /// Unix time in milliseconds of the last status report update for this fragment
     /// instance. Set in Update(). Uses UnixMillis() instead of MonotonicMillis() as
@@ -244,57 +288,55 @@ class Coordinator::BackendState {
     /// persists for the duration of one status report update (5 seconds by default).
     int64_t last_report_time_ms_ = 0;
 
-    /// The sequence number of the last report.
+    /// The sequence number of the last report for this fragment instance.
     int64_t last_report_seq_no_ = 0;
 
-    /// owned by coordinator object pool provided in the c'tor, created in Update()
+    /// The profile tree for this instance. Owned by coordinator object pool provided in
+    /// the c'tor, created in Update().
+    /// When --gen_experimental_profile=true, this contains a copy of the full instance
+    /// profile from the executor, which is updated with each status report. It also
+    /// includes some additional counters added by the coordinator.
+    /// When --gen_experimental_profile=false, this only includes the counters added
+    /// by the coordinator.
     RuntimeProfile* profile_ = nullptr;
 
+    /// True if 'agg_profile_' for the fragment is up-to-date with 'profile_'.
+    /// Set to false in Update() when a new profile is received for this instance
+    /// and back to true in UpdateExecStats() when the updated profile is merged into
+    /// 'agg_profile_'.
+    bool agg_profile_up_to_date_ = true;
+
+    /// If true, the completion time for this instance has been set.
+    bool completion_time_set_ = false;
+
     /// true if the final report has been received for the fragment instance.
     /// Used to handle duplicate done ReportExecStatus RPC messages. Used only
     /// in ApplyExecStatusReport()
     bool done_ = false;
 
-    /// true after the first call to profile->Update()
-    bool profile_created_ = false;
-
     /// cumulative size of all splits; set in c'tor
     int64_t total_split_size_ = 0;
 
     /// wall clock timer for this instance
     MonotonicStopWatch stopwatch_;
 
-    /// total scan ranges complete across all scan nodes
-    int64_t total_ranges_complete_ = 0;
-
-    /// SCAN_RANGES_COMPLETE_COUNTERs in profile_
-    std::vector<RuntimeProfile::Counter*> scan_ranges_complete_counters_;
-
-    /// Collection of BYTES_READ_COUNTERs of all scan nodes in this fragment instance.
-    std::vector<RuntimeProfile::Counter*> bytes_read_counters_;
-
-    /// Collection of TotalBytesSent of all data stream senders in this fragment instance.
-    std::vector<RuntimeProfile::Counter*> bytes_sent_counters_;
-
     /// Descriptor string for the last query status report time in the profile.
     static const char* LAST_REPORT_TIME_DESC;
 
     /// The current state of this fragment instance's execution. This gets serialized in
     /// ToJson() and is displayed in the debug webpages.
     FInstanceExecStatePB current_state_ = FInstanceExecStatePB::WAITING_FOR_EXEC;
-
-    /// Extracts scan_ranges_complete_counters_ and  bytes_read_counters_ from profile_.
-    void InitCounters();
   };
 
-  /// QuerySchedule associated with the Coordinator that owns this BackendState.
-  const QuerySchedule& schedule_;
+  /// ExecParams associated with the Coordinator that owns this BackendState.
+  const QueryExecParams& exec_params_;
 
   const int state_idx_;  /// index of 'this' in Coordinator::backend_states_
   const TRuntimeFilterMode::type filter_mode_;
 
-  /// Backend exec params, owned by the QuerySchedule and has query lifetime.
-  const BackendExecParams* backend_exec_params_;
+  /// Backend exec params, owned by the QuerySchedulePB in ClientRequestState and has
+  /// query lifetime.
+  const BackendExecParamsPB& backend_exec_params_;
 
   /// map from instance idx to InstanceStats, the latter live in the obj_pool parameter
   /// of Init()
@@ -308,16 +350,17 @@ class Coordinator::BackendState {
   /// Owned by coordinator object pool provided in the c'tor, created in Update().
   RuntimeProfile* host_profile_ = nullptr;
 
-  /// Thrift address of execution backend.
-  TNetworkAddress host_;
-  /// Krpc address of execution backend.
-  TNetworkAddress krpc_host_;
+  /// Address of execution backend: hostname + krpc_port.
+  NetworkAddressPB host_;
+  /// Krpc address of execution backend: ip_address + krpc_port.
+  NetworkAddressPB krpc_host_;
 
   /// The query context of the Coordinator that owns this BackendState.
   const TQueryCtx& query_ctx_;
 
-  /// The query id of the Coordinator that owns this BackendState.
-  const TUniqueId& query_id_;
+  /// The query id of the Coordinator that owns this BackendState. Owned by
+  /// 'exec_params_'.
+  const UniqueIdPB& query_id_;
 
   /// Used to issue the ExecQueryFInstances() rpc.
   kudu::rpc::RpcController exec_rpc_controller_;
@@ -356,6 +399,10 @@ class Coordinator::BackendState {
   /// Invalid if no fragment instance has reported an error status.
   TUniqueId failed_instance_id_;
 
+  /// If true, the backend reported that the query failure was caused by disk IO error
+  /// on its local disk.
+  bool local_disk_faulty_ = false;
+
   /// Errors reported by this fragment instance.
   ErrorLogMap error_log_;
 
@@ -374,6 +421,21 @@ class Coordinator::BackendState {
   /// Uses GenerateReportTimeout().
   int64_t last_report_time_ms_ = 0;
 
+  /// The sequence number of the last report for this backend.
+  int64_t last_backend_report_seq_no_ = 0;
+
+  /// True if a CancelQueryFInstances RPC was already sent to this backend.
+  bool sent_cancel_rpc_ = false;
+
+  /// True if Exec() RPC is cancelled.
+  bool cancel_exec_rpc_ = false;
+
+  /// Total scan ranges complete across all scan nodes. Set in ApplyExecStatusReport().
+  int64_t total_ranges_complete_ = 0;
+
+  /// Resource utilization values. Set in ApplyExecStatusReport().
+  ResourceUtilization backend_utilization_;
+
   /// END: Members that are protected by 'lock_'.
   /////////////////////////////////////////
 
@@ -399,8 +461,12 @@ class Coordinator::BackendState {
   /// Version of IsDone() where caller must hold lock_ via lock;
   bool IsDoneLocked(const std::unique_lock<std::mutex>& lock) const;
 
-  /// Same as ComputeResourceUtilization() but caller must hold lock.
-  ResourceUtilization ComputeResourceUtilizationLocked();
+  /// Same as GetResourceUtilization() but caller must hold lock.
+  ResourceUtilization GetResourceUtilizationLocked();
+
+  /// Implementation of UpdateExecStats(). Caller must hold 'lock_' via 'lock'.
+  void UpdateExecStatsLocked(const std::unique_lock<std::mutex>& lock,
+      const std::vector<FragmentStats*>& fragment_stats, bool finalize);
 
   /// Logs 'msg' at the VLOG_QUERY level, along with 'query_id_' and 'krpc_host_'.
   void VLogForBackend(const std::string& msg);
@@ -418,16 +484,15 @@ class Coordinator::FragmentStats {
       boost::accumulators::tag::variance>
   > SummaryStats;
 
-  /// Create avg and root profiles in obj_pool.
-  FragmentStats(const std::string& avg_profile_name,
-      const std::string& root_profile_name,
+  /// Create aggregated and root profiles in obj_pool.
+  FragmentStats(const std::string& agg_profile_name, const std::string& root_profile_name,
       int num_instances, ObjectPool* obj_pool);
 
-  RuntimeProfile* avg_profile() { return avg_profile_; }
+  AggregatedRuntimeProfile* agg_profile() { return agg_profile_; }
   RuntimeProfile* root_profile() { return root_profile_; }
   SummaryStats* bytes_assigned() { return &bytes_assigned_; }
 
-  /// Compute stats for 'bytes_assigned' and add as info string to avg_profile.
+  /// Compute stats for 'bytes_assigned' and add as info string to agg_profile.
   void AddSplitStats();
 
   /// Add summary string with execution stats to avg profile.
@@ -441,7 +506,7 @@ class Coordinator::FragmentStats {
   /// counters in the fragment instance profiles.
   /// Note that the individual fragment instance profiles themselves are stored and
   /// displayed as children of the root_profile below.
-  RuntimeProfile* avg_profile_;
+  AggregatedRuntimeProfile* agg_profile_;
 
   /// root profile for all fragment instances for this fragment; resides in obj_pool
   RuntimeProfile* root_profile_;
@@ -450,12 +515,15 @@ class Coordinator::FragmentStats {
   int num_instances_;
 
   /// Bytes assigned for instances of this fragment
+  /// TODO: IMPALA-9846: can remove when we switch to the transposed profile.
   SummaryStats bytes_assigned_;
 
   /// Completion times for instances of this fragment
+  /// TODO: IMPALA-9846: can remove when we switch to the transposed profile.
   SummaryStats completion_times_;
 
   /// Execution rates for instances of this fragment
+  /// TODO: IMPALA-9846: can remove when we switch to the transposed profile.
   SummaryStats rates_;
 };
 
@@ -518,8 +586,7 @@ class Coordinator::BackendResourceState {
  public:
   /// Create the BackendResourceState with the given vector of BackendStates. All
   /// BackendStates are initially in the IN_USE state.
-  BackendResourceState(
-      const std::vector<BackendState*>& backend_states, const QuerySchedule& schedule);
+  BackendResourceState(const std::vector<BackendState*>& backend_states);
   ~BackendResourceState();
 
   /// Mark a BackendState as finished and transition it to the PENDING state. Applies
@@ -582,9 +649,6 @@ class Coordinator::BackendResourceState {
   // True if the BackendResourceState is closed, false otherwise.
   bool closed_ = false;
 
-  /// QuerySchedule associated with the Coordinator that owns this BackendResourceState.
-  const QuerySchedule& schedule_;
-
   /// Configured value of FLAGS_release_backend_states_delay_ms in nanoseconds.
   const int64_t release_backend_states_delay_ns_;
 
