@@ -31,10 +31,13 @@
 #include "gen-cpp/Types_types.h"
 #include "gen-cpp/data_stream_service.pb.h"
 #include "runtime/dml-exec-state.h"
+#include "runtime/query-exec-params.h"
 #include "util/counting-barrier.h"
 #include "util/progress-updater.h"
+#include "util/promise.h"
 #include "util/runtime-profile-counters.h"
 #include "util/spinlock.h"
+#include "util/container-util.h"
 
 namespace kudu {
 namespace rpc {
@@ -46,12 +49,13 @@ namespace impala {
 
 class AuxErrorInfoPB;
 class ClientRequestState;
+class FragmentExecParamsPB;
 class FragmentInstanceState;
 class MemTracker;
 class ObjectPool;
 class PlanRootSink;
+class QueryDriver;
 class QueryResultSet;
-class QuerySchedule;
 class QueryState;
 class ReportExecStatusRequestPB;
 class RuntimeProfile;
@@ -59,7 +63,6 @@ class RuntimeState;
 class TPlanExecRequest;
 class TRuntimeProfileTree;
 class TUpdateCatalogRequest;
-struct FragmentExecParams;
 
 /// Query coordinator: handles execution of fragment instances on remote nodes, given a
 /// TQueryExecRequest. As part of that, it handles all interactions with the executing
@@ -92,10 +95,16 @@ struct FragmentExecParams;
 ///    executing (without an error status).
 /// 3. The query has returned all rows. The overall query status is OK (and remains
 ///    OK). Client cancellation is no longer possible and subsequent backend errors are
-///    ignored. (TODO: IMPALA-6984 initiate backend cancellation in this case).
+///    ignored. The coordinator initiates cancellation for all backends and waits for the
+///    final exec status reports.
 ///
 /// Lifecycle: this object must not be destroyed until after one of the three states
 /// above is reached (error, cancelled, or EOS) to ensure resources are released.
+/// There is work that needs to be done as part of the transition into each of the
+/// above states. That work is done by the thread that triggered the state transition
+/// *after* the atomic state transition by calling HandleExecStateTransition(). If
+/// another thread depends on that work being complete, they can wait on the 'finalized_'
+/// promise.
 ///
 /// Lock ordering: (lower-numbered acquired before higher-numbered)
 /// 1. wait_lock_
@@ -107,8 +116,8 @@ struct FragmentExecParams;
 /// and unnest them
 class Coordinator { // NOLINT: The member variables could be re-ordered to save space
  public:
-  Coordinator(ClientRequestState* parent, const QuerySchedule& schedule,
-      RuntimeProfile::EventSequence* events);
+  Coordinator(ClientRequestState* parent, const TExecRequest& exec_request,
+      const QuerySchedulePB& query_schedule, RuntimeProfile::EventSequence* events);
   ~Coordinator();
 
   /// Initiate asynchronous execution of a query with the given schedule. When it
@@ -139,8 +148,12 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
       int64_t block_on_wait_time_us) WARN_UNUSED_RESULT;
 
   /// Cancel execution of query and sets the overall query status to CANCELLED if the
-  /// query is still executing. Idempotent.
-  void Cancel();
+  /// query is still executing. Idempotent. If 'wait_until_finalized' is true and another
+  /// thread is cancelling 'coord_', block until the query profile is completed, e.g.
+  /// if ComputeQuerySummary() is running concurrently. This is only used by the thread
+  /// unregistering the query - other threads should return immediately if the query
+  /// is in the process of being cancelled.
+  void Cancel(bool wait_until_finalized=false);
 
   /// Called by the report status RPC handler to update execution status of a particular
   /// backend as well as dml_exec_state_ and the profile. This may block if exec RPCs are
@@ -154,7 +167,7 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// for that backend. Only valid to call if Exec() has returned successfully. May return
   /// 0, for example if all of the backends have already completed, in which case
   /// 'address' will not be set.
-  int64_t GetMaxBackendStateLagMs(TNetworkAddress* address);
+  int64_t GetMaxBackendStateLagMs(NetworkAddressPB* address);
 
   /// Get cumulative profile aggregated over all fragments of the query.
   /// This is a snapshot of the current state of execution and will change in
@@ -214,6 +227,9 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
     /// Total system cpu consumed.
     int64_t cpu_sys_ns = 0;
 
+    /// Total num rows produced by each join node. The key is join node id.
+    std::map<int32_t, int64_t> per_join_rows_produced;
+
     /// Merge utilization from 'other' into this.
     void Merge(const ResourceUtilization& other) {
       peak_per_host_mem_consumption =
@@ -223,6 +239,17 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
       scan_bytes_sent += other.scan_bytes_sent;
       cpu_user_ns += other.cpu_user_ns;
       cpu_sys_ns += other.cpu_sys_ns;
+      MergeMapValues(other.per_join_rows_produced, &per_join_rows_produced);
+    }
+
+    /// Max join rows produced across join nodes
+    const std::pair<int32_t, int64_t> MaxJoinNodeRowsProduced() const {
+      std::pair<int32_t, int64_t> entryWithMaxValue = std::make_pair(0, 0);
+      for (const auto& entry : per_join_rows_produced) {
+        if (entry.second > entryWithMaxValue.second)
+          entryWithMaxValue = std::make_pair(entry.first, entry.second);
+      }
+      return entryWithMaxValue;
     }
   };
 
@@ -232,8 +259,8 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
 
   /// Return the backends in 'candidates' that still have at least one fragment instance
   /// executing on them. The returned backends may not be in the same order as the input.
-  std::vector<TNetworkAddress> GetActiveBackends(
-      const std::vector<TNetworkAddress>& candidates);
+  std::vector<NetworkAddressPB> GetActiveBackends(
+      const std::vector<NetworkAddressPB>& candidates);
 
  private:
   class BackendState;
@@ -243,12 +270,17 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   class FilterState;
   class FragmentStats;
 
+  /// The parent QueryDriver object for this coordinator. The reference is set in the
+  /// constructor. It always outlives the coordinator.
+  QueryDriver* parent_query_driver_;
+
   /// The parent ClientRequestState object for this coordinator. The reference is set in
-  /// the constructor. It always outlives the this coordinator.
+  /// the constructor. It always outlives the coordinator.
   ClientRequestState* parent_request_state_;
 
-  /// owned by the ClientRequestState that owns this coordinator
-  const QuerySchedule& schedule_;
+  /// Contains references to Thrift and Protobuf structs constituting the execution
+  /// parameters. These structs are owned by the QueryDriver and ClientRequestState.
+  QueryExecParams exec_params_;
 
   /// Copied from TQueryExecRequest, governs when finalization occurs. Set in Exec().
   TStmtType::type stmt_type_;
@@ -257,12 +289,12 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// are non-nullptr and owned by obj_pool(). Populated by Exec()/InitBackendStates().
   std::vector<BackendState*> backend_states_;
 
-  /// A map from the TNetworkAddress of a backend to the BackendState running on the
-  /// TNetworkAddress. All values are non-nullptr and owned by obj_pool(). The address
+  /// A map from the NetworkAddressPB of a backend to the BackendState running on the
+  /// NetworkAddressPB. All values are non-nullptr and owned by obj_pool(). The address
   /// is the kRPC address (Coordinator::BackendState::krpc_impalad_address) of the
-  /// Backend. This map is distinct from QuerySchedule::per_backend_exec_params(),
+  /// Backend. This map is distinct from QuerySchedulePB::per_backend_exec_params(),
   /// which uses the Thrift address as the key rather than the kRPC address.
-  boost::unordered_map<TNetworkAddress, BackendState*> addr_to_backend_state_;
+  boost::unordered_map<NetworkAddressPB, BackendState*> addr_to_backend_state_;
 
   /// Protects the population of backend_states_ vector (not the BackendState objects).
   /// Used when accessing backend_states_ if it's not guaranteed that
@@ -348,7 +380,7 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
     /// A mapping of fragment data sink ids to index into thrift_exec_summary.nodes
     boost::unordered_map<TPlanNodeId, int> data_sink_id_to_idx_map;
 
-    void Init(const QuerySchedule& query_schedule);
+    void Init(const QueryExecParams& exec_params);
   };
 
   // Initialized by Exec().
@@ -404,6 +436,10 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// - ERROR: error status
   Status exec_status_;
 
+  /// Set when all of the work to transition into a terminal state is complete,
+  /// including ComputeQuerySummary().
+  Promise<bool> finalized_;
+
   /// Contains all the state about filters being handled by this coordinator.
   std::unique_ptr<FilterRoutingTable> filter_routing_table_;
 
@@ -462,7 +498,6 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   void HandleExecStateTransition(const ExecState old_state, const ExecState new_state);
 
   /// Return true if 'exec_state_' is RETURNED_RESULTS.
-  /// TODO: remove with IMPALA-6984.
   bool ReturnedAllResults() WARN_UNUSED_RESULT {
     return exec_state_.Load() == ExecState::RETURNED_RESULTS;
   }
@@ -476,9 +511,10 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   }
 
   /// Helper for HandleExecStateTransition(). Sends cancellation request to all
-  /// executing backends but does not wait for acknowledgement from the backends. The
-  /// ExecState state-machine ensures this is called at most once.
-  void CancelBackends();
+  /// executing backends. If 'fire_and_forget' is true, does not wait for the
+  /// final exec status reports from the backends. The ExecState state-machine
+  /// ensures this is called at most once.
+  void CancelBackends(bool fire_and_forget);
 
   /// Returns only when either all execution backends have reported success or a request
   /// to cancel the backends has already been sent. It is safe to call this concurrently,
@@ -508,6 +544,12 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// finishing the INSERT in flight.
   Status FinalizeHdfsDml() WARN_UNUSED_RESULT;
 
+  /// Perform any post-query cleanup required for Result Sink writing. Called by Wait()
+  /// only after all fragment instances have returned. It is the responsibility of an
+  /// external frontend to clean up the Result Sink location on success or error of the
+  /// running query.
+  Status FinalizeResultSink() WARN_UNUSED_RESULT;
+
   /// Helper for Exec(). Populates backend_states_, starts query execution at all
   /// backends in parallel, and blocks until startup completes.
   Status StartBackendExec();
@@ -528,7 +570,7 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// fragment containing the join node (if build is integrated) or build sink (if the
   /// build is separate). 'num_instances' and 'num_backends' are the number of instances
   /// and backends that the fragment runs on.
-  void AddFilterSource(const FragmentExecParams& src_fragment_params,int num_instances,
+  void AddFilterSource(const FragmentExecParamsPB& src_fragment_params, int num_instances,
       int num_backends, const TRuntimeFilterDesc& filter, int join_node_id);
 
   /// Helper for HandleExecStateTransition(). Releases all resources associated with
@@ -582,7 +624,28 @@ class Coordinator { // NOLINT: The member variables could be re-ordered to save
   /// AuxErrorInfoPB to classify specific nodes as "faulty" and then blacklists them. A
   /// node might be considered "faulty" if, for example, a RPC to that node failed, or a
   /// fragment on that node failed due to a disk IO error.
-  void UpdateBlacklistWithAuxErrorInfo(std::vector<AuxErrorInfoPB>* aux_error_info);
+  /// 'status' is the Status of the given BackendState. 'backend_state' is the
+  /// BackendState that reported an error.
+  /// Returns the Status object used when blacklisting a backend, or Status::OK if no
+  /// backends were blacklisted.
+  Status UpdateBlacklistWithAuxErrorInfo(std::vector<AuxErrorInfoPB>* aux_error_info,
+      const Status& status, BackendState* backend_state) WARN_UNUSED_RESULT;
+
+  /// Helper function for UpdateBackendExecStatus that check if the backend node should
+  /// be blacklisted based on the reported backend error.
+  /// 'status' is the Status of the given BackendState. 'backend_state' is the
+  /// BackendState that reported an error.
+  /// Returns the Status object used when blacklisting a backend, or Status::OK if no
+  /// backends were blacklisted.
+  Status UpdateBlacklistWithBackendState(
+      const Status& status, BackendState* backend_state) WARN_UNUSED_RESULT;
+
+  /// Called if the Exec RPC to the given vector of BackendStates failed. Currently, just
+  /// triggers a retry of the query.
+  void HandleFailedExecRpcs(std::vector<BackendState*> failed_backend_states);
+
+  /// Deletes the query-level staging directory.
+  Status DeleteQueryLevelStagingDir();
 
   /// BackendState and BackendResourceState are private to the Coordinator class, so mark
   /// all tests in CoordinatorBackendStateTest as friends.
