@@ -142,7 +142,7 @@ string SlotDescriptor::DebugString() const {
     out << " collection_item_tuple_id=" << collection_item_descriptor_->id();
   }
   out << " offset=" << tuple_offset_ << " null=" << null_indicator_offset_.DebugString()
-      << " slot_idx=" << slot_idx_ << " field_idx=" << llvm_field_idx_
+      << " slot_idx=" << slot_idx_ << " field_idx=" << slot_idx_
       << ")";
   return out.str();
 }
@@ -159,10 +159,17 @@ bool SlotDescriptor::LayoutEquals(const SlotDescriptor& other_desc) const {
 ColumnDescriptor::ColumnDescriptor(const TColumnDescriptor& tdesc)
   : name_(tdesc.name),
     type_(ColumnType::FromThrift(tdesc.type)) {
+  if (tdesc.__isset.icebergFieldId) {
+    field_id_ = tdesc.icebergFieldId;
+    // Get key and value field_id for Iceberg table column with Map type.
+    field_map_key_id_ = tdesc.icebergFieldMapKeyId;
+    field_map_value_id_ = tdesc.icebergFieldMapValueId;
+  }
 }
 
 string ColumnDescriptor::DebugString() const {
-  return Substitute("$0: $1", name_, type_.DebugString());
+  return Substitute("$0: $1$2", name_, type_.DebugString(),
+      field_id_ != -1 ? " field_id: " + std::to_string(field_id_): "");
 }
 
 TableDescriptor::TableDescriptor(const TTableDescriptor& tdesc)
@@ -195,14 +202,15 @@ string TableDescriptor::DebugString() const {
 
 HdfsPartitionDescriptor::HdfsPartitionDescriptor(
     const THdfsTable& thrift_table, const THdfsPartition& thrift_partition)
-  : line_delim_(thrift_partition.lineDelim),
-    field_delim_(thrift_partition.fieldDelim),
-    collection_delim_(thrift_partition.collectionDelim),
-    escape_char_(thrift_partition.escapeChar),
-    block_size_(thrift_partition.blockSize),
-    id_(thrift_partition.id),
-    thrift_partition_key_exprs_(thrift_partition.partitionKeyExprs),
-    file_format_(thrift_partition.fileFormat) {
+  : id_(thrift_partition.id),
+    thrift_partition_key_exprs_(thrift_partition.partitionKeyExprs) {
+  THdfsStorageDescriptor sd = thrift_partition.hdfs_storage_descriptor;
+  line_delim_ = sd.lineDelim;
+  field_delim_ = sd.fieldDelim;
+  collection_delim_ = sd.collectionDelim;
+  escape_char_ = sd.escapeChar;
+  block_size_ = sd.blockSize;
+  file_format_ = sd.fileFormat;
   DecompressLocation(thrift_table, thrift_partition, &location_);
 }
 
@@ -235,6 +243,17 @@ HdfsTableDescriptor::HdfsTableDescriptor(const TTableDescriptor& tdesc, ObjectPo
   prototype_partition_descriptor_ = pool->Add(new HdfsPartitionDescriptor(
     tdesc.hdfsTable, tdesc.hdfsTable.prototype_partition));
   avro_schema_ = tdesc.hdfsTable.__isset.avroSchema ? tdesc.hdfsTable.avroSchema : "";
+  is_full_acid_ = tdesc.hdfsTable.is_full_acid;
+  valid_write_id_list_ = tdesc.hdfsTable.valid_write_ids;
+  if (tdesc.__isset.icebergTable) {
+    is_iceberg_ = true;
+    iceberg_table_location_ = tdesc.icebergTable.table_location;
+    const TIcebergPartitionSpec& spec = tdesc.icebergTable.partition_spec[
+        tdesc.icebergTable.default_partition_spec_id];
+    for (const TIcebergPartitionField& spec_field : spec.partition_fields) {
+      iceberg_partition_names_.push_back(spec_field.field_name);
+    }
+  }
 }
 
 void HdfsTableDescriptor::ReleaseResources() {
@@ -263,6 +282,7 @@ string HdfsTableDescriptor::DebugString() const {
 
   out << " null_partition_key_value='" << null_partition_key_value_ << "'";
   out << " null_column_value='" << null_column_value_ << "'";
+  out << " is_full_acid=" << std::boolalpha << is_full_acid_;
   return out.str();
 }
 
@@ -551,6 +571,7 @@ Status DescriptorTbl::CreateTblDescriptorInternal(const TTableDescriptor& tdesc,
     ObjectPool* pool, TableDescriptor** desc) {
   *desc = nullptr;
   switch (tdesc.tableType) {
+    case TTableType::ICEBERG_TABLE:
     case TTableType::HDFS_TABLE: {
       HdfsTableDescriptor* hdfs_tbl = pool->Add(new HdfsTableDescriptor(tdesc, pool));
       *desc = hdfs_tbl;
@@ -724,6 +745,10 @@ llvm::Value* SlotDescriptor::CodegenGetNullByte(
 vector<SlotDescriptor*> TupleDescriptor::SlotsOrderedByIdx() const {
   vector<SlotDescriptor*> sorted_slots(slots().size());
   for (SlotDescriptor* slot: slots()) sorted_slots[slot->slot_idx_] = slot;
+  // Check that the size of sorted_slots has not changed. This ensures that the series
+  // of slot indexes starts at 0 and increases by 1 for each slot. This also ensures that
+  // the returned vector has no nullptr elements.
+  DCHECK_EQ(slots().size(), sorted_slots.size());
   return sorted_slots;
 }
 
@@ -736,7 +761,6 @@ llvm::StructType* TupleDescriptor::GetLlvmStruct(LlvmCodeGen* codegen) const {
   int curr_struct_offset = 0;
   for (SlotDescriptor* slot: sorted_slots) {
     DCHECK_EQ(curr_struct_offset, slot->tuple_offset());
-    slot->llvm_field_idx_ = struct_fields.size();
     struct_fields.push_back(codegen->GetSlotType(slot->type()));
     curr_struct_offset = slot->tuple_offset() + slot->slot_size();
   }
