@@ -17,23 +17,23 @@
 
 #pragma once
 
+#include <cstdint>
 #include <iosfwd>
+#include <map>
 #include <unordered_map>
+#include <utility>
 #include <vector>
-#include <boost/scoped_ptr.hpp>
 
 #include "codegen/impala-ir.h"
 #include "common/global-types.h"
 #include "common/status.h"
 #include "runtime/types.h"
 
-#include "gen-cpp/Descriptors_types.h"  // for TTupleId
+#include "gen-cpp/CatalogObjects_types.h"
 #include "gen-cpp/Types_types.h"
 
 namespace llvm {
   class Constant;
-  class Function;
-  class PointerType;
   class StructType;
   class Value;
 };
@@ -43,14 +43,14 @@ namespace impala {
 class LlvmBuilder;
 class LlvmCodeGen;
 class ObjectPool;
-class RuntimeState;
-class ScalarExpr;
 class ScalarExprEvaluator;
+class TColumnDescriptor;
 class TDescriptorTable;
+class TDescriptorTableSerialized;
+class TExpr;
 class TSlotDescriptor;
-class TTable;
-class TTupleDescriptor;
 class TTableDescriptor;
+class TTupleDescriptor;
 
 /// A path into a table schema (e.g. a vector of ColumnTypes) pointing to a particular
 /// column/field. The i-th element of the path is the ordinal position of the column/field
@@ -124,8 +124,12 @@ class SlotDescriptor {
   /// convenient for table formats for which we only support flat data.
   int col_pos() const { return col_path_[0]; }
   const SchemaPath& col_path() const { return col_path_; }
-  /// Returns the field index in the generated llvm struct for this slot's tuple
-  int llvm_field_idx() const { return llvm_field_idx_; }
+
+  /// Returns the field index in the generated llvm struct for this slot's tuple.
+  /// The 'llvm_field_idx' is the index of the slot in the llvm codegen'd tuple struct.
+  /// This takes into account any padding bytes.
+  int llvm_field_idx() const { return slot_idx_; }
+
   int tuple_offset() const { return tuple_offset_; }
   const NullIndicatorOffset& null_indicator_offset() const {
     return null_indicator_offset_;
@@ -185,11 +189,6 @@ class SlotDescriptor {
   /// the byte size of this slot.
   const int slot_size_;
 
-  /// The idx of the slot in the llvm codegen'd tuple struct
-  /// This is set by TupleDescriptor during codegen and takes into account
-  /// any padding bytes.
-  int llvm_field_idx_ = -1;
-
   /// collection_item_descriptor should be non-NULL iff this is a collection slot
   SlotDescriptor(const TSlotDescriptor& tdesc, const TupleDescriptor* parent,
       const TupleDescriptor* collection_item_descriptor);
@@ -207,11 +206,17 @@ class ColumnDescriptor {
   ColumnDescriptor(const TColumnDescriptor& tdesc);
   const std::string& name() const { return name_; }
   const ColumnType& type() const { return type_; }
+  int field_id() const { return field_id_; }
+  int field_map_key_id() const { return field_map_key_id_; }
+  int field_map_value_id() const { return field_map_value_id_; }
   std::string DebugString() const;
 
  private:
   std::string name_;
   ColumnType type_;
+  int field_id_ = -1;
+  int field_map_key_id_ = -1;
+  int field_map_value_id_ = -1;
 };
 
 /// Base class for table descriptors.
@@ -330,6 +335,15 @@ class HdfsTableDescriptor : public TableDescriptor {
     return prototype_partition_descriptor_;
   }
 
+  bool IsTableFullAcid() const { return is_full_acid_; }
+  const TValidWriteIdList& ValidWriteIdList() const { return valid_write_id_list_; }
+
+  bool IsIcebergTable() const { return is_iceberg_; }
+  const std::string& IcebergTableLocation() const { return iceberg_table_location_; }
+  const std::vector<std::string>& IcebergPartitionNames() const {
+    return iceberg_partition_names_;
+  }
+
   virtual std::string DebugString() const;
 
  protected:
@@ -341,6 +355,11 @@ class HdfsTableDescriptor : public TableDescriptor {
   HdfsPartitionDescriptor* prototype_partition_descriptor_;
   /// Set to the table's Avro schema if this is an Avro table, empty string otherwise
   std::string avro_schema_;
+  bool is_full_acid_;
+  TValidWriteIdList valid_write_id_list_;
+  bool is_iceberg_ = false;
+  std::string iceberg_table_location_;
+  std::vector<std::string> iceberg_partition_names_;
 };
 
 class HBaseTableDescriptor : public TableDescriptor {
