@@ -36,12 +36,15 @@
 DEFINE_bool(use_hdfs_pread, false, "Enables using hdfsPread() instead of hdfsRead() "
     "when performing HDFS read operations. This is necessary to use HDFS hedged reads "
     "(assuming the HDFS client is configured to do so). Preads are always enabled for "
-    "S3A reads.");
+    "S3A and ABFS reads.");
 
 DEFINE_int64(fs_slow_read_log_threshold_ms, 10L * 1000L,
     "Log diagnostics about I/Os issued via the HDFS client that take longer than this "
     "threshold.");
 
+DEFINE_bool(fs_trace_remote_reads, false,
+    "(Advanced) Log block locations for remote reads.");
+
 #ifndef NDEBUG
 DECLARE_int32(stress_disk_read_delay_ms);
 #endif
@@ -78,6 +81,30 @@ Status HdfsFileReader::Open(bool use_file_handle_cache) {
   return Status::OK();
 }
 
+std::string HdfsFileReader::GetHostList(int64_t file_offset,
+    int64_t bytes_to_read) const {
+  char*** hosts = hdfsGetHosts(hdfs_fs_, scan_range_->file_string()->c_str(),
+      file_offset, bytes_to_read);
+  if (hosts) {
+    std::ostringstream ostr;
+    int blockIndex = 0;
+    while (hosts[blockIndex]) {
+      ostr << " [" << blockIndex << "] { ";
+      int hostIndex = 0;
+      while (hosts[blockIndex][hostIndex]) {
+        if(hostIndex > 0) ostr << ", ";
+        ostr << hosts[blockIndex][hostIndex];
+        hostIndex++;
+      }
+      ostr << " }";
+      blockIndex++;
+    }
+    hdfsFreeHosts(hosts);
+    return ostr.str();
+  }
+  return "";
+}
+
 Status HdfsFileReader::ReadFromPos(DiskQueue* queue, int64_t file_offset, uint8_t* buffer,
     int64_t bytes_to_read, int64_t* bytes_read, bool* eof) {
   DCHECK(scan_range_->read_in_flight());
@@ -96,27 +123,6 @@ Status HdfsFileReader::ReadFromPos(DiskQueue* queue, int64_t file_offset, uint8_
   *eof = false;
   *bytes_read = 0;
 
-  CachedHdfsFileHandle* borrowed_hdfs_fh = nullptr;
-  hdfsFile hdfs_file;
-
-  // If the reader has an exclusive file handle, use it. Otherwise, borrow
-  // a file handle from the cache.
-  if (exclusive_hdfs_fh_ != nullptr) {
-    hdfs_file = exclusive_hdfs_fh_->file();
-  } else {
-    RETURN_IF_ERROR(io_mgr->GetCachedHdfsFileHandle(hdfs_fs_,
-        scan_range_->file_string(),
-        scan_range_->mtime(), request_context, &borrowed_hdfs_fh));
-    hdfs_file = borrowed_hdfs_fh->file();
-  }
-  // Make sure to release any borrowed file handle.
-  auto release_borrowed_hdfs_fh = MakeScopeExitTrigger([this, &borrowed_hdfs_fh]() {
-    if (borrowed_hdfs_fh != nullptr) {
-      scan_range_->io_mgr_->ReleaseCachedHdfsFileHandle(scan_range_->file_string(),
-          borrowed_hdfs_fh);
-    }
-  });
-
   Status status = Status::OK();
   {
     ScopedTimer<MonotonicStopWatch> req_context_read_timer(
@@ -133,6 +139,35 @@ Status HdfsFileReader::ReadFromPos(DiskQueue* queue, int64_t file_offset, uint8_
       *bytes_read = cached_read;
     }
 
+    if (*bytes_read == bytes_to_read) {
+      // All bytes successfully read from data cache. We can safely return.
+      return status;
+    }
+
+    // If we get here, the next bytes are not available in data cache, so we need to get
+    // file handle in order to read the rest of data from file.
+    // If the reader has an exclusive file handle, use it. Otherwise, borrow
+    // a file handle from the cache.
+    req_context_read_timer.Stop();
+    CachedHdfsFileHandle* borrowed_hdfs_fh = nullptr;
+    hdfsFile hdfs_file;
+    if (exclusive_hdfs_fh_ != nullptr) {
+      hdfs_file = exclusive_hdfs_fh_->file();
+    } else {
+      RETURN_IF_ERROR(
+          io_mgr->GetCachedHdfsFileHandle(hdfs_fs_, scan_range_->file_string(),
+              scan_range_->mtime(), request_context, &borrowed_hdfs_fh));
+      hdfs_file = borrowed_hdfs_fh->file();
+    }
+    // Make sure to release any borrowed file handle.
+    auto release_borrowed_hdfs_fh = MakeScopeExitTrigger([this, &borrowed_hdfs_fh]() {
+      if (borrowed_hdfs_fh != nullptr) {
+        scan_range_->io_mgr_->ReleaseCachedHdfsFileHandle(
+            scan_range_->file_string(), borrowed_hdfs_fh);
+      }
+    });
+    req_context_read_timer.Start();
+
     while (*bytes_read < bytes_to_read) {
       int bytes_remaining = bytes_to_read - *bytes_read;
       DCHECK_GT(bytes_remaining, 0);
@@ -201,8 +236,20 @@ Status HdfsFileReader::ReadFromPos(DiskQueue* queue, int64_t file_offset, uint8_
       }
       *bytes_read += current_bytes_read;
 
+      bool is_first_read = (num_remote_bytes_ == 0);
       // Collect and accumulate statistics
       GetHdfsStatistics(hdfs_file, log_slow_read);
+      if (FLAGS_fs_trace_remote_reads && expected_local_ &&
+          num_remote_bytes_ > 0 && is_first_read) {
+        // Only log the first unexpected remote read for scan range
+        LOG(INFO)
+            << "First remote read of scan range on file "
+            << *scan_range_->file_string()
+            << " " << num_remote_bytes_
+            << " bytes. offsets " << file_offset
+            << "-" << file_offset+bytes_to_read-1
+            << GetHostList(file_offset, bytes_to_read);
+      }
     }
 
     int64_t cached_bytes_missed = *bytes_read - cached_read;
@@ -221,7 +268,8 @@ Status HdfsFileReader::ReadFromPosInternal(hdfsFile hdfs_file, DiskQueue* queue,
   ScopedHistogramTimer read_timer(queue->read_latency());
   // For file handles from the cache, any of the below file operations may fail
   // due to a bad file handle.
-  if (FLAGS_use_hdfs_pread || IsS3APath(scan_range_->file_string()->c_str())) {
+  if (FLAGS_use_hdfs_pread || IsS3APath(scan_range_->file_string()->c_str())
+      || IsABFSPath(scan_range_->file_string()->c_str())) {
     if (hdfsPreadFully(
           hdfs_fs_, hdfs_file, position_in_file, buffer, bytes_to_read) == -1) {
       return Status(TErrorCode::DISK_IO_ERROR, GetBackendString(),
@@ -286,6 +334,7 @@ int64_t HdfsFileReader::ReadDataCache(DataCache* remote_data_cache, int64_t file
       scan_range_->reader_->data_cache_partial_hit_counter_->Add(1);
     }
     ImpaladMetrics::IO_MGR_REMOTE_DATA_CACHE_HIT_BYTES->Increment(cached_read);
+    ImpaladMetrics::IO_MGR_REMOTE_DATA_CACHE_HIT_COUNT->Increment(1);
   }
   return cached_read;
 }
@@ -301,6 +350,7 @@ void HdfsFileReader::WriteDataCache(DataCache* remote_data_cache, int64_t file_o
   scan_range_->reader_->data_cache_miss_bytes_counter_->Add(bytes_missed);
   scan_range_->reader_->data_cache_miss_counter_->Add(1);
   ImpaladMetrics::IO_MGR_REMOTE_DATA_CACHE_MISS_BYTES->Increment(bytes_missed);
+  ImpaladMetrics::IO_MGR_REMOTE_DATA_CACHE_MISS_COUNT->Increment(1);
 }
 
 void HdfsFileReader::Close() {
