@@ -16,8 +16,8 @@
 // under the License.
 
 #include "runtime/exec-env.h"
-#include "runtime/io/disk-io-mgr.h"
 #include "runtime/io/disk-io-mgr-internal.h"
+#include "runtime/io/disk-io-mgr.h"
 #include "runtime/io/hdfs-file-reader.h"
 #include "runtime/io/local-file-reader.h"
 #include "util/error-util.h"
@@ -27,9 +27,13 @@
 
 using namespace impala;
 using namespace impala::io;
+using namespace std;
+using std::mutex;
+using std::unique_lock;
 
 DECLARE_bool(cache_remote_file_handles);
 DECLARE_bool(cache_s3_file_handles);
+DECLARE_bool(cache_abfs_file_handles);
 
 // Implementation of the ScanRange functionality. Each ScanRange contains a queue
 // of ready buffers. For each ScanRange, there is only a single producer and
@@ -159,11 +163,13 @@ unique_ptr<BufferDescriptor> ScanRange::GetUnusedBuffer(
   return result;
 }
 
-ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
+ReadOutcome ScanRange::DoReadInternal(
+    DiskQueue* queue, int disk_id, bool use_local_buff) {
   int64_t bytes_remaining = bytes_to_read_ - bytes_read_;
   DCHECK_GT(bytes_remaining, 0);
 
   unique_ptr<BufferDescriptor> buffer_desc;
+  FileReader* file_reader = nullptr;
   {
     unique_lock<mutex> lock(lock_);
     DCHECK(!read_in_flight_);
@@ -185,25 +191,31 @@ ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
       iomgr_buffer_cumulative_bytes_used_ += buffer_desc->buffer_len();
     }
     read_in_flight_ = true;
+    if (use_local_buff) {
+      file_reader = local_buffer_reader_.get();
+      file_ = disk_buffer_file_->path();
+    } else {
+      file_reader = file_reader_.get();
+    }
+    use_local_buffer_ = use_local_buff;
   }
+  DCHECK(file_reader != nullptr);
 
   // No locks in this section.  Only working on local vars.  We don't want to hold a
   // lock across the read call.
   // To use the file handle cache:
   // 1. It must be enabled at the daemon level.
-  // 2. The file cannot be erasure coded.
-  // 3. The file is a local HDFS file (expected_local_) OR it is a remote HDFS file and
+  // 2. The file is a local HDFS file (expected_local_) OR it is a remote HDFS file and
   //    'cache_remote_file_handles' is true
-  // Note: S3, ADLS, and ABFS file handles are not cached. Erasure coded HDFS files
-  // are also not cached (IMPALA-8178), due to excessive memory usage (see HDFS-14308).
   bool use_file_handle_cache = false;
-  if (is_file_handle_caching_enabled() && !is_erasure_coded_ &&
+  if (is_file_handle_caching_enabled() &&
       (expected_local_ ||
        (FLAGS_cache_remote_file_handles && disk_id_ == io_mgr_->RemoteDfsDiskId()) ||
-       (FLAGS_cache_s3_file_handles && disk_id_ == io_mgr_->RemoteS3DiskId()))) {
+       (FLAGS_cache_s3_file_handles && disk_id_ == io_mgr_->RemoteS3DiskId()) ||
+       (FLAGS_cache_abfs_file_handles && disk_id_ == io_mgr_->RemoteAbfsDiskId()))) {
     use_file_handle_cache = true;
   }
-  Status read_status = file_reader_->Open(use_file_handle_cache);
+  Status read_status = file_reader->Open(use_file_handle_cache);
   bool eof = false;
   if (read_status.ok()) {
     COUNTER_ADD_IF_NOT_NULL(reader_->active_read_thread_counter_, 1L);
@@ -211,12 +223,12 @@ ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
 
     if (sub_ranges_.empty()) {
       DCHECK(cache_.data == nullptr);
-      read_status = file_reader_->ReadFromPos(queue, offset_ + bytes_read_,
-          buffer_desc->buffer_,
-          min(bytes_to_read() - bytes_read_, buffer_desc->buffer_len_),
-          &buffer_desc->len_, &eof);
+      read_status =
+          file_reader->ReadFromPos(queue, offset_ + bytes_read_, buffer_desc->buffer_,
+              min(bytes_to_read() - bytes_read_, buffer_desc->buffer_len_),
+              &buffer_desc->len_, &eof);
     } else {
-      read_status = ReadSubRanges(queue, buffer_desc.get(), &eof);
+      read_status = ReadSubRanges(queue, buffer_desc.get(), &eof, file_reader);
     }
 
     COUNTER_ADD_IF_NOT_NULL(reader_->bytes_read_counter_, buffer_desc->len_);
@@ -224,8 +236,8 @@ ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
   }
 
   DCHECK(buffer_desc->buffer_ != nullptr);
-  DCHECK(!buffer_desc->is_cached()) <<
-      "Pure HDFS cache reads don't go through this code path.";
+  DCHECK(!buffer_desc->is_cached())
+      << "Pure HDFS cache reads don't go through this code path.";
   if (!read_status.ok()) {
     // Free buffer to release resources before we cancel the range so that all buffers
     // are freed at cancellation.
@@ -253,7 +265,7 @@ ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
   // Store the state we need before calling EnqueueReadyBuffer().
   bool eosr = buffer_desc->eosr();
   // No more reads for this scan range - we can close it.
-  if (eosr) file_reader_->Close();
+  if (eosr) file_reader->Close();
   // Read successful - enqueue the buffer and return the appropriate outcome.
   if (!EnqueueReadyBuffer(move(buffer_desc))) return ReadOutcome::CANCELLED;
   // At this point, if eosr=true, then we cannot touch the state of this scan range
@@ -261,8 +273,38 @@ ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
   return eosr ? ReadOutcome::SUCCESS_EOSR : ReadOutcome::SUCCESS_NO_EOSR;
 }
 
+ReadOutcome ScanRange::DoRead(DiskQueue* queue, int disk_id) {
+  bool use_local_buffer = false;
+  if (disk_file_ != nullptr && disk_file_->disk_type() != DiskFileType::LOCAL) {
+    // The sequence for acquiring the locks should always be from the local to
+    // the remote to avoid deadlocks.
+    shared_lock<shared_mutex> local_file_lock(*(disk_buffer_file_->GetFileLock()));
+    shared_lock<shared_mutex> remote_file_lock(*(disk_file_->GetFileLock()));
+    {
+      unique_lock<SpinLock> buffer_file_lock(*(disk_buffer_file_->GetStatusLock()));
+      unique_lock<SpinLock> file_lock(*(disk_file_->GetStatusLock()));
+      if (disk_buffer_file_->is_deleted(buffer_file_lock)
+          && disk_file_->is_deleted(file_lock)) {
+        // If both of the local buffer file and the remote file have been deleted,
+        // the only case could be the query is cancelled, so that both files are deleted.
+        return ReadOutcome::CANCELLED;
+      }
+      // If the local buffer exists, we can read from the local buffer, otherwise,
+      // we will read from the remote file system.
+      if (!disk_buffer_file_->is_deleted(buffer_file_lock)) {
+        use_local_buffer = true;
+      } else {
+        // Read from the remote file. The remote file must be in persisted status.
+        DCHECK(disk_file_->is_persisted(file_lock));
+      }
+    }
+    return DoReadInternal(queue, disk_id, use_local_buffer);
+  }
+  return DoReadInternal(queue, disk_id, use_local_buffer);
+}
+
 Status ScanRange::ReadSubRanges(
-    DiskQueue* queue, BufferDescriptor* buffer_desc, bool* eof) {
+    DiskQueue* queue, BufferDescriptor* buffer_desc, bool* eof, FileReader* file_reader) {
   buffer_desc->len_ = 0;
   while (buffer_desc->len() < buffer_desc->buffer_len()
       && sub_range_pos_.index < sub_ranges_.size()) {
@@ -276,7 +318,7 @@ Status ScanRange::ReadSubRanges(
           cache_.data + offset, bytes_to_read);
     } else {
       int64_t current_bytes_read;
-      Status read_status = file_reader_->ReadFromPos(queue, offset,
+      Status read_status = file_reader->ReadFromPos(queue, offset,
           buffer_desc->buffer_ + buffer_desc->len(), bytes_to_read, &current_bytes_read,
           eof);
       if (!read_status.ok()) return read_status;
@@ -314,7 +356,11 @@ void ScanRange::CleanUpBuffer(const unique_lock<mutex>& scan_range_lock,
     // Close the scan range if there are no more buffers in the reader and no more buffers
     // will be returned to readers in future. Close() is idempotent so it is ok to call
     // multiple times during cleanup so long as the range is actually finished.
-    file_reader_->Close();
+    if (!use_local_buffer_) {
+      file_reader_->Close();
+    } else {
+      local_buffer_reader_->Close();
+    }
   }
 }
 
@@ -342,12 +388,14 @@ void ScanRange::Cancel(const Status& status) {
 void ScanRange::CancelInternal(const Status& status, bool read_error) {
   DCHECK(io_mgr_ != nullptr);
   DCHECK(!status.ok());
+  FileReader* file_reader = nullptr;
   {
     // Grab both locks to make sure that we don't change 'cancel_status_' while other
     // threads are in critical sections.
     unique_lock<mutex> scan_range_lock(lock_);
     {
-      unique_lock<SpinLock> fs_lock(file_reader_->lock());
+      file_reader = use_local_buffer_ ? local_buffer_reader_.get() : file_reader_.get();
+      unique_lock<SpinLock> fs_lock(file_reader->lock());
       DCHECK(Validate()) << DebugString();
       // If already cancelled, preserve the original reason for cancellation. Most of the
       // cleanup is not required if already cancelled, but we need to set
@@ -376,7 +424,10 @@ void ScanRange::CancelInternal(const Status& status, bool read_error) {
   // TODO: IMPALA-4249 - this Close() call makes it unsafe to reuse a cancelled scan
   // range, because there is no synchronisation between this Close() call and the
   // client adding the ScanRange back into the IoMgr.
-  if (external_buffer_tag_ != ExternalBufferTag::CACHED_BUFFER) file_reader_->Close();
+  if (external_buffer_tag_ != ExternalBufferTag::CACHED_BUFFER) {
+    DCHECK(file_reader != nullptr);
+    file_reader->Close();
+  }
 }
 
 void ScanRange::WaitForInFlightRead() {
@@ -443,15 +494,30 @@ ScanRange::~ScanRange() {
 }
 
 void ScanRange::Reset(hdfsFS fs, const char* file, int64_t len, int64_t offset,
-    int disk_id, bool expected_local, bool is_erasure_coded, int64_t mtime,
-    const BufferOpts& buffer_opts, void* meta_data) {
-  Reset(fs, file, len, offset, disk_id, expected_local, is_erasure_coded, mtime,
-      buffer_opts, {}, meta_data);
+    int disk_id, bool expected_local, int64_t mtime, const BufferOpts& buffer_opts,
+    void* meta_data, DiskFile* disk_file, DiskFile* disk_buffer_file) {
+  Reset(fs, file, len, offset, disk_id, expected_local, mtime, buffer_opts, {}, meta_data,
+      disk_file, disk_buffer_file);
+}
+
+ScanRange* ScanRange::AllocateScanRange(ObjectPool* obj_pool, hdfsFS fs, const char* file,
+    int64_t len, int64_t offset, std::vector<SubRange>&& sub_ranges, void* metadata,
+    int disk_id, bool expected_local, int64_t mtime, const BufferOpts& buffer_opts) {
+  DCHECK_GE(disk_id, -1);
+  DCHECK_GE(offset, 0);
+  DCHECK_GE(len, 0);
+  disk_id = ExecEnv::GetInstance()->disk_io_mgr()->AssignQueue(
+      file, disk_id, expected_local, /* check_default_fs */ true);
+  ScanRange* range = obj_pool->Add(new ScanRange);
+  range->Reset(fs, file, len, offset, disk_id, expected_local, mtime, buffer_opts,
+      move(sub_ranges), metadata);
+  return range;
 }
 
 void ScanRange::Reset(hdfsFS fs, const char* file, int64_t len, int64_t offset,
-    int disk_id, bool expected_local, bool is_erasure_coded, int64_t mtime,
-    const BufferOpts& buffer_opts, vector<SubRange>&& sub_ranges, void* meta_data) {
+    int disk_id, bool expected_local, int64_t mtime, const BufferOpts& buffer_opts,
+    vector<SubRange>&& sub_ranges, void* meta_data, DiskFile* disk_file,
+    DiskFile* disk_buffer_file) {
   DCHECK(ready_buffers_.empty());
   DCHECK(!read_in_flight_);
   DCHECK(file != nullptr);
@@ -460,8 +526,9 @@ void ScanRange::Reset(hdfsFS fs, const char* file, int64_t len, int64_t offset,
   DCHECK(buffer_opts.client_buffer_ == nullptr ||
          buffer_opts.client_buffer_len_ >= len_);
   fs_ = fs;
-  if (fs_) {
-    file_reader_ = make_unique<HdfsFileReader>(this, fs_, expected_local);
+  if (fs != nullptr) {
+    file_reader_ = make_unique<HdfsFileReader>(this, fs_, false);
+    local_buffer_reader_ = make_unique<LocalFileReader>(this);
   } else {
     file_reader_ = make_unique<LocalFileReader>(this);
   }
@@ -471,6 +538,8 @@ void ScanRange::Reset(hdfsFS fs, const char* file, int64_t len, int64_t offset,
   offset_ = offset;
   disk_id_ = disk_id;
   cache_options_ = buffer_opts.cache_options_;
+  disk_file_ = disk_file;
+  disk_buffer_file_ = disk_buffer_file;
 
   // HDFS ranges must have an mtime > 0. Local ranges do not use mtime.
   if (fs_) DCHECK_GT(mtime, 0);
@@ -483,10 +552,7 @@ void ScanRange::Reset(hdfsFS fs, const char* file, int64_t len, int64_t offset,
   } else {
     external_buffer_tag_ = ExternalBufferTag::NO_BUFFER;
   }
-  // Erasure coded should not be considered local (see IMPALA-7019).
-  DCHECK(!(expected_local && is_erasure_coded));
   expected_local_ = expected_local;
-  is_erasure_coded_ = is_erasure_coded;
   io_mgr_ = nullptr;
   reader_ = nullptr;
   sub_ranges_.clear();
@@ -556,6 +622,7 @@ void ScanRange::InitInternal(DiskIoMgr* io_mgr, RequestContext* reader) {
   bytes_read_ = 0;
   sub_range_pos_ = {};
   file_reader_->ResetState();
+  if (local_buffer_reader_ != nullptr) local_buffer_reader_->ResetState();
   DCHECK(Validate()) << DebugString();
 }
 
