@@ -38,6 +38,7 @@
 #include "rpc/rpc-mgr.h"
 #include "runtime/descriptors.h"
 #include "runtime/exec-env.h"
+#include "runtime/fragment-state.h"
 #include "runtime/mem-tracker.h"
 #include "runtime/raw-value.inline.h"
 #include "runtime/row-batch.h"
@@ -76,7 +77,7 @@ const char* KrpcDataStreamSender::LLVM_CLASS_NAME = "class.impala::KrpcDataStrea
 const char* KrpcDataStreamSender::TOTAL_BYTES_SENT_COUNTER = "TotalBytesSent";
 
 Status KrpcDataStreamSenderConfig::Init(
-    const TDataSink& tsink, const RowDescriptor* input_row_desc, RuntimeState* state) {
+    const TDataSink& tsink, const RowDescriptor* input_row_desc, FragmentState* state) {
   RETURN_IF_ERROR(DataSinkConfig::Init(tsink, input_row_desc, state));
   DCHECK(tsink_->__isset.stream_sink);
   partition_type_ = tsink_->stream_sink.output_partition.type;
@@ -85,17 +86,21 @@ Status KrpcDataStreamSenderConfig::Init(
     RETURN_IF_ERROR(
         ScalarExpr::Create(tsink_->stream_sink.output_partition.partition_exprs,
             *input_row_desc_, state, &partition_exprs_));
+    exchange_hash_seed_ =
+        KrpcDataStreamSender::EXCHANGE_HASH_SEED_CONST ^ state->query_id().hi;
   }
+  num_channels_ = state->fragment_ctx().destinations().size();
+  state->CheckAndAddCodegenDisabledMessage(codegen_status_msgs_);
   return Status::OK();
 }
 
-DataSink* KrpcDataStreamSenderConfig::CreateSink(const TPlanFragmentCtx& fragment_ctx,
-    const TPlanFragmentInstanceCtx& fragment_instance_ctx, RuntimeState* state) const {
+DataSink* KrpcDataStreamSenderConfig::CreateSink(RuntimeState* state) const {
   // We have one fragment per sink, so we can use the fragment index as the sink ID.
-  TDataSinkId sink_id = fragment_ctx.fragment.idx;
-  return state->obj_pool()->Add(new KrpcDataStreamSender(sink_id,
-      fragment_instance_ctx.sender_id, *this, tsink_->stream_sink,
-      fragment_ctx.destinations, FLAGS_data_stream_sender_buffer_size, state));
+  TDataSinkId sink_id = state->fragment().idx;
+  return state->obj_pool()->Add(
+      new KrpcDataStreamSender(sink_id, state->instance_ctx().sender_id, *this,
+          tsink_->stream_sink, state->fragment_ctx().destinations(),
+          FLAGS_data_stream_sender_buffer_size, state));
 }
 
 void KrpcDataStreamSenderConfig::Close() {
@@ -143,8 +148,8 @@ class KrpcDataStreamSender::Channel : public CacheLineAligned {
   // data is getting accumulated before being sent; it only applies when data is added via
   // AddRow() and not sent directly via SendBatch().
   Channel(KrpcDataStreamSender* parent, const RowDescriptor* row_desc,
-      const std::string& hostname, const TNetworkAddress& destination,
-      const TUniqueId& fragment_instance_id, PlanNodeId dest_node_id, int buffer_size)
+      const std::string& hostname, const NetworkAddressPB& destination,
+      const UniqueIdPB& fragment_instance_id, PlanNodeId dest_node_id, int buffer_size)
     : parent_(parent),
       row_desc_(row_desc),
       hostname_(hostname),
@@ -211,8 +216,8 @@ class KrpcDataStreamSender::Channel : public CacheLineAligned {
 
   // The triplet of IP-address:port/finst-id/node-id uniquely identifies the receiver.
   const std::string hostname_;
-  const TNetworkAddress address_;
-  const TUniqueId fragment_instance_id_;
+  const NetworkAddressPB address_;
+  const UniqueIdPB fragment_instance_id_;
   const PlanNodeId dest_node_id_;
 
   // The row batch for accumulating rows copied from AddRow().
@@ -368,7 +373,8 @@ Status KrpcDataStreamSender::Channel::Init(RuntimeState* state) {
   batch_.reset(new RowBatch(row_desc_, capacity, parent_->mem_tracker()));
 
   // Create a DataStreamService proxy to the destination.
-  RETURN_IF_ERROR(DataStreamService::GetProxy(address_, hostname_, &proxy_));
+  RETURN_IF_ERROR(
+      DataStreamService::GetProxy(FromNetworkAddressPB(address_), hostname_, &proxy_));
   return Status::OK();
 }
 
@@ -384,18 +390,18 @@ void KrpcDataStreamSender::Channel::MarkDone(const Status& status) {
 template <typename ResponsePBType>
 void KrpcDataStreamSender::Channel::LogSlowRpc(
     const char* rpc_name, int64_t total_time_ns, const ResponsePBType& resp) {
-  int64_t network_time_ns = total_time_ns - resp_.receiver_latency_ns();
-  LOG(INFO) << "Slow " << rpc_name << " RPC to " << TNetworkAddressToString(address_)
+  int64_t network_time_ns = total_time_ns - resp.receiver_latency_ns();
+  LOG(INFO) << "Slow " << rpc_name << " RPC to " << address_
             << " (fragment_instance_id=" << PrintId(fragment_instance_id_) << "): "
             << "took " << PrettyPrinter::Print(total_time_ns, TUnit::TIME_NS) << ". "
             << "Receiver time: "
-            << PrettyPrinter::Print(resp_.receiver_latency_ns(), TUnit::TIME_NS)
+            << PrettyPrinter::Print(resp.receiver_latency_ns(), TUnit::TIME_NS)
             << " Network time: " << PrettyPrinter::Print(network_time_ns, TUnit::TIME_NS);
 }
 
 void KrpcDataStreamSender::Channel::LogSlowFailedRpc(
     const char* rpc_name, int64_t total_time_ns, const kudu::Status& err) {
-  LOG(INFO) << "Slow " << rpc_name << " RPC to " << TNetworkAddressToString(address_)
+  LOG(INFO) << "Slow " << rpc_name << " RPC to " << address_
             << " (fragment_instance_id=" << PrintId(fragment_instance_id_) << "): "
             << "took " << PrettyPrinter::Print(total_time_ns, TUnit::TIME_NS) << ". "
             << "Error: " << err.ToString();
@@ -419,7 +425,7 @@ Status KrpcDataStreamSender::Channel::WaitForRpcLocked(std::unique_lock<SpinLock
   }
   int64_t elapsed_time_ns = timer.ElapsedTime();
   if (IsSlowRpc(elapsed_time_ns)) {
-    LOG(INFO) << "Long delay waiting for RPC to " << TNetworkAddressToString(address_)
+    LOG(INFO) << "Long delay waiting for RPC to " << address_
               << " (fragment_instance_id=" << PrintId(fragment_instance_id_) << "): "
               << "took " << PrettyPrinter::Print(elapsed_time_ns, TUnit::TIME_NS);
   }
@@ -432,9 +438,9 @@ Status KrpcDataStreamSender::Channel::WaitForRpcLocked(std::unique_lock<SpinLock
 
   DCHECK(!rpc_in_flight_);
   if (UNLIKELY(!rpc_status_.ok())) {
-    LOG(ERROR) << "channel send to " << TNetworkAddressToString(address_) << " failed: "
-               << "(fragment_instance_id=" << PrintId(fragment_instance_id_) << "): "
-               << rpc_status_.GetDetail();
+    LOG(ERROR) << "channel send to " << address_ << " failed: (fragment_instance_id="
+               << PrintId(fragment_instance_id_)
+               << "): " << rpc_status_.GetDetail();
     return rpc_status_;
   }
   return Status::OK();
@@ -483,13 +489,16 @@ void KrpcDataStreamSender::Channel::HandleFailedRPC(const DoRpcFn& rpc_fn,
 }
 
 void KrpcDataStreamSender::Channel::TransmitDataCompleteCb() {
-  DCHECK_NE(rpc_start_time_ns_, 0);
-  int64_t total_time = MonotonicNanos() - rpc_start_time_ns_;
   std::unique_lock<SpinLock> l(lock_);
   DCHECK(rpc_in_flight_);
+  DCHECK_NE(rpc_start_time_ns_, 0);
+  int64_t total_time = MonotonicNanos() - rpc_start_time_ns_;
   const kudu::Status controller_status = rpc_controller_.status();
   if (LIKELY(controller_status.ok())) {
     DCHECK(rpc_in_flight_batch_ != nullptr);
+    // 'receiver_latency_ns' is calculated with MonoTime, so it must be non-negative.
+    DCHECK_GE(resp_.receiver_latency_ns(), 0);
+    DCHECK_GE(total_time, resp_.receiver_latency_ns());
     int64_t row_batch_size = RowBatch::GetSerializedSize(*rpc_in_flight_batch_);
     int64_t network_time = total_time - resp_.receiver_latency_ns();
     COUNTER_ADD(parent_->bytes_sent_counter_, row_batch_size);
@@ -518,7 +527,7 @@ void KrpcDataStreamSender::Channel::TransmitDataCompleteCb() {
     DoRpcFn rpc_fn =
         boost::bind(&KrpcDataStreamSender::Channel::DoTransmitDataRpc, this);
     const string& prepend =
-        Substitute("TransmitData() to $0 failed", TNetworkAddressToString(address_));
+        Substitute("TransmitData() to $0 failed", NetworkAddressPBToString(address_));
     HandleFailedRPC(rpc_fn, controller_status, prepend);
   }
 }
@@ -529,9 +538,7 @@ Status KrpcDataStreamSender::Channel::DoTransmitDataRpc() {
 
   // Initialize some constant fields in the request protobuf.
   TransmitDataRequestPB req;
-  UniqueIdPB* finstance_id_pb = req.mutable_dest_fragment_instance_id();
-  finstance_id_pb->set_lo(fragment_instance_id_.lo);
-  finstance_id_pb->set_hi(fragment_instance_id_.hi);
+  *req.mutable_dest_fragment_instance_id() = fragment_instance_id_;
   req.set_sender_id(parent_->sender_id_);
   req.set_dest_node_id(dest_node_id_);
 
@@ -624,7 +631,10 @@ void KrpcDataStreamSender::Channel::EndDataStreamCompleteCb() {
   int64_t total_time_ns = MonotonicNanos() - rpc_start_time_ns_;
   const kudu::Status controller_status = rpc_controller_.status();
   if (LIKELY(controller_status.ok())) {
-    int64_t network_time_ns = total_time_ns - resp_.receiver_latency_ns();
+    // 'receiver_latency_ns' is calculated with MonoTime, so it must be non-negative.
+    DCHECK_GE(eos_resp_.receiver_latency_ns(), 0);
+    DCHECK_GE(total_time_ns, eos_resp_.receiver_latency_ns());
+    int64_t network_time_ns = total_time_ns - eos_resp_.receiver_latency_ns();
     parent_->network_time_stats_->UpdateCounter(network_time_ns);
     parent_->recvr_time_stats_->UpdateCounter(eos_resp_.receiver_latency_ns());
     if (IsSlowRpc(total_time_ns)) LogSlowRpc("EndDataStream", total_time_ns, eos_resp_);
@@ -636,7 +646,7 @@ void KrpcDataStreamSender::Channel::EndDataStreamCompleteCb() {
     DoRpcFn rpc_fn =
         boost::bind(&KrpcDataStreamSender::Channel::DoEndDataStreamRpc, this);
     const string& prepend =
-        Substitute("EndDataStream() to $0 failed", TNetworkAddressToString(address_));
+        Substitute("EndDataStream() to $0 failed", NetworkAddressPBToString(address_));
     HandleFailedRPC(rpc_fn, controller_status, prepend);
   }
 }
@@ -645,9 +655,7 @@ Status KrpcDataStreamSender::Channel::DoEndDataStreamRpc() {
   DCHECK(rpc_in_flight_);
   EndDataStreamRequestPB eos_req;
   rpc_controller_.Reset();
-  UniqueIdPB* finstance_id_pb = eos_req.mutable_dest_fragment_instance_id();
-  finstance_id_pb->set_lo(fragment_instance_id_.lo);
-  finstance_id_pb->set_hi(fragment_instance_id_.hi);
+  *eos_req.mutable_dest_fragment_instance_id() = fragment_instance_id_;
   eos_req.set_sender_id(parent_->sender_id_);
   eos_req.set_dest_node_id(dest_node_id_);
   eos_resp_.Clear();
@@ -705,7 +713,7 @@ void KrpcDataStreamSender::Channel::Teardown(RuntimeState* state) {
 
 KrpcDataStreamSender::KrpcDataStreamSender(TDataSinkId sink_id, int sender_id,
     const KrpcDataStreamSenderConfig& sink_config, const TDataStreamSink& sink,
-    const std::vector<TPlanFragmentDestination>& destinations,
+    const google::protobuf::RepeatedPtrField<PlanFragmentDestinationPB>& destinations,
     int per_channel_buffer_size, RuntimeState* state)
   : DataSink(sink_id, sink_config,
         Substitute("KrpcDataStreamSender (dst_id=$0)", sink.dest_node_id), state),
@@ -715,6 +723,7 @@ KrpcDataStreamSender::KrpcDataStreamSender(TDataSinkId sink_id, int sender_id,
     partition_exprs_(sink_config.partition_exprs_),
     dest_node_id_(sink.dest_node_id),
     next_unknown_partition_(0),
+    exchange_hash_seed_(sink_config.exchange_hash_seed_),
     hash_and_add_rows_fn_(sink_config.hash_and_add_rows_fn_) {
   DCHECK_GT(destinations.size(), 0);
   DCHECK(sink.output_partition.type == TPartitionType::UNPARTITIONED
@@ -722,11 +731,10 @@ KrpcDataStreamSender::KrpcDataStreamSender(TDataSinkId sink_id, int sender_id,
       || sink.output_partition.type == TPartitionType::RANDOM
       || sink.output_partition.type == TPartitionType::KUDU);
 
-  for (int i = 0; i < destinations.size(); ++i) {
-    channels_.emplace_back(
-        new Channel(this, row_desc_, destinations[i].thrift_backend.hostname,
-            destinations[i].krpc_backend, destinations[i].fragment_instance_id,
-            sink.dest_node_id, per_channel_buffer_size));
+  for (const auto& destination : destinations) {
+    channels_.emplace_back(new Channel(this, row_desc_, destination.address().hostname(),
+        destination.krpc_backend(), destination.fragment_instance_id(), sink.dest_node_id,
+        per_channel_buffer_size));
   }
 
   if (partition_type_ == TPartitionType::UNPARTITIONED
@@ -770,12 +778,12 @@ Status KrpcDataStreamSender::Prepare(
   for (int i = 0; i < channels_.size(); ++i) {
     RETURN_IF_ERROR(channels_[i]->Init(state));
   }
-  state->CheckAndAddCodegenDisabledMessage(profile());
   return Status::OK();
 }
 
 Status KrpcDataStreamSender::Open(RuntimeState* state) {
   SCOPED_TIMER(profile_->total_time_counter());
+  RETURN_IF_ERROR(DataSink::Open(state));
   return ScalarExprEvaluator::Open(partition_expr_evals_, state);
 }
 
@@ -831,7 +839,7 @@ Status KrpcDataStreamSenderConfig::CodegenHashRow(
 
   // Store the initial seed to hash_val
   llvm::Value* hash_val =
-      codegen->GetI64Constant(KrpcDataStreamSender::EXCHANGE_HASH_SEED);
+      codegen->GetI64Constant(exchange_hash_seed_);
 
   // Unroll the loop and codegen each of the partition expressions
   for (int i = 0; i < partition_exprs_.size(); ++i) {
@@ -918,22 +926,15 @@ string KrpcDataStreamSenderConfig::PartitionTypeName() const {
   }
 }
 
-void KrpcDataStreamSender::Codegen(RuntimeState* state) {
-  const KrpcDataStreamSenderConfig& config =
-      static_cast<const KrpcDataStreamSenderConfig&>(sink_config_);
-  KrpcDataStreamSenderConfig& non_const_config =
-      const_cast<KrpcDataStreamSenderConfig&>(config);
-  non_const_config.Codegen(state, profile());
-}
-
-void KrpcDataStreamSenderConfig::Codegen(RuntimeState* state, RuntimeProfile* profile) {
+void KrpcDataStreamSenderConfig::Codegen(FragmentState* state) {
   LlvmCodeGen* codegen = state->codegen();
   DCHECK(codegen != nullptr);
   const string sender_name = PartitionTypeName() + " Sender";
   if (partition_type_ != TPartitionType::HASH_PARTITIONED) {
     const string& msg = Substitute("not $0",
         partition_type_ == TPartitionType::KUDU ? "supported" : "needed");
-    profile->AddCodegenMsg(false, msg, sender_name);
+    codegen_status_msgs_.emplace_back(
+        FragmentState::GenerateCodegenMsg(false, msg, sender_name));
     return;
   }
 
@@ -946,9 +947,8 @@ void KrpcDataStreamSenderConfig::Codegen(RuntimeState* state, RuntimeProfile* pr
 
     int num_replaced;
     // Replace GetNumChannels() with a constant.
-    int num_channels =  state->fragment_ctx().destinations.size();
     num_replaced = codegen->ReplaceCallSitesWithValue(hash_and_add_rows_fn,
-        codegen->GetI32Constant(num_channels), "GetNumChannels");
+        codegen->GetI32Constant(num_channels_), "GetNumChannels");
     DCHECK_EQ(num_replaced, 1);
 
     // Replace HashRow() with the handcrafted IR function.
@@ -961,11 +961,10 @@ void KrpcDataStreamSenderConfig::Codegen(RuntimeState* state, RuntimeProfile* pr
       codegen_status =
           Status("Codegen'd HashAndAddRows() failed verification. See log");
     } else {
-      codegen->AddFunctionToJit(hash_and_add_rows_fn,
-          reinterpret_cast<void**>(&hash_and_add_rows_fn_));
+      codegen->AddFunctionToJit(hash_and_add_rows_fn, &hash_and_add_rows_fn_);
     }
   }
-  profile->AddCodegenMsg(codegen_status.ok(), codegen_status, sender_name);
+  AddCodegenStatus(codegen_status, sender_name);
 }
 
 Status KrpcDataStreamSender::AddRowToChannel(const int channel_id, TupleRow* row) {
@@ -973,7 +972,7 @@ Status KrpcDataStreamSender::AddRowToChannel(const int channel_id, TupleRow* row
 }
 
 uint64_t KrpcDataStreamSender::HashRow(TupleRow* row) {
-  uint64_t hash_val = EXCHANGE_HASH_SEED;
+  uint64_t hash_val = exchange_hash_seed_;
   for (ScalarExprEvaluator* eval : partition_expr_evals_) {
     void* partition_val = eval->GetValue(row);
     // We can't use the crc hash function here because it does not result in
@@ -1035,8 +1034,10 @@ Status KrpcDataStreamSender::Send(RuntimeState* state, RowBatch* batch) {
     }
   } else {
     DCHECK_EQ(partition_type_, TPartitionType::HASH_PARTITIONED);
-    if (hash_and_add_rows_fn_ != nullptr) {
-      RETURN_IF_ERROR(hash_and_add_rows_fn_(this, batch));
+    const KrpcDataStreamSenderConfig::HashAndAddRowsFn hash_and_add_rows_fn =
+        hash_and_add_rows_fn_.load();
+    if (hash_and_add_rows_fn != nullptr) {
+      RETURN_IF_ERROR(hash_and_add_rows_fn(this, batch));
     } else {
       RETURN_IF_ERROR(HashAndAddRows(batch));
     }
