@@ -38,16 +38,17 @@ class MemTracker;
 class RowDescriptor;
 class TDataStreamSink;
 class TNetworkAddress;
-class TPlanFragmentDestination;
+class PlanFragmentDestinationPB;
 
 class KrpcDataStreamSenderConfig : public DataSinkConfig {
  public:
-  DataSink* CreateSink(const TPlanFragmentCtx& fragment_ctx,
-      const TPlanFragmentInstanceCtx& fragment_instance_ctx,
-      RuntimeState* state) const override;
+  DataSink* CreateSink(RuntimeState* state) const override;
   void Close() override;
 
-  void Codegen(RuntimeState* state, RuntimeProfile* profile);
+  /// Codegen KrpcDataStreamSender::HashAndAddRows() if partitioning type is
+  /// HASH_PARTITIONED. Replaces KrpcDataStreamSender::HashRow() and
+  /// KrpcDataStreamSender::GetNumChannels() based on runtime information.
+  void Codegen(FragmentState* state) override;
 
   /// The type of partitioning to perform.
   TPartitionType::type partition_type_ = TPartitionType::UNPARTITIONED;
@@ -56,16 +57,22 @@ class KrpcDataStreamSenderConfig : public DataSinkConfig {
   /// per-row partition values for shuffling exchange;
   std::vector<ScalarExpr*> partition_exprs_;
 
+  /// The number of channels that this node will create.
+  int  num_channels_;
+
+  /// Hash seed used for exchanges. Query id will be used to seed the hash function.
+  uint64_t exchange_hash_seed_;
+
   /// Type and pointer for the codegen'd KrpcDataStreamSender::HashAndAddRows()
   /// function. NULL if codegen is disabled or failed.
   typedef Status (*HashAndAddRowsFn)(KrpcDataStreamSender*, RowBatch* row);
-  HashAndAddRowsFn hash_and_add_rows_fn_ = nullptr;
+  CodegenFnPtr<HashAndAddRowsFn> hash_and_add_rows_fn_;
 
   ~KrpcDataStreamSenderConfig() override {}
 
  protected:
   Status Init(const TDataSink& tsink, const RowDescriptor* input_row_desc,
-      RuntimeState* state) override;
+      FragmentState* state) override;
 
  private:
   /// Codegen the KrpcDataStreamSender::HashRow() function and returns the codegen'd
@@ -99,7 +106,7 @@ class KrpcDataStreamSender : public DataSink {
   /// and RANDOM.
   KrpcDataStreamSender(TDataSinkId sink_id, int sender_id,
       const KrpcDataStreamSenderConfig& sink_config, const TDataStreamSink& sink,
-      const std::vector<TPlanFragmentDestination>& destinations,
+      const google::protobuf::RepeatedPtrField<PlanFragmentDestinationPB>& destinations,
       int per_channel_buffer_size, RuntimeState* state);
 
   virtual ~KrpcDataStreamSender();
@@ -108,10 +115,6 @@ class KrpcDataStreamSender : public DataSink {
   /// the stat counters. Return error status if any channels failed to initialize.
   virtual Status Prepare(RuntimeState* state, MemTracker* parent_mem_tracker) override;
 
-  /// Codegen HashAndAddRows() if partitioning type is HASH_PARTITIONED.
-  /// Replaces HashRow() and GetNumChannels() based on runtime information.
-  virtual void Codegen(RuntimeState* state) override;
-
   /// Initializes the evaluator of the partitioning expressions. Return error status
   /// if initialization failed.
   virtual Status Open(RuntimeState* state) override;
@@ -139,8 +142,8 @@ class KrpcDataStreamSender : public DataSink {
   /// KrpcDataStreamSender::HashRow() symbol. Used for call-site replacement.
   static const char* HASH_ROW_SYMBOL;
 
-  /// An arbitrary hash seed used for exchanges.
-  static constexpr uint64_t EXCHANGE_HASH_SEED = 0x66bd68df22c3ef37;
+  /// An arbitrary constant used to seed the hash.
+  static constexpr uint64_t EXCHANGE_HASH_SEED_CONST = 0x66bd68df22c3ef37;
 
   static const char* LLVM_CLASS_NAME;
 
@@ -262,9 +265,12 @@ class KrpcDataStreamSender : public DataSink {
   /// or when errors are encountered.
   int next_unknown_partition_;
 
+  /// Hash seed used for exchanges. Query id will be used to seed the hash function.
+  uint64_t exchange_hash_seed_;
+
   /// Pointer for the codegen'd HashAndAddRows() function.
   /// NULL if codegen is disabled or failed.
-  const KrpcDataStreamSenderConfig::HashAndAddRowsFn& hash_and_add_rows_fn_;
+  const CodegenFnPtr<KrpcDataStreamSenderConfig::HashAndAddRowsFn>& hash_and_add_rows_fn_;
 };
 
 } // namespace impala
