@@ -19,6 +19,7 @@
 
 #include <mutex>
 
+#include "codegen/llvm-codegen.h"
 #include "common/thread-debug-info.h"
 #include "exec/kudu-util.h"
 #include "exprs/expr.h"
@@ -28,23 +29,27 @@
 #include "kudu/util/monotime.h"
 #include "kudu/util/status.h"
 #include "rpc/rpc-mgr.h"
-#include "runtime/backend-client.h"
+#include "rpc/thrift-util.h"
 #include "runtime/bufferpool/buffer-pool.h"
 #include "runtime/bufferpool/reservation-tracker.h"
 #include "runtime/bufferpool/reservation-util.h"
 #include "runtime/exec-env.h"
 #include "runtime/fragment-instance-state.h"
+#include "runtime/fragment-state.h"
 #include "runtime/initial-reservations.h"
+#include "runtime/krpc-data-stream-mgr.h"
 #include "runtime/mem-tracker.h"
 #include "runtime/query-exec-mgr.h"
 #include "runtime/runtime-filter-bank.h"
 #include "runtime/runtime-state.h"
 #include "runtime/scanner-mem-limiter.h"
+#include "runtime/tmp-file-mgr.h"
 #include "service/control-service.h"
 #include "service/data-stream-service.h"
 #include "util/container-util.h"
 #include "util/debug-util.h"
 #include "util/impalad-metrics.h"
+#include "util/memory-metrics.h"
 #include "util/metrics.h"
 #include "util/system-state-info.h"
 #include "util/thread.h"
@@ -66,7 +71,27 @@ DECLARE_int64(rpc_max_message_size);
 DEFINE_int32_hidden(stress_status_report_delay_ms, 0, "Stress option to inject a delay "
     "before status reports. Has no effect on release builds.");
 
-using namespace impala;
+namespace impala {
+
+PROFILE_DEFINE_DERIVED_COUNTER(GcCount, STABLE_LOW, TUnit::UNIT,
+    "Per-Impalad Counter: The number of GC collections that have occurred in the Impala "
+    "process over the duration of the query. Reported by JMX.");
+PROFILE_DEFINE_DERIVED_COUNTER(GcTimeMillis, STABLE_LOW, TUnit::TIME_MS,
+    "Per-Impalad Counter: The amount of time spent in GC in the Impala process over the "
+    "duration of the query. Reported by JMX.");
+PROFILE_DEFINE_DERIVED_COUNTER(GcNumWarnThresholdExceeded, STABLE_LOW,
+    TUnit::UNIT,
+    "Per-Impalad Counter: The number of JVM process pauses that occurred in this Impala "
+    "process over the duration of the query. Tracks the number of pauses at the WARN "
+    "threshold. See JvmPauseMonitor for details. Reported by the JvmPauseMonitor.");
+PROFILE_DEFINE_DERIVED_COUNTER(GcNumInfoThresholdExceeded, STABLE_LOW,
+    TUnit::UNIT,
+    "Per-Impalad Counter: The number of JVM process pauses that occurred in this Impala "
+    "process over the duration of the query. Tracks the number of pauses at the INFO "
+    "threshold. See JvmPauseMonitor for details. Reported by the JvmPauseMonitor.");
+PROFILE_DEFINE_DERIVED_COUNTER(GcTotalExtraSleepTimeMillis, STABLE_LOW, TUnit::TIME_MS,
+    "Per-Impalad Counter: The amount of time the JVM process paused over the duration "
+    "of the query. See JvmPauseMonitor for details. Reported by the JvmPauseMonitor.");
 
 QueryState::ScopedRef::ScopedRef(const TUniqueId& query_id) {
   DCHECK(ExecEnv::GetInstance()->query_exec_mgr() != nullptr);
@@ -115,6 +140,10 @@ void QueryState::ReleaseBackendResources() {
   if (initial_reservations_ != nullptr) initial_reservations_->ReleaseResources();
   if (buffer_reservation_ != nullptr) buffer_reservation_->Close();
   if (desc_tbl_ != nullptr) desc_tbl_->ReleaseResources();
+  // Release any memory associated with codegen.
+  for (auto& elem : fragment_state_map_) {
+    elem.second->ReleaseResources();
+  }
   // Mark the query as finished on the query MemTracker so that admission control will
   // not consider the whole query memory limit to be "reserved".
   query_mem_tracker_->set_query_exec_finished();
@@ -140,15 +169,57 @@ QueryState::~QueryState() {
 
 Status QueryState::Init(const ExecQueryFInstancesRequestPB* exec_rpc_params,
     const TExecPlanFragmentInfo& fragment_info) {
+  std::lock_guard<std::mutex> l(init_lock_);
   // Decremented in QueryExecMgr::StartQueryHelper() on success or by the caller of
   // Init() on failure. We need to do this before any returns because Init() always
   // returns a resource refcount to its caller.
   AcquireBackendResourceRefcount();
 
+  if (IsCancelled()) return Status::CANCELLED;
+
   RETURN_IF_ERROR(DebugAction(query_options(), "QUERY_STATE_INIT"));
 
   ExecEnv* exec_env = ExecEnv::GetInstance();
 
+  RuntimeProfile* jvm_host_profile = RuntimeProfile::Create(&obj_pool_, "JVM");
+  host_profile_->AddChild(jvm_host_profile);
+
+  int64_t gc_count = JvmMemoryCounterMetric::GC_COUNT->GetValue();
+  PROFILE_GcCount.Instantiate(jvm_host_profile,
+      [gc_count]() {
+        return JvmMemoryCounterMetric::GC_COUNT->GetValue() - gc_count;
+      });
+
+  int64_t gc_time_millis = JvmMemoryCounterMetric::GC_TIME_MILLIS->GetValue();
+  PROFILE_GcTimeMillis.Instantiate(jvm_host_profile,
+      [gc_time_millis]() {
+        return JvmMemoryCounterMetric::GC_TIME_MILLIS->GetValue() - gc_time_millis;
+      });
+
+  int64_t gc_num_warn_threshold_exceeded =
+      JvmMemoryCounterMetric::GC_NUM_WARN_THRESHOLD_EXCEEDED->GetValue();
+  PROFILE_GcNumWarnThresholdExceeded.Instantiate(jvm_host_profile,
+      [gc_num_warn_threshold_exceeded]() {
+        return JvmMemoryCounterMetric::GC_NUM_WARN_THRESHOLD_EXCEEDED->GetValue()
+            - gc_num_warn_threshold_exceeded;
+      });
+
+  int64_t gc_num_info_threshold_exceeded =
+      JvmMemoryCounterMetric::GC_NUM_INFO_THRESHOLD_EXCEEDED->GetValue();
+  PROFILE_GcNumInfoThresholdExceeded.Instantiate(jvm_host_profile,
+      [gc_num_info_threshold_exceeded]() {
+        return JvmMemoryCounterMetric::GC_NUM_INFO_THRESHOLD_EXCEEDED->GetValue()
+            - gc_num_info_threshold_exceeded;
+      });
+
+  int64_t gc_total_extra_sleep_time_millis =
+      JvmMemoryCounterMetric::GC_TOTAL_EXTRA_SLEEP_TIME_MILLIS->GetValue();
+  PROFILE_GcTotalExtraSleepTimeMillis.Instantiate(jvm_host_profile,
+      [gc_total_extra_sleep_time_millis]() {
+        return JvmMemoryCounterMetric::GC_TOTAL_EXTRA_SLEEP_TIME_MILLIS->GetValue()
+            - gc_total_extra_sleep_time_millis;
+      });
+
   // Initialize resource tracking counters.
   if (query_ctx().trace_resource_usage) {
     SystemStateInfo* system_state_info = exec_env->system_state_info();
@@ -198,24 +269,25 @@ Status QueryState::Init(const ExecQueryFInstancesRequestPB* exec_rpc_params,
   RETURN_IF_ERROR(InitBufferPoolState());
 
   // Initialize the RPC proxy once and report any error.
-  RETURN_IF_ERROR(ControlService::GetProxy(query_ctx().coord_krpc_address,
-      query_ctx().coord_address.hostname, &proxy_));
+  RETURN_IF_ERROR(ControlService::GetProxy(
+      query_ctx().coord_ip_address, query_ctx().coord_hostname, &proxy_));
 
   // don't copy query_ctx, it's large and we already did that in the c'tor
   exec_rpc_params_.set_coord_state_idx(exec_rpc_params->coord_state_idx());
+  exec_rpc_params_.mutable_fragment_ctxs()->Swap(
+      const_cast<google::protobuf::RepeatedPtrField<impala::PlanFragmentCtxPB>*>(
+          &exec_rpc_params->fragment_ctxs()));
+  exec_rpc_params_.mutable_fragment_instance_ctxs()->Swap(
+      const_cast<google::protobuf::RepeatedPtrField<impala::PlanFragmentInstanceCtxPB>*>(
+          &exec_rpc_params->fragment_instance_ctxs()));
   TExecPlanFragmentInfo& non_const_fragment_info =
       const_cast<TExecPlanFragmentInfo&>(fragment_info);
-  fragment_info_.fragment_ctxs.swap(non_const_fragment_info.fragment_ctxs);
-  fragment_info_.__isset.fragment_ctxs = true;
+  fragment_info_.fragments.swap(non_const_fragment_info.fragments);
+  fragment_info_.__isset.fragments = true;
   fragment_info_.fragment_instance_ctxs.swap(
       non_const_fragment_info.fragment_instance_ctxs);
   fragment_info_.__isset.fragment_instance_ctxs = true;
 
-  instances_prepared_barrier_.reset(
-      new CountingBarrier(fragment_info_.fragment_instance_ctxs.size()));
-  instances_finished_barrier_.reset(
-      new CountingBarrier(fragment_info_.fragment_instance_ctxs.size()));
-
   // Claim the query-wide minimum reservation. Do this last so that we don't need
   // to handle releasing it if a later step fails.
   initial_reservations_ =
@@ -225,11 +297,24 @@ Status QueryState::Init(const ExecQueryFInstancesRequestPB* exec_rpc_params,
       query_id(), exec_rpc_params->min_mem_reservation_bytes()));
   RETURN_IF_ERROR(InitFilterBank());
   scanner_mem_limiter_ = obj_pool_.Add(new ScannerMemLimiter);
+
+  // Set barriers only for successful initialization. Otherwise the barriers
+  // never be notified.
+  instances_prepared_barrier_.reset(
+      new CountingBarrier(fragment_info_.fragment_instance_ctxs.size()));
+  instances_finished_barrier_.reset(
+      new CountingBarrier(fragment_info_.fragment_instance_ctxs.size()));
+  is_initialized_ = true;
   return Status::OK();
 }
 
-Status QueryState::InitBufferPoolState() {
-  ExecEnv* exec_env = ExecEnv::GetInstance();
+UniqueIdPB QueryState::GetCoordinatorBackendId() const {
+  UniqueIdPB backend_id_pb;
+  TUniqueIdToUniqueIdPB(query_ctx_.coord_backend_id, &backend_id_pb);
+  return backend_id_pb;
+}
+
+int64_t QueryState::GetMaxReservation() {
   int64_t mem_limit = query_mem_tracker_->GetLowestLimit(MemLimit::HARD);
   int64_t max_reservation;
   if (query_options().__isset.buffer_pool_limit
@@ -243,6 +328,12 @@ Status QueryState::InitBufferPoolState() {
     DCHECK_GE(mem_limit, 0);
     max_reservation = ReservationUtil::GetReservationLimitFromMemLimit(mem_limit);
   }
+  return max_reservation;
+}
+
+Status QueryState::InitBufferPoolState() {
+  ExecEnv* exec_env = ExecEnv::GetInstance();
+  int64_t max_reservation = GetMaxReservation();
   VLOG(2) << "Buffer pool limit for " << PrintId(query_id()) << ": " << max_reservation;
 
   buffer_reservation_ = obj_pool_.Add(new ReservationTracker);
@@ -251,22 +342,44 @@ Status QueryState::InitBufferPoolState() {
 
   if (query_options().scratch_limit != 0 && !query_ctx_.disable_spilling) {
     file_group_ = obj_pool_.Add(
-        new TmpFileMgr::FileGroup(exec_env->tmp_file_mgr(), exec_env->disk_io_mgr(),
+        new TmpFileGroup(exec_env->tmp_file_mgr(), exec_env->disk_io_mgr(),
             host_profile_, query_id(), query_options().scratch_limit));
+    if (!query_options().debug_action.empty()) {
+      file_group_->SetDebugAction(query_options().debug_action);
+    }
   }
   return Status::OK();
 }
 
+// Verifies the filters produced by all instances on the same backend are the same.
+bool VerifyFiltersProduced(const vector<TPlanFragmentInstanceCtx>& instance_ctxs) {
+  int fragment_idx = -1;
+  std::unordered_set<int> first_set;
+  for (const TPlanFragmentInstanceCtx& instance_ctx : instance_ctxs) {
+    bool first_instance_of_fragment =
+        fragment_idx == -1 || fragment_idx != instance_ctx.fragment_idx;
+    if (first_instance_of_fragment) {
+      fragment_idx = instance_ctx.fragment_idx;
+      first_set.clear();
+      for (auto f : instance_ctx.filters_produced) first_set.insert(f.filter_id);
+    }
+    if (first_set.size() != instance_ctx.filters_produced.size()) return false;
+    for (auto f : instance_ctx.filters_produced) {
+      if (first_set.find(f.filter_id) == first_set.end()) return false;
+    }
+  }
+  return true;
+}
+
 Status QueryState::InitFilterBank() {
   int64_t runtime_filters_reservation_bytes = 0;
   int fragment_ctx_idx = -1;
-  const vector<TPlanFragmentCtx>& fragment_ctxs = fragment_info_.fragment_ctxs;
+  const vector<TPlanFragment>& fragments = fragment_info_.fragments;
   const vector<TPlanFragmentInstanceCtx>& instance_ctxs =
       fragment_info_.fragment_instance_ctxs;
   // Add entries for all produced and consumed filters.
   unordered_map<int32_t, FilterRegistration> filters;
-  for (const TPlanFragmentCtx& fragment_ctx : fragment_ctxs) {
-    const TPlanFragment& fragment = fragment_ctx.fragment;
+  for (const TPlanFragment& fragment : fragments) {
     for (const TPlanNode& plan_node : fragment.plan.nodes) {
       if (!plan_node.__isset.runtime_filters) continue;
       for (const TRuntimeFilterDesc& filter : plan_node.runtime_filters) {
@@ -285,17 +398,19 @@ Status QueryState::InitFilterBank() {
       }
     }
   }
+  DCHECK(VerifyFiltersProduced(instance_ctxs))
+      << "Filters produced by all instances on the same backend should be the same";
   for (const TPlanFragmentInstanceCtx& instance_ctx : instance_ctxs) {
     bool first_instance_of_fragment = fragment_ctx_idx == -1
-        || fragment_ctxs[fragment_ctx_idx].fragment.idx != instance_ctx.fragment_idx;
+        || fragments[fragment_ctx_idx].idx != instance_ctx.fragment_idx;
     if (first_instance_of_fragment) {
       ++fragment_ctx_idx;
-      DCHECK_EQ(fragment_ctxs[fragment_ctx_idx].fragment.idx, instance_ctx.fragment_idx);
+      DCHECK_EQ(fragments[fragment_ctx_idx].idx, instance_ctx.fragment_idx);
     }
     // TODO: this over-reserves memory a bit in a couple of cases:
     // * if different fragments on this backend consume or produce the same filter.
     // * if a finstance was chosen not to produce a global broadcast filter.
-    const TPlanFragment& fragment = fragment_ctxs[fragment_ctx_idx].fragment;
+    const TPlanFragment& fragment = fragments[fragment_ctx_idx];
     runtime_filters_reservation_bytes +=
         fragment.produced_runtime_filters_reservation_bytes;
     if (first_instance_of_fragment) {
@@ -347,10 +462,13 @@ void QueryState::UpdateBackendExecState() {
           BackendExecState::EXECUTING : BackendExecState::FINISHED;
     }
   }
-  // Send one last report if the query has reached the terminal state.
+  // Send one last report if the query has reached the terminal state
+  // and the coordinator is active.
   if (IsTerminalState()) {
     VLOG_QUERY << "UpdateBackendExecState(): last report for " << PrintId(query_id());
-    while (!ReportExecStatus()) SleepForMs(GetReportWaitTimeMs());
+    while (is_coord_active_.Load() && !ReportExecStatus()) {
+      SleepForMs(GetReportWaitTimeMs());
+    }
   }
 }
 
@@ -363,40 +481,144 @@ Status QueryState::GetFInstanceState(
   return Status::OK();
 }
 
+int64_t QueryState::AsyncCodegenThreadHelper(const std::string& suffix) const {
+  int64_t res = 0;
+  vector<RuntimeProfile::Counter*> counters;
+  host_profile_->GetCounters(
+      LlvmCodeGen::ASYNC_CODEGEN_THREAD_COUNTERS_PREFIX + suffix, &counters);
+
+  for (const RuntimeProfile::Counter* counter : counters) {
+    DCHECK(counter != nullptr);
+    res += counter->value();
+  }
+
+  return res;
+}
+
+int64_t QueryState::AsyncCodegenThreadUserTime() const {
+  return AsyncCodegenThreadHelper("UserTime");
+}
+
+int64_t QueryState::AsyncCodegenThreadSysTime() const {
+  return AsyncCodegenThreadHelper("SysTime");
+}
+
 void QueryState::ConstructReport(bool instances_started,
     ReportExecStatusRequestPB* report, TRuntimeProfileForest* profiles_forest) {
   report->Clear();
+  report->set_backend_report_seq_no(++last_report_seq_no_);
   TUniqueIdToUniqueIdPB(query_id(), report->mutable_query_id());
   DCHECK(exec_rpc_params_.has_coord_state_idx());
   report->set_coord_state_idx(exec_rpc_params_.coord_state_idx());
+  Status report_overall_status;
   {
     unique_lock<SpinLock> l(status_lock_);
+
+    Status debug_action_status =
+        DebugAction(query_options(), "CONSTRUCT_QUERY_STATE_REPORT");
+    if (UNLIKELY(!debug_action_status.ok())) overall_status_ = debug_action_status;
+
     overall_status_.ToProto(report->mutable_overall_status());
+    report_overall_status = overall_status_;
     if (IsValidFInstanceId(failed_finstance_id_)) {
       TUniqueIdToUniqueIdPB(failed_finstance_id_, report->mutable_fragment_instance_id());
     }
   }
+  if (!report_overall_status.ok() && query_spilled_.Load() == 1 && file_group_ != nullptr
+      && file_group_->IsSpillingDiskFaulty()) {
+    report->set_local_disk_faulty(true);
+  }
 
   // Add profile to report
   host_profile_->ToThrift(&profiles_forest->host_profile);
   profiles_forest->__isset.host_profile = true;
+
   // Free resources in chunked counters in the profile
   host_profile_->ClearChunkedTimeSeriesCounters();
 
   if (instances_started) {
+    // Map from fragment idx to the averaged profile. When aggregated profiles are
+    // enabled (IMPALA-9382), we populate this map with profiles that are aggregated
+    // from all the instances on this backend.
+    unordered_map<int, AggregatedRuntimeProfile*> agg_profiles;
+    ObjectPool agg_profile_pool;
+
+    // Stats that we aggregate across the instances.
+    int64_t cpu_user_ns = AsyncCodegenThreadUserTime();
+    int64_t cpu_sys_ns = AsyncCodegenThreadSysTime();
+    int64_t bytes_read = 0;
+    int64_t scan_ranges_complete = 0;
+    int64_t exchange_bytes_sent = 0;
+    int64_t scan_bytes_sent = 0;
+    std::map<int32_t, int64_t> per_join_rows_produced;
+
     for (const auto& entry : fis_map_) {
       FragmentInstanceState* fis = entry.second;
 
       // If this fragment instance has already sent its last report, skip it.
       if (fis->final_report_sent()) {
         DCHECK(fis->IsDone());
-        continue;
+      } else {
+        // Update the status and profiles of this fragment instance.
+        FragmentInstanceExecStatusPB* instance_status =
+            report->add_instance_exec_status();
+        if (query_ctx_.gen_aggregated_profile) {
+          int fragment_idx = fis->instance_ctx().fragment_idx;
+          AggregatedRuntimeProfile*& agg_profile = agg_profiles[fragment_idx];
+          if (agg_profile == nullptr) {
+            const auto it = fragment_state_map_.find(fragment_idx);
+            DCHECK(it != fragment_state_map_.end());
+            agg_profile = AggregatedRuntimeProfile::Create(&agg_profile_pool,
+                "tmp profile", it->second->instance_ctxs().size(), /*is_root=*/ true);
+          }
+          fis->GetStatusReport(
+              instance_status, nullptr, agg_profile, report_overall_status);
+        } else {
+          profiles_forest->profile_trees.emplace_back();
+          fis->GetStatusReport(instance_status, &profiles_forest->profile_trees.back(),
+              nullptr, report_overall_status);
+        }
       }
 
-      // Update the status and profiles of this fragment instance.
-      FragmentInstanceExecStatusPB* instance_status = report->add_instance_exec_status();
+      // Include these values for running and completed finstances in the status report.
+      cpu_user_ns += fis->cpu_user_ns();
+      cpu_sys_ns += fis->cpu_sys_ns();
+      bytes_read += fis->bytes_read();
+      scan_ranges_complete += fis->scan_ranges_complete();
+      // Determine whether this instance had a scan node in its plan.
+      // Note: this is hacky. E.g. it doesn't work for Kudu scans.
+      if (fis->bytes_read() > 0) {
+        scan_bytes_sent += fis->total_bytes_sent();
+      } else {
+        exchange_bytes_sent += fis->total_bytes_sent();
+      }
+      MergeMapValues(fis->per_join_rows_produced(), &per_join_rows_produced);
+    }
+
+    // Construct the per-fragment status reports, including runtime profiles.
+    for (const auto& entry : agg_profiles) {
+      int fragment_idx = entry.first;
+      const AggregatedRuntimeProfile* agg_profile = entry.second;
+      const auto it = fragment_state_map_.find(fragment_idx);
+      DCHECK(it != fragment_state_map_.end());
+
+      // Add the aggregated runtime profile and additional metadata to the report.
+      FragmentExecStatusPB* fragment_status = report->add_fragment_exec_status();
+      fragment_status->set_fragment_idx(fragment_idx);
+      fragment_status->set_min_per_fragment_instance_idx(
+          it->second->min_per_fragment_instance_idx());
       profiles_forest->profile_trees.emplace_back();
-      fis->GetStatusReport(instance_status, &profiles_forest->profile_trees.back());
+      agg_profile->ToThrift(&profiles_forest->profile_trees.back());
+    }
+    report->set_peak_mem_consumption(query_mem_tracker_->peak_consumption());
+    report->set_cpu_user_ns(cpu_user_ns);
+    report->set_cpu_sys_ns(cpu_sys_ns);
+    report->set_bytes_read(bytes_read);
+    report->set_scan_ranges_complete(scan_ranges_complete);
+    report->set_exchange_bytes_sent(exchange_bytes_sent);
+    report->set_scan_bytes_sent(scan_bytes_sent);
+    for (const auto& entry : per_join_rows_produced) {
+      (*report->mutable_per_join_rows_produced())[entry.first] = entry.second;
     }
   }
 }
@@ -508,6 +730,7 @@ bool QueryState::ReportExecStatus() {
     if (!rpc_status.ok()) {
       LOG(ERROR) << "Cancelling fragment instances due to failure to reach the "
                  << "coordinator. (" << rpc_status.GetDetail() << ").";
+      is_coord_active_.Store(false);
     } else if (!result_status.ok()) {
       // If the ReportExecStatus RPC succeeded in reaching the coordinator and we get
       // back a non-OK status, it means that the coordinator expects us to cancel the
@@ -526,7 +749,23 @@ int64_t QueryState::GetReportWaitTimeMs() const {
   int64_t report_interval = query_ctx().status_report_interval_ms > 0 ?
       query_ctx().status_report_interval_ms :
       DEFAULT_REPORT_WAIT_TIME_MS;
-  return report_interval * (num_failed_reports_ + 1);
+  if (num_failed_reports_ == 0) {
+    return report_interval;
+  } else {
+    // Generate a random number between 0 and 1 - we'll retry sometime evenly distributed
+    // between 'report_interval' and 'report_interval * (num_failed_reports_ + 1)', so we
+    // won't hit the "thundering herd" problem.
+    float jitter = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
+    return report_interval * (num_failed_reports_ * jitter + 1);
+  }
+}
+
+void QueryState::ErrorDuringFragmentCodegen(const Status& status) {
+  unique_lock<SpinLock> l(status_lock_);
+  if (!HasErrorStatus()) {
+    overall_status_ = status;
+    failed_finstance_id_ = TUniqueId();
+  }
 }
 
 void QueryState::ErrorDuringPrepare(const Status& status, const TUniqueId& finst_id) {
@@ -548,7 +787,6 @@ void QueryState::ErrorDuringExecute(const Status& status, const TUniqueId& finst
       failed_finstance_id_ = finst_id;
     }
   }
-  instances_finished_barrier_->NotifyRemaining();
 }
 
 Status QueryState::WaitForPrepare() {
@@ -573,10 +811,9 @@ bool QueryState::StartFInstances() {
   DCHECK_GT(refcnt_.Load(), 0);
   DCHECK_GT(backend_resource_refcnt_.Load(), 0) << "Should have been taken in Init()";
 
-  DCHECK_GT(fragment_info_.fragment_ctxs.size(), 0);
-  TPlanFragmentCtx* fragment_ctx = &fragment_info_.fragment_ctxs[0];
+  DCHECK_GT(fragment_info_.fragments.size(), 0);
+  vector<unique_ptr<Thread>> codegen_threads;
   int num_unstarted_instances = fragment_info_.fragment_instance_ctxs.size();
-  int fragment_ctx_idx = 0;
 
   // set up desc tbl
   DCHECK(query_ctx().__isset.desc_tbl_serialized);
@@ -586,49 +823,53 @@ bool QueryState::StartFInstances() {
   VLOG(2) << "descriptor table for query=" << PrintId(query_id())
           << "\n" << desc_tbl_->DebugString();
 
+  start_finstances_status = FragmentState::CreateFragmentStateMap(
+      fragment_info_, exec_rpc_params_, this, fragment_state_map_);
+  if (UNLIKELY(!start_finstances_status.ok())) goto error;
+
   fragment_events_start_time_ = MonotonicStopWatch::Now();
-  for (const TPlanFragmentInstanceCtx& instance_ctx :
-      fragment_info_.fragment_instance_ctxs) {
-    // determine corresponding TPlanFragmentCtx
-    if (fragment_ctx->fragment.idx != instance_ctx.fragment_idx) {
-      ++fragment_ctx_idx;
-      DCHECK_LT(fragment_ctx_idx, fragment_info_.fragment_ctxs.size());
-      fragment_ctx = &fragment_info_.fragment_ctxs[fragment_ctx_idx];
-      // we expect fragment and instance contexts to follow the same order
-      DCHECK_EQ(fragment_ctx->fragment.idx, instance_ctx.fragment_idx);
+  for (auto& fragment : fragment_state_map_) {
+    FragmentState* fragment_state = fragment.second;
+    for (int i = 0; i < fragment_state->instance_ctxs().size(); ++i) {
+      const TPlanFragmentInstanceCtx* instance_ctx = fragment_state->instance_ctxs()[i];
+      const PlanFragmentInstanceCtxPB* instance_ctx_pb =
+          fragment_state->instance_ctx_pbs()[i];
+      DCHECK_EQ(instance_ctx->fragment_idx, instance_ctx_pb->fragment_idx());
+      FragmentInstanceState* fis = obj_pool_.Add(new FragmentInstanceState(
+          this, fragment_state, *instance_ctx, *instance_ctx_pb));
+
+      // start new thread to execute instance
+      refcnt_.Add(1); // decremented in ExecFInstance()
+      AcquireBackendResourceRefcount(); // decremented in ExecFInstance()
+
+      // Add the fragment instance ID to the 'fis_map_'. Has to happen before the thread
+      // is spawned or we may race with users of 'fis_map_'.
+      fis_map_.emplace(fis->instance_id(), fis);
+
+      string thread_name =
+          Substitute("$0 (finst:$1)", FragmentInstanceState::FINST_THREAD_NAME_PREFIX,
+              PrintId(instance_ctx->fragment_instance_id));
+      unique_ptr<Thread> t;
+
+      // Inject thread creation failures through debug actions if enabled.
+      Status debug_action_status =
+          DebugAction(query_options(), "FIS_FAIL_THREAD_CREATION");
+      start_finstances_status = !debug_action_status.ok() ?
+          debug_action_status :
+          Thread::Create(FragmentInstanceState::FINST_THREAD_GROUP_NAME, thread_name,
+              [this, fis]() { this->ExecFInstance(fis); }, &t, true);
+      if (!start_finstances_status.ok()) {
+        fis_map_.erase(fis->instance_id());
+        // Undo refcnt increments done immediately prior to Thread::Create(). The
+        // reference counts were both greater than zero before the increments, so
+        // neither of these decrements will free any structures.
+        ReleaseBackendResourceRefcount();
+        ExecEnv::GetInstance()->query_exec_mgr()->ReleaseQueryState(this);
+        goto error;
+      }
+      t->Detach();
+      --num_unstarted_instances;
     }
-    FragmentInstanceState* fis = obj_pool_.Add(
-        new FragmentInstanceState(this, *fragment_ctx, instance_ctx));
-
-    // start new thread to execute instance
-    refcnt_.Add(1); // decremented in ExecFInstance()
-    AcquireBackendResourceRefcount(); // decremented in ExecFInstance()
-
-    // Add the fragment instance ID to the 'fis_map_'. Has to happen before the thread is
-    // spawned or we may race with users of 'fis_map_'.
-    fis_map_.emplace(fis->instance_id(), fis);
-
-    string thread_name = Substitute("$0 (finst:$1)",
-        FragmentInstanceState::FINST_THREAD_NAME_PREFIX,
-        PrintId(instance_ctx.fragment_instance_id));
-    unique_ptr<Thread> t;
-
-    // Inject thread creation failures through debug actions if enabled.
-    Status debug_action_status = DebugAction(query_options(), "FIS_FAIL_THREAD_CREATION");
-    start_finstances_status = !debug_action_status.ok() ? debug_action_status :
-        Thread::Create(FragmentInstanceState::FINST_THREAD_GROUP_NAME, thread_name,
-            [this, fis]() { this->ExecFInstance(fis); }, &t, true);
-    if (!start_finstances_status.ok()) {
-      fis_map_.erase(fis->instance_id());
-      // Undo refcnt increments done immediately prior to Thread::Create(). The
-      // reference counts were both greater than zero before the increments, so
-      // neither of these decrements will free any structures.
-      ReleaseBackendResourceRefcount();
-      ExecEnv::GetInstance()->query_exec_mgr()->ReleaseQueryState(this);
-      goto error;
-    }
-    t->Detach();
-    --num_unstarted_instances;
   }
   return true;
 
@@ -681,6 +922,9 @@ done:
       DCHECK(entry.second->IsDone());
     }
   } else {
+    // If the query execution hit an error, when the final status report is sent, the
+    // coordinator's response will instruct the QueryState to cancel itself, so Cancel()
+    // should have always been called by this point.
     DCHECK_EQ(is_cancelled_.Load(), 1);
   }
 }
@@ -728,10 +972,19 @@ void QueryState::ExecFInstance(FragmentInstanceState* fis) {
 
 void QueryState::Cancel() {
   VLOG_QUERY << "Cancel: query_id=" << PrintId(query_id());
+  {
+    std::lock_guard<std::mutex> l(init_lock_);
+    if (!is_initialized_) {
+      discard_result(is_cancelled_.CompareAndSwap(0, 1));
+      return;
+    }
+  }
   discard_result(WaitForPrepare());
   if (!is_cancelled_.CompareAndSwap(0, 1)) return;
   if (filter_bank_ != nullptr) filter_bank_->Cancel();
   for (auto entry: fis_map_) entry.second->Cancel();
+  // Cancel data streams for all fragment instances.
+  ExecEnv::GetInstance()->stream_mgr()->Cancel(query_id());
 }
 
 void QueryState::PublishFilter(const PublishFilterParamsPB& params, RpcContext* context) {
@@ -756,3 +1009,4 @@ Status QueryState::StartSpilling(RuntimeState* runtime_state, MemTracker* mem_tr
   }
   return Status::OK();
 }
+}
