@@ -32,7 +32,6 @@
 #include "gen-cpp/control_service.pb.h"
 #include "gutil/macros.h"
 #include "gutil/threading/thread_collision_warner.h" // for DFAKE_*
-#include "runtime/tmp-file-mgr.h"
 #include "util/counting-barrier.h"
 #include "util/spinlock.h"
 #include "util/unique-id-hash.h"
@@ -46,16 +45,22 @@ class RpcContext;
 namespace impala {
 
 class ControlServiceProxy;
+class DataSinkConfig;
 class DescriptorTbl;
+class FragmentState;
 class FragmentInstanceState;
 class InitialReservations;
+class LlvmCodeGen;
 class MemTracker;
+class PlanNode;
 class PublishFilterParamsPB;
 class ReservationTracker;
 class RuntimeFilterBank;
 class RuntimeProfile;
 class RuntimeState;
+class ScalarExpr;
 class ScannerMemLimiter;
+class TmpFileGroup;
 class TRuntimeProfileForest;
 
 /// Central class for all backend execution state (example: the FragmentInstanceStates
@@ -104,7 +109,10 @@ class TRuntimeProfileForest;
 /// indicator). If execution ended with an error, that error status will be part of
 /// the final report (it will not be overridden by the resulting cancellation).
 ///
-/// Thread-safe, unless noted otherwise.
+/// Thread-safe Notes:
+/// - Init() must be called first. After Init() returns successfully, all other public
+///   functions are thread-safe to be called unless noted otherwise.
+/// - Cancel() is safe to be called at any time (before or during Init()).
 ///
 /// TODO:
 /// - set up kudu clients in Init(), remove related locking
@@ -144,6 +152,7 @@ class QueryState {
   }
   MemTracker* query_mem_tracker() const { return query_mem_tracker_; }
   RuntimeProfile* host_profile() const { return host_profile_; }
+  UniqueIdPB GetCoordinatorBackendId() const;
 
   /// The following getters are only valid after Init().
   ScannerMemLimiter* scanner_mem_limiter() const { return scanner_mem_limiter_; }
@@ -156,7 +165,7 @@ class QueryState {
     return buffer_reservation_;
   }
   InitialReservations* initial_reservations() const { return initial_reservations_; }
-  TmpFileMgr::FileGroup* file_group() const {
+  TmpFileGroup* file_group() const {
     DCHECK_GT(backend_resource_refcnt_.Load(), 0);
     return file_group_;
   }
@@ -178,7 +187,7 @@ class QueryState {
   /// it to the caller on both success and failure. The caller must release it by
   /// calling ReleaseBackendResourceRefcount().
   ///
-  /// Uses few cycles and never blocks. Not idempotent, not thread-safe.
+  /// Uses few cycles and blocks Cancel() to execute. Not idempotent.
   /// The remaining public functions must be called only after Init().
   Status Init(const ExecQueryFInstancesRequestPB* exec_rpc_params,
       const TExecPlanFragmentInfo& fragment_info) WARN_UNUSED_RESULT;
@@ -208,8 +217,13 @@ class QueryState {
 
   /// Cancels all actively executing fragment instances. Blocks until all fragment
   /// instances have finished their Prepare phase. Idempotent.
+  /// For uninitialized QueryState, just set is_cancelled_ and don't need to cancel
+  /// fragment instances.
   void Cancel();
 
+  /// Return true if the executing fragment instances have been cancelled.
+  bool IsCancelled() const { return (is_cancelled_.Load() == 1); }
+
   /// Increment the resource refcount. Must be decremented before the query state
   /// reference is released. A refcount should be held by a fragment or other entity
   /// for as long as it is consuming query backend execution resources (e.g. memory).
@@ -240,6 +254,16 @@ class QueryState {
   /// Called by a FragmentInstanceState thread to notify that it's done executing.
   void DoneExecuting() { discard_result(instances_finished_barrier_->Notify()); }
 
+  /// Called by a FragmentInstanceState thread to notify anyone waiting on
+  /// WaitForFinishOrTimeout() or waiting on WaitForFinish().
+  /// This function must be called after calling ErrorDuringExecute().
+  void DoneRemainingExecuting() { instances_finished_barrier_->NotifyRemaining(); }
+
+  /// Called to notify that an error was encountered during codegen. This is called by the
+  /// first fragment instance thread that invoked its corresponding FragmentState's
+  /// Codegen() and encountered the error.
+  void ErrorDuringFragmentCodegen(const Status& status);
+
   /// Called by a fragment instance thread to notify that it hit an error during Prepare()
   /// Updates the query status and the failed instance ID if it's not set already.
   /// Also notifies anyone waiting on WaitForPrepare() if this is called by the last
@@ -248,9 +272,15 @@ class QueryState {
 
   /// Called by a fragment instance thread to notify that it hit an error during Execute()
   /// Updates the query status and records the failed instance ID if they're not set
-  /// already. Also notifies anyone waiting on WaitForFinishOrTimeout().
+  /// already.
+  /// Caller must call DoneRemainingExecuting() after calling this function to notify
+  /// anyone waiting on WaitForFinishOrTimeout() or waiting on WaitForFinish().
   void ErrorDuringExecute(const Status& status, const TUniqueId& finst_id);
 
+  /// Get maximum reservation allowed for this query. MAX_INT64 means effectively
+  /// unlimited.
+  int64_t GetMaxReservation();
+
   /// The default BATCH_SIZE.
   static const int DEFAULT_BATCH_SIZE = 1024;
 
@@ -297,6 +327,11 @@ class QueryState {
   /// Thread-safety: Only updated by the query state thread.
   BackendExecState backend_exec_state_ = BackendExecState::PREPARING;
 
+  /// Monotonically increasing sequence number for status reports, incremented for each
+  /// report sent.
+  /// Thread-safety: Only updated by the query state thread.
+  int64_t last_report_seq_no_ = 0;
+
   /// Protects 'overall_status_' and 'failed_finstance_id_'.
   SpinLock status_lock_;
 
@@ -350,7 +385,7 @@ class QueryState {
 
   /// Temporary files for this query (owned by obj_pool_). Non-null if spilling is
   /// enabled. Set in Prepare().
-  TmpFileMgr::FileGroup* file_group_ = nullptr;
+  TmpFileGroup* file_group_ = nullptr;
 
   /// Manages runtime filters that are either produced or consumed (or both!) by plan
   /// nodes on this backend.
@@ -370,18 +405,33 @@ class QueryState {
   /// and so is 'failed_instance_id_' if an error is hit.
   std::unique_ptr<CountingBarrier> instances_finished_barrier_;
 
-  /// map from instance id to its state (owned by obj_pool_), populated in
+  /// Map from instance id to its state (owned by obj_pool_), populated in
   /// StartFInstances(); Not valid to read from until 'instances_prepared_barrier_'
   /// is set (i.e. readers should always call WaitForPrepare()).
   std::unordered_map<TUniqueId, FragmentInstanceState*> fis_map_;
 
+  /// Map from fragment index to its fragment state (owned by obj_pool_), populated in
+  /// StartFInstances();
+  std::unordered_map<TFragmentIdx, FragmentState*> fragment_state_map_;
+
   ObjectPool obj_pool_;
   AtomicInt32 refcnt_;
 
+  /// Protects 'is_initialized_'.
+  std::mutex init_lock_;
+
+  /// Set as true on successful initialization.
+  /// Protected by 'init_lock_'.
+  bool is_initialized_ = false;
+
   /// set to 1 when any fragment instance fails or when Cancel() is called; used to
   /// initiate cancellation exactly once
   AtomicInt32 is_cancelled_;
 
+  /// set to false when the coordinator has been detected as inactive in the cluster;
+  /// used to avoid sending the last execution report to the inactive/failed coordinator.
+  AtomicBool is_coord_active_{true};
+
   /// True if and only if ReleaseExecResources() has been called.
   bool released_backend_resources_ = false;
 
@@ -428,6 +478,11 @@ class QueryState {
   /// thread-safe.
   void ReleaseBackendResources();
 
+  /// Functions for calculating user or system time spent on async codegen threads.
+  int64_t AsyncCodegenThreadHelper(const std::string& suffix) const;
+  int64_t AsyncCodegenThreadUserTime() const;
+  int64_t AsyncCodegenThreadSysTime() const;
+
   /// Helper for ReportExecStatus() to construct a status report to be sent to the
   /// coordinator. The execution statuses (e.g. 'done' indicator) of all fragment
   /// instances belonging to this query state are stored in 'report'. The Thrift
