@@ -27,7 +27,6 @@
 #include "kudu/rpc/rpc_context.h"
 #include "kudu/rpc/rpc_controller.h"
 #include "kudu/rpc/rpc_sidecar.h"
-#include "runtime/backend-client.h"
 #include "runtime/bufferpool/reservation-tracker.h"
 #include "runtime/client-cache.h"
 #include "runtime/exec-env.h"
@@ -40,6 +39,7 @@
 #include "service/impala-server.h"
 #include "util/bit-util.h"
 #include "util/bloom-filter.h"
+#include "util/debug-util.h"
 #include "util/min-max-filter.h"
 #include "util/pretty-printer.h"
 #include "util/uid-util.h"
@@ -52,8 +52,9 @@ using kudu::rpc::RpcSidecar;
 using namespace impala;
 using namespace strings;
 
-DEFINE_double(max_filter_error_rate, 0.75, "(Advanced) The maximum probability of false "
-    "positives in a runtime filter before it is disabled.");
+DEFINE_double(max_filter_error_rate, 0.75, "(Advanced) The target false positive "
+    "probability used to determine the ideal size for each bloom filter size. This value "
+    "can be overriden by the RUNTIME_FILTER_ERROR_RATE query option.");
 
 const int64_t RuntimeFilterBank::MIN_BLOOM_FILTER_SIZE;
 const int64_t RuntimeFilterBank::MAX_BLOOM_FILTER_SIZE;
@@ -269,18 +270,17 @@ void RuntimeFilterBank::UpdateFilterFromLocal(
       DCHECK_EQ(type, TRuntimeFilterType::MIN_MAX);
       min_max_filter->ToProtobuf(params.mutable_min_max_filter());
     }
-    const TNetworkAddress& krpc_address = query_state_->query_ctx().coord_krpc_address;
-    const TNetworkAddress& host_address = query_state_->query_ctx().coord_address;
+    const TNetworkAddress& krpc_address = query_state_->query_ctx().coord_ip_address;
+    const std::string& hostname = query_state_->query_ctx().coord_hostname;
 
     // Use 'proxy' to send the filter to the coordinator.
     unique_ptr<DataStreamServiceProxy> proxy;
-    Status get_proxy_status =
-        DataStreamService::GetProxy(krpc_address, host_address.hostname, &proxy);
+    Status get_proxy_status = DataStreamService::GetProxy(krpc_address, hostname, &proxy);
     if (!get_proxy_status.ok()) {
       // Failing to send a filter is not a query-wide error - the remote fragment will
       // continue regardless.
-      LOG(INFO) << Substitute("Failed to get proxy to coordinator $0: $1",
-          host_address.hostname, get_proxy_status.msg().msg());
+      LOG(INFO) << Substitute("Failed to get proxy to coordinator $0: $1", hostname,
+          get_proxy_status.msg().msg());
       return;
     }
     // Increment 'num_inflight_rpcs_' to make sure that the filter will not be deallocated
@@ -341,8 +341,8 @@ void RuntimeFilterBank::PublishGlobalFilter(
       }
 
       if (bloom_filter != BloomFilter::ALWAYS_TRUE_FILTER) {
-        Status status = bloom_filter->Init(
-            params.bloom_filter(), sidecar_slice.data(), sidecar_slice.size());
+        Status status = bloom_filter->Init(params.bloom_filter(), sidecar_slice.data(),
+            sidecar_slice.size(), DefaultHashSeed());
         if (!status.ok()) {
           LOG(ERROR) << "Unable to allocate memory for bloom filter: "
                      << status.GetDetail();
@@ -382,7 +382,7 @@ BloomFilter* RuntimeFilterBank::AllocateScratchBloomFilter(int32_t filter_id) {
       << "BufferPool Client should have enough reservation to fulfill bloom filter "
          "allocation";
   BloomFilter* bloom_filter = obj_pool_.Add(new BloomFilter(&buffer_pool_client_));
-  Status status = bloom_filter->Init(log_filter_size);
+  Status status = bloom_filter->Init(log_filter_size, DefaultHashSeed());
   if (!status.ok()) {
     LOG(ERROR) << "Unable to allocate memory for bloom filter: " << status.GetDetail();
     return nullptr;
@@ -407,12 +407,6 @@ MinMaxFilter* RuntimeFilterBank::AllocateScratchMinMaxFilter(
   return min_max_filter;
 }
 
-bool RuntimeFilterBank::FpRateTooHigh(int64_t filter_size, int64_t observed_ndv) {
-  double fpp =
-      BloomFilter::FalsePositiveProb(observed_ndv, BitUtil::Log2Ceiling64(filter_size));
-  return fpp > FLAGS_max_filter_error_rate;
-}
-
 vector<unique_lock<SpinLock>> RuntimeFilterBank::LockAllFilters() {
   vector<unique_lock<SpinLock>> locks;
   for (auto& entry : filters_) locks.emplace_back(entry.second->lock);
@@ -425,7 +419,9 @@ void RuntimeFilterBank::Cancel() {
 }
 
 void RuntimeFilterBank::CancelLocked() {
-  if (cancelled_) return;
+  // IMPALA-9730: if no filters are present, we did not acquire any filter locks. Avoid a
+  // TSAN data race on 'cancelled_' by exiting early.
+  if (filters_.empty() || cancelled_) return;
   // Cancel all filters that a thread might be waiting on.
   for (auto& entry : filters_) {
     if (entry.second->consumed_filter != nullptr) entry.second->consumed_filter->Cancel();
