@@ -19,10 +19,12 @@
 #ifndef IMPALA_RUNTIME_RUNTIME_STATE_H
 #define IMPALA_RUNTIME_RUNTIME_STATE_H
 
-#include <boost/scoped_ptr.hpp>
+#include <mutex>
+#include <string>
 #include <utility>
 #include <vector>
-#include <string>
+
+#include <boost/scoped_ptr.hpp>
 
 // NOTE: try not to add more headers here: runtime-state.h is included in many many files.
 #include "common/global-types.h"  // for PlanNodeId
@@ -52,10 +54,11 @@ class ThreadResourcePool;
 class TUniqueId;
 class ExecEnv;
 class HBaseTableFactory;
-class TPlanFragmentCtx;
+class TPlanFragment;
 class TPlanFragmentInstanceCtx;
 class QueryState;
 class ConditionVariable;
+class CyclicBarrier;
 
 namespace io {
   class DiskIoMgr;
@@ -78,8 +81,10 @@ class RuntimeState {
  public:
   /// query_state, fragment_ctx, and instance_ctx need to be alive at least as long as
   /// the constructed RuntimeState
-  RuntimeState(QueryState* query_state, const TPlanFragmentCtx& fragment_ctx,
-      const TPlanFragmentInstanceCtx& instance_ctx, ExecEnv* exec_env);
+  RuntimeState(QueryState* query_state, const TPlanFragment& fragment,
+      const TPlanFragmentInstanceCtx& instance_ctx,
+      const PlanFragmentCtxPB& fragment_ctx,
+      const PlanFragmentInstanceCtxPB& instance_ctx_pb, ExecEnv* exec_env);
 
   /// RuntimeState for test execution and fe-support.cc. Creates its own QueryState and
   /// installs desc_tbl, if set. If query_ctx.request_pool isn't set, sets it to "test-pool".
@@ -97,10 +102,13 @@ class RuntimeState {
   int batch_size() const { return query_options().batch_size; }
   bool abort_on_error() const { return query_options().abort_on_error; }
   bool strict_mode() const { return query_options().strict_mode; }
+  bool utf8_mode() const { return query_options().utf8_mode; }
   bool decimal_v2() const { return query_options().decimal_v2; }
   const TQueryCtx& query_ctx() const;
-  const TPlanFragmentCtx& fragment_ctx() const { return *fragment_ctx_; }
+  const TPlanFragment& fragment() const { return *fragment_; }
   const TPlanFragmentInstanceCtx& instance_ctx() const { return *instance_ctx_; }
+  const PlanFragmentCtxPB& fragment_ctx() const { return *fragment_ctx_; }
+  const PlanFragmentInstanceCtxPB& instance_ctx_pb() const { return *instance_ctx_pb_; }
   const TUniqueId& session_id() const { return query_ctx().session.session_id; }
   const std::string& do_as_user() const { return query_ctx().session.delegated_user; }
   const std::string& connected_user() const {
@@ -113,6 +121,9 @@ class RuntimeState {
   const Timezone* time_zone_for_unix_time_conversions() const {
     return time_zone_for_unix_time_conversions_;
   }
+  const Timezone* time_zone_for_legacy_parquet_time_conversions() const {
+    return time_zone_for_legacy_parquet_time_conversions_;
+  }
   const TUniqueId& query_id() const { return query_ctx().query_id; }
   const TUniqueId& fragment_instance_id() const {
     return instance_ctx_ != nullptr
@@ -142,67 +153,10 @@ class RuntimeState {
   /// Returns runtime state profile
   RuntimeProfile* runtime_profile() { return profile_; }
 
-  /// Returns the LlvmCodeGen object for this fragment instance.
-  LlvmCodeGen* codegen() { return codegen_.get(); }
-
   const std::string& GetEffectiveUser() const;
 
-  /// Add ScalarExpr expression 'expr' to be codegen'd later if it's not disabled by
-  /// query option. If 'is_codegen_entry_point' is true, 'expr' will be an entry
-  /// point into codegen'd evaluation (i.e. it will have a function pointer populated).
-  /// Adding an expr here ensures that it will be codegen'd (i.e. fragment execution
-  /// will fail with an error if the expr cannot be codegen'd).
-  void AddScalarExprToCodegen(ScalarExpr* expr, bool is_codegen_entry_point) {
-    scalar_exprs_to_codegen_.push_back({expr, is_codegen_entry_point});
-  }
-
-  /// Returns true if there are ScalarExpr expressions in the fragments that we want
-  /// to codegen (because they can't be interpreted or based on options/hints).
-  /// This should only be used after the Prepare() phase in which all expressions'
-  /// Prepare() are invoked.
-  bool ScalarExprNeedsCodegen() const { return !scalar_exprs_to_codegen_.empty(); }
-
-  /// Check if codegen was disabled and if so, add a message to the runtime profile.
-  void CheckAndAddCodegenDisabledMessage(RuntimeProfile* profile) {
-    if (CodegenDisabledByQueryOption()) {
-      profile->AddCodegenMsg(false, "disabled by query option DISABLE_CODEGEN");
-    } else if (CodegenDisabledByHint()) {
-      profile->AddCodegenMsg(false, "disabled due to optimization hints");
-    }
-  }
-
-  /// Returns true if there is a hint to disable codegen. This can be true for single node
-  /// optimization or expression evaluation request from FE to BE (see fe-support.cc).
-  /// Note that this internal flag is advisory and it may be ignored if the fragment has
-  /// any UDF which cannot be interpreted. See ScalarExpr::Prepare() for details.
-  inline bool CodegenHasDisableHint() const {
-    return query_ctx().disable_codegen_hint;
-  }
-
-  /// Returns true iff there is a hint to disable codegen and all expressions in the
-  /// fragment can be interpreted. This should only be used after the Prepare() phase
-  /// in which all expressions' Prepare() are invoked.
-  inline bool CodegenDisabledByHint() const {
-    return CodegenHasDisableHint() && !ScalarExprNeedsCodegen();
-  }
-
-  /// Returns true if codegen is disabled by query option.
-  inline bool CodegenDisabledByQueryOption() const {
-    return query_options().disable_codegen;
-  }
-
-  /// Returns true if codegen should be enabled for this fragment. Codegen is enabled
-  /// if all the following conditions hold:
-  /// 1. it's enabled by query option
-  /// 2. it's not disabled by internal hints or there are expressions in the fragment
-  ///    which cannot be interpreted.
-  inline bool ShouldCodegen() const {
-    return !CodegenDisabledByQueryOption() && !CodegenDisabledByHint();
-  }
-
   inline Status GetQueryStatus() {
-    // Do a racy check for query_status_ to avoid unnecessary spinlock acquisition.
-    if (UNLIKELY(!query_status_.ok())) {
+    if (UNLIKELY(!is_query_status_ok_.Load())) {
       std::lock_guard<SpinLock> l(query_status_lock_);
       return query_status_;
     }
@@ -242,12 +196,27 @@ class RuntimeState {
   Status LogOrReturnError(const ErrorMsg& message);
 
   bool is_cancelled() const { return is_cancelled_.Load(); }
+
+  /// Cancel this runtime state, signalling all condition variables and cancelling all
+  /// barriers added in AddCancellationCV() and AddBarrierToCancel(). This function will
+  /// acquire mutexes added in AddCancellationCV(), so the caller must not hold any locks
+  /// that must acquire after those mutexes in the lock order.
   void Cancel();
+
   /// Add a condition variable to be signalled when this RuntimeState is cancelled.
   /// Adding a condition variable multiple times is a no-op. Each distinct 'cv' will be
-  /// signalled once with NotifyAll() when is_cancelled() becomes true.
+  /// signalled once with NotifyAll() when is_cancelled() becomes true. 'mutex' will
+  /// be acquired by the cancelling thread after is_cancelled() becomes true. The caller
+  /// must hold 'mutex' when checking is_cancelled() to avoid a race like IMPALA-9611
+  /// where the notification on 'cv' is lost.
   /// The condition variable must have query lifetime.
-  void AddCancellationCV(ConditionVariable* cv);
+  void AddCancellationCV(std::mutex* mutex, ConditionVariable* cv);
+
+  /// Add a barrier to be cancelled when this RuntimeState is cancelled. Adding a barrier
+  /// multiple times is a no-op. Each distinct 'cb' will be cancelled with status code
+  /// CANCELLED_INTERNALLY when is_cancelled() becomes true. 'cb' must have query
+  /// lifetime.
+  void AddBarrierToCancel(CyclicBarrier* cb);
 
   RuntimeProfile::Counter* total_storage_wait_timer() {
     return total_storage_wait_timer_;
@@ -282,6 +251,8 @@ class RuntimeState {
     std::lock_guard<SpinLock> l(query_status_lock_);
     if (!query_status_.ok()) return;
     query_status_ = Status(err_msg);
+    bool set_query_status_ok_ = is_query_status_ok_.CompareAndSwap(true, false);
+    DCHECK(set_query_status_ok_);
   }
 
   /// Sets query_status_ to MEM_LIMIT_EXCEEDED and logs all the registered trackers.
@@ -302,15 +273,6 @@ class RuntimeState {
   /// called after ReleaseResources().
   Status CheckQueryState();
 
-  /// Create a codegen object accessible via codegen() if it doesn't exist already.
-  Status CreateCodegen();
-
-  /// Codegen all ScalarExpr expressions in 'scalar_exprs_to_codegen_'. If codegen fails
-  /// for any expressions, return immediately with the error status. Once IMPALA-4233 is
-  /// fixed, it's not fatal to fail codegen if the expression can be interpreted.
-  /// TODO: Fix IMPALA-4233
-  Status CodegenScalarExprs();
-
   /// Helper to call QueryState::StartSpilling().
   Status StartSpilling(MemTracker* mem_tracker);
 
@@ -322,7 +284,7 @@ class RuntimeState {
   /// the posix error code of the failed RPC. The target node address and posix error code
   /// will be included in the AuxErrorInfo returned by GetAuxErrorInfo. This method is
   /// idempotent.
-  void SetRPCErrorInfo(TNetworkAddress dest_node, int16_t posix_error_code);
+  void SetRPCErrorInfo(NetworkAddressPB dest_node, int16_t posix_error_code);
 
   /// Returns true if this RuntimeState has any auxiliary error information, false
   /// otherwise. Currently, only SetRPCErrorInfo() sets aux error info.
@@ -355,8 +317,10 @@ class RuntimeState {
 
   /// Global QueryState and original thrift descriptors for this fragment instance.
   QueryState* const query_state_;
-  const TPlanFragmentCtx* const fragment_ctx_;
+  const TPlanFragment* const fragment_;
   const TPlanFragmentInstanceCtx* const instance_ctx_;
+  const PlanFragmentCtxPB* const fragment_ctx_;
+  const PlanFragmentInstanceCtxPB* const instance_ctx_pb_;
 
   /// only populated by the (const QueryCtx&, ExecEnv*, DescriptorTbl*) c'tor
   boost::scoped_ptr<QueryState> local_query_state_;
@@ -380,11 +344,10 @@ class RuntimeState {
   /// instead.
   const Timezone* time_zone_for_unix_time_conversions_;
 
-  boost::scoped_ptr<LlvmCodeGen> codegen_;
-
-  /// Contains all ScalarExpr expressions which need to be codegen'd. The second element
-  /// is true if we want to generate a codegen entry point for this expr.
-  std::vector<std::pair<ScalarExpr*, bool>> scalar_exprs_to_codegen_;
+  /// Query-global timezone used to read INT96 timestamps written by Hive. UTC by default,
+  /// but if convert_legacy_hive_parquet_utc_timestamps=1, then the local_time_zone_ is
+  /// used instead.
+  const Timezone* time_zone_for_legacy_parquet_time_conversions_;
 
   /// Thread resource management object for this fragment's execution.  The runtime
   /// state is responsible for returning this pool to the thread mgr.
@@ -435,7 +398,14 @@ class RuntimeState {
 
   /// Condition variables that will be signalled by Cancel(). Protected by
   /// 'cancellation_cvs_lock_'.
-  std::vector<ConditionVariable*> cancellation_cvs_;
+  std::vector<std::pair<std::mutex*, ConditionVariable*>> cancellation_cvs_;
+
+  /// Cyclic barriers that will be signalled by Cancel(). Protected by
+  /// 'cancellation_cvs_lock_'.
+  std::vector<CyclicBarrier*> cancellation_cbs_;
+
+  /// Condition variables that will be signalled by Cancel(). Protected by
+  /// 'cancellation_cvs_lock_'.
   SpinLock cancellation_cvs_lock_;
 
   /// if true, ReleaseResources() was called.
@@ -447,6 +417,11 @@ class RuntimeState {
   SpinLock query_status_lock_;
   Status query_status_;
 
+  /// True if the query_status_ is OK, false otherwise. Used to check if the
+  /// query_status_ is OK without incurring the overhead of acquiring the
+  /// query_status_lock_.
+  AtomicBool is_query_status_ok_{true};
+
   /// This is the node id of the root node for this plan fragment.
   ///
   /// This is used as the hash seed within the fragment so we do not run into hash
