@@ -18,12 +18,14 @@
 #include "runtime/sorter-internal.h"
 
 #include <boost/bind.hpp>
-#include <boost/random/mersenne_twister.hpp>
-#include <boost/random/uniform_int.hpp>
 #include <gutil/strings/substitute.h>
 
+#include "codegen/llvm-codegen.h"
+#include "exprs/scalar-expr-evaluator.h"
 #include "runtime/bufferpool/reservation-tracker.h"
+#include "runtime/bufferpool/reservation-util.h"
 #include "runtime/exec-env.h"
+#include "runtime/fragment-state.h"
 #include "runtime/mem-tracker.h"
 #include "runtime/query-state.h"
 #include "runtime/runtime-state.h"
@@ -33,8 +35,6 @@
 
 #include "common/names.h"
 
-using boost::uniform_int;
-using boost::mt19937_64;
 using namespace strings;
 
 namespace impala {
@@ -753,7 +753,13 @@ Status Sorter::TupleSorter::Sort(Run* run) {
   DCHECK(run->is_finalized());
   DCHECK(!run->is_sorted());
   run_ = run;
-  RETURN_IF_ERROR(SortHelper(TupleIterator::Begin(run_), TupleIterator::End(run_)));
+  const SortHelperFn sort_helper_fn = parent_->codegend_sort_helper_fn_.load();
+  if (sort_helper_fn != nullptr) {
+    RETURN_IF_ERROR(
+        sort_helper_fn(this, TupleIterator::Begin(run_), TupleIterator::End(run_)));
+  } else {
+    RETURN_IF_ERROR(SortHelper(TupleIterator::Begin(run_), TupleIterator::End(run_)));
+  }
   run_->set_sorted();
   return Status::OK();
 }
@@ -762,13 +768,16 @@ Sorter::Sorter(const TupleRowComparatorConfig& tuple_row_comparator_config,
     const vector<ScalarExpr*>& sort_tuple_exprs, RowDescriptor* output_row_desc,
     MemTracker* mem_tracker, BufferPool::ClientHandle* buffer_pool_client,
     int64_t page_len, RuntimeProfile* profile, RuntimeState* state,
-    const string& node_label, bool enable_spilling)
+    const string& node_label, bool enable_spilling,
+    const CodegenFnPtr<SortHelperFn>& codegend_sort_helper_fn,
+    int64_t estimated_input_size)
   : node_label_(node_label),
     state_(state),
     expr_perm_pool_(mem_tracker),
     expr_results_pool_(mem_tracker),
     compare_less_than_(nullptr),
     in_mem_tuple_sorter_(nullptr),
+    codegend_sort_helper_fn_(codegend_sort_helper_fn),
     buffer_pool_client_(buffer_pool_client),
     page_len_(page_len),
     has_var_len_slots_(false),
@@ -795,6 +804,8 @@ Sorter::Sorter(const TupleRowComparatorConfig& tuple_row_comparator_config,
     default:
       DCHECK(false);
   }
+
+  if (estimated_input_size > 0) ComputeSpillEstimate(estimated_input_size);
 }
 
 Sorter::~Sorter() {
@@ -804,6 +815,21 @@ Sorter::~Sorter() {
   DCHECK(merge_output_run_ == nullptr);
 }
 
+void Sorter::ComputeSpillEstimate(int64_t estimated_input_size) {
+  int64_t max_reservation = state_->query_state()->GetMaxReservation();
+  if (estimated_input_size > max_reservation) {
+    CheckSortRunBytesLimitEnforcement();
+    if (enforce_sort_run_bytes_limit_) {
+      VLOG(3) << Substitute(
+          "Enforcing sort_run_bytes_limit because spill is estimated to happen: "
+          "max_reservation=$0 estimated_input_size=$1 sort_run_bytes_limit=$2",
+          PrettyPrinter::PrintBytes(max_reservation),
+          PrettyPrinter::PrintBytes(estimated_input_size),
+          PrettyPrinter::PrintBytes(GetSortRunBytesLimit()));
+    }
+  }
+}
+
 Status Sorter::Prepare(ObjectPool* obj_pool) {
   DCHECK(in_mem_tuple_sorter_ == nullptr) << "Already prepared";
   // Page byte offsets are packed into uint32_t values, which limits the supported
@@ -872,21 +898,74 @@ Status Sorter::AddBatch(RowBatch* batch) {
     RETURN_IF_ERROR(AddBatchNoSpill(batch, cur_batch_index, &num_processed));
 
     cur_batch_index += num_processed;
-    if (cur_batch_index < batch->num_rows()) {
+    if (MustSortAndSpill(cur_batch_index, batch->num_rows())) {
       // The current run is full. Sort it, spill it and begin the next one.
+      int64_t unsorted_run_bytes = unsorted_run_->TotalBytes();
       RETURN_IF_ERROR(state_->StartSpilling(mem_tracker_));
       RETURN_IF_ERROR(SortCurrentInputRun());
       RETURN_IF_ERROR(sorted_runs_.back()->UnpinAllPages());
       unsorted_run_ =
           run_pool_.Add(new Run(this, output_row_desc_->tuple_descriptors()[0], true));
       RETURN_IF_ERROR(unsorted_run_->Init());
+
+      bool prev_enforcement = enforce_sort_run_bytes_limit_;
+      CheckSortRunBytesLimitEnforcement();
+      if (!prev_enforcement && enforce_sort_run_bytes_limit_) {
+        VLOG(3) << Substitute(
+            "Enforcing sort_run_bytes_limit because reservation limit exceeded: "
+            "prev_run_total_bytes=$0 sort_run_bytes_limit=$1",
+            PrettyPrinter::PrintBytes(unsorted_run_bytes),
+            PrettyPrinter::PrintBytes(GetSortRunBytesLimit()));
+      }
     }
   }
+  if (enforce_sort_run_bytes_limit_) TryLowerMemUpToSortRunBytesLimit();
   // Clear any temporary allocations made while materializing the sort tuples.
   expr_results_pool_.Clear();
   return Status::OK();
 }
 
+void Sorter::CheckSortRunBytesLimitEnforcement() {
+  enforce_sort_run_bytes_limit_ = GetSortRunBytesLimit() > 0
+      && state_->query_options().scratch_limit != 0
+      && !state_->query_options().disable_unsafe_spills;
+}
+
+int64_t Sorter::GetSortRunBytesLimit() const {
+  if (state_->query_options().sort_run_bytes_limit > 0
+      && state_->query_options().sort_run_bytes_limit < MIN_SORT_RUN_BYTES_LIMIT) {
+    return MIN_SORT_RUN_BYTES_LIMIT;
+  } else {
+    return state_->query_options().sort_run_bytes_limit;
+  }
+}
+
+bool Sorter::MustSortAndSpill(const int rows_added, const int batch_num_rows) {
+  if (rows_added < batch_num_rows) {
+    return true;
+  } else if (enforce_sort_run_bytes_limit_) {
+    int used_reservation = buffer_pool_client_->GetUsedReservation();
+    return GetSortRunBytesLimit() < used_reservation;
+  }
+  return false;
+}
+
+void Sorter::TryLowerMemUpToSortRunBytesLimit() {
+  DCHECK(enforce_sort_run_bytes_limit_);
+  int unused_reservation = buffer_pool_client_->GetUnusedReservation();
+  int current_reservation = buffer_pool_client_->GetReservation();
+  int threshold = GetSortRunBytesLimit() + page_len_; // tolerate +1 page
+
+  if (current_reservation > threshold) {
+    Status status = buffer_pool_client_->DecreaseReservationTo(
+        unused_reservation, GetSortRunBytesLimit());
+    DCHECK(status.ok());
+    VLOG(3) << Substitute("Sorter reservation decreased from $0 to $1",
+        PrettyPrinter::PrintBytes(current_reservation),
+        PrettyPrinter::PrintBytes(buffer_pool_client_->GetReservation()));
+  }
+}
+
 Status Sorter::AddBatchNoSpill(RowBatch* batch, int start_index, int* num_processed) {
   DCHECK(batch != nullptr);
   RETURN_IF_ERROR(unsorted_run_->AddInputBatch(batch, start_index, num_processed));
@@ -1124,4 +1203,37 @@ bool Sorter::HasSpilledRuns() const {
   return !merging_runs_.empty() || sorted_runs_.size() > 1 ||
       (sorted_runs_.size() == 1 && !sorted_runs_.back()->is_pinned());
 }
+
+const char* Sorter::TupleSorter::SORTER_HELPER_SYMBOL =
+    "_ZN6impala6Sorter11TupleSorter10SortHelperENS0_13TupleIteratorES2_";
+
+const char* Sorter::TupleSorter::LLVM_CLASS_NAME = "class.impala::Sorter::TupleSorter";
+
+// A method to code-gen for TupleSorter::SortHelper().
+Status Sorter::TupleSorter::Codegen(FragmentState* state, llvm::Function* compare_fn,
+    CodegenFnPtr<SortHelperFn>* codegened_fn) {
+  LlvmCodeGen* codegen = state->codegen();
+  DCHECK(codegen != nullptr);
+
+  llvm::Function* fn = codegen->GetFunction(IRFunction::TUPLE_SORTER_SORT_HELPER, true);
+  DCHECK(fn != NULL);
+
+  // There are 6 calls to Less() which calls comparator_.Less() once in each.
+  int replaced =
+      codegen->ReplaceCallSites(fn, compare_fn, TupleRowComparator::COMPARE_SYMBOL);
+  DCHECK_REPLACE_COUNT(replaced, 6) << LlvmCodeGen::Print(fn);
+
+  // There are 2 recursive calls within SorterHelper() to replace with.
+  replaced = codegen->ReplaceCallSites(fn, fn, SORTER_HELPER_SYMBOL);
+  DCHECK_REPLACE_COUNT(replaced, 2) << LlvmCodeGen::Print(fn);
+
+  fn = codegen->FinalizeFunction(fn);
+  if (fn == nullptr) {
+    return Status("Sorter::TupleSorter::Codegen(): failed to finalize function");
+  }
+  codegen->AddFunctionToJit(fn, codegened_fn);
+
+  return Status::OK();
+}
+
 } // namespace impala
