@@ -15,18 +15,23 @@
 // specific language governing permissions and limitations
 // under the License.
 
+#pragma once
 
-#ifndef IMPALA_RUNTIME_TIMESTAMP_VALUE_H
-#define IMPALA_RUNTIME_TIMESTAMP_VALUE_H
-
-#include <boost/date_time/compiler_config.hpp>
-#include <boost/date_time/gregorian/gregorian.hpp>
-#include <boost/date_time/local_time/local_time.hpp>
+#include <cstdint>
+#include <cstring>
+#include <iosfwd>
 #include <string>
 
+#include <boost/date_time/gregorian/greg_date.hpp>
+#include <boost/date_time/posix_time/posix_time_config.hpp>
+#include <boost/date_time/posix_time/ptime.hpp>
+#include <boost/date_time/special_defs.hpp>
+#include <boost/date_time/time_duration.hpp>
+
+#include "common/compiler-util.h"
 #include "common/global-types.h"
-#include "gen-cpp/Data_types.h"
-#include "gen-cpp/data_stream_service.pb.h"
+#include "common/logging.h"
+#include "gen-cpp/common.pb.h"
 #include "udf/udf.h"
 #include "util/hash-util.h"
 
@@ -147,7 +152,7 @@ class TimestampValue {
   /// Returns a TimestampVal representation in the output variable.
   /// Returns null if the TimestampValue instance doesn't have a valid date or time.
   void ToTimestampVal(impala_udf::TimestampVal* tv) const {
-    if (!HasDateOrTime()) {
+    if (!HasDate()) {
       tv->is_null = true;
       return;
     }
@@ -176,7 +181,6 @@ class TimestampValue {
 
   bool HasDate() const { return !date_.is_special(); }
   bool HasTime() const { return !time_.is_special(); }
-  bool HasDateOrTime() const { return HasDate() || HasTime(); }
   bool HasDateAndTime() const { return HasDate() && HasTime(); }
 
   std::string ToString() const;
@@ -196,7 +200,6 @@ class TimestampValue {
 
   /// Verifies that the time is not negative and is less than a whole day.
   static inline bool IsValidTime(const boost::posix_time::time_duration& time) {
-    static const int64_t NANOS_PER_DAY = 1'000'000'000LL * SECONDS_PER_DAY;
     return !time.is_negative()
         && time.total_nanoseconds() < NANOS_PER_DAY;
   }
@@ -276,6 +279,14 @@ class TimestampValue {
   const boost::gregorian::date& date() const { return date_; }
   const boost::posix_time::time_duration& time() const { return time_; }
 
+  /// Return this added with a time duration when t is valid and HasDateAndTime() is
+  /// true on this. Return an empty TimestampValue object othewise.
+  TimestampValue Add(const boost::posix_time::time_duration& t) const;
+
+  /// Return this subtracted from a time duration when t is valid and HasDateAndTime()
+  /// is true on this. Return an empty TimestampValue object othewise.
+  TimestampValue Subtract(const boost::posix_time::time_duration& t) const;
+
   TimestampValue& operator=(const boost::posix_time::ptime& ptime) {
     date_ = ptime.date();
     time_ = ptime.time_of_day();
@@ -337,6 +348,8 @@ class TimestampValue {
 
   static const uint64_t SECONDS_PER_DAY = 24 * 60 * 60;
 
+  static const int64_t NANOS_PER_DAY = 1'000'000'000LL * SECONDS_PER_DAY;
+
   /// Boost ptime leaves a gap in the structure, so we swap the order to make it
   /// 12 contiguous bytes.  We then must convert to and from the boost ptime data type.
   /// See IMP-87 for more information on why using ptime with the 4 byte gap is
@@ -379,5 +392,3 @@ inline std::size_t hash_value(const TimestampValue& v) {
 
 std::ostream& operator<<(std::ostream& os, const TimestampValue& timestamp_value);
 }
-
-#endif
