@@ -21,8 +21,6 @@
 #include "common/object-pool.h"
 #include "common/status.h"
 #include "exec/catalog-op-executor.h"
-#include "runtime/timestamp-value.h"
-#include "scheduling/query-schedule.h"
 #include "service/child-query.h"
 #include "service/impala-server.h"
 #include "service/query-result-set.h"
@@ -36,9 +34,9 @@
 
 namespace impala {
 
+class AdmissionControlClient;
 class ClientRequestStateCleaner;
 class Coordinator;
-class ExecEnv;
 class Expr;
 class Frontend;
 class ReportExecStatusRequestPB;
@@ -47,7 +45,7 @@ class RuntimeState;
 class Thread;
 class TRuntimeProfileTree;
 class TupleRow;
-enum class AdmissionOutcome;
+class QuerySchedulePB;
 
 /// Execution state of the client-facing side of a query. This captures everything
 /// necessary to convert row batches received by the coordinator into results
@@ -66,25 +64,36 @@ enum class AdmissionOutcome;
 /// the TStmtType). Successful QUERY / DML queries transition from INITIALIZED to PENDING
 /// to RUNNING to FINISHED whereas DDL queries skip the PENDING phase.
 ///
+/// Retry State:
+/// The retry state is only used when the query corresponding to this ClientRequestState
+/// is retried. Otherwise, the default state is NOT_RETRIED. MarkAsRetrying and
+/// MarkAsRetried set the state to RETRYING and RETRIED, respectively. Queries can only
+/// transition from NOT_RETRIED to RETRYING and then finally to RETRIED.
+///
 /// TODO: Compute stats is the only stmt that requires child queries. Once the
 /// CatalogService performs background stats gathering the concept of child queries
 /// will likely become obsolete. Remove all child-query related code from this class.
 class ClientRequestState {
  public:
-  ClientRequestState(const TQueryCtx& query_ctx, ExecEnv* exec_env, Frontend* frontend,
-      ImpalaServer* server, std::shared_ptr<ImpalaServer::SessionState> session);
+  ClientRequestState(const TQueryCtx& query_ctx, Frontend* frontend, ImpalaServer* server,
+      std::shared_ptr<ImpalaServer::SessionState> session, TExecRequest* exec_request,
+      QueryDriver* query_driver);
 
   ~ClientRequestState();
 
-  enum class ExecState {
-    INITIALIZED, PENDING, RUNNING, FINISHED, ERROR
-  };
+  enum class ExecState { INITIALIZED, PENDING, RUNNING, FINISHED, ERROR };
+
+  enum class RetryState { RETRYING, RETRIED, NOT_RETRIED };
 
   /// Sets the profile that is produced by the frontend. The frontend creates the
   /// profile during planning and returns it to the backend via TExecRequest,
   /// which then sets the frontend profile.
   void SetFrontendProfile(TRuntimeProfileNode profile);
 
+  /// Sets the coordinator time that the plan request was submitted at so that
+  /// the backend timeline starts where the frontend timeline ends
+  void SetRemoteSubmitTime(int64_t remote_submit_time);
+
   /// Based on query type, this either initiates execution of this ClientRequestState's
   /// TExecRequest or submits the query to the Admission controller for asynchronous
   /// admission control. When this returns the operation state is either RUNNING_STATE or
@@ -163,18 +172,26 @@ class ClientRequestState {
   Status UpdateQueryStatus(const Status& status) WARN_UNUSED_RESULT;
 
   /// Cancels the child queries and the coordinator with the given cause.
-  /// If cause is NULL, it assume this was deliberately cancelled by the user while in
+  /// If cause is NULL, it assumes this was deliberately cancelled by the user while in
   /// FINISHED state. Otherwise, sets state to ERROR (TODO: IMPALA-1262: use CANCELLED).
-  /// Does nothing if the query has reached EOS or already cancelled.
+  /// Does nothing if the query has reached EOS or already cancelled. If
+  /// 'wait_until_finalized' is true and another thread is cancelling 'coord_', block
+  /// until cancellaton of 'coord_' finishes and it is finalized.
+  /// 'wait_until_finalized' should only used by the single thread finalizing the query,
+  /// to avoid many threads piling up waiting for query cancellation.
   ///
   /// Only returns an error if 'check_inflight' is true and the query is not yet
   /// in-flight. Otherwise, proceed and return Status::OK() even if the query isn't
   /// in-flight (for cleaning up after an error on the query issuing path).
-  Status Cancel(bool check_inflight, const Status* cause) WARN_UNUSED_RESULT;
+  Status Cancel(bool check_inflight, const Status* cause, bool wait_until_finalized=false);
 
-  /// This is called when the query is done (finished, cancelled, or failed).
+  /// This is called when the query is done (finished, cancelled, or failed). This runs
+  /// synchronously within the last client RPC and does any work that is required before
+  /// the query is finished from the client's point of view, including cancelling the
+  /// query with 'cause'. Returns an error if 'check_inflight' is true and the query is
+  /// not yet in-flight, or if another thread has already started finalizing the query.
   /// Takes lock_: callers must not hold lock() before calling.
-  void Done();
+  Status Finalize(bool check_inflight, const Status* cause);
 
   /// Sets the API-specific (Beeswax, HS2) result cache and its size bound.
   /// The given cache is owned by this client request state, even if an error is returned.
@@ -197,14 +214,19 @@ class ClientRequestState {
   /// Caller must not hold 'lock()'.
   bool GetDmlStats(TDmlResult* dml_result, Status* query_status);
 
-  /// Creates and sets the TExecRequest for the query associated with this
-  /// ClientRequestState. The TExecRequest is created by the Impala frontend via the
-  /// method Frontend::GetExecRequest(TQueryCtx, TExecRequest). The TQueryCtx is created
-  /// by the ImpalaServer and contains the full query string
-  /// (TQueryCtx::TClientRequest::stmt).
-  Status InitExecRequest(const TQueryCtx& query_ctx);
+  /// Blocks until this query has been retried. Waits until the ExecState has transitioned
+  /// to RETRIED (e.g. once MarkAsRetried() has been called). Can only be called if the
+  /// current state is either RETRIED or RETRYING. Takes lock_.
+  void WaitUntilRetried();
 
-  ImpalaServer::SessionState* session() const { return session_.get(); }
+  /// Converts the given ExecState to a string representation.
+  static std::string ExecStateToString(ExecState state);
+
+  /// Converts the given RetryState to a string representation.
+  static std::string RetryStateToString(RetryState state);
+
+  /// Returns the session for this query.
+  std::shared_ptr<ImpalaServer::SessionState> session() const { return session_; }
 
   /// Queries are run and authorized on behalf of the effective_user.
   const std::string& effective_user() const;
@@ -215,7 +237,7 @@ class ClientRequestState {
   const TUniqueId& session_id() const { return query_ctx_.session.session_id; }
   const std::string& default_db() const { return query_ctx_.session.database; }
   bool eos() const { return eos_.Load(); }
-  const QuerySchedule* schedule() const { return schedule_.get(); }
+  const QuerySchedulePB* schedule() const { return schedule_.get(); }
 
   /// Returns the Coordinator for 'QUERY' and 'DML' requests once Coordinator::Exec()
   /// completes successfully. Otherwise returns null.
@@ -241,23 +263,32 @@ class ClientRequestState {
   /// Contents are only valid after InitExecRequest(TQueryCtx) initializes the
   /// TExecRequest.
   const TExecRequest& exec_request() const {
-    return exec_request_;
+    DCHECK(exec_request_ != nullptr);
+    return *exec_request_;
   }
-  TStmtType::type stmt_type() const { return exec_request_.stmt_type; }
+  TStmtType::type stmt_type() const { return exec_request_->stmt_type; }
   TCatalogOpType::type catalog_op_type() const {
-    return exec_request_.catalog_op_request.op_type;
+    return exec_request_->catalog_op_request.op_type;
   }
   TDdlType::type ddl_type() const {
-    return exec_request_.catalog_op_request.ddl_params.ddl_type;
+    return exec_request_->catalog_op_request.ddl_params.ddl_type;
   }
   std::mutex* lock() { return &lock_; }
   std::mutex* fetch_rows_lock() { return &fetch_rows_lock_; }
+
   /// ExecState is stored using an AtomicEnum, so reads do not require holding lock_.
   ExecState exec_state() const { return exec_state_.Load(); }
-  /// Translate exec_state_ to a TOperationState.
+
+  /// RetryState is stored using an AtomicEnum, so reads do not require holding lock_.
+  RetryState retry_state() const { return retry_state_.Load(); }
+
+  /// Translate exec_state_ to a TOperationState. Returns the current TOperationState.
   apache::hive::service::cli::thrift::TOperationState::type TOperationState() const;
-  /// Translate exec_state_ to a beeswax::QueryState.
+
+  /// Translate exec_state_ to a beeswax::QueryState. Returns the current
+  /// beeswax::QueryState.
   beeswax::QueryState::type BeeswaxQueryState() const;
+
   const Status& query_status() const { return query_status_; }
   void set_result_metadata(const TResultSetMetadata& md) { result_metadata_ = md; }
   void set_user_profile_access(bool user_has_profile_access) {
@@ -275,7 +306,7 @@ class ClientRequestState {
   TUniqueId parent_query_id() const { return query_ctx_.parent_query_id; }
 
   const std::vector<std::string>& GetAnalysisWarnings() const {
-    return exec_request_.analysis_warnings;
+    return exec_request_->analysis_warnings;
   }
 
   inline int64_t last_active_ms() const {
@@ -310,7 +341,91 @@ class ClientRequestState {
   /// Returns the FETCH_ROWS_TIMEOUT_MS value for this query (converted to microseconds).
   int64_t fetch_rows_timeout_us() const { return fetch_rows_timeout_us_; }
 
-protected:
+  /// Returns the max size of the result_cache_ in number of rows.
+  int64_t result_cache_max_size() const { return result_cache_max_size_; }
+
+  /// Sets the RetryState to RETRYING. Updates the runtime profile with the retry status
+  /// and cause. Must be called while 'lock_' is held. Sets the query_status_. Future
+  /// calls to UpdateQueryStatus will not have any effect. This is necessary to prevent
+  /// any future calls to UpdateQueryStatus from updating the ExecState to ERROR. The
+  /// ExecState should not be set to ERROR when a query is being retried in order to
+  /// prevent any error statuses from being exposed to the client.
+  void MarkAsRetrying(const Status& status);
+
+  /// Sets the RetryState to RETRIED and wakes up any threads waiting for the query to be
+  /// RETRIED. 'retried_id' is the query id of the newly retried query (e.g. not the id
+  /// of the "original" query that was submitted by the user, but the id of the new query
+  /// that was created because the "original" query failed and had to be retried).
+  void MarkAsRetried(const TUniqueId& retried_id);
+
+  /// Returns true if this ClientRequestState was created as a retry of a previously
+  /// failed query, false otherwise. This is different from WasRetried() which tracks
+  /// if this ClientRequestState was retried (retries are done in a new
+  /// ClientRequestState).
+  bool IsRetriedQuery() const { return original_id_ != nullptr; }
+
+  /// Only called if this is a "retried" query - e.g. it was created as a result of
+  /// retrying a failed query. It sets the 'original_id_' field, which is the query id of
+  /// the original query attempt that failed. The original query id is added to the
+  /// runtime profile as well.
+  void SetOriginalId(const TUniqueId& original_id);
+
+  /// Returns true if this ClientRequestState has already been retried, e.g. the
+  /// RetryState is either RETRYING or RETRIED.
+  bool WasRetried() const {
+    RetryState retry_state = retry_state_.Load();
+    return retry_state == RetryState::RETRYING || retry_state == RetryState::RETRIED;
+  }
+
+  /// Can only be called if this query is the result of retrying a previously failed
+  /// query. Returns the query id of the original query.
+  const TUniqueId& original_id() const {
+    DCHECK(original_id_ != nullptr);
+    return *original_id_;
+  }
+
+  /// Can only be called if this query was retried. Returns the query id of the retried
+  /// query.
+  const TUniqueId& retried_id() const {
+    DCHECK(retried_id_ != nullptr);
+    return *retried_id_;
+  }
+
+  bool IsSetRetriedId() const {
+    return retried_id_ != nullptr;
+  }
+
+  /// Returns the QueryDriver that owns this ClientRequestState.
+  QueryDriver* parent_driver() const { return parent_driver_; }
+
+  AdmissionControlClient* admission_control_client() const {
+    return admission_control_client_.get();
+  }
+
+  /// Returns true if results cacheing is enabled, false otherwise.
+  bool IsResultCacheingEnabled() const { return result_cache_max_size_ >= 0; }
+
+  /// Add an executor address to the set of blacklisted executor addresses. The scheduler
+  /// will not schedule any query fragment instance on the given executor.
+  /// Note: The given address is added to the set of blacklisted addresses only if the
+  /// request has not been retried, e.g retry_state_ is not equal RETRYING nor RETRIED.
+  void AddBlacklistedExecutorAddress(const NetworkAddressPB& addr);
+
+  /// Return blacklisted executor addresses for the query
+  /// This function should be called after the request is retried, no more executor
+  /// address will be added to the set of blacklisted addresses so there is no data race.
+  std::unordered_set<NetworkAddressPB>& GetBlacklistedExecutorAddresses() {
+    return blacklisted_executor_addresses_;
+  }
+
+  /// Set a set of blacklisted executor addresses. The scheduler will not schedule any
+  /// query fragment instance on the given set of executors.
+  /// This function is called before the retried request is executed so don't need
+  /// mutex protection.
+  void SetBlacklistedExecutorAddresses(
+      std::unordered_set<NetworkAddressPB>& executor_addresses);
+
+ protected:
   /// Updates the end_time_us_ of this query if it isn't set. The end time is determined
   /// when this function is called for the first time, calling it multiple times does not
   /// change the end time.
@@ -351,12 +466,6 @@ protected:
   /// Executor for any child queries (e.g. compute stats subqueries). Always non-NULL.
   const boost::scoped_ptr<ChildQueryExecutor> child_query_executor_;
 
-  /// Promise used by the admission controller. AdmissionController:AdmitQuery() will
-  /// block on this promise until the query is either rejected, admitted, times out, or is
-  /// cancelled. Can be set to CANCELLED by the ClientRequestState in order to cancel, but
-  /// otherwise is set by AdmissionController with the admission decision.
-  Promise<AdmissionOutcome, PromiseMode::MULTIPLE_PRODUCER> admit_outcome_;
-
   /// Protects all following fields. Acquirers should be careful not to hold it for too
   /// long, e.g. during RPCs because this lock is required to make progress on various
   /// ImpalaServer requests. If held for too long it can block progress of client
@@ -365,9 +474,6 @@ protected:
   /// See "Locking" in the class comment for lock acquisition order.
   std::mutex lock_;
 
-  /// TODO: remove and use ExecEnv::GetInstance() instead
-  ExecEnv* exec_env_;
-
   /// Thread for asynchronously running Wait().
   std::unique_ptr<Thread> wait_thread_;
 
@@ -381,8 +487,8 @@ protected:
   /// Session that this query is from
   std::shared_ptr<ImpalaServer::SessionState> session_;
 
-  /// Resource assignment determined by scheduler. Owned by obj_pool_.
-  std::unique_ptr<QuerySchedule> schedule_;
+  /// Resource assignment determined by scheduler.
+  std::unique_ptr<QuerySchedulePB> schedule_;
 
   /// Thread for asynchronously running the admission control code-path and starting
   /// execution in the following cases:
@@ -402,6 +508,10 @@ protected:
   /// the async-exec-thread only after Exec() has been successfully called on 'coord_'.
   AtomicBool coord_exec_called_;
 
+  /// Set of executor addresses, which allow caller to tell the scheduler to not schedule
+  /// the query on this set of executors.
+  std::unordered_set<NetworkAddressPB> blacklisted_executor_addresses_;
+
   /// Runs statements that query or modify the catalog via the CatalogService.
   boost::scoped_ptr<CatalogOpExecutor> catalog_op_executor_;
 
@@ -474,14 +584,17 @@ protected:
   /// UpdateExecState(), to ensure that the query profile is also updated.
   AtomicEnum<ExecState> exec_state_{ExecState::INITIALIZED};
 
+  /// The current RetryState of the query.
+  AtomicEnum<RetryState> retry_state_{RetryState::NOT_RETRIED};
+
   /// The current status of the query tracked by this ClientRequestState. Updated by
-  /// UpdateQueryStatus(Status).
+  /// UpdateQueryStatus(Status) or MarkAsRetrying(Status).
   Status query_status_;
 
   /// The TExecRequest for the query tracked by this ClientRequestState. The TExecRequest
-  /// is initialized in InitExecRequest(TQueryCtx). It should not be used until
-  /// InitExecRequest(TQueryCtx) has been called.
-  TExecRequest exec_request_;
+  /// is initialized in QueryDriver::RunFrontendPlanner(TQueryCtx).The TExecRequest is
+  /// owned by the parent QueryDriver.
+  TExecRequest* exec_request_;
 
   /// If true, effective_user() has access to the runtime profile and execution
   /// summary.
@@ -516,14 +629,37 @@ protected:
   /// Start/end time of the query, in Unix microseconds.
   int64_t start_time_us_;
   /// end_time_us_ is initialized to 0, which is used to indicate that the query is not
-  /// yet done. It is assinged the final value in ClientRequestState::Done() or when the
-  /// coordinator relases its admission control resources.
+  /// yet done. It is assigned the final value in ClientRequestState::Finalize() or when
+  /// the coordinator releases its admission control resources.
   AtomicInt64 end_time_us_{0};
 
   /// Timeout, in microseconds, when waiting for rows to become available. Derived from
   /// the query option FETCH_ROWS_TIMEOUT_MS.
   const int64_t fetch_rows_timeout_us_;
 
+  /// If this ClientRequestState was created as a retry of a previously failed query, the
+  /// original_id_ is set to the query id of the original query that failed. The
+  /// "original" query is the query that was submitted by the user that failed and had to
+  /// be retried.
+  std::unique_ptr<const TUniqueId> original_id_ = nullptr;
+
+  /// Query id of the retried query. The retried query is the new query that is run
+  /// whenever the original query fails with a retryable error. See 'original_id_' for
+  /// an explanation of what the "original" query is.
+  std::unique_ptr<const TUniqueId> retried_id_ = nullptr;
+
+  /// Condition variable used to signal any threads that are waiting until the query has
+  /// been retried.
+  ConditionVariable block_until_retried_cv_;
+
+  /// The QueryDriver that owns this ClientRequestState. The reference is set in the
+  /// constructor. It always outlives the ClientRequestState.
+  QueryDriver* parent_driver_;
+
+  /// Manages interactions with the AdmissionController, which may be either local or
+  /// remote.
+  std::unique_ptr<AdmissionControlClient> admission_control_client_;
+
   /// Executes a local catalog operation (an operation that does not need to execute
   /// against the catalog service). Includes USE, SHOW, DESCRIBE, and EXPLAIN statements.
   Status ExecLocalCatalogOp(const TCatalogOpRequest& catalog_op) WARN_UNUSED_RESULT;
@@ -543,7 +679,7 @@ protected:
   Status ExecAsyncQueryOrDmlRequest(const TQueryExecRequest& query_exec_request)
       WARN_UNUSED_RESULT;
 
-  /// Submits the QuerySchedule to the admission controller and on successful admission,
+  /// Submits the exec request to the admission controller and on successful admission,
   /// starts up the coordinator execution, makes it accessible by setting
   /// 'coord_exec_called_' to true and advances operation_state_ to RUNNING. Handles
   /// async cancellation of queries and cleans up state if needed.
@@ -636,9 +772,6 @@ protected:
   /// Logs audit and column lineage events. Expects that Wait() has already finished.
   /// Grabs lock_ for polling the query_status(). Hence do not call it under lock_.
   void LogQueryEvents();
-
-  /// Converts the given ExecState to a string representation.
-  std::string ExecStateToString(ExecState state) const;
 };
 
 }
