@@ -21,6 +21,7 @@
 #include <mutex>
 #include <sstream>
 #include <boost/lexical_cast.hpp>
+#include <boost/unordered_set.hpp>
 #include <rapidjson/prettywriter.h>
 #include <rapidjson/stringbuffer.h>
 
@@ -31,16 +32,20 @@
 #include "runtime/coordinator.h"
 #include "runtime/exec-env.h"
 #include "runtime/mem-tracker.h"
+#include "runtime/query-driver.h"
 #include "runtime/query-state.h"
 #include "runtime/timestamp-value.h"
 #include "runtime/timestamp-value.inline.h"
+#include "scheduling/admission-control-service.h"
 #include "scheduling/admission-controller.h"
+#include "scheduling/admissiond-env.h"
 #include "scheduling/cluster-membership-mgr.h"
 #include "service/client-request-state.h"
 #include "service/frontend.h"
 #include "service/impala-server.h"
 #include "thrift/protocol/TDebugProtocol.h"
 #include "util/coding-util.h"
+#include "util/debug-util.h"
 #include "util/logging-support.h"
 #include "util/pretty-printer.h"
 #include "util/redactor.h"
@@ -59,7 +64,9 @@ using namespace rapidjson;
 using namespace strings;
 
 DECLARE_int32(query_log_size);
+DECLARE_int32(query_stmt_size);
 DECLARE_bool(use_local_catalog);
+DECLARE_string(admission_service_host);
 
 namespace {
 
@@ -87,15 +94,39 @@ static Status ParseIdFromRequest(const Webserver::WebRequest& req, TUniqueId* id
 
 }
 
-void ImpalaHttpHandler::RegisterHandlers(Webserver* webserver) {
+ImpalaHttpHandler::ImpalaHttpHandler(ImpalaServer* server,
+    AdmissionController* admission_controller,
+    ClusterMembershipMgr* cluster_membership_mgr, bool is_admissiond)
+  : server_(server),
+    admission_controller_(admission_controller),
+    cluster_membership_mgr_(cluster_membership_mgr),
+    is_admissiond_(is_admissiond) {}
+
+void ImpalaHttpHandler::RegisterHandlers(Webserver* webserver, bool metrics_only) {
   DCHECK(webserver != NULL);
 
+  if (is_admissiond_) {
+    // The admissiond only exposes a subset of endpoints that have info relevant to
+    // admission control.
+    webserver->RegisterUrlCallback("/backends", "backends.tmpl",
+        MakeCallback(this, &ImpalaHttpHandler::BackendsHandler), true);
+
+    webserver->RegisterUrlCallback("/admission", "admission_controller.tmpl",
+        MakeCallback(this, &ImpalaHttpHandler::AdmissionStateHandler), true);
+
+    webserver->RegisterUrlCallback("/resource_pool_reset", "",
+        MakeCallback(this, &ImpalaHttpHandler::ResetResourcePoolStatsHandler), false);
+    return;
+  }
+
   Webserver::RawUrlCallback healthz_callback =
     [this](const auto& req, auto* data, auto* response) {
       return this->HealthzHandler(req, data, response);
     };
   webserver->RegisterUrlCallback("/healthz", healthz_callback);
 
+  if (metrics_only) return;
+
   webserver->RegisterUrlCallback("/backends", "backends.tmpl",
       MakeCallback(this, &ImpalaHttpHandler::BackendsHandler), true);
 
@@ -164,11 +195,15 @@ void ImpalaHttpHandler::RegisterHandlers(Webserver* webserver) {
       [this](const auto& req, auto* doc) {
         this->QuerySummaryHandler(false, false, req, doc); }, false);
 
+  // Only enable the admission control endpoints for impalads if the admission service is
+  // not enabled, otherwise these will be exposed on the admissiond.
+  if (!ExecEnv::GetInstance()->AdmissionServiceEnabled()) {
     webserver->RegisterUrlCallback("/admission", "admission_controller.tmpl",
-      MakeCallback(this, &ImpalaHttpHandler::AdmissionStateHandler), true);
+        MakeCallback(this, &ImpalaHttpHandler::AdmissionStateHandler), true);
 
     webserver->RegisterUrlCallback("/resource_pool_reset", "",
-      MakeCallback(this, &ImpalaHttpHandler::ResetResourcePoolStatsHandler), false);
+        MakeCallback(this, &ImpalaHttpHandler::ResetResourcePoolStatsHandler), false);
+  }
 
   RegisterLogLevelCallbacks(webserver, true);
 }
@@ -262,9 +297,11 @@ void ImpalaHttpHandler::QueryProfileHandler(const Webserver::WebRequest& req,
     return;
   }
 
+  ImpalaServer::RuntimeProfileOutput runtime_profile;
   stringstream ss;
+  runtime_profile.string_output = &ss;
   Status status = server_->GetRuntimeProfileOutput(
-      unique_id, "", TRuntimeProfileFormat::STRING, &ss, nullptr, nullptr);
+      unique_id, "", TRuntimeProfileFormat::STRING, &runtime_profile);
   if (!status.ok()) {
     Value error(status.GetDetail().c_str(), document->GetAllocator());
     document->AddMember("error", error, document->GetAllocator());
@@ -286,8 +323,11 @@ void ImpalaHttpHandler::QueryProfileHelper(const Webserver::WebRequest& req,
   if (!status.ok()) {
     ss << status.GetDetail();
   } else {
-    Status status = server_->GetRuntimeProfileOutput(
-      unique_id, "", format, &ss, nullptr, document);
+    ImpalaServer::RuntimeProfileOutput runtime_profile;
+    runtime_profile.string_output = &ss;
+    runtime_profile.json_output = document;
+    Status status =
+        server_->GetRuntimeProfileOutput(unique_id, "", format, &runtime_profile);
     if (!status.ok()) {
       ss.str(Substitute("Could not obtain runtime profile: $0", status.GetDetail()));
     }
@@ -323,9 +363,9 @@ void ImpalaHttpHandler::QueryProfileJsonHandler(const Webserver::WebRequest& req
 void ImpalaHttpHandler::InflightQueryIdsHandler(const Webserver::WebRequest& req,
     Document* document) {
   stringstream ss;
-  server_->client_request_state_map_.DoFuncForAllEntries(
-      [&](const std::shared_ptr<ClientRequestState>& request_state) {
-          ss << PrintId(request_state->query_id()) << "\n";
+  server_->query_driver_map_.DoFuncForAllEntries(
+      [&](const std::shared_ptr<QueryDriver>& query_driver) {
+        ss << PrintId(query_driver->GetActiveClientRequestState()->query_id()) << "\n";
       });
   document->AddMember(rapidjson::StringRef(Webserver::ENABLE_RAW_HTML_KEY), true,
       document->GetAllocator());
@@ -368,7 +408,11 @@ void ImpalaHttpHandler::QueryStateToJson(const ImpalaServer::QueryStateRecord& r
   value->AddMember("default_db", default_db, document->GetAllocator());
 
   // Redact the query string
-  Value stmt(RedactCopy(record.stmt).c_str(), document->GetAllocator());
+  std::string tmp_stmt = RedactCopy(record.stmt);
+  if(FLAGS_query_stmt_size && tmp_stmt.length() > FLAGS_query_stmt_size) {
+    tmp_stmt = tmp_stmt.substr(0, FLAGS_query_stmt_size).append("...");
+  }
+  Value stmt(tmp_stmt.c_str(), document->GetAllocator());
   value->AddMember("stmt", stmt, document->GetAllocator());
 
   Value stmt_type(_TStmtType_VALUES_TO_NAMES.find(record.stmt_type)->second,
@@ -409,8 +453,7 @@ void ImpalaHttpHandler::QueryStateToJson(const ImpalaServer::QueryStateRecord& r
   Value progress_json(progress.c_str(), document->GetAllocator());
   value->AddMember("progress", progress_json, document->GetAllocator());
 
-  Value state(_QueryState_VALUES_TO_NAMES.find(record.query_state)->second,
-      document->GetAllocator());
+  Value state(record.query_state.c_str(), document->GetAllocator());
   value->AddMember("state", state, document->GetAllocator());
 
   value->AddMember("rows_fetched", record.num_rows_fetched, document->GetAllocator());
@@ -425,7 +468,7 @@ void ImpalaHttpHandler::QueryStateToJson(const ImpalaServer::QueryStateRecord& r
   }
 
   // Waiting to be closed.
-  bool waiting = record.query_state == beeswax::QueryState::EXCEPTION ||
+  bool waiting = record.beeswax_query_state == beeswax::QueryState::EXCEPTION ||
       record.all_rows_returned;
   value->AddMember("waiting", waiting, document->GetAllocator());
   value->AddMember("executing", !waiting, document->GetAllocator());
@@ -447,11 +490,13 @@ void ImpalaHttpHandler::QueryStateHandler(const Webserver::WebRequest& req,
   set<ImpalaServer::QueryStateRecord, ImpalaServer::QueryStateRecordLessThan>
       sorted_query_records;
 
-  server_->client_request_state_map_.DoFuncForAllEntries(
-      [&](const std::shared_ptr<ClientRequestState>& request_state) {
-          sorted_query_records.insert(ImpalaServer::QueryStateRecord(*request_state));
+  server_->query_driver_map_.DoFuncForAllEntries(
+      [&](const std::shared_ptr<QueryDriver>& query_driver) {
+        sorted_query_records.insert(
+            ImpalaServer::QueryStateRecord(*query_driver->GetActiveClientRequestState()));
       });
 
+  unordered_set<TUniqueId> in_flight_query_ids;
   Value in_flight_queries(kArrayType);
   int64_t num_waiting_queries = 0;
   for (const ImpalaServer::QueryStateRecord& record: sorted_query_records) {
@@ -461,6 +506,7 @@ void ImpalaHttpHandler::QueryStateHandler(const Webserver::WebRequest& req,
     if (record_json["waiting"].GetBool()) ++num_waiting_queries;
 
     in_flight_queries.PushBack(record_json, document->GetAllocator());
+    in_flight_query_ids.insert(record.id);
   }
   document->AddMember("in_flight_queries", in_flight_queries, document->GetAllocator());
   document->AddMember("num_in_flight_queries",
@@ -482,6 +528,8 @@ void ImpalaHttpHandler::QueryStateHandler(const Webserver::WebRequest& req,
     lock_guard<mutex> l(server_->query_log_lock_);
     for (const unique_ptr<ImpalaServer::QueryStateRecord>& log_entry :
         server_->query_log_) {
+      // Don't show duplicated entries between in-flight and completed queries.
+      if (in_flight_query_ids.find(log_entry->id) != in_flight_query_ids.end()) continue;
       Value record_json(kObjectType);
       QueryStateToJson(*log_entry, &record_json, document);
       completed_queries.PushBack(record_json, document->GetAllocator());
@@ -497,7 +545,7 @@ void ImpalaHttpHandler::QueryStateHandler(const Webserver::WebRequest& req,
     for (const ImpalaServer::QueryLocations::value_type& location :
         server_->query_locations_) {
       Value location_json(kObjectType);
-      Value location_name(TNetworkAddressToString(location.second.address).c_str(),
+      Value location_name(NetworkAddressPBToString(location.second.address).c_str(),
           document->GetAllocator());
       location_json.AddMember("location", location_name, document->GetAllocator());
       Value backend_id_str(PrintId(location.first).c_str(), document->GetAllocator());
@@ -794,12 +842,14 @@ void ImpalaHttpHandler::QueryBackendsHandler(
     return;
   }
 
-  shared_ptr<ClientRequestState> request_state = server_->GetClientRequestState(query_id);
-  if (request_state.get() == nullptr || request_state->GetCoordinator() == nullptr) {
-    return;
+  QueryHandle query_handle;
+  status = server_->GetQueryHandle(query_id, &query_handle);
+  if (status.ok()) {
+    if (query_handle->GetCoordinator() == nullptr) {
+      return;
+    }
+    query_handle->GetCoordinator()->BackendsToJson(document);
   }
-
-  request_state->GetCoordinator()->BackendsToJson(document);
 }
 
 void ImpalaHttpHandler::QueryFInstancesHandler(
@@ -815,12 +865,14 @@ void ImpalaHttpHandler::QueryFInstancesHandler(
     return;
   }
 
-  shared_ptr<ClientRequestState> request_state = server_->GetClientRequestState(query_id);
-  if (request_state.get() == nullptr || request_state->GetCoordinator() == nullptr) {
-    return;
+  QueryHandle query_handle;
+  status = server_->GetQueryHandle(query_id, &query_handle);
+  if (status.ok()) {
+    if (query_handle->GetCoordinator() == nullptr) {
+      return;
+    }
+    query_handle->GetCoordinator()->FInstanceStatsToJson(document);
   }
-
-  request_state->GetCoordinator()->FInstanceStatsToJson(document);
 }
 
 void ImpalaHttpHandler::QuerySummaryHandler(bool include_json_plan, bool include_summary,
@@ -845,30 +897,30 @@ void ImpalaHttpHandler::QuerySummaryHandler(bool include_json_plan, bool include
 
   // Search the in-flight queries first, followed by the archived ones.
   {
-    shared_ptr<ClientRequestState> request_state =
-        server_->GetClientRequestState(query_id);
-    if (request_state != nullptr) {
+    QueryHandle query_handle;
+    status = server_->GetQueryHandle(query_id, &query_handle);
+    if (status.ok()) {
       found = true;
       // If the query plan isn't generated, avoid waiting for the request
       // state lock to be acquired, since it could potentially be an expensive
       // call, if the table Catalog metadata loading is in progress. Instead
       // update the caller that the plan information is unavailable.
-      if (request_state->exec_state() == ClientRequestState::ExecState::INITIALIZED) {
+      if (query_handle->exec_state() == ClientRequestState::ExecState::INITIALIZED) {
         document->AddMember(
             "plan_metadata_unavailable", "true", document->GetAllocator());
         return;
       }
-      lock_guard<mutex> l(*request_state->lock());
-      query_status = request_state->query_status();
-      stmt = request_state->sql_stmt();
-      plan = request_state->exec_request().query_exec_request.query_plan;
+      lock_guard<mutex> l(*(*query_handle).lock());
+      query_status = query_handle->query_status();
+      stmt = query_handle->sql_stmt();
+      plan = query_handle->exec_request().query_exec_request.query_plan;
       if ((include_json_plan || include_summary)
-          && request_state->GetCoordinator() != nullptr) {
-        request_state->GetCoordinator()->GetTExecSummary(&summary);
+          && query_handle->GetCoordinator() != nullptr) {
+        query_handle->GetCoordinator()->GetTExecSummary(&summary);
       }
       if (include_json_plan) {
         for (const TPlanExecInfo& plan_exec_info:
-            request_state->exec_request().query_exec_request.plan_exec_info) {
+            query_handle->exec_request().query_exec_request.plan_exec_info) {
           for (const TPlanFragment& fragment: plan_exec_info.fragments) {
             fragments.push_back(fragment);
           }
@@ -912,6 +964,8 @@ void ImpalaHttpHandler::QuerySummaryHandler(bool include_json_plan, bool include
   document->AddMember("stmt", json_stmt, document->GetAllocator());
   Value json_plan_text(RedactCopy(plan).c_str(), document->GetAllocator());
   document->AddMember("plan", json_plan_text, document->GetAllocator());
+  Value json_inflight(found);
+  document->AddMember("inflight", json_inflight, document->GetAllocator());
 
   // Redact the error in case the query is contained in the error message.
   Value json_status(query_status.ok() ? "OK" :
@@ -922,46 +976,45 @@ void ImpalaHttpHandler::QuerySummaryHandler(bool include_json_plan, bool include
 void ImpalaHttpHandler::BackendsHandler(const Webserver::WebRequest& req,
     Document* document) {
   AdmissionController::PerHostStats host_stats;
-  ExecEnv::GetInstance()->admission_controller()->PopulatePerHostMemReservedAndAdmitted(
-      &host_stats);
+  DCHECK(admission_controller_ != nullptr);
+  admission_controller_->PopulatePerHostMemReservedAndAdmitted(&host_stats);
   Value backends_list(kArrayType);
   int num_active_backends = 0;
   int num_quiescing_backends = 0;
   int num_blacklisted_backends = 0;
-  ClusterMembershipMgr* cluster_membership_mgr =
-      ExecEnv::GetInstance()->cluster_membership_mgr();
-  DCHECK(cluster_membership_mgr != nullptr);
+  DCHECK(cluster_membership_mgr_ != nullptr);
   ClusterMembershipMgr::SnapshotPtr membership_snapshot =
-      cluster_membership_mgr->GetSnapshot();
+      cluster_membership_mgr_->GetSnapshot();
   DCHECK(membership_snapshot.get() != nullptr);
   for (const auto& entry : membership_snapshot->current_backends) {
-    TBackendDescriptor backend = entry.second;
+    BackendDescriptorPB backend = entry.second;
     Value backend_obj(kObjectType);
-    string address = TNetworkAddressToString(backend.address);
+    string address = NetworkAddressPBToString(backend.address());
     Value str(address.c_str(), document->GetAllocator());
-    Value krpc_address(
-        TNetworkAddressToString(backend.krpc_address).c_str(), document->GetAllocator());
+    Value krpc_address(NetworkAddressPBToString(backend.krpc_address()).c_str(),
+        document->GetAllocator());
     backend_obj.AddMember("address", str, document->GetAllocator());
     backend_obj.AddMember("krpc_address", krpc_address, document->GetAllocator());
-    Value backend_id_str(PrintId(backend.backend_id).c_str(), document->GetAllocator());
+    Value backend_id_str(PrintId(backend.backend_id()).c_str(), document->GetAllocator());
     backend_obj.AddMember("backend_id", backend_id_str, document->GetAllocator());
     string webserver_url =
-        Substitute("$0://$1", backend.secure_webserver ? "https" : "http",
-            TNetworkAddressToString(backend.debug_http_address));
+        Substitute("$0://$1", backend.secure_webserver() ? "https" : "http",
+            NetworkAddressPBToString(backend.debug_http_address()));
     Value webserver_url_val(webserver_url.c_str(), document->GetAllocator());
     backend_obj.AddMember("webserver_url", webserver_url_val, document->GetAllocator());
-    backend_obj.AddMember("is_coordinator", backend.is_coordinator,
-        document->GetAllocator());
-    backend_obj.AddMember("is_executor", backend.is_executor, document->GetAllocator());
-    backend_obj.AddMember("is_quiescing", backend.is_quiescing, document->GetAllocator());
+    backend_obj.AddMember(
+        "is_coordinator", backend.is_coordinator(), document->GetAllocator());
+    backend_obj.AddMember("is_executor", backend.is_executor(), document->GetAllocator());
+    backend_obj.AddMember(
+        "is_quiescing", backend.is_quiescing(), document->GetAllocator());
     Status blacklist_cause;
     int64_t blacklist_time_remaining_ms;
     bool is_blacklisted = membership_snapshot->executor_blacklist.IsBlacklisted(
         backend, &blacklist_cause, &blacklist_time_remaining_ms);
     backend_obj.AddMember("is_blacklisted", is_blacklisted, document->GetAllocator());
-    backend_obj.AddMember(
-        "is_active", !is_blacklisted && !backend.is_quiescing, document->GetAllocator());
-    if (backend.is_quiescing) {
+    backend_obj.AddMember("is_active", !is_blacklisted && !backend.is_quiescing(),
+        document->GetAllocator());
+    if (backend.is_quiescing()) {
       // Backends cannot be both blacklisted and quiescing.
       DCHECK(!is_blacklisted);
       ++num_quiescing_backends;
@@ -979,7 +1032,7 @@ void ImpalaHttpHandler::BackendsHandler(const Webserver::WebRequest& req,
     } else {
       ++num_active_backends;
     }
-    Value admit_mem_limit(PrettyPrinter::PrintBytes(backend.admit_mem_limit).c_str(),
+    Value admit_mem_limit(PrettyPrinter::PrintBytes(backend.admit_mem_limit()).c_str(),
         document->GetAllocator());
     backend_obj.AddMember("admit_mem_limit", admit_mem_limit, document->GetAllocator());
     // If the host address does not exist in the 'host_stats', this would ensure that a
@@ -990,14 +1043,16 @@ void ImpalaHttpHandler::BackendsHandler(const Webserver::WebRequest& req,
     Value mem_admitted(PrettyPrinter::PrintBytes(
         host_stats[address].mem_admitted).c_str(), document->GetAllocator());
     backend_obj.AddMember("mem_admitted", mem_admitted, document->GetAllocator());
-    backend_obj.AddMember("admission_slots", backend.admission_slots,
-        document->GetAllocator());
+    backend_obj.AddMember(
+        "admission_slots", backend.admission_slots(), document->GetAllocator());
     backend_obj.AddMember("num_admitted", host_stats[address].num_admitted,
         document->GetAllocator());
     backend_obj.AddMember("admission_slots_in_use", host_stats[address].slots_in_use,
         document->GetAllocator());
     vector<string> group_names;
-    for (const auto& group : backend.executor_groups) group_names.push_back(group.name);
+    for (const auto& group : backend.executor_groups()) {
+      group_names.push_back(group.name());
+    }
     Value executor_groups(JoinStrings(group_names, ", ").c_str(),
         document->GetAllocator());
     backend_obj.AddMember("executor_groups", executor_groups, document->GetAllocator());
@@ -1021,14 +1076,13 @@ void ImpalaHttpHandler::BackendsHandler(const Webserver::WebRequest& req,
 void ImpalaHttpHandler::AdmissionStateHandler(
     const Webserver::WebRequest& req, Document* document) {
   const auto& args = req.parsed_args;
-  AdmissionController* ac = ExecEnv::GetInstance()->admission_controller();
   Webserver::ArgumentMap::const_iterator pool_name_arg = args.find("pool_name");
   bool get_all_pools = (pool_name_arg == args.end());
   Value resource_pools(kArrayType);
   if (get_all_pools) {
-    ac->AllPoolsToJson( &resource_pools, document);
+    admission_controller_->AllPoolsToJson(&resource_pools, document);
   } else {
-    ac->PoolToJson(pool_name_arg->second, &resource_pools, document);
+    admission_controller_->PoolToJson(pool_name_arg->second, &resource_pools, document);
   }
 
   // Now get running queries from CRS map.
@@ -1041,21 +1095,45 @@ void ImpalaHttpHandler::AdmissionStateHandler(
     unsigned long num_backends;
   };
   unordered_map<string, vector<QueryInfo>> running_queries;
-  server_->client_request_state_map_.DoFuncForAllEntries([&running_queries](
-      const std::shared_ptr<ClientRequestState>& request_state) {
-    // Make sure only queries past admission control are added.
-    auto query_state = request_state->exec_state();
-    if (query_state != ClientRequestState::ExecState::INITIALIZED
-        && query_state != ClientRequestState::ExecState::PENDING
-        && request_state->schedule() != nullptr)
-      running_queries[request_state->request_pool()].push_back(
-          {request_state->query_id(), request_state->schedule()->per_backend_mem_limit(),
-              request_state->schedule()->per_backend_mem_to_admit(),
-              request_state->schedule()->coord_backend_mem_limit(),
-              request_state->schedule()->coord_backend_mem_to_admit(),
-              static_cast<unsigned long>(
-                  request_state->schedule()->per_backend_exec_params().size())});
-  });
+  if (is_admissiond_) {
+    AdmissiondEnv::GetInstance()
+        ->admission_control_service()
+        ->admission_state_map_.DoFuncForAllEntries(
+            [&running_queries](
+                const std::shared_ptr<AdmissionControlService::AdmissionState>&
+                    query_info) {
+              lock_guard<mutex> l(query_info->lock);
+              if (query_info->schedule.get() != nullptr) {
+                TUniqueId query_id;
+                UniqueIdPBToTUniqueId(query_info->query_id, &query_id);
+                running_queries[query_info->request_pool].push_back(
+                    {query_id, query_info->schedule->per_backend_mem_limit(),
+                        query_info->schedule->per_backend_mem_to_admit(),
+                        query_info->schedule->coord_backend_mem_limit(),
+                        query_info->schedule->coord_backend_mem_to_admit(),
+                        static_cast<unsigned long>(
+                            query_info->schedule->backend_exec_params().size())});
+              };
+            });
+  } else {
+    server_->query_driver_map_.DoFuncForAllEntries(
+        [&running_queries](const std::shared_ptr<QueryDriver>& query_driver) {
+          // Make sure only queries past admission control are added.
+          ClientRequestState* request_state = query_driver->GetActiveClientRequestState();
+          auto query_state = request_state->exec_state();
+          if (query_state != ClientRequestState::ExecState::INITIALIZED
+              && query_state != ClientRequestState::ExecState::PENDING
+              && request_state->schedule() != nullptr)
+            running_queries[request_state->request_pool()].push_back(
+                {request_state->query_id(),
+                    request_state->schedule()->per_backend_mem_limit(),
+                    request_state->schedule()->per_backend_mem_to_admit(),
+                    request_state->schedule()->coord_backend_mem_limit(),
+                    request_state->schedule()->coord_backend_mem_to_admit(),
+                    static_cast<unsigned long>(
+                        request_state->schedule()->backend_exec_params().size())});
+        });
+  }
 
   // Add the running queries to the resource_pools json.
   for (int i = 0; i < resource_pools.Size(); i++) {
@@ -1088,12 +1166,15 @@ void ImpalaHttpHandler::AdmissionStateHandler(
         "running_queries", queries_in_pool, document->GetAllocator());
   }
   int64_t ms_since_last_statestore_update;
-  string staleness_detail = ac->GetStalenessDetail("", &ms_since_last_statestore_update);
+  string staleness_detail =
+      admission_controller_->GetStalenessDetail("", &ms_since_last_statestore_update);
 
   // In order to embed a plain json inside the webpage generated by mustache, we need
-  // to stringify it and write it out as a json element.
+  // to stringify it and write it out as a json element. We do not need to pretty-print
+  // it, so use the basic writer.
   rapidjson::StringBuffer strbuf;
-  PrettyWriter<rapidjson::StringBuffer> writer(strbuf);
+  Writer<rapidjson::StringBuffer> writer(strbuf);
+
   resource_pools.Accept(writer);
   Value raw_json(strbuf.GetString(), document->GetAllocator());
   document->AddMember("resource_pools_plain_json", raw_json, document->GetAllocator());
@@ -1116,9 +1197,8 @@ void ImpalaHttpHandler::ResetResourcePoolStatsHandler(
   Webserver::ArgumentMap::const_iterator pool_name_arg = args.find("pool_name");
   bool reset_all_pools = (pool_name_arg == args.end());
   if (reset_all_pools) {
-    ExecEnv::GetInstance()->admission_controller()->ResetAllPoolInformationalStats();
+    admission_controller_->ResetAllPoolInformationalStats();
   } else {
-    ExecEnv::GetInstance()->admission_controller()->ResetPoolInformationalStats(
-        pool_name_arg->second);
+    admission_controller_->ResetPoolInformationalStats(pool_name_arg->second);
   }
 }
