@@ -25,7 +25,6 @@
 #include "exprs/timezone_db.h"
 #include "gen-cpp/ImpalaInternalService_types.h"
 
-#include <zstd.h>
 #include <sstream>
 #include <boost/algorithm/string.hpp>
 #include <gutil/strings/strip.h>
@@ -186,35 +185,6 @@ static bool IsRemovedQueryOption(const string& key) {
   return false;
 }
 
-// Return all enum values in a string format, e.g. FOO(1), BAR(2), BAZ(3).
-static string GetThriftEnumValues(const map<int, const char*>& enum_values_to_names) {
-  bool first = true;
-  stringstream ss;
-  for (const auto& e : enum_values_to_names) {
-    if (!first) {
-      ss << ", ";
-    } else {
-      first = false;
-    }
-    ss << e.second << "(" << e.first << ")";
-  }
-  return ss.str();
-}
-
-// Return false for an invalid Thrift enum value.
-template<typename ENUM_TYPE>
-static Status GetThriftEnum(const string& value, const string& key,
-    const map<int, const char*>& enum_values_to_names, ENUM_TYPE* enum_value) {
-  for (const auto& e : enum_values_to_names) {
-    if (iequals(value, to_string(e.first)) || iequals(value, e.second)) {
-      *enum_value = static_cast<ENUM_TYPE>(e.first);
-      return Status::OK();
-    }
-  }
-  return Status(Substitute("Invalid $0: '$1'. Valid values are $2.", key, value,
-      GetThriftEnumValues(enum_values_to_names)));
-}
-
 // Return true if the given value is true (case-insensitive) or 1.
 static bool IsTrue(const string& value) {
   return iequals(value, "true") || iequals(value, "1");
@@ -279,36 +249,10 @@ Status impala::SetQueryOption(const string& key, const string& value,
         query_options->__set_debug_action(value.c_str());
         break;
       case TImpalaQueryOptions::COMPRESSION_CODEC: {
-        // Acceptable values are:
-        // - zstd:compression_level
-        // - codec
-        vector<string> tokens;
-        split(tokens, value, is_any_of(":"), token_compress_on);
-        if (tokens.size() > 2) return Status("Invalid compression codec value");
-
-        string& codec_name = tokens[0];
-        trim(codec_name);
-        int compression_level = ZSTD_CLEVEL_DEFAULT;
         THdfsCompression::type enum_type;
-        RETURN_IF_ERROR(GetThriftEnum(codec_name, "compression codec",
-            _THdfsCompression_VALUES_TO_NAMES, &enum_type));
-
-        if (tokens.size() == 2) {
-          if (enum_type != THdfsCompression::ZSTD) {
-            return Status("Compression level only supported for ZSTD");
-          }
-          StringParser::ParseResult status;
-          string& clevel = tokens[1];
-          trim(clevel);
-          compression_level = StringParser::StringToInt<int>(
-            clevel.c_str(), static_cast<int>(clevel.size()), &status);
-          if (status != StringParser::PARSE_SUCCESS || compression_level < 1
-              || compression_level > ZSTD_maxCLevel()) {
-            return Status(Substitute("Invalid ZSTD compression level '$0'."
-                " Valid values are in [1,$1]", clevel, ZSTD_maxCLevel()));
-          }
-        }
-
+        int compression_level;
+        RETURN_IF_ERROR(
+            ParseUtil::ParseCompressionCodec(value, &enum_type, &compression_level));
         TCompressionCodec compression_codec;
         compression_codec.__set_codec(enum_type);
         if (enum_type == THdfsCompression::ZSTD) {
@@ -385,6 +329,9 @@ Status impala::SetQueryOption(const string& key, const string& value,
       case TImpalaQueryOptions::OPTIMIZE_PARTITION_KEY_SCANS:
         query_options->__set_optimize_partition_key_scans(IsTrue(value));
         break;
+      case TImpalaQueryOptions::OPTIMIZE_SIMPLE_LIMIT:
+        query_options->__set_optimize_simple_limit(IsTrue(value));
+        break;
       case TImpalaQueryOptions::REPLICA_PREFERENCE: {
         map<int, const char *> valid_enums_values = {
             {0, "CACHE_LOCAL"},
@@ -457,6 +404,26 @@ Status impala::SetQueryOption(const string& key, const string& value,
       case TImpalaQueryOptions::DISABLE_ROW_RUNTIME_FILTERING:
         query_options->__set_disable_row_runtime_filtering(IsTrue(value));
         break;
+      case TImpalaQueryOptions::MINMAX_FILTERING_LEVEL:
+        // Parse the enabled runtime filter types and validate it.
+        TMinmaxFilteringLevel::type enum_type;
+        RETURN_IF_ERROR(GetThriftEnum(value, "minmax filter level",
+            _TMinmaxFilteringLevel_VALUES_TO_NAMES, &enum_type));
+        query_options->__set_minmax_filtering_level(enum_type);
+        break;
+      case TImpalaQueryOptions::MINMAX_FILTER_THRESHOLD: {
+        StringParser::ParseResult status;
+        double val = StringParser::StringToFloat<double>(
+            value.c_str(), static_cast<int>(value.size()), &status);
+        if (status != StringParser::PARSE_SUCCESS || val < 0.0 || val > 1.0) {
+          return Status(
+              Substitute("Invalid minmax filter threshold '$0'. Valid sizes are in"
+                         "[0.0, 1.0]",
+                  value));
+        }
+        query_options->__set_minmax_filter_threshold(val);
+        break;
+      }
       case TImpalaQueryOptions::MAX_NUM_RUNTIME_FILTERS: {
         StringParser::ParseResult status;
         int val = StringParser::StringToInt<int>(value.c_str(), value.size(), &status);
@@ -531,6 +498,10 @@ Status impala::SetQueryOption(const string& key, const string& value,
         query_options->__set_enable_expr_rewrites(IsTrue(value));
         break;
       }
+      case TImpalaQueryOptions::ENABLE_CNF_REWRITES: {
+        query_options->__set_enable_cnf_rewrites(IsTrue(value));
+        break;
+      }
       case TImpalaQueryOptions::DECIMAL_V2: {
         query_options->__set_decimal_v2(IsTrue(value));
         break;
@@ -630,6 +601,14 @@ Status impala::SetQueryOption(const string& key, const string& value,
         query_options->__set_compute_stats_min_sample_size(min_sample_size);
         break;
       }
+      case TImpalaQueryOptions::COMPUTE_COLUMN_MINMAX_STATS: {
+        query_options->__set_compute_column_minmax_stats(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::SHOW_COLUMN_MINMAX_STATS: {
+        query_options->__set_show_column_minmax_stats(IsTrue(value));
+        break;
+      }
       case TImpalaQueryOptions::EXEC_TIME_LIMIT_S: {
         StringParser::ParseResult result;
         const int32_t time_limit =
@@ -734,6 +713,7 @@ Status impala::SetQueryOption(const string& key, const string& value,
               value));
         }
         query_options->__set_resource_trace_ratio(val);
+        break;
       }
       case TImpalaQueryOptions::PLANNER_TESTCASE_MODE: {
         query_options->__set_planner_testcase_mode(IsTrue(value));
@@ -916,6 +896,186 @@ Status impala::SetQueryOption(const string& key, const string& value,
         query_options->__set_broadcast_bytes_limit(broadcast_bytes_limit);
         break;
       }
+      case TImpalaQueryOptions::RETRY_FAILED_QUERIES: {
+        query_options->__set_retry_failed_queries(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::PREAGG_BYTES_LIMIT: {
+        // Parse the preaggregation bytes limit and validate it
+        int64_t preagg_bytes_limit;
+        RETURN_IF_ERROR(
+            ParseMemValue(value, "preaggregation bytes limit", &preagg_bytes_limit));
+        query_options->__set_preagg_bytes_limit(preagg_bytes_limit);
+        break;
+      }
+      case TImpalaQueryOptions::MAX_CNF_EXPRS: {
+        StringParser::ParseResult result;
+        const int32_t requested_max_cnf_exprs =
+            StringParser::StringToInt<int32_t>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || requested_max_cnf_exprs < -1) {
+          return Status(
+              Substitute("Invalid max cnf exprs : '$0'. "
+                         "Only -1 and non-negative numbers are allowed.", value));
+        }
+        query_options->__set_max_cnf_exprs(requested_max_cnf_exprs);
+        break;
+      }
+      case TImpalaQueryOptions::KUDU_SNAPSHOT_READ_TIMESTAMP_MICROS: {
+        StringParser::ParseResult result;
+        const int64_t timestamp =
+            StringParser::StringToInt<int64_t>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || timestamp < 0) {
+          return Status(Substitute("Invalid Kudu snapshot read timestamp: Only "
+              "non-negative numbers are allowed.", value));
+        }
+        query_options->__set_kudu_snapshot_read_timestamp_micros(timestamp);
+        break;
+      }
+      case TImpalaQueryOptions::ENABLED_RUNTIME_FILTER_TYPES: {
+        // Parse the enabled runtime filter types and validate it.
+        TEnabledRuntimeFilterTypes::type enum_type;
+        RETURN_IF_ERROR(GetThriftEnum(value, "enabled runtime filter types",
+            _TEnabledRuntimeFilterTypes_VALUES_TO_NAMES, &enum_type));
+        query_options->__set_enabled_runtime_filter_types(enum_type);
+        break;
+      }
+      case TImpalaQueryOptions::ASYNC_CODEGEN: {
+        query_options->__set_async_codegen(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::ENABLE_DISTINCT_SEMI_JOIN_OPTIMIZATION: {
+        query_options->__set_enable_distinct_semi_join_optimization(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::SORT_RUN_BYTES_LIMIT: {
+        // Parse the sort bytes limit and validate it
+        int64_t sort_run_bytes_limit;
+        RETURN_IF_ERROR(
+            ParseMemValue(value, "sort run bytes limit", &sort_run_bytes_limit));
+        query_options->__set_sort_run_bytes_limit(sort_run_bytes_limit);
+        break;
+      }
+      case TImpalaQueryOptions::MAX_FS_WRITERS: {
+        StringParser::ParseResult result;
+        const int32_t max_fs_writers =
+            StringParser::StringToInt<int32_t>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || max_fs_writers < 0) {
+          return Status(Substitute("$0 is not valid for MAX_FS_WRITERS. Only "
+                                   "non-negative numbers are allowed.",
+              value));
+        }
+        query_options->__set_max_fs_writers(max_fs_writers);
+        break;
+      }
+      case TImpalaQueryOptions::REFRESH_UPDATED_HMS_PARTITIONS: {
+        query_options->__set_refresh_updated_hms_partitions(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::SPOOL_ALL_RESULTS_FOR_RETRIES: {
+        query_options->__set_spool_all_results_for_retries(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::RUNTIME_FILTER_ERROR_RATE: {
+        StringParser::ParseResult result;
+        const double val =
+            StringParser::StringToFloat<double>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || val <= 0 || val >= 1) {
+          return Status(Substitute("Invalid runtime filter error rate: "
+                "'$0'. Only values between 0 to 1 (exclusive) are allowed.", value));
+        }
+        query_options->__set_runtime_filter_error_rate(val);
+        break;
+      }
+      case TImpalaQueryOptions::USE_LOCAL_TZ_FOR_UNIX_TIMESTAMP_CONVERSIONS: {
+        query_options->__set_use_local_tz_for_unix_timestamp_conversions(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::CONVERT_LEGACY_HIVE_PARQUET_UTC_TIMESTAMPS: {
+        query_options->__set_convert_legacy_hive_parquet_utc_timestamps(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::ENABLE_OUTER_JOIN_TO_INNER_TRANSFORMATION: {
+         query_options->__set_enable_outer_join_to_inner_transformation(IsTrue(value));
+         break;
+      }
+      case TImpalaQueryOptions::TARGETED_KUDU_SCAN_RANGE_LENGTH: {
+        int64_t scan_length = 0;
+        RETURN_IF_ERROR(
+            ParseMemValue(value, "targeted kudu scan range length", &scan_length));
+        query_options->__set_targeted_kudu_scan_range_length(scan_length);
+        break;
+      }
+      case TImpalaQueryOptions::REPORT_SKEW_LIMIT: {
+        StringParser::ParseResult result;
+        const double skew_threshold =
+            StringParser::StringToFloat<double>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS) {
+          return Status(
+              Substitute("$0 is not valid for REPORT_SKEW_LIMIT. Only numeric "
+                         "values (such as 1.0) are allowed.",
+                  value));
+        }
+        query_options->__set_report_skew_limit(skew_threshold);
+        break;
+      }
+      case TImpalaQueryOptions::USE_DOP_FOR_COSTING: {
+        query_options->__set_use_dop_for_costing(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::BROADCAST_TO_PARTITION_FACTOR: {
+        StringParser::ParseResult result;
+        const double val =
+            StringParser::StringToFloat<double>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || val < 0 || val > 1000) {
+          return Status(Substitute("Invalid broadcast to partition factor '$0'. "
+                                   "Only values from 0 to 1000 are allowed.",
+              value));
+        }
+        query_options->__set_broadcast_to_partition_factor(val);
+        break;
+      }
+      case TImpalaQueryOptions::JOIN_ROWS_PRODUCED_LIMIT: {
+        StringParser::ParseResult result;
+        const int64_t join_rows_produced_limit =
+            StringParser::StringToInt<int64_t>(value.c_str(), value.length(), &result);
+        if (result != StringParser::PARSE_SUCCESS || join_rows_produced_limit < 0) {
+          return Status(Substitute("Invalid join rows produced limit: '$0'. "
+                                   "Only non-negative numbers are allowed.", value));
+        }
+        query_options->__set_join_rows_produced_limit(join_rows_produced_limit);
+        break;
+      }
+      case TImpalaQueryOptions::UTF8_MODE: {
+        query_options->__set_utf8_mode(IsTrue(value));
+        break;
+      }
+      case TImpalaQueryOptions::ANALYTIC_RANK_PUSHDOWN_THRESHOLD: {
+        StringParser::ParseResult status;
+        int64_t val =
+            StringParser::StringToInt<int64_t>(value.c_str(), value.size(), &status);
+        if (status != StringParser::PARSE_SUCCESS) {
+          return Status(Substitute("Invalid threshold: '$0'.", value));
+        }
+        if (val < -1) {
+          return Status(Substitute("Invalid threshold: '$0'. Only non-negative values "
+                "and -1 are allowed.", val));
+        }
+        query_options->__set_analytic_rank_pushdown_threshold(val);
+        break;
+      }
+      case TImpalaQueryOptions::DEFAULT_NDV_SCALE: {
+        StringParser::ParseResult result;
+        const int32_t scale =
+            StringParser::StringToInt<int32_t>(value.c_str(), value.length(), &result);
+        if (value == nullptr || result != StringParser::PARSE_SUCCESS || scale < 1 ||
+            scale > 10) {
+          return Status(
+              Substitute("Invalid NDV scale: '$0'. "
+                         "Only integer value in [1, 10] is allowed.", value));
+        }
+        query_options->__set_default_ndv_scale(scale);
+        break;
+      }
       default:
         if (IsRemovedQueryOption(key)) {
           LOG(WARNING) << "Ignoring attempt to set removed query option '" << key << "'";
