@@ -34,13 +34,15 @@
 
 #include "common/logging.h"
 #include "gutil/endian.h"
-#include "gutil/strings/substitute.h"
+#include "gutil/strings/escaping.h"
 #include "gutil/strings/strip.h"
+#include "gutil/strings/substitute.h"
+#include "kudu/security/gssapi.h"
 #include "kudu/util/env.h"
 #include "kudu/util/logging.h"
 #include "kudu/util/net/sockaddr.h"
-#include "kudu/security/gssapi.h"
-#include "rpc/cookie-util.h"
+#include "rpc/authentication.h"
+#include "rpc/authentication-util.h"
 #include "rpc/thrift-util.h"
 #include "runtime/exec-env.h"
 #include "service/impala-server.h"
@@ -105,7 +107,8 @@ DEFINE_string(webserver_authentication_domain, "",
     "Domain used for debug webserver authentication");
 DEFINE_string(webserver_password_file, "",
     "(Optional) Location of .htpasswd file containing user names and hashed passwords for"
-    " debug webserver authentication");
+    " debug webserver authentication. Cannot be used with --webserver_require_ldap or "
+    "--webserver_require_spnego.");
 
 DEFINE_string(webserver_x_frame_options, "DENY",
     "webserver will add X-Frame-Options HTTP header with this value");
@@ -114,14 +117,39 @@ DEFINE_int32(webserver_max_post_length_bytes, 1024 * 1024,
              "the embedded web server.");
 
 DEFINE_bool(webserver_require_spnego, false,
-            "Require connections to the web server to authenticate via Kerberos "
-            "using SPNEGO.");
+    "Require connections to the web server to authenticate via Kerberos using SPNEGO. "
+    "Cannot be used with --webserver_require_ldap or --webserver_password_file.");
 
+DEFINE_bool(webserver_require_ldap, false,
+    "Require connections to the web server to authenticate via LDAP using HTTP Basic "
+    "authentication. Cannot be used with --webserver_require_spnego or "
+    "--webserver_password_file.");
+DEFINE_string(webserver_ldap_user_filter, "",
+    "Used as filter for both simple and search bind mechanisms for the webserver "
+    "authentication. For simple bind it is a comma separated list of user names. If "
+    "specified, users must be on this list for authentication to succeed. For search "
+    "bind it is an LDAP filter that will be used during LDAP search, it can contain "
+    "'{0}' pattern which will be replaced with the user name.");
+DEFINE_string(webserver_ldap_group_filter, "",
+    "Used as filter for both simple and search bind mechanisms for the webserver "
+    "authentication. For simple bind it is a comma separated list of groups. If "
+    "specified, users must belong to one of these groups for authentication to succeed. "
+    "For search bind it is an LDAP filter that will be used during LDAP group search, it "
+    "can contain '{0}' pattern which will be replaced with the user name and/or '{1}' "
+    "which will be replace with the user dn.");
+
+DEFINE_bool(webserver_ldap_passwords_in_clear_ok, false,
+    "(Advanced) If true, allows the webserver to start with LDAP authentication even if "
+    "SSL is not enabled, a potentially insecure configuration.");
+
+DECLARE_bool(enable_ldap_auth);
 DECLARE_string(hostname);
 DECLARE_bool(is_coordinator);
 DECLARE_int64(max_cookie_lifetime_s);
 DECLARE_string(ssl_minimum_version);
 DECLARE_string(ssl_cipher_list);
+DECLARE_string(trusted_domain);
+DECLARE_bool(trusted_domain_use_xff_header);
 
 static const char* DOC_FOLDER = "/www/";
 static const int DOC_FOLDER_LEN = strlen(DOC_FOLDER);
@@ -231,7 +259,7 @@ kudu::Status RunSpnegoStep(
 
   string resp_token_b64;
   bool is_complete;
-  RETURN_NOT_OK(kudu::gssapi::SpnegoStep(
+  KUDU_RETURN_NOT_OK(kudu::gssapi::SpnegoStep(
       neg_token, &resp_token_b64, &is_complete, authn_user));
 
   if (!resp_token_b64.empty()) {
@@ -244,27 +272,42 @@ kudu::Status RunSpnegoStep(
 } // anonymous namespace
 
 Webserver::Webserver(MetricGroup* metrics)
-  : Webserver(FLAGS_webserver_interface, FLAGS_webserver_port, metrics) {}
+  : Webserver(FLAGS_webserver_interface, FLAGS_webserver_port, metrics,
+        GetConfiguredAuthMode()) {}
 
-Webserver::Webserver(const string& interface, const int port, MetricGroup* metrics)
-  : context_(nullptr),
+Webserver::Webserver(const string& interface, const int port, MetricGroup* metrics,
+    const AuthMode& auth_mode)
+  : auth_mode_(auth_mode),
+    context_(nullptr),
     error_handler_(UrlHandler(
         bind<void>(&Webserver::ErrorHandler, this, _1, _2), "error.tmpl", false)),
-    use_cookies_(FLAGS_max_cookie_lifetime_s > 0) {
+    use_cookies_(FLAGS_max_cookie_lifetime_s > 0),
+    check_trusted_domain_(!FLAGS_trusted_domain.empty()) {
   http_address_ = MakeNetworkAddress(interface.empty() ? "0.0.0.0" : interface, port);
   Init();
 
-  if (FLAGS_webserver_require_spnego) {
+  if (auth_mode_ == AuthMode::SPNEGO) {
     total_negotiate_auth_success_ =
         metrics->AddCounter("impala.webserver.total-negotiate-auth-success", 0);
     total_negotiate_auth_failure_ =
         metrics->AddCounter("impala.webserver.total-negotiate-auth-failure", 0);
-    if (use_cookies_) {
-      total_cookie_auth_success_ =
-          metrics->AddCounter("impala.webserver.total-cookie-auth-success", 0);
-      total_cookie_auth_failure_ =
-          metrics->AddCounter("impala.webserver.total-cookie-auth-failure", 0);
-    }
+  }
+  if (auth_mode_ == AuthMode::LDAP) {
+    total_basic_auth_success_ =
+        metrics->AddCounter("impala.webserver.total-basic-auth-success", 0);
+    total_basic_auth_failure_ =
+        metrics->AddCounter("impala.webserver.total-basic-auth-failure", 0);
+  }
+  if (use_cookies_ && (auth_mode_ == AuthMode::SPNEGO || auth_mode_ == AuthMode::LDAP)) {
+    total_cookie_auth_success_ =
+        metrics->AddCounter("impala.webserver.total-cookie-auth-success", 0);
+    total_cookie_auth_failure_ =
+        metrics->AddCounter("impala.webserver.total-cookie-auth-failure", 0);
+  }
+  if (check_trusted_domain_
+      && (auth_mode_ == AuthMode::SPNEGO || auth_mode_ == AuthMode::LDAP)) {
+    total_trusted_domain_check_success_ =
+        metrics->AddCounter("impala.webserver.total-trusted-domain-check-success", 0);
   }
 }
 
@@ -364,19 +407,24 @@ Status Webserver::Start() {
   }
 
   if (!FLAGS_webserver_password_file.empty()) {
-    // Squeasel doesn't log anything if it can't stat the password file (but will if it
-    // can't open it, which it tries to do during a request)
-    if (!exists(FLAGS_webserver_password_file)) {
-      stringstream ss;
-      ss << "Webserver: Password file does not exist: " << FLAGS_webserver_password_file;
-      return Status(ss.str());
+    if (FIPS_mode()) {
+      return Status("HTTP digest authorization is not supported in FIPS approved mode.");
+    } else {
+      // Squeasel doesn't log anything if it can't stat the password file (but will if it
+      // can't open it, which it tries to do during a request)
+      if (!exists(FLAGS_webserver_password_file)) {
+        stringstream ss;
+        ss << "Webserver: Password file does not exist: "
+           << FLAGS_webserver_password_file;
+        return Status(ss.str());
+      }
+      LOG(INFO) << "Webserver: Password file is " << FLAGS_webserver_password_file;
+      options.push_back("global_auth_file");
+      options.push_back(FLAGS_webserver_password_file.c_str());
     }
-    LOG(INFO) << "Webserver: Password file is " << FLAGS_webserver_password_file;
-    options.push_back("global_auth_file");
-    options.push_back(FLAGS_webserver_password_file.c_str());
   }
 
-  if (FLAGS_webserver_require_spnego) {
+  if (auth_mode_ == AuthMode::SPNEGO) {
     // If Kerberos has been configured, security::InitKerberosForServer() will
     // already have been called, ensuring that the keytab path has been
     // propagated into this environment variable where the GSSAPI calls will
@@ -387,6 +435,23 @@ Status Webserver::Start() {
       return Status("Unable to configure web server for SPNEGO authentication: "
                     "must configure a keytab file for the server");
     }
+    LOG(INFO) << "Webserver: secured with SPNEGO authentication.";
+  }
+
+  if (auth_mode_ == AuthMode::LDAP) {
+    if (!FLAGS_enable_ldap_auth) {
+      return Status("Unable to secure web server with LDAP: LDAP authentication must be "
+                    "configured for this daemon.");
+    }
+    if (!IsSecure() && !FLAGS_webserver_ldap_passwords_in_clear_ok) {
+      return Status("Unable to secure web server with LDAP: must either enable SSL or "
+                    "set --webserver_ldap_passwords_in_clear_ok=true");
+    }
+
+    RETURN_IF_ERROR(ImpalaLdap::CreateLdap(
+        &ldap_, FLAGS_webserver_ldap_user_filter, FLAGS_webserver_ldap_group_filter));
+
+    LOG(INFO) << "Webserver: secured with LDAP authentication.";
   }
 
   options.push_back("listening_ports");
@@ -481,16 +546,19 @@ void Webserver::GetCommonJson(
   }
 
   Value lst(kArrayType);
-  for (const UrlHandlerMap::value_type& handler: url_handlers_) {
-    if (handler.second.is_on_nav_bar()) {
-      Value hdl(kObjectType);
-      // Though we set link and title the same value, be careful with RapidJSON's MOVE
-      // semantic. We create the values by deep-copy here.
-      Value link(handler.first.c_str(), document->GetAllocator());
-      Value title(handler.first.c_str(), document->GetAllocator());
-      hdl.AddMember("link", link, document->GetAllocator());
-      hdl.AddMember("title", title, document->GetAllocator());
-      lst.PushBack(hdl, document->GetAllocator());
+  {
+    shared_lock<shared_mutex> lock(url_handlers_lock_);
+    for (const UrlHandlerMap::value_type& handler : url_handlers_) {
+      if (handler.second.is_on_nav_bar()) {
+        Value hdl(kObjectType);
+        // Though we set link and title the same value, be careful with RapidJSON's MOVE
+        // semantic. We create the values by deep-copy here.
+        Value link(handler.first.c_str(), document->GetAllocator());
+        Value title(handler.first.c_str(), document->GetAllocator());
+        hdl.AddMember("link", link, document->GetAllocator());
+        hdl.AddMember("title", title, document->GetAllocator());
+        lst.PushBack(hdl, document->GetAllocator());
+      }
     }
   }
 
@@ -516,6 +584,15 @@ sq_callback_result_t Webserver::BeginRequestCallbackStatic(
 
 sq_callback_result_t Webserver::BeginRequestCallback(struct sq_connection* connection,
     struct sq_request_info* request_info) {
+  if (VLOG_IS_ON(4)) {
+    VLOG(4) << request_info->request_method << " " << request_info->uri << " "
+            << request_info->http_version;
+    for (int i = 0; i < request_info->num_headers; ++i) {
+      VLOG(4) << "  " << request_info->http_headers[i].name << ": "
+              << request_info->http_headers[i].value;
+    }
+  }
+
   if (strncmp("OPTIONS", request_info->request_method, 7) == 0) {
     // Let Squeasel deal with the request. OPTIONS requests should not require
     // authentication, so do this before doing SPNEGO.
@@ -523,50 +600,79 @@ sq_callback_result_t Webserver::BeginRequestCallback(struct sq_connection* conne
   }
 
   vector<string> response_headers;
-  if (FLAGS_webserver_require_spnego){
-    bool authenticated = false;
-    // Try authenticating with a cookie first, if enabled.
-    if (use_cookies_) {
-      const char* cookie_header = sq_get_header(connection, "Cookie");
-      string username;
-      if (cookie_header != nullptr) {
-        Status cookie_status = AuthenticateCookie(hash_, cookie_header, &username);
-        if (cookie_status.ok()) {
-          authenticated = true;
-          request_info->remote_user = strdup(username.c_str());
-          total_cookie_auth_success_->Increment(1);
-        } else {
-          LOG(INFO) << "Invalid cookie provided: " << cookie_header << ": "
-                    << cookie_status.GetDetail();
-          response_headers.push_back(Substitute("Set-Cookie: $0", GetDeleteCookie()));
-          total_cookie_auth_failure_->Increment(1);
-        }
+  bool authenticated = auth_mode_ != AuthMode::SPNEGO && auth_mode_ != AuthMode::LDAP;
+  // Try authenticating with a cookie first, if enabled.
+  if (!authenticated && use_cookies_) {
+    const char* cookie_header = sq_get_header(connection, "Cookie");
+    string username;
+    if (cookie_header != nullptr) {
+      Status cookie_status = AuthenticateCookie(hash_, cookie_header, &username);
+      if (cookie_status.ok()) {
+        authenticated = true;
+        request_info->remote_user = strdup(username.c_str());
+        total_cookie_auth_success_->Increment(1);
+      } else {
+        LOG(INFO) << "Invalid cookie provided: " << cookie_header << ": "
+                  << cookie_status.GetDetail();
+        response_headers.push_back(Substitute("Set-Cookie: $0", GetDeleteCookie()));
+        total_cookie_auth_failure_->Increment(1);
       }
     }
+  }
 
-    if (!authenticated) {
+  // Connections originating from trusted domains should not require authentication.
+  // Returns a cookie on the first successful auth attempt. This check is performed after
+  // checking for cookie to avoid subsequent reverse DNS lookups which can be
+  // unpredictably costly.
+  if (!authenticated && check_trusted_domain_) {
+    const char* xff_origin = sq_get_header(connection, "X-Forwarded-For");
+    string xff_origin_string = !xff_origin ? "" : string(xff_origin);
+    string origin = FLAGS_trusted_domain_use_xff_header ?
+        xff_origin_string :
+        GetRemoteAddress(request_info).ToString();
+    StripWhiteSpace(&origin);
+    if (!origin.empty()) {
+      if (TrustedDomainCheck(origin, connection, request_info)) {
+        total_trusted_domain_check_success_->Increment(1);
+        authenticated = true;
+        AddCookie(request_info, &response_headers);
+      }
+    }
+  }
+
+  if (!authenticated) {
+    if (auth_mode_ == AuthMode::SPNEGO) {
       sq_callback_result_t spnego_result =
           HandleSpnego(connection, request_info, &response_headers);
       if (spnego_result == SQ_CONTINUE_HANDLING) {
         // Spnego negotiation was successful.
-        if (use_cookies_) {
-          // If cookie auth failed above and we generated a 'delete cookie' header,
-          // remove it.
-          auto eq = [](const string& header) {
-            return header.rfind("Set-Cookie", 0) == 0;
-          };
-          auto it = find_if(response_headers.begin(), response_headers.end(), eq);
-          if (it != response_headers.end()) {
-            response_headers.erase(it);
-          }
-          // Generate a cookie to return.
-          response_headers.push_back(Substitute(
-              "Set-Cookie: $0", GenerateCookie(request_info->remote_user, hash_)));
-        }
+        AddCookie(request_info, &response_headers);
       } else {
         // Spnego negotiation is incomplete or failed, stop processing the request.
         return spnego_result;
       }
+    } else {
+      DCHECK(auth_mode_ == AuthMode::LDAP);
+      Status basic_status = HandleBasic(connection, request_info, &response_headers);
+      if (basic_status.ok()) {
+        // Basic auth was successful.
+        total_basic_auth_success_->Increment(1);
+        AddCookie(request_info, &response_headers);
+      } else {
+        total_basic_auth_failure_->Increment(1);
+        if (!sq_get_header(connection, "Authorization")) {
+          // This case is expected, as some clients will always initially try to connect
+          // without an 'Authorization' and only provide one after getting the
+          // 'WWW-Authenticate' back, so we don't log it as an error.
+          VLOG(2) << "Not authenticated: no Authorization header provided.";
+        } else {
+          LOG(ERROR) << "Failed to authenticate: " << basic_status.GetDetail();
+        }
+        response_headers.push_back("WWW-Authenticate: Basic");
+        SendResponse(connection, "401 Authentication Required", "text/plain",
+            "Must authenticate with Basic authentication.", response_headers);
+        return SQ_HANDLED_OK;
+      }
     }
   }
 
@@ -701,6 +807,73 @@ sq_callback_result_t Webserver::HandleSpnego(struct sq_connection* connection,
   return SQ_CONTINUE_HANDLING;
 }
 
+bool Webserver::TrustedDomainCheck(const string& origin, struct sq_connection* connection,
+    struct sq_request_info* request_info) {
+  if (!IsTrustedDomain(origin, FLAGS_trusted_domain)) return false;
+  const char* authz_header = sq_get_header(connection, "Authorization");
+  if (!authz_header) {
+    LOG(ERROR) << "Passed TrustedDomainCheck but no Authorization header provided.";
+    return false;
+  }
+
+  string base64;
+  if (!TryStripPrefixString(authz_header, "Basic ", &base64)) {
+    LOG(ERROR) << "Passed TrustedDomainCheck but No Basic authentication provided.";
+    return false;
+  }
+  string username, password;
+  Status status = BasicAuthExtractCredentials(base64, username, password);
+  if (!status.ok()) {
+    LOG(ERROR) << "Error parsing basic authentication token from: "
+               << GetRemoteAddress(request_info).ToString() << " Error: " << status;
+    return false;
+  }
+  request_info->remote_user = strdup(username.c_str());
+  return true;
+}
+
+Status Webserver::HandleBasic(struct sq_connection* connection,
+    struct sq_request_info* request_info, vector<string>* response_headers) {
+  const char* authz_header = sq_get_header(connection, "Authorization");
+  if (!authz_header) {
+    return Status::Expected("No Authorization header provided.");
+  }
+
+  string base64;
+  if (!TryStripPrefixString(authz_header, "Basic ", &base64)) {
+    return Status::Expected("No Basic authentication provided.");
+  }
+  string username, password;
+  Status status = BasicAuthExtractCredentials(base64, username, password);
+  if (!status.ok()) {
+    LOG(ERROR) << "Error parsing basic authentication token from: "
+               << GetRemoteAddress(request_info).ToString() << " Error: " << status;
+    return status;
+  }
+  if (ldap_->LdapCheckPass(username.c_str(), password.c_str(), password.length())
+      && ldap_->LdapCheckFilters(username)) {
+    request_info->remote_user = strdup(username.c_str());
+    return Status::OK();
+  }
+
+  return Status::Expected("Failed to authenticate to LDAP.");
+}
+
+void Webserver::AddCookie(
+    struct sq_request_info* request_info, vector<string>* response_headers) {
+  if (use_cookies_) {
+    // If cookie auth failed and we generated a 'delete cookie' header, remove it.
+    auto eq = [](const string& header) { return header.rfind("Set-Cookie", 0) == 0; };
+    auto it = find_if(response_headers->begin(), response_headers->end(), eq);
+    if (it != response_headers->end()) {
+      response_headers->erase(it);
+    }
+    // Generate a cookie to return.
+    response_headers->push_back(
+        Substitute("Set-Cookie: $0", GenerateCookie(request_info->remote_user, hash_)));
+  }
+}
+
 void Webserver::RenderUrlWithTemplate(const struct sq_connection* connection,
     const WebRequest& req, const UrlHandler& url_handler, stringstream* output,
     ContentType* content_type) {
@@ -716,7 +889,11 @@ void Webserver::RenderUrlWithTemplate(const struct sq_connection* connection,
     // Callbacks may optionally be rendered as a text-only, pretty-printed Json document
     // (mostly for debugging or integration with third-party tools).
     StringBuffer strbuf;
+    // Write the JSON out with human-readable formatting. The settings are tweaked to
+    // reduce extraneous whitespace characters, compared to the default formatting.
     PrettyWriter<StringBuffer> writer(strbuf);
+    writer.SetIndent('\t', 1);
+    writer.SetFormatOptions(kFormatSingleLineArray);
     document.Accept(writer);
     (*output) << strbuf.GetString();
     *content_type = JSON;
@@ -777,4 +954,15 @@ const string Webserver::GetMimeType(const ContentType& content_type) {
       return "";
   }
 }
+
+Webserver::AuthMode Webserver::GetConfiguredAuthMode() {
+  if (!FLAGS_webserver_password_file.empty()) {
+    return AuthMode::HTPASSWD;
+  } else if (FLAGS_webserver_require_spnego) {
+    return AuthMode::SPNEGO;
+  } else if (FLAGS_webserver_require_ldap) {
+    return AuthMode::LDAP;
+  }
+  return AuthMode::NONE;
+}
 }
