@@ -1,4 +1,4 @@
-#!/usr/bin/env impala-python
+#!/usr/bin/env python
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
@@ -23,29 +23,24 @@
 # has been sourced by verifying that IMPALA_HOME is set. This script will fail if an
 # expected environment variable is not present.
 #
+# To share a toolchain directory between multiple checkouts of Impala (or to use a
+# cached copy to avoid downloading a new one), it is best to override IMPALA_TOOLCHAIN
+# in bin/impala-config-local.sh or in the environment prior to sourcing
+# bin/impala-config.sh. This sets IMPALA_TOOLCHAIN_PACKAGES_HOME as well as
+# CDP_COMPONENTS_HOME.
+#
 # The following environment variables control the behavior of this script:
-# IMPALA_TOOLCHAIN - Directory in which to place the artifacts. It can be useful to
-#   override this to share a toolchain directory between multiple checkouts of Impala
-#   or to use a cached copy to avoid downloading a new one.
+# IMPALA_TOOLCHAIN_PACKAGES_HOME - Directory in which to place the native-toolchain
+#   packages.
 # IMPALA_TOOLCHAIN_HOST - The host to use for downloading the artifacts
-# CDH_COMPONENTS_HOME - Directory to store CDH Hadoop component artifacts
 # CDP_COMPONENTS_HOME - Directory to store CDP Hadoop component artifacts
-# CDH_BUILD_NUMBER - CDH Hadoop components are built with consistent versions so that
+# CDP_BUILD_NUMBER - CDP Hadoop components are built with consistent versions so that
 #   Hadoop, Hive, Kudu, etc are all built with versions that are compatible with each
 #   other. The way to specify a single consistent set of components is via a build
 #   number. This determines the location in s3 to get the artifacts.
-# CDP_BUILD_NUMBER - The CDP equivalent of a CDH_BUILD_NUMBER.
-# USE_CDP_HIVE - If false, this will use the CDH version of all Hadoop components
-#   (except Ranger, which is CDP only). If true, this will use the CDP version of all
-#   Hadoop components (except Sentry, which is CDH only).
 # DOWNLOAD_CDH_COMPONENTS - When set to true, this script will also download and extract
-#   the CDH/CDP Hadoop components (i.e. Hadoop, Hive, HBase, Sentry, Ranger, etc) into
-#   CDH_COMPONENTS_HOME/CDP_COMPONENTS_HOME as appropriate.
-# USE_CDH_KUDU - Kudu can be downloaded either from the toolchain or as a CDH component,
-#   depending on the value of USE_CDH_KUDU.
-# KUDU_IS_SUPPORTED - If KUDU_IS_SUPPORTED is false, Kudu is disabled and we download
-#   the toolchain Kudu and use the symbols to compile a non-functional stub library so
-#   that Impala has something to link against.
+#   the CDP Hadoop components (i.e. Hadoop, Hive, HBase, Ranger, etc) into
+#   CDP_COMPONENTS_HOME as appropriate.
 # IMPALA_<PACKAGE>_VERSION - The version expected for <PACKAGE>. This is typically
 #   configured in bin/impala-config.sh and must exist for every package. This is used
 #   to construct an appropriate URL and expected archive name.
@@ -58,18 +53,13 @@
 #
 # The script is directly executable, and it takes no parameters:
 #     ./bootstrap_toolchain.py
-# It should NOT be run via 'python bootstrap_toolchain.py', as it relies on a specific
-# python environment.
 import logging
 import glob
 import multiprocessing.pool
 import os
+import platform
 import random
 import re
-# TODO: This file should be runnable without using impala-python, and system python
-# does not have 'sh' available. Rework code to avoid importing sh (and anything else
-# that gets in the way).
-import sh
 import shutil
 import subprocess
 import sys
@@ -80,31 +70,56 @@ from collections import namedtuple
 from string import Template
 
 # Maps return values from 'lsb_release -irs' to the corresponding OS labels for both the
-# toolchain and the CDH components.
+# toolchain and the CDP components.
+# For Ubuntu20.04, the toolchain and CDP components to be mapped to are still 18.04
+# based, due to the unavailability of 20.04 parts on EC2.
 OsMapping = namedtuple('OsMapping', ['lsb_release', 'toolchain', 'cdh'])
 OS_MAPPING = [
   OsMapping("centos5", "ec2-package-centos-5", None),
   OsMapping("centos6", "ec2-package-centos-6", "redhat6"),
   OsMapping("centos7", "ec2-package-centos-7", "redhat7"),
+  OsMapping("centos8", "ec2-package-centos-8", "redhat8"),
   OsMapping("redhatenterpriseserver5", "ec2-package-centos-5", None),
   OsMapping("redhatenterpriseserver6", "ec2-package-centos-6", "redhat6"),
   OsMapping("redhatenterpriseserver7", "ec2-package-centos-7", "redhat7"),
+  OsMapping("redhatenterprise8", "ec2-package-centos-8", "redhat8"),
+  OsMapping("redhatenterpriseserver8", "ec2-package-centos-8", "redhat8"),
   OsMapping("debian6", "ec2-package-debian-6", None),
   OsMapping("debian7", "ec2-package-debian-7", None),
   OsMapping("debian8", "ec2-package-debian-8", "debian8"),
   OsMapping("suselinux11", "ec2-package-sles-11", None),
   OsMapping("suselinux12", "ec2-package-sles-12", "sles12"),
-  OsMapping("suse12.2", "ec2-package-sles-12", "sles12"),
-  OsMapping("suse12.3", "ec2-package-sles-12", "sles12"),
+  OsMapping("suse12", "ec2-package-sles-12", "sles12"),
   OsMapping("ubuntu12.04", "ec2-package-ubuntu-12-04", None),
   OsMapping("ubuntu14.04", "ec2-package-ubuntu-14-04", None),
   OsMapping("ubuntu15.04", "ec2-package-ubuntu-14-04", None),
   OsMapping("ubuntu15.10", "ec2-package-ubuntu-14-04", None),
   OsMapping('ubuntu16.04', "ec2-package-ubuntu-16-04", "ubuntu1604"),
-  OsMapping('ubuntu18.04', "ec2-package-ubuntu-18-04", "ubuntu1804")
+  OsMapping('ubuntu18.04', "ec2-package-ubuntu-18-04", "ubuntu1804"),
+  OsMapping('ubuntu20.04', "ec2-package-ubuntu-18-04", "ubuntu1804")
 ]
 
 
+def check_output(cmd_args):
+  """Run the command and return the output. Raise an exception if the command returns
+     a non-zero return code. Similar to subprocess.check_output() which is only provided
+     in python 2.7.
+  """
+  process = subprocess.Popen(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+  stdout, _ = process.communicate()
+  if process.wait() != 0:
+    raise Exception("Command with args '%s' failed with exit code %s:\n%s"
+        % (cmd_args, process.returncode, stdout))
+  return stdout
+
+
+def get_toolchain_compiler():
+  """Return the <name>-<version> string for the compiler package to use for the
+  toolchain."""
+  # Currently we always use GCC.
+  return "gcc-{0}".format(os.environ["IMPALA_GCC_VERSION"])
+
+
 def wget_and_unpack_package(download_path, file_name, destination, wget_no_clobber):
   if not download_path.endswith("/" + file_name):
     raise Exception("URL {0} does not match with expected file_name {1}"
@@ -115,7 +130,10 @@ def wget_and_unpack_package(download_path, file_name, destination, wget_no_clobb
       download_path, destination, file_name, attempt))
     # --no-clobber avoids downloading the file if a file with the name already exists
     try:
-      sh.wget(download_path, directory_prefix=destination, no_clobber=wget_no_clobber)
+      cmd = ["wget", download_path, "--directory-prefix={0}".format(destination)]
+      if wget_no_clobber:
+        cmd.append("--no-clobber")
+      check_output(cmd)
       break
     except Exception, e:
       if attempt == NUM_ATTEMPTS:
@@ -123,8 +141,9 @@ def wget_and_unpack_package(download_path, file_name, destination, wget_no_clobb
       logging.error("Download failed; retrying after sleep: " + str(e))
       time.sleep(10 + random.random() * 5)  # Sleep between 10 and 15 seconds.
   logging.info("Extracting {0}".format(file_name))
-  sh.tar(z=True, x=True, f=os.path.join(destination, file_name), directory=destination)
-  sh.rm(os.path.join(destination, file_name))
+  check_output(["tar", "xzf", os.path.join(destination, file_name),
+                "--directory={0}".format(destination)])
+  os.unlink(os.path.join(destination, file_name))
 
 
 class DownloadUnpackTarball(object):
@@ -235,12 +254,12 @@ class EnvVersionedPackage(TemplatedDownloadUnpackTarball):
 
 class ToolchainPackage(EnvVersionedPackage):
   def __init__(self, name, explicit_version=None, platform_release=None):
-    toolchain_root = os.environ.get("IMPALA_TOOLCHAIN")
-    if not toolchain_root:
+    toolchain_packages_home = os.environ.get("IMPALA_TOOLCHAIN_PACKAGES_HOME")
+    if not toolchain_packages_home:
       logging.error("Impala environment not set up correctly, make sure "
-          "$IMPALA_TOOLCHAIN is set.")
+          "$IMPALA_TOOLCHAIN_PACKAGES_HOME is set.")
       sys.exit(1)
-    compiler = "gcc-{0}".format(os.environ["IMPALA_GCC_VERSION"])
+    compiler = get_toolchain_compiler()
     label = get_platform_release_label(release=platform_release).toolchain
     toolchain_build_id = os.environ["IMPALA_TOOLCHAIN_BUILD_ID"]
     toolchain_host = os.environ["IMPALA_TOOLCHAIN_HOST"]
@@ -251,7 +270,8 @@ class ToolchainPackage(EnvVersionedPackage):
     url_prefix_tmpl = "https://${toolchain_host}/build/${toolchain_build_id}/" + \
         "${name}/${version}-${compiler}/"
     unpack_directory_tmpl = "${name}-${version}"
-    super(ToolchainPackage, self).__init__(name, url_prefix_tmpl, toolchain_root,
+    super(ToolchainPackage, self).__init__(name, url_prefix_tmpl,
+                                           toolchain_packages_home,
                                            explicit_version=explicit_version,
                                            archive_basename_tmpl=archive_basename_tmpl,
                                            unpack_directory_tmpl=unpack_directory_tmpl,
@@ -280,32 +300,10 @@ class ToolchainPackage(EnvVersionedPackage):
       f.write(self.archive_basename)
 
 
-class CdhComponent(EnvVersionedPackage):
-  def __init__(self, name, explicit_version=None, archive_basename_tmpl=None,
-               unpack_directory_tmpl=None):
-    # Compute the CDH base URL (based on the IMPALA_TOOLCHAIN_HOST and CDH_BUILD_NUMBER)
-    if "IMPALA_TOOLCHAIN_HOST" not in os.environ or "CDH_BUILD_NUMBER" not in os.environ:
-      logging.error("Impala environment not set up correctly, make sure "
-                    "impala-config.sh is sourced.")
-      sys.exit(1)
-    template_subs = {"toolchain_host": os.environ["IMPALA_TOOLCHAIN_HOST"],
-                     "cdh_build_number": os.environ["CDH_BUILD_NUMBER"]}
-    url_prefix_tmpl = "https://${toolchain_host}/build/cdh_components/" + \
-        "${cdh_build_number}/tarballs/"
-
-    # Get the output base directory from CDH_COMPONENTS_HOME
-    destination_basedir = os.environ["CDH_COMPONENTS_HOME"]
-    super(CdhComponent, self).__init__(name, url_prefix_tmpl, destination_basedir,
-                                       explicit_version=explicit_version,
-                                       archive_basename_tmpl=archive_basename_tmpl,
-                                       unpack_directory_tmpl=unpack_directory_tmpl,
-                                       template_subs_in=template_subs)
-
-
 class CdpComponent(EnvVersionedPackage):
   def __init__(self, name, explicit_version=None, archive_basename_tmpl=None,
                unpack_directory_tmpl=None, makedir=False):
-    # Compute the CDH base URL (based on the IMPALA_TOOLCHAIN_HOST and CDP_BUILD_NUMBER)
+    # Compute the CDP base URL (based on the IMPALA_TOOLCHAIN_HOST and CDP_BUILD_NUMBER)
     if "IMPALA_TOOLCHAIN_HOST" not in os.environ or "CDP_BUILD_NUMBER" not in os.environ:
       logging.error("Impala environment not set up correctly, make sure "
                     "impala-config.sh is sourced.")
@@ -341,49 +339,6 @@ class ToolchainKudu(ToolchainPackage):
     return False
 
 
-class CdhKudu(CdhComponent):
-  def __init__(self, platform_label):
-    kudu_archive_tmpl = "kudu-${version}-" + platform_label
-    # IMPALA_KUDU_URL can contain '%(platform_label)', which needs to be replaced
-    # with the platform. We override this in os.environ so that it is picked up
-    # in EnvVersionedPackage.
-    kudu_url = os.environ.get("IMPALA_KUDU_URL")
-    if kudu_url:
-       kudu_url = kudu_url.replace("%(platform_label)", platform_label)
-       os.environ["IMPALA_KUDU_URL"] = kudu_url
-    super(CdhKudu, self).__init__('kudu',
-                                  archive_basename_tmpl=kudu_archive_tmpl,
-                                  unpack_directory_tmpl="kudu-${version}")
-
-  def needs_download(self):
-    # This verifies that the unpack directory exists
-    if super(CdhKudu, self).needs_download():
-      return True
-    # Additional check to distinguish this from the Kudu Java package
-    # Regardless of the actual build type, the 'kudu' tarball will always contain a
-    # 'debug' and a 'release' directory.
-    if not os.path.exists(os.path.join(self.pkg_directory(), "debug")):
-      return True
-    # Both the pkg_directory and the debug directory exist
-    return False
-
-
-class CdhKuduJava(CdhComponent):
-  def __init__(self):
-    super(CdhKuduJava, self).__init__('kudu-java',
-                                      archive_basename_tmpl="kudu-${version}")
-
-  def needs_download(self):
-    # This verify that the unpack directory exists
-    if super(CdhKuduJava, self).needs_download():
-      return True
-    # Additional check to distinguish this from the Kudu package
-    # There should be jars under the kudu directory.
-    if len(glob.glob("{0}/*jar".format(self.pkg_directory()))) == 0:
-      return True
-    return False
-
-
 def try_get_platform_release_label():
   """Gets the right package label from the OS version. Returns an OsMapping with both
      'toolchain' and 'cdh' labels. Return None if not found.
@@ -408,17 +363,17 @@ def get_platform_release_label(release=None):
     if lsb_release_cache:
       release = lsb_release_cache
     else:
-      release = "".join(map(lambda x: x.lower(), sh.lsb_release("-irs").split()))
-      # Only need to check against the major release if RHEL or CentOS
-      for platform in ['centos', 'redhatenterpriseserver']:
-        if platform in release:
+      lsb_release = check_output(["lsb_release", "-irs"])
+      release = "".join(map(lambda x: x.lower(), lsb_release.split()))
+      # Only need to check against the major release if RHEL, CentOS or Suse
+      for distro in ['centos', 'redhatenterprise', 'redhatenterpriseserver', 'suse']:
+        if distro in release:
           release = release.split('.')[0]
           break
       lsb_release_cache = release
   for mapping in OS_MAPPING:
     if re.search(mapping.lsb_release, release):
       return mapping
-
   raise Exception("Could not find package label for OS version: {0}.".format(release))
 
 
@@ -435,7 +390,7 @@ def check_output(cmd_args):
   return stdout
 
 
-def check_custom_toolchain(toolchain_root, packages):
+def check_custom_toolchain(toolchain_packages_home, packages):
   missing = []
   for p in packages:
     if not os.path.isdir(p.pkg_directory()):
@@ -452,114 +407,6 @@ def check_custom_toolchain(toolchain_root, packages):
     raise Exception("Toolchain bootstrap failed: required packages were missing")
 
 
-def build_kudu_stub(kudu_dir, gcc_dir):
-  """When Kudu isn't supported, the CentOS 7 Kudu package is downloaded from the
-     toolchain. This replaces the client lib with a stubbed client. The
-     'kudu_dir' specifies the location of the unpacked CentOS 7 Kudu package.
-     The 'gcc_dir' specifies the location of the unpacked GCC/G++."""
-
-  print "Building kudu stub"
-  # Find the client lib files in the Kudu dir. There may be several files with
-  # various extensions. Also there will be a debug version.
-  client_lib_paths = []
-  for path, _, files in os.walk(kudu_dir):
-    for file in files:
-      if not file.startswith("libkudu_client.so"):
-        continue
-      file_path = os.path.join(path, file)
-      if os.path.islink(file_path):
-        continue
-      client_lib_paths.append(file_path)
-  if not client_lib_paths:
-    raise Exception("Unable to find Kudu client lib under '%s'" % kudu_dir)
-
-  # The client stub will be create by inspecting a real client and extracting the
-  # symbols. The choice of which client file to use shouldn't matter.
-  client_lib_path = client_lib_paths[0]
-
-  # Use a newer version of binutils because on older systems the default binutils may
-  # not be able to read the newer binary.
-  binutils_dir = ToolchainPackage("binutils").pkg_directory()
-  nm_path = os.path.join(binutils_dir, "bin", "nm")
-  objdump_path = os.path.join(binutils_dir, "bin", "objdump")
-
-  # Extract the symbols and write the stubbed client source. There is a special method
-  # kudu::client::GetShortVersionString() that is overridden so that the stub can be
-  # identified by the caller.
-  get_short_version_sig = "kudu::client::GetShortVersionString()"
-  nm_out = check_output([nm_path, "--defined-only", "-D", client_lib_path])
-  stub_build_dir = tempfile.mkdtemp()
-  stub_client_src_file = open(os.path.join(stub_build_dir, "kudu_client.cc"), "w")
-  try:
-    stub_client_src_file.write("""
-#include <string>
-
-static const std::string kFakeKuduVersion = "__IMPALA_KUDU_STUB__";
-
-static void KuduNotSupported() {
-    *((char*)0) = 0;
-}
-
-namespace kudu { namespace client {
-std::string GetShortVersionString() { return kFakeKuduVersion; }
-}}
-""")
-    found_start_version_symbol = False
-    cpp_filt_path = os.path.join(binutils_dir, "bin", "c++filt")
-    for line in nm_out.splitlines():
-      addr, sym_type, mangled_name = line.split(" ")
-      # Skip special functions an anything that isn't a strong symbol. Any symbols that
-      # get passed this check must be related to Kudu. If a symbol unrelated to Kudu
-      # (ex: a boost symbol) gets defined in the stub, there's a chance the symbol could
-      # get used and crash Impala.
-      if mangled_name in ["_init", "_fini"] or sym_type not in "Tt":
-        continue
-      demangled_name = check_output([cpp_filt_path, mangled_name]).strip()
-      assert "kudu" in demangled_name, \
-          "Symbol doesn't appear to be related to Kudu: " + demangled_name
-      if demangled_name == get_short_version_sig:
-        found_start_version_symbol = True
-        continue
-      stub_client_src_file.write("""
-extern "C" void %s() {
-  KuduNotSupported();
-}
-""" % mangled_name)
-
-    if not found_start_version_symbol:
-      raise Exception("Expected to find symbol a corresponding to"
-          " %s but it was not found." % get_short_version_sig)
-    stub_client_src_file.flush()
-
-    # The soname is needed to avoid problem in packaging builds. Without the soname,
-    # the library dependency as listed in the impalad binary will be a full path instead
-    # of a short name. Debian in particular has problems with packaging when that happens.
-    objdump_out = check_output([objdump_path, "-p", client_lib_path])
-    for line in objdump_out.splitlines():
-      if "SONAME" not in line:
-        continue
-      # The line that needs to be parsed should be something like:
-      # "  SONAME               libkudu_client.so.0"
-      so_name = line.split()[1]
-      break
-    else:
-      raise Exception("Unable to extract soname from %s" % client_lib_path)
-
-    # Compile the library.
-    stub_client_lib_path = os.path.join(stub_build_dir, "libkudu_client.so")
-    toolchain_root = os.environ.get("IMPALA_TOOLCHAIN")
-    gpp = os.path.join(
-        toolchain_root, "gcc-%s" % os.environ.get("IMPALA_GCC_VERSION"), "bin", "g++")
-    subprocess.check_call([gpp, stub_client_src_file.name, "-shared", "-fPIC",
-        "-Wl,-soname,%s" % so_name, "-o", stub_client_lib_path])
-
-    # Replace the real libs with the stub.
-    for client_lib_path in client_lib_paths:
-      shutil.copyfile(stub_client_lib_path, client_lib_path)
-  finally:
-    shutil.rmtree(stub_build_dir)
-
-
 def execute_many(f, args):
   """
   Executes f(a) for a in args using a threadpool to execute in parallel.
@@ -592,17 +439,15 @@ def get_toolchain_downloads():
   toolchain_packages += map(ToolchainPackage,
       ["avro", "binutils", "boost", "breakpad", "bzip2", "cctz", "cmake", "crcutil",
        "flatbuffers", "gdb", "gflags", "glog", "gperftools", "gtest", "libev",
-       "libunwind", "lz4", "openldap", "openssl", "orc", "protobuf",
+       "libunwind", "lz4", "openldap", "openssl", "orc", "protobuf", "python",
        "rapidjson", "re2", "snappy", "thrift", "tpc-h", "tpc-ds", "zlib", "zstd"])
-  toolchain_packages += [ToolchainPackage("thrift",
-      explicit_version=os.environ.get("IMPALA_THRIFT11_VERSION"))]
   # Check whether this platform is supported (or whether a valid custom toolchain
   # has been provided).
   if not try_get_platform_release_label() \
      or not try_get_platform_release_label().toolchain:
-    toolchain_root = os.environ.get("IMPALA_TOOLCHAIN")
+    toolchain_packages_home = os.environ.get("IMPALA_TOOLCHAIN_PACKAGES_HOME")
     # This would throw an exception if the custom toolchain were not valid
-    check_custom_toolchain(toolchain_root, toolchain_packages)
+    check_custom_toolchain(toolchain_packages_home, toolchain_packages)
     # Nothing to download
     return []
   return toolchain_packages
@@ -610,77 +455,50 @@ def get_toolchain_downloads():
 
 def get_hadoop_downloads():
   cluster_components = []
-  use_cdp_hive = os.environ["USE_CDP_HIVE"] == "true"
-  if use_cdp_hive:
-    hadoop = CdpComponent("hadoop")
-    hbase = CdpComponent("hbase", archive_basename_tmpl="hbase-${version}-bin",
-                         unpack_directory_tmpl="hbase-${version}")
-    hive = CdpComponent("hive", archive_basename_tmpl="apache-hive-${version}-bin")
-    hive_src = CdpComponent("hive-source",
-                            explicit_version=os.environ.get("IMPALA_HIVE_VERSION"),
-                            archive_basename_tmpl="hive-${version}-source",
-                            unpack_directory_tmpl="hive-${version}")
-    tez = CdpComponent("tez", archive_basename_tmpl="tez-${version}-minimal",
-                       makedir=True)
-    cluster_components.extend([hadoop, hbase, hive, hive_src, tez])
+  hadoop = CdpComponent("hadoop")
+  hbase = CdpComponent("hbase", archive_basename_tmpl="hbase-${version}-bin",
+                       unpack_directory_tmpl="hbase-${version}")
+  hive = CdpComponent("hive", archive_basename_tmpl="apache-hive-${version}-bin")
+  hive_src = CdpComponent("hive-source",
+                          explicit_version=os.environ.get("IMPALA_HIVE_VERSION"),
+                          archive_basename_tmpl="hive-${version}-source",
+                          unpack_directory_tmpl="hive-${version}")
+  tez = CdpComponent("tez", archive_basename_tmpl="tez-${version}-minimal",
+                     makedir=True)
+  use_override_hive = \
+      "HIVE_VERSION_OVERRIDE" in os.environ and os.environ["HIVE_VERSION_OVERRIDE"] != ""
+  # If we are using a locally built Hive we do not have a need to pull hive as a
+  # dependency
+  if use_override_hive:
+    cluster_components.extend([hadoop, hbase, tez])
   else:
-    cluster_components.extend(map(CdhComponent, ["hadoop", "hbase", "hive"]))
-  # Sentry is always CDH
-  cluster_components.append(CdhComponent("sentry"))
+    cluster_components.extend([hadoop, hbase, hive, hive_src, tez])
   # Ranger is always CDP
   cluster_components.append(CdpComponent("ranger",
                                          archive_basename_tmpl="ranger-${version}-admin"))
   return cluster_components
 
 
-def get_kudu_downloads(use_kudu_stub):
-  # If Kudu is not supported, we download centos7 kudu to build the kudu stub.
-  # TODO: Should this be from toolchain or CDH? Does it matter?
-  kudu_downloads = []
-  if use_kudu_stub:
-    kudu_downloads += [ToolchainKudu("centos7")]
-  else:
-    use_cdh_kudu = os.getenv("USE_CDH_KUDU") == "true"
-    if use_cdh_kudu:
-      if not try_get_platform_release_label() \
-         or not try_get_platform_release_label().cdh:
-        logging.error("CDH Kudu is not supported on this platform. Set "
-                      "USE_CDH_KUDU=false to use the toolchain Kudu.")
-        sys.exit(1)
-      kudu_downloads += [CdhKudu(get_platform_release_label().cdh)]
-      # There is also a Kudu Java package.
-      kudu_downloads += [CdhKuduJava()]
-    else:
-      # Toolchain Kudu includes Java artifacts.
-      kudu_downloads += [ToolchainKudu()]
-
-  return kudu_downloads
+def get_kudu_downloads():
+  # Toolchain Kudu includes Java artifacts.
+  return [ToolchainKudu()]
 
 
 def main():
-  """Validates that bin/impala-config.sh has been sourced by verifying that $IMPALA_HOME
-  and $IMPALA_TOOLCHAIN are in the environment. We assume that if these are set, then
-  IMPALA_<PACKAGE>_VERSION environment variables are also set. This will create the
-  directory specified by $IMPALA_TOOLCHAIN if it does not already exist. Then, it will
-  compute what packages need to be downloaded. Packages are only downloaded if they are
-  not already present in $IMPALA_TOOLCHAIN. There are two main categories of packages.
-  Toolchain packages are native packages built using the native toolchain. These are
-  always downloaded. Hadoop component packages are the CDH or CDP builds of Hadoop
-  components such as Hadoop, Hive, HBase, etc. Hadoop component packages are organized
-  as a consistent set of compatible version via a build number (i.e. CDH_BUILD_NUMBER
-  and CDP_BUILD_NUMBER). Hadoop component packages are only downloaded if
-  $DOWNLOAD_CDH_COMPONENTS is true. CDH Hadoop packages are downloaded into
-  $CDH_COMPONENTS_HOME. CDP Hadoop packages are downloaded into $CDP_COMPONENTS_HOME.
-  The versions used for Hadoop components depend on whether USE_CDP_HIVE is true or
-  false. If true, most components get the CDP versions based on the $CDP_BUILD_NUMBER.
-  If false, most components get the CDH versions based on the $CDH_BUILD_NUMBER.
-  The exceptions are:
-  - sentry (always downloaded from $IMPALA_TOOLCHAIN_HOST for a given $CDH_BUILD_NUMBER)
-  - ranger (always downloaded from $IMPALA_TOOLCHAIN_HOST for a given $CDP_BUILD_NUMBER)
-  - kudu (currently always downloaded from $IMPALA_TOOLCHAIN_HOST for a given
-    $CDH_BUILD_NUMBER)
-  If Kudu is not supported on this platform (or KUDU_IS_SUPPORTED=false), then this
-  builds a Kudu stub to allow for compilation without Kudu support.
+  """
+  Validates that bin/impala-config.sh has been sourced by verifying that $IMPALA_HOME
+  and $IMPALA_TOOLCHAIN_PACKAGES_HOME are in the environment. We assume that if these
+  are set, then IMPALA_<PACKAGE>_VERSION environment variables are also set. This will
+  create the directory specified by $IMPALA_TOOLCHAIN_PACKAGES_HOME if it does not
+  already exist. Then, it will compute what packages need to be downloaded. Packages are
+  only downloaded if they are not already present. There are two main categories of
+  packages. Toolchain packages are native packages built using the native toolchain.
+  These are always downloaded. Hadoop component packages are the CDP builds of Hadoop
+  components such as Hadoop, Hive, HBase, etc. Hadoop component packages are organized as
+  a consistent set of compatible version via a build number (i.e. CDP_BUILD_NUMBER).
+  Hadoop component packages are only downloaded if $DOWNLOAD_CDH_COMPONENTS is true. The
+  versions used for Hadoop components come from the CDP versions based on the
+  $CDP_BUILD_NUMBER. CDP Hadoop packages are downloaded into $CDP_COMPONENTS_HOME.
   """
   logging.basicConfig(level=logging.INFO,
       format='%(asctime)s %(threadName)s %(levelname)s: %(message)s')
@@ -693,17 +511,16 @@ def main():
     sys.exit(1)
 
   # Create the toolchain directory if necessary
-  create_directory_from_env_var("IMPALA_TOOLCHAIN")
-
-  use_kudu_stub = os.environ["KUDU_IS_SUPPORTED"] != "true"
+  create_directory_from_env_var("IMPALA_TOOLCHAIN_PACKAGES_HOME")
 
   downloads = []
-  downloads += get_toolchain_downloads()
+  if os.getenv("SKIP_TOOLCHAIN_BOOTSTRAP", "false") != "true":
+    downloads += get_toolchain_downloads()
   kudu_download = None
   if os.getenv("DOWNLOAD_CDH_COMPONENTS", "false") == "true":
-    create_directory_from_env_var("CDH_COMPONENTS_HOME")
     create_directory_from_env_var("CDP_COMPONENTS_HOME")
-    downloads += get_kudu_downloads(use_kudu_stub)
+    if platform.processor() != "aarch64":
+      downloads += get_kudu_downloads()
     downloads += get_hadoop_downloads()
 
   components_needing_download = [d for d in downloads if d.needs_download()]
@@ -713,11 +530,5 @@ def main():
 
   execute_many(download, components_needing_download)
 
-  if use_kudu_stub:
-    # Find the kudu package directory and the gcc package directory
-    kudu_download = [d for d in downloads if d.name == 'kudu'][0]
-    gcc_download = [d for d in downloads if d.name == 'gcc'][0]
-    build_kudu_stub(kudu_download.pkg_directory(), gcc_download.pkg_directory())
-
 
 if __name__ == "__main__": main()
