@@ -63,32 +63,23 @@ fi
 # variable from a previously-sourced impala-config.sh overrides the new value.        #
 #######################################################################################
 
-# The current Impala version that will be embedded in the Impala binary. This is
-# also used to find the Impala frontend jar files, so the version must match
-# the version in our Maven pom.xml files. This is validated via
-# bin/validate-java-pom-version.sh during the build.
-# WARNING: If changing this value, also run these commands:
-# cd ${IMPALA_HOME}/java
-# mvn versions:set -DnewVersion=YOUR_NEW_VERSION
-export IMPALA_VERSION=3.4.2-RELEASE
-
 # The unique build id of the toolchain to use if bootstrapping. This is generated by the
 # native-toolchain build when publishing its build artifacts. This should be changed when
 # moving to a different build of the toolchain, e.g. when a version is bumped or a
 # compile option is changed. The build id can be found in the output of the toolchain
 # build jobs, it is constructed from the build number and toolchain git hash prefix.
-export IMPALA_TOOLCHAIN_BUILD_ID=159-0989d236fd
+export IMPALA_TOOLCHAIN_BUILD_ID=10-425e7fd862
 # Versions of toolchain dependencies.
 # -----------------------------------
 export IMPALA_AVRO_VERSION=1.7.4-p5
 unset IMPALA_AVRO_URL
-export IMPALA_BINUTILS_VERSION=2.26.1
+export IMPALA_BINUTILS_VERSION=2.28
 unset IMPALA_BINUTILS_URL
 export IMPALA_BOOST_VERSION=1.61.0-p2
 unset IMPALA_BOOST_URL
 export IMPALA_BREAKPAD_VERSION=97a98836768f8f0154f8f86e5e14c2bb7e74132e-p2
 unset IMPALA_BREAKPAD_URL
-export IMPALA_BZIP2_VERSION=1.0.6-p2
+export IMPALA_BZIP2_VERSION=1.0.8-p2
 unset IMPALA_BZIP2_URL
 export IMPALA_CCTZ_VERSION=2.2
 unset IMPALA_CCTZ_URL
@@ -100,7 +91,7 @@ export IMPALA_CYRUS_SASL_VERSION=2.1.23
 unset IMPALA_CYRUS_SASL_URL
 export IMPALA_FLATBUFFERS_VERSION=1.6.0
 unset IMPALA_FLATBUFFERS_URL
-export IMPALA_GCC_VERSION=4.9.2
+export IMPALA_GCC_VERSION=7.5.0
 unset IMPALA_GCC_URL
 export IMPALA_GDB_VERSION=7.9.1-p1
 unset IMPALA_GDB_URL
@@ -108,7 +99,7 @@ export IMPALA_GFLAGS_VERSION=2.2.0-p2
 unset IMPALA_GFLAGS_URL
 export IMPALA_GLOG_VERSION=0.3.4-p3
 unset IMPALA_GLOG_URL
-export IMPALA_GPERFTOOLS_VERSION=2.5
+export IMPALA_GPERFTOOLS_VERSION=2.5-p2
 unset IMPALA_GPERFTOOLS_URL
 export IMPALA_GTEST_VERSION=1.6.0
 unset IMPALA_GTEST_URL
@@ -116,11 +107,14 @@ export IMPALA_LIBEV_VERSION=4.20
 unset IMPALA_LIBEV_URL
 export IMPALA_LIBUNWIND_VERSION=1.3-rc1-p3
 unset IMPALA_LIBUNWIND_URL
-export IMPALA_LLVM_VERSION=5.0.1-p2
+export IMPALA_LLVM_VERSION=5.0.1-p3
 unset IMPALA_LLVM_URL
-export IMPALA_LLVM_ASAN_VERSION=5.0.1-p2
+export IMPALA_LLVM_ASAN_VERSION=5.0.1-p3
 unset IMPALA_LLVM_ASAN_URL
 
+# Maximum memory available for mini-cluster and CDH cluster
+export IMPALA_CLUSTER_MAX_MEM_GB
+
 # LLVM stores some files in subdirectories that are named after what
 # version it thinks it is. We might think it is 5.0.1-p1, based on a
 # patch we have applied, but LLVM thinks its version is 5.0.1.
@@ -128,11 +122,11 @@ export IMPALA_LLVM_UBSAN_BASE_VERSION=5.0.1
 
 # Debug builds should use the release+asserts build to get additional coverage.
 # Don't use the LLVM debug build because the binaries are too large to distribute.
-export IMPALA_LLVM_DEBUG_VERSION=5.0.1-asserts-p2
+export IMPALA_LLVM_DEBUG_VERSION=5.0.1-asserts-p3
 unset IMPALA_LLVM_DEBUG_URL
-export IMPALA_LZ4_VERSION=1.7.5
+export IMPALA_LZ4_VERSION=1.9.3
 unset IMPALA_LZ4_URL
-export IMPALA_ZSTD_VERSION=1.4.0
+export IMPALA_ZSTD_VERSION=1.4.9
 unset IMPALA_ZSTD_URL
 export IMPALA_OPENLDAP_VERSION=2.4.47
 unset IMPALA_OPENLDAP_URL
@@ -142,13 +136,15 @@ export IMPALA_ORC_VERSION=1.6.2-p7
 unset IMPALA_ORC_URL
 export IMPALA_PROTOBUF_VERSION=3.5.1
 unset IMPALA_PROTOBUF_URL
-export IMPALA_POSTGRES_JDBC_DRIVER_VERSION=42.2.5
+export IMPALA_POSTGRES_JDBC_DRIVER_VERSION=42.2.14
 unset IMPALA_POSTGRES_JDBC_DRIVER_URL
+export IMPALA_PYTHON_VERSION=2.7.16
+unset IMPALA_PYTHON_URL
 export IMPALA_RAPIDJSON_VERSION=1.1.0
 unset IMPALA_RAPIDJSON_URL
 export IMPALA_RE2_VERSION=20190301
 unset IMPALA_RE2_URL
-export IMPALA_SNAPPY_VERSION=1.1.4
+export IMPALA_SNAPPY_VERSION=1.1.8
 unset IMPALA_SNAPPY_URL
 export IMPALA_SQUEASEL_VERSION=3.3
 unset IMPALA_SQUEASEL_URL
@@ -157,11 +153,9 @@ export IMPALA_TPC_DS_VERSION=2.1.0
 unset IMPALA_TPC_DS_URL
 export IMPALA_TPC_H_VERSION=2.17.0
 unset IMPALA_TPC_H_URL
-export IMPALA_THRIFT_VERSION=0.9.3-p7
+export IMPALA_THRIFT_VERSION=0.11.0-p4
 unset IMPALA_THRIFT_URL
-export IMPALA_THRIFT11_VERSION=0.11.0-p2
-unset IMPALA_THRIFT11_URL
-export IMPALA_ZLIB_VERSION=1.2.8
+export IMPALA_ZLIB_VERSION=1.2.11
 unset IMPALA_ZLIB_URL
 
 if [[ $OSTYPE == "darwin"* ]]; then
@@ -175,40 +169,39 @@ fi
 
 : ${IMPALA_TOOLCHAIN_HOST:=native-toolchain.s3.amazonaws.com}
 export IMPALA_TOOLCHAIN_HOST
-export CDH_BUILD_NUMBER=1814051
-export CDH_MAVEN_REPOSITORY=\
-"https://${IMPALA_TOOLCHAIN_HOST}/build/cdh_components/${CDH_BUILD_NUMBER}/maven"
-export CDH_HADOOP_VERSION=3.0.0-cdh6.x-SNAPSHOT
-export CDH_HBASE_VERSION=2.1.0-cdh6.x-SNAPSHOT
-export CDH_HIVE_VERSION=2.1.1-cdh6.x-SNAPSHOT
-export CDH_SENTRY_VERSION=2.1.0-cdh6.x-SNAPSHOT
 
-export CDP_BUILD_NUMBER=1617729
+#TODO: On next CDP_BUILD_NUMBER bump check if the following issues are resolved
+# and adjust existing tests that mentions them:
+# * HIVE-23995
+# * HIVE-24175
+export CDP_BUILD_NUMBER=11920537
 export CDP_MAVEN_REPOSITORY=\
 "https://${IMPALA_TOOLCHAIN_HOST}/build/cdp_components/${CDP_BUILD_NUMBER}/maven"
-export CDP_HADOOP_VERSION=3.1.1.7.0.2.0-212
-export CDP_HBASE_VERSION=2.2.0.7.0.2.0-212
-export CDP_HIVE_VERSION=3.1.2000.7.0.2.0-212
-export CDP_RANGER_VERSION=2.0.0.7.0.2.0-212
-export CDP_TEZ_VERSION=0.9.1.7.0.2.0-212
-export CDP_KNOX_VERSION=1.3.0.7.0.2.0-212
-export CDP_OZONE_VERSION=0.4.0.7.0.2.0-212
+export CDP_AVRO_JAVA_VERSION=1.8.2.7.2.9.0-146
+export CDP_HADOOP_VERSION=3.1.1.7.2.9.0-146
+export CDP_HBASE_VERSION=2.2.6.7.2.9.0-146
+export CDP_HIVE_VERSION=3.1.3000.7.2.9.0-146
+export CDP_ICEBERG_VERSION=0.9.1.7.2.9.0-146
+export CDP_KNOX_VERSION=1.3.0.7.2.9.0-146
+export CDP_OZONE_VERSION=1.0.0.7.2.9.0-146
+export CDP_PARQUET_VERSION=1.10.99.7.2.9.0-146
+export CDP_RANGER_VERSION=2.1.0.7.2.9.0-146
+export CDP_TEZ_VERSION=0.9.1.7.2.9.0-146
+export CDP_GCS_VERSION=2.1.2.7.2.9.0-146
+
+export ARCH_NAME=$(uname -p)
 
-export IMPALA_PARQUET_VERSION=1.10.99-cdh6.x-SNAPSHOT
-export IMPALA_AVRO_JAVA_VERSION=1.8.2-cdh6.x-SNAPSHOT
 export IMPALA_HUDI_VERSION=0.5.0-incubating
-export IMPALA_KITE_VERSION=1.0.0-cdh6.x-SNAPSHOT
+export IMPALA_KITE_VERSION=1.1.0
 export IMPALA_ORC_JAVA_VERSION=1.6.2
 
-# When IMPALA_(CDH_COMPONENT)_URL are overridden, they may contain '$(platform_label)'
-# which will be substituted for the CDH platform label in bootstrap_toolchain.py
+# When IMPALA_(CDP_COMPONENT)_URL are overridden, they may contain '$(platform_label)'
+# which will be substituted for the CDP platform label in bootstrap_toolchain.py
 unset IMPALA_HADOOP_URL
 unset IMPALA_HBASE_URL
 unset IMPALA_HIVE_URL
 unset IMPALA_KUDU_URL
 unset IMPALA_KUDU_VERSION
-unset IMPALA_KUDU_JAVA_VERSION
-unset IMPALA_SENTRY_URL
 
 export IMPALA_KERBERIZE=false
 
@@ -223,66 +216,56 @@ if [ -f "$IMPALA_HOME/bin/impala-config-local.sh" ]; then
   . "$IMPALA_HOME/bin/impala-config-local.sh"
 fi
 
-export CDH_HIVE_URL=${CDH_HIVE_URL-}
-export CDH_HADOOP_URL=${CDH_HADOOP_URL-}
-export CDH_HBASE_URL=${CDH_HBASE_URL-}
-export CDH_SENTRY_URL=${CDH_SENTRY_URL-}
+# IMPALA_TOOLCHAIN_PACKAGES_HOME is the location inside IMPALA_TOOLCHAIN where native
+# toolchain packages are placed. This uses a subdirectory that contains the information
+# about the compiler to allow using different compiler versions.
+export IMPALA_TOOLCHAIN_PACKAGES_HOME=\
+${IMPALA_TOOLCHAIN}/toolchain-packages-gcc${IMPALA_GCC_VERSION}
 
-export CDP_HIVE_URL=${CDP_HIVE_URL-}
-export CDP_HIVE_SOURCE_URL=${CDP_HIVE_SOURCE_URL-}
 export CDP_HADOOP_URL=${CDP_HADOOP_URL-}
 export CDP_HBASE_URL=${CDP_HBASE_URL-}
-export CDP_TEZ_URL=${CDP_TEZ_URL-}
+export CDP_HIVE_URL=${CDP_HIVE_URL-}
+export CDP_HIVE_SOURCE_URL=${CDP_HIVE_SOURCE_URL-}
+export CDP_ICEBERG_URL=${CDP_ICEBERG_URL-}
 export CDP_RANGER_URL=${CDP_RANGER_URL-}
+export CDP_TEZ_URL=${CDP_TEZ_URL-}
 
-export CDH_COMPONENTS_HOME="$IMPALA_TOOLCHAIN/cdh_components-$CDH_BUILD_NUMBER"
 export CDP_COMPONENTS_HOME="$IMPALA_TOOLCHAIN/cdp_components-$CDP_BUILD_NUMBER"
-export USE_CDP_HIVE=${USE_CDP_HIVE-false}
-if $USE_CDP_HIVE; then
-  # When USE_CDP_HIVE is set we use the CDP hive version to build as well as deploy in
-  # the minicluster
-  export CDH_MAJOR_VERSION=7
-  export IMPALA_HIVE_VERSION=${CDP_HIVE_VERSION}
-  export IMPALA_HIVE_URL=${CDP_HIVE_URL-}
-  export IMPALA_HIVE_SOURCE_URL=${CDP_HIVE_SOURCE_URL-}
-  export IMPALA_HADOOP_VERSION=${CDP_HADOOP_VERSION}
-  export IMPALA_HADOOP_URL=${CDP_HADOOP_URL-}
-  export IMPALA_HBASE_VERSION=${CDP_HBASE_VERSION}
-  export IMPALA_HBASE_URL=${CDP_HBASE_URL-}
-  export IMPALA_TEZ_VERSION=${CDP_TEZ_VERSION}
-  export IMPALA_TEZ_URL=${CDP_TEZ_URL-}
-  export IMPALA_KNOX_VERSION=${CDP_KNOX_VERSION}
-  export HADOOP_HOME="$CDP_COMPONENTS_HOME/hadoop-${IMPALA_HADOOP_VERSION}/"
-else
-  # CDH hive version is used to build and deploy in minicluster when USE_CDP_HIVE is
-  # false
-  export CDH_MAJOR_VERSION=6
-  export IMPALA_HIVE_VERSION=${CDH_HIVE_VERSION}
-  export IMPALA_HIVE_URL=${CDH_HIVE_URL-}
-  export IMPALA_HADOOP_VERSION=${CDH_HADOOP_VERSION}
-  export IMPALA_HADOOP_URL=${CDH_HADOOP_URL-}
-  export IMPALA_HBASE_VERSION=${CDH_HBASE_VERSION}
-  export IMPALA_HBASE_URL=${CDH_HBASE_URL-}
-  export HADOOP_HOME="$CDH_COMPONENTS_HOME/hadoop-${IMPALA_HADOOP_VERSION}/"
-fi
-
-# Ozone always uses the CDP version
+export CDH_MAJOR_VERSION=7
+export IMPALA_AVRO_JAVA_VERSION=${CDP_AVRO_JAVA_VERSION}
+export IMPALA_HADOOP_VERSION=${CDP_HADOOP_VERSION}
+export IMPALA_HADOOP_URL=${CDP_HADOOP_URL-}
+export HADOOP_HOME="$CDP_COMPONENTS_HOME/hadoop-${IMPALA_HADOOP_VERSION}/"
+export IMPALA_HBASE_VERSION=${CDP_HBASE_VERSION}
+export IMPALA_HBASE_URL=${CDP_HBASE_URL-}
+export IMPALA_HIVE_VERSION=${HIVE_VERSION_OVERRIDE:-"$CDP_HIVE_VERSION"}
+export IMPALA_HIVE_URL=${CDP_HIVE_URL-}
+export IMPALA_HIVE_SOURCE_URL=${CDP_HIVE_SOURCE_URL-}
+export IMPALA_ICEBERG_VERSION=${CDP_ICEBERG_VERSION}
+export IMPALA_ICEBERG_URL=${CDP_ICEBERG_URL-}
+export IMPALA_KNOX_VERSION=${CDP_KNOX_VERSION}
 export IMPALA_OZONE_VERSION=${CDP_OZONE_VERSION}
-
-# Ranger always uses the CDP version
+export IMPALA_PARQUET_VERSION=${CDP_PARQUET_VERSION}
 export IMPALA_RANGER_VERSION=${CDP_RANGER_VERSION}
 export IMPALA_RANGER_URL=${CDP_RANGER_URL-}
-
-# Sentry always uses the CDH version
-export IMPALA_SENTRY_VERSION=${CDH_SENTRY_VERSION}
-export IMPALA_SENTRY_URL=${CDH_SENTRY_URL-}
+export IMPALA_TEZ_VERSION=${CDP_TEZ_VERSION}
+export IMPALA_TEZ_URL=${CDP_TEZ_URL-}
+export IMPALA_HIVE_STORAGE_API_VERSION=${HIVE_STORAGE_API_VERSION_OVERRIDE:-"2.3.0.$IMPALA_HIVE_VERSION"}
+export IMPALA_GCS_VERSION=${CDP_GCS_VERSION}
 
 # Extract the first component of the hive version.
 # Allow overriding of Hive source location in case we want to build Impala without
-# a complete Hive build. This is used by fe/pom.xml to activate compatibility shims
-# for Hive-2 or Hive-3
+# a complete Hive build. This is used by various tests and scripts to enable and
+# disable tests and functionality.
 export IMPALA_HIVE_MAJOR_VERSION=$(echo "$IMPALA_HIVE_VERSION" | cut -d . -f 1)
 
+# Hive 1 and 2 are no longer supported.
+if [[ "${IMPALA_HIVE_MAJOR_VERSION}" == "1" ||
+      "${IMPALA_HIVE_MAJOR_VERSION}" == "2" ]]; then
+  echo "Hive 1 and 2 are no longer supported"
+  return 1
+fi
+
 # It is important to have a coherent view of the JAVA_HOME and JAVA executable.
 # The JAVA_HOME should be determined first, then the JAVA executable should be
 # derived from JAVA_HOME. bin/bootstrap_development.sh adds code to
@@ -309,6 +292,11 @@ if [[ ! -e "$JAVA" ]]; then
   return 1
 fi
 
+# Java libraries required by executables and java tests.
+export LIB_JAVA=$(find "${JAVA_HOME}/" -name libjava.so | head -1)
+export LIB_JSIG=$(find "${JAVA_HOME}/" -name libjsig.so | head -1)
+export LIB_JVM=$(find "${JAVA_HOME}/" -name libjvm.so  | head -1)
+
 #########################################################################################
 # Below here are variables that can be overridden by impala-config-*.sh and environment #
 # vars, variables computed based on other variables, and variables that cannot be       #
@@ -318,6 +306,9 @@ fi
 # If true, will not call $IMPALA_HOME/bin/bootstrap_toolchain.py.
 export SKIP_TOOLCHAIN_BOOTSTRAP=${SKIP_TOOLCHAIN_BOOTSTRAP-false}
 
+# If true, will not download python dependencies.
+export SKIP_PYTHON_DOWNLOAD=${SKIP_PYTHON_DOWNLOAD-false}
+
 # This flag is used in $IMPALA_HOME/cmake_modules/toolchain.cmake.
 # If it's 0, Impala will be built with the compiler in the toolchain directory.
 export USE_SYSTEM_GCC=${USE_SYSTEM_GCC-0}
@@ -344,8 +335,6 @@ export DOWNLOAD_CDH_COMPONENTS=${DOWNLOAD_CDH_COMPONENTS-true}
 
 export IS_OSX="$(if [[ "$OSTYPE" == "darwin"* ]]; then echo true; else echo false; fi)"
 
-export HADOOP_LZO="${HADOOP_LZO-$IMPALA_HOME/../hadoop-lzo}"
-export IMPALA_LZO="${IMPALA_LZO-$IMPALA_HOME/../Impala-lzo}"
 export IMPALA_AUX_TEST_HOME="${IMPALA_AUX_TEST_HOME-$IMPALA_HOME/../Impala-auxiliary-tests}"
 export TARGET_FILESYSTEM="${TARGET_FILESYSTEM-hdfs}"
 export ERASURE_CODING="${ERASURE_CODING-false}"
@@ -360,6 +349,10 @@ export azure_client_secret="${azure_client_secret-DummyAdlsClientSecret}"
 export azure_data_lake_store_name="${azure_data_lake_store_name-}"
 export azure_storage_account_name="${azure_storage_account_name-}"
 export azure_storage_container_name="${azure_storage_container_name-}"
+export GOOGLE_CLOUD_PROJECT_ID="${GOOGLE_CLOUD_PROJECT_ID-}"
+export GOOGLE_CLOUD_SERVICE_ACCOUNT="${GOOGLE_CLOUD_SERVICE_ACCOUNT-}"
+export GOOGLE_APPLICATION_CREDENTIALS="${GOOGLE_APPLICATION_CREDENTIALS-}"
+export GCS_BUCKET="${GCS_BUCKET-}"
 export HDFS_REPLICATION="${HDFS_REPLICATION-3}"
 export ISILON_NAMENODE="${ISILON_NAMENODE-}"
 # Internal and external interfaces that test cluster services will listen on. The
@@ -374,37 +367,19 @@ export LOCAL_FS="file:${WAREHOUSE_LOCATION_PREFIX}"
 export IMPALA_CLUSTER_NODES_DIR="${IMPALA_CLUSTER_NODES_DIR-$IMPALA_HOME/testdata/cluster/cdh$CDH_MAJOR_VERSION}"
 
 ESCAPED_IMPALA_HOME=$(sed "s/[^0-9a-zA-Z]/_/g" <<< "$IMPALA_HOME")
-if $USE_CDP_HIVE; then
-  export HIVE_HOME="$CDP_COMPONENTS_HOME/apache-hive-${IMPALA_HIVE_VERSION}-bin"
-  export HIVE_SRC_DIR=${HIVE_SRC_DIR_OVERRIDE:-"${CDP_COMPONENTS_HOME}/hive-\
-${IMPALA_HIVE_VERSION}"}
-  # Set the path to the hive_metastore.thrift which is used to build thrift code
-  export HIVE_METASTORE_THRIFT_DIR=$HIVE_SRC_DIR/standalone-metastore/src/main/thrift
-  export TEZ_HOME="$CDP_COMPONENTS_HOME/tez-${IMPALA_TEZ_VERSION}-minimal"
-  export HBASE_HOME="$CDP_COMPONENTS_HOME/hbase-${IMPALA_HBASE_VERSION}/"
-  # It is likely that devs will want to work with both the versions of metastore
-  # if cdp hive is being used change the metastore db name, so we don't have to
-  # format the metastore db everytime we switch between hive versions
-  export METASTORE_DB=${METASTORE_DB-"$(cut -c-59 <<< HMS$ESCAPED_IMPALA_HOME)_cdp"}
-else
-  export HIVE_HOME="$CDH_COMPONENTS_HOME/hive-${IMPALA_HIVE_VERSION}"
-  # Allow overriding of Hive source location in case we want to build Impala without
-# a complete Hive build.
-  export HIVE_SRC_DIR=${HIVE_SRC_DIR_OVERRIDE:-"${HIVE_HOME}/src"}
-  export HIVE_METASTORE_THRIFT_DIR=$HIVE_SRC_DIR/metastore/if
-  export HBASE_HOME="$CDH_COMPONENTS_HOME/hbase-${IMPALA_HBASE_VERSION}/"
-  export METASTORE_DB=${METASTORE_DB-$(cut -c-63 <<< HMS$ESCAPED_IMPALA_HOME)}
-fi
+export HIVE_HOME=${HIVE_HOME_OVERRIDE:-"$CDP_COMPONENTS_HOME/apache-hive-${IMPALA_HIVE_VERSION}-bin"}
+export HIVE_SRC_DIR=${HIVE_SRC_DIR_OVERRIDE:-"${CDP_COMPONENTS_HOME}/hive-${IMPALA_HIVE_VERSION}"}
+# Set the path to the hive_metastore.thrift which is used to build thrift code
+export HIVE_METASTORE_THRIFT_DIR=${HIVE_METASTORE_THRIFT_DIR_OVERRIDE:-\
+"$HIVE_SRC_DIR/standalone-metastore/src/main/thrift"}
+export TEZ_HOME="$CDP_COMPONENTS_HOME/tez-${IMPALA_TEZ_VERSION}-minimal"
+export HBASE_HOME="$CDP_COMPONENTS_HOME/hbase-${IMPALA_HBASE_VERSION}/"
+# Previously, there were multiple configurations and the "_cdp" included below
+# allowed the two to be distinct. We keep this "_cdp" for historical reasons.
+export METASTORE_DB=${METASTORE_DB-"$(cut -c-59 <<< HMS$ESCAPED_IMPALA_HOME)_cdp"}
 # Set the Hive binaries in the path
 export PATH="$HIVE_HOME/bin:$PATH"
 
-export SENTRY_POLICY_DB=${SENTRY_POLICY_DB-$(cut -c-63 <<< SP$ESCAPED_IMPALA_HOME)}
-if [[ "${TARGET_FILESYSTEM}" == "s3" ]]; then
-    # On S3, disable Sentry HDFS sync plugin.
-    export SENTRY_PROCESSOR_FACTORIES="org.apache.sentry.api.service.thrift.SentryPolicyStoreProcessorFactory"
-else
-    export SENTRY_PROCESSOR_FACTORIES="org.apache.sentry.api.service.thrift.SentryPolicyStoreProcessorFactory,org.apache.sentry.hdfs.SentryHDFSServiceProcessorFactory"
-fi
 RANGER_POLICY_DB=${RANGER_POLICY_DB-$(cut -c-63 <<< ranger$ESCAPED_IMPALA_HOME)}
 # The DB script in Ranger expects the database name to be in lower case.
 export RANGER_POLICY_DB=$(echo ${RANGER_POLICY_DB} | tr '[:upper:]' '[:lower:]')
@@ -522,6 +497,15 @@ elif [ "${TARGET_FILESYSTEM}" = "abfs" ]; then
   domain="${azure_storage_account_name}.dfs.core.windows.net"
   DEFAULT_FS="abfss://${azure_storage_container_name}@${domain}"
   export DEFAULT_FS
+elif [ "${TARGET_FILESYSTEM}" = "gs" ]; then
+  # Basic error checking
+  if [[ "${GOOGLE_APPLICATION_CREDENTIALS}" = "" ]]; then
+    echo "GOOGLE_APPLICATION_CREDENTIALS should be set to the JSON file that contains
+      your service account key."
+    return 1
+  fi
+  DEFAULT_FS="gs://${GCS_BUCKET}"
+  export DEFAULT_FS
 elif [ "${TARGET_FILESYSTEM}" = "isilon" ]; then
   if [ "${ISILON_NAMENODE}" = "" ]; then
     echo "In order to access the Isilon filesystem, ISILON_NAMENODE"
@@ -555,7 +539,7 @@ elif [ "${TARGET_FILESYSTEM}" = "hdfs" ]; then
   fi
 else
   echo "Unsupported filesystem '$TARGET_FILESYSTEM'"
-  echo "Valid values are: hdfs, isilon, s3, local"
+  echo "Valid values are: hdfs, isilon, s3, abfs, adls, gs, local"
   return 1
 fi
 
@@ -595,8 +579,9 @@ export IMPALA_AUX_WORKLOAD_DIR="$IMPALA_AUX_TEST_HOME/testdata/workloads"
 export IMPALA_DATASET_DIR="$IMPALA_HOME/testdata/datasets"
 export IMPALA_AUX_DATASET_DIR="$IMPALA_AUX_TEST_HOME/testdata/datasets"
 export IMPALA_COMMON_DIR="$IMPALA_HOME/common"
-export PATH="$IMPALA_TOOLCHAIN/gdb-$IMPALA_GDB_VERSION/bin:$PATH"
-export PATH="$IMPALA_HOME/bin:$IMPALA_TOOLCHAIN/cmake-$IMPALA_CMAKE_VERSION/bin/:$PATH"
+export PATH="$IMPALA_TOOLCHAIN_PACKAGES_HOME/gdb-$IMPALA_GDB_VERSION/bin:$PATH"
+export PATH="$IMPALA_TOOLCHAIN_PACKAGES_HOME/cmake-$IMPALA_CMAKE_VERSION/bin/:$PATH"
+export PATH="$IMPALA_HOME/bin:$PATH"
 
 export HADOOP_CONF_DIR="$IMPALA_FE_DIR/src/test/resources"
 # The include and lib paths are needed to pick up hdfs.h and libhdfs.*
@@ -604,25 +589,17 @@ export HADOOP_CONF_DIR="$IMPALA_FE_DIR/src/test/resources"
 export HADOOP_INCLUDE_DIR=${HADOOP_INCLUDE_DIR_OVERRIDE:-"${HADOOP_HOME}/include"}
 export HADOOP_LIB_DIR=${HADOOP_LIB_DIR_OVERRIDE:-"${HADOOP_HOME}/lib"}
 
-# Please note that the * is inside quotes, thus it won't get expanded by bash but
-# by java, see "Understanding class path wildcards" at http://goo.gl/f0cfft
-export HADOOP_CLASSPATH="${HADOOP_CLASSPATH-}:${HADOOP_HOME}/share/hadoop/tools/lib/*"
-# YARN is configured to use LZO so the LZO jar needs to be in the hadoop classpath.
-export LZO_JAR_PATH="$HADOOP_LZO/build/hadoop-lzo-0.4.15.jar"
-HADOOP_CLASSPATH+=":$LZO_JAR_PATH"
-
 # Beware of adding entries from $HADOOP_HOME here, because they can change
 # the order of the classpath, leading to configuration not showing up first.
-HADOOP_CLASSPATH="$LZO_JAR_PATH"
+export HADOOP_CLASSPATH="${HADOOP_CLASSPATH-}"
 # Add the path containing the hadoop-aws jar, which is required to access AWS from the
 # minicluster.
+# Please note that the * is inside quotes, thus it won't get expanded by bash but
+# by java, see "Understanding class path wildcards" at http://goo.gl/f0cfft
 HADOOP_CLASSPATH="${HADOOP_CLASSPATH}:${HADOOP_HOME}/share/hadoop/tools/lib/*"
 
 export PATH="$HADOOP_HOME/bin:$PATH"
 
-export SENTRY_HOME="$CDH_COMPONENTS_HOME/sentry-${IMPALA_SENTRY_VERSION}"
-export SENTRY_CONF_DIR="$IMPALA_HOME/fe/src/test/resources"
-
 export RANGER_HOME="${CDP_COMPONENTS_HOME}/ranger-${IMPALA_RANGER_VERSION}-admin"
 export RANGER_CONF_DIR="$IMPALA_HOME/fe/src/test/resources"
 
@@ -645,11 +622,11 @@ export HIVE_CONF_DIR="$IMPALA_FE_DIR/./src/test/resources"
 # Hive looks for jar files in a single directory from HIVE_AUX_JARS_PATH plus
 # any jars in AUX_CLASSPATH. (Or a list of jars in HIVE_AUX_JARS_PATH.)
 # The Postgres JDBC driver is downloaded by maven when building the frontend.
-# Export the location of Postgres JDBC driver so Sentry can pick it up.
+# Export the location of Postgres JDBC driver so Ranger can pick it up.
 export POSTGRES_JDBC_DRIVER="${IMPALA_FE_DIR}/target/dependency/postgresql-${IMPALA_POSTGRES_JDBC_DRIVER_VERSION}.jar"
 
 export HIVE_AUX_JARS_PATH="$POSTGRES_JDBC_DRIVER"
-export AUX_CLASSPATH="${LZO_JAR_PATH}"
+export AUX_CLASSPATH=""
 ### Tell hive not to use jline
 export HADOOP_USER_CLASSPATH_FIRST=true
 
@@ -689,63 +666,19 @@ fi
 # overall build type) and does not apply when using a local Kudu build.
 export USE_KUDU_DEBUG_BUILD=${USE_KUDU_DEBUG_BUILD-false}
 
-# Kudu doesn't compile on some old Linux distros. KUDU_IS_SUPPORTED enables building Kudu
-# into the backend. We prefer to pull Kudu in from the toolchain, but will fall back to
-# using the CDH Kudu by setting USE_CDH_KUDU to true.
-export USE_CDH_KUDU=${USE_CDH_KUDU-false}
-if [[ -z "${KUDU_IS_SUPPORTED-}" ]]; then
-  if [[ -n "$KUDU_BUILD_DIR" ]]; then
-    KUDU_IS_SUPPORTED=true
-  elif $IS_OSX; then
-    USE_CDH_KUDU=false
-    KUDU_IS_SUPPORTED=false
-  else
-    KUDU_IS_SUPPORTED=true
-    if $USE_CDH_KUDU; then
-      if ! which lsb_release &>/dev/null; then
-        echo Unable to find the 'lsb_release' command. \
-            Please ensure it is available in your PATH. 1>&2
-        return 1
-      fi
-      DISTRO_VERSION="$(lsb_release -sir 2>&1)"
-      if [[ $? -ne 0 ]]; then
-        echo lsb_release command failed, output was: "$DISTRO_VERSION" 1>&2
-        return 1
-      fi
-      # Remove spaces, trim minor versions, and convert to lowercase.
-      DISTRO_VERSION="$(tr -d ' \n' <<< "$DISTRO_VERSION" | cut -d. -f1 | tr "A-Z" "a-z")"
-      if [[ "$DISTRO_VERSION" == "ubuntu14" ]]; then
-        USE_CDH_KUDU=false
-      fi
-    fi
-  fi
-fi
-export KUDU_IS_SUPPORTED
-
-if $USE_CDH_KUDU; then
-  export IMPALA_KUDU_VERSION=${IMPALA_KUDU_VERSION-"1.11.0-cdh6.x-SNAPSHOT"}
-  export IMPALA_KUDU_JAVA_VERSION=${IMPALA_KUDU_JAVA_VERSION-"1.11.0-cdh6.x-SNAPSHOT"}
-  export IMPALA_KUDU_HOME=${CDH_COMPONENTS_HOME}/kudu-$IMPALA_KUDU_VERSION
-  export IMPALA_KUDU_JAVA_HOME=${CDH_COMPONENTS_HOME}/kudu-$IMPALA_KUDU_VERSION
-  # If USE_CDH_KUDU is true, Toolchain Kudu maven repository should be disabled.
-  # We get Kudu Java artifacts from CDH.
-  export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY="file:///non/existing/repo"
-  export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY_ENABLED=false
-else
-  export IMPALA_KUDU_VERSION=${IMPALA_KUDU_VERSION-"4ed0dbbd1"}
-  export IMPALA_KUDU_JAVA_VERSION=${IMPALA_KUDU_JAVA_VERSION-"1.12.0-SNAPSHOT"}
-  export IMPALA_KUDU_HOME=${IMPALA_TOOLCHAIN}/kudu-$IMPALA_KUDU_VERSION
-  export IMPALA_KUDU_JAVA_HOME=${IMPALA_TOOLCHAIN}/kudu-${IMPALA_KUDU_VERSION}/java
-  export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY=\
+export IMPALA_KUDU_VERSION=${IMPALA_KUDU_VERSION-"b5e7362e69"}
+export IMPALA_KUDU_HOME=${IMPALA_TOOLCHAIN_PACKAGES_HOME}/kudu-$IMPALA_KUDU_VERSION
+export IMPALA_KUDU_JAVA_HOME=\
+${IMPALA_TOOLCHAIN_PACKAGES_HOME}/kudu-${IMPALA_KUDU_VERSION}/java
+export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY=\
 "file://${IMPALA_KUDU_JAVA_HOME}/repository"
-  export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY_ENABLED=true
-fi
+export IMPALA_TOOLCHAIN_KUDU_MAVEN_REPOSITORY_ENABLED=true
 
 # Set $THRIFT_HOME to the Thrift directory in toolchain.
-export THRIFT_HOME="${IMPALA_TOOLCHAIN}/thrift-${IMPALA_THRIFT_VERSION}"
+export THRIFT_HOME="${IMPALA_TOOLCHAIN_PACKAGES_HOME}/thrift-${IMPALA_THRIFT_VERSION}"
 
 # ASAN needs a matching version of llvm-symbolizer to symbolize stack traces.
-export ASAN_SYMBOLIZER_PATH="${IMPALA_TOOLCHAIN}/llvm-${IMPALA_LLVM_ASAN_VERSION}/bin/llvm-symbolizer"
+export ASAN_SYMBOLIZER_PATH="${IMPALA_TOOLCHAIN_PACKAGES_HOME}/llvm-${IMPALA_LLVM_ASAN_VERSION}/bin/llvm-symbolizer"
 
 export CLUSTER_DIR="${IMPALA_HOME}/testdata/cluster"
 
@@ -773,39 +706,13 @@ LIBHDFS_OPTS+=" -XX:ErrorFile=${IMPALA_LOGS_DIR}/hs_err_pid%p.log"
 # running tests that load UDF jars.
 LIBHDFS_OPTS="${LIBHDFS_OPTS} -XX:MaxPermSize=128m"
 
-export IMPALA_SNAPPY_PATH="${IMPALA_TOOLCHAIN}/snappy-${IMPALA_SNAPPY_VERSION}/lib"
-
-export JAVA_LIBRARY_PATH="${IMPALA_SNAPPY_PATH}"
-
-# So that the frontend tests and PlanService can pick up required libraries
-LIB_JAVA=`find "${JAVA_HOME}/"   -name libjava.so | head -1`
-LIB_JSIG=`find "${JAVA_HOME}/"   -name libjsig.so | head -1`
-LIB_JVM=` find "${JAVA_HOME}/"   -name libjvm.so  | head -1`
-export LD_LIBRARY_PATH="${LD_LIBRARY_PATH:-}:`dirname ${LIB_JAVA}`:`dirname ${LIB_JSIG}`"
-LD_LIBRARY_PATH="${LD_LIBRARY_PATH}:`dirname ${LIB_JVM}`"
-LD_LIBRARY_PATH="${LD_LIBRARY_PATH}:${HADOOP_LIB_DIR}/native"
-LD_LIBRARY_PATH="${LD_LIBRARY_PATH}:${IMPALA_SNAPPY_PATH}"
-LD_LIBRARY_PATH="${LD_LIBRARY_PATH}:${IMPALA_LZO}/build"
-
-if [ $USE_SYSTEM_GCC -eq 0 ]; then
-  IMPALA_TOOLCHAIN_GCC_LIB="${IMPALA_TOOLCHAIN}/gcc-${IMPALA_GCC_VERSION}/lib64"
-  LD_LIBRARY_PATH="${LD_LIBRARY_PATH}:${IMPALA_TOOLCHAIN_GCC_LIB}"
-fi
-
-export LD_PRELOAD="${LD_PRELOAD-}:${LIB_JSIG}"
-
-export CLASSPATH="$IMPALA_FE_DIR/target/dependency:${CLASSPATH-}"
+export CLASSPATH="$IMPALA_FE_DIR/target/dependency:${CLASSPATH:+:${CLASSPATH}}"
 CLASSPATH="$IMPALA_FE_DIR/target/classes:$CLASSPATH"
 CLASSPATH="$IMPALA_FE_DIR/src/test/resources:$CLASSPATH"
-CLASSPATH="$LZO_JAR_PATH:$CLASSPATH"
 
 # A marker in the environment to prove that we really did source this file
 export IMPALA_CONFIG_SOURCED=1
 
-# Whether to strip the impalad binary when generating deployment package
-export STRIP_DEPLOYMENT_IMPALAD=true
-
-echo "IMPALA_VERSION          = $IMPALA_VERSION"
 echo "IMPALA_HOME             = $IMPALA_HOME"
 echo "HADOOP_HOME             = $HADOOP_HOME"
 echo "HADOOP_CONF_DIR         = $HADOOP_CONF_DIR"
@@ -817,36 +724,31 @@ echo "HIVE_CONF_DIR           = $HIVE_CONF_DIR"
 echo "HIVE_SRC_DIR            = $HIVE_SRC_DIR"
 echo "HBASE_HOME              = $HBASE_HOME"
 echo "HBASE_CONF_DIR          = $HBASE_CONF_DIR"
-echo "SENTRY_HOME             = $SENTRY_HOME"
-echo "SENTRY_CONF_DIR         = $SENTRY_CONF_DIR"
 echo "RANGER_HOME             = $RANGER_HOME"
 echo "RANGER_CONF_DIR         = $RANGER_CONF_DIR "
 echo "THRIFT_HOME             = $THRIFT_HOME"
-echo "HADOOP_LZO              = $HADOOP_LZO"
-echo "IMPALA_LZO              = $IMPALA_LZO"
 echo "CLASSPATH               = $CLASSPATH"
 echo "LIBHDFS_OPTS            = $LIBHDFS_OPTS"
 echo "JAVA_HOME               = $JAVA_HOME"
-echo "LD_LIBRARY_PATH         = $LD_LIBRARY_PATH"
-echo "LD_PRELOAD              = $LD_PRELOAD"
 echo "POSTGRES_JDBC_DRIVER    = $POSTGRES_JDBC_DRIVER"
 echo "IMPALA_TOOLCHAIN        = $IMPALA_TOOLCHAIN"
+echo "IMPALA_TOOLCHAIN_PACKAGES_HOME = $IMPALA_TOOLCHAIN_PACKAGES_HOME"
 echo "METASTORE_DB            = $METASTORE_DB"
 echo "DOWNLOAD_CDH_COMPONENTS = $DOWNLOAD_CDH_COMPONENTS"
 echo "IMPALA_MAVEN_OPTIONS    = $IMPALA_MAVEN_OPTIONS"
 echo "IMPALA_TOOLCHAIN_HOST   = $IMPALA_TOOLCHAIN_HOST"
-echo "CDH_BUILD_NUMBER        = $CDH_BUILD_NUMBER"
-echo "CDH_COMPONENTS_HOME     = $CDH_COMPONENTS_HOME"
 echo "CDP_BUILD_NUMBER        = $CDP_BUILD_NUMBER"
 echo "CDP_COMPONENTS_HOME     = $CDP_COMPONENTS_HOME"
 echo "IMPALA_HADOOP_VERSION   = $IMPALA_HADOOP_VERSION"
+echo "IMPALA_AVRO_JAVA_VERSION= $IMPALA_AVRO_JAVA_VERSION"
+echo "IMPALA_PARQUET_VERSION  = $IMPALA_PARQUET_VERSION"
 echo "IMPALA_HIVE_VERSION     = $IMPALA_HIVE_VERSION"
 echo "IMPALA_HBASE_VERSION    = $IMPALA_HBASE_VERSION"
 echo "IMPALA_HUDI_VERSION     = $IMPALA_HUDI_VERSION"
-echo "IMPALA_SENTRY_VERSION   = $IMPALA_SENTRY_VERSION"
 echo "IMPALA_KUDU_VERSION     = $IMPALA_KUDU_VERSION"
-echo "IMPALA_KUDU_JAVA_VERSION= $IMPALA_KUDU_JAVA_VERSION"
 echo "IMPALA_RANGER_VERSION   = $IMPALA_RANGER_VERSION"
+echo "IMPALA_ICEBERG_VERSION  = $IMPALA_ICEBERG_VERSION"
+echo "IMPALA_GCS_VERSION      = $IMPALA_GCS_VERSION"
 
 # Kerberos things.  If the cluster exists and is kerberized, source
 # the required environment.  This is required for any hadoop tool to
@@ -854,7 +756,9 @@ echo "IMPALA_RANGER_VERSION   = $IMPALA_RANGER_VERSION"
 # kerberized cluster is created, it will have to be sourced again
 # *after* the cluster is created in order to pick up these settings.
 export MINIKDC_ENV="${IMPALA_HOME}/testdata/bin/minikdc_env.sh"
-if "${CLUSTER_DIR}/admin" is_kerberized; then
+if "${CLUSTER_DIR}/admin" is_kerberized ||
+  ( ! "${CLUSTER_DIR}/admin" cluster_exists && [[ "$IMPALA_KERBERIZE" == "true" ]] ); then
+
   . "${MINIKDC_ENV}"
   echo " *** This cluster is kerberized ***"
   echo "KRB5_KTNAME            = $KRB5_KTNAME"
