@@ -26,11 +26,9 @@ import logging
 import multiprocessing
 import os
 import re
-import shutil
 import sqlparse
 import subprocess
 import sys
-import tempfile
 import time
 import traceback
 
@@ -148,27 +146,9 @@ def exec_hive_query_from_file_beeline(file_name):
 
   LOG.info("Beginning execution of hive SQL: {0}".format(file_name))
 
-  # When HiveServer2 is configured to use "local" mode (i.e., MR jobs are run
-  # in-process rather than on YARN), Hadoop's LocalDistributedCacheManager has a
-  # race, wherein it tires to localize jars into
-  # /tmp/hadoop-$USER/mapred/local/<millis>. Two simultaneous Hive queries
-  # against HS2 can conflict here. Weirdly LocalJobRunner handles a similar issue
-  # (with the staging directory) by appending a random number. To over come this,
-  # in the case that HS2 is on the local machine (which we conflate with also
-  # running MR jobs locally), we move the temporary directory into a unique
-  # directory via configuration. This block can be removed when
-  # https://issues.apache.org/jira/browse/MAPREDUCE-6441 is resolved.
-  hive_args = HIVE_ARGS
-  unique_dir = None
-  if options.hive_hs2_hostport.startswith("localhost:"):
-    unique_dir = tempfile.mkdtemp(prefix="hive-data-load-")
-    hive_args += ' --hiveconf "mapreduce.cluster.local.dir=%s"' % unique_dir
-
   output_file = file_name + ".log"
-  hive_cmd = "{0} {1} -f {2}".format(HIVE_CMD, hive_args, file_name)
+  hive_cmd = "{0} {1} -f {2}".format(HIVE_CMD, HIVE_ARGS, file_name)
   is_success = exec_cmd(hive_cmd, exit_on_error=False, out_file=output_file)
-  if unique_dir:
-    shutil.rmtree(unique_dir)
 
   if is_success:
     LOG.info("Finished execution of hive SQL: {0}".format(file_name))
@@ -415,6 +395,7 @@ def main():
 
     impala_create_files = []
     hive_load_text_files = []
+    hive_load_orc_files = []
     hive_load_nontext_files = []
     hbase_create_files = []
     hbase_postload_files = []
@@ -426,6 +407,8 @@ def main():
       elif hive_load_match in filename:
         if 'text-none-none' in filename:
           hive_load_text_files.append(filename)
+        elif 'orc-def-block' in filename:
+          hive_load_orc_files.append(filename)
         else:
           hive_load_nontext_files.append(filename)
       elif hbase_create_match in filename:
@@ -448,6 +431,7 @@ def main():
 
     log_file_list("Impala Create Files:", impala_create_files)
     log_file_list("Hive Load Text Files:", hive_load_text_files)
+    log_file_list("Hive Load Orc Files:", hive_load_orc_files)
     log_file_list("Hive Load Non-Text Files:", hive_load_nontext_files)
     log_file_list("HBase Create Files:", hbase_create_files)
     log_file_list("HBase Post-Load Files:", hbase_postload_files)
@@ -472,6 +456,13 @@ def main():
     # need to be loaded first
     assert(len(hive_load_text_files) <= 1)
     hive_exec_query_files_parallel(thread_pool, hive_load_text_files)
+    # IMPALA-9923: Run ORC serially separately from other non-text formats. This hacks
+    # around flakiness seen when loading this in parallel. This should be removed as
+    # soon as possible.
+    assert(len(hive_load_orc_files) <= 1)
+    hive_exec_query_files_parallel(thread_pool, hive_load_orc_files)
+
+    # Load all non-text formats (goes parallel)
     hive_exec_query_files_parallel(thread_pool, hive_load_nontext_files)
 
     assert(len(hbase_postload_files) <= 1)
