@@ -58,7 +58,7 @@ TESTDATA_ACTION=0
 TESTS_ACTION=1
 FORMAT_CLUSTER=0
 FORMAT_METASTORE=0
-FORMAT_SENTRY_POLICY_DB=0
+UPGRADE_METASTORE_SCHEMA=0
 FORMAT_RANGER_POLICY_DB=0
 NEED_MINICLUSTER=0
 START_IMPALA_CLUSTER=0
@@ -69,7 +69,6 @@ BUILD_ASAN=0
 BUILD_FE_ONLY=0
 BUILD_TESTS=1
 GEN_CMAKE_ONLY=0
-GEN_PACKAGE=0
 BUILD_RELEASE_AND_DEBUG=0
 BUILD_TIDY=0
 BUILD_UBSAN=0
@@ -108,7 +107,6 @@ do
     -format)
       FORMAT_CLUSTER=1
       FORMAT_METASTORE=1
-      FORMAT_SENTRY_POLICY_DB=1
       FORMAT_RANGER_POLICY_DB=1
       ;;
     -format_cluster)
@@ -117,8 +115,8 @@ do
     -format_metastore)
       FORMAT_METASTORE=1
       ;;
-    -format_sentry_policy_db)
-      FORMAT_SENTRY_POLICY_DB=1
+    -upgrade_metastore_db)
+      UPGRADE_METASTORE_SCHEMA=1
       ;;
     -format_ranger_policy_db)
       FORMAT_RANGER_POLICY_DB=1
@@ -199,18 +197,16 @@ do
     -cmake_only)
       GEN_CMAKE_ONLY=1
       ;;
-    -package)
-      GEN_PACKAGE=1
-      ;;
     -help|*)
       echo "buildall.sh - Builds Impala and runs all tests."
       echo "[-noclean] : Omits cleaning all packages before building. Will not kill"\
            "running Hadoop services unless any -format* is True"
-      echo "[-format] : Format the minicluster, metastore db, and sentry policy db"\
+      echo "[-format] : Format the minicluster, metastore db, and ranger policy db"\
            "[Default: False]"
       echo "[-format_cluster] : Format the minicluster [Default: False]"
       echo "[-format_metastore] : Format the metastore db [Default: False]"
-      echo "[-format_sentry_policy_db] : Format the Sentry policy db [Default: False]"
+      echo "[-upgrade_metastore_db] : Upgrades the schema of metastore db"\
+           "[Default: False]"
       echo "[-format_ranger_policy_db] : Format the Ranger policy db [Default: False]"
       echo "[-release_and_debug] : Build both release and debug binaries. Overrides "\
            "other build types [Default: false]"
@@ -251,7 +247,6 @@ do
       echo "[-fe_only] : Build just the frontend"
       echo "[-ninja] : Use ninja instead of make"
       echo "[-cmake_only] : Generate makefiles only, instead of doing a full build"
-      echo "[-package] : Generate a package for deployment."
       echo "-----------------------------------------------------------------------------
 Examples of common tasks:
 
@@ -280,7 +275,10 @@ Examples of common tasks:
   ./buildall.sh -testdata
 
   # Build, format mini-cluster and metastore, load all test data, run tests
-  ./buildall.sh -testdata -format"
+  ./buildall.sh -testdata -format
+
+  # Build and upgrade metastore schema to latest.
+  ./buildall.sh -upgrade_metastore_db"
       exit 1
       ;;
     esac
@@ -359,9 +357,8 @@ if [[ -z "$METASTORE_SNAPSHOT_FILE" && "${TARGET_FILESYSTEM}" != "hdfs" &&
 fi
 
 if [[ $TESTS_ACTION -eq 1 || $TESTDATA_ACTION -eq 1 || $FORMAT_CLUSTER -eq 1 ||
-      $FORMAT_METASTORE -eq 1 || $FORMAT_SENTRY_POLICY_DB -eq 1 ||
-      $FORMAT_RANGER_POLICY_DB -eq 1 || -n "$SNAPSHOT_FILE" ||
-      -n "$METASTORE_SNAPSHOT_FILE" ]]; then
+      $FORMAT_METASTORE -eq 1 || $FORMAT_RANGER_POLICY_DB -eq 1 || -n "$SNAPSHOT_FILE" ||
+      -n "$METASTORE_SNAPSHOT_FILE" || $UPGRADE_METASTORE_SCHEMA -eq 1 ]]; then
   NEED_MINICLUSTER=1
 fi
 
@@ -383,10 +380,8 @@ create_log_dirs() {
 }
 
 bootstrap_dependencies() {
-  # Populate necessary thirdparty components unless it's set to be skipped.
-  if [[ "${SKIP_TOOLCHAIN_BOOTSTRAP}" = true ]]; then
-    echo "SKIP_TOOLCHAIN_BOOTSTRAP is true, skipping download of Python dependencies."
-    echo "SKIP_TOOLCHAIN_BOOTSTRAP is true, skipping toolchain bootstrap."
+  if [[ "${SKIP_PYTHON_DOWNLOAD}" = true ]]; then
+    echo "SKIP_PYTHON_DOWNLOAD is true, skipping python dependencies download."
   else
     echo ">>> Downloading Python dependencies"
     # Download all the Python dependencies we need before doing anything
@@ -397,7 +392,29 @@ bootstrap_dependencies() {
     else
       echo "Finished downloading Python dependencies"
     fi
+  fi
 
+  # Populate necessary thirdparty components unless it's set to be skipped.
+  if [[ "${SKIP_TOOLCHAIN_BOOTSTRAP}" = true ]]; then
+    echo "SKIP_TOOLCHAIN_BOOTSTRAP is true, skipping toolchain bootstrap."
+    if [[ "${DOWNLOAD_CDH_COMPONENTS}" = true ]]; then
+      echo ">>> Downloading and extracting cdh components."
+      "$IMPALA_HOME/bin/bootstrap_toolchain.py"
+    fi
+    # Create soft link to locally builded native-toolchain on aarch64
+    if [[ "$(uname -p)" = "aarch64" ]]; then
+      mkdir -p $IMPALA_TOOLCHAIN_PACKAGES_HOME
+      cd "$IMPALA_TOOLCHAIN_PACKAGES_HOME"
+      ln -f -s ${NATIVE_TOOLCHAIN_HOME}/build/* .
+      cd -
+      if ! [[ -d "$IMPALA_HOME/../hadoopAarch64NativeLibs" ]]; then
+        git clone https://github.com/zhaorenhai/hadoopAarch64NativeLibs \
+          "$IMPALA_HOME/../hadoopAarch64NativeLibs"
+      fi
+      cp $IMPALA_HOME/../hadoopAarch64NativeLibs/lib*  $HADOOP_HOME/lib/native/
+    fi
+
+  else
     echo ">>> Downloading and extracting toolchain dependencies."
     "$IMPALA_HOME/bin/bootstrap_toolchain.py"
     echo "Toolchain bootstrap complete."
@@ -407,7 +424,7 @@ bootstrap_dependencies() {
 # Build the Impala frontend and its dependencies.
 build_fe() {
   generate_cmake_files $CMAKE_BUILD_TYPE
-  ${MAKE_CMD} ${IMPALA_MAKE_FLAGS} fe
+  ${MAKE_CMD} ${IMPALA_MAKE_FLAGS} java
 }
 
 # Build all components. The build type is specified as the first argument, and the
@@ -428,14 +445,10 @@ build_all_components() {
   # If we skip specifying targets, everything we need gets built.
   local MAKE_TARGETS=""
   if [[ $BUILD_TESTS -eq 0 ]]; then
-    # Specify all the non-test targets
-    MAKE_TARGETS="impalad statestored catalogd fesupport loggingsupport ImpalaUdf \
-        udasample udfsample"
     if (( build_independent_targets )); then
-      MAKE_TARGETS+=" cscope fe tarballs"
-    fi
-    if [[ -e "$IMPALA_LZO" ]]; then
-      MAKE_TARGETS+=" impala-lzo"
+      MAKE_TARGETS="notests_all_targets"
+    else
+      MAKE_TARGETS="notests_regular_targets"
     fi
   fi
   ${MAKE_CMD} -j${IMPALA_BUILD_THREADS:-4} ${IMPALA_MAKE_FLAGS} ${MAKE_TARGETS}
@@ -454,9 +467,6 @@ generate_cmake_files() {
   if [[ $BUILD_SHARED_LIBS -eq 1 ]]; then
     CMAKE_ARGS+=(-DBUILD_SHARED_LIBS=ON)
   fi
-  if [[ "$BUILD_TESTS" -eq 0 && "$GEN_PACKAGE" -eq 1 ]]; then
-    CMAKE_ARGS+=(-DBUILD_WITH_NO_TESTS=ON)
-  fi
   if [[ "${MAKE_CMD}" = "ninja" ]]; then
     CMAKE_ARGS+=(-GNinja)
   fi
@@ -470,9 +480,23 @@ generate_cmake_files() {
   else
     CMAKE_ARGS+=(-DCMAKE_TOOLCHAIN_FILE=$IMPALA_HOME/cmake_modules/toolchain.cmake)
   fi
-  if [[ "$GEN_PACKAGE" -eq 1 ]]; then
-    CMAKE_ARGS+=(-DBUILD_PACKAGES=ON)
+
+  # ARM64's L3 cacheline size is different according to CPU vendor's implementations of
+  # architecture. so here we will let user decide this value.
+  # If user defined CACHELINESIZE_AARCH64 in impala-config-local.sh, then we will use that
+  # value, if user did not define it, then we will get the value from OS, if fail, then
+  # we will use the default value 64.
+  if [[ "$(uname -p)" = "aarch64" &&  "$(uname -s)" = "Linux" ]]; then
+    local cachelinesize=$(cat /sys/devices/system/cpu/cpu0/cache/index3/coherency_line_size)
+    if [[ $cachelinesize -gt 0 ]]; then
+      CACHELINESIZE_AARCH64=${CACHELINESIZE_AARCH64-$cachelinesize}
+    else
+      CACHELINESIZE_AARCH64=${CACHELINESIZE_AARCH64-64}
+    fi
+    echo "CACHELINESIZE_AARCH64:$CACHELINESIZE_AARCH64"
+    CMAKE_ARGS+=(-DCACHELINESIZE_AARCH64=$CACHELINESIZE_AARCH64)
   fi
+
   cmake . ${CMAKE_ARGS[@]}
 }
 
@@ -484,8 +508,8 @@ reconfigure_test_cluster() {
   "${IMPALA_HOME}/bin/start-impala-cluster.py" --kill --force
 
   if [[ "$FORMAT_METASTORE" -eq 1 || "$FORMAT_CLUSTER" -eq 1 ||
-        "$FORMAT_SENTRY_POLICY_DB" -eq 1 || "$FORMAT_RANGER_POLICY_DB" -eq 1 ||
-        -n "$METASTORE_SNAPSHOT_FILE" ]]
+        "$FORMAT_RANGER_POLICY_DB" -eq 1 || -n "$METASTORE_SNAPSHOT_FILE" ||
+        "$UPGRADE_METASTORE_SCHEMA" -eq 1 ]]
   then
     # Kill any processes that may be accessing postgres metastore. To be safe, this is
     # done before we make any changes to the config files.
@@ -493,10 +517,6 @@ reconfigure_test_cluster() {
   fi
 
   local CREATE_TEST_CONFIG_ARGS=""
-  if [[ "$FORMAT_SENTRY_POLICY_DB" -eq 1 ]]; then
-    CREATE_TEST_CONFIG_ARGS+=" -create_sentry_policy_db"
-  fi
-
   if [[ "$FORMAT_RANGER_POLICY_DB" -eq 1 ]]; then
     CREATE_TEST_CONFIG_ARGS+=" -create_ranger_policy_db"
   fi
@@ -505,15 +525,12 @@ reconfigure_test_cluster() {
     CREATE_TEST_CONFIG_ARGS+=" -create_metastore"
   fi
 
+  if [[ "$UPGRADE_METASTORE_SCHEMA" -eq 1 ]]; then
+    CREATE_TEST_CONFIG_ARGS+=" -upgrade_metastore_db"
+  fi
+
   # Generate the Hadoop configs needed by Impala
   "${IMPALA_HOME}/bin/create-test-configuration.sh" ${CREATE_TEST_CONFIG_ARGS}
-
-  # Copy Hadoop-lzo dependencies if available (required to generate Lzo data).
-  if stat "$HADOOP_LZO"/build/native/Linux-*-*/lib/libgplcompression.* > /dev/null ; then
-    cp "$HADOOP_LZO"/build/native/Linux-*-*/lib/libgplcompression.* "$HADOOP_LIB_DIR/native"
-  else
-    echo "No hadoop-lzo found"
-  fi
 }
 
 # Starts the test cluster processes except for Impala.
@@ -564,18 +581,14 @@ create_log_dirs
 
 bootstrap_dependencies
 
-# Create .cdh and .cdp files that contains the CDH_BUILD_NUMBER and CDP_BUILD_NUMBER
-# respectively. If the content of the files are different than the ones in the
-# environment variable, append -U into IMPALA_MAVEN_OPTION to force Maven to update its
-# local cache.
-CDH_FILE="${IMPALA_HOME}/.cdh"
+# Create .cdp file that contains the CDP_BUILD_NUMBER. If the content of the files
+# are different than the ones in the environment variable, append -U into
+# IMPALA_MAVEN_OPTION to force Maven to update its local cache.
+# TODO: Look into removing this. The CDP components do not use SNAPSHOT versions.
 CDP_FILE="${IMPALA_HOME}/.cdp"
-if [[ ! -f ${CDH_FILE} || ! -f ${CDP_FILE} || \
-      $(cat ${CDH_FILE}) != ${CDH_BUILD_NUMBER} || \
-      $(cat ${CDP_FILE}) != ${CDP_BUILD_NUMBER} ]]; then
+if [[ ! -f ${CDP_FILE} || $(cat ${CDP_FILE}) != ${CDP_BUILD_NUMBER} ]]; then
   export IMPALA_MAVEN_OPTIONS="${IMPALA_MAVEN_OPTIONS} -U"
 fi
-echo "${CDH_BUILD_NUMBER}" > ${CDH_FILE}
 echo "${CDP_BUILD_NUMBER}" > ${CDP_FILE}
 
 if [[ "$BUILD_FE_ONLY" -eq 1 ]]; then
@@ -598,10 +611,6 @@ else
   build_all_components $CMAKE_BUILD_TYPE 1
 fi
 
-if [[ "$GEN_PACKAGE" -eq 1 ]]; then
-  ${MAKE_CMD} -j${IMPALA_BUILD_THREADS:-4} package
-fi
-
 if [[ $NEED_MINICLUSTER -eq 1 ]]; then
   reconfigure_test_cluster
 fi
