@@ -193,7 +193,7 @@ error_codes = (
    "Try running \\\"refresh $1\\\" to reload the file metadata."),
 
   ("PARQUET_BAD_VERSION_NUMBER", 60, "File '$0' has an invalid Parquet version number: "
-   "$1\\n. Please check that it is a valid Parquet file. "
+   "$1.\\nPlease check that it is a valid Parquet file. "
    "This error can also occur due to stale metadata. "
    "If you believe this is a valid Parquet file, try running \\\"refresh $2\\\"."),
 
@@ -233,7 +233,7 @@ error_codes = (
 
   ("IMPALA_KUDU_TYPE_MISSING", 73, "Impala type $0 is not available in Kudu."),
 
-  ("KUDU_NOT_SUPPORTED_ON_OS", 74, "Kudu is not supported on this operating system."),
+  ("KUDU_NOT_SUPPORTED_ON_OS", 74, "Not in use."),
 
   ("KUDU_NOT_ENABLED", 75, "Kudu features are disabled by the startup flag "
    "--disable_kudu."),
@@ -334,7 +334,7 @@ error_codes = (
   ("ADMISSION_REJECTED", 107, "Rejected query from pool $0: $1"),
 
   ("ADMISSION_TIMED_OUT", 108, "Admission for query exceeded timeout $0ms in pool $1. "
-     "Queued reason: $2"),
+          "Queued reason: $2 Additional Details: $3"),
 
   ("THREAD_CREATION_FAILED", 109, "Failed to create thread $0 in category $1: $2"),
 
@@ -454,6 +454,23 @@ error_codes = (
   ("ORC_TYPE_NOT_ROOT_AT_STRUCT", 148,
    "Root of the $0 type returned by the ORC lib is not STRUCT: $1. "
    "Either there are bugs in the ORC lib or ORC file '$2' is corrupt."),
+
+  ("NAAJ_OUT_OF_MEMORY", 149,
+   "Unable to perform Null-Aware Anti-Join. Could not get enough reservation to fit "
+   "all rows with NULLs from the build side in memory. Memory required for $0 rows "
+   "was $1. $2/$3 of the join's reservation was available for the rows."),
+
+  # Note: impala_shell uses a regex to search for this specific error message, so
+  # changing it may break older shell version.
+  ("INVALID_QUERY_HANDLE", 150, "Invalid or unknown query handle: $0."),
+
+  ("JOIN_ROWS_PRODUCED_LIMIT_EXCEEDED", 151,
+   "Query $0 terminated due to join rows produced exceeds the limit of $1 "
+   "at node with id $2. Unset or increase JOIN_ROWS_PRODUCED_LIMIT query option "
+   "to produce more rows."),
+
+  ("LOCAL_DISK_FAULTY", 152,
+   "Query execution failure caused by local disk IO fatal error on backend: $0."),
 )
 
 import sys
