@@ -25,15 +25,20 @@ import static org.apache.impala.service.MetadataOp.TABLE_TYPE_TABLE;
 import static org.apache.impala.service.MetadataOp.TABLE_TYPE_VIEW;
 
 import com.google.common.annotations.VisibleForTesting;
+import com.google.common.base.Stopwatch;
 import com.google.common.collect.ImmutableMap;
 
+import com.google.common.collect.Iterables;
 import java.net.InetAddress;
 import java.net.UnknownHostException;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.BitSet;
+import java.util.Collections;
 import java.util.EnumSet;
 import java.util.List;
+import java.util.concurrent.TimeUnit;
+import java.util.stream.Collectors;
 
 import org.apache.hadoop.hive.common.StatsSetupConst;
 import org.apache.hadoop.hive.common.ValidReaderWriteIdList;
@@ -52,6 +57,10 @@ import org.apache.hadoop.hive.metastore.api.ColumnStatistics;
 import org.apache.hadoop.hive.metastore.api.ColumnStatisticsObj;
 import org.apache.hadoop.hive.metastore.api.Database;
 import org.apache.hadoop.hive.metastore.api.FieldSchema;
+import org.apache.hadoop.hive.metastore.api.FireEventRequest;
+import org.apache.hadoop.hive.metastore.api.FireEventRequestData;
+import org.apache.hadoop.hive.metastore.api.FireEventResponse;
+import org.apache.hadoop.hive.metastore.api.InsertEventRequestData;
 import org.apache.hadoop.hive.metastore.api.InvalidInputException;
 import org.apache.hadoop.hive.metastore.api.InvalidObjectException;
 import org.apache.hadoop.hive.metastore.api.InvalidOperationException;
@@ -68,6 +77,7 @@ import org.apache.hadoop.hive.metastore.api.SetPartitionsStatsRequest;
 import org.apache.hadoop.hive.metastore.api.Table;
 import org.apache.hadoop.hive.metastore.api.TableValidWriteIds;
 import org.apache.hadoop.hive.metastore.api.TxnAbortedException;
+import org.apache.hadoop.hive.metastore.api.WriteNotificationLogRequest;
 import org.apache.hadoop.hive.metastore.conf.MetastoreConf;
 import org.apache.hadoop.hive.metastore.messaging.AlterTableMessage;
 import org.apache.hadoop.hive.metastore.messaging.EventMessage;
@@ -83,6 +93,7 @@ import org.apache.hive.service.rpc.thrift.TGetFunctionsReq;
 import org.apache.hive.service.rpc.thrift.TGetSchemasReq;
 import org.apache.hive.service.rpc.thrift.TGetTablesReq;
 import org.apache.impala.authorization.User;
+import org.apache.impala.catalog.MetaStoreClientPool.MetaStoreClient;
 import org.apache.impala.common.ImpalaException;
 import org.apache.impala.common.ImpalaRuntimeException;
 import org.apache.impala.common.Pair;
@@ -92,19 +103,23 @@ import org.apache.impala.service.Frontend;
 import org.apache.impala.service.MetadataOp;
 import org.apache.impala.thrift.TMetadataOpRequest;
 import org.apache.impala.thrift.TResultSet;
+import org.apache.impala.thrift.TValidWriteIdList;
 import org.apache.impala.util.AcidUtils;
 import org.apache.impala.util.AcidUtils.TblTransaction;
-import org.apache.log4j.Logger;
+import org.apache.impala.util.MetaStoreUtil.TableInsertEventInfo;
 import org.apache.thrift.TException;
 
 import com.google.common.base.Preconditions;
 import com.google.common.collect.Lists;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
+
 /**
  * A wrapper around some of Hive's Metastore API's to abstract away differences
  * between major versions of Hive. This implements the shimmed methods for Hive 3.
  */
 public class MetastoreShim {
-  private static final Logger LOG = Logger.getLogger(MetastoreShim.class);
+  private static final Logger LOG = LoggerFactory.getLogger(MetastoreShim.class);
 
   private static final String EXTWRITE = "EXTWRITE";
   private static final String EXTREAD = "EXTREAD";
@@ -230,7 +245,7 @@ public class MetastoreShim {
   public static void createTableWithConstraints(IMetaStoreClient client,
       Table newTbl, List<SQLPrimaryKey> primaryKeys, List<SQLForeignKey> foreignKeys)
       throws InvalidOperationException, MetaException, TException {
-    client.createTableWithConstraints(newTbl,primaryKeys, foreignKeys, null, null,
+    client.createTableWithConstraints(newTbl, primaryKeys, foreignKeys, null, null,
         null, null);
   }
 
@@ -461,6 +476,15 @@ public class MetastoreShim {
     return HiveMetadataFormatUtils.getTableInformation(table.getTTable(), false);
   }
 
+  /**
+   * Wrapper method around Hive-3's MetadataFormatUtils.getConstraintsInformation
+   * @return
+   */
+  public static String getConstraintsInformation(
+      org.apache.hadoop.hive.ql.metadata.Table table) {
+    return HiveMetadataFormatUtils.getConstraintsInformation(table);
+  }
+
   /**
    * This method has been copied from BaseSemanticAnalyzer class of Hive and is fairly
    * stable now (last change was in mid 2016 as of April 2019). Copying is preferred over
@@ -588,6 +612,57 @@ public class MetastoreShim {
     return new ValidReaderWriteIdList(validWriteIds);
   }
 
+  /**
+   * Converts a TValidWriteIdList object to ValidWriteIdList.
+   * @param tableName the name of the table.
+   * @param validWriteIds the thrift object.
+   * @return ValidWriteIdList object
+   */
+  public static ValidWriteIdList getValidWriteIdListFromThrift(String tableName,
+      TValidWriteIdList validWriteIds) {
+    Preconditions.checkNotNull(validWriteIds);
+    BitSet abortedBits;
+    if (validWriteIds.getAborted_indexesSize() > 0) {
+      abortedBits = new BitSet(validWriteIds.getInvalid_write_idsSize());
+      for (int aborted_index : validWriteIds.getAborted_indexes()) {
+        abortedBits.set(aborted_index);
+      }
+    } else {
+      abortedBits = new BitSet();
+    }
+    long highWatermark = validWriteIds.isSetHigh_watermark() ?
+        validWriteIds.high_watermark : Long.MAX_VALUE;
+    long minOpenWriteId = validWriteIds.isSetMin_open_write_id() ?
+        validWriteIds.min_open_write_id : Long.MAX_VALUE;
+    return new ValidReaderWriteIdList(tableName,
+        validWriteIds.getInvalid_write_ids().stream().mapToLong(i -> i).toArray(),
+        abortedBits, highWatermark, minOpenWriteId);
+  }
+
+  /**
+   * Converts a ValidWriteIdList object to TValidWriteIdList.
+   */
+  public static TValidWriteIdList convertToTValidWriteIdList(
+      ValidWriteIdList validWriteIdList) {
+    Preconditions.checkNotNull(validWriteIdList);
+    TValidWriteIdList ret = new TValidWriteIdList();
+    long minOpenWriteId = validWriteIdList.getMinOpenWriteId() != null ?
+        validWriteIdList.getMinOpenWriteId() : Long.MAX_VALUE;
+    ret.setHigh_watermark(validWriteIdList.getHighWatermark());
+    ret.setMin_open_write_id(minOpenWriteId);
+    ret.setInvalid_write_ids(Arrays.stream(
+        validWriteIdList.getInvalidWriteIds()).boxed().collect(Collectors.toList()));
+    List<Integer> abortedIndexes = new ArrayList<>();
+    for (int i = 0; i < validWriteIdList.getInvalidWriteIds().length; ++i) {
+      long writeId = validWriteIdList.getInvalidWriteIds()[i];
+      if (validWriteIdList.isWriteIdAborted(writeId)) {
+        abortedIndexes.add(i);
+      }
+    }
+    ret.setAborted_indexes(abortedIndexes);
+    return ret;
+  }
+
   /**
    * Returns a ValidTxnList object that helps to identify in-progress and aborted
    * transactions.
@@ -843,6 +918,8 @@ public class MetastoreShim {
         EXTREAD,  // External table read
         HIVEMANAGEDINSERTREAD, // Insert-only table read
         HIVEMANAGEDINSERTWRITE, // Insert-only table write
+        HIVEFULLACIDREAD,
+        HIVEFULLACIDWRITE,
         HIVESQL,
         HIVEMQT,
         HIVEBUCKET2 // Includes the capability to get the correct bucket number.
@@ -982,4 +1059,116 @@ public class MetastoreShim {
     return wh.getDefaultTablePath(db, tbl.getTableName().toLowerCase(), isExternal)
         .toString();
   }
+
+  /**
+   * Fire insert events for table and partition.
+   * In case of any exception, we just log the failure of firing insert events.
+   */
+  public static List<Long> fireInsertEvents(MetaStoreClient msClient,
+      TableInsertEventInfo insertEventInfo, String dbName, String tableName) {
+    Stopwatch sw = Stopwatch.createStarted();
+    try {
+      if (insertEventInfo.isTransactional()) {
+        // if the table is transactional we use a different API to fire these
+        // events. Note that we don't really need the event ids here for self-event
+        // detection since these events are not fetched by the EventsProcessor later
+        // These events are mostly required for incremental replication in Hive
+        fireInsertTransactionalEventHelper(msClient.getHiveClient(),
+            insertEventInfo, dbName, tableName);
+      } else {
+        return fireInsertEventHelper(msClient.getHiveClient(),
+            insertEventInfo.getInsertEventReqData(), dbName,
+            tableName);
+      }
+    } catch (Exception e) {
+      LOG.error("Failed to fire insert event. Some tables might not be"
+              + " refreshed on other impala clusters.", e);
+    } finally {
+      LOG.info("Time taken to fire insert events on table {}.{}: {} msec", dbName,
+          tableName, sw.stop().elapsed(TimeUnit.MILLISECONDS));
+      msClient.close();
+    }
+    return Collections.emptyList();
+  }
+
+  /**
+   * Fires a listener event of the type ACID_WRITE on a transactional table in metastore.
+   * This event is polled by other external systems to detect insert operations into
+   * ACID tables.
+   * @throws TException in case of errors during HMS API call.
+   */
+  private static void fireInsertTransactionalEventHelper(
+      IMetaStoreClient hiveClient, TableInsertEventInfo insertEventInfo, String dbName,
+      String tableName) throws TException {
+    for (InsertEventRequestData insertData : insertEventInfo.getInsertEventReqData()) {
+      // TODO(Vihang) unfortunately there is no bulk insert event API for transactional
+      // tables. It is possible that this may take long time here if there are lots of
+      // partitions which were inserted.
+      if (LOG.isDebugEnabled()) {
+        String msg =
+            "Firing write notification log request for table " + dbName + "." + tableName
+                + (insertData.isSetPartitionVal() ? " on partition " + insertData
+                .getPartitionVal() : "");
+        LOG.debug(msg);
+      }
+      WriteNotificationLogRequest rqst = new WriteNotificationLogRequest(
+          insertEventInfo.getTxnId(), insertEventInfo.getWriteId(), dbName, tableName,
+          insertData);
+      if (insertData.isSetPartitionVal()) {
+        rqst.setPartitionVals(insertData.getPartitionVal());
+      }
+      // TODO(Vihang) metastore should return the event id here so that we can get rid
+      // of firing INSERT event types for transactional tables.
+      hiveClient.addWriteNotificationLog(rqst);
+    }
+  }
+
+  /**
+   *  Fires an insert event to HMS notification log. In Hive-3 for partitioned table,
+   *  all partition insert events will be fired by a bulk API.
+   *
+   * @param msClient Metastore client,
+   * @param insertEventDataList A list of insert event info encapsulating the information
+   *                            needed to fire insert events.
+   * @param dbName
+   * @param tableName
+   * @return a list of eventIds for the insert events
+   */
+  @VisibleForTesting
+  public static List<Long> fireInsertEventHelper(IMetaStoreClient msClient,
+      List<InsertEventRequestData> insertEventDataList, String dbName, String tableName)
+      throws TException {
+    Preconditions.checkNotNull(msClient);
+    Preconditions.checkNotNull(dbName);
+    Preconditions.checkNotNull(tableName);
+    Preconditions.checkState(!insertEventDataList.isEmpty(), "Atleast one insert event "
+        + "info must be provided.");
+    LOG.debug(String.format(
+        "Firing %s insert event(s) for %s.%s", insertEventDataList.size(), dbName,
+        tableName));
+    FireEventRequestData data = new FireEventRequestData();
+    FireEventRequest rqst = new FireEventRequest(true, data);
+    rqst.setDbName(dbName);
+    rqst.setTableName(tableName);
+    if (insertEventDataList.size() == 1) {
+      InsertEventRequestData insertEventData = Iterables
+          .getOnlyElement(insertEventDataList);
+      if (insertEventData.getPartitionVal() != null) {
+        rqst.setPartitionVals(insertEventData.getPartitionVal());
+      }
+      // single insert event API
+      data.setInsertData(insertEventData);
+    } else {
+      // use bulk insert API
+      data.setInsertDatas(insertEventDataList);
+    }
+    FireEventResponse response = msClient.fireListenerEvent(rqst);
+    if (!response.isSetEventIds()) {
+      LOG.error("FireEventResponse does not have event ids set for table {}.{}. This "
+              + "may cause the table to unnecessarily be refreshed when the insert event "
+              + "is received.", dbName, tableName);
+      return Collections.EMPTY_LIST;
+    }
+    return response.getEventIds();
+  }
 }
