@@ -17,6 +17,7 @@
 
 package org.apache.impala.analysis;
 
+import java.nio.charset.Charset;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.HashMap;
@@ -760,8 +761,8 @@ public class ColumnLineageGraph {
   }
 
   private String getQueryHash(String queryStr) {
-    Hasher hasher = Hashing.md5().newHasher();
-    hasher.putString(queryStr);
+    Hasher hasher = Hashing.murmur3_128().newHasher();
+    hasher.putUnencodedChars(queryStr);
     return hasher.hash().toString();
   }
 
