@@ -27,6 +27,7 @@ import org.apache.hadoop.hive.metastore.api.SQLForeignKey;
 import org.apache.hadoop.hive.metastore.api.SQLPrimaryKey;
 import org.apache.impala.authorization.AuthorizationConfig;
 import org.apache.impala.catalog.KuduTable;
+import org.apache.impala.catalog.IcebergTable;
 import org.apache.impala.catalog.RowFormat;
 import org.apache.impala.catalog.Table;
 import org.apache.impala.common.AnalysisException;
@@ -35,11 +36,14 @@ import org.apache.impala.common.RuntimeEnv;
 import org.apache.impala.service.BackendConfig;
 import org.apache.impala.thrift.TCreateTableParams;
 import org.apache.impala.thrift.THdfsFileFormat;
+import org.apache.impala.thrift.TIcebergCatalog;
+import org.apache.impala.thrift.TIcebergPartitionTransformType;
 import org.apache.impala.thrift.TSortingOrder;
 import org.apache.impala.thrift.TTableName;
 import org.apache.impala.util.AvroSchemaConverter;
 import org.apache.impala.util.AvroSchemaParser;
 import org.apache.impala.util.AvroSchemaUtils;
+import org.apache.impala.util.IcebergUtil;
 import org.apache.impala.util.KuduUtil;
 import org.apache.impala.util.MetaStoreUtil;
 
@@ -139,6 +143,10 @@ public class CreateTableStmt extends StatementBase {
     return tableDef_.getPrimaryKeyColumnNames();
   }
 
+  public List<IcebergPartitionSpec> getIcebergPartitionSpecs() {
+    return tableDef_.getIcebergPartitionSpecs();
+  }
+
   /**
    * Get foreign keys information as strings. Useful for toSqlUtils.
    * @return List of strings of the form "(col1, col2,..) REFERENCES [pk_db].pk_table
@@ -218,6 +226,13 @@ public class CreateTableStmt extends StatementBase {
       params.addToForeign_keys(fk);
     }
     params.setServer_name(serverName_);
+
+    // Create table stmt only have one PartitionSpec
+    if (!getIcebergPartitionSpecs().isEmpty()) {
+      Preconditions.checkState(getIcebergPartitionSpecs().size() == 1);
+      params.setPartition_spec(getIcebergPartitionSpecs().get(0).toThrift());
+    }
+
     return params;
   }
 
@@ -243,7 +258,8 @@ public class CreateTableStmt extends StatementBase {
     // Avro tables can have empty column defs because they can infer them from the Avro
     // schema. Likewise for external Kudu tables, the schema can be read from Kudu.
     if (getColumnDefs().isEmpty() && getFileFormat() != THdfsFileFormat.AVRO
-        && getFileFormat() != THdfsFileFormat.KUDU) {
+        && getFileFormat() != THdfsFileFormat.KUDU && getFileFormat() !=
+        THdfsFileFormat.ICEBERG) {
       throw new AnalysisException("Table requires at least 1 column");
     }
     if (getFileFormat() == THdfsFileFormat.AVRO) {
@@ -255,6 +271,17 @@ public class CreateTableStmt extends StatementBase {
       AvroSchemaUtils.setFromSerdeComment(getColumnDefs());
     }
 
+    if (getFileFormat() == THdfsFileFormat.ICEBERG) {
+      analyzeIcebergColumns();
+      analyzeIcebergFormat(analyzer);
+    } else {
+      List<IcebergPartitionSpec> iceSpec = tableDef_.getIcebergPartitionSpecs();
+      if (iceSpec != null && !iceSpec.isEmpty()) {
+        throw new AnalysisException(
+            "PARTITION BY SPEC is only valid for Iceberg tables.");
+      }
+    }
+
     // If lineage logging is enabled, compute minimal lineage graph.
     if (BackendConfig.INSTANCE.getComputeLineage() || RuntimeEnv.INSTANCE.isTestEnv()) {
        computeLineageGraph(analyzer);
@@ -305,7 +332,7 @@ public class CreateTableStmt extends StatementBase {
   private void analyzeKuduTableProperties(Analyzer analyzer) throws AnalysisException {
     AuthorizationConfig authzConfig = analyzer.getAuthzConfig();
     if (authzConfig.isEnabled()) {
-      // Today there is no comprehensive way of enforcing a Sentry authorization policy
+      // Today there is no comprehensive way of enforcing a Ranger authorization policy
       // against tables stored in Kudu. This is why only users with ALL privileges on
       // SERVER may create external Kudu tables or set the master addresses.
       // See IMPALA-4000 for details.
@@ -542,4 +569,170 @@ public class CreateTableStmt extends StatementBase {
           new StringLiteral(kv.getValue()).getUnescapedValue());
     }
   }
+
+  /**
+   * For iceberg file format, add related storage handler
+   */
+  private void analyzeIcebergFormat(Analyzer analyzer) throws AnalysisException {
+    // A managed table cannot have 'external.table.purge' property set
+    if (!isExternal() && Boolean.parseBoolean(
+        getTblProperties().get(IcebergTable.TBL_PROP_EXTERNAL_TABLE_PURGE))) {
+      throw new AnalysisException(String.format("Table property '%s' cannot be set to " +
+          "true with a managed Iceberg table.",
+          IcebergTable.TBL_PROP_EXTERNAL_TABLE_PURGE));
+    }
+
+    // Check for managed table
+    if (!isExternal() || Boolean.parseBoolean(getTblProperties().get(
+        Table.TBL_PROP_EXTERNAL_TABLE_PURGE))) {
+      if (getColumnDefs().isEmpty()) {
+        // External iceberg table can have empty column, but managed iceberg table
+        // requires at least one column.
+        throw new AnalysisException("Table requires at least 1 column for " +
+            "managed iceberg table.");
+      }
+
+      // Check partition columns for managed iceberg table
+      checkPartitionColumns(analyzer);
+    }
+
+    String handler = getTblProperties().get(IcebergTable.KEY_STORAGE_HANDLER);
+    if (handler != null && !IcebergTable.isIcebergStorageHandler(handler)) {
+      throw new AnalysisException("Invalid storage handler " +
+          "specified for Iceberg format: " + handler);
+    }
+    putGeneratedKuduProperty(IcebergTable.KEY_STORAGE_HANDLER,
+        IcebergTable.ICEBERG_STORAGE_HANDLER);
+
+    String fileformat = getTblProperties().get(IcebergTable.ICEBERG_FILE_FORMAT);
+    if (fileformat != null && IcebergUtil.getIcebergFileFormat(fileformat) == null) {
+      throw new AnalysisException("Invalid fileformat for Iceberg table: " + fileformat);
+    }
+    if (fileformat == null || fileformat.isEmpty()) {
+      putGeneratedKuduProperty(IcebergTable.ICEBERG_FILE_FORMAT, "parquet");
+    }
+
+    // Determine the Iceberg catalog being used. The default catalog is HiveCatalog.
+    String catalogStr = getTblProperties().get(IcebergTable.ICEBERG_CATALOG);
+    TIcebergCatalog catalog;
+    if (catalogStr == null || catalogStr.isEmpty()) {
+      catalog = TIcebergCatalog.HIVE_CATALOG;
+      putGeneratedKuduProperty(IcebergTable.ICEBERG_CATALOG, "hive.catalog");
+    } else {
+      catalog = IcebergUtil.getTIcebergCatalog(catalogStr);
+    }
+    validateIcebergTableProperties(catalog);
+  }
+
+  private void validateIcebergTableProperties(TIcebergCatalog catalog)
+      throws AnalysisException {
+    // Metadata location is only used by HiveCatalog, but we shouldn't allow setting this
+    // for any catalogs to avoid confusion.
+    if (getTblProperties().get(IcebergTable.METADATA_LOCATION) != null) {
+      throw new AnalysisException(String.format("%s cannot be set for Iceberg tables",
+          IcebergTable.METADATA_LOCATION));
+    }
+    switch(catalog) {
+      case HIVE_CATALOG: validateTableInHiveCatalog();
+      break;
+      case HADOOP_CATALOG: validateTableInHadoopCatalog();
+      break;
+      case HADOOP_TABLES: validateTableInHadoopTables();
+      break;
+      default: throw new AnalysisException(String.format(
+          "Unknown Iceberg catalog type: %s", catalog));
+    }
+  }
+
+  private void validateTableInHiveCatalog() throws AnalysisException {
+    if (getTblProperties().get(IcebergTable.ICEBERG_CATALOG_LOCATION) != null) {
+      throw new AnalysisException(String.format("%s cannot be set for Iceberg table " +
+          "stored in hive.catalog", IcebergTable.ICEBERG_CATALOG_LOCATION));
+    }
+  }
+
+  private void validateTableInHadoopCatalog() throws AnalysisException {
+    // Table location cannot be set in SQL when using 'hadoop.catalog'
+    if (getLocation() != null) {
+      throw new AnalysisException(String.format("Location cannot be set for Iceberg " +
+          "table with 'hadoop.catalog'."));
+    }
+    String catalogLoc = getTblProperties().get(IcebergTable.ICEBERG_CATALOG_LOCATION);
+    if (catalogLoc == null || catalogLoc.isEmpty()) {
+      throw new AnalysisException(String.format("Table property '%s' is necessary " +
+          "for Iceberg table with 'hadoop.catalog'.",
+          IcebergTable.ICEBERG_CATALOG_LOCATION));
+    }
+  }
+
+  private void validateTableInHadoopTables() throws AnalysisException {
+    if (getTblProperties().get(IcebergTable.ICEBERG_CATALOG_LOCATION) != null) {
+      throw new AnalysisException(String.format("%s cannot be set for Iceberg table " +
+          "stored in hadoop.tables", IcebergTable.ICEBERG_CATALOG_LOCATION));
+    }
+  }
+
+  /**
+   * For iceberg table, partition column must be from source column
+   */
+  private void checkPartitionColumns(Analyzer analyzer) throws AnalysisException {
+    // This check is unnecessary for iceberg table without partition spec
+    List<IcebergPartitionSpec> specs = tableDef_.getIcebergPartitionSpecs();
+    if (specs == null || specs.isEmpty()) return;
+
+    // Iceberg table only has one partition spec now
+    IcebergPartitionSpec spec = specs.get(0);
+    // Analyzes the partition spec and the underlying partition fields.
+    spec.analyze(analyzer);
+
+    List<IcebergPartitionField> fields = spec.getIcebergPartitionFields();
+    Preconditions.checkState(fields != null && !fields.isEmpty());
+    for (IcebergPartitionField field : fields) {
+      String fieldName = field.getFieldName();
+      boolean containFlag = false;
+      for (ColumnDef columnDef : tableDef_.getColumnDefs()) {
+        if (columnDef.getColName().equalsIgnoreCase(fieldName)) {
+          containFlag = true;
+          break;
+        }
+      }
+      if (!containFlag) {
+        throw new AnalysisException("Cannot find source column: " + fieldName);
+      }
+    }
+  }
+
+  /**
+   * Set column's nullable as true for default situation, so we can create optional
+   * Iceberg field
+   */
+  private void analyzeIcebergColumns() {
+    if (!getPartitionColumnDefs().isEmpty()) {
+      createIcebergPartitionSpecFromPartitionColumns();
+    }
+    for (ColumnDef def : getColumnDefs()) {
+      if (!def.isNullabilitySet()) {
+        def.setNullable(true);
+      }
+    }
+  }
+
+  /**
+   * Creates Iceberg partition spec from partition columns. Needed to support old-style
+   * CREATE TABLE .. PARTITIONED BY (<cols>) syntax. In this case the column list in
+   * 'cols' is appended to the table-level columns, but also Iceberg-level IDENTITY
+   * partitions are created from this list.
+   */
+  private void createIcebergPartitionSpecFromPartitionColumns() {
+    Preconditions.checkState(!getPartitionColumnDefs().isEmpty());
+    Preconditions.checkState(getIcebergPartitionSpecs().isEmpty());
+    List<IcebergPartitionField> partFields = new ArrayList<>();
+    for (ColumnDef colDef : getPartitionColumnDefs()) {
+      partFields.add(new IcebergPartitionField(colDef.getColName(),
+          new IcebergPartitionTransform(TIcebergPartitionTransformType.IDENTITY)));
+    }
+    getIcebergPartitionSpecs().add(new IcebergPartitionSpec(partFields));
+    getColumnDefs().addAll(getPartitionColumnDefs());
+    getPartitionColumnDefs().clear();
+  }
 }
