@@ -50,7 +50,7 @@ import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
 import com.google.common.base.Joiner;
-import com.google.common.base.Objects;
+import com.google.common.base.MoreObjects;
 import com.google.common.base.Preconditions;
 import com.google.common.base.Predicates;
 import com.google.common.collect.Iterables;
@@ -83,7 +83,6 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
   public static final float ARITHMETIC_OP_COST = 1;
   public static final float BINARY_PREDICATE_COST = 1;
   public static final float VAR_LEN_BINARY_PREDICATE_COST = 5;
-  public static final float CAST_COST = 1;
   public static final float COMPOUND_PREDICATE_COST = 1;
   public static final float FUNCTION_CALL_COST = 10;
   public static final float IS_NOT_EMPTY_COST = 1;
@@ -140,6 +139,18 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
         }
       };
 
+  // Returns true if an Expr has a subquery as a direct child.
+  public static final com.google.common.base.Predicate<Expr> HAS_SUBQUERY_CHILD =
+      new com.google.common.base.Predicate<Expr>() {
+        @Override
+        public boolean apply(Expr arg) {
+          for (Expr child : arg.getChildren()) {
+            if (child instanceof Subquery) return true;
+          }
+          return false;
+        }
+      };
+
   // Returns true if an Expr is an aggregate function that returns non-null on
   // an empty set (e.g. count).
   public static final com.google.common.base.Predicate<Expr>
@@ -305,6 +316,53 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
       }
     };
 
+  public static final com.google.common.base.Predicate<Expr>
+    IS_CONDITIONAL_BUILTIN_FN_PREDICATE =
+    new com.google.common.base.Predicate<Expr>() {
+      @Override
+      public boolean apply(Expr arg) {
+        return arg instanceof FunctionCallExpr
+            && ((FunctionCallExpr) arg).isConditionalBuiltinFn();
+      }
+    };
+
+  public static final com.google.common.base.Predicate<Expr> IS_IS_NULL_PREDICATE =
+    new com.google.common.base.Predicate<Expr>() {
+      @Override
+      public boolean apply(Expr arg) {
+        return arg instanceof IsNullPredicate
+            && !((IsNullPredicate) arg).isNotNull();
+      }
+    };
+
+  public static final com.google.common.base.Predicate<Expr>
+    IS_DISTINCT_FROM_OR_NOT_DISTINCT_PREDICATE =
+      new com.google.common.base.Predicate<Expr>() {
+        @Override
+        public boolean apply(Expr arg) {
+          return arg instanceof BinaryPredicate
+              && (((BinaryPredicate) arg).getOp() == Operator.DISTINCT_FROM
+              || ((BinaryPredicate) arg).getOp() == Operator.NOT_DISTINCT);
+        }
+      };
+
+  public static final com.google.common.base.Predicate<Expr> IS_CASE_EXPR_PREDICATE =
+      new com.google.common.base.Predicate<Expr>() {
+        @Override
+        public boolean apply(Expr arg) {
+          return arg instanceof CaseExpr;
+        }
+      };
+
+  public static final com.google.common.base.Predicate<Expr> IS_ALWAYS_TRUE_PREDICATE =
+      new com.google.common.base.Predicate<Expr>() {
+        @Override
+        public boolean apply(Expr arg) {
+          return arg instanceof Predicate
+              && ((Predicate) arg).hasAlwaysTrueHint();
+        }
+      };
+
   // id that's unique across the entire query statement and is assigned by
   // Analyzer.registerConjuncts(); only assigned for the top-level terms of a
   // conjunction, and therefore null for most Exprs
@@ -352,6 +410,9 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
   // True if this has already been counted towards the number of statement expressions
   private boolean isCountedForNumStmtExprs_ = false;
 
+  // For exprs of type Predicate, this keeps track of predicate hints
+  protected List<PlanHint> predicateHints_;
+
   protected Expr() {
     type_ = Type.INVALID;
     selectivity_ = -1.0;
@@ -376,6 +437,10 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     fn_ = other.fn_;
     isCountedForNumStmtExprs_ = other.isCountedForNumStmtExprs_;
     children_ = Expr.cloneList(other.children_);
+    if (other.predicateHints_ != null) {
+      predicateHints_ = new ArrayList<>();
+      predicateHints_.addAll(other.predicateHints_);
+    }
   }
 
   public boolean isAnalyzed() { return isAnalyzed_; }
@@ -440,6 +505,29 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     analysisDone();
   }
 
+  protected void analyzeHints(Analyzer analyzer) throws AnalysisException {
+    if (predicateHints_ != null && !predicateHints_.isEmpty()) {
+      if (!(this instanceof Predicate)) {
+        throw new AnalysisException("Expr hints are only supported for predicates");
+      }
+      for (PlanHint hint : predicateHints_) {
+        if (hint.is("ALWAYS_TRUE")) {
+          ((Predicate) this).setHasAlwaysTrueHint(true);
+          // If the top level expr has always_true hint, its conjuncts must
+          // also have the same hint (note that there's a TODO in the parser
+          // grammar to allow hints on a per expr basis).
+          List<Expr> conjuncts = getConjuncts();
+          if (conjuncts.size() > 1) {
+            for (Expr e : conjuncts) ((Predicate) e).setHasAlwaysTrueHint(true);
+          }
+          analyzer.setHasPlanHints();
+        } else {
+          analyzer.addWarning("Predicate hint not recognized: " + hint);
+        }
+      }
+    }
+  }
+
   /**
    * Does subclass-specific analysis. Subclasses should override analyzeImpl().
    */
@@ -961,6 +1049,17 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     return list;
   }
 
+  /**
+   * Returns true if this expression is trivially true. Currently we only check if
+   * it is comprised of conjuncts that are all TRUE literals.
+   */
+  public boolean isTriviallyTrue() {
+    for (Expr conjunct : getConjuncts()) {
+      if (!Expr.IS_TRUE_LITERAL.apply(conjunct)) return false;
+    }
+    return true;
+  }
+
   /**
    * Returns an analyzed clone of 'this' with exprs substituted according to smap.
    * Removes implicit casts and analysis state while cloning/substituting exprs within
@@ -1098,6 +1197,38 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     return result;
   }
 
+  /**
+   * Create a deep copy of 'ls'. The elements of the returned list are of the same
+   * type as the input list.
+   */
+  public static <C extends Expr> List<List<C>> deepCopy(List<List<C>> ls) {
+    Preconditions.checkNotNull(ls);
+    List<List<C>> result = new ArrayList<>(ls.size());
+    for (List<C> l : ls) {
+      if (l == null) {
+        result.add(null);
+        continue;
+      }
+      List<C> l2 = new ArrayList<>(l.size());
+      for (Expr element : l) {
+        l2.add((C) element.clone());
+      }
+      result.add(l2);
+    }
+    return result;
+  }
+
+  /**
+   * Create a clone of the expression including analysis state with a different
+   * selectivity.
+   */
+  public Expr cloneAndOverrideSelectivity(double selectivity) {
+    Preconditions.checkArgument(selectivity >= 0.0 && selectivity <= 1.0, selectivity);
+    Expr e = clone();
+    e.selectivity_ = selectivity;
+    return e;
+  }
+
   /**
    * Removes duplicate exprs (according to equals()).
    */
@@ -1140,36 +1271,28 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
   }
 
   /**
-   * Propagates constant expressions of the form <slot ref> = <constant> to
+   * Propagates constant expressions of the form <slot ref> = <constant> or,
+   * for certain data types, <slot ref> range_operator <constant> to
    * other uses of slot ref in the given conjuncts; returns a BitSet with
    * bits set to true in all changed indices.  Only one round of substitution
    * is performed.  The candidates BitSet is used to determine which members of
-   * conjuncts are considered for propagation.
+   * conjuncts are considered for propagation. The keepConjuncts list is
+   * populated in specific cases (e.g for date/time range predicate propagation)
+   * with the original conjuncts that need to be preserved even after rewrite.
    */
   private static BitSet propagateConstants(List<Expr> conjuncts, BitSet candidates,
-      Analyzer analyzer) {
+      List<Expr> keepConjuncts, Analyzer analyzer) {
     Preconditions.checkState(conjuncts.size() <= candidates.size());
+
+    // first pass: gather the constant predicates into separate buckets
+    // for range and equality predicates
+    ConstantPredicateHandler handler = new ConstantPredicateHandler();
+    handler.classifyPredicates(conjuncts, candidates);
+
+    // second pass: propagate constants to the other predicates and keep track
+    // of which other predicates have been changed
     BitSet changed = new BitSet(conjuncts.size());
-    for (int i = candidates.nextSetBit(0); i >= 0; i = candidates.nextSetBit(i+1)) {
-      if (!(conjuncts.get(i) instanceof BinaryPredicate)) continue;
-      BinaryPredicate bp = (BinaryPredicate) conjuncts.get(i);
-      if (bp.getOp() != BinaryPredicate.Operator.EQ) continue;
-      SlotRef slotRef = bp.getBoundSlot();
-      if (slotRef == null || !bp.getChild(1).isConstant()) continue;
-      Expr subst = bp.getSlotBinding(slotRef.getSlotId());
-      ExprSubstitutionMap smap = new ExprSubstitutionMap();
-      smap.put(slotRef, subst);
-      for (int j = 0; j < conjuncts.size(); ++j) {
-        // Don't rewrite with our own substitution!
-        if (j == i) continue;
-        Expr toRewrite = conjuncts.get(j);
-        Expr rewritten = toRewrite.substitute(smap, analyzer, true);
-        if (!rewritten.equals(toRewrite)) {
-          conjuncts.set(j, rewritten);
-          changed.set(j, true);
-        }
-      }
-    }
+    handler.propagateConstantPreds(conjuncts, changed, keepConjuncts, analyzer);
     return changed;
   }
 
@@ -1183,6 +1306,7 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
   public static boolean optimizeConjuncts(List<Expr> conjuncts, Analyzer analyzer) {
     Preconditions.checkNotNull(conjuncts);
     List<Expr> tmpConjuncts = new ArrayList<>();
+    List<Expr> keepConjuncts = new ArrayList<>();
     try {
       BitSet candidates = new BitSet(conjuncts.size());
       candidates.set(0, Math.min(conjuncts.size(), CONST_PROPAGATION_EXPR_LIMIT));
@@ -1195,7 +1319,8 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
         // change the content of the first input param. We do not want to
         // change the expr in conjucts before make sure the constant propagation
         // does not cause analysis failures.
-        BitSet changed = propagateConstants(tmpConjuncts, candidates, analyzer);
+        BitSet changed = propagateConstants(tmpConjuncts, candidates, keepConjuncts,
+            analyzer);
         candidates.clear();
         int pruned = 0;
         for (int i = changed.nextSetBit(0); i >= 0; i = changed.nextSetBit(i+1)) {
@@ -1235,6 +1360,8 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
       LOG.warn("Not able to analyze after rewrite: " + e.toString() + " conjuncts: " +
           Expr.debugString(conjuncts));
     }
+    // add the conjuncts that need to be preserved
+    conjuncts.addAll(keepConjuncts);
     Expr.removeDuplicates(conjuncts);
     return true;
   }
@@ -1273,6 +1400,11 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     return null;
   }
 
+  /**
+   * Find all unique slot and/or tuple ids referenced by this expr tree.
+   * @param tupleIds unique tuple IDs from this expr tree are appended here.
+   * @param slotIds unique slot IDs from this expr tree are appended here.
+   */
   public void getIds(List<TupleId> tupleIds, List<SlotId> slotIds) {
     Set<TupleId> tupleIdSet = new HashSet<>();
     Set<SlotId> slotIdSet = new HashSet<>();
@@ -1295,6 +1427,19 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     }
   }
 
+  /**
+   * Returns true if this expression tree references a slot in the tuple identified
+   * by tid.
+   */
+  public boolean referencesTuple(TupleId tid) {
+    // This is the default implementation. Expr subclasses that reference slots in
+    // tuples, i.e. SlotRef, must override this.
+    for (Expr child: children_) {
+      if (child.referencesTuple(tid)) return true;
+    }
+    return false;
+  }
+
   /**
    * Returns true if this expression should be treated as constant. I.e. if the frontend
    * and backend should assume that two evaluations of the expression within a query will
@@ -1343,7 +1488,7 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     if (!(this instanceof Subquery)) return false;
     Subquery subq = (Subquery) this;
     SelectStmt stmt = (SelectStmt) subq.getStatement();
-    return stmt.returnsSingleRow() && getType().isScalarType();
+    return stmt.returnsAtMostOneRow() && getType().isScalarType();
   }
 
   /**
@@ -1449,7 +1594,7 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
 
   @Override
   public String toString() {
-    return Objects.toStringHelper(this.getClass())
+    return MoreObjects.toStringHelper(this.getClass())
         .add("id", id_)
         .add("type", type_)
         .add("toSql", toSql(ToSqlOptions.SHOW_IMPLICIT_CASTS))
@@ -1470,11 +1615,14 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
   }
 
   /**
-   * Returns the first child if this Expr is a CastExpr. Otherwise, returns 'this'.
+   * Returns the first child if this Expr is a CastExpr or builtin cast
+   * function. Otherwise, returns 'this'.
    */
   public Expr unwrapExpr(boolean implicitOnly) {
-    if (this instanceof CastExpr
-        && (!implicitOnly || ((CastExpr) this).isImplicit())) {
+    if ((this instanceof CastExpr
+        && (!implicitOnly || ((CastExpr) this).isImplicit()))
+        || (this instanceof FunctionCallExpr
+            && ((FunctionCallExpr) this).isBuiltinCastFunction())) {
       return children_.get(0);
     }
     return this;
@@ -1683,4 +1831,28 @@ abstract public class Expr extends TreeNode<Expr> implements ParseNode, Cloneabl
     }
     return value;
   }
+
+  public void setPredicateHints(List<PlanHint> hints) {
+    Preconditions.checkNotNull(hints);
+    predicateHints_ = hints;
+  }
+
+  public List<PlanHint> getPredicateHints() { return predicateHints_; }
+
+  // A wrapper method to create null literal. This can be overriden
+  // in a derived class
+  protected Expr createNullLiteral() {
+    return new NullLiteral();
+  }
+
+  /**
+   * Subclass that contains query statements, e.g SubQuery, should override this.
+   */
+  public boolean resolveTableMask(Analyzer analyzer) throws AnalysisException {
+    boolean hasChanges = false;
+    for (Expr child : children_) {
+      hasChanges |= child.resolveTableMask(analyzer);
+    }
+    return hasChanges;
+  }
 }
