@@ -20,7 +20,9 @@ package org.apache.impala.analysis;
 import java.util.ArrayList;
 import java.util.List;
 import java.util.Set;
+import java.util.stream.Collectors;
 
+import org.apache.impala.authorization.AuthorizationContext;
 import org.apache.impala.authorization.TableMask;
 import org.apache.impala.catalog.Column;
 import org.apache.impala.catalog.ColumnStats;
@@ -141,31 +143,35 @@ public class InlineViewRef extends TableRef {
    * Creates an inline-view doing table masking for column masking and row filtering
    * policies. Callers should replace 'tableRef' with the returned view.
    *
-   * @param resolvedPath resolved path for the original table/view
    * @param tableRef original resolved table/view
    * @param tableMask TableMask providing column masking and row filtering policies
+   * @param authzCtx AuthorizationContext containing RangerBufferAuditHandler
    */
-  static InlineViewRef createTableMaskView(Path resolvedPath, TableRef tableRef,
-      TableMask tableMask) throws AnalysisException, InternalException {
-    Preconditions.checkNotNull(resolvedPath);
-    Preconditions.checkNotNull(resolvedPath.getRootTable());
+  static InlineViewRef createTableMaskView(TableRef tableRef, TableMask tableMask,
+      AuthorizationContext authzCtx) throws AnalysisException, InternalException {
     Preconditions.checkNotNull(tableRef);
+    Preconditions.checkNotNull(authzCtx);
     Preconditions.checkState(tableRef instanceof InlineViewRef
         || tableRef instanceof BaseTableRef);
-    List<Column> columns = resolvedPath.getRootTable().getColumnsInHiveOrder();
+    List<Column> columns = tableMask.getRequiredColumns();
     List<SelectListItem> items = Lists.newArrayListWithCapacity(columns.size());
     for (Column col: columns) {
-      if (col.getType().isComplexType()) continue;
-      // TODO: only add materialized columns to avoid introducing new privilege
-      //  requirements (IMPALA-9223)
+      Preconditions.checkState(!col.getType().isComplexType(),
+          "Complex type columns should not be registered to table mask");
       items.add(new SelectListItem(
-          tableMask.createColumnMask(col.getName(), col.getType()),
+          tableMask.createColumnMask(col.getName(), col.getType(), authzCtx),
           /*alias*/ col.getName()));
     }
+    if (columns.isEmpty()) {
+      // No columns so use "SELECT 1 FROM tbl" to make a valid statement.
+      items.add(new SelectListItem(NumericLiteral.create(1), /*alias*/null));
+    }
     SelectList selectList = new SelectList(items);
     FromClause fromClause = new FromClause(Lists.newArrayList(tableRef));
-    SelectStmt tableMaskStmt = new SelectStmt(selectList, fromClause,
-        null, null, null, null, null);
+    Expr wherePredicate = tableMask.createRowFilter(authzCtx);
+    SelectStmt tableMaskStmt = new SelectStmt(selectList, fromClause, wherePredicate,
+        null, null, null, null);
+
     InlineViewRef viewRef = new InlineViewRef(/*alias*/ null, tableMaskStmt,
         (TableSampleClause) null);
     tableRef.migratePropertiesTo(viewRef);
@@ -191,8 +197,7 @@ public class InlineViewRef extends TableRef {
     inlineViewAnalyzer_ = new Analyzer(analyzer);
 
     // Catalog views refs require special analysis settings for authorization.
-    boolean isCatalogView = (view_ != null && !view_.isLocalView());
-    if (isCatalogView) {
+    if (isCatalogView()) {
       analyzer.registerAuthAndAuditEvent(view_, priv_, requireGrantOption_);
       if (inlineViewAnalyzer_.isExplain()) {
         // If the user does not have privileges on the view's definition
@@ -210,7 +215,7 @@ public class InlineViewRef extends TableRef {
     }
 
     inlineViewAnalyzer_.setUseHiveColLabels(
-        isCatalogView ? true : analyzer.useHiveColLabels());
+        isCatalogView() ? true : analyzer.useHiveColLabels());
     queryStmt_.analyze(inlineViewAnalyzer_);
     correlatedTupleIds_.addAll(queryStmt_.getCorrelatedTupleIds());
     if (explicitColLabels_ != null) {
@@ -388,6 +393,8 @@ public class InlineViewRef extends TableRef {
 
   public boolean isTableMaskingView() { return isTableMaskingView_; }
 
+  public boolean isCatalogView() { return view_ != null && !view_.isLocalView(); }
+
   /**
    * Return the unmasked TableRef if this is an inline view for table masking.
    */
@@ -396,7 +403,9 @@ public class InlineViewRef extends TableRef {
     Preconditions.checkState(queryStmt_ instanceof SelectStmt);
     SelectStmt selectStmt = (SelectStmt) queryStmt_;
     Preconditions.checkNotNull(selectStmt.fromClause_);
-    Preconditions.checkState(selectStmt.fromClause_.size() == 1);
+    // FromClause could have several table refs due to subquery rewrite, i.e. subquery
+    // could be rewritten to joins. The first table refs is the original table ref.
+    Preconditions.checkState(selectStmt.fromClause_.size() > 0);
     return selectStmt.fromClause_.get(0);
   }
 
