@@ -23,27 +23,39 @@ import java.util.HashSet;
 import java.util.List;
 import java.util.Set;
 
+import org.apache.iceberg.types.Types;
 import org.apache.impala.authorization.Privilege;
+import org.apache.impala.catalog.BuiltinsDb;
 import org.apache.impala.catalog.Column;
+import org.apache.impala.catalog.FeFsPartition;
 import org.apache.impala.catalog.FeFsTable;
 import org.apache.impala.catalog.FeHBaseTable;
+import org.apache.impala.catalog.FeIcebergTable;
 import org.apache.impala.catalog.FeKuduTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.catalog.FeView;
+import org.apache.impala.catalog.HdfsFileFormat;
+import org.apache.impala.catalog.HdfsTable;
+import org.apache.impala.catalog.IcebergColumn;
 import org.apache.impala.catalog.KuduColumn;
+import org.apache.impala.catalog.PrunablePartition;
 import org.apache.impala.catalog.Type;
 import org.apache.impala.catalog.View;
 import org.apache.impala.common.AnalysisException;
 import org.apache.impala.common.FileSystemUtil;
 import org.apache.impala.common.Pair;
 import org.apache.impala.planner.DataSink;
+import org.apache.impala.planner.HdfsTableSink;
 import org.apache.impala.planner.TableSink;
 import org.apache.impala.rewrite.ExprRewriter;
+import org.apache.impala.thrift.TIcebergPartitionTransformType;
 import org.apache.impala.thrift.TSortingOrder;
+import org.apache.impala.util.IcebergUtil;
 
 import com.google.common.base.Joiner;
 import com.google.common.base.Preconditions;
 import com.google.common.collect.Iterables;
+import com.google.common.collect.Sets;
 
 /**
  * Representation of a single insert or upsert statement, including the select statement
@@ -117,6 +129,10 @@ public class InsertStmt extends StatementBase {
   // Set in analyze(). Contains metadata of target table to determine type of sink.
   private FeTable table_;
 
+  // Set in analyze(). Set the limit on the maximum number of table sink instances.
+  // A value of 0 means no limit.
+  private int maxTableSinks_ = 0;
+
   // Set in analyze(). Exprs correspond to the partitionKeyValues, if specified, or to
   // the partition columns for Kudu tables.
   private List<Expr> partitionKeyExprs_ = new ArrayList<>();
@@ -287,6 +303,13 @@ public class InsertStmt extends StatementBase {
       selectListExprs = new ArrayList<>();
     }
 
+    // Make sure static partition key values only contain const exprs.
+    if (partitionKeyValues_ != null) {
+      for (PartitionKeyValue kv: partitionKeyValues_) {
+        kv.analyze(analyzer);
+      }
+    }
+
     // Set target table and perform table-type specific analysis and auth checking.
     // Also checks if the target table is missing.
     analyzeTargetTable(analyzer);
@@ -350,7 +373,7 @@ public class InsertStmt extends StatementBase {
     }
 
     int numStaticPartitionExprs = 0;
-    if (partitionKeyValues_ != null) {
+    if (partitionKeyValues_ != null && !isIcebergTarget()) {
       for (PartitionKeyValue pkv: partitionKeyValues_) {
         Column column = table_.getColumn(pkv.getColName());
         if (column == null) {
@@ -379,13 +402,6 @@ public class InsertStmt extends StatementBase {
     checkColumnCoverage(selectExprTargetColumns, mentionedColumnNames,
         selectListExprs.size(), numStaticPartitionExprs);
 
-    // Make sure static partition key values only contain const exprs.
-    if (partitionKeyValues_ != null) {
-      for (PartitionKeyValue kv: partitionKeyValues_) {
-        kv.analyze(analyzer);
-      }
-    }
-
     // Check that we can write to the target table/partition. This must be
     // done after the partition expression has been analyzed above.
     analyzeWriteAccess();
@@ -435,14 +451,15 @@ public class InsertStmt extends StatementBase {
           builder.onTable(table_).allOf(privilegeRequired).build());
     }
 
-    // We do not support (in|up)serting into views.
+    // We do not support (in|up)serting into views and iceberg table
     if (table_ instanceof FeView) {
       throw new AnalysisException(
           String.format("Impala does not support %sing into views: %s", getOpName(),
               table_.getFullName()));
     }
 
-    analyzer.checkTableCapability(table_, Analyzer.OperationType.WRITE);
+    Analyzer.ensureTableNotFullAcid(table_, "INSERT");
+    Analyzer.checkTableCapability(table_, Analyzer.OperationType.WRITE);
 
     // We do not support (in|up)serting into tables with unsupported column types.
     for (Column c: table_.getColumns()) {
@@ -486,15 +503,30 @@ public class InsertStmt extends StatementBase {
       if (isHBaseTable) {
         throw new AnalysisException("PARTITION clause is not valid for INSERT into " +
             "HBase tables. '" + targetTableName_ + "' is an HBase table");
-
       } else {
-        // Unpartitioned table, but INSERT has PARTITION clause
-        throw new AnalysisException("PARTITION clause is only valid for INSERT into " +
-            "partitioned table. '" + targetTableName_ + "' is not partitioned");
+        if (isIcebergTarget()) {
+          IcebergPartitionSpec partSpec =
+              ((FeIcebergTable)table_).getDefaultPartitionSpec();
+          if (partSpec == null || !partSpec.hasPartitionFields()) {
+            throw new AnalysisException("PARTITION clause is only valid for INSERT " +
+                "into partitioned table. '" + targetTableName_ + "' is not partitioned");
+          }
+          for (PartitionKeyValue pkv: partitionKeyValues_) {
+            if (pkv.isStatic()) {
+              throw new AnalysisException("Static partitioning is not supported for " +
+                  "Iceberg tables.");
+            }
+          }
+        } else {
+          // Unpartitioned table, but INSERT has PARTITION clause
+          throw new AnalysisException("PARTITION clause is only valid for INSERT into " +
+              "partitioned table. '" + targetTableName_ + "' is not partitioned");
+        }
       }
     }
 
     if (table_ instanceof FeFsTable) {
+      setMaxTableSinks(analyzer_.getQueryOptions().getMax_fs_writers());
       FeFsTable fsTable = (FeFsTable) table_;
       StringBuilder error = new StringBuilder();
       fsTable.parseSkipHeaderLineCount(error);
@@ -521,6 +553,33 @@ public class InsertStmt extends StatementBase {
               targetTableName_));
         }
       }
+      // Check if the target partition is supported to write. For static partitioning the
+      // file format is available on partition level, however for dynamic partitioning
+      // the target partition formats are unknown during analysis and planning, therefore
+      // it will be verified based on the table metadata.
+      if (isStaticPartitionTarget()) {
+        PrunablePartition partition =
+            HdfsTable.getPartition(fsTable, partitionKeyValues_);
+        if (partition != null && partition instanceof FeFsPartition) {
+          HdfsFileFormat fileFormat = ((FeFsPartition) partition).getFileFormat();
+          Boolean notSupported =
+              !HdfsTableSink.SUPPORTED_FILE_FORMATS.contains(fileFormat);
+          if (notSupported) {
+            throw new AnalysisException(String.format("Writing the destination " +
+                "partition format '" + fileFormat + "' is not supported."));
+          }
+        }
+      } else {
+        Set<HdfsFileFormat> formats = fsTable.getFileFormats();
+        Set<HdfsFileFormat> unsupportedFormats =
+            Sets.difference(formats, HdfsTableSink.SUPPORTED_FILE_FORMATS);
+        if (!unsupportedFormats.isEmpty()) {
+          throw new AnalysisException(String.format("Destination table '" +
+              fsTable.getFullName() + "' contains partition format(s) that are not " +
+              "supported to write: '" + Joiner.on(',').join(unsupportedFormats) + "', " +
+              "dynamic partition clauses are forbidden."));
+        }
+      }
     }
 
     if (table_ instanceof FeKuduTable) {
@@ -533,30 +592,68 @@ public class InsertStmt extends StatementBase {
       }
     }
 
+    if (table_ instanceof FeIcebergTable) {
+      if (overwrite_) validateNoBucketTransform((FeIcebergTable)table_);
+      validateIcebergColumnsForInsert((FeIcebergTable)table_);
+    }
+
     if (isHBaseTable && overwrite_) {
       throw new AnalysisException("HBase doesn't have a way to perform INSERT OVERWRITE");
     }
   }
 
+  private void validateIcebergColumnsForInsert(FeIcebergTable iceTable)
+      throws AnalysisException {
+    for (Types.NestedField field : iceTable.getIcebergSchema().columns()) {
+      org.apache.iceberg.types.Type iceType = field.type();
+      if (iceType.isPrimitiveType() && iceType instanceof Types.TimestampType) {
+        Types.TimestampType tsType = (Types.TimestampType)iceType;
+        if (tsType.shouldAdjustToUTC()) {
+          throw new AnalysisException("The Iceberg table has a TIMESTAMPTZ " +
+              "column that Impala cannot write.");
+        }
+      }
+    }
+  }
+
+  private void validateNoBucketTransform(FeIcebergTable iceTable)
+      throws AnalysisException {
+    IcebergPartitionSpec spec = iceTable.getDefaultPartitionSpec();
+    if (!spec.hasPartitionFields()) return;
+    for (IcebergPartitionField field : spec.getIcebergPartitionFields()) {
+      if (field.getTransformType() == TIcebergPartitionTransformType.BUCKET) {
+          throw new AnalysisException("The Iceberg table has BUCKET partitioning. " +
+              "This means the outcome of dynamic partition overwrite is unforeseeable. " +
+              "Consider using TRUNCATE and INSERT INTO to overwrite your table.");
+      }
+    }
+  }
+
   private void analyzeWriteAccess() throws AnalysisException {
     if (!(table_ instanceof FeFsTable)) return;
     FeFsTable fsTable = (FeFsTable) table_;
 
-    FeFsTable.Utils.checkWriteAccess(fsTable,
-        hasStaticPartitionTarget() ? partitionKeyValues_ : null, "INSERT");
-  }
-
-  private boolean hasStaticPartitionTarget() {
-    if (partitionKeyValues_ == null) return false;
-
     // If the partition target is fully static, then check for write access against
     // the specific partition. Otherwise, check the whole table.
+    FeFsTable.Utils.checkWriteAccess(fsTable,
+        isStaticPartitionTarget() ? partitionKeyValues_ : null, "INSERT");
+  }
+
+  /**
+   * Returns true if all the partition key values of the target table are static.
+   */
+  private boolean isStaticPartitionTarget() {
+    if (partitionKeyValues_ == null) return false;
     for (PartitionKeyValue pkv : partitionKeyValues_) {
       if (pkv.isDynamic()) return false;
     }
     return true;
   }
 
+  private boolean isIcebergTarget() {
+    return table_ instanceof FeIcebergTable;
+  }
+
   /**
    * Checks that the column permutation + select list + static partition exprs + dynamic
    * partition exprs collectively cover exactly all required columns in the target table,
@@ -701,9 +798,13 @@ public class InsertStmt extends StatementBase {
     if (isKuduTable) {
       kuduPartitionColumnNames = getKuduPartitionColumnNames((FeKuduTable) table_);
     }
+    IcebergPartitionSpec icebergPartSpec = null;
+    if (isIcebergTarget()) {
+      icebergPartSpec = ((FeIcebergTable)table_).getDefaultPartitionSpec();
+    }
 
-    UnionStmt unionStmt =
-        (queryStmt_ instanceof UnionStmt) ? (UnionStmt) queryStmt_ : null;
+    SetOperationStmt unionStmt =
+        (queryStmt_ instanceof SetOperationStmt) ? (SetOperationStmt) queryStmt_ : null;
     List<Expr> widestTypeExprList = null;
     if (unionStmt != null && unionStmt.getWidestExprs() != null
         && unionStmt.getWidestExprs().size() > 0) {
@@ -747,24 +848,38 @@ public class InsertStmt extends StatementBase {
       }
     }
 
-    // Reorder the partition key exprs and names to be consistent with the target table
-    // declaration, and store their column positions.  We need those exprs in the original
-    // order to create the corresponding Hdfs folder structure correctly, or the indexes
-    // to construct rows to pass to the Kudu partitioning API.
-    for (int i = 0; i < table_.getColumns().size(); ++i) {
-      Column c = table_.getColumns().get(i);
-      for (int j = 0; j < tmpPartitionKeyNames.size(); ++j) {
-        if (c.getName().equals(tmpPartitionKeyNames.get(j))) {
-          partitionKeyExprs_.add(tmpPartitionKeyExprs.get(j));
-          partitionColPos_.add(i);
-          break;
+    if (isIcebergTarget()) {
+      // Add partition key expressions in the order of the Iceberg partition fields.
+      addIcebergPartExprs(analyzer, widestTypeExprList, selectExprTargetColumns,
+          selectListExprs, icebergPartSpec);
+    } else {
+      // Reorder the partition key exprs and names to be consistent with the target table
+      // declaration, and store their column positions.  We need those exprs in the
+      // original order to create the corresponding Hdfs folder structure correctly, or
+      // the indexes to construct rows to pass to the Kudu partitioning API.
+      for (int i = 0; i < table_.getColumns().size(); ++i) {
+        Column c = table_.getColumns().get(i);
+        for (int j = 0; j < tmpPartitionKeyNames.size(); ++j) {
+          if (c.getName().equals(tmpPartitionKeyNames.get(j))) {
+            partitionKeyExprs_.add(tmpPartitionKeyExprs.get(j));
+            partitionColPos_.add(i);
+            break;
+          }
         }
       }
     }
 
-    Preconditions.checkState(
-        (isKuduTable && partitionKeyExprs_.size() == kuduPartitionColumnNames.size())
-        || partitionKeyExprs_.size() == numClusteringCols);
+    if (isIcebergTarget()) {
+      Preconditions.checkState(
+          partitionKeyExprs_.size() == icebergPartSpec.getIcebergPartitionFieldsSize());
+    }
+    else if (isKuduTable) {
+      Preconditions.checkState(
+          partitionKeyExprs_.size() == kuduPartitionColumnNames.size());
+    } else {
+      Preconditions.checkState(partitionKeyExprs_.size() == numClusteringCols);
+    }
+
     // Make sure we have stats for partitionKeyExprs
     for (Expr expr: partitionKeyExprs_) {
       expr.analyze(analyzer);
@@ -829,6 +944,77 @@ public class InsertStmt extends StatementBase {
     }
   }
 
+  /**
+   * Adds partition key expressions in the order of Iceberg partition spec fields.
+   */
+  private void addIcebergPartExprs(Analyzer analyzer, List<Expr> widestTypeExprList,
+      List<Column> selectExprTargetColumns, List<Expr> selectListExprs,
+      IcebergPartitionSpec icebergPartSpec) throws AnalysisException {
+    if (!icebergPartSpec.hasPartitionFields()) return;
+    for (IcebergPartitionField partField : icebergPartSpec.getIcebergPartitionFields()) {
+      for (int i = 0; i < selectListExprs.size(); ++i) {
+        IcebergColumn targetColumn = (IcebergColumn)selectExprTargetColumns.get(i);
+        if (targetColumn.getFieldId() != partField.getSourceId()) continue;
+        // widestTypeExpr is widest type expression for column i
+        Expr widestTypeExpr =
+            (widestTypeExprList != null) ? widestTypeExprList.get(i) : null;
+        Expr compatibleExpr = checkTypeCompatibility(targetTableName_.toString(),
+            targetColumn, selectListExprs.get(i),
+            analyzer.getQueryOptions().isDecimal_v2(),
+            widestTypeExpr);
+        Expr icebergPartitionTransformExpr =
+            getIcebergPartitionTransformExpr(partField, compatibleExpr);
+        partitionKeyExprs_.add(icebergPartitionTransformExpr);
+        partitionColPos_.add(targetColumn.getPosition());
+        break;
+      }
+    }
+  }
+
+  /**
+   * Returns the partition transform expression. E.g. if the partition transform is DAY,
+   * it returns 'to_date(compatibleExpr)'. If the partition transform is BUCKET,
+   * it returns 'iceberg_bucket_transform(compatibleExpr, transformParam)'.
+   */
+  private Expr getIcebergPartitionTransformExpr(IcebergPartitionField partField,
+      Expr compatibleExpr) {
+    String funcNameStr = transformTypeToFunctionName(partField.getTransformType());
+    if (funcNameStr == null || funcNameStr.equals("")) return compatibleExpr;
+    FunctionName fnName = new FunctionName(BuiltinsDb.NAME, funcNameStr);
+    List<Expr> paramList = new ArrayList<>();
+    paramList.add(compatibleExpr);
+    Integer transformParam = partField.getTransformParam();
+    if (transformParam != null) {
+      paramList.add(NumericLiteral.create(transformParam));
+    }
+    if (partField.getTransformType() == TIcebergPartitionTransformType.MONTH) {
+      paramList.add(new StringLiteral("yyyy-MM"));
+    }
+    else if (partField.getTransformType() == TIcebergPartitionTransformType.HOUR) {
+      paramList.add(new StringLiteral("yyyy-MM-dd-HH"));
+    }
+    FunctionCallExpr fnCall = new FunctionCallExpr(fnName, new FunctionParams(paramList));
+    fnCall.setIsInternalFnCall(true);
+    return fnCall;
+  }
+
+  /**
+   * Returns the builtin function to use for the given partition transform.
+   */
+  private static String transformTypeToFunctionName(
+      TIcebergPartitionTransformType transformType) {
+    switch (transformType) {
+      case IDENTITY: return "";
+      case HOUR:
+      case MONTH: return "from_timestamp";
+      case DAY: return "to_date";
+      case YEAR: return "year";
+      case BUCKET: return "iceberg_bucket_transform";
+      case TRUNCATE: return "iceberg_truncate_transform";
+    }
+    return "";
+  }
+
   private static Set<String> getKuduPartitionColumnNames(FeKuduTable table) {
     Set<String> ret = new HashSet<>();
     for (KuduPartitionParam partitionParam : table.getPartitionBy()) {
@@ -921,6 +1107,7 @@ public class InsertStmt extends StatementBase {
   public TableName getTargetTableName() { return targetTableName_; }
   public FeTable getTargetTable() { return table_; }
   public void setTargetTable(FeTable table) { this.table_ = table; }
+  public void setMaxTableSinks(int maxTableSinks) { this.maxTableSinks_ = maxTableSinks; }
   public void setWriteId(long writeId) { this.writeId_ = writeId; }
   public boolean isOverwrite() { return overwrite_; }
   public TSortingOrder getSortingOrder() { return sortingOrder_; }
@@ -958,7 +1145,8 @@ public class InsertStmt extends StatementBase {
     Preconditions.checkState(table_ != null);
     return TableSink.create(table_, isUpsert_ ? TableSink.Op.UPSERT : TableSink.Op.INSERT,
         partitionKeyExprs_, resultExprs_, mentionedColumns_, overwrite_,
-        requiresClustering(), new Pair<>(sortColumns_, sortingOrder_), writeId_);
+        requiresClustering(), new Pair<>(sortColumns_, sortingOrder_), writeId_,
+        maxTableSinks_);
   }
 
   /**
@@ -1026,4 +1214,12 @@ public class InsertStmt extends StatementBase {
     tblRefs.add(new TableRef(targetTableName_.toPath(), null));
     if (queryStmt_ != null) queryStmt_.collectTableRefs(tblRefs);
   }
+
+  @Override
+  public boolean resolveTableMask(Analyzer analyzer) throws AnalysisException {
+    boolean hasChange = false;
+    if (withClause_ != null) hasChange = withClause_.resolveTableMask(analyzer);
+    if (queryStmt_ != null) hasChange |= queryStmt_.resolveTableMask(analyzer);
+    return hasChange;
+  }
 }
