@@ -38,7 +38,7 @@ import com.google.common.collect.Lists;
 /**
  * Abstract base class for any statement that returns results
  * via a list of result expressions, for example a
- * SelectStmt or UnionStmt. Also maintains a map of expression substitutions
+ * SelectStmt or SetOperationStmt. Also maintains a map of expression substitutions
  * for replacing expressions from ORDER BY or GROUP BY clauses with
  * their corresponding result expressions.
  * Used for sharing members/methods and some of the analysis code, in particular the
@@ -167,12 +167,15 @@ public abstract class QueryStmt extends StatementBase {
     if (hasWithClause()) withClause_.analyze(analyzer);
   }
 
-  /**
-   * Disable table masking when analyzing the FromClauses. Used in CreateView and
-   * AlterView statements since they don't actually read the data.
-   * @param doTableMasking
-   */
-  public void setDoTableMasking(boolean doTableMasking) {}
+  @Override
+  public boolean resolveTableMask(Analyzer analyzer) throws AnalysisException {
+    boolean hasChanges = false;
+    if (hasWithClause()) {
+      // Local views may be used in rewriting. Make sure they are masked as well.
+      hasChanges = withClause_.resolveTableMask(analyzer);
+    }
+    return hasChanges;
+  }
 
   /**
    * Returns a list containing all the materialized tuple ids that this stmt is
