@@ -19,12 +19,14 @@ package org.apache.impala.analysis;
 
 import java.util.ArrayList;
 import java.util.Collections;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Set;
 
 import org.apache.impala.analysis.Path.PathType;
 import org.apache.impala.authorization.Privilege;
 import org.apache.impala.catalog.Column;
+import org.apache.impala.catalog.FeKuduTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.catalog.FeView;
 import org.apache.impala.catalog.StructField;
@@ -32,9 +34,11 @@ import org.apache.impala.catalog.StructType;
 import org.apache.impala.catalog.TableLoadingException;
 import org.apache.impala.common.AnalysisException;
 import org.apache.impala.common.ColumnAliasGenerator;
+import org.apache.impala.common.Pair;
 import org.apache.impala.common.TableAliasGenerator;
 import org.apache.impala.common.TreeNode;
 import org.apache.impala.rewrite.ExprRewriter;
+import org.apache.impala.service.BackendConfig;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -50,39 +54,6 @@ import com.google.common.collect.Lists;
 public class SelectStmt extends QueryStmt {
   private final static Logger LOG = LoggerFactory.getLogger(SelectStmt.class);
 
-  /**
-   * Impala 3.x and earlier provide non-standard support for a single ordinal
-   * in the HAVING clause:
-   *
-   * SELECT region, count(*), count(*) > 10 FROM sales
-   * HAVING 3
-   *
-   * Most other SQL dialects do not allow ordinals in expressions (such as
-   * HAVING), only in lists (GROUP BY, ORDER BY). The standard way to express
-   * the above is:
-   *
-   * SELECT region, count(*) FROM sales
-   * HAVING count(*) > 10
-   *
-   * Or, better (not currently supported by IMPALA):
-   *
-   * SELECT region, count(*) c FROM SALES
-   * HAVING c > 10
-   *
-   * We intend to disable this non-standard feature in the next major
-   * release. For now, we leave the feature/bug enabled, controlled by
-   * the following constant.
-   *
-   * In that same release, we will enable the third example above.
-   * At present, all that is supported is the odd:
-   *
-   * SELECT region, count(*), count(*) > 10 p FROM sales
-   * HAVING p
-   *
-   * See IMPALA-7844.
-   */
-  public static final boolean ALLOW_ORDINALS_IN_HAVING = true;
-
   /////////////////////////////////////////
   // BEGIN: Members that need to be reset()
 
@@ -90,7 +61,17 @@ public class SelectStmt extends QueryStmt {
   protected final List<String> colLabels_; // lower case column labels
   protected final FromClause fromClause_;
   protected Expr whereClause_;
+
+  // Grouping expressions for this select block, including expressions from the original
+  // query statement and additional expressions that may be added during rewriting.
   protected List<Expr> groupingExprs_;
+
+  // The original GROUP BY clause with information about grouping sets, etc.
+  // If there was no original GROUP BY clause, but grouping exprs are added during
+  // rewriting, this is initialized as a placeholder empty group by list.
+  // Non-null iff groupingExprs_ is non-null.
+  private GroupByClause groupByClause_;
+
   protected Expr havingClause_;  // original having clause
 
   // havingClause with aliases and agg output resolved
@@ -115,7 +96,7 @@ public class SelectStmt extends QueryStmt {
 
   SelectStmt(SelectList selectList,
              FromClause fromClause,
-             Expr wherePredicate, List<Expr> groupingExprs,
+             Expr wherePredicate, GroupByClause groupByClause,
              Expr havingPredicate, List<OrderByElement> orderByElements,
              LimitElement limitElement) {
     super(orderByElements, limitElement);
@@ -126,7 +107,12 @@ public class SelectStmt extends QueryStmt {
       fromClause_ = fromClause;
     }
     whereClause_ = wherePredicate;
-    groupingExprs_ = groupingExprs;
+    groupByClause_ = groupByClause;
+    if (groupByClause != null) {
+      groupingExprs_ = Expr.cloneList(groupByClause.getOrigGroupingExprs());
+    } else {
+      groupingExprs_ = null;
+    }
     havingClause_ = havingPredicate;
     colLabels_ = new ArrayList<>();
     havingPred_ = null;
@@ -159,6 +145,74 @@ public class SelectStmt extends QueryStmt {
   public boolean hasHavingClause() { return havingClause_ != null; }
   public ExprSubstitutionMap getBaseTblSmap() { return baseTblSmap_; }
 
+  /**
+   * A simple limit statement has a limit but no order-by,
+   * group-by, aggregates or analytic functions. Joins are
+   * allowed if one of the table refs has a limit_to_sample
+   * hint. The statement can have a WHERE clause if it has an
+   * always_true hint.
+   * This method does not explicitly check for subqueries. If the
+   * subquery occurs in the WHERE, the initial check for simple
+   * limit may succeed. Subsequently, in the planning process
+   * the statement rewriter may transform it to a join or an
+   * equivalent plan and analyze will be called again and this time
+   * it may or may not satisfy the simple limit criteria.
+   * Note that FROM clause subquery is generally ok since it is
+   * a table ref.
+   */
+  public Pair<Boolean, Long> checkSimpleLimitStmt() {
+    if (hasSimpleLimitEligibleTableRefs()
+        && !hasGroupByClause() && !hasOrderByClause()
+        && !hasMultiAggInfo() && !hasAnalyticInfo()) {
+      if (hasWhereClause() && !Expr.IS_ALWAYS_TRUE_PREDICATE.apply(getWhereClause())) {
+        return null;
+      }
+      if (hasLimit()) {
+        return new Pair<>(new Boolean(true), getLimit());
+      } else {
+        // even if this SELECT statement does not have a LIMIT, it is a
+        // simple select which may be an inline view and eligible for a
+        // limit pushdown from an outer block, so we return a non-null value
+        return new Pair<>(new Boolean(false), null);
+      }
+    }
+    return null;
+  }
+
+  /**
+   * Append additional grouping expressions to the select list. Used by StmtRewriter.
+   */
+  protected void addGroupingExprs(List<Expr> addtlGroupingExprs) {
+    if (groupingExprs_ == null) {
+      groupByClause_ = new GroupByClause(
+          Collections.emptyList(), GroupByClause.GroupingSetsType.NONE);
+      groupingExprs_ = new ArrayList<>();
+    }
+    groupingExprs_.addAll(addtlGroupingExprs);
+  }
+
+  private boolean hasSimpleLimitEligibleTableRefs() {
+    // for single table query blocks, limit pushdown can be considered
+    // even if no table hint is present
+    if (getTableRefs().size() == 1) return true;
+    // if there are multiple table refs, at least 1 should have the hint
+    // for limit to sample
+    for (TableRef ref : getTableRefs()) {
+      if (ref.hasConvertLimitToSampleHint()) return true;
+    }
+    return false;
+  }
+
+  /**
+   * Remove the group by clause. Used by StmtRewriter. This changes the semantics
+   * of this statement and should only be called when the query is being rewritten
+   * in a way such that the GROUP BY is *not* required for correctness.
+   */
+  protected void removeGroupBy() {
+    groupByClause_ = null;
+    groupingExprs_ = null;
+  }
+
   // Column alias generator used during query rewriting.
   private ColumnAliasGenerator columnAliasGenerator_ = null;
   public ColumnAliasGenerator getColumnAliasGenerator() {
@@ -178,8 +232,20 @@ public class SelectStmt extends QueryStmt {
   }
 
   @Override
-  public void setDoTableMasking(boolean doTableMasking) {
-    fromClause_.setDoTableMasking(doTableMasking);
+  public boolean resolveTableMask(Analyzer analyzer) throws AnalysisException {
+    boolean hasChanges = super.resolveTableMask(analyzer);
+    // Recurse in all places that could have subqueries. After resolveTableMask() is done
+    // on the whole AST, SlotRefs referencing the original table refs will be reset and
+    // re-analyzed to reference the table masking views. So we don't need to deal with
+    // SlotRefs here.
+    hasChanges |= fromClause_.resolveTableMask(analyzer);
+    for (SelectListItem item : selectList_.getItems()) {
+      if (item.isStar()) continue;
+      hasChanges |= item.getExpr().resolveTableMask(analyzer);
+    }
+    if (whereClause_ != null) hasChanges |= whereClause_.resolveTableMask(analyzer);
+    if (havingClause_ != null) hasChanges |= havingClause_.resolveTableMask(analyzer);
+    return hasChanges;
   }
 
   /**
@@ -201,7 +267,6 @@ public class SelectStmt extends QueryStmt {
     private final Analyzer analyzer_;
     private List<Expr> groupingExprsCopy_;
     private List<FunctionCallExpr> aggExprs_;
-    private ExprSubstitutionMap ndvSmap_;
     private ExprSubstitutionMap countAllMap_;
 
     private SelectAnalyzer(Analyzer analyzer) {
@@ -227,7 +292,6 @@ public class SelectStmt extends QueryStmt {
         verifyAggSemantics();
         analyzeGroupingExprs();
         collectAggExprs();
-        rewriteCountDistinct();
         buildAggregateExprs();
         buildResultExprs();
         verifyAggregation();
@@ -248,6 +312,7 @@ public class SelectStmt extends QueryStmt {
       }
 
       buildColumnLineageGraph();
+      analyzer_.setSimpleLimitStatus(checkSimpleLimitStmt());
     }
 
     private void analyzeSelectClause() throws AnalysisException {
@@ -274,9 +339,24 @@ public class SelectStmt extends QueryStmt {
           // Analyze the resultExpr before generating a label to ensure enforcement
           // of expr child and depth limits (toColumn() label may call toSql()).
           item.getExpr().analyze(analyzer_);
-          if (item.getExpr().contains(Predicates.instanceOf(Subquery.class))) {
-            throw new AnalysisException(
-                "Subqueries are not supported in the select list.");
+          // Check for scalar subquery types which are not supported
+          List<Subquery> subqueryExprs = new ArrayList<>();
+          item.getExpr().collect(Subquery.class, subqueryExprs);
+          for (Subquery s : subqueryExprs) {
+            Preconditions.checkState(s.getStatement() instanceof SelectStmt);
+            if (!s.returnsScalarColumn()) {
+              throw new AnalysisException("A non-scalar subquery is not supported in "
+                  + "the expression: " + item.getExpr().toSql());
+            }
+            if (s.getStatement().isRuntimeScalar()) {
+              throw new AnalysisException(
+                  "A subquery which may return more than one row is not supported in "
+                  + "the expression: " + item.getExpr().toSql());
+            }
+            if (!((SelectStmt)s.getStatement()).returnsAtMostOneRow()) {
+              throw new AnalysisException("Only subqueries that are guaranteed to return "
+                   + "a single row are supported: " + item.getExpr().toSql());
+            }
           }
           resultExprs_.add(item.getExpr());
           String label = item.toColumnLabel(i, analyzer_.useHiveColLabels());
@@ -462,6 +542,7 @@ public class SelectStmt extends QueryStmt {
       }
       // expand in From clause order
       for (TableRef tableRef: fromClause_) {
+        if (tableRef.isHidden()) continue;
         if (analyzer_.isSemiJoined(tableRef.getId())) continue;
         Path resolvedPath = new Path(tableRef.getDesc(),
             Collections.<String>emptyList());
@@ -539,7 +620,16 @@ public class SelectStmt extends QueryStmt {
       if (p.destType().isComplexType()) return;
       SlotDescriptor slotDesc = analyzer_.registerSlotRef(p);
       SlotRef slotRef = new SlotRef(slotDesc);
-      slotRef.analyze(analyzer_);
+      Preconditions.checkState(slotRef.isAnalyzed(),
+          "Analysis should be done in constructor");
+      // Empty matched types means this is expanded from star of a catalog table.
+      // For star of complex types, e.g. my_struct.*, my_array.*, my_map.*, the matched
+      // types will have the complex type so it's not empty.
+      if (resolvedPath.getMatchedTypes().isEmpty()) {
+        Preconditions.checkState(!slotDesc.getType().isComplexType(),
+            "Star expansion should only introduce scalar columns");
+        analyzer_.registerScalarColumnForMasking(slotDesc);
+      }
       resultExprs_.add(slotRef);
       colLabels_.add(relRawPath[relRawPath.length - 1]);
     }
@@ -607,14 +697,16 @@ public class SelectStmt extends QueryStmt {
       // Analyze the HAVING clause first so we can check if it contains aggregates.
       // We need to analyze/register it even if we are not computing aggregates.
       if (havingClause_ == null) return;
-      // can't contain subqueries
-      if (havingClause_.contains(Predicates.instanceOf(Subquery.class))) {
+      List<Expr> subqueries = new ArrayList<>();
+      havingClause_.collectAll(Predicates.instanceOf(Subquery.class), subqueries);
+      if (subqueries.size() > 1) {
         throw new AnalysisException(
-            "Subqueries are not supported in the HAVING clause.");
+            "Multiple subqueries are not supported in expression: "
+            + havingClause_.toSql());
       }
       // Resolve (top-level) aliases and analyzes
       havingPred_ = resolveReferenceExpr(havingClause_, "HAVING", analyzer_,
-          ALLOW_ORDINALS_IN_HAVING);
+          BackendConfig.INSTANCE.getAllowOrdinalsInHaving());
       // can't contain analytic exprs
       Expr analyticExpr = havingPred_.findFirstOf(AnalyticExpr.class);
       if (analyticExpr != null) {
@@ -626,13 +718,7 @@ public class SelectStmt extends QueryStmt {
     }
 
     private boolean checkForAggregates() throws AnalysisException {
-      if (groupingExprs_ == null && !selectList_.isDistinct()
-          && !TreeNode.contains(resultExprs_, Expr.IS_AGGREGATE)
-          && (havingPred_ == null
-              || !havingPred_.contains(Expr.IS_AGGREGATE))
-          && (sortInfo_ == null
-              || !TreeNode.contains(sortInfo_.getSortExprs(),
-                                    Expr.IS_AGGREGATE))) {
+      if (!hasAggregate(/*includeDistinct=*/ true)) {
         // We're not computing aggregates but we still need to register the HAVING
         // clause which could, e.g., contain a constant expression evaluating to false.
         if (havingPred_ != null) analyzer_.registerConjuncts(havingPred_, true);
@@ -706,6 +792,22 @@ public class SelectStmt extends QueryStmt {
                   + groupingExprsCopy_.get(i).toSql());
         }
       }
+      // initialize groupingExprs_ with the analyzed version
+      // use the original ordinal if the analyzed expr is a INT literal
+      List<Expr> groupingExprs = new ArrayList<>();
+      for (int i = 0; i < groupingExprsCopy_.size(); ++i) {
+        Expr expr = groupingExprsCopy_.get(i);
+        if (expr instanceof NumericLiteral && Expr.IS_INT_LITERAL.apply(expr)) {
+          groupingExprs.add(groupingExprs_.get(i).clone());
+        } else {
+          groupingExprs.add(expr);
+        }
+      }
+      groupingExprs_ = groupingExprs;
+
+      if (groupByClause_ != null && groupByClause_.hasGroupingSets()) {
+        groupByClause_.analyzeGroupingSets(groupingExprsCopy_);
+      }
     }
 
     private void collectAggExprs() {
@@ -723,34 +825,6 @@ public class SelectStmt extends QueryStmt {
       }
     }
 
-    private void rewriteCountDistinct() {
-      // Optionally rewrite all count(distinct <expr>) into equivalent NDV() calls.
-      if (!analyzer_.getQueryCtx().client_request.query_options.appx_count_distinct) {
-        return;
-      }
-      ndvSmap_ = new ExprSubstitutionMap();
-      for (FunctionCallExpr aggExpr: aggExprs_) {
-        if (!aggExpr.isDistinct()
-            || !aggExpr.getFnName().getFunction().equals("count")
-            || aggExpr.getParams().size() != 1) {
-          continue;
-        }
-        FunctionCallExpr ndvFnCall =
-            new FunctionCallExpr("ndv", aggExpr.getParams().exprs());
-        ndvFnCall.analyzeNoThrow(analyzer_);
-        Preconditions.checkState(ndvFnCall.getType().equals(aggExpr.getType()));
-        ndvSmap_.put(aggExpr, ndvFnCall);
-      }
-      // Replace all count(distinct <expr>) with NDV(<expr>).
-      List<Expr> substAggExprs = Expr.substituteList(aggExprs_,
-          ndvSmap_, analyzer_, false);
-      aggExprs_.clear();
-      for (Expr aggExpr: substAggExprs) {
-        Preconditions.checkState(aggExpr instanceof FunctionCallExpr);
-        aggExprs_.add((FunctionCallExpr) aggExpr);
-      }
-    }
-
     private void buildAggregateExprs() throws AnalysisException {
       // When DISTINCT aggregates are present, non-distinct (i.e. ALL) aggregates are
       // evaluated in two phases (see AggregateInfo for more details). In particular,
@@ -764,7 +838,6 @@ public class SelectStmt extends QueryStmt {
       // i) There is no GROUP-BY clause, and
       // ii) Other DISTINCT aggregates are present.
       countAllMap_ = createCountAllMap();
-      countAllMap_ = ExprSubstitutionMap.compose(ndvSmap_, countAllMap_, analyzer_);
       List<Expr> substitutedAggs =
           Expr.substituteList(aggExprs_, countAllMap_, analyzer_, false);
       aggExprs_.clear();
@@ -779,9 +852,14 @@ public class SelectStmt extends QueryStmt {
         Preconditions.checkState(aggExprs_.isEmpty());
         groupingExprs = Expr.cloneList(resultExprs_);
       }
+      // All expressions passed into the MultiAggregateInfo must be deduplicated.
+      // The analyzed grouping sets were already deduplicated.
       Expr.removeDuplicates(aggExprs_);
       Expr.removeDuplicates(groupingExprs);
-      multiAggInfo_ = new MultiAggregateInfo(groupingExprs, aggExprs_);
+      List<List<Expr>> groupingSets =
+          (groupByClause_ != null && groupByClause_.hasGroupingSets()) ?
+          groupByClause_.getAnalyzedGroupingSets() : null;
+      multiAggInfo_ = new MultiAggregateInfo(groupingExprs, aggExprs_, groupingSets);
       multiAggInfo_.analyze(analyzer_);
     }
 
@@ -802,10 +880,6 @@ public class SelectStmt extends QueryStmt {
         LOG.trace("post-agg selectListExprs: " + Expr.debugString(resultExprs_));
       }
       if (havingPred_ != null) {
-        // Make sure the predicate in the HAVING clause does not contain a
-        // subquery.
-        Preconditions.checkState(!havingPred_.contains(
-            Predicates.instanceOf(Subquery.class)));
         havingPred_ = havingPred_.substitute(combinedSmap, analyzer_, false);
         analyzer_.registerConjuncts(havingPred_, true);
         if (LOG.isTraceEnabled()) {
@@ -1011,6 +1085,22 @@ public class SelectStmt extends QueryStmt {
       List<TupleId> tids = new ArrayList<>();
       getMaterializedTupleIds(tids); // includes the analytic tuple
       List<Expr> conjuncts = analyzer.getUnassignedConjuncts(tids, false);
+      // The predicates that can be bounded to KuduScanNode don't need to materialize
+      // here. Because we don't need to materialize the predicates that can be evaluated
+      // by Kudu.
+      for (TupleId tid : tids) {
+        if (analyzer.getTupleDesc(tid).getTable() instanceof FeKuduTable) {
+          Iterator<Expr> iterator = conjuncts.iterator();
+          while (iterator.hasNext()) {
+            Expr e = iterator.next();
+            List<TupleId> etids = new ArrayList<>();
+            e.getIds(etids, null);
+            if (1 == etids.size() && etids.get(0) == tid) {
+              iterator.remove();
+            }
+          }
+        }
+      }
       materializeSlots(analyzer, conjuncts);
       analyticInfo_.materializeRequiredSlots(analyzer, baseTblSmap_);
     }
@@ -1053,16 +1143,16 @@ public class SelectStmt extends QueryStmt {
     Preconditions.checkState(isAnalyzed());
     selectList_.rewriteExprs(rewriter, analyzer_);
     for (TableRef ref: fromClause_.getTableRefs()) ref.rewriteExprs(rewriter, analyzer_);
+    List<Subquery> subqueryExprs = new ArrayList<>();
     if (whereClause_ != null) {
       whereClause_ = rewriter.rewrite(whereClause_, analyzer_);
-      // Also rewrite exprs in the statements of subqueries.
-      List<Subquery> subqueryExprs = new ArrayList<>();
       whereClause_.collect(Subquery.class, subqueryExprs);
-      for (Subquery s: subqueryExprs) s.getStatement().rewriteExprs(rewriter);
     }
     if (havingClause_ != null) {
       havingClause_ = rewriteCheckOrdinalResult(rewriter, havingClause_);
+      havingClause_.collect(Subquery.class, subqueryExprs);
     }
+    for (Subquery s : subqueryExprs) s.getStatement().rewriteExprs(rewriter);
     if (groupingExprs_ != null) {
       for (int i = 0; i < groupingExprs_.size(); ++i) {
         groupingExprs_.set(i, rewriteCheckOrdinalResult(
@@ -1107,20 +1197,21 @@ public class SelectStmt extends QueryStmt {
     // Where clause
     if (whereClause_ != null) {
       strBuilder.append(" WHERE ");
+      List<PlanHint> predHints = whereClause_.getPredicateHints();
+      if (predHints != null && predHints.size() > 0) {
+        strBuilder.append(ToSqlUtils.getPlanHintsSql(options,
+            predHints)).append(" ");
+      }
       strBuilder.append(whereClause_.toSql(options));
     }
     // Group By clause
-    if (groupingExprs_ != null) {
-      strBuilder.append(" GROUP BY ");
+    if (groupByClause_ != null) {
       // Handle both analyzed (multiAggInfo_ != null) and unanalyzed cases.
       // Unanalyzed case us used to generate SQL such as for views.
       // See ToSqlUtils.getCreateViewSql().
       List<Expr> groupingExprs = multiAggInfo_ == null
           ? groupingExprs_ : multiAggInfo_.getGroupingExprs();
-      for (int i = 0; i < groupingExprs.size(); ++i) {
-        strBuilder.append(groupingExprs.get(i).toSql(options));
-        strBuilder.append((i+1 != groupingExprs.size()) ? ", " : "");
-      }
+      strBuilder.append(groupByClause_.toSql(groupingExprs, options));
     }
     // Having clause
     if (havingClause_ != null) {
@@ -1181,6 +1272,8 @@ public class SelectStmt extends QueryStmt {
     whereClause_ = (other.whereClause_ != null) ? other.whereClause_.clone() : null;
     groupingExprs_ =
         (other.groupingExprs_ != null) ? Expr.cloneList(other.groupingExprs_) : null;
+    groupByClause_ =
+        (other.groupByClause_ != null) ? other.groupByClause_.clone() : null;
     havingClause_ = (other.havingClause_ != null) ? other.havingClause_.clone() : null;
     colLabels_ = Lists.newArrayList(other.colLabels_);
     multiAggInfo_ = (other.multiAggInfo_ != null) ? other.multiAggInfo_.clone() : null;
@@ -1195,12 +1288,20 @@ public class SelectStmt extends QueryStmt {
     if (fromClauseOnly) {
       fromClause_.collectFromClauseTableRefs(tblRefs);
     } else {
+      // Collect TableRefs in all subqueries.
       fromClause_.collectTableRefs(tblRefs);
-    }
-    if (!fromClauseOnly && whereClause_ != null) {
-      // Collect TableRefs in WHERE-clause subqueries.
       List<Subquery> subqueries = new ArrayList<>();
-      whereClause_.collect(Subquery.class, subqueries);
+      if (whereClause_ != null) {
+        whereClause_.collect(Subquery.class, subqueries);
+      }
+      if (havingClause_ != null) {
+        havingClause_.collect(Subquery.class, subqueries);
+      }
+      for (SelectListItem item : selectList_.getItems()) {
+        if (item.isStar()) continue;
+        item.getExpr().collect(Subquery.class, subqueries);
+      }
+
       for (Subquery sq: subqueries) {
         sq.getStatement().collectTableRefs(tblRefs, fromClauseOnly);
       }
@@ -1209,8 +1310,6 @@ public class SelectStmt extends QueryStmt {
 
   @Override
   public void collectInlineViews(Set<FeView> inlineViews) {
-    // Impala currently supports sub queries only in FROM, WHERE & WITH clauses. Hence,
-    // this function does not carry out any checks on HAVING clause.
     super.collectInlineViews(inlineViews);
     List<TableRef> fromTblRefs = getTableRefs();
     Preconditions.checkNotNull(inlineViews);
@@ -1234,6 +1333,17 @@ public class SelectStmt extends QueryStmt {
         whereSubQueries.get(0).getStatement().collectInlineViews(inlineViews);
       }
     }
+    List<Subquery> subqueries = Lists.newArrayList();
+    for (SelectListItem item : selectList_.getItems()) {
+      if (item.isStar()) continue;
+      item.getExpr().collect(Subquery.class, subqueries);
+    }
+    if (havingClause_ != null) {
+      havingClause_.collect(Subquery.class, subqueries);
+    }
+    for (Subquery sq : subqueries) {
+      sq.getStatement().collectInlineViews(inlineViews);
+    }
   }
 
   @Override
@@ -1243,6 +1353,7 @@ public class SelectStmt extends QueryStmt {
     colLabels_.clear();
     fromClause_.reset();
     if (whereClause_ != null) whereClause_.reset();
+    if (groupByClause_ != null) groupByClause_.reset();
     if (groupingExprs_ != null) Expr.resetList(groupingExprs_);
     if (havingClause_ != null) havingClause_.reset();
     havingPred_ = null;
@@ -1264,27 +1375,96 @@ public class SelectStmt extends QueryStmt {
    *
    * This function may produce false negatives because the cardinality of the
    * result set also depends on the data a stmt is processing.
+   *
+   * TODO: IMPALA-1285 to cover more cases that can be determinded at plan time such has a
+   * group by clause where all grouping expressions are bound to constant expressions.
    */
-  public boolean returnsSingleRow() {
-    Preconditions.checkState(isAnalyzed());
-    // limit 1 clause
-    if (limitElement_ != null && hasLimit() && limitElement_.getLimit() == 1) return true;
-    // No from clause (base tables or inline views)
-    if (fromClause_.isEmpty()) return true;
-    // Aggregation with no group by and no DISTINCT
+  public boolean returnsAtMostOneRow() {
+    return returnsSingleRow(false);
+  }
+
+  /**
+   * Check if the stmt returns exactly one row. This can happen
+   * in the following cases:
+   * 1. select stmt with an aggregate function and no group by and no having clause.
+   * 2. select stmt with no from clause and no where or having clause.
+   * 3. select from an inline view that returns exactly one row and no where or
+   *    having clause.
+   *
+   * This function may produce false negatives because the cardinality of the
+   * result set also depends on the data a stmt is processing.
+   *
+   * TODO: IMPALA-1285 to cover more cases that can be determinded at plan time such has a
+   * group by clause where all grouping expressions are bound to constant expressions.
+   */
+  public boolean returnsExactlyOneRow() {
+    return returnsSingleRow(true);
+  }
+
+  /**
+   * Helper for returnsAtMostOneRow() and returnsExactlyOneRow(). If exactlyOne
+   * is true, returns true iff this SELECT always returns exactly one row. If
+   * exactlyOne is false, returns true iff this SELECT returns zero or one rows.
+   */
+  private boolean returnsSingleRow(boolean exactlyOne) {
+    // This function checks the clauses in reverse order in which they are applied.
+    // We must return false as soon as we determine there is no upper bound on the rows
+    // returned.
+    // If exactlyOne is true, we also must return false when we determine there is no
+    // lower bound on the rows returned.
+    // If exactlyOne is false, we can also return true as soon as we determine an upper
+    // bound on the rows returned.
+
+    if (limitElement_ != null && hasLimit()) {
+      // Limit 0 or 1 clause. This is an upper bound on rows returned.
+      if (!exactlyOne && limitElement_.getLimit() <= 1) return true;
+      // Limit 0 clause - the query will return 0 rows.
+      if (exactlyOne && limitElement_.getLimit() == 0) return false;
+    }
+
+    // HAVING cause can eliminate any rows produced by the statement, therefore we have
+    // no lower bound on rows returned.
+    if (exactlyOne && havingClause_ != null && !havingClause_.isTriviallyTrue()) {
+      return false;
+    }
+
+    // Aggregation with no GROUP BY and no DISTINCT. The aggregation produces exactly one
+    // row and the WHERE clause cannot eliminate that row.
     if (hasMultiAggInfo() && !hasGroupByClause() && !selectList_.isDistinct()) {
       return true;
     }
-    // Select from an inline view that returns at most one row.
+
+    // The WHERE clause can eliminate the singular row produced by the FROM clause
+    // (or lack of FROM clause).
+    if (exactlyOne && whereClause_ != null && !whereClause_.isTriviallyTrue()) {
+      return false;
+    }
+
+    // No FROM clause (base tables or inline views). Exactly one row is produced.
+    if (fromClause_.isEmpty()) return true;
+
     List<TableRef> tableRefs = fromClause_.getTableRefs();
     if (tableRefs.size() == 1 && tableRefs.get(0) instanceof InlineViewRef) {
       InlineViewRef inlineView = (InlineViewRef)tableRefs.get(0);
       if (inlineView.queryStmt_ instanceof SelectStmt) {
         SelectStmt selectStmt = (SelectStmt)inlineView.queryStmt_;
-        return selectStmt.returnsSingleRow();
+        return selectStmt.returnsSingleRow(exactlyOne);
       }
     }
     // In all other cases, return false.
     return false;
   }
+
+  /**
+   * @param includeDistinct if true, a distinct in the select list is counted
+   *    as requiring an aggregation
+   * @returns true if query block has an aggregate function or grouping.
+   */
+  public boolean hasAggregate(boolean includeDistinct) throws AnalysisException {
+    return groupingExprs_ != null || (includeDistinct && selectList_.isDistinct()) ||
+          TreeNode.contains(resultExprs_, Expr.IS_AGGREGATE) ||
+          (havingPred_ != null && havingPred_.contains(Expr.IS_AGGREGATE)) ||
+          (sortInfo_ != null &&
+           TreeNode.contains(sortInfo_.getSortExprs(), Expr.IS_AGGREGATE));
+  }
 }
