@@ -22,11 +22,13 @@ import static org.apache.impala.analysis.ToSqlOptions.DEFAULT;
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.HashSet;
+import java.util.LinkedHashMap;
 import java.util.List;
+import java.util.Map;
 import java.util.Set;
-import java.util.UUID;
 
 import org.apache.impala.authorization.Privilege;
+import org.apache.impala.catalog.Column;
 import org.apache.impala.catalog.FeFsTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.common.AnalysisException;
@@ -91,10 +93,16 @@ public class TableRef extends StmtNode {
   protected JoinOperator joinOp_;
   protected List<PlanHint> joinHints_ = new ArrayList<>();
   protected List<String> usingColNames_;
+  // Whether we should allow an empty ON clause in cases where it would usually be
+  // rejected in analysis. Used for TableRefs generated during query rewriting.
+  protected boolean allowEmptyOn_ = false;
 
   protected List<PlanHint> tableHints_ = new ArrayList<>();
   protected TReplicaPreference replicaPreference_;
   protected boolean randomReplica_;
+  // A sampling percent if convert_limit_to_sample hint is supplied. Valid range
+  // is 1-100. -1 indicates no hint.
+  protected int convertLimitToSampleHintPercent_;
 
   // Hinted distribution mode for this table ref; set after analyzeJoinHints()
   // TODO: Move join-specific members out of TableRef.
@@ -136,9 +144,27 @@ public class TableRef extends StmtNode {
   // columns via the view.
   protected boolean exposeNestedColumnsByTableMaskView_ = false;
 
+  // Scalar columns referenced in the query. Used in resolving column mask.
+  protected Map<String, Column> scalarColumns_ = new LinkedHashMap<>();
+
   // END: Members that need to be reset()
   /////////////////////////////////////////
 
+  // true if this table ref is hidden, because e.g. it was generated during statement
+  // rewrite.
+  private boolean isHidden_ = false;
+
+  /**
+   * Returns a new, resolved, and analyzed table ref.
+   */
+  public static TableRef newTableRef(Analyzer analyzer, List<String> rawPath,
+      String alias) throws AnalysisException {
+    TableRef ret = new TableRef(rawPath, alias);
+    ret = analyzer.resolveTableRef(ret);
+    ret.analyze(analyzer);
+    return ret;
+  }
+
   public TableRef(List<String> path, String alias) {
     this(path, alias, Privilege.SELECT);
   }
@@ -171,6 +197,7 @@ public class TableRef extends StmtNode {
     isAnalyzed_ = false;
     replicaPreference_ = null;
     randomReplica_ = false;
+    convertLimitToSampleHintPercent_ = -1;
   }
 
   /**
@@ -189,8 +216,10 @@ public class TableRef extends StmtNode {
     onClause_ = (other.onClause_ != null) ? other.onClause_.clone() : null;
     usingColNames_ =
         (other.usingColNames_ != null) ? Lists.newArrayList(other.usingColNames_) : null;
+    allowEmptyOn_ = other.allowEmptyOn_;
     tableHints_ = Lists.newArrayList(other.tableHints_);
     replicaPreference_ = other.replicaPreference_;
+    convertLimitToSampleHintPercent_ = other.convertLimitToSampleHintPercent_;
     randomReplica_ = other.randomReplica_;
     distrMode_ = other.distrMode_;
     // The table ref links are created at the statement level, so cloning a set of linked
@@ -202,6 +231,8 @@ public class TableRef extends StmtNode {
     correlatedTupleIds_ = Lists.newArrayList(other.correlatedTupleIds_);
     desc_ = other.desc_;
     exposeNestedColumnsByTableMaskView_ = other.exposeNestedColumnsByTableMaskView_;
+    scalarColumns_ = new LinkedHashMap<>(other.scalarColumns_);
+    isHidden_ = other.isHidden_;
   }
 
   @Override
@@ -233,6 +264,7 @@ public class TableRef extends StmtNode {
     this.tableHints_ = other.tableHints_;
     this.onClause_ = other.onClause_;
     this.usingColNames_ = other.usingColNames_;
+    this.allowEmptyOn_ = other.allowEmptyOn_;
   }
 
   public JoinOperator getJoinOp() {
@@ -240,8 +272,17 @@ public class TableRef extends StmtNode {
     return (joinOp_ == null ? JoinOperator.INNER_JOIN : joinOp_);
   }
 
+  public boolean allowEmptyOn() { return allowEmptyOn_ ; }
+  public void setAllowEmptyOn(boolean v) { allowEmptyOn_ = v; }
+
   public TReplicaPreference getReplicaPreference() { return replicaPreference_; }
   public boolean getRandomReplica() { return randomReplica_; }
+  public boolean hasConvertLimitToSampleHint() {
+    return convertLimitToSampleHintPercent_ != -1;
+  }
+  public int getConvertLimitToSampleHintPercent() {
+    return convertLimitToSampleHintPercent_;
+  }
 
   /**
    * Returns true if this table ref has a resolved path that is rooted at a registered
@@ -307,6 +348,9 @@ public class TableRef extends StmtNode {
     return exposeNestedColumnsByTableMaskView_;
   }
 
+  public void setHidden(boolean isHidden) { isHidden_ = isHidden; }
+  public boolean isHidden() { return isHidden_; }
+
   public void setJoinHints(List<PlanHint> hints) {
     Preconditions.checkNotNull(hints);
     joinHints_ = hints;
@@ -317,6 +361,10 @@ public class TableRef extends StmtNode {
     tableHints_ = hints;
   }
 
+  public void setTableSampleClause(TableSampleClause sampleParams) {
+    sampleParams_ = sampleParams;
+  }
+
   public boolean isBroadcastJoin() { return distrMode_ == DistributionMode.BROADCAST; }
 
   public boolean isPartitionedJoin() {
@@ -416,14 +464,36 @@ public class TableRef extends StmtNode {
       } else if (hint.is("SCHEDULE_RANDOM_REPLICA")) {
         analyzer.setHasPlanHints();
         randomReplica_ = true;
+      } else if (hint.is("CONVERT_LIMIT_TO_SAMPLE")) {
+        List<String> args = hint.getArgs();
+        if (args == null || args.size() != 1) {
+          addHintWarning(hint, analyzer);
+          return;
+        } else {
+          try {
+            int samplePercent = Integer.parseInt(args.get(0));
+            if (samplePercent < 1 || samplePercent > 100) {
+              addHintWarning(hint, analyzer);
+              return;
+            }
+            analyzer.setHasPlanHints();
+            convertLimitToSampleHintPercent_ = samplePercent;
+          } catch (NumberFormatException e) {
+            addHintWarning(hint, analyzer);
+          }
+        }
       } else {
-        Preconditions.checkState(getAliases() != null && getAliases().length > 0);
-        analyzer.addWarning("Table hint not recognized for table " + getUniqueAlias() +
-            ": " + hint);
+        addHintWarning(hint, analyzer);
       }
     }
   }
 
+  private void addHintWarning(PlanHint hint, Analyzer analyzer) {
+    Preconditions.checkState(getAliases() != null && getAliases().length > 0);
+    analyzer.addWarning("Table hint not recognized for table " + getUniqueAlias() +
+        ": " + hint);
+  }
+
   private void analyzeJoinHints(Analyzer analyzer) throws AnalysisException {
     if (joinHints_.isEmpty()) return;
     for (PlanHint hint: joinHints_) {
@@ -571,7 +641,7 @@ public class TableRef extends StmtNode {
         e.getIds(tupleIds, null);
         onClauseTupleIds.addAll(tupleIds);
       }
-    } else if (!isRelative() && !isCorrelated()
+    } else if (!allowEmptyOn() && !isRelative() && !isCorrelated()
         && (getJoinOp().isOuterJoin() || getJoinOp().isSemiJoin())) {
       throw new AnalysisException(
           joinOp_.toString() + " requires an ON or USING clause.");
@@ -651,6 +721,14 @@ public class TableRef extends StmtNode {
 
   public boolean isTableMaskingView() { return false; }
 
+  public void registerScalarColumn(Column column) {
+    scalarColumns_.put(column.getName(), column);
+  }
+
+  public List<Column> getScalarColumns() {
+    return new ArrayList<>(scalarColumns_.values());
+  }
+
   void migratePropertiesTo(TableRef other) {
     other.aliases_ = aliases_;
     other.onClause_ = onClause_;
