@@ -36,18 +36,24 @@ import org.apache.impala.catalog.CatalogException;
 import org.apache.impala.catalog.Column;
 import org.apache.impala.catalog.FeFsTable;
 import org.apache.impala.catalog.FeHBaseTable;
+import org.apache.impala.catalog.FeIcebergTable;
 import org.apache.impala.catalog.FeKuduTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.catalog.FeView;
 import org.apache.impala.catalog.Function;
 import org.apache.impala.catalog.HdfsCompression;
 import org.apache.impala.catalog.HdfsFileFormat;
+import org.apache.impala.catalog.IcebergColumn;
+import org.apache.impala.catalog.IcebergTable;
 import org.apache.impala.catalog.KuduColumn;
 import org.apache.impala.catalog.KuduTable;
 import org.apache.impala.catalog.RowFormat;
 import org.apache.impala.catalog.Table;
 import org.apache.impala.common.Pair;
+import org.apache.impala.thrift.TIcebergCatalog;
 import org.apache.impala.thrift.TSortingOrder;
+import org.apache.impala.util.AcidUtils;
+import org.apache.impala.util.IcebergUtil;
 import org.apache.impala.util.KuduUtil;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -243,6 +249,20 @@ public class ToSqlUtils {
     return result.toString();
   }
 
+  /**
+   * Gets the single IcebergPartitionSpec from a CreateTableStmt and returns its
+   * toSql(). If there is no IcebergPartitionSpec in the statement then returns null.
+   */
+  private static String getIcebergPartitionSpecsSql(CreateTableStmt stmt) {
+    List<IcebergPartitionSpec> partitionSpecs = stmt.getIcebergPartitionSpecs();
+    // It's safe to get the first element from 'partitionSpecs' as in create table
+    // statement only one is allowed.
+    if (!partitionSpecs.isEmpty() && partitionSpecs.get(0).hasPartitionFields()) {
+      return partitionSpecs.get(0).toSql();
+    }
+    return null;
+  }
+
   /**
    * Returns the "CREATE TABLE" SQL string corresponding to the given CreateTableStmt
    * statement.
@@ -264,13 +284,14 @@ public class ToSqlUtils {
     removeHiddenKuduTableProperties(generatedProperties);
     properties.putAll(generatedProperties);
     String kuduParamsSql = getKuduPartitionByParams(stmt);
+    String icebergPartitionSpecs = getIcebergPartitionSpecsSql(stmt);
     // TODO: Pass the correct compression, if applicable.
     return getCreateTableSql(stmt.getDb(), stmt.getTbl(), stmt.getComment(), colsSql,
         partitionColsSql, stmt.getTblPrimaryKeyColumnNames(), stmt.getForeignKeysSql(),
         kuduParamsSql, new Pair<>(stmt.getSortColumns(), stmt.getSortingOrder()),
         properties, stmt.getSerdeProperties(), stmt.isExternal(), stmt.getIfNotExists(),
         stmt.getRowFormat(), HdfsFileFormat.fromThrift(stmt.getFileFormat()),
-        HdfsCompression.NONE, null, stmt.getLocation());
+        HdfsCompression.NONE, null, stmt.getLocation(), icebergPartitionSpecs);
   }
 
   /**
@@ -298,6 +319,7 @@ public class ToSqlUtils {
     removeHiddenKuduTableProperties(generatedProperties);
     properties.putAll(generatedProperties);
     String kuduParamsSql = getKuduPartitionByParams(innerStmt);
+    String icebergPartitionSpecs = getIcebergPartitionSpecsSql(innerStmt);
     // TODO: Pass the correct compression, if applicable.
     String createTableSql = getCreateTableSql(innerStmt.getDb(), innerStmt.getTbl(),
         innerStmt.getComment(), null, partitionColsSql,
@@ -306,7 +328,7 @@ public class ToSqlUtils {
         innerStmt.getSortingOrder()), properties, innerStmt.getSerdeProperties(),
         innerStmt.isExternal(), innerStmt.getIfNotExists(), innerStmt.getRowFormat(),
         HdfsFileFormat.fromThrift(innerStmt.getFileFormat()), HdfsCompression.NONE, null,
-        innerStmt.getLocation());
+        innerStmt.getLocation(), icebergPartitionSpecs);
     return createTableSql + " AS " + stmt.getQueryStmt().toSql(options);
   }
 
@@ -332,11 +354,17 @@ public class ToSqlUtils {
     List<String> colsSql = new ArrayList<>();
     List<String> partitionColsSql = new ArrayList<>();
     boolean isHbaseTable = table instanceof FeHBaseTable;
+    boolean isFullAcid = AcidUtils.isFullAcidTable(
+        table.getMetaStoreTable().getParameters());
     for (int i = 0; i < table.getColumns().size(); i++) {
+      Column col = table.getColumns().get(i);
       if (!isHbaseTable && i < table.getNumClusteringCols()) {
-        partitionColsSql.add(columnToSql(table.getColumns().get(i)));
+        partitionColsSql.add(columnToSql(col));
+      } else if (isFullAcid && i == table.getNumClusteringCols()) {
+        Preconditions.checkState(col.getName().equals("row__id"));
+        continue;
       } else {
-        colsSql.add(columnToSql(table.getColumns().get(i)));
+        colsSql.add(columnToSql(col));
       }
     }
     RowFormat rowFormat = RowFormat.fromStorageDescriptor(msTable.getSd());
@@ -349,6 +377,7 @@ public class ToSqlUtils {
     List<String> primaryKeySql = new ArrayList<>();
     List<String> foreignKeySql = new ArrayList<>();
     String kuduPartitionByParams = null;
+    String icebergPartitions = null;
     if (table instanceof FeKuduTable) {
       FeKuduTable kuduTable = (FeKuduTable) table;
       // Kudu tables don't use LOCATION syntax
@@ -383,6 +412,24 @@ public class ToSqlUtils {
         colsSql = null;
       }
     } else if (table instanceof FeFsTable) {
+      if (table instanceof FeIcebergTable) {
+        storageHandlerClassName = null;
+        // Internal properties, should not be exposed to the user.
+        properties.remove(IcebergTable.KEY_STORAGE_HANDLER);
+        properties.remove(StatsSetupConst.DO_NOT_UPDATE_STATS);
+        properties.remove(IcebergTable.METADATA_LOCATION);
+
+        // Fill "PARTITION BY SPEC" part if the Iceberg table is partitioned.
+        FeIcebergTable feIcebergTable= (FeIcebergTable)table;
+        if (!feIcebergTable.getPartitionSpecs().isEmpty()) {
+          IcebergPartitionSpec latestPartitionSpec =
+              feIcebergTable.getDefaultPartitionSpec();
+          if (latestPartitionSpec.hasPartitionFields()) {
+            icebergPartitions = feIcebergTable.getDefaultPartitionSpec().toSql();
+          }
+        }
+      }
+
       String inputFormat = msTable.getSd().getInputFormat();
       format = HdfsFileFormat.fromHdfsInputFormatClass(inputFormat);
       compression = HdfsCompression.fromHdfsInputFormatClass(inputFormat);
@@ -393,12 +440,13 @@ public class ToSqlUtils {
         throw new CatalogException("Could not get primary key/foreign keys sql.", e);
       }
     }
+
     HdfsUri tableLocation = location == null ? null : new HdfsUri(location);
     return getCreateTableSql(table.getDb().getName(), table.getName(), comment, colsSql,
         partitionColsSql, primaryKeySql, foreignKeySql, kuduPartitionByParams,
         new Pair<>(sortColsSql, sortingOrder), properties, serdeParameters,
         isExternal, false, rowFormat, format, compression,
-        storageHandlerClassName, tableLocation);
+        storageHandlerClassName, tableLocation, icebergPartitions);
   }
 
   /**
@@ -413,7 +461,7 @@ public class ToSqlUtils {
       Map<String, String> tblProperties, Map<String, String> serdeParameters,
       boolean isExternal, boolean ifNotExists, RowFormat rowFormat,
       HdfsFileFormat fileFormat, HdfsCompression compression,
-      String storageHandlerClass, HdfsUri location) {
+      String storageHandlerClass, HdfsUri location, String icebergPartitions) {
     Preconditions.checkNotNull(tableName);
     StringBuilder sb = new StringBuilder("CREATE ");
     if (isExternal) sb.append("EXTERNAL ");
@@ -454,6 +502,11 @@ public class ToSqlUtils {
       sb.append(String.format("SORT BY %s (\n  %s\n)\n", sortProperties.second.toString(),
           Joiner.on(", \n  ").join(sortProperties.first)));
     }
+    if (icebergPartitions != null && !icebergPartitions.isEmpty()) {
+      sb.append("PARTITION BY SPEC\n");
+      sb.append(icebergPartitions);
+      sb.append("\n");
+    }
 
     if (tableComment != null) sb.append(" COMMENT '" + tableComment + "'\n");
 
@@ -475,8 +528,7 @@ public class ToSqlUtils {
     }
 
     if (storageHandlerClass == null) {
-      // TODO: Remove this special case when we have the LZO_TEXT writer
-      // We must handle LZO_TEXT specially because Impala does not yet support creating
+      // We must handle LZO_TEXT specially because Impala does not support creating
       // tables with this row format. In this case, we cannot output "WITH
       // SERDEPROPERTIES" because Hive does not support it with "STORED AS". For any
       // other HdfsFileFormat we want to output the serdeproperties because it is
@@ -500,7 +552,14 @@ public class ToSqlUtils {
             "WITH SERDEPROPERTIES " + propertyMapToSql(serdeParameters) + "\n");
       }
     }
-    if (location != null) {
+
+    // Iceberg table with 'hadoop.catalog' do not display table LOCATION when using
+    // 'show create table', user can use 'describe formatted/extended' to get location
+    TIcebergCatalog icebergCatalog =
+        IcebergUtil.getTIcebergCatalog(tblProperties.get(IcebergTable.ICEBERG_CATALOG));
+    boolean isHadoopCatalog = fileFormat == HdfsFileFormat.ICEBERG &&
+        icebergCatalog == TIcebergCatalog.HADOOP_CATALOG;
+    if (location != null && !isHadoopCatalog) {
       sb.append("LOCATION '" + location.toString() + "'\n");
     }
     if (tblProperties != null && !tblProperties.isEmpty()) {
@@ -547,6 +606,10 @@ public class ToSqlUtils {
       if (kuduCol.getBlockSize() != 0) {
         sb.append(String.format(" BLOCK_SIZE %d", kuduCol.getBlockSize()));
       }
+    } else if (col instanceof IcebergColumn) {
+      IcebergColumn icebergCol = (IcebergColumn) col;
+      Boolean isNullable = icebergCol.isNullable();
+      if (isNullable != null) sb.append(isNullable ? " NULL" : " NOT NULL");
     }
     if (!Strings.isNullOrEmpty(col.getComment())) {
       sb.append(String.format(" COMMENT '%s'", col.getComment()));
