@@ -41,6 +41,7 @@ import org.apache.impala.common.TransactionKeepalive.HeartbeatContext;
 import org.apache.impala.compat.MetastoreShim;
 import org.apache.impala.thrift.TCatalogObject;
 import org.apache.impala.thrift.TFunction;
+import org.apache.impala.thrift.THdfsPartition;
 import org.apache.impala.thrift.TPartitionKeyValue;
 import org.apache.impala.thrift.TPrincipalType;
 import org.apache.impala.thrift.TTable;
@@ -123,6 +124,11 @@ public abstract class Catalog implements AutoCloseable {
     this(new MetaStoreClientPool(0, 0));
   }
 
+  /**
+   * Returns the Hive ACID user id used by this catalog.
+   */
+  public abstract String getAcidUserId();
+
   /**
    * Adds a new database to the catalog, replacing any existing database with the same
    * name.
@@ -504,16 +510,17 @@ public abstract class Catalog implements AutoCloseable {
           throw new CatalogException("Table not found: " +
               objectDesc.getTable().getTbl_name());
         }
-        table.getLock().lock();
+        table.takeReadLock();
         try {
           result.setType(table.getCatalogObjectType());
           result.setCatalog_version(table.getCatalogVersion());
           result.setTable(table.toThrift());
         } finally {
-          table.getLock().unlock();
+          table.releaseReadLock();
         }
         break;
       }
+      // TODO(IMPALA-9935): support HDFS_PARTITION
       case FUNCTION: {
         TFunction tfn = objectDesc.getFn();
         Function desc = Function.fromThrift(tfn);
@@ -624,6 +631,11 @@ public abstract class Catalog implements AutoCloseable {
         TTable tbl = catalogObject.getTable();
         return "TABLE:" + tbl.getDb_name().toLowerCase() + "." +
             tbl.getTbl_name().toLowerCase();
+      case HDFS_PARTITION:
+        THdfsPartition part = catalogObject.getHdfs_partition();
+        return "HDFS_PARTITION:" + part.getDb_name().toLowerCase() + "." +
+            part.getTbl_name().toLowerCase() + ":" +
+            Preconditions.checkNotNull(part.getPartition_name());
       case FUNCTION:
         return "FUNCTION:" + catalogObject.getFn().getName() + "(" +
             catalogObject.getFn().getSignature() + ")";
@@ -677,7 +689,7 @@ public abstract class Catalog implements AutoCloseable {
    */
   public Transaction openTransaction(IMetaStoreClient hmsClient, HeartbeatContext ctx)
       throws TransactionException {
-    return new Transaction(hmsClient, transactionKeepalive_, "Impala Catalog", ctx);
+    return new Transaction(hmsClient, transactionKeepalive_, getAcidUserId(), ctx);
   }
 
   /**
