@@ -32,6 +32,7 @@ import org.apache.impala.analysis.KuduPartitionParam;
 import org.apache.impala.catalog.events.InFlightEvents;
 import org.apache.impala.common.ImpalaException;
 import org.apache.impala.common.ImpalaRuntimeException;
+import org.apache.impala.thrift.TBriefTableMeta;
 import org.apache.impala.thrift.TCatalogObject;
 import org.apache.impala.thrift.TCatalogObjectType;
 import org.apache.impala.thrift.TDatabase;
@@ -53,6 +54,8 @@ import com.google.common.base.Preconditions;
 import com.google.common.collect.ImmutableList;
 import com.google.common.collect.Lists;
 
+import static org.apache.impala.service.MetadataOp.TABLE_COMMENT_KEY;
+
 /**
  * Internal representation of db-related metadata. Owned by Catalog instance.
  * Not thread safe.
@@ -462,6 +465,12 @@ public class Db extends CatalogObjectImpl implements FeDb {
     catalogObject.setDb(toThrift());
   }
 
+  public TCatalogObject toMinimalTCatalogObject() {
+    TCatalogObject min = new TCatalogObject(getCatalogObjectType(), getCatalogVersion());
+    min.setDb(new TDatabase(getName()));
+    return min;
+  }
+
   /**
    * Get partial information about this DB in order to service CatalogdMetaProvider
    * running in a remote impalad.
@@ -480,8 +489,19 @@ public class Db extends CatalogObjectImpl implements FeDb {
       // instead to avoid this copy.
       resp.db_info.hms_database = getMetaStoreDb().deepCopy();
     }
-    if (selector.want_table_names) {
-      resp.db_info.table_names = getAllTableNames();
+    if (selector.want_brief_meta_of_tables) {
+      List<TBriefTableMeta> briefTableMetaList = Lists.newArrayListWithCapacity(
+          tableCache_.keySet().size());
+      for (Table tbl : tableCache_.getValues()) {
+        TBriefTableMeta meta = new TBriefTableMeta(tbl.getName());
+        org.apache.hadoop.hive.metastore.api.Table msTbl = tbl.getMetaStoreTable();
+        if (msTbl != null) {
+          meta.setMsType(msTbl.getTableType());
+          meta.setComment(msTbl.getParameters().get(TABLE_COMMENT_KEY));
+        }
+        briefTableMetaList.add(meta);
+      }
+      resp.db_info.brief_meta_of_tables = briefTableMetaList;
     }
     if (selector.want_function_names) {
       resp.db_info.function_names = ImmutableList.copyOf(functions_.keySet());
@@ -511,7 +531,7 @@ public class Db extends CatalogObjectImpl implements FeDb {
     Preconditions.checkState(dbLock_.isHeldByCurrentThread(),
         "removeFromVersionsForInflightEvents called without getting the db lock for "
             + getName() + " database.");
-    return inFlightEvents_.remove(versionNumber);
+    return inFlightEvents_.remove(false, versionNumber);
   }
 
   /**
@@ -525,7 +545,7 @@ public class Db extends CatalogObjectImpl implements FeDb {
     Preconditions.checkState(dbLock_.isHeldByCurrentThread(),
         "addToVersionsForInFlightEvents called without getting the db lock for "
             + getName() + " database.");
-    if (!inFlightEvents_.add(versionNumber)) {
+    if (!inFlightEvents_.add(false, versionNumber)) {
       LOG.warn(String.format("Could not add version %s to the list of in-flight "
           + "events. This could cause unnecessary database %s invalidation when the "
           + "event is processed", versionNumber, getName()));
