@@ -25,8 +25,9 @@ import java.util.Map;
 import java.util.Set;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicLong;
-import java.util.concurrent.locks.ReentrantLock;
-
+import java.util.concurrent.locks.ReentrantReadWriteLock;
+import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
+import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
 import org.apache.hadoop.hive.common.ValidWriteIdList;
 import org.apache.hadoop.hive.metastore.IMetaStoreClient;
 import org.apache.hadoop.hive.metastore.TableType;
@@ -40,9 +41,7 @@ import org.apache.impala.compat.MetastoreShim;
 import org.apache.impala.common.ImpalaRuntimeException;
 import org.apache.impala.common.Metrics;
 import org.apache.impala.common.Pair;
-import org.apache.impala.common.PrintUtils;
 import org.apache.impala.common.RuntimeEnv;
-import org.apache.impala.compat.MetastoreShim;
 import org.apache.impala.service.MetadataOp;
 import org.apache.impala.thrift.TAccessLevel;
 import org.apache.impala.thrift.TCatalogObject;
@@ -62,8 +61,6 @@ import org.apache.log4j.Logger;
 
 import com.codahale.metrics.Timer;
 import com.google.common.base.Preconditions;
-import com.google.common.base.Stopwatch;
-import com.google.common.collect.ImmutableList;
 import com.google.common.collect.Lists;
 
 /**
@@ -82,8 +79,14 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
   protected final String name_;
   protected final String owner_;
   protected TAccessLevel accessLevel_ = TAccessLevel.READ_WRITE;
-  // Lock protecting this table
-  private final ReentrantLock tableLock_ = new ReentrantLock();
+  // Lock protecting this table. A read lock must be table when we are serializing
+  // the table contents over thrift (e.g when returning the table to clients over thrift
+  // or when topic-update thread serializes the table in the topic update)
+  // A write lock must be table when the table is being modified (e.g. DDLs or refresh)
+  private final ReentrantReadWriteLock tableLock_ = new ReentrantReadWriteLock(
+      true /*fair ordering*/);
+  private final ReadLock readLock_ = tableLock_.readLock();
+  private final WriteLock writeLock_ = tableLock_.writeLock();
 
   // Number of clustering columns.
   protected int numClusteringCols_;
@@ -114,7 +117,7 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
   protected final ArrayList<Column> colsByPos_ = new ArrayList<>();
 
   // map from lowercase column name to Column object.
-  private final Map<String, Column> colsByName_ = new HashMap<>();
+  protected final Map<String, Column> colsByName_ = new HashMap<>();
 
   // List of SQL constraints associated with the table.
   private final SqlConstraints sqlConstraints_ = new SqlConstraints(new ArrayList<>(),
@@ -134,11 +137,6 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
   // impalad.
   protected long lastUsedTime_;
 
-  // Valid write id list for this table.
-  // null in the case that this table is not transactional.
-  // TODO(todd) this should probably be a ValidWriteIdList in memory instead of a String.
-  protected String validWriteIds_ = null;
-
   // tracks the in-flight metastore events for this table. Used by Events processor to
   // avoid unnecessary refresh when the event is received
   private final InFlightEvents inFlightEvents = new InFlightEvents();
@@ -212,7 +210,52 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
         .parseBoolean(msTbl.getParameters().get(TBL_PROP_EXTERNAL_TABLE_PURGE));
   }
 
-  public ReentrantLock getLock() { return tableLock_; }
+  public void takeReadLock() {
+    readLock_.lock();
+  }
+
+  public ReadLock readLock() { return readLock_; }
+  public WriteLock writeLock() { return writeLock_; }
+
+  public void releaseReadLock() {
+    readLock_.unlock();
+  }
+
+  public boolean isReadLockedByCurrentThread() {
+    return tableLock_.getReadHoldCount() > 0;
+  }
+
+  public boolean tryReadLock() {
+    try {
+      // a tryLock with a 0 timeout honors the fairness for lock acquisition
+      // in case there are other threads waiting to acquire a read lock when
+      // compared to tryLock() which "barges" in and takes a read lock if
+      // available with no consideration to other threads waiting for a read lock.
+      return readLock_.tryLock(0, TimeUnit.SECONDS);
+    } catch (InterruptedException e) {
+      // ignored
+    }
+    return false;
+  }
+
+  public void releaseWriteLock() {
+    writeLock_.unlock();
+  }
+
+  /**
+   * Returns true if this table is write locked by current thread.
+   */
+  public boolean isWriteLockedByCurrentThread() {
+    return writeLock_.isHeldByCurrentThread();
+  }
+
+  /**
+   * Returns true if this table is write locked by any thread.
+   */
+  public boolean isWriteLocked() {
+    return tableLock_.isWriteLocked();
+  }
+
   @Override
   public abstract TTableDescriptor toThriftDescriptor(
       int tableId, Set<Long> referencedPartitions);
@@ -379,48 +422,6 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
     }
   }
 
-  /**
-   * Get valid write ids for the acid table.
-   * @param client the client to access HMS
-   * @return the list of valid write IDs for the table
-   */
-  protected String fetchValidWriteIds(IMetaStoreClient client)
-      throws TableLoadingException {
-    String tblFullName = getFullName();
-    if (LOG.isTraceEnabled()) LOG.trace("Get valid writeIds for table: " + tblFullName);
-    String writeIds = null;
-    try {
-      ValidWriteIdList validWriteIds = MetastoreShim.fetchValidWriteIds(client,
-          tblFullName);
-      writeIds = validWriteIds == null ? null : validWriteIds.writeToString();
-    } catch (Exception e) {
-      throw new TableLoadingException(String.format("Error loading ValidWriteIds for " +
-          "table '%s'", getName()), e);
-    }
-    if (LOG.isTraceEnabled()) {
-      LOG.trace("Valid writeIds: " + writeIds);
-    }
-    return writeIds;
-  }
-
-  /**
-   * Set ValistWriteIdList with stored writeId
-   * @param client the client to access HMS
-   */
-  protected void loadValidWriteIdList(IMetaStoreClient client)
-      throws TableLoadingException {
-    Stopwatch sw = new Stopwatch().start();
-    Preconditions.checkState(msTable_ != null && msTable_.getParameters() != null);
-    if (MetastoreShim.getMajorVersion() > 2 &&
-        AcidUtils.isTransactionalTable(msTable_.getParameters())) {
-      validWriteIds_ = fetchValidWriteIds(client);
-    } else {
-      validWriteIds_ = null;
-    }
-    LOG.debug("Load Valid Write Id List Done. Time taken: " +
-        PrintUtils.printTimeNs(sw.elapsed(TimeUnit.NANOSECONDS)));
-  }
-
   /**
    * Creates a table of the appropriate type based on the given hive.metastore.api.Table
    * object.
@@ -437,6 +438,8 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
       table = new HBaseTable(msTbl, db, msTbl.getTableName(), msTbl.getOwner());
     } else if (KuduTable.isKuduTable(msTbl)) {
       table = new KuduTable(msTbl, db, msTbl.getTableName(), msTbl.getOwner());
+    } else if (IcebergTable.isIcebergTable(msTbl)) {
+      table = new IcebergTable(msTbl, db, msTbl.getTableName(), msTbl.getOwner());
     } else if (DataSourceTable.isDataSourceTable(msTbl)) {
       // It's important to check if this is a DataSourceTable before HdfsTable because
       // DataSourceTables are still represented by HDFS tables in the metastore but
@@ -485,8 +488,7 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
         Column col = Column.fromThrift(columns.get(i));
         colsByPos_.add(col.getPosition(), col);
         colsByName_.put(col.getName().toLowerCase(), col);
-        ((StructType) type_.getItemType()).addField(
-            new StructField(col.getName(), col.getType(), col.getComment()));
+        ((StructType) type_.getItemType()).addField(getStructFieldFromColumn(col));
       }
     } catch (ImpalaRuntimeException e) {
       throw new TableLoadingException(String.format("Error loading schema for " +
@@ -503,8 +505,20 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
     storageMetadataLoadTime_ = thriftTable.getStorage_metadata_load_time_ns();
 
     storedInImpaladCatalogCache_ = true;
-    validWriteIds_ = thriftTable.isSetValid_write_ids() ?
-        thriftTable.getValid_write_ids() : null;
+  }
+
+  /**
+   * If column is 'IcebergColumn', we return 'IcebergStructField', otherwise, we
+   * just return 'StructField'.
+   */
+  private StructField getStructFieldFromColumn(Column col) {
+    if (col instanceof IcebergColumn) {
+      IcebergColumn iCol = (IcebergColumn) col;
+      return new IcebergStructField(iCol.getName(), iCol.getType(),
+          iCol.getComment(), iCol.getFieldId());
+    } else {
+      return new StructField(col.getName(), col.getType(), col.getComment());
+    }
   }
 
   /**
@@ -530,7 +544,7 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
     // the table lock should already be held, and we want the toThrift() to be consistent
     // with the modification. So this check helps us identify places where the lock
     // acquisition is probably missing entirely.
-    if (!tableLock_.isHeldByCurrentThread()) {
+    if (!isLockedByCurrentThread()) {
       throw new IllegalStateException(
           "Table.toThrift() called without holding the table lock: " +
               getFullName() + " " + getClass().getName());
@@ -553,32 +567,53 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
       }
     }
 
-    table.setMetastore_table(getMetaStoreTable());
+    org.apache.hadoop.hive.metastore.api.Table msTable = getMetaStoreTable();
+    // IMPALA-10243: We should get our own copy of the metastore table, otherwise other
+    // threads might modify it when the table lock is not held.
+    if (msTable != null) msTable = msTable.deepCopy();
+    table.setMetastore_table(msTable);
     table.setTable_stats(tableStats_);
-    if (validWriteIds_ != null) {
-      table.setValid_write_ids(validWriteIds_);
-    }
     return table;
   }
 
+  private boolean isLockedByCurrentThread() {
+    return isReadLockedByCurrentThread() || tableLock_.isWriteLockedByCurrentThread();
+  }
+
   public TCatalogObject toMinimalTCatalogObject() {
+    return toMinimalTCatalogObjectHelper();
+  }
+
+  /**
+   * Helper method for generating the minimal catalog object. toMinimalTCatalogObject()
+   * may be overrided by subclasses so we put the general implementation here.
+   */
+  private TCatalogObject toMinimalTCatalogObjectHelper() {
     TCatalogObject catalogObject =
         new TCatalogObject(getCatalogObjectType(), getCatalogVersion());
-    catalogObject.setTable(new TTable());
-    catalogObject.getTable().setDb_name(getDb().getName());
-    catalogObject.getTable().setTbl_name(getName());
+    catalogObject.setTable(new TTable(getDb().getName(), getName()));
     return catalogObject;
   }
 
+  /**
+   * Returns a TCatalogObject with only the table name for invalidation. For non-hdfs
+   * tables, it's the same as toMinimalTCatalogObject(). For hdfs tables, their
+   * toMinimalTCatalogObject() will also return the partition names. So we use this method
+   * to get a light-weight object for invalidation in LocalCatalog.
+   */
+  public TCatalogObject toInvalidationObject() {
+    return toMinimalTCatalogObjectHelper();
+  }
+
   /**
    * Override parent implementation that will finally call toThrift() which requires
    * holding the table lock. However, it's not guaranteed that caller holds the table
-   * lock (IMPALA-9136). Here we use toMinimalTCatalogObject() directly since only db
-   * name and table name are needed.
+   * lock (IMPALA-9136). Here we use toMinimalTCatalogObjectHelper() directly since only
+   * db name and table name are needed.
    */
   @Override
   public final String getUniqueName() {
-    return Catalog.toCatalogObjectKey(toMinimalTCatalogObject());
+    return Catalog.toCatalogObjectKey(toMinimalTCatalogObjectHelper());
   }
 
   @Override
@@ -586,12 +621,27 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
     catalogObject.setTable(toThrift());
   }
 
+  /**
+   * Generates a TCatalogObject based on the required form. See more details in comments
+   * of ThriftObjectType.
+   */
+  public TCatalogObject toTCatalogObject(ThriftObjectType resultType) {
+    switch (resultType) {
+      case FULL: return toTCatalogObject();
+      case DESCRIPTOR_ONLY: return toMinimalTCatalogObject();
+      case INVALIDATION: return toInvalidationObject();
+      case NONE:
+      default:
+        return null;
+    }
+  }
+
   /**
    * Return partial info about this table. This is called only on the catalogd to
    * service GetPartialCatalogObject RPCs.
    */
   public TGetPartialCatalogObjectResponse getPartialInfo(
-      TGetPartialCatalogObjectRequest req) throws TableLoadingException {
+      TGetPartialCatalogObjectRequest req) throws CatalogException {
     Preconditions.checkState(isLoaded(), "unloaded table: %s", getFullName());
     TTableInfoSelector selector = Preconditions.checkNotNull(req.table_info_selector,
         "no table_info_selector");
@@ -633,7 +683,12 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
       }
       resp.table_info.setColumn_stats(statsList);
     }
-
+    if (getMetaStoreTable() != null &&
+        AcidUtils.isTransactionalTable(getMetaStoreTable().getParameters())) {
+      Preconditions.checkState(getValidWriteIds() != null);
+      resp.table_info.setValid_write_ids(
+          MetastoreShim.convertToTValidWriteIdList(getValidWriteIds()));
+    }
     return resp;
   }
   /**
@@ -683,6 +738,12 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
   @Override // FeTable
   public List<Column> getColumnsInHiveOrder() {
     List<Column> columns = Lists.newArrayList(getNonClusteringColumns());
+    if (getMetaStoreTable() != null &&
+        AcidUtils.isFullAcidTable(getMetaStoreTable().getParameters())) {
+      // Remove synthetic "row__id" column.
+      Preconditions.checkState(columns.get(0).getName().equals("row__id"));
+      columns.remove(0);
+    }
     columns.addAll(getClusteringColumns());
     return Collections.unmodifiableList(columns);
   }
@@ -795,14 +856,19 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
 
   /**
    * Removes a given version from the collection of version numbers for in-flight events
-   * @param versionNumber version number to remove from the collection
+   * @param isInsertEvent If true, remove eventId from list of eventIds for in-flight
+   * Insert events. If false, remove versionNumber from list of versions for in-flight DDL
+   * events
+   * @param versionNumber when isInsertEvent is true, it's eventId to remove
+   * when isInsertEvent is false, it's version number to remove
    * @return true if version was successfully removed, false if didn't exist
    */
-  public boolean removeFromVersionsForInflightEvents(long versionNumber) {
-    Preconditions.checkState(tableLock_.isHeldByCurrentThread(),
+  public boolean removeFromVersionsForInflightEvents(
+      boolean isInsertEvent, long versionNumber) {
+    Preconditions.checkState(isWriteLockedByCurrentThread(),
         "removeFromVersionsForInFlightEvents called without taking the table lock on "
             + getFullName());
-    return inFlightEvents.remove(versionNumber);
+    return inFlightEvents.remove(isInsertEvent, versionNumber);
   }
 
   /**
@@ -810,16 +876,19 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
    * collection is already at the max size defined by
    * <code>MAX_NUMBER_OF_INFLIGHT_EVENTS</code>, then it ignores the given version and
    * does not add it
-   * @param versionNumber version number to add
+   * @param isInsertEvent if true, add eventId to list of eventIds for in-flight Insert
+   * events. If false, add versionNumber to list of versions for in-flight DDL events
+   * @param versionNumber when isInsertEvent is true, it's eventId to add
+   * when isInsertEvent is false, it's version number to add
    * @return True if version number was added, false if the collection is at its max
    * capacity
    */
-  public void addToVersionsForInflightEvents(long versionNumber) {
+  public void addToVersionsForInflightEvents(boolean isInsertEvent, long versionNumber) {
     // we generally don't take locks on Incomplete tables since they are atomically
     // replaced during load
     Preconditions.checkState(
-        this instanceof IncompleteTable || tableLock_.isHeldByCurrentThread());
-    if (!inFlightEvents.add(versionNumber)) {
+        this instanceof IncompleteTable || isWriteLockedByCurrentThread());
+    if (!inFlightEvents.add(isInsertEvent, versionNumber)) {
       LOG.warn(String.format("Could not add %s version to the table %s. This could "
           + "cause unnecessary refresh of the table when the event is received by the "
               + "Events processor.", versionNumber, getFullName()));
@@ -832,7 +901,22 @@ public abstract class Table extends CatalogObjectImpl implements FeTable {
   }
 
   @Override
-  public String getValidWriteIds() {
-    return validWriteIds_;
+  public ValidWriteIdList getValidWriteIds() {
+    return null;
   }
+
+  /**
+   * When altering a table we modify the cached metadata and apply the new metadata to
+   * HMS. Some DDL/DMLs require reloading the HMS and file metadata to update the cached
+   * metadata (See CatalogOpExecutor#alterTable(TAlterTableParams, TDdlExecResponse) for
+   * more details). Before reloading the metadata, these modifications are not finalized.
+   * @return true if there are any in-progress modifications need to be finalized in an
+   * incremental table reload.
+   */
+  public boolean hasInProgressModification() { return false; }
+
+  /**
+   * Clears the in-progress modifications in case of failures.
+   */
+  public void resetInProgressModification() { }
 }
