@@ -187,6 +187,7 @@ public abstract class Type {
   public boolean isDate() { return isScalarType(PrimitiveType.DATE); }
   public boolean isDecimal() { return isScalarType(PrimitiveType.DECIMAL); }
   public boolean isFullySpecifiedDecimal() { return false; }
+  public boolean isVarchar() { return isScalarType(PrimitiveType.VARCHAR); }
   public boolean isWildcardDecimal() { return false; }
   public boolean isWildcardVarchar() { return false; }
   public boolean isWildcardChar() { return false; }
@@ -469,7 +470,13 @@ public abstract class Type {
           if (thriftField.isSetComment()) comment = thriftField.getComment();
           Pair<Type, Integer> res = fromThrift(col, nodeIdx);
           nodeIdx = res.second.intValue();
-          structFields.add(new StructField(name, res.first, comment));
+          if (thriftField.isSetField_id()) {
+            // We create 'IcebergStructField' for Iceberg tables which have field id.
+            structFields.add(new IcebergStructField(name, res.first, comment,
+                thriftField.getField_id()));
+          } else {
+            structFields.add(new StructField(name, res.first, comment));
+          }
         }
         type = new StructType(structFields);
         break;
