@@ -213,8 +213,8 @@ public class MetastoreEvents {
         String eventDb = currentEvent.getDbName();
         String eventTbl = currentEvent.getTableName();
         // if the event is on blacklisted db or table we should filter it out
-        if (catalog_.isBlacklistedDb(eventDb) || (eventTbl != null && catalog_
-            .isBlacklistedTable(eventDb, eventTbl))) {
+        if ((eventDb != null && catalog_.isBlacklistedDb(eventDb)) || (eventTbl != null
+            && catalog_.isBlacklistedTable(eventDb, eventTbl))) {
           String blacklistedObject = eventTbl != null ? new TableName(eventDb,
               eventTbl).toString() : eventDb;
           LOG.info(currentEvent.debugString("Filtering out this event since it is on a "
@@ -269,7 +269,7 @@ public class MetastoreEvents {
     protected final String tblName_;
 
     // eventId of the event. Used instead of calling getter on event_ everytime
-    protected final long eventId_;
+    protected long eventId_;
 
     // eventType from the NotificationEvent
     protected final MetastoreEventType eventType_;
@@ -419,13 +419,15 @@ public class MetastoreEvents {
      * serviceId and version. More details on complete flow of self-event handling
      * logic can be read in <code>MetastoreEventsProcessor</code> documentation.
      *
+     * @param isInsertEvent if true, check in flight events list of Insert event
+     * if false, check events list of DDL
      * @return True if this event is a self-generated event. If the returned value is
      * true, this method also clears the version number from the catalog database/table.
      * Returns false if the version numbers or service id don't match
      */
-    protected boolean isSelfEvent() {
+    protected boolean isSelfEvent(boolean isInsertEvent) {
       try {
-        if (catalog_.evaluateSelfEvent(getSelfEventContext())) {
+        if (catalog_.evaluateSelfEvent(isInsertEvent, getSelfEventContext())) {
           metrics_.getCounter(MetastoreEventsProcessor.NUMBER_OF_SELF_EVENTS).inc();
           return true;
         }
@@ -434,6 +436,8 @@ public class MetastoreEvents {
       }
       return false;
     }
+
+    protected boolean isSelfEvent() { return isSelfEvent(false); }
   }
 
   public static String getStringProperty(
@@ -555,17 +559,10 @@ public class MetastoreEvents {
      */
     protected boolean reloadTableFromCatalog(String operation, boolean isTransactional)
         throws CatalogException {
-      try {
-        if (!catalog_.reloadTableIfExists(dbName_, tblName_,
-            "Processing " + operation + " event from HMS")) {
-          debugLog("Automatic refresh on table {} failed as the table "
-              + "either does not exist anymore or is not in loaded state.",
-              getFullyQualifiedTblName());
-          return false;
-        }
-      } catch (DatabaseNotFoundException e) {
-        debugLog("Refresh table {} failed as "
-                + "the database was not present in the catalog.",
+      if (!catalog_.reloadTableIfExists(dbName_, tblName_,
+          "Processing " + operation + " event from HMS")) {
+        debugLog("Automatic refresh on table {} failed as the table "
+            + "either does not exist anymore or is not in loaded state.",
             getFullyQualifiedTblName());
         return false;
       }
@@ -766,19 +763,25 @@ public class MetastoreEvents {
 
     @Override
     public SelfEventContext getSelfEventContext() {
-      throw new UnsupportedOperationException("Self-event evaluation is not implemented"
-          + " for insert event type");
+      if (insertPartition_ != null) {
+        // create selfEventContext for insert partition event
+        List<TPartitionKeyValue> tPartSpec =
+            getTPartitionSpecFromHmsPartition(msTbl_, insertPartition_);
+        return new SelfEventContext(dbName_, tblName_, Arrays.asList(tPartSpec),
+            insertPartition_.getParameters(), eventId_);
+      } else {
+        // create selfEventContext for insert table event
+        return new SelfEventContext(
+            dbName_, tblName_, null, msTbl_.getParameters(), eventId_);
+      }
     }
 
-    /**
-     * Currently we do not check for self-events in Inserts. Existing self-events logic
-     * cannot be used for insert events since firing insert event does not allow us to
-     * modify table parameters in HMS. Hence, we cannot get CatalogServiceIdentifiers in
-     * Insert Events.
-     * TODO: Handle self-events for insert case.
-     */
     @Override
     public void process() throws MetastoreNotificationException {
+      if (isSelfEvent(true)) {
+        infoLog("Not processing the event as it is a self-event");
+        return;
+      }
       // Reload the whole table if it's a transactional table.
       if (AcidUtils.isTransactionalTable(msTbl_.getParameters())) {
         insertPartition_ = null;
