@@ -34,11 +34,13 @@ import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.atomic.AtomicReference;
 
+import com.google.common.collect.ImmutableCollection;
+import com.google.common.collect.ImmutableMap;
+import org.apache.commons.lang3.StringUtils;
 import org.apache.hadoop.hive.metastore.api.ColumnStatisticsObj;
 import org.apache.hadoop.hive.metastore.api.Database;
 import org.apache.hadoop.hive.metastore.api.MetaException;
 import org.apache.hadoop.hive.metastore.api.NoSuchObjectException;
-import org.apache.hadoop.hive.metastore.api.Partition;
 import org.apache.hadoop.hive.metastore.api.SQLForeignKey;
 import org.apache.hadoop.hive.metastore.api.SQLPrimaryKey;
 import org.apache.hadoop.hive.metastore.api.Table;
@@ -51,7 +53,10 @@ import org.apache.impala.catalog.CatalogDeltaLog;
 import org.apache.impala.catalog.CatalogException;
 import org.apache.impala.catalog.CatalogObjectCache;
 import org.apache.impala.catalog.Function;
+import org.apache.impala.catalog.HdfsCachePool;
 import org.apache.impala.catalog.HdfsPartition.FileDescriptor;
+import org.apache.impala.catalog.HdfsPartitionLocationCompressor;
+import org.apache.impala.catalog.HdfsStorageDescriptor;
 import org.apache.impala.catalog.ImpaladCatalog.ObjectUpdateSequencer;
 import org.apache.impala.catalog.Principal;
 import org.apache.impala.catalog.PrincipalPrivilege;
@@ -60,8 +65,10 @@ import org.apache.impala.common.InternalException;
 import org.apache.impala.common.Pair;
 import org.apache.impala.service.FeSupport;
 import org.apache.impala.service.FrontendProfile;
+import org.apache.impala.service.MetadataOp;
 import org.apache.impala.thrift.CatalogLookupStatus;
 import org.apache.impala.thrift.TBackendGflags;
+import org.apache.impala.thrift.TBriefTableMeta;
 import org.apache.impala.thrift.TCatalogInfoSelector;
 import org.apache.impala.thrift.TCatalogObject;
 import org.apache.impala.thrift.TCatalogObjectType;
@@ -81,6 +88,7 @@ import org.apache.impala.thrift.TUniqueId;
 import org.apache.impala.thrift.TUnit;
 import org.apache.impala.thrift.TUpdateCatalogCacheRequest;
 import org.apache.impala.thrift.TUpdateCatalogCacheResponse;
+import org.apache.impala.thrift.TValidWriteIdList;
 import org.apache.impala.util.ListMap;
 import org.apache.impala.util.TByteBuffer;
 import org.apache.thrift.TDeserializer;
@@ -108,6 +116,8 @@ import com.google.common.util.concurrent.Uninterruptibles;
 import com.google.errorprone.annotations.Immutable;
 import com.google.errorprone.annotations.concurrent.GuardedBy;
 
+import static org.apache.impala.service.MetadataOp.TABLE_COMMENT_KEY;
+
 /**
  * MetaProvider which fetches metadata in a granular fashion from the catalogd.
  *
@@ -203,7 +213,7 @@ public class CatalogdMetaProvider implements MetaProvider {
   private static final String CATALOG_FETCH_PREFIX = "CatalogFetch";
   private static final String DB_LIST_STATS_CATEGORY = "DatabaseList";
   private static final String DB_METADATA_STATS_CATEGORY = "Databases";
-  private static final String TABLE_NAMES_STATS_CATEGORY = "TableNames";
+  private static final String TABLE_LIST_STATS_CATEGORY = "TableList";
   private static final String TABLE_METADATA_CACHE_CATEGORY = "Tables";
   private static final String PARTITION_LIST_STATS_CATEGORY = "PartitionLists";
   private static final String PARTITIONS_STATS_CATEGORY = "Partitions";
@@ -320,6 +330,11 @@ public class CatalogdMetaProvider implements MetaProvider {
       new CatalogObjectCache<>();
   private AtomicReference<? extends AuthorizationChecker> authzChecker_;
 
+  // Cache of known HDFS cache pools. Allows for checking the existence
+  // of pools without hitting HDFS. This is _not_ "fetch-on-demand".
+  private final CatalogObjectCache<HdfsCachePool> hdfsCachePools_ =
+      new CatalogObjectCache<>(false);
+
   public CatalogdMetaProvider(TBackendGflags flags) {
     Preconditions.checkArgument(flags.isSetLocal_catalog_cache_expiration_s());
     Preconditions.checkArgument(flags.isSetLocal_catalog_cache_mb());
@@ -350,6 +365,11 @@ public class CatalogdMetaProvider implements MetaProvider {
     return cache_.stats();
   }
 
+  @Override
+  public Iterable<HdfsCachePool> getHdfsCachePools() {
+    return hdfsCachePools_;
+  }
+
   @Override
   public AuthorizationPolicy getAuthPolicy() {
     return authPolicy_;
@@ -375,7 +395,7 @@ public class CatalogdMetaProvider implements MetaProvider {
       throws TException {
     TGetPartialCatalogObjectResponse resp;
     byte[] ret = null;
-    Stopwatch sw = new Stopwatch().start();
+    Stopwatch sw = Stopwatch.createStarted();
     try {
       ret = FeSupport.GetPartialCatalogObject(new TSerializer().serialize(req));
     } catch (InternalException e) {
@@ -445,7 +465,7 @@ public class CatalogdMetaProvider implements MetaProvider {
     // around invalidation (IMPALA-7534). Namely, we have the following interleaving to
     // worry about:
     //
-    //  Thread 1: loadTableNames() misses and sends a request to fetch table names
+    //  Thread 1: loadTableList() misses and sends a request to fetch table names
     //  Catalogd: sends a response with table list ['foo']
     //  Thread 2:    creates a table 'bar'
     //  Catalogd:    returns an invalidation for the table name list
@@ -482,7 +502,7 @@ public class CatalogdMetaProvider implements MetaProvider {
     // NOTE: we don't need to perform this dance for cache keys which embed a version
     // number, because invalidation is not handled by removing cache entries, but
     // rather by bumping top-level version numbers.
-    Stopwatch sw = new Stopwatch().start();
+    Stopwatch sw = Stopwatch.createStarted();
     boolean hit = false;
     boolean isPiggybacked = false;
     try {
@@ -637,22 +657,28 @@ public class CatalogdMetaProvider implements MetaProvider {
   }
 
   @Override
-  public ImmutableList<String> loadTableNames(final String dbName)
+  public ImmutableCollection<TBriefTableMeta> loadTableList(final String dbName)
       throws MetaException, UnknownDBException, TException {
-    return loadWithCaching("table names for database " + dbName,
-        TABLE_NAMES_STATS_CATEGORY,
-        new DbCacheKey(dbName.toLowerCase(), DbCacheKey.DbInfoType.TABLE_NAMES),
-        new Callable<ImmutableList<String>>() {
+    ImmutableMap<String, TBriefTableMeta> res = loadWithCaching(
+        "table list of database " + dbName, TABLE_LIST_STATS_CATEGORY,
+        new DbCacheKey(dbName.toLowerCase(), DbCacheKey.DbInfoType.TABLE_LIST),
+        new Callable<ImmutableMap<String, TBriefTableMeta>>() {
           @Override
-          public ImmutableList<String> call() throws Exception {
+          public ImmutableMap<String, TBriefTableMeta> call() throws Exception {
             TGetPartialCatalogObjectRequest req = newReqForDb(dbName);
-            req.db_info_selector.want_table_names = true;
+            req.db_info_selector.want_brief_meta_of_tables = true;
             TGetPartialCatalogObjectResponse resp = sendRequest(req);
-            checkResponse(resp.db_info != null && resp.db_info.table_names != null,
+            checkResponse(
+                resp.db_info != null && resp.db_info.brief_meta_of_tables != null,
                 req, "missing expected table names");
-            return ImmutableList.copyOf(resp.db_info.table_names);
+            ImmutableMap.Builder<String, TBriefTableMeta> map = ImmutableMap.builder();
+            for (TBriefTableMeta meta : resp.db_info.brief_meta_of_tables) {
+              map.put(meta.getName(), meta);
+            }
+            return map.build();
           }
       });
+    return res.values();
   }
 
   private TGetPartialCatalogObjectRequest newReqForTable(String dbName,
@@ -703,16 +729,62 @@ public class CatalogdMetaProvider implements MetaProvider {
                 new ArrayList<>() : resp.table_info.sql_constraints.getForeign_keys();
             return new TableMetaRefImpl(
                 dbName, tableName, resp.table_info.hms_table, resp.object_version_number,
-                new SqlConstraints(primaryKeys, foreignKeys));
+                new SqlConstraints(primaryKeys, foreignKeys),
+                resp.table_info.valid_write_ids, resp.table_info.is_marked_cached,
+                resp.table_info.partition_prefixes);
            }
       });
+    // The table list is populated based on tables in a given Db in catalogd. If a table
+    // is not loaded at the time in catalogd, we assume that the table type is "TABLE" and
+    // comment is empty (See MetadataOp.getTableType and MetadataOp.getTableComment()).
+    // However, after issuing this load request, if we find that our assumption was
+    // incorrect, we need to invalidate the table list immediately. The table list gets
+    // invalidated automatically when we receive the catalog topic-update but until then,
+    // all the getTables calls will show incorrect table type and comment for this table.
+    // TODO: consider removing this after IMPALA-9670.
+    invalidateStaleTableList(dbName.toLowerCase(), ref);
     return Pair.create(ref.msTable_, (TableMetaRef)ref);
   }
 
+  /**
+   * Invalidate table list of the given DB if it's loaded and it contains stale
+   * type/comment comparing to the newly loaded msTable.
+   */
+  private void invalidateStaleTableList(final String dbName,
+      final TableMetaRefImpl latestMeta) {
+    Preconditions.checkNotNull(latestMeta.msTable_, "loaded table should have a msTable");
+    DbCacheKey dbKey = new DbCacheKey(dbName, DbCacheKey.DbInfoType.TABLE_LIST);
+    Object dbValue = cache_.getIfPresent(dbKey);
+    if (dbValue instanceof ImmutableMap) {
+      TBriefTableMeta cachedMeta = ((ImmutableMap<String, TBriefTableMeta>) dbValue).get(
+          latestMeta.tableName_);
+      String latestComment = latestMeta.msTable_.getParameters().get(TABLE_COMMENT_KEY);
+      boolean hasStaleComment = !StringUtils.equals(latestComment, cachedMeta.comment);
+      // It's possible that the cached type is null (due to previously unloaded) and the
+      // latest msType we get is a table type (e.g. MANAGED_TABLE, EXTERNAL_TABLE).
+      // This is the most usual case and we don't consider the cached type is stale since
+      // they have the same result after applying MetadataOp.getImpalaTableType().
+      boolean hasStaleType = !StringUtils.equals(
+          MetadataOp.getImpalaTableType(cachedMeta.msType),
+          MetadataOp.getImpalaTableType(latestMeta.msTable_.getTableType()));
+      if (hasStaleType || hasStaleComment) {
+        List<String> invalidated = new ArrayList<>();
+        invalidateCacheForDb(dbName, ImmutableList.of(DbCacheKey.DbInfoType.TABLE_LIST),
+            invalidated);
+        if (!invalidated.isEmpty()) {
+          Preconditions.checkState(invalidated.size() == 1);
+          LOG.debug("Invalidated stale {} after loading table {}: hasStaleType={}, " +
+                  "hasStaleComment={}",
+              invalidated.get(0), latestMeta.tableName_, hasStaleType, hasStaleComment);
+        }
+      }
+    }
+  }
+
   @Override
   public List<ColumnStatisticsObj> loadTableColumnStatistics(final TableMetaRef table,
       List<String> colNames) throws TException {
-    Stopwatch sw = new Stopwatch().start();
+    Stopwatch sw = Stopwatch.createStarted();
     List<ColumnStatisticsObj> ret = Lists.newArrayListWithCapacity(colNames.size());
     // Look up in cache first, keeping track of which ones are missing.
     // We can't use 'loadWithCaching' since we need to fetch several entries batched
@@ -815,7 +887,7 @@ public class CatalogdMetaProvider implements MetaProvider {
       throws MetaException, TException {
     Preconditions.checkArgument(table instanceof TableMetaRefImpl);
     TableMetaRefImpl refImpl = (TableMetaRefImpl)table;
-    Stopwatch sw = new Stopwatch().start();
+    Stopwatch sw = Stopwatch.createStarted();
     // Load what we can from the cache.
     Map<PartitionRef, PartitionMetadata> refToMeta = loadPartitionsFromCache(refImpl,
         hostIndex, partitionRefs);
@@ -833,7 +905,7 @@ public class CatalogdMetaProvider implements MetaProvider {
           refImpl, hostIndex, missingRefs);
       refToMeta.putAll(fromCatalogd);
       // Write back to the cache.
-      storePartitionsInCache(refImpl, hostIndex, fromCatalogd);
+      storePartitionsInCache(hostIndex, fromCatalogd);
     }
     sw.stop();
     addStatsToProfile(PARTITIONS_STATS_CATEGORY, numHits, numMisses, sw);
@@ -881,30 +953,42 @@ public class CatalogdMetaProvider implements MetaProvider {
     for (int i = 0; i < ids.size(); i++) {
       PartitionRef partRef = partRefs.get(i);
       TPartialPartitionInfo part = resp.table_info.partitions.get(i);
-      Partition msPart = part.getHms_partition();
-      if (msPart == null) {
+      HdfsStorageDescriptor hdfsStorageDescriptor = null;
+      HdfsPartitionLocationCompressor.Location location;
+      if (part.isSetHdfs_storage_descriptor()) {
+        Preconditions.checkNotNull(part.location, "location should not be null");
+        hdfsStorageDescriptor = HdfsStorageDescriptor.fromThrift(
+            part.hdfs_storage_descriptor, table.tableName_);
+        location = table.getPartitionLocationCompressor().new Location(part.location);
+      } else {
         checkResponse(table.msTable_.getPartitionKeysSize() == 0, req,
-            "Should not return a partition with missing HMS partition unless " +
+            "Should not return a partition with missing partition meta unless " +
             "the table is unpartitioned");
-        msPart = DirectMetaProvider.msTableToPartition(table.msTable_);
+        // For the only partition of a nonpartitioned table, reuse table-level metadata.
+        try {
+          hdfsStorageDescriptor = HdfsStorageDescriptor.fromStorageDescriptor(
+              table.tableName_, table.msTable_.getSd());
+        } catch (HdfsStorageDescriptor.InvalidStorageDescriptorException e) {
+          Preconditions.checkState(false, "Failed to create HdfsStorageDescriptor " +
+              "using sd of table");
+        }
+        location = table.getPartitionLocationCompressor().new Location(
+            table.msTable_.getSd().getLocation());
+        part.setHms_parameters(table.msTable_.getParameters());
       }
 
       // Transform the file descriptors to the caller's index.
       checkResponse(part.file_descriptors != null, req, "missing file descriptors");
-      List<FileDescriptor> fds = Lists.newArrayListWithCapacity(
-          part.file_descriptors.size());
-      for (THdfsFileDesc thriftFd: part.file_descriptors) {
-        FileDescriptor fd = FileDescriptor.fromThrift(thriftFd);
-        // The file descriptors returned via the RPC use host indexes that reference
-        // the 'network_addresses' list in the RPC. However, the caller may have already
-        // loaded some addresses into 'hostIndex'. So, the returned FDs need to be
-        // remapped to point to the caller's 'hostIndex' instead of the list in the
-        // RPC response.
-        fds.add(fd.cloneWithNewHostIndex(resp.table_info.network_addresses, hostIndex));
-      }
-      PartitionMetadataImpl metaImpl = new PartitionMetadataImpl(msPart,
-          ImmutableList.copyOf(fds), part.getPartition_stats(),
-          part.has_incremental_stats);
+      ImmutableList<FileDescriptor> fds = convertThriftFdList(part.file_descriptors,
+          resp.table_info.network_addresses, hostIndex);
+      ImmutableList<FileDescriptor> insertFds = convertThriftFdList(
+          part.insert_file_descriptors, resp.table_info.network_addresses, hostIndex);
+      ImmutableList<FileDescriptor> deleteFds = convertThriftFdList(
+          part.delete_file_descriptors, resp.table_info.network_addresses, hostIndex);
+      PartitionMetadataImpl metaImpl = new PartitionMetadataImpl(part.getHms_parameters(),
+          part.write_id, hdfsStorageDescriptor,
+          fds, insertFds, deleteFds, part.getPartition_stats(),
+          part.has_incremental_stats, part.is_marked_cached, location);
 
       checkResponse(partRef != null, req, "returned unexpected partition id %s", part.id);
 
@@ -917,6 +1001,21 @@ public class CatalogdMetaProvider implements MetaProvider {
     return ret;
   }
 
+  private ImmutableList<FileDescriptor> convertThriftFdList(List<THdfsFileDesc> thriftFds,
+      List<TNetworkAddress> networkAddresses, ListMap<TNetworkAddress> hostIndex) {
+    List<FileDescriptor> fds = Lists.newArrayListWithCapacity(thriftFds.size());
+    for (THdfsFileDesc thriftFd: thriftFds) {
+      FileDescriptor fd = FileDescriptor.fromThrift(thriftFd);
+      // The file descriptors returned via the RPC use host indexes that reference
+      // the 'network_addresses' list in the RPC. However, the caller may have already
+      // loaded some addresses into 'hostIndex'. So, the returned FDs need to be
+      // remapped to point to the caller's 'hostIndex' instead of the list in the
+      // RPC response.
+      fds.add(fd.cloneWithNewHostIndex(networkAddresses, hostIndex));
+    }
+    return ImmutableList.copyOf(fds);
+  }
+
   /**
    * Load all partitions from 'partitionRefs' that are currently present in the cache.
    * Any partitions that miss the cache are left unset in the resulting map.
@@ -932,7 +1031,7 @@ public class CatalogdMetaProvider implements MetaProvider {
         partitionRefs.size());
     for (PartitionRef ref: partitionRefs) {
       PartitionRefImpl prefImpl = (PartitionRefImpl)ref;
-      PartitionCacheKey cacheKey = new PartitionCacheKey(table, prefImpl.getId());
+      PartitionCacheKey cacheKey = new PartitionCacheKey(prefImpl.getId());
       PartitionMetadataImpl val = (PartitionMetadataImpl)getIfPresent(cacheKey);
       if (val == null) continue;
 
@@ -948,12 +1047,12 @@ public class CatalogdMetaProvider implements MetaProvider {
    * Write back the partitions in 'metas' into the cache. The file descriptors in these
    * partitions must be relative to the 'hostIndex'.
    */
-  private void storePartitionsInCache(TableMetaRefImpl table,
-      ListMap<TNetworkAddress> hostIndex, Map<PartitionRef, PartitionMetadata> metas) {
+  private void storePartitionsInCache(ListMap<TNetworkAddress> hostIndex,
+      Map<PartitionRef, PartitionMetadata> metas) {
     for (Map.Entry<PartitionRef, PartitionMetadata> e: metas.entrySet()) {
       PartitionRefImpl prefImpl = (PartitionRefImpl)e.getKey();
       PartitionMetadataImpl metaImpl = (PartitionMetadataImpl)e.getValue();
-      PartitionCacheKey cacheKey = new PartitionCacheKey(table, prefImpl.getId());
+      PartitionCacheKey cacheKey = new PartitionCacheKey(prefImpl.getId());
       PartitionMetadataImpl cacheVal = metaImpl.cloneRelativeToHostIndex(hostIndex,
           cacheHostIndex_);
       cache_.put(cacheKey, cacheVal);
@@ -1043,6 +1142,8 @@ public class CatalogdMetaProvider implements MetaProvider {
   public synchronized TUpdateCatalogCacheResponse updateCatalogCache(
       TUpdateCatalogCacheRequest req) {
     if (req.isSetCatalog_service_id()) {
+      // Requests from processing statestore updates won't have this. Only requests from
+      // DDLs will set this. See more details in ImpalaServer::ProcessCatalogUpdateResult.
       witnessCatalogServiceId(req.catalog_service_id);
     }
 
@@ -1052,6 +1153,7 @@ public class CatalogdMetaProvider implements MetaProvider {
     Long nextCatalogVersion = null;
 
     ObjectUpdateSequencer authObjectSequencer = new ObjectUpdateSequencer();
+    ObjectUpdateSequencer hdfsCachePoolSequencer = new ObjectUpdateSequencer();
 
     Pair<Boolean, ByteBuffer> update;
     while ((update = FeSupport.NativeGetNextCatalogObjectUpdate(req.native_iterator_ptr))
@@ -1076,8 +1178,26 @@ public class CatalogdMetaProvider implements MetaProvider {
         continue;
       }
 
+      if (!isDelete && obj.type == TCatalogObjectType.HDFS_PARTITION) {
+        // Skip if this is the update for a new partition.
+        if (!obj.hdfs_partition.isSetPrev_id()) continue;
+        // This is an update from an existing partiton. Invalidate the previous partition
+        // instance by resetting the id to the previous one.
+        obj.hdfs_partition.setId(obj.hdfs_partition.prev_id);
+        obj.hdfs_partition.unsetPrev_id();
+      }
       invalidateCacheForObject(obj);
 
+      if (obj.type == TCatalogObjectType.HDFS_CACHE_POOL) {
+        // HdfsCachePools have no dependency to each other. But we can't update
+        // hdfsCachePools_ directly since a new catalog service id (due to catalogd
+        // restart) will cause hdfsCachePools_ to be clear (see mode details in
+        // witnessCatalogServiceId()). So we have to deal with HDFS_CACHE_POOL objects
+        // after the CATALOG object.
+        hdfsCachePoolSequencer.add(obj, isDelete);
+        continue;
+      }
+
       // The sequencing of updates to authorization objects is important since they
       // may be cross-referential. So, just add them to the sequencer which ensures
       // we handle them in the right order later.
@@ -1085,6 +1205,7 @@ public class CatalogdMetaProvider implements MetaProvider {
           obj.type == TCatalogObjectType.PRIVILEGE ||
           obj.type == TCatalogObjectType.AUTHZ_CACHE_INVALIDATION) {
         authObjectSequencer.add(obj, isDelete);
+        continue;
       }
 
       // Handle CATALOG objects. These are sent only via the updates published via
@@ -1101,10 +1222,23 @@ public class CatalogdMetaProvider implements MetaProvider {
           // Detected a new reset() finishes in Catalogd, clear the cache in case some
           // tables are skipped in this topic update.
           cache_.invalidateAll();
+          // Don't need to clear hdfsCachePools_ if this comes from a catalogd restart,
+          // because we already clear it in witnessCatalogServiceId().
+          // Shouldn't clear hdfsCachePools_ if this comes from a global invalidation,
+          // because HdfsCachePool updates may already come in the previous statestore
+          // update (see how the version lock is held in CatalogServiceCatalog.reset()).
+          // Clear hdfsCachePools_ will also clear them.
         }
       }
     }
 
+    for (TCatalogObject obj : hdfsCachePoolSequencer.getUpdatedObjects()) {
+      updateHdfsCachePools(obj, /*isDelete=*/false);
+    }
+    for (TCatalogObject obj : hdfsCachePoolSequencer.getDeletedObjects()) {
+      updateHdfsCachePools(obj, /*isDelete=*/true);
+    }
+
     for (TCatalogObject obj : authObjectSequencer.getUpdatedObjects()) {
       updateAuthPolicy(obj, /*isDelete=*/false);
     }
@@ -1134,6 +1268,29 @@ public class CatalogdMetaProvider implements MetaProvider {
     }
   }
 
+  private void updateHdfsCachePools(TCatalogObject obj, boolean isDelete) {
+    Preconditions.checkState(obj.type == TCatalogObjectType.HDFS_CACHE_POOL);
+    String poolName = obj.getCache_pool().getPool_name();
+    if (isDelete) {
+      HdfsCachePool existingItem = hdfsCachePools_.get(poolName);
+      if (existingItem != null
+          && existingItem.getCatalogVersion() <= obj.getCatalog_version()) {
+        hdfsCachePools_.remove(poolName);
+        LOG.trace("Removed HdfsCachePool {}", poolName);
+      }
+      return;
+    }
+    HdfsCachePool cachePool = new HdfsCachePool(obj.getCache_pool());
+    cachePool.setCatalogVersion(obj.getCatalog_version());
+    if (hdfsCachePools_.add(cachePool)) {
+      LOG.trace("Added HdfsCachePool name={}, version={}", poolName,
+          obj.getCatalog_version());
+      return;
+    }
+    LOG.warn("Ignored stale HdfsCachePool update: name={}, version={}", poolName,
+        obj.getCatalog_version());
+  }
+
   private void updateAuthPolicy(TCatalogObject obj, boolean isDelete) {
     LOG.trace("Updating authorization policy: {} isDelete={}", obj, isDelete);
     switch (obj.type) {
@@ -1201,6 +1358,10 @@ public class CatalogdMetaProvider implements MetaProvider {
         }
         catalogServiceId_ = serviceId;
         cache_.invalidateAll();
+        // Clear cached items from the previous catalogd instance. Otherwise, we'll
+        // ignore new updates from the new catalogd instance since they have lower
+        // versions.
+        hdfsCachePools_.clear();
         // TODO(todd): we probably need to invalidate the auth policy too.
         // we are probably better off detecting this at a higher level and
         // reinstantiating the metaprovider entirely, similar to how ImpaladCatalog
@@ -1236,9 +1397,13 @@ public class CatalogdMetaProvider implements MetaProvider {
       // DB, so we'll be coarse-grained here and invalidate the DB table list when
       // any table change happens. It's relatively cheap to re-fetch this.
       invalidateCacheForDb(obj.table.db_name,
-          ImmutableList.of(DbCacheKey.DbInfoType.TABLE_NAMES),
+          ImmutableList.of(DbCacheKey.DbInfoType.TABLE_LIST),
           invalidated);
       break;
+    case HDFS_PARTITION:
+      invalidateCacheForPartition(obj.hdfs_partition.db_name, obj.hdfs_partition.tbl_name,
+          obj.hdfs_partition.partition_name, obj.hdfs_partition.id, invalidated);
+      break;
     case FUNCTION:
       // Same as above: if we see a function, it might be new or deleted and we should
       // refresh the list of functions in the DB to be safe.
@@ -1260,11 +1425,10 @@ public class CatalogdMetaProvider implements MetaProvider {
         invalidated.add("list of database names");
       }
       invalidateCacheForDb(obj.db.db_name, ImmutableList.of(
-          DbCacheKey.DbInfoType.TABLE_NAMES,
+          DbCacheKey.DbInfoType.TABLE_LIST,
           DbCacheKey.DbInfoType.HMS_METADATA,
           DbCacheKey.DbInfoType.FUNCTION_NAMES), invalidated);
       break;
-
     default:
       break;
     }
@@ -1300,6 +1464,19 @@ public class CatalogdMetaProvider implements MetaProvider {
     }
   }
 
+  /**
+   * Invalidate cached metadata for the given partition. If anything was invalidated, adds
+   * a human-readable string to 'invalidated' indicating the invalidated metadata.
+   */
+  private void invalidateCacheForPartition(String dbName, String tblName, String partName,
+      long partitionId, List<String> invalidated) {
+    PartitionCacheKey key = new PartitionCacheKey(partitionId);
+    if (cache_.asMap().remove(key) != null) {
+      invalidated.add(String.format("partition %s.%s:%s (id=%d)",
+          dbName, tblName, partName, partitionId));
+    }
+  }
+
   /**
    * Invalidate cached metadata for the given function. If anything was invalidated, adds
    * a human-readable string to 'invalidated' indicating the invalidated metadata.
@@ -1339,17 +1516,37 @@ public class CatalogdMetaProvider implements MetaProvider {
   }
 
   public static class PartitionMetadataImpl implements PartitionMetadata {
-    private final Partition msPartition_;
+    private final Map<String, String> hmsParameters_;
+    private final long writeId_;
+    private final HdfsStorageDescriptor hdfsStorageDescriptor_;
+    // Prefix-compressed location. The corresponding prefix map, i.e. the
+    // HdfsPartitionLocationCompressor object is in the corresponding TableMetaRefImpl
+    // object. TableMetaRefImpl may be evicted. But when this partition is accessed,
+    // we can assure that the table meta is already loaded, so the prefix map won't be
+    // missing.
+    private final HdfsPartitionLocationCompressor.Location location_;
     private final ImmutableList<FileDescriptor> fds_;
+    private final ImmutableList<FileDescriptor> insertFds_;
+    private final ImmutableList<FileDescriptor> deleteFds_;
     private final byte[] partitionStats_;
     private final boolean hasIncrementalStats_;
+    private final boolean isMarkedCached_;
 
-    public PartitionMetadataImpl(Partition msPartition, ImmutableList<FileDescriptor> fds,
-        byte[] partitionStats, boolean hasIncrementalStats) {
-      this.msPartition_ = Preconditions.checkNotNull(msPartition);
+    public PartitionMetadataImpl(Map<String, String> hmsParameters, long writeId,
+        HdfsStorageDescriptor hdfsStorageDescriptor, ImmutableList<FileDescriptor> fds,
+        ImmutableList<FileDescriptor> insertFds, ImmutableList<FileDescriptor> deleteFds,
+        byte[] partitionStats, boolean hasIncrementalStats, boolean isMarkedCached,
+        HdfsPartitionLocationCompressor.Location location) {
+      this.hmsParameters_ = hmsParameters;
+      this.writeId_ = writeId;
+      this.hdfsStorageDescriptor_ = hdfsStorageDescriptor;
+      this.location_ = location;
       this.fds_ = fds;
+      this.insertFds_ = insertFds;
+      this.deleteFds_ = deleteFds;
       this.partitionStats_ = partitionStats;
       this.hasIncrementalStats_ = hasIncrementalStats;
+      this.isMarkedCached_ = isMarkedCached;
     }
 
     /**
@@ -1359,22 +1556,59 @@ public class CatalogdMetaProvider implements MetaProvider {
     public PartitionMetadataImpl cloneRelativeToHostIndex(
         ListMap<TNetworkAddress> origIndex,
         ListMap<TNetworkAddress> dstIndex) {
-      List<FileDescriptor> fds = Lists.newArrayListWithCapacity(fds_.size());
-      for (FileDescriptor fd: fds_) {
-        fds.add(fd.cloneWithNewHostIndex(origIndex.getList(), dstIndex));
+      ImmutableList<FileDescriptor> fds = cloneFdsRelativeToHostIndex(
+          fds_, origIndex, dstIndex);
+      ImmutableList<FileDescriptor> insertFds = cloneFdsRelativeToHostIndex(
+          insertFds_, origIndex, dstIndex);
+      ImmutableList<FileDescriptor> deleteFds = cloneFdsRelativeToHostIndex(
+          deleteFds_, origIndex, dstIndex);
+      return new PartitionMetadataImpl(hmsParameters_, writeId_, hdfsStorageDescriptor_,
+          fds, insertFds, deleteFds, partitionStats_, hasIncrementalStats_,
+          isMarkedCached_, location_);
+    }
+
+    private static ImmutableList<FileDescriptor> cloneFdsRelativeToHostIndex(
+        ImmutableList<FileDescriptor> fds, ListMap<TNetworkAddress> origIndex,
+        ListMap<TNetworkAddress> dstIndex) {
+      List<FileDescriptor> ret = Lists.newArrayListWithCapacity(fds.size());
+      for (FileDescriptor fd: fds) {
+        ret.add(fd.cloneWithNewHostIndex(origIndex.getList(), dstIndex));
       }
-      return new PartitionMetadataImpl(msPartition_, ImmutableList.copyOf(fds),
-          partitionStats_, hasIncrementalStats_);
+      return ImmutableList.copyOf(ret);
     }
 
     @Override
-    public Partition getHmsPartition() {
-      return msPartition_;
+    public Map<String, String> getHmsParameters() { return hmsParameters_; }
+
+    @Override
+    public long getWriteId() { return writeId_; }
+
+    @Override
+    public HdfsStorageDescriptor getInputFormatDescriptor() {
+      return hdfsStorageDescriptor_;
     }
 
+    @Override
+    public HdfsPartitionLocationCompressor.Location getLocation() { return location_; }
+
     @Override
     public ImmutableList<FileDescriptor> getFileDescriptors() {
-      return fds_;
+      if (insertFds_.isEmpty()) return fds_;
+      List<FileDescriptor> ret = Lists.newArrayListWithCapacity(
+          insertFds_.size() + deleteFds_.size());
+      ret.addAll(insertFds_);
+      ret.addAll(deleteFds_);
+      return ImmutableList.copyOf(ret);
+    }
+
+    @Override
+    public ImmutableList<FileDescriptor> getInsertFileDescriptors() {
+      return insertFds_;
+    }
+
+    @Override
+    public ImmutableList<FileDescriptor> getDeleteFileDescriptors() {
+      return deleteFds_;
     }
 
     @Override
@@ -1382,6 +1616,9 @@ public class CatalogdMetaProvider implements MetaProvider {
 
     @Override
     public boolean hasIncrementalStats() { return hasIncrementalStats_; }
+
+    @Override
+    public boolean isMarkedCached() { return isMarkedCached_; }
   }
 
   /**
@@ -1409,19 +1646,50 @@ public class CatalogdMetaProvider implements MetaProvider {
      */
     private final long catalogVersion_;
 
+    /**
+     * Valid write id list of ACID tables.
+     */
+    private final TValidWriteIdList validWriteIds_;
+
+    /**
+     * True if this table's is marked as cached by hdfs caching. See comments in
+     * LocalFsTable.
+     */
+    private final boolean isMarkedCached_;
+
+    private final HdfsPartitionLocationCompressor partitionLocationCompressor_;
+
     public TableMetaRefImpl(String dbName, String tableName,
-        Table msTable, long catalogVersion, SqlConstraints sqlConstraints) {
+        Table msTable, long catalogVersion, SqlConstraints sqlConstraints,
+        TValidWriteIdList validWriteIds, boolean isMarkedCached,
+        List<String> locationPrefixes) {
       this.dbName_ = dbName;
       this.tableName_ = tableName;
       this.msTable_ = msTable;
       this.catalogVersion_ = catalogVersion;
       this.sqlConstraints_ = sqlConstraints;
+      this.validWriteIds_ = validWriteIds;
+      this.isMarkedCached_ = isMarkedCached;
+      // Non-hdfs tables will have null locationPrefixes.
+      this.partitionLocationCompressor_ = (locationPrefixes == null) ? null :
+          new HdfsPartitionLocationCompressor(
+              msTable.getPartitionKeysSize(), locationPrefixes);
     }
 
     @Override
     public String toString() {
       return String.format("TableMetaRef %s.%s@%d", dbName_, tableName_, catalogVersion_);
     }
+
+    public boolean isMarkedCached() { return isMarkedCached_; }
+
+    public HdfsPartitionLocationCompressor getPartitionLocationCompressor() {
+      return partitionLocationCompressor_;
+    }
+
+    public List<String> getPartitionPrefixes() {
+      return partitionLocationCompressor_.getPrefixes();
+    }
   }
 
   /**
@@ -1527,9 +1795,7 @@ public class CatalogdMetaProvider implements MetaProvider {
 
     @Override
     public boolean equals(Object obj) {
-      if (obj == null || !(obj instanceof ColStatsCacheKey)) {
-        return false;
-      }
+      if (!(obj instanceof ColStatsCacheKey)) return false;
       ColStatsCacheKey other = (ColStatsCacheKey)obj;
       return super.equals(obj) && colName_.equals(other.colName_);
     }
@@ -1549,34 +1815,26 @@ public class CatalogdMetaProvider implements MetaProvider {
   /**
    * Key for caching information about a single partition.
    *
-   * TODO(todd): currently this inherits from VersionedTableCacheKey. This means that, if
-   * a table's version number changes, all of its partitions must be reloaded. However,
-   * since partition IDs are globally unique within a catalogd instance, we could
-   * optimize this to just key based on the partition ID. However, currently, there are
-   * some cases where partitions are mutated in place rather than replaced with a new ID.
-   * We need to eliminate those or add a partition sequence number before we can make
-   * this optimization.
+   * Values for these keys are 'PartitionMetadataImpl' objects.
    */
-  private static class PartitionCacheKey extends VersionedTableCacheKey {
+  @Immutable
+  private static class PartitionCacheKey {
     private final long partId_;
 
-    PartitionCacheKey(TableMetaRefImpl table, long partId) {
-      super(table);
+    PartitionCacheKey(long partId) {
       partId_ = partId;
     }
 
     @Override
     public int hashCode() {
-      return Objects.hashCode(super.hashCode(), partId_);
+      return Objects.hashCode(partId_, getClass());
     }
 
     @Override
     public boolean equals(Object obj) {
-      if (obj == null || !(obj instanceof PartitionCacheKey)) {
-        return false;
-      }
+      if (!(obj instanceof PartitionCacheKey)) return false;
       PartitionCacheKey other = (PartitionCacheKey)obj;
-      return super.equals(obj) && partId_ == other.partId_;
+      return partId_ == other.partId_;
     }
   }
 
@@ -1587,8 +1845,8 @@ public class CatalogdMetaProvider implements MetaProvider {
     static enum DbInfoType {
       /** Cache the HMS Database object */
       HMS_METADATA,
-      /** Cache an ImmutableList<String> for table names within the DB */
-      TABLE_NAMES,
+      /** Cache an ImmutableList<BriefTableMeta> for table names within the DB */
+      TABLE_LIST,
       /** Cache an ImmutableList<String> for function names within the DB */
       FUNCTION_NAMES
     }
@@ -1640,4 +1898,11 @@ public class CatalogdMetaProvider implements MetaProvider {
       return (int)size;
     }
   }
+
+  @Override
+  public TValidWriteIdList getValidWriteIdList(TableMetaRef ref) {
+    Preconditions.checkArgument(ref instanceof TableMetaRefImpl,
+        "table ref %s was not created by CatalogdMetaProvider", ref);
+    return ((TableMetaRefImpl)ref).validWriteIds_;
+  }
 }
