@@ -62,7 +62,7 @@ import org.apache.impala.util.ExecutorMembershipSnapshot;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
-import com.google.common.base.Objects;
+import com.google.common.base.MoreObjects;
 import com.google.common.base.Preconditions;
 import com.google.common.collect.Lists;
 
@@ -303,6 +303,7 @@ public class HBaseScanNode extends ScanNode {
       // We have run computeStats successfully. Don't need to estimate cardinality again
       // (IMPALA-8912). Check some invariants if computeStats has been called.
       Preconditions.checkState(numNodes_ > 0);
+      Preconditions.checkState(numInstances_ > 0);
       Preconditions.checkState(cardinality_ >= 0);
       cardinality_ = inputCardinality_;
       if (LOG.isTraceEnabled()) {
@@ -358,19 +359,24 @@ public class HBaseScanNode extends ScanNode {
       LOG.trace("computeStats HbaseScan: cardinality=" + cardinality_);
     }
 
-    // Assume that each executor in the cluster gets a scan range, unless there are fewer
-    // scan ranges than nodes.
-    numNodes_ = Math.max(1, Math.min(scanRangeSpecs_.getConcrete_rangesSize(),
-                                ExecutorMembershipSnapshot.getCluster().numExecutors()));
+    // Assume that each node/instance in the cluster gets a scan range, unless there are
+    // fewer scan ranges than nodes/instances.
+    int numExecutors = ExecutorMembershipSnapshot.getCluster().numExecutors();
+    numNodes_ =
+        Math.max(1, Math.min(scanRangeSpecs_.getConcrete_rangesSize(), numExecutors));
+    int maxInstances = numNodes_ * getMaxInstancesPerNode(analyzer);
+    numInstances_ =
+        Math.max(1, Math.min(scanRangeSpecs_.getConcrete_rangesSize(), maxInstances));
     if (LOG.isTraceEnabled()) {
-      LOG.trace("computeStats HbaseScan: #nodes=" + numNodes_);
+      LOG.trace(
+          "computeStats HbaseScan: #nodes=" + numNodes_ + " #instances=" + numInstances_);
     }
   }
 
   @Override
   protected String debugString() {
     FeHBaseTable tbl = (FeHBaseTable) desc_.getTable();
-    return Objects.toStringHelper(this)
+    return MoreObjects.toStringHelper(this)
         .add("tid", desc_.getId().asInt())
         .add("hiveTblName", tbl.getFullName())
         .add("hbaseTblName", tbl.getHBaseTableName())
