@@ -24,6 +24,7 @@ import java.util.Set;
 import org.apache.impala.analysis.DescriptorTable;
 import org.apache.impala.analysis.Expr;
 import org.apache.impala.catalog.FeFsTable;
+import org.apache.impala.catalog.FeIcebergTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.catalog.HdfsFileFormat;
 import org.apache.impala.common.Pair;
@@ -61,9 +62,12 @@ public class HdfsTableSink extends TableSink {
   // be opened, written, and closed one by one.
   protected final boolean inputIsClustered_;
 
-  private static final Set<HdfsFileFormat> SUPPORTED_FILE_FORMATS = ImmutableSet.of(
-      HdfsFileFormat.PARQUET, HdfsFileFormat.TEXT, HdfsFileFormat.LZO_TEXT,
-      HdfsFileFormat.RC_FILE, HdfsFileFormat.SEQUENCE_FILE, HdfsFileFormat.AVRO);
+  // Indicates that this sink is being used to write query results
+  protected final boolean isResultSink_;
+
+  public static final Set<HdfsFileFormat> SUPPORTED_FILE_FORMATS = ImmutableSet.of(
+      HdfsFileFormat.PARQUET, HdfsFileFormat.TEXT, HdfsFileFormat.RC_FILE,
+      HdfsFileFormat.SEQUENCE_FILE, HdfsFileFormat.AVRO, HdfsFileFormat.ICEBERG);
 
   // Stores the indices into the list of non-clustering columns of the target table that
   // are stored in the 'sort.columns' table property. This is sent to the backend to
@@ -75,10 +79,29 @@ public class HdfsTableSink extends TableSink {
   // Stores the allocated write id if the target table is transactional, otherwise -1.
   private long writeId_;
 
+  // Set the limit on the maximum number of hdfs table sink instances.
+  // A value of 0 means no limit.
+  private int maxHdfsSinks_;
+
+  // Stores the desired location for the sink output. Primarily intended to be used by
+  // external FEs that expect the results at a certain location for the purposes of
+  // post processing and finalization.
+  protected String externalOutputDir_;
+
+  // This specifies the position in the partition specifition in which to start
+  // creating partition directories.
+  // Example:
+  // Partition specification: year, month, day
+  // The external frontend has provided a path that has year, month created
+  // by them and thus provides a depth of 2 which hints to the TableSink to skip the
+  // first two partition keys and create the day directory for new results. This is
+  // usually used by tables with dynamic partition columns.
+  private int externalOutputPartitionDepth_;
+
   public HdfsTableSink(FeTable targetTable, List<Expr> partitionKeyExprs,
-      List<Expr> outputExprs,
-      boolean overwrite, boolean inputIsClustered,
-      Pair<List<Integer>, TSortingOrder> sortProperties, long writeId) {
+      List<Expr> outputExprs, boolean overwrite, boolean inputIsClustered,
+      Pair<List<Integer>, TSortingOrder> sortProperties, long writeId,
+      int maxTableSinks, boolean isResultSink) {
     super(targetTable, Op.INSERT, outputExprs);
     Preconditions.checkState(targetTable instanceof FeFsTable);
     partitionKeyExprs_ = partitionKeyExprs;
@@ -87,12 +110,22 @@ public class HdfsTableSink extends TableSink {
     sortColumns_ = sortProperties.first;
     sortingOrder_ = sortProperties.second;
     writeId_ = writeId;
+    maxHdfsSinks_ = maxTableSinks;
+    isResultSink_ = isResultSink;
+  }
+
+  public void setExternalOutputDir(String externalOutputDir) {
+    externalOutputDir_ = externalOutputDir;
+  }
+
+  public void setExternalOutputPartitionDepth(int partitionDepth) {
+    externalOutputPartitionDepth_ = partitionDepth;
   }
 
   @Override
   public void computeResourceProfile(TQueryOptions queryOptions) {
     PlanNode inputNode = fragment_.getPlanRoot();
-    int numInstances = fragment_.getNumInstances(queryOptions.getMt_dop());
+    int numInstances = fragment_.getNumInstances();
     // Compute the number of partitions buffered in memory at the same time, taking into
     // account the number of nodes and the data partition of the fragment executing this
     // sink.
@@ -137,12 +170,6 @@ public class HdfsTableSink extends TableSink {
    * set of file formats.
    */
   private long getPerPartitionMemReq(Set<HdfsFileFormat> formats) {
-    Set<HdfsFileFormat> unsupportedFormats =
-        Sets.difference(formats, SUPPORTED_FILE_FORMATS);
-    if (!unsupportedFormats.isEmpty()) {
-      Preconditions.checkState(false,
-          "Unsupported TableSink format(s): " + Joiner.on(',').join(unsupportedFormats));
-    }
     if (formats.contains(HdfsFileFormat.PARQUET)) {
       // Writing to a Parquet partition requires up to 1GB of buffer. From a resource
       // management purview, even if there are non-Parquet partitions, we want to be
@@ -150,7 +177,7 @@ public class HdfsTableSink extends TableSink {
       return 1024L * 1024L * 1024L;
     }
 
-    // For all other supported formats (TEXT, LZO_TEXT, RC_FILE, SEQUENCE_FILE & AVRO)
+    // For all other supported formats (TEXT, RC_FILE, SEQUENCE_FILE & AVRO)
     // 100KB is a very approximate estimate of the amount of data buffered.
     return 100L * 1024L;
   }
@@ -209,6 +236,11 @@ public class HdfsTableSink extends TableSink {
     }
     hdfsTableSink.setSort_columns(sortColumns_);
     hdfsTableSink.setSorting_order(sortingOrder_);
+    hdfsTableSink.setIs_result_sink(isResultSink_);
+    if (externalOutputDir_ != null) {
+      hdfsTableSink.setExternal_output_dir(externalOutputDir_);
+      hdfsTableSink.setExternal_output_partition_depth(externalOutputPartitionDepth_);
+    }
     if (writeId_ != -1) hdfsTableSink.setWrite_id(writeId_);
     TTableSink tTableSink = new TTableSink(DescriptorTable.TABLE_SINK_ID,
         TTableSinkType.HDFS, sinkOp_.toThrift());
@@ -224,8 +256,43 @@ public class HdfsTableSink extends TableSink {
 
   @Override
   public void collectExprs(List<Expr> exprs) {
-    exprs.addAll(partitionKeyExprs_);
-    // Avoid adding any partition exprs redundantly.
-    exprs.addAll(outputExprs_.subList(0, targetTable_.getNonClusteringColumns().size()));
+    if (!(targetTable_ instanceof FeIcebergTable)) exprs.addAll(partitionKeyExprs_);
+    if (isResultSink_) {
+      // It is expected outputExprs_ is fully populated with the expected result schema
+      // when writing query results
+      exprs.addAll(outputExprs_);
+    } else {
+      // Avoid adding any partition exprs redundantly.
+      exprs.addAll(outputExprs_.subList(0,
+          targetTable_.getNonClusteringColumns().size()));
+    }
+  }
+
+  /**
+   * Return an estimate of the number of nodes the fragment with this sink will
+   * run on. This is based on the number of nodes set for the plan root and has an
+   * upper limit set by the MAX_HDFS_WRITER query option.
+   */
+  public int getNumNodes() {
+    int num_nodes = getFragment().getPlanRoot().getNumNodes();
+    if (maxHdfsSinks_ > 0) {
+      // If there are more nodes than instances where the fragment was initially
+      // planned to run then, then the instances will be distributed evenly across them.
+      num_nodes = Math.min(num_nodes, getNumInstances());
+    }
+    return num_nodes;
+  }
+
+  /**
+   * Return an estimate of the number of instances the fragment with this sink
+   * will run on. This is based on the number of instances set for the plan root
+   * and has an upper limit set by the MAX_HDFS_WRITER query option.
+   */
+  public int getNumInstances() {
+    int num_instances = getFragment().getPlanRoot().getNumInstances();
+    if (maxHdfsSinks_ > 0) {
+      num_instances =  Math.min(num_instances, maxHdfsSinks_);
+    }
+    return num_instances;
   }
 }
