@@ -80,6 +80,14 @@ public abstract class JoinNode extends PlanNode {
   // joinTableId_
   protected JoinTableId joinTableId_ = JoinTableId.INVALID;
 
+  // True if this join is used to do the join between insert and delete delta files.
+  protected boolean isAcidJoin_ = false;
+
+  public void setIsAcidJoin() {
+    isAcidJoin_ = true;
+    displayName_ = "DELETE EVENTS " + displayName_;
+  }
+
   // List of equi-join conjuncts believed to be involved in a FK/PK relationship.
   // The conjuncts are grouped by the tuple ids of the joined base table refs. A conjunct
   // is only included in this list if it is of the form <SlotRef> = <SlotRef> and the
@@ -191,6 +199,12 @@ public abstract class JoinNode extends PlanNode {
     return true;
   }
 
+  // Returns true if we can share a join build between multiple consuming fragment
+  // instances.
+  public boolean canShareBuild() {
+    return distrMode_ == JoinNode.DistributionMode.BROADCAST;
+  }
+
   public JoinOperator getJoinOp() { return joinOp_; }
   public List<BinaryPredicate> getEqJoinConjuncts() { return eqJoinConjuncts_; }
   public List<Expr> getOtherJoinConjuncts() { return otherJoinConjuncts_; }
@@ -551,7 +565,7 @@ public abstract class JoinNode extends PlanNode {
    * Unwraps the SlotRef in expr and returns the NDVs of it.
    * Returns -1 if the NDVs are unknown or if expr is not a SlotRef.
    */
-  private long getNdv(Expr expr) {
+  protected long getNdv(Expr expr) {
     SlotRef slotRef = expr.unwrapSlotRef(false);
     if (slotRef == null) return -1;
     SlotDescriptor slotDesc = slotRef.getDesc();
@@ -561,6 +575,14 @@ public abstract class JoinNode extends PlanNode {
     return stats.getNumDistinctValues();
   }
 
+  /**
+   * Reset the numNodes_ and numInstances_ based on the left child
+   */
+  public void recomputeNodes() {
+    numNodes_ = getChild(0).numNodes_;
+    numInstances_ = getChild(0).numInstances_;
+  }
+
   @Override
   public void computeStats(Analyzer analyzer) {
     super.computeStats(analyzer);
