@@ -19,30 +19,28 @@ package org.apache.impala.planner;
 
 import java.io.IOException;
 import java.util.ArrayList;
+import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
 import java.util.ListIterator;
+import java.util.Map;
 import java.util.Set;
 
-import org.apache.impala.analysis.AggregateInfo;
 import org.apache.impala.analysis.Analyzer;
 import org.apache.impala.analysis.BinaryPredicate;
 import org.apache.impala.analysis.BoolLiteral;
 import org.apache.impala.analysis.Expr;
 import org.apache.impala.analysis.ExprSubstitutionMap;
-import org.apache.impala.analysis.FunctionCallExpr;
-import org.apache.impala.analysis.FunctionName;
-import org.apache.impala.analysis.FunctionParams;
 import org.apache.impala.analysis.InPredicate;
 import org.apache.impala.analysis.IsNullPredicate;
 import org.apache.impala.analysis.LiteralExpr;
+import org.apache.impala.analysis.MultiAggregateInfo;
 import org.apache.impala.analysis.NumericLiteral;
 import org.apache.impala.analysis.SlotDescriptor;
 import org.apache.impala.analysis.SlotRef;
 import org.apache.impala.analysis.StringLiteral;
 import org.apache.impala.analysis.TupleDescriptor;
 import org.apache.impala.catalog.FeKuduTable;
-import org.apache.impala.catalog.HdfsFileFormat;
 import org.apache.impala.catalog.KuduColumn;
 import org.apache.impala.catalog.Type;
 import org.apache.impala.common.ImpalaRuntimeException;
@@ -58,6 +56,7 @@ import org.apache.impala.thrift.TScanRangeLocation;
 import org.apache.impala.thrift.TScanRangeLocationList;
 import org.apache.impala.thrift.TScanRangeSpec;
 import org.apache.impala.util.KuduUtil;
+import org.apache.impala.util.ExecutorMembershipSnapshot;
 import org.apache.kudu.ColumnSchema;
 import org.apache.kudu.Schema;
 import org.apache.kudu.client.KuduClient;
@@ -123,23 +122,23 @@ public class KuduScanNode extends ScanNode {
   boolean isPointLookupQuery_ = false;
 
   public KuduScanNode(PlanNodeId id, TupleDescriptor desc, List<Expr> conjuncts,
-      AggregateInfo aggInfo) {
+      MultiAggregateInfo aggInfo) {
     super(id, desc, "SCAN KUDU");
     kuduTable_ = (FeKuduTable) desc_.getTable();
     conjuncts_ = conjuncts;
     aggInfo_ = aggInfo;
   }
 
-  public ExprSubstitutionMap getOptimizedAggSmap() { return optimizedAggSmap_; }
-
   @Override
   public void init(Analyzer analyzer) throws ImpalaRuntimeException {
     conjuncts_ = orderConjunctsByCost(conjuncts_);
 
     KuduClient client = KuduUtil.getKuduClient(kuduTable_.getKuduMasterHosts());
     try {
+      // Get the KuduTable from the analyzer to retrieve the cached KuduTable
+      // for this query and prevent multiple openTable calls for a single query.
       org.apache.kudu.client.KuduTable rpcTable =
-          client.openTable(kuduTable_.getKuduTableName());
+          analyzer.getKuduTable(kuduTable_);
       validateSchema(rpcTable);
 
       if (canApplyCountStarOptimization(analyzer)) {
@@ -218,7 +217,7 @@ public class KuduScanNode extends ScanNode {
       throws ImpalaRuntimeException {
     scanRangeSpecs_ = new TScanRangeSpec();
 
-    List<KuduScanToken> scanTokens = createScanTokens(client, rpcTable);
+    List<KuduScanToken> scanTokens = createScanTokens(analyzer, client, rpcTable);
     for (KuduScanToken token: scanTokens) {
       LocatedTablet tablet = token.getTablet();
       List<TScanRangeLocation> locations = new ArrayList<>();
@@ -257,7 +256,7 @@ public class KuduScanNode extends ScanNode {
    * will be pushed to Kudu. The projected Kudu columns are ordered by offset in an
    * Impala tuple to make the Impala and Kudu tuple layouts identical.
    */
-  private List<KuduScanToken> createScanTokens(KuduClient client,
+  private List<KuduScanToken> createScanTokens(Analyzer analyzer, KuduClient client,
       org.apache.kudu.client.KuduTable rpcTable) {
     List<String> projectedCols = new ArrayList<>();
     for (SlotDescriptor desc: getTupleDesc().getSlotsOrderedByOffset()) {
@@ -267,6 +266,9 @@ public class KuduScanNode extends ScanNode {
     }
     KuduScanTokenBuilder tokenBuilder = client.newScanTokenBuilder(rpcTable);
     tokenBuilder.setProjectedColumnNames(projectedCols);
+    long split_size_hint = analyzer.getQueryOptions()
+        .getTargeted_kudu_scan_range_length();
+    if (split_size_hint > 0) tokenBuilder.setSplitSizeBytes(split_size_hint);
     for (KuduPredicate predicate: kuduPredicates_) tokenBuilder.addPredicate(predicate);
     return tokenBuilder.build();
   }
@@ -278,11 +280,81 @@ public class KuduScanNode extends ScanNode {
     return computeCombinedSelectivity(allConjuncts);
   }
 
+  /**
+   * Estimate the number of impalad nodes that this scan node will execute on (which is
+   * ultimately determined by the scheduling done by the backend's Scheduler).
+   * Assume that scan ranges that can be scheduled locally will be, and that scan
+   * ranges that cannot will be round-robined across the cluster.
+   */
+  protected void computeNumNodes(Analyzer analyzer) {
+    ExecutorMembershipSnapshot cluster = ExecutorMembershipSnapshot.getCluster();
+    final int maxInstancesPerNode = getMaxInstancesPerNode(analyzer);
+    final int maxPossibleInstances = cluster.numExecutors() * maxInstancesPerNode;
+    int totalNodes = 0;
+    int totalInstances = 0;
+    int numLocalRanges = 0;
+    int numRemoteRanges = 0;
+    // Counts the number of local ranges, capped at maxInstancesPerNode.
+    Map<TNetworkAddress, Integer> localRangeCounts = new HashMap<>();
+    // Sum of the counter values in localRangeCounts.
+    int totalLocalParallelism = 0;
+    if (scanRangeSpecs_.isSetConcrete_ranges()) {
+      for (TScanRangeLocationList range : scanRangeSpecs_.concrete_ranges) {
+        boolean anyLocal = false;
+        if (range.isSetLocations()) {
+          for (TScanRangeLocation loc : range.locations) {
+            TNetworkAddress address =
+                analyzer.getHostIndex().getEntry(loc.getHost_idx());
+            if (cluster.contains(address)) {
+              anyLocal = true;
+              // Use the full tserver address (including port) to account for the test
+              // minicluster where there are multiple tservers and impalads on a single
+              // host.  This assumes that when an impalad is colocated with a tserver,
+              // there are the same number of impalads as tservers on this host in this
+              // cluster.
+              int count = localRangeCounts.getOrDefault(address, 0);
+              if (count < maxInstancesPerNode) {
+                ++totalLocalParallelism;
+                localRangeCounts.put(address, count + 1);
+              }
+            }
+          }
+        }
+        // This range has at least one replica with a colocated impalad, so assume it
+        // will be scheduled on one of those nodes.
+        if (anyLocal) {
+          ++numLocalRanges;
+        } else {
+          ++numRemoteRanges;
+        }
+        // Approximate the number of nodes that will execute locally assigned ranges to
+        // be the smaller of the number of locally assigned ranges and the number of
+        // hosts that hold replica for those ranges.
+        int numLocalNodes = Math.min(numLocalRanges, localRangeCounts.size());
+        // The remote ranges are round-robined across all the impalads.
+        int numRemoteNodes = Math.min(numRemoteRanges, cluster.numExecutors());
+        // The local and remote assignments may overlap, but we don't know by how much
+        // so conservatively assume no overlap.
+        totalNodes = Math.min(numLocalNodes + numRemoteNodes, cluster.numExecutors());
+
+        int numLocalInstances = Math.min(numLocalRanges, totalLocalParallelism);
+        totalInstances = Math.min(numLocalInstances + numRemoteRanges,
+            totalNodes * maxInstancesPerNode);
+
+        // Exit early if we have maxed out our estimate of hosts/instances, to avoid
+        // extraneous work in case the number of scan ranges dominates the number of
+        // nodes.
+        if (totalInstances == maxPossibleInstances) break;
+      }
+    }
+    numNodes_ = Math.max(totalNodes, 1);
+    numInstances_ = Math.max(totalInstances, 1);
+  }
+
   @Override
-  protected void computeStats(Analyzer analyzer) {
+  public void computeStats(Analyzer analyzer) {
     super.computeStats(analyzer);
-    // Update the number of nodes to reflect the hosts that have relevant data.
-    numNodes_ = Math.max(1, hostIndexSet_.size());
+    computeNumNodes(analyzer);
 
     // Update the cardinality
     inputCardinality_ = cardinality_ = kuduTable_.getNumRows();
@@ -316,12 +388,11 @@ public class KuduScanNode extends ScanNode {
     int perHostScanRanges = estimatePerHostScanRanges(numOfScanRanges);
     int maxScannerThreads = computeMaxNumberOfScannerThreads(queryOptions,
         perHostScanRanges);
-    int num_cols = desc_.getSlots().size();
     long estimated_bytes_per_column_per_thread = BackendConfig.INSTANCE.getBackendCfg().
         kudu_scanner_thread_estimated_bytes_per_column;
     long max_estimated_bytes_per_thread = BackendConfig.INSTANCE.getBackendCfg().
         kudu_scanner_thread_max_estimated_bytes;
-    long mem_estimate_per_thread = Math.min(num_cols *
+    long mem_estimate_per_thread = Math.min(getNumMaterializedSlots(desc_) *
         estimated_bytes_per_column_per_thread, max_estimated_bytes_per_thread);
     useMtScanNode_ = queryOptions.mt_dop > 0;
     nodeResourceProfile_ = new ResourceProfileBuilder()
