@@ -26,9 +26,11 @@ import java.util.List;
 import java.util.Set;
 
 import org.apache.impala.analysis.Analyzer;
+import org.apache.impala.analysis.BinaryPredicate;
 import org.apache.impala.analysis.Expr;
 import org.apache.impala.analysis.ExprId;
 import org.apache.impala.analysis.ExprSubstitutionMap;
+import org.apache.impala.analysis.SlotDescriptor;
 import org.apache.impala.analysis.ToSqlOptions;
 import org.apache.impala.analysis.TupleDescriptor;
 import org.apache.impala.analysis.TupleId;
@@ -122,10 +124,16 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
   // invalid: -1
   protected long cardinality_;
 
-  // number of nodes on which the plan tree rooted at this node would execute;
+  // Estimated number of nodes on which the plan tree rooted at this node would be
+  // scheduled;
   // set in computeStats(); invalid: -1
   protected int numNodes_;
 
+  // Estimated number of instances across all nodes that the scheduler would generate for
+  // a fragment with the plan tree rooted at this node;
+  // set in computeStats(); invalid: -1
+  protected int numInstances_;
+
   // resource requirements and estimates for this plan node.
   // Initialized with a dummy value. Gets set correctly in
   // computeResourceProfile().
@@ -163,6 +171,7 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
     tblRefIds_ = new ArrayList<>();
     cardinality_ = -1;
     numNodes_ = -1;
+    numInstances_ = -1;
     displayName_ = displayName;
     disableCodegen_ = false;
   }
@@ -179,6 +188,7 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
     conjuncts_ = Expr.cloneList(node.conjuncts_);
     cardinality_ = -1;
     numNodes_ = -1;
+    numInstances_ = -1;
     displayName_ = displayName;
     disableCodegen_ = node.disableCodegen_;
   }
@@ -210,6 +220,7 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
   public boolean hasLimit() { return limit_ > -1; }
   public long getCardinality() { return cardinality_; }
   public int getNumNodes() { return numNodes_; }
+  public int getNumInstances() { return numInstances_; }
   public ResourceProfile getNodeResourceProfile() { return nodeResourceProfile_; }
   public float getAvgRowSize() { return avgRowSize_; }
   public void setFragment(PlanFragment fragment) { fragment_ = fragment; }
@@ -507,6 +518,44 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
     analyzer.markConjunctsAssigned(unassigned);
   }
 
+  /**
+   * Apply the provided conjuncts to the this node, returning the new root of
+   * the plan tree. Also add any slot equivalences for tupleIds that have not
+   * yet been enforced.
+   * TODO: change this to assign the unassigned conjuncts to root itself, if that is
+   * semantically correct
+   */
+  public PlanNode addConjunctsToNode(PlannerContext plannerCtx,
+          Analyzer analyzer, List<TupleId> tupleIds, List<Expr> conjuncts) {
+    if (LOG.isTraceEnabled()) {
+      LOG.trace(String.format("conjuncts for (Node %s): %s",
+          getDisplayLabel(), Expr.debugString(conjuncts)));
+      LOG.trace("all conjuncts: " + analyzer.conjunctAssignmentsDebugString());
+    }
+    if (this instanceof EmptySetNode) return this;
+    for (TupleId tid: tupleIds) {
+      analyzer.createEquivConjuncts(tid, conjuncts);
+    }
+    if (conjuncts.isEmpty()) return this;
+
+    List<Expr> finalConjuncts = new ArrayList<>();
+    // Check if this is an inferred identity predicate i.e for c1 = c2 both
+    // sides are pointing to the same source slot. In such cases it is wrong
+    // to add the predicate to the SELECT node because it will incorrectly
+    // eliminate rows with NULL values.
+    for (Expr e : conjuncts) {
+      if (e instanceof BinaryPredicate && ((BinaryPredicate) e).isInferred()) {
+        SlotDescriptor lhs = ((BinaryPredicate) e).getChild(0).findSrcScanSlot();
+        SlotDescriptor rhs = ((BinaryPredicate) e).getChild(1).findSrcScanSlot();
+        if (lhs != null && rhs != null && lhs.equals(rhs)) continue;
+      }
+      finalConjuncts.add(e);
+    }
+    if (finalConjuncts.isEmpty()) return this;
+    // The conjuncts need to be evaluated in a SelectNode.
+    return SelectNode.create(plannerCtx, analyzer, this, finalConjuncts);
+  }
+
   /**
    * Returns an smap that combines the childrens' smaps.
    */
@@ -541,13 +590,16 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
    * from init() (to facilitate inserting additional nodes during plan
    * partitioning w/o the need to call init() recursively on the whole tree again).
    */
-  protected void computeStats(Analyzer analyzer) {
+  public void computeStats(Analyzer analyzer) {
     avgRowSize_ = 0.0F;
     for (TupleId tid: tupleIds_) {
       TupleDescriptor desc = analyzer.getTupleDesc(tid);
       avgRowSize_ += desc.getAvgSerializedSize();
     }
-    if (!children_.isEmpty()) numNodes_ = getChild(0).numNodes_;
+    if (!children_.isEmpty()) {
+      numNodes_ = getChild(0).numNodes_;
+      numInstances_ = getChild(0).numInstances_;
+    }
   }
 
   protected long capCardinalityAtLimit(long cardinality) {
@@ -557,7 +609,7 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
     return cardinality;
   }
 
-  static long capCardinalityAtLimit(long cardinality, long limit) {
+  public static long capCardinalityAtLimit(long cardinality, long limit) {
     return cardinality == -1 ? limit : Math.min(cardinality, limit);
   }
 
@@ -654,7 +706,7 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
 
   protected String getSortingOrderExplainString(List<? extends Expr> exprs,
       List<Boolean> isAscOrder, List<Boolean> nullsFirstParams,
-      TSortingOrder sortingOrder) {
+      TSortingOrder sortingOrder, int numLexicalKeysInZOrder) {
     StringBuilder output = new StringBuilder();
     switch (sortingOrder) {
     case LEXICAL:
@@ -670,9 +722,20 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
       }
       break;
     case ZORDER:
+      if (numLexicalKeysInZOrder > 0) {
+        Preconditions.checkElementIndex(numLexicalKeysInZOrder, exprs.size());
+        output.append("LEXICAL: ");
+        for (int i = 0; i < numLexicalKeysInZOrder; ++i) {
+          if (i > 0) output.append(", ");
+          // Pre-insert sorting uses ASC and NULLS LAST. See more in
+          // Planner.createPreInsertSort().
+          output.append(exprs.get(i).toSql()).append(" ASC NULLS LAST");
+        }
+        output.append(", ");
+      }
       output.append("ZORDER: ");
-      for (int i = 0; i < exprs.size(); ++i) {
-        if (i > 0) output.append(", ");
+      for (int i = numLexicalKeysInZOrder; i < exprs.size(); ++i) {
+        if (i > numLexicalKeysInZOrder) output.append(", ");
         output.append(exprs.get(i).toSql());
       }
       break;
@@ -684,8 +747,9 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
   /**
    * Returns true if stats-related variables are valid.
    */
-  protected boolean hasValidStats() {
+  public boolean hasValidStats() {
     return (numNodes_ == -1 || numNodes_ >= 0) &&
+           (numInstances_ == -1 || numInstances_ >= 0) &&
            (cardinality_ == -1 || cardinality_ >= 0);
   }
 
@@ -860,6 +924,14 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
     return sum;
   }
 
+  /**
+   * Returns the max number of instances of a fragment that can be scheduled on a single
+   * node - the mt_dop value if mt_dop > 0, or 1 otherwise.
+   */
+  protected int getMaxInstancesPerNode(Analyzer analyzer) {
+    return Math.max(1, analyzer.getQueryOptions().getMt_dop());
+  }
+
   protected void addRuntimeFilter(RuntimeFilter filter) { runtimeFilters_.add(filter); }
 
   protected Collection<RuntimeFilter> getRuntimeFilters() { return runtimeFilters_; }
@@ -969,4 +1041,17 @@ abstract public class PlanNode extends TreeNode<PlanNode> {
   public void setDisableCodegen(boolean disableCodegen) {
     disableCodegen_ = disableCodegen;
   }
+
+  public boolean allowPartitioned() { return !hasLimit(); }
+
+  public int getNumMaterializedSlots(TupleDescriptor desc) {
+    int numCols = 0;
+    for (SlotDescriptor slot: desc.getSlots()) {
+      // Columns used in this query
+      if (slot.isMaterialized()) {
+        ++numCols;
+      }
+    }
+    return numCols;
+  }
 }
