@@ -94,6 +94,7 @@ public class Planner {
   }
 
   public TQueryCtx getQueryCtx() { return ctx_.getQueryCtx(); }
+  public PlannerContext getPlannerCtx() { return ctx_; }
   public AnalysisContext.AnalysisResult getAnalysisResult() {
     return ctx_.getAnalysisResult();
   }
@@ -125,26 +126,7 @@ public class Planner {
     invertJoins(singleNodePlan, ctx_.isSingleNodeExec());
     singleNodePlan = useNljForSingularRowBuilds(singleNodePlan, ctx_.getRootAnalyzer());
 
-    // Parallel plans are not supported by default for plans with base table joins or
-    // table sinks: we only allow such plans if --unlock_mt_dop=true is specified.
-    if (useParallelPlan()
-        && (!RuntimeEnv.INSTANCE.isTestEnv()
-               || RuntimeEnv.INSTANCE.isMtDopValidationEnabled())
-        && !BackendConfig.INSTANCE.isMtDopUnlocked()
-        && (ctx_.hasTableSink()
-               || singleNodePlanner.hasUnsupportedMtDopJoin(singleNodePlan))) {
-      if (BackendConfig.INSTANCE.mtDopAutoFallback()) {
-        // Fall back to non-dop mode. This assumes that the mt_dop value is only used
-        // in the distributed planning process, which should be generally true as long
-        // as the value isn't cached in any plan nodes.
-        ctx_.getQueryOptions().setMt_dop(0);
-      } else {
-        throw new NotImplementedException(
-            "MT_DOP not supported for plans with base table joins or table sinks.");
-      }
-    }
-
-    singleNodePlanner.validatePlan(singleNodePlan);
+    SingleNodePlanner.validatePlan(ctx_, singleNodePlan);
 
     if (ctx_.isSingleNodeExec()) {
       // create one fragment containing the entire single-node plan tree
@@ -263,7 +245,7 @@ public class Planner {
   public List<PlanFragment> createPlans() throws ImpalaException {
     List<PlanFragment> distrPlan = createPlanFragments();
     Preconditions.checkNotNull(distrPlan);
-    if (!useParallelPlan()) {
+    if (!useParallelPlan(ctx_)) {
       return Collections.singletonList(distrPlan.get(0));
     }
     ParallelPlanner planner = new ParallelPlanner(ctx_);
@@ -276,9 +258,9 @@ public class Planner {
    * Return true if we should generate a parallel plan for this query, based on the
    * current mt_dop value and whether a single-node plan was chosen.
    */
-  private boolean useParallelPlan() {
-    Preconditions.checkState(ctx_.getQueryOptions().isSetMt_dop());
-    return ctx_.getQueryOptions().mt_dop > 0 && !ctx_.isSingleNodeExec();
+  public static boolean useParallelPlan(PlannerContext planCtx) {
+    Preconditions.checkState(planCtx.getQueryOptions().isSetMt_dop());
+    return planCtx.getQueryOptions().mt_dop > 0 && !planCtx.isSingleNodeExec();
   }
 
   /**
@@ -413,11 +395,12 @@ public class Planner {
    * consumed by all fragment instances belonging to the query per host. Sets the
    * per-host resource values in 'request'.
    */
-  public void computeResourceReqs(List<PlanFragment> planRoots,
-      TQueryCtx queryCtx, TQueryExecRequest request) {
+  public static void computeResourceReqs(List<PlanFragment> planRoots,
+      TQueryCtx queryCtx, TQueryExecRequest request, PlannerContext planCtx,
+      boolean isQueryStmt) {
     Preconditions.checkState(!planRoots.isEmpty());
     Preconditions.checkNotNull(request);
-    TQueryOptions queryOptions = ctx_.getRootAnalyzer().getQueryOptions();
+    TQueryOptions queryOptions = planCtx.getRootAnalyzer().getQueryOptions();
     int mtDop = queryOptions.getMt_dop();
 
     // Peak per-host peak resources for all plan fragments, assuming that all fragments
@@ -432,7 +415,7 @@ public class Planner {
     List<PlanFragment> allFragments = rootFragment.getNodesPostOrder();
     for (PlanFragment fragment: allFragments) {
       // Compute the per-node, per-sink and aggregate profiles for the fragment.
-      fragment.computeResourceProfile(ctx_.getRootAnalyzer());
+      fragment.computeResourceProfile(planCtx.getRootAnalyzer());
 
       // Different fragments do not synchronize their Open() and Close(), so the backend
       // does not provide strong guarantees about whether one fragment instance releases
@@ -463,7 +446,7 @@ public class Planner {
         maxPerHostPeakResources.getMinMemReservationBytes());
     request.setMax_per_host_thread_reservation(
         maxPerHostPeakResources.getThreadReservation());
-    if (getAnalysisResult().isQueryStmt()) {
+    if (isQueryStmt) {
       // Only one instance of root fragment, so don't need to multiply instance resource
       // profile.
       ResourceProfile rootFragmentResourceProfile =
@@ -504,7 +487,7 @@ public class Planner {
    * The 'isLocalPlan' parameter indicates whether the plan tree rooted at 'root'
    * will be executed locally within one machine, i.e., without any data exchanges.
    */
-  private void invertJoins(PlanNode root, boolean isLocalPlan) {
+  public static void invertJoins(PlanNode root, boolean isLocalPlan) {
     if (root instanceof SubplanNode) {
       invertJoins(root.getChild(0), isLocalPlan);
       invertJoins(root.getChild(1), true);
@@ -535,11 +518,14 @@ public class Planner {
       } else if (isInvertedJoinCheaper(joinNode, isLocalPlan)) {
         joinNode.invertJoin();
       }
+      // Re-compute the numNodes and numInstances based on the new input order
+      joinNode.recomputeNodes();
     }
 
     // Re-compute tuple ids because the backend assumes that their order corresponds to
     // the order of children.
     root.computeTupleIds();
+
   }
 
   /**
@@ -588,7 +574,7 @@ public class Planner {
    * parallelism significantly, then a significant difference between lhs and rhs
    * bytes is needed to justify inversion.
    */
-  private boolean isInvertedJoinCheaper(JoinNode joinNode, boolean isLocalPlan) {
+  public static boolean isInvertedJoinCheaper(JoinNode joinNode, boolean isLocalPlan) {
     long lhsCard = joinNode.getChild(0).getCardinality();
     long rhsCard = joinNode.getChild(1).getCardinality();
     // Need cardinality estimates to make a decision.
@@ -656,32 +642,37 @@ public class Planner {
     return newJoinNode;
   }
 
-  private void checkForSmallQueryOptimization(PlanNode singleNodePlan) {
+  public static void checkForSmallQueryOptimization(PlanNode singleNodePlan,
+      PlannerContext ctx) {
     MaxRowsProcessedVisitor visitor = new MaxRowsProcessedVisitor();
     singleNodePlan.accept(visitor);
-    // TODO: IMPALA-3335: support the optimization for plans with joins.
-    if (!visitor.valid() || visitor.foundJoinNode()) return;
+    if (!visitor.valid()) return;
     // This optimization executes the plan on a single node so the threshold must
     // be based on the total number of rows processed.
     long maxRowsProcessed = visitor.getMaxRowsProcessed();
-    int threshold = ctx_.getQueryOptions().exec_single_node_rows_threshold;
+    int threshold = ctx.getQueryOptions().exec_single_node_rows_threshold;
     if (maxRowsProcessed < threshold) {
       // Execute on a single node and disable codegen for small results
       LOG.trace("Query is small enough to execute on a single node: maxRowsProcessed = "
           + maxRowsProcessed);
-      ctx_.getQueryOptions().setNum_nodes(1);
-      ctx_.getQueryCtx().disable_codegen_hint = true;
-      if (maxRowsProcessed < ctx_.getQueryOptions().batch_size ||
-          maxRowsProcessed < 1024 && ctx_.getQueryOptions().batch_size == 0) {
+      ctx.getQueryOptions().setNum_nodes(1);
+      ctx.getQueryCtx().disable_codegen_hint = true;
+      if (maxRowsProcessed < ctx.getQueryOptions().batch_size ||
+          maxRowsProcessed < 1024 && ctx.getQueryOptions().batch_size == 0) {
         // Only one scanner thread for small queries
-        ctx_.getQueryOptions().setNum_scanner_threads(1);
+        ctx.getQueryOptions().setNum_scanner_threads(1);
       }
       // disable runtime filters
-      ctx_.getQueryOptions().setRuntime_filter_mode(TRuntimeFilterMode.OFF);
+      ctx.getQueryOptions().setRuntime_filter_mode(TRuntimeFilterMode.OFF);
     }
   }
 
-  private void checkForDisableCodegen(PlanNode distributedPlan) {
+  private void checkForSmallQueryOptimization(PlanNode singleNodePlan) {
+      checkForSmallQueryOptimization(singleNodePlan, ctx_);
+  }
+
+  public static void checkForDisableCodegen(PlanNode distributedPlan,
+      PlannerContext ctx) {
     MaxRowsProcessedVisitor visitor = new MaxRowsProcessedVisitor();
     distributedPlan.accept(visitor);
     if (!visitor.valid()) return;
@@ -689,11 +680,15 @@ public class Planner {
     // reduce per-node execution time enough to justify the cost of codegen. Per-node
     // execution time is correlated with the number of rows flowing through the plan.
     if (visitor.getMaxRowsProcessedPerNode()
-        < ctx_.getQueryOptions().getDisable_codegen_rows_threshold()) {
-      ctx_.getQueryCtx().disable_codegen_hint = true;
+        < ctx.getQueryOptions().getDisable_codegen_rows_threshold()) {
+      ctx.getQueryCtx().disable_codegen_hint = true;
     }
   }
 
+  private void checkForDisableCodegen(PlanNode distributedPlan) {
+      checkForDisableCodegen(distributedPlan, ctx_);
+  }
+
   /**
    * Insert a sort node on top of the plan, depending on the clustered/noclustered
    * plan hint and on the 'sort.columns' table property. If clustering is enabled in
@@ -710,6 +705,7 @@ public class Planner {
     List<Expr> orderingExprs = new ArrayList<>();
 
     boolean partialSort = false;
+    int numPartitionKeys = 0;
     if (insertStmt.getTargetTable() instanceof FeKuduTable) {
       // Always sort if the 'clustered' hint is present. Otherwise, don't sort if either
       // the 'noclustered' hint is present, or this is a single node exec, or if the
@@ -722,7 +718,11 @@ public class Planner {
         partialSort = true;
       }
     } else if (insertStmt.requiresClustering()) {
-      orderingExprs.addAll(insertStmt.getPartitionKeyExprs());
+      List<Expr> partKeys = insertStmt.getPartitionKeyExprs();
+      orderingExprs.addAll(partKeys);
+      // Ignore constants. Only dynamic partition inserts have non constant keys.
+      Expr.removeConstants(orderingExprs);
+      numPartitionKeys = orderingExprs.size();
     }
     orderingExprs.addAll(insertStmt.getSortExprs());
     // Ignore constants for the sake of clustering.
@@ -735,6 +735,7 @@ public class Planner {
     List<Boolean> nullsFirstParams = Collections.nCopies(orderingExprs.size(), false);
     SortInfo sortInfo = new SortInfo(orderingExprs, isAscOrder, nullsFirstParams,
         insertStmt.getSortingOrder());
+    sortInfo.setNumLexicalKeysInZOrder(numPartitionKeys);
     sortInfo.createSortTupleInfo(insertStmt.getResultExprs(), analyzer);
     sortInfo.getSortTupleDescriptor().materializeSlots();
     insertStmt.substituteResultExprs(sortInfo.getOutputSmap(), analyzer);
