@@ -33,7 +33,10 @@ import org.apache.impala.analysis.BinaryPredicate.Operator;
 import org.apache.impala.analysis.CastExpr;
 import org.apache.impala.analysis.Expr;
 import org.apache.impala.analysis.ExprSubstitutionMap;
+import org.apache.impala.analysis.FunctionCallExpr;
 import org.apache.impala.analysis.IsNullPredicate;
+import org.apache.impala.analysis.LiteralExpr;
+import org.apache.impala.analysis.NullLiteral;
 import org.apache.impala.analysis.Predicate;
 import org.apache.impala.analysis.SlotDescriptor;
 import org.apache.impala.analysis.SlotId;
@@ -42,13 +45,17 @@ import org.apache.impala.analysis.TupleDescriptor;
 import org.apache.impala.analysis.TupleId;
 import org.apache.impala.analysis.TupleIsNullPredicate;
 import org.apache.impala.catalog.FeTable;
+import org.apache.impala.catalog.Column;
 import org.apache.impala.catalog.KuduColumn;
 import org.apache.impala.catalog.Type;
+import org.apache.impala.common.AnalysisException;
 import org.apache.impala.common.IdGenerator;
 import org.apache.impala.common.InternalException;
 import org.apache.impala.planner.JoinNode.DistributionMode;
 import org.apache.impala.service.BackendConfig;
 import org.apache.impala.service.FeSupport;
+import org.apache.impala.thrift.TColumnValue;
+import org.apache.impala.thrift.TEnabledRuntimeFilterTypes;
 import org.apache.impala.thrift.TQueryOptions;
 import org.apache.impala.thrift.TRuntimeFilterDesc;
 import org.apache.impala.thrift.TRuntimeFilterMode;
@@ -120,6 +127,9 @@ public final class RuntimeFilterGenerator {
     // Pre-computed default filter size, in bytes, rounded up to a power of two.
     public final long defaultVal;
 
+    // Target false positive probability, between 0 and 1 exclusive.
+    public final double targetFpp;
+
     public FilterSizeLimits(TQueryOptions tQueryOptions) {
       // Round up all limits to a power of two and make sure filter size is more
       // than the min buffer size that can be allocated by the buffer pool.
@@ -135,6 +145,13 @@ public final class RuntimeFilterGenerator {
       long defaultValue = tQueryOptions.getRuntime_bloom_filter_size();
       defaultValue = Math.max(defaultValue, minVal);
       defaultVal = BitUtil.roundUpToPowerOf2(Math.min(defaultValue, maxVal));
+
+      // Target FPP is determined by runtime_filter_error_rate query option, or if that
+      // is not set, --max_filter_error_rate, which was the legacy option controlling
+      // this that also had other effects.
+      targetFpp = tQueryOptions.isSetRuntime_filter_error_rate() ?
+          tQueryOptions.getRuntime_filter_error_rate() :
+          BackendConfig.INSTANCE.getMaxFilterErrorRate();
     }
   };
 
@@ -193,6 +210,9 @@ public final class RuntimeFilterGenerator {
     private boolean finalized_ = false;
     // The type of filter to build.
     private final TRuntimeFilterType type_;
+    // If set, indicates that the filter is targeted for Kudu scan node with source
+    // timestamp truncation.
+    private boolean isTimestampTruncation_ = false;
 
     /**
      * Internal representation of a runtime filter target.
@@ -206,14 +226,20 @@ public final class RuntimeFilterGenerator {
       public final boolean isBoundByPartitionColumns;
       // Indicates if 'node' is in the same fragment as the join that produces the filter
       public final boolean isLocalTarget;
+      // The low and high value of the column on which the filter is applied
+      public final TColumnValue lowValue;
+      public final TColumnValue highValue;
 
       public RuntimeFilterTarget(ScanNode targetNode, Expr targetExpr,
-          boolean isBoundByPartitionColumns, boolean isLocalTarget) {
+          boolean isBoundByPartitionColumns, boolean isLocalTarget, TColumnValue lowValue,
+          TColumnValue highValue) {
         Preconditions.checkState(targetExpr.isBoundByTupleIds(targetNode.getTupleIds()));
         node = targetNode;
         expr = targetExpr;
         this.isBoundByPartitionColumns = isBoundByPartitionColumns;
         this.isLocalTarget = isLocalTarget;
+        this.lowValue = lowValue;
+        this.highValue = highValue;
       }
 
       public TRuntimeFilterTargetDesc toThrift() {
@@ -235,6 +261,8 @@ public final class RuntimeFilterGenerator {
           tFilterTarget.setKudu_col_name(col.getKuduName());
           tFilterTarget.setKudu_col_type(col.getType().toThrift());
         }
+        tFilterTarget.setLow_value(lowValue);
+        tFilterTarget.setHigh_value(highValue);
         return tFilterTarget;
       }
 
@@ -245,13 +273,16 @@ public final class RuntimeFilterGenerator {
             .append("Target expr: " + expr.debugString() + " ")
             .append("Partition columns: " + isBoundByPartitionColumns)
             .append("Is local: " + isLocalTarget)
+            .append("lowValue: " + lowValue.toString())
+            .append("highValue: " + highValue.toString())
             .toString();
       }
     }
 
     private RuntimeFilter(RuntimeFilterId filterId, JoinNode filterSrcNode, Expr srcExpr,
         Expr origTargetExpr, Operator exprCmpOp, Map<TupleId, List<SlotId>> targetSlots,
-        TRuntimeFilterType type, FilterSizeLimits filterSizeLimits) {
+        TRuntimeFilterType type, FilterSizeLimits filterSizeLimits,
+        boolean isTimestampTruncation) {
       id_ = filterId;
       src_ = filterSrcNode;
       srcExpr_ = srcExpr;
@@ -259,6 +290,7 @@ public final class RuntimeFilterGenerator {
       exprCmpOp_ = exprCmpOp;
       targetSlotsByTid_ = targetSlots;
       type_ = type;
+      isTimestampTruncation_ = isTimestampTruncation;
       computeNdvEstimate();
       calculateFilterSize(filterSizeLimits);
     }
@@ -308,17 +340,23 @@ public final class RuntimeFilterGenerator {
      */
     public static RuntimeFilter create(IdGenerator<RuntimeFilterId> idGen,
         Analyzer analyzer, Expr joinPredicate, JoinNode filterSrcNode,
-        TRuntimeFilterType type, FilterSizeLimits filterSizeLimits) {
+        TRuntimeFilterType type, FilterSizeLimits filterSizeLimits,
+        boolean isTimestampTruncation) {
       Preconditions.checkNotNull(idGen);
       Preconditions.checkNotNull(joinPredicate);
       Preconditions.checkNotNull(filterSrcNode);
       // Only consider binary equality predicates
-      if (!Predicate.isEquivalencePredicate(joinPredicate)) return null;
-
-      BinaryPredicate normalizedJoinConjunct =
-          SingleNodePlanner.getNormalizedEqPred(joinPredicate,
-              filterSrcNode.getChild(0).getTupleIds(),
-              filterSrcNode.getChild(1).getTupleIds(), analyzer);
+      if (type == TRuntimeFilterType.BLOOM
+          && !Predicate.isEquivalencePredicate(joinPredicate)) {
+        return null;
+      }
+      if (type == TRuntimeFilterType.MIN_MAX
+          && !Predicate.isSqlEquivalencePredicate(joinPredicate)) {
+        return null;
+      }
+      BinaryPredicate normalizedJoinConjunct = SingleNodePlanner.getNormalizedEqPred(
+          joinPredicate, filterSrcNode.getChild(0).getTupleIds(),
+          filterSrcNode.getChild(1).getTupleIds(), analyzer);
       if (normalizedJoinConjunct == null) return null;
 
       // Ensure that the target expr does not contain TupleIsNull predicates as these
@@ -327,15 +365,66 @@ public final class RuntimeFilterGenerator {
           TupleIsNullPredicate.unwrapExpr(normalizedJoinConjunct.getChild(0).clone());
       Expr srcExpr = normalizedJoinConjunct.getChild(1);
 
+      if (isTimestampTruncation) {
+        Preconditions.checkArgument(srcExpr.isAnalyzed());
+        Preconditions.checkArgument(srcExpr.getType() == Type.TIMESTAMP);
+        Expr toUnixTimeExpr =
+            new FunctionCallExpr("utc_to_unix_micros", Lists.newArrayList(srcExpr));
+        try {
+          toUnixTimeExpr.analyze(analyzer);
+        } catch (AnalysisException e) {
+          // Expr analysis failed. Skip this runtime filter since we cannot serialize
+          // it to thrift.
+          LOG.warn("Skipping runtime filter because analysis failed: "
+                  + toUnixTimeExpr.toSql(),
+              e);
+          return null;
+        }
+        srcExpr = toUnixTimeExpr;
+      }
+
       Map<TupleId, List<SlotId>> targetSlots = getTargetSlots(analyzer, targetExpr);
       Preconditions.checkNotNull(targetSlots);
       if (targetSlots.isEmpty()) return null;
 
+      if (filterSrcNode.getJoinOp().isLeftOuterJoin() ||
+              filterSrcNode.getJoinOp().isFullOuterJoin()) {
+        try {
+          LiteralExpr nullSlotVal = analyzer.evalWithNullSlots(srcExpr);
+          if (nullSlotVal == null || !(nullSlotVal instanceof NullLiteral)) {
+            // IMPALA-10252: if the source expression returns a non-NULL value for a NULL
+            // input then it is not safe to generate a runtime filter from a LEFT OUTER
+            // JOIN or FULL OUTER JOIN because the generated filter would be missing a
+            // possible non-NULL value of the expression.
+            // E.g. If the source expression is zeroifnull(y), column y has values
+            // [1, 2, 3], and y comes from the right input of a left outer join , then
+            // the generated runtime filter would only contain the values [1, 2, 3] but
+            // not the value zeroifnull(NULL) = 0 that would be present for an unmatched
+            // row. For now we avoid the problem by skipping this filter. In future we
+            // could generate valid runtime filters for these join types by adding
+            // backend support to calculate and insert the missing value.
+            // It's also possible that the expression did not evaluate to a constant after
+            // null slot substitution. In this case, we also want to disable runtime
+            // filtering.
+            if (LOG.isTraceEnabled()) {
+              LOG.trace("Skipping runtime filter because source expression returns " +
+                      "non-null after null substitution: " + srcExpr.toSql());
+            }
+            return null;
+          }
+        } catch (AnalysisException e) {
+          LOG.warn("Skipping runtime filter because analysis after null substitution " +
+              "failed: " + srcExpr.toSql(), e);
+          return null;
+        }
+      }
+
       if (LOG.isTraceEnabled()) {
         LOG.trace("Generating runtime filter from predicate " + joinPredicate);
       }
       return new RuntimeFilter(idGen.getNextId(), filterSrcNode, srcExpr, targetExpr,
-          normalizedJoinConjunct.getOp(), targetSlots, type, filterSizeLimits);
+          normalizedJoinConjunct.getOp(), targetSlots, type, filterSizeLimits,
+          isTimestampTruncation);
     }
 
     /**
@@ -442,6 +531,20 @@ public final class RuntimeFilterGenerator {
     public TRuntimeFilterType getType() { return type_; }
     public Operator getExprCompOp() { return exprCmpOp_; }
     public long getFilterSize() { return filterSizeBytes_; }
+    public boolean isTimestampTruncation() { return isTimestampTruncation_; }
+
+    /**
+     * Return TIMESTAMP if the isTimestampTruncation_ is set as true so that
+     * the source expr type could be matching with target expr type when
+     * assigning bloom filter to target scan node.
+     */
+    public Type getSrcExprType() {
+      if (!isTimestampTruncation_) {
+        return srcExpr_.getType();
+      } else {
+        return Type.TIMESTAMP;
+      }
+    }
 
     /**
      * Estimates the selectivity of a runtime filter as the cardinality of the
@@ -485,8 +588,8 @@ public final class RuntimeFilterGenerator {
         filterSizeBytes_ = filterSizeLimits.defaultVal;
         return;
       }
-      double fpp = BackendConfig.INSTANCE.getMaxFilterErrorRate();
-      int logFilterSize = FeSupport.GetMinLogSpaceForBloomFilter(ndvEstimate_, fpp);
+      double targetFpp = filterSizeLimits.targetFpp;
+      int logFilterSize = FeSupport.GetMinLogSpaceForBloomFilter(ndvEstimate_, targetFpp);
       filterSizeBytes_ = 1L << logFilterSize;
       filterSizeBytes_ = Math.max(filterSizeBytes_, filterSizeLimits.minVal);
       filterSizeBytes_ = Math.min(filterSizeBytes_, filterSizeLimits.maxVal);
@@ -553,6 +656,19 @@ public final class RuntimeFilterGenerator {
       if (LOG.isTraceEnabled()) LOG.trace("Runtime filter: " + filter.debugString());
       filter.assignToPlanNodes();
     }
+
+    arrangeRuntimefiltersForParquet(plan);
+  }
+
+  public static void arrangeRuntimefiltersForParquet(
+      PlanNode root) {
+    if (root instanceof HdfsScanNode) {
+      ((HdfsScanNode)root).arrangeRuntimefiltersForParquet();
+    } else {
+      for (PlanNode childNode: root.getChildren()) {
+        arrangeRuntimefiltersForParquet(childNode);
+      }
+    }
   }
 
   /**
@@ -594,14 +710,32 @@ public final class RuntimeFilterGenerator {
         joinConjuncts.addAll(joinNode.getEqJoinConjuncts());
       }
       joinConjuncts.addAll(joinNode.getConjuncts());
+
       List<RuntimeFilter> filters = new ArrayList<>();
-      for (TRuntimeFilterType type : TRuntimeFilterType.values()) {
+      for (TRuntimeFilterType filterType : TRuntimeFilterType.values()) {
         for (Expr conjunct : joinConjuncts) {
-          RuntimeFilter filter = RuntimeFilter.create(filterIdGenerator,
-              ctx.getRootAnalyzer(), conjunct, joinNode, type, bloomFilterSizeLimits_);
-          if (filter == null) continue;
-          registerRuntimeFilter(filter);
-          filters.add(filter);
+          RuntimeFilter filter =
+              RuntimeFilter.create(filterIdGenerator, ctx.getRootAnalyzer(), conjunct,
+                  joinNode, filterType, bloomFilterSizeLimits_,
+                  /* isTimestampTruncation */ false);
+          if (filter != null) {
+            registerRuntimeFilter(filter);
+            filters.add(filter);
+          }
+          // For timestamp bloom filters, we also generate a RuntimeFilter with the
+          // src timestamp truncated for Kudu scan node targets.
+          if (filterType == TRuntimeFilterType.BLOOM
+              && Predicate.isEquivalencePredicate(conjunct)
+              && conjunct.getChild(0).getType().isTimestamp()
+              && conjunct.getChild(1).getType().isTimestamp()) {
+            RuntimeFilter filter2 =
+                RuntimeFilter.create(filterIdGenerator, ctx.getRootAnalyzer(), conjunct,
+                    joinNode, filterType, bloomFilterSizeLimits_,
+                    /* isTimestampTruncation */ true);
+            if (filter2 == null) continue;
+            registerRuntimeFilter(filter2);
+            filters.add(filter2);
+          }
         }
       }
       generateFilters(ctx, root.getChild(0));
@@ -674,10 +808,15 @@ public final class RuntimeFilterGenerator {
    *    to 'scanNode' if the filter is produced within the same fragment that contains the
    *    scan node.
    * 3. Only Hdfs and Kudu scan nodes are supported:
-   *     a. If the target is an HdfsScanNode, the filter must be type BLOOM.
-   *     b. If the target is a KuduScanNode, the filter must be type MIN_MAX, the target
-   *         must be a slot ref on a column, and the comp op cannot be 'not distinct'.
-   * A scan node may be used as a destination node for multiple runtime filters.
+   *     a. If the target is an HdfsScanNode, the filter must be type BLOOM for non
+   *        Parquet tables, or type BLOOM and/or MIN_MAX for Parquet tables.
+   *     b. If the target is a KuduScanNode, the filter could be type MIN_MAX, and/or
+   *        BLOOM, the target must be a slot ref on a column, and the comp op cannot
+   *        be 'not distinct'.
+   * A scan node may be used as a destination node for multiple runtime filters. This
+   * method is called once per scan node to process all filters accumulated for it
+   * for the entire query, per top-down traversal nature of the calling method
+   * generateFilters().
    */
   private void assignRuntimeFilters(PlannerContext ctx, ScanNode scanNode) {
     if (!(scanNode instanceof HdfsScanNode || scanNode instanceof KuduScanNode)) return;
@@ -686,7 +825,17 @@ public final class RuntimeFilterGenerator {
     Analyzer analyzer = ctx.getRootAnalyzer();
     boolean disableRowRuntimeFiltering =
         ctx.getQueryOptions().isDisable_row_runtime_filtering();
+    boolean disable_overlap_filter =
+        ctx.getQueryOptions().getMinmax_filter_threshold() == 0.0;
     TRuntimeFilterMode runtimeFilterMode = ctx.getQueryOptions().getRuntime_filter_mode();
+    TEnabledRuntimeFilterTypes enabledRuntimeFilterTypes =
+        ctx.getQueryOptions().getEnabled_runtime_filter_types();
+
+    // Init the overlap predicate for the hdfs scan node.
+    if (scanNode instanceof HdfsScanNode && !disable_overlap_filter) {
+      ((HdfsScanNode) scanNode).initOverlapPredicate(analyzer);
+    }
+
     for (RuntimeFilter filter: runtimeFiltersByTid_.get(tid)) {
       if (filter.isFinalized()) continue;
       Expr targetExpr = computeTargetExpr(filter, tid, analyzer);
@@ -698,30 +847,97 @@ public final class RuntimeFilterGenerator {
       if (runtimeFilterMode == TRuntimeFilterMode.LOCAL && !isLocalTarget) continue;
 
       // Check that the scan node supports applying filters of this type and targetExpr.
-      if (scanNode instanceof HdfsScanNode
-          && filter.getType() != TRuntimeFilterType.BLOOM) {
-        continue;
-      } else if (scanNode instanceof KuduScanNode) {
-        if (filter.getType() != TRuntimeFilterType.MIN_MAX) continue;
-        // TODO: IMPALA-9294: Support Kudu Date Min/Max Filters
-        if (targetExpr.getType().isDate()) continue;
-        SlotRef slotRef = targetExpr.unwrapSlotRef(true);
-        // Kudu only supports targeting a single column, not general exprs, so the target
-        // must be a SlotRef pointing to a column. We can allow implicit integer casts
-        // by casting the min/max values before sending them to Kudu.
-        // Kudu also cannot currently return nulls if a filter is applied, so it does not
-        // work with "is not distinct".
-        if (slotRef == null || slotRef.getDesc().getColumn() == null
-            || (targetExpr instanceof CastExpr && !targetExpr.getType().isIntegerType())
-            || filter.getExprCompOp() == Operator.NOT_DISTINCT) {
+      if (scanNode instanceof HdfsScanNode) {
+        if (filter.isTimestampTruncation()) {
           continue;
         }
+        if (filter.getType() == TRuntimeFilterType.MIN_MAX) {
+          boolean allow_min_max =
+              enabledRuntimeFilterTypes == TEnabledRuntimeFilterTypes.MIN_MAX
+              || enabledRuntimeFilterTypes == TEnabledRuntimeFilterTypes.ALL;
+          if (!allow_min_max) {
+            continue;
+          }
+          // TODO: Apply min/max filters on partition columns.
+          if (isBoundByPartitionColumns) {
+            continue;
+          }
+          if (!disable_overlap_filter) {
+            // If the filter is not defined on partition columns, try to compute
+            // an overlap predicate for it. This predicate will be used to filter
+            // out row groups or pages in Parquet data files.
+            if (!((HdfsScanNode) scanNode)
+                .tryToComputeOverlapPredicate(analyzer, filter, targetExpr)) {
+              continue;
+            }
+          } else {
+            continue;
+          }
+        }
+      } else {
+        Preconditions.checkState(scanNode instanceof KuduScanNode);
+        if (filter.getType() == TRuntimeFilterType.BLOOM) {
+          if (enabledRuntimeFilterTypes != TEnabledRuntimeFilterTypes.BLOOM
+              && enabledRuntimeFilterTypes != TEnabledRuntimeFilterTypes.ALL) {
+            continue;
+          }
+          // TODO: Support Kudu VARCHAR Bloom Filter
+          if (targetExpr.getType().isVarchar()) continue;
+          // Kudu only supports targeting a single column, not general exprs, so the
+          // target must be a SlotRef pointing to a column without casting.
+          // For timestamp bloom filter, assign it to Kudu if it has src timestamp
+          // truncation.
+          if (!(targetExpr instanceof SlotRef)
+              || filter.getExprCompOp() == Operator.NOT_DISTINCT
+              || (targetExpr.getType().isTimestamp()
+                     && !filter.isTimestampTruncation())) {
+            continue;
+          }
+          SlotRef slotRef = (SlotRef) targetExpr;
+          if (slotRef.getDesc().getColumn() == null) continue;
+        } else {
+          Preconditions.checkState(filter.getType() == TRuntimeFilterType.MIN_MAX);
+          if (enabledRuntimeFilterTypes != TEnabledRuntimeFilterTypes.MIN_MAX
+              && enabledRuntimeFilterTypes != TEnabledRuntimeFilterTypes.ALL) {
+            continue;
+          }
+          // TODO: IMPALA-9580: Support Kudu VARCHAR Min/Max Filters
+          if (targetExpr.getType().isVarchar()) continue;
+          SlotRef slotRef = targetExpr.unwrapSlotRef(true);
+          // Kudu only supports targeting a single column, not general exprs, so the
+          // target must be a SlotRef pointing to a column. We can allow implicit
+          // integer casts by casting the min/max values before sending them to Kudu.
+          // Kudu also cannot currently return nulls if a filter is applied, so it
+          // does not work with "is not distinct".
+          if (slotRef == null || slotRef.getDesc().getColumn() == null
+              || (targetExpr instanceof CastExpr && !targetExpr.getType().isIntegerType())
+              || filter.getExprCompOp() == Operator.NOT_DISTINCT) {
+            continue;
+          }
+        }
       }
-
-      RuntimeFilter.RuntimeFilterTarget target = new RuntimeFilter.RuntimeFilterTarget(
-          scanNode, targetExpr, isBoundByPartitionColumns, isLocalTarget);
+      TColumnValue lowValue = null;
+      TColumnValue highValue = null;
+      if (scanNode instanceof HdfsScanNode) {
+        SlotRef slotRefInScan = targetExpr.unwrapSlotRef(true);
+        if (slotRefInScan != null) {
+          Column col = slotRefInScan.getDesc().getColumn();
+          if (col != null) {
+            lowValue = col.getStats().getLowValue();
+            highValue = col.getStats().getHighValue();
+          }
+        }
+      }
+      RuntimeFilter.RuntimeFilterTarget target =
+          new RuntimeFilter.RuntimeFilterTarget(scanNode, targetExpr,
+              isBoundByPartitionColumns, isLocalTarget, lowValue, highValue);
       filter.addTarget(target);
     }
+
+    // finalize the overlap predicate for the hdfs scan node.
+    if (scanNode instanceof HdfsScanNode && !disable_overlap_filter) {
+      ((HdfsScanNode) scanNode).finalizeOverlapPredicate();
+    }
   }
 
   /**
@@ -784,7 +1000,7 @@ public final class RuntimeFilterGenerator {
         return null;
       }
     }
-    Type srcType = filter.getSrcExpr().getType();
+    Type srcType = filter.getSrcExprType();
     // Types of targetExpr and srcExpr must be exactly the same since runtime filters are
     // based on hashing.
     if (!targetExpr.getType().equals(srcType)) {
