@@ -42,10 +42,16 @@ public class BackendConfig {
   }
 
   public static void create(TBackendGflags cfg) {
+    BackendConfig.create(cfg, true);
+  }
+
+  public static void create(TBackendGflags cfg, boolean initialize) {
     Preconditions.checkNotNull(cfg);
     INSTANCE = new BackendConfig(cfg);
-    SqlScanner.init(cfg.getReserved_words_version());
-    initAuthToLocal();
+    if (initialize) {
+      SqlScanner.init(cfg.getReserved_words_version());
+      initAuthToLocal();
+    }
   }
 
   public TBackendGflags getBackendCfg() { return backendCfg_; }
@@ -70,9 +76,6 @@ public class BackendConfig {
   public String getImpalaBuildVersion() { return backendCfg_.impala_build_version; }
   public int getImpalaLogLevel() { return backendCfg_.impala_log_lvl; }
   public int getNonImpalaJavaVlogLevel() { return backendCfg_.non_impala_java_vlog; }
-  public long getSentryCatalogPollingFrequency() {
-    return backendCfg_.sentry_catalog_polling_frequency_s;
-  }
 
   public int maxHdfsPartsParallelLoad() {
     return backendCfg_.max_hdfs_partitions_parallel_load;
@@ -126,6 +129,8 @@ public class BackendConfig {
     return backendCfg_.hms_event_polling_interval_s;
   }
 
+  public boolean isInsertEventsEnabled() { return backendCfg_.enable_insert_events; }
+
   public boolean isOrcScannerEnabled() {
     return backendCfg_.enable_orc_scanner;
   }
@@ -141,14 +146,6 @@ public class BackendConfig {
     return "".equals(val) ? null : val;
   }
 
-  public boolean isMtDopUnlocked() {
-    return backendCfg_.unlock_mt_dop;
-  }
-
-  public boolean mtDopAutoFallback() {
-    return backendCfg_.mt_dop_auto_fallback;
-  }
-
   public boolean recursivelyListPartitions() {
     return backendCfg_.recursively_list_partitions;
   }
@@ -209,7 +206,60 @@ public class BackendConfig {
     backendCfg_.setEnable_column_masking(columnMaskingEnabled);
   }
 
+  public void setRowFilteringEnabled(boolean rowFilteringEnabled) {
+    backendCfg_.setEnable_row_filtering(rowFilteringEnabled);
+  }
+
   public boolean isColumnMaskingEnabled() { return backendCfg_.enable_column_masking; }
+  public boolean isRowFilteringEnabled() { return backendCfg_.enable_row_filtering; }
+
+  public boolean isCompactCatalogTopic() { return backendCfg_.compact_catalog_topic; }
+
+  public boolean isIncrementalMetadataUpdatesEnabled() {
+    return backendCfg_.enable_incremental_metadata_updates;
+  }
+
+  public String getSaml2KeystorePath() { return backendCfg_.saml2_keystore_path; }
+
+  public String getSaml2KeystorePassword() {
+    return backendCfg_.saml2_keystore_password;
+  }
+
+  public String getSaml2PrivateKeyPassword() {
+    return backendCfg_.saml2_private_key_password;
+  }
+
+  public String getSaml2IdpMetadata() { return backendCfg_.saml2_idp_metadata; }
+
+  public String getSaml2SpEntityId() { return backendCfg_.saml2_sp_entity_id; }
+
+  public String getSaml2SpCallbackUrl() { return backendCfg_.saml2_sp_callback_url; }
+
+  public boolean getSaml2WantAsserationsSigned() {
+    return backendCfg_.saml2_want_assertations_signed;
+  }
+
+  public boolean getSaml2SignRequest() { return backendCfg_.saml2_sign_requests; }
+
+  public int getSaml2CallbackTokenTtl() {
+    return backendCfg_.saml2_callback_token_ttl;
+  }
+
+  public String getSaml2GroupAttibuteName() { return backendCfg_.saml2_group_attribute_name; }
+
+  public String getSaml2GroupFilter() { return backendCfg_.saml2_group_filter; }
+
+  public boolean getSaml2EETestMode() { return backendCfg_.saml2_ee_test_mode; }
+
+  public String getScratchDirs() { return backendCfg_.scratch_dirs; }
+
+  public boolean getAllowOrdinalsInHaving() {
+    return backendCfg_.allow_ordinals_in_having;
+  }
+
+  public void setAllowOrdinalsInHaving(boolean allow_ordinals_in_having) {
+    backendCfg_.allow_ordinals_in_having = allow_ordinals_in_having;
+  }
 
   // Inits the auth_to_local configuration in the static KerberosName class.
   private static void initAuthToLocal() {
@@ -234,4 +284,8 @@ public class BackendConfig {
   public boolean isDedicatedCoordinator() {
     return (backendCfg_.is_executor == false) && (backendCfg_.is_coordinator == true);
   }
+
+  public int getMaxWaitTimeForSyncDdlSecs() {
+    return backendCfg_.max_wait_time_for_sync_ddl_s;
+  }
 }
