@@ -69,10 +69,11 @@ import com.google.common.base.Preconditions;
 public class FeSupport {
   private final static Logger LOG = LoggerFactory.getLogger(FeSupport.class);
   private static boolean loaded_ = false;
+  private static boolean externalFE_ = false;
 
-  // Only called if this library is explicitly loaded. This only happens
-  // when running FE tests.
-  public native static void NativeFeTestInit();
+  // Only called if this library is explicitly loaded. This happens
+  // when running FE tests or external FE
+  public native static void NativeFeInit(boolean externalFE);
 
   // Returns a serialized TResultRow
   public native static byte[] NativeEvalExprsWithoutRow(
@@ -86,8 +87,8 @@ public class FeSupport {
 
   // Adds a topic item to the backend's pending metadata-topic update.
   // 'serializationBuffer' is a serialized TCatalogObject.
-  // The return value is true if the operation succeeds and false otherwise.
-  public native static boolean NativeAddPendingTopicItem(long nativeCatalogServerPtr,
+  // Returns the actual value size and -1 if the operation fails.
+  public native static int NativeAddPendingTopicItem(long nativeCatalogServerPtr,
       String key, long version, byte[] serializationBuffer, boolean deleted);
 
   // Get a catalog object update from the backend. A pair of isDeletion flag and
@@ -115,9 +116,6 @@ public class FeSupport {
 
   public native static byte[] NativeUpdateTableUsage(byte[] thriftReq);
 
-  // Does an RPC to the Catalog Server to check if the given user is a Sentry admin.
-  public native static byte[] NativeSentryAdminCheck(byte[] thriftReq);
-
   // Parses a string of comma-separated key=value query options ('csvQueryOptions'),
   // updates the existing query options ('queryOptions') with them and returns the
   // resulting serialized TQueryOptions object.
@@ -444,15 +442,6 @@ public class FeSupport {
     return NativeGetPartialCatalogObject(thriftReq);
   }
 
-  public static byte[] CheckSentryAdmin(byte[] thriftReq) {
-    try {
-      return NativeSentryAdminCheck(thriftReq);
-    } catch (UnsatisfiedLinkError e) {
-      loadLibrary();
-    }
-    return NativeSentryAdminCheck(thriftReq);
-  }
-
   private static byte[] parseDateStringUtil(String date) {
     try {
       return nativeParseDateString(date);
@@ -482,6 +471,15 @@ public class FeSupport {
     }
   }
 
+  /**
+   * Calling this function before loadLibrary() causes external frontend
+   * initialization to be used during NativeFeInit()
+   */
+  public static synchronized void setExternalFE() {
+    Preconditions.checkState(!loaded_);
+    externalFE_ = true;
+  }
+
   /**
    * This function should be called explicitly by the FeSupport to ensure that
    * native functions are loaded. Tests that depend on JniCatalog or JniFrontend
@@ -493,6 +491,6 @@ public class FeSupport {
     NativeLibUtil.loadLibrary("libfesupport.so");
     LOG.info("Loaded libfesupport.so");
     loaded_ = true;
-    NativeFeTestInit();
+    NativeFeInit(externalFE_);
   }
 }
