@@ -46,6 +46,11 @@ import org.apache.hadoop.hive.metastore.api.LockLevel;
 import org.apache.hadoop.hive.metastore.api.LockType;
 import org.apache.hadoop.hive.metastore.api.SQLForeignKey;
 import org.apache.hadoop.hive.metastore.api.SQLPrimaryKey;
+import org.apache.iceberg.Table;
+import org.apache.iceberg.HistoryEntry;
+import org.apache.iceberg.Snapshot;
+import org.apache.iceberg.TableMetadata;
+import org.apache.iceberg.util.SnapshotUtil;
 import org.apache.impala.analysis.AlterDbStmt;
 import org.apache.impala.analysis.AnalysisContext;
 import org.apache.impala.analysis.AnalysisContext.AnalysisResult;
@@ -75,8 +80,12 @@ import org.apache.impala.analysis.StmtMetadataLoader;
 import org.apache.impala.analysis.StmtMetadataLoader.StmtTableCache;
 import org.apache.impala.analysis.TableName;
 import org.apache.impala.analysis.TruncateStmt;
+import org.apache.impala.authentication.saml.ImpalaSamlClient;
+import org.apache.impala.authorization.Authorizable;
+import org.apache.impala.authorization.AuthorizableTable;
 import org.apache.impala.authorization.AuthorizationChecker;
 import org.apache.impala.authorization.AuthorizationConfig;
+import org.apache.impala.authorization.AuthorizationException;
 import org.apache.impala.authorization.AuthorizationManager;
 import org.apache.impala.authorization.AuthorizationFactory;
 import org.apache.impala.authorization.ImpalaInternalAdminUser;
@@ -94,6 +103,7 @@ import org.apache.impala.catalog.FeDataSourceTable;
 import org.apache.impala.catalog.FeDb;
 import org.apache.impala.catalog.FeFsTable;
 import org.apache.impala.catalog.FeHBaseTable;
+import org.apache.impala.catalog.FeIcebergTable;
 import org.apache.impala.catalog.FeKuduTable;
 import org.apache.impala.catalog.FeTable;
 import org.apache.impala.catalog.Function;
@@ -101,8 +111,10 @@ import org.apache.impala.catalog.ImpaladCatalog;
 import org.apache.impala.catalog.ImpaladTableUsageTracker;
 import org.apache.impala.catalog.MetaStoreClientPool;
 import org.apache.impala.catalog.MetaStoreClientPool.MetaStoreClient;
+import org.apache.impala.catalog.TableLoadingException;
 import org.apache.impala.catalog.Type;
 import org.apache.impala.catalog.local.InconsistentMetadataFetchException;
+import org.apache.impala.common.AnalysisException;
 import org.apache.impala.common.FileSystemUtil;
 import org.apache.impala.common.ImpalaException;
 import org.apache.impala.common.InternalException;
@@ -119,6 +131,7 @@ import org.apache.impala.planner.PlanFragment;
 import org.apache.impala.planner.Planner;
 import org.apache.impala.planner.ScanNode;
 import org.apache.impala.thrift.TAlterDbParams;
+import org.apache.impala.thrift.TBackendGflags;
 import org.apache.impala.thrift.TCatalogOpRequest;
 import org.apache.impala.thrift.TCatalogOpType;
 import org.apache.impala.thrift.TCatalogServiceRequestHeader;
@@ -136,6 +149,8 @@ import org.apache.impala.thrift.TExplainResult;
 import org.apache.impala.thrift.TFinalizeParams;
 import org.apache.impala.thrift.TFunctionCategory;
 import org.apache.impala.thrift.TGetCatalogMetricsResult;
+import org.apache.impala.thrift.TGetTableHistoryResult;
+import org.apache.impala.thrift.TGetTableHistoryResultItem;
 import org.apache.impala.thrift.TGrantRevokePrivParams;
 import org.apache.impala.thrift.TGrantRevokeRoleParams;
 import org.apache.impala.thrift.TLineageGraph;
@@ -162,6 +177,7 @@ import org.apache.impala.thrift.TUpdateExecutorMembershipRequest;
 import org.apache.impala.util.AcidUtils;
 import org.apache.impala.util.EventSequence;
 import org.apache.impala.util.ExecutorMembershipSnapshot;
+import org.apache.impala.util.IcebergUtil;
 import org.apache.impala.util.PatternMatcher;
 import org.apache.impala.util.TResultRowBuilder;
 import org.apache.impala.util.TSessionStateUtil;
@@ -281,8 +297,11 @@ public class Frontend {
 
   private static ExecutorService checkAuthorizationPool_;
 
-  public Frontend(AuthorizationFactory authzFactory) throws ImpalaException {
-    this(authzFactory, FeCatalogManager.createFromBackendConfig());
+  private final ImpalaSamlClient saml2Client_;
+
+  public Frontend(AuthorizationFactory authzFactory, boolean isBackendTest)
+      throws ImpalaException {
+    this(authzFactory, FeCatalogManager.createFromBackendConfig(), isBackendTest);
   }
 
   /**
@@ -292,11 +311,12 @@ public class Frontend {
   @VisibleForTesting
   public Frontend(AuthorizationFactory authzFactory, FeCatalog testCatalog)
       throws ImpalaException {
-    this(authzFactory, FeCatalogManager.createForTests(testCatalog));
+    // This signature is only used for frontend tests, so pass false for isBackendTest
+    this(authzFactory, FeCatalogManager.createForTests(testCatalog), false);
   }
 
-  private Frontend(AuthorizationFactory authzFactory, FeCatalogManager catalogManager)
-      throws ImpalaException {
+  private Frontend(AuthorizationFactory authzFactory, FeCatalogManager catalogManager,
+      boolean isBackendTest) throws ImpalaException {
     catalogManager_ = catalogManager;
     authzFactory_ = authzFactory;
 
@@ -308,7 +328,7 @@ public class Frontend {
       Preconditions.checkState(numThreads > 0
         && numThreads <= MAX_CHECK_AUTHORIZATION_POOL_SIZE);
       if (numThreads == 1) {
-        checkAuthorizationPool_ = MoreExecutors.sameThreadExecutor();
+        checkAuthorizationPool_ = MoreExecutors.newDirectExecutorService();
       } else {
         LOG.info("Using a thread pool of size {} for authorization", numThreads);
         checkAuthorizationPool_ = Executors.newFixedThreadPool(numThreads);
@@ -323,12 +343,23 @@ public class Frontend {
     impaladTableUsageTracker_ = ImpaladTableUsageTracker.createFromConfig(
         BackendConfig.INSTANCE);
     queryHookManager_ = QueryEventHookManager.createFromConfig(BackendConfig.INSTANCE);
-    metaStoreClientPool_ = new MetaStoreClientPool(1, 0);
-    if (MetastoreShim.getMajorVersion() > 2) {
-      transactionKeepalive_ = new TransactionKeepalive(metaStoreClientPool_);
+    if (!isBackendTest) {
+      TBackendGflags cfg = BackendConfig.INSTANCE.getBackendCfg();
+      metaStoreClientPool_ = new MetaStoreClientPool(1, cfg.initial_hms_cnxn_timeout_s);
+      if (MetastoreShim.getMajorVersion() > 2) {
+        transactionKeepalive_ = new TransactionKeepalive(metaStoreClientPool_);
+      } else {
+        transactionKeepalive_ = null;
+      }
     } else {
+      metaStoreClientPool_ = null;
       transactionKeepalive_ = null;
     }
+    if (!BackendConfig.INSTANCE.getSaml2IdpMetadata().isEmpty()) {
+      saml2Client_ =  ImpalaSamlClient.get();
+    } else {
+      saml2Client_ = null;
+    }
   }
 
   /**
@@ -351,6 +382,8 @@ public class Frontend {
 
   public FeCatalog getCatalog() { return catalogManager_.getOrCreateCatalog(); }
 
+  public AuthorizationFactory getAuthzFactory() { return authzFactory_; }
+
   public AuthorizationChecker getAuthzChecker() { return authzChecker_.get(); }
 
   public AuthorizationManager getAuthzManager() { return authzManager_; }
@@ -359,6 +392,8 @@ public class Frontend {
     return impaladTableUsageTracker_;
   }
 
+  public ImpalaSamlClient getSaml2Client() { return saml2Client_; }
+
   public TUpdateCatalogCacheResponse updateCatalogCache(
       TUpdateCatalogCacheRequest req) throws ImpalaException, TException {
     TUpdateCatalogCacheResponse resp = catalogManager_.updateCatalogCache(req);
@@ -371,19 +406,12 @@ public class Frontend {
     return resp;
   }
 
-  /**
-   * Update the cluster membership snapshot with the latest snapshot from the backend.
-   */
-  public void updateExecutorMembership(TUpdateExecutorMembershipRequest req) {
-    ExecutorMembershipSnapshot.update(req);
-  }
-
   /**
    * Constructs a TCatalogOpRequest and attaches it, plus any metadata, to the
    * result argument.
    */
-  private void createCatalogOpRequest(AnalysisResult analysis,
-      TExecRequest result) throws InternalException {
+  private void createCatalogOpRequest(AnalysisResult analysis, TExecRequest result)
+      throws InternalException {
     TCatalogOpRequest ddl = new TCatalogOpRequest();
     TResultSetMetadata metadata = new TResultSetMetadata();
     if (analysis.isUseStmt()) {
@@ -437,6 +465,14 @@ public class Frontend {
       ddl.op_type = TCatalogOpType.SHOW_FILES;
       ddl.setShow_files_params(analysis.getShowFilesStmt().toThrift());
       metadata.setColumns(Collections.<TColumn>emptyList());
+    } else if (analysis.isDescribeHistoryStmt()) {
+      ddl.op_type = TCatalogOpType.DESCRIBE_HISTORY;
+      ddl.setDescribe_history_params(analysis.getDescribeHistoryStmt().toThrift());
+      metadata.setColumns(Arrays.asList(
+          new TColumn("creation_time", Type.STRING.toThrift()),
+          new TColumn("snapshot_id", Type.STRING.toThrift()),
+          new TColumn("parent_id", Type.STRING.toThrift()),
+          new TColumn("is_current_ancestor", Type.STRING.toThrift())));
     } else if (analysis.isDescribeDbStmt()) {
       ddl.op_type = TCatalogOpType.DESCRIBE_DB;
       ddl.setDescribe_db_params(analysis.getDescribeDbStmt().toThrift());
@@ -460,6 +496,9 @@ public class Frontend {
         columns.add(new TColumn("encoding", Type.STRING.toThrift()));
         columns.add(new TColumn("compression", Type.STRING.toThrift()));
         columns.add(new TColumn("block_size", Type.STRING.toThrift()));
+      } else if (descStmt.getTable() instanceof FeIcebergTable
+          && descStmt.getOutputStyle() == TDescribeOutputStyle.MINIMAL) {
+        columns.add(new TColumn("nullable", Type.STRING.toThrift()));
       }
       metadata.setColumns(columns);
     } else if (analysis.isAlterTableStmt()) {
@@ -658,11 +697,26 @@ public class Frontend {
       TClientRequest clientRequest = queryCtx.getClient_request();
       header.setRedacted_sql_stmt(clientRequest.isSetRedacted_stmt() ?
           clientRequest.getRedacted_stmt() : clientRequest.getStmt());
+      header.setWant_minimal_response(
+          BackendConfig.INSTANCE.getBackendCfg().use_local_catalog);
       ddl.getDdl_params().setHeader(header);
       ddl.getDdl_params().setSync_ddl(ddl.isSync_ddl());
-    }
-    if (ddl.getOp_type() == TCatalogOpType.RESET_METADATA) {
+      // forward debug_actions to the catalogd
+      if (result.getQuery_options().isSetDebug_action()) {
+        ddl.getDdl_params()
+            .setDebug_action(result.getQuery_options().getDebug_action());
+      }
+    } else if (ddl.getOp_type() == TCatalogOpType.RESET_METADATA) {
       ddl.getReset_metadata_params().setSync_ddl(ddl.isSync_ddl());
+      ddl.getReset_metadata_params().setRefresh_updated_hms_partitions(
+          result.getQuery_options().isRefresh_updated_hms_partitions());
+      ddl.getReset_metadata_params().getHeader().setWant_minimal_response(
+          BackendConfig.INSTANCE.getBackendCfg().use_local_catalog);
+      // forward debug_actions to the catalogd
+      if (result.getQuery_options().isSetDebug_action()) {
+        ddl.getReset_metadata_params()
+            .setDebug_action(result.getQuery_options().getDebug_action());
+      }
     }
   }
 
@@ -1037,6 +1091,35 @@ public class Frontend {
     return dbs;
   }
 
+  /**
+   *  Handles DESCRIBE HISTORY queries.
+   */
+  public TGetTableHistoryResult getTableHistory(String dbName, String tableName)
+      throws DatabaseNotFoundException, TableLoadingException {
+    FeTable feTable = getCatalog().getTable(dbName, tableName);
+    Preconditions.checkState(feTable instanceof FeIcebergTable);
+    FeIcebergTable feIcebergTable = (FeIcebergTable) feTable;
+    TableMetadata metadata = IcebergUtil.getIcebergTableMetadata(feIcebergTable);
+    Table table = IcebergUtil.loadTable(feIcebergTable);
+    Set<Long> ancestorIds = Sets.newHashSet(SnapshotUtil.currentAncestors(table));
+
+    TGetTableHistoryResult result = new TGetTableHistoryResult();
+    result.result = Lists.newArrayList();
+    for (HistoryEntry historyEntry : metadata.snapshotLog()) {
+      TGetTableHistoryResultItem resultItem = new TGetTableHistoryResultItem();
+      long snapshotId = historyEntry.snapshotId();
+      resultItem.setCreation_time(historyEntry.timestampMillis());
+      resultItem.setSnapshot_id(snapshotId);
+      Snapshot snapshot = metadata.snapshot(snapshotId);
+      if (snapshot != null && snapshot.parentId() != null) {
+        resultItem.setParent_id(snapshot.parentId());
+      }
+      resultItem.setIs_current_ancestor(ancestorIds.contains(snapshotId));
+      result.result.add(resultItem);
+    }
+    return result;
+  }
+
   /**
    * Check whether table/database is accessible to given user.
    */
@@ -1099,8 +1182,9 @@ public class Frontend {
    * level objects in the hierarchy.
    */
   private boolean authEngineSupportsDenyRules() {
+    // Sentry did not support deny rules, but Ranger does. So, this now returns true.
     // TODO: could check config for Ranger and return true if deny rules are disabled
-    return !authzFactory_.getAuthorizationConfig().getProviderName().equals("sentry");
+    return true;
   }
 
   /**
@@ -1116,20 +1200,20 @@ public class Frontend {
   /**
    * Generate result set and schema for a SHOW COLUMN STATS command.
    */
-  public TResultSet getColumnStats(String dbName, String tableName)
+  public TResultSet getColumnStats(String dbName, String tableName, boolean showMinMax)
       throws ImpalaException {
     RetryTracker retries = new RetryTracker(
         String.format("fetching column stats from %s.%s", dbName, tableName));
     while (true) {
       try {
-        return doGetColumnStats(dbName, tableName);
+        return doGetColumnStats(dbName, tableName, showMinMax);
       } catch(InconsistentMetadataFetchException e) {
         retries.handleRetryOrThrow(e);
       }
     }
   }
 
-  private TResultSet doGetColumnStats(String dbName, String tableName)
+  private TResultSet doGetColumnStats(String dbName, String tableName, boolean showMinMax)
       throws ImpalaException {
     FeTable table = getCatalog().getTable(dbName, tableName);
     TResultSet result = new TResultSet();
@@ -1142,13 +1226,38 @@ public class Frontend {
     resultSchema.addToColumns(new TColumn("#Nulls", Type.BIGINT.toThrift()));
     resultSchema.addToColumns(new TColumn("Max Size", Type.BIGINT.toThrift()));
     resultSchema.addToColumns(new TColumn("Avg Size", Type.DOUBLE.toThrift()));
+    resultSchema.addToColumns(new TColumn("#Trues", Type.BIGINT.toThrift()));
+    resultSchema.addToColumns(new TColumn("#Falses", Type.BIGINT.toThrift()));
+    if (showMinMax) {
+      resultSchema.addToColumns(new TColumn("Min", Type.STRING.toThrift()));
+      resultSchema.addToColumns(new TColumn("Max", Type.STRING.toThrift()));
+    }
 
     for (Column c: table.getColumnsInHiveOrder()) {
       TResultRowBuilder rowBuilder = new TResultRowBuilder();
-      // Add name, type, NDVs, numNulls, max size and avg size.
-      rowBuilder.add(c.getName()).add(c.getType().toSql())
-          .add(c.getStats().getNumDistinctValues()).add(c.getStats().getNumNulls())
-          .add(c.getStats().getMaxSize()).add(c.getStats().getAvgSize());
+      // Add name, type, NDVs, numNulls, max size, avg size, and conditionally
+      // the min value and max value.
+      if (showMinMax) {
+        rowBuilder.add(c.getName())
+            .add(c.getType().toSql())
+            .add(c.getStats().getNumDistinctValues())
+            .add(c.getStats().getNumNulls())
+            .add(c.getStats().getMaxSize())
+            .add(c.getStats().getAvgSize())
+            .add(c.getStats().getNumTrues())
+            .add(c.getStats().getNumFalses())
+            .add(c.getStats().getLowValueAsString())
+            .add(c.getStats().getHighValueAsString());
+      } else {
+        rowBuilder.add(c.getName())
+            .add(c.getType().toSql())
+            .add(c.getStats().getNumDistinctValues())
+            .add(c.getStats().getNumNulls())
+            .add(c.getStats().getMaxSize())
+            .add(c.getStats().getAvgSize())
+            .add(c.getStats().getNumTrues())
+            .add(c.getStats().getNumFalses());
+      }
       result.addToRows(rowBuilder.get());
     }
     return result;
@@ -1174,6 +1283,9 @@ public class Frontend {
       throws ImpalaException {
     FeTable table = getCatalog().getTable(dbName, tableName);
     if (table instanceof FeFsTable) {
+      if (table instanceof FeIcebergTable && op == TShowStatsOp.PARTITIONS) {
+        return FeIcebergTable.Utils.getPartitionSpecs((FeIcebergTable) table);
+      }
       return ((FeFsTable) table).getTableStats();
     } else if (table instanceof FeHBaseTable) {
       return ((FeHBaseTable) table).getTableStats();
@@ -1182,7 +1294,10 @@ public class Frontend {
     } else if (table instanceof FeKuduTable) {
       if (op == TShowStatsOp.RANGE_PARTITIONS) {
         return FeKuduTable.Utils.getRangePartitions((FeKuduTable) table);
+      } else if (op == TShowStatsOp.PARTITIONS) {
+        return FeKuduTable.Utils.getPartitions((FeKuduTable) table);
       } else {
+        Preconditions.checkState(op == TShowStatsOp.TABLE_STATS);
         return FeKuduTable.Utils.getTableStats((FeKuduTable) table);
       }
     } else {
@@ -1272,6 +1387,18 @@ public class Frontend {
       PrivilegeRequest privilegeRequest = new PrivilegeRequestBuilder(
           authzFactory_.getAuthorizableFactory())
           .allOf(Privilege.VIEW_METADATA).onTable(table).build();
+      // TODO(IMPALA-10122): Remove the following if-then statement once we can
+      // properly process a PrivilegeRequest for a view whose creation was not
+      // authorized.
+      if (privilegeRequest.getAuthorizable().getType() == Authorizable.Type.TABLE) {
+        if (((AuthorizableTable) privilegeRequest.getAuthorizable())
+            .isViewCreatedWithoutAuthz()) {
+          throw new AuthorizationException(String.format("User '%s' does not have " +
+                  "privileges to execute '%s' on: %s", user.getName(),
+              privilegeRequest.getPrivilege().toString(),
+              privilegeRequest.getAuthorizable().getFullTableName()));
+        }
+      }
       if (!authzChecker_.get().hasAccess(user, privilegeRequest)) {
         // Filter out columns that the user is not authorized to see.
         filteredColumns = new ArrayList<Column>();
@@ -1295,11 +1422,14 @@ public class Frontend {
       filteredColumns = table.getColumnsInHiveOrder();
     }
     if (outputStyle == TDescribeOutputStyle.MINIMAL) {
-      if (!(table instanceof FeKuduTable)) {
+      if (table instanceof FeKuduTable) {
+        return DescribeResultFactory.buildKuduDescribeMinimalResult(filteredColumns);
+      } else if (table instanceof FeIcebergTable) {
+        return DescribeResultFactory.buildIcebergDescribeMinimalResult(filteredColumns);
+      } else {
         return DescribeResultFactory.buildDescribeMinimalResult(
             Column.columnsToStruct(filteredColumns));
       }
-      return DescribeResultFactory.buildKuduDescribeMinimalResult(filteredColumns);
     } else {
       Preconditions.checkArgument(outputStyle == TDescribeOutputStyle.FORMATTED ||
           outputStyle == TDescribeOutputStyle.EXTENDED);
@@ -1352,8 +1482,8 @@ public class Frontend {
   /**
    * Return a TPlanExecInfo corresponding to the plan with root fragment 'planRoot'.
    */
-  private TPlanExecInfo createPlanExecInfo(PlanFragment planRoot, Planner planner,
-      TQueryCtx queryCtx, TQueryExecRequest queryExecRequest) {
+  public static TPlanExecInfo createPlanExecInfo(PlanFragment planRoot,
+      TQueryCtx queryCtx) {
     TPlanExecInfo result = new TPlanExecInfo();
     List<PlanFragment> fragments = planRoot.getFragmentsInPlanPreorder();
 
@@ -1417,14 +1547,15 @@ public class Frontend {
 
     // Compute resource requirements of the final plans.
     TQueryExecRequest result = new TQueryExecRequest();
-    planner.computeResourceReqs(planRoots, queryCtx, result);
+    Planner.computeResourceReqs(planRoots, queryCtx, result,
+        planner.getPlannerCtx(), planner.getAnalysisResult().isQueryStmt());
 
     // create per-plan exec info;
     // also assemble list of names of tables with missing or corrupt stats for
     // assembling a warning message
     for (PlanFragment planRoot: planRoots) {
       result.addToPlan_exec_info(
-          createPlanExecInfo(planRoot, planner, queryCtx, result));
+          createPlanExecInfo(planRoot, queryCtx));
     }
 
     // Optionally disable spilling in the backend. Allow spilling if there are plan hints
@@ -1517,8 +1648,9 @@ public class Frontend {
     // Parse stmt and collect/load metadata to populate a stmt-local table cache
     StatementBase stmt = Parser.parse(
         queryCtx.client_request.stmt, queryCtx.client_request.query_options);
+    User user = new User(TSessionStateUtil.getEffectiveUser(queryCtx.session));
     StmtMetadataLoader metadataLoader =
-        new StmtMetadataLoader(this, queryCtx.session.database, timeline);
+        new StmtMetadataLoader(this, queryCtx.session.database, timeline, user);
     //TODO (IMPALA-8788): should load table write ids in transaction context.
     StmtTableCache stmtTableCache = metadataLoader.loadTables(stmt);
 
@@ -1544,8 +1676,9 @@ public class Frontend {
         if (!analysisResult.isCreateTableAsSelectStmt()) {
           return result;
         }
-      } else if (analysisResult.isInsertStmt() ||
-          analysisResult.isCreateTableAsSelectStmt()) {
+      }
+      if (!analysisResult.isExplainStmt() &&
+          (analysisResult.isInsertStmt() || analysisResult.isCreateTableAsSelectStmt())) {
         InsertStmt insertStmt = analysisResult.getInsertStmt();
         FeTable targetTable = insertStmt.getTargetTable();
         if (AcidUtils.isTransactionalTable(
@@ -1683,21 +1816,39 @@ public class Frontend {
    */
   private static void addFinalizationParamsForInsert(
       TQueryCtx queryCtx, TQueryExecRequest queryExecRequest, InsertStmt insertStmt) {
-    if (insertStmt.getTargetTable() instanceof FeFsTable) {
+    FeTable targetTable = insertStmt.getTargetTable();
+    addFinalizationParamsForInsert(queryCtx, queryExecRequest,
+        targetTable, insertStmt.getWriteId(), insertStmt.isOverwrite());
+  }
+
+  // This is public to allow external frontends to utilize this method to fill in the
+  // finalization parameters for externally generated INSERTs.
+  public static void addFinalizationParamsForInsert(
+      TQueryCtx queryCtx, TQueryExecRequest queryExecRequest, FeTable targetTable,
+          long writeId, boolean isOverwrite) {
+    if (targetTable instanceof FeFsTable) {
       TFinalizeParams finalizeParams = new TFinalizeParams();
-      finalizeParams.setIs_overwrite(insertStmt.isOverwrite());
-      finalizeParams.setTable_name(insertStmt.getTargetTableName().getTbl());
+      finalizeParams.setIs_overwrite(isOverwrite);
+      finalizeParams.setTable_name(targetTable.getTableName().getTbl());
       finalizeParams.setTable_id(DescriptorTable.TABLE_SINK_ID);
-      String db = insertStmt.getTargetTableName().getDb();
+      String db = targetTable.getTableName().getDb();
       finalizeParams.setTable_db(db == null ? queryCtx.session.database : db);
-      FeFsTable hdfsTable = (FeFsTable) insertStmt.getTargetTable();
+      FeFsTable hdfsTable = (FeFsTable) targetTable;
       finalizeParams.setHdfs_base_dir(hdfsTable.getHdfsBaseDir());
-      finalizeParams.setStaging_dir(
-          hdfsTable.getHdfsBaseDir() + "/_impala_insert_staging");
-      if (insertStmt.getWriteId() != -1) {
+      if (writeId != -1) {
         Preconditions.checkState(queryCtx.isSetTransaction_id());
         finalizeParams.setTransaction_id(queryCtx.getTransaction_id());
-        finalizeParams.setWrite_id(insertStmt.getWriteId());
+        finalizeParams.setWrite_id(writeId);
+      } else if (targetTable instanceof FeIcebergTable) {
+        FeIcebergTable iceTable = (FeIcebergTable)targetTable;
+        finalizeParams.setSpec_id(iceTable.getDefaultPartitionSpecId());
+      } else {
+        // TODO: Currently this flag only controls the removal of the query-level staging
+        // directory. HdfsTableSink (that creates the staging dir) calculates the path
+        // independently. So it'd be better to either remove this option, or make it used
+        // everywhere where the staging directory is referenced.
+        finalizeParams.setStaging_dir(
+            hdfsTable.getHdfsBaseDir() + "/_impala_insert_staging");
       }
       queryExecRequest.setFinalize_params(finalizeParams);
     }
@@ -1904,6 +2055,19 @@ public class Frontend {
         = this.queryHookManager_.executeQueryCompleteHooks(context);
   }
 
+  /**
+   * Adds a transaction to the keepalive object.
+   * @param queryCtx context that the transaction is associated with
+   * @throws TransactionException
+   */
+  public void addTransaction(TQueryCtx queryCtx) throws TransactionException {
+    Preconditions.checkState(queryCtx.isSetTransaction_id());
+    long transactionId = queryCtx.getTransaction_id();
+    HeartbeatContext ctx = new HeartbeatContext(queryCtx, System.nanoTime());
+    transactionKeepalive_.addTransaction(transactionId, ctx);
+    LOG.info("Opened transaction: " + Long.toString(transactionId));
+  }
+
   /**
    * Opens a new transaction and registers it to the keepalive object.
    * @param queryCtx context of the query that requires the transaction.
@@ -1913,13 +2077,10 @@ public class Frontend {
   private long openTransaction(TQueryCtx queryCtx) throws TransactionException {
     try (MetaStoreClient client = metaStoreClientPool_.getClient()) {
       IMetaStoreClient hmsClient = client.getHiveClient();
-      long transactionId = MetastoreShim.openTransaction(hmsClient);
-      HeartbeatContext ctx = new HeartbeatContext(queryCtx, System.nanoTime());
-      transactionKeepalive_.addTransaction(transactionId, ctx);
-      LOG.info("Opened transaction: " + Long.toString(transactionId));
-      queryCtx.setTransaction_id(transactionId);
-      return transactionId;
+      queryCtx.setTransaction_id(MetastoreShim.openTransaction(hmsClient));
+      addTransaction(queryCtx);
     }
+    return queryCtx.getTransaction_id();
   }
 
   /**
@@ -1980,7 +2141,9 @@ public class Frontend {
     Preconditions.checkState(
         AcidUtils.isTransactionalTable(targetTable.getMetaStoreTable().getParameters()));
     List<LockComponent> lockComponents = new ArrayList<>(tables.size());
-    for (FeTable table : tables) {
+    List<FeTable> lockTables = new ArrayList<>(tables);
+    if (!lockTables.contains(targetTable)) lockTables.add(targetTable);
+    for (FeTable table : lockTables) {
       if (!AcidUtils.isTransactionalTable(table.getMetaStoreTable().getParameters())) {
         continue;
       }
