@@ -170,6 +170,10 @@ public class KuduUtil {
         checkCorrectType(literal.isSetInt_literal(), type, colName, literal);
         key.addLong(pos, literal.getInt_literal().getValue());
         break;
+      case VARCHAR:
+        checkCorrectType(literal.isSetString_literal(), type, colName, literal);
+        key.addVarchar(pos, literal.getString_literal().getValue());
+        break;
       case STRING:
         checkCorrectType(literal.isSetString_literal(), type, colName, literal);
         key.addString(pos, literal.getString_literal().getValue());
@@ -227,6 +231,7 @@ public class KuduUtil {
       case DOUBLE:
         checkCorrectType(literal.isSetFloat_literal(), type, colName, literal);
         return (double) literal.getFloat_literal().getValue();
+      case VARCHAR:
       case STRING:
         checkCorrectType(literal.isSetString_literal(), type, colName, literal);
         return literal.getString_literal().getValue();
@@ -440,13 +445,13 @@ public class KuduUtil {
       case TIMESTAMP: return org.apache.kudu.Type.UNIXTIME_MICROS;
       case DECIMAL: return org.apache.kudu.Type.DECIMAL;
       case DATE: return org.apache.kudu.Type.DATE;
+      case VARCHAR: return org.apache.kudu.Type.VARCHAR;
       /* Fall through below */
       case INVALID_TYPE:
       case NULL_TYPE:
       case BINARY:
       case DATETIME:
       case CHAR:
-      case VARCHAR:
       default:
         throw new ImpalaRuntimeException(format(
             "Type %s is not supported in Kudu", s.toSql()));
@@ -469,6 +474,7 @@ public class KuduUtil {
       case DECIMAL:
         return ScalarType.createDecimalType(
             typeAttributes.getPrecision(), typeAttributes.getScale());
+      case VARCHAR: return ScalarType.createVarcharType(typeAttributes.getLength());
       default:
         throw new ImpalaRuntimeException(String.format(
             "Kudu type '%s' is not supported in Impala", t.getName()));
