@@ -19,7 +19,6 @@ package org.apache.impala.analysis;
 
 import org.apache.impala.catalog.KuduTable;
 import org.apache.impala.common.FrontendTestBase;
-import org.apache.impala.service.BackendConfig;
 import org.apache.impala.testutil.TestUtils;
 import org.apache.kudu.ColumnSchema.CompressionAlgorithm;
 import org.apache.kudu.ColumnSchema.Encoding;
@@ -72,7 +71,6 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
   }
 
   private void testDDlsOnKuduTable(boolean isExternalPurgeTbl) {
-    TestUtils.assumeKuduIsSupported();
     // Test primary keys and partition by clauses
     AnalyzesOk("create table tab (x int primary key) partition by hash(x) " +
         "partitions 8 stored as kudu", isExternalPurgeTbl);
@@ -298,7 +296,7 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
 
 
     // Test unsupported Kudu types
-    List<String> unsupportedTypes = Lists.newArrayList("VARCHAR(20)", "CHAR(20)",
+    List<String> unsupportedTypes = Lists.newArrayList("CHAR(20)",
         "STRUCT<f1:INT,f2:STRING>", "ARRAY<INT>", "MAP<STRING,STRING>");
     for (String t: unsupportedTypes) {
       String expectedError = String.format(
@@ -351,7 +349,7 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
         "i2 smallint default null, i3 int default null, i4 bigint default null, " +
         "vals string default null, valf float default null, vald double default null, " +
         "valb boolean default null, valdec decimal(10, 5) default null, " +
-        "valdate date default null) " +
+        "valdate date default null, valvc varchar(10) default null) " +
         "partition by hash (x) partitions 3 stored as kudu", isExternalPurgeTbl);
     // Use NULL as a default value on a non-nullable column
     AnalysisError("create table tab (x int primary key, y int not null default null) " +
@@ -387,6 +385,7 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
         "cast(3.1452 as double), valb boolean default true, " +
         "valdec decimal(10, 5) default 3.14159, " +
         "valdate date default date '1970-01-01', " +
+        "valvc varchar(10) default cast('test' as varchar(10)), " +
         "primary key (i1, i2, i3, i4, vals)) partition by hash (i1) partitions 3 " +
         "stored as kudu", isExternalPurgeTbl);
     AnalyzesOk("create table tab (i int primary key default 1+1+1) " +
@@ -431,15 +430,11 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
         "partitions 8 sort by(i) stored as kudu", "SORT BY is not supported for Kudu " +
         "tables.", isExternalPurgeTbl);
 
-    // Z-Sort columns are not supported for Kudu tables.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
-
+    // Z-Order sorted columns are not supported for Kudu tables.
     AnalysisError("create table tab (i int, x int primary key) partition by hash(x) " +
         "partitions 8 sort by zorder(i) stored as kudu", "SORT BY is not " +
         "supported for Kudu tables.", isExternalPurgeTbl);
 
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
-
     // Range partitions with TIMESTAMP
     AnalyzesOk("create table ts_ranges (ts timestamp primary key) " +
         "partition by range (partition cast('2009-01-01 00:00:00' as timestamp) " +
@@ -485,8 +480,7 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
     AnalysisError("create table tdefault (id int primary key, " +
         "ts timestamp not null default cast('00:00:00' as timestamp)) " +
         "partition by hash(id) partitions 3 stored as kudu",
-        "CAST('00:00:00' AS TIMESTAMP) cannot be cast to a TIMESTAMP literal.",
-        isExternalPurgeTbl);
+        "Default value of NULL not allowed on non-nullable column:", isExternalPurgeTbl);
     AnalysisError("create table tdefault (id int primary key, " +
         "ts timestamp not null default '2009-1 foo') " +
         "partition by hash(id) partitions 3 stored as kudu",
@@ -523,7 +517,6 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
 
   @Test
   public void TestCreateExternalKuduTable() {
-    TestUtils.assumeKuduIsSupported();
     final String kuduMasters = catalog_.getDefaultKuduMasterHosts();
     AnalyzesOk("create external table t stored as kudu " +
         "tblproperties('kudu.table_name'='t')");
@@ -591,7 +584,6 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
 
   @Test
   public void TestAlterKuduTable() {
-    TestUtils.assumeKuduIsSupported();
     // ALTER TABLE ADD/DROP range partitions
     String[] addDrop = {"add if not exists", "add", "drop if exists", "drop"};
     for (String kw: addDrop) {
@@ -717,8 +709,6 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
     AnalysisError("alter table functional_kudu.alltypes sort by (int_col)",
         "ALTER TABLE SORT BY not supported on Kudu tables.");
 
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
-
     // ALTER TABLE SORT BY ZORDER
     AnalysisError("alter table functional_kudu.alltypes sort by zorder (int_col)",
         "ALTER TABLE SORT BY not supported on Kudu tables.");
@@ -733,6 +723,7 @@ public class AnalyzeKuduDDLTest extends FrontendTestBase {
         + "'sort.order'='true')",
         "'sort.*' table properties are not supported for Kudu tables.");
 
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
+    // ALTER TABLE SET OWNER USER
+    AnalyzesOk("ALTER TABLE functional_kudu.testtbl SET OWNER USER new_owner");
   }
 }
