@@ -300,6 +300,16 @@ public class AuditingTest extends FrontendTestBase {
             "functional_seq_snap.alltypes", TCatalogObjectType.TABLE, "ALTER"),
         new TAccessEvent(
             "functional_seq_snap.alltypes", TCatalogObjectType.TABLE, "SELECT")));
+
+    // COMPUTE STATS results in two registrations of the ALTER event. Check we do not have
+    // duplicate ALTER events when the fully-qualified table name is not in lowercase.
+    accessEvents = AnalyzeAccessEvents(
+        "COMPUTE STATS FUNCTIONAL_SEQ_SNAP.ALLTYPES");
+    Assert.assertEquals(accessEvents, Sets.newHashSet(
+        new TAccessEvent(
+            "functional_seq_snap.alltypes", TCatalogObjectType.TABLE, "ALTER"),
+        new TAccessEvent(
+            "functional_seq_snap.alltypes", TCatalogObjectType.TABLE, "SELECT")));
   }
 
   @Test
@@ -345,7 +355,7 @@ public class AuditingTest extends FrontendTestBase {
   }
 
   @Test
-  public void TestShow() throws AnalysisException, AuthorizationException{
+  public void TestShowViewMetadata() throws AnalysisException, AuthorizationException{
     String[] statsQuals = new String[]{ "partitions", "table stats", "column stats" };
     for (String qual: statsQuals) {
       Set<TAccessEvent> accessEvents =
@@ -355,6 +365,18 @@ public class AuditingTest extends FrontendTestBase {
     }
   }
 
+  @Test
+  public void TestShowAny() throws AnalysisException, AuthorizationException {
+    String[] statsQuals = new String[] { "tables", "functions" };
+    for (String qual : statsQuals) {
+        Set<TAccessEvent> accessEvents = AnalyzeAccessEvents(String.format(
+            "show %s in functional", qual));
+        Assert.assertEquals(accessEvents,
+            Sets.newHashSet(new TAccessEvent("functional", TCatalogObjectType
+            .DATABASE, "ANY")));
+    }
+  }
+
   @Test
   public void TestShowCreateTable() throws AuthorizationException, AnalysisException {
     Set<TAccessEvent> accessEvents =
