@@ -32,10 +32,12 @@ import org.apache.impala.common.ImpalaException;
 import org.apache.impala.common.QueryFixture;
 import org.apache.impala.common.SqlCastException;
 import org.apache.impala.rewrite.BetweenToCompoundRule;
+import org.apache.impala.rewrite.ConvertToCNFRule;
 import org.apache.impala.rewrite.EqualityDisjunctsToInRule;
 import org.apache.impala.rewrite.ExprRewriteRule;
 import org.apache.impala.rewrite.ExprRewriter;
 import org.apache.impala.rewrite.ExtractCommonConjunctRule;
+import org.apache.impala.rewrite.ExtractCompoundVerticalBarExprRule;
 import org.apache.impala.rewrite.FoldConstantsRule;
 import org.apache.impala.rewrite.NormalizeBinaryPredicatesRule;
 import org.apache.impala.rewrite.NormalizeCountStarRule;
@@ -175,6 +177,19 @@ public class ExprRewriteRulesTest extends FrontendTestBase {
     return RewritesOk(tableName, exprStr, Lists.newArrayList(rule), expectedExprStr);
   }
 
+  /**
+   * Given a list of `rule`, this function checks whether the rewritten <expr> is as
+   * expected.
+   *
+   * If no rule in `rules` is expected to be fired, callers should set expectedExprStr
+   * to null or "NULL". Otherwise, this function would throw an exception like
+   * "Rule xxx didn't fire"
+   *
+   * @param exprStr: origin expr
+   * @param rules: list of rewrite rules
+   * @param expectedExprStr: expected expr
+   * @return: Expr: rewritten expr
+   */
   public Expr RewritesOk(String exprStr, List<ExprRewriteRule> rules,
       String expectedExprStr)
       throws ImpalaException {
@@ -424,8 +439,8 @@ public class ExprRewriteRulesTest extends FrontendTestBase {
   public void testCompoundPredicate() throws ImpalaException {
     ExprRewriteRule rule = SimplifyConditionalsRule.INSTANCE;
 
-    RewritesOk("false || id = 0", rule, "id = 0");
-    RewritesOk("true || id = 0", rule, "TRUE");
+    RewritesOk("false OR id = 0", rule, "id = 0");
+    RewritesOk("true OR id = 0", rule, "TRUE");
     RewritesOk("false && id = 0", rule, "FALSE");
     RewritesOk("true && id = 0", rule, "id = 0");
   }
@@ -771,6 +786,53 @@ public class ExprRewriteRulesTest extends FrontendTestBase {
     RewritesOk("if(bool_col <=> NULL, 1, 2)", rules, null);
   }
 
+  @Test
+  public void testCountDistinctToNdvRule() throws ImpalaException {
+    List<ExprRewriteRule> rules = Lists.newArrayList(
+            org.apache.impala.rewrite.CountDistinctToNdvRule.INSTANCE
+    );
+    session.options().setAppx_count_distinct(true);
+    RewritesOk("count(distinct bool_col)", rules, "ndv(bool_col)");
+  }
+
+  @Test
+  public void testDefaultNdvScaleRule() throws ImpalaException {
+    List<ExprRewriteRule> rules = Lists.newArrayList(
+            org.apache.impala.rewrite.DefaultNdvScaleRule.INSTANCE
+    );
+    session.options().setDefault_ndv_scale(10);
+    RewritesOk("ndv(bool_col)", rules, "ndv(bool_col, 10)");
+  }
+
+  @Test
+  public void testDefaultNdvScaleRuleNotSet() throws ImpalaException {
+    List<ExprRewriteRule> rules = Lists.newArrayList(
+            org.apache.impala.rewrite.DefaultNdvScaleRule.INSTANCE
+    );
+    RewritesOk("ndv(bool_col)", rules, null);
+  }
+
+  @Test
+  public void testDefaultNdvScaleRuleSetDefault() throws ImpalaException {
+    List<ExprRewriteRule> rules = Lists.newArrayList(
+            org.apache.impala.rewrite.DefaultNdvScaleRule.INSTANCE
+    );
+    session.options().setDefault_ndv_scale(2);
+    RewritesOk("ndv(bool_col)", rules, null);
+  }
+
+
+  @Test
+  public void testCountDistinctToNdvAndDefaultNdvScaleRule() throws ImpalaException {
+    List<ExprRewriteRule> rules = Lists.newArrayList(
+            org.apache.impala.rewrite.CountDistinctToNdvRule.INSTANCE,
+            org.apache.impala.rewrite.DefaultNdvScaleRule.INSTANCE
+    );
+    session.options().setAppx_count_distinct(true);
+    session.options().setDefault_ndv_scale(10);
+    RewritesOk("count(distinct bool_col)", rules, "ndv(bool_col, 10)");
+  }
+
   @Test
   public void testSimplifyCastStringToTimestamp() throws ImpalaException {
     ExprRewriteRule rule = SimplifyCastStringToTimestamp.INSTANCE;
@@ -804,4 +866,53 @@ public class ExprRewriteRulesTest extends FrontendTestBase {
     // are not simplified
     RewritesOk("nullif(1 + int_col, 1 + int_col)", rules, "NULL");
   }
+
+  @Test
+  public void testConvertToCNFRule() throws ImpalaException {
+    ExprRewriteRule rule = new ConvertToCNFRule(-1, false);
+
+    RewritesOk("(int_col > 10 AND int_col < 20) OR float_col < 5.0", rule,
+            "int_col < 20 OR float_col < 5.0 AND int_col > 10 OR float_col < 5.0");
+    RewritesOk("float_col < 5.0 OR (int_col > 10 AND int_col < 20)", rule,
+            "float_col < 5.0 OR int_col < 20 AND float_col < 5.0 OR int_col > 10");
+    RewritesOk("(int_col > 10 AND float_col < 5.0) OR " +
+            "(int_col < 20 AND float_col > 15.0)", rule,
+            "float_col < 5.0 OR float_col > 15.0 AND " +
+                    "float_col < 5.0 OR int_col < 20 " +
+                    "AND int_col > 10 OR float_col > 15.0 AND int_col > 10 " +
+                    "OR int_col < 20");
+    RewritesOk("NOT(int_col > 10 OR int_col < 20)", rule,
+            "NOT int_col < 20 AND NOT int_col > 10");
+  }
+
+  @Test
+  public void testExtractCompoundVerticalBarExprRule() throws ImpalaException {
+    ExprRewriteRule extractCompoundVerticalBarExprRule =
+        ExtractCompoundVerticalBarExprRule.INSTANCE;
+    ExprRewriteRule simplifyConditionalsRule = SimplifyConditionalsRule.INSTANCE;
+    List<ExprRewriteRule> rules = new ArrayList<>();
+    rules.add(extractCompoundVerticalBarExprRule);
+    rules.add(simplifyConditionalsRule);
+
+    RewritesOk("string_col || string_col", extractCompoundVerticalBarExprRule,
+        "concat(string_col, string_col)");
+
+    RewritesOk("bool_col || bool_col", extractCompoundVerticalBarExprRule,
+        "bool_col OR bool_col");
+
+    RewritesOk("string_col || 'TEST'", extractCompoundVerticalBarExprRule,
+        "concat(string_col, 'TEST')");
+
+    RewritesOk("functional.chars_tiny", "cl || cs", extractCompoundVerticalBarExprRule,
+        "concat(cl, cs)");
+
+    RewritesOk("FALSE || id = 0", extractCompoundVerticalBarExprRule, "FALSE OR id = 0");
+
+    RewritesOk("FALSE || id = 0", rules, "id = 0");
+
+    RewritesOk("'' || ''", extractCompoundVerticalBarExprRule, "concat('', '')");
+
+    RewritesOk(
+        "NULL || bool_col", extractCompoundVerticalBarExprRule, "NULL OR bool_col");
+  }
 }
