@@ -23,7 +23,6 @@ import static org.junit.Assert.fail;
 import org.apache.impala.authorization.Privilege;
 import org.apache.impala.common.AnalysisException;
 import org.apache.impala.common.FrontendTestBase;
-import org.apache.impala.service.BackendConfig;
 import org.apache.impala.testutil.TestUtils;
 import org.junit.Test;
 
@@ -345,15 +344,11 @@ public class ToSqlTest extends FrontendTestBase {
         "SORT BY LEXICAL ( a, b ) STORED AS TEXTFILE" , true);
 
     // Table with SORT BY ZORDER clause.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
-
     testToSql("create table p (a int, b int) partitioned by (day string) sort by zorder" +
         "(a ,b) ", "default",
         "CREATE TABLE default.p ( a INT, b INT ) PARTITIONED BY ( day STRING ) " +
         "SORT BY ZORDER ( a, b ) STORED AS TEXTFILE" , true);
 
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
-
     // Kudu table with a TIMESTAMP column default value
     String kuduMasters = catalog_.getDefaultKuduMasterHosts();
     testToSql(String.format("create table p (a bigint primary key, " +
@@ -412,14 +407,12 @@ public class ToSqlTest extends FrontendTestBase {
         "( string_col ) STORED AS TEXTFILE AS SELECT double_col, string_col, int_col " +
         "FROM functional.alltypes", true);
     // Table with SORT BY ZORDER clause.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
     testToSql("create table p partitioned by (string_col) sort by zorder (int_col, " +
         "bool_col) as select int_col, bool_col, string_col from functional.alltypes",
         "default",
         "CREATE TABLE default.p PARTITIONED BY ( string_col ) SORT BY ZORDER " +
         "( int_col, bool_col ) STORED AS TEXTFILE AS SELECT " +
         "int_col, bool_col, string_col FROM functional.alltypes", true);
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
     // Kudu table with multiple partition params
     String kuduMasters = catalog_.getDefaultKuduMasterHosts();
     testToSql(String.format("create table p primary key (a,b) " +
@@ -445,11 +438,9 @@ public class ToSqlTest extends FrontendTestBase {
     testToSql("create table p sort by LEXICAL (id) like functional.alltypes", "default",
         "CREATE TABLE p SORT BY LEXICAL (id) LIKE functional.alltypes");
     // Table with ZORDER sort columns.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
     testToSql("create table p sort by zorder (bool_col, int_col) like " +
         "functional.alltypes",  "default",
         "CREATE TABLE p SORT BY ZORDER (bool_col,int_col) LIKE functional.alltypes");
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
   }
 
   @Test
@@ -473,13 +464,11 @@ public class ToSqlTest extends FrontendTestBase {
         "'hdfs://localhost:20500/test-warehouse/schemas/alltypestiny.parquet' " +
         "SORT BY LEXICAL ( int_col, id ) STORED AS TEXTFILE", true);
     // Table with ZORDER sort columns.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
     testToSql("create table if not exists p like parquet " +
         "'/test-warehouse/schemas/alltypestiny.parquet' sort by zorder (int_col, id)",
         "default", "CREATE TABLE IF NOT EXISTS default.p LIKE PARQUET " +
         "'hdfs://localhost:20500/test-warehouse/schemas/alltypestiny.parquet' " +
         "SORT BY ZORDER ( int_col, id ) STORED AS TEXTFILE", true);
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
   }
 
   @Test
@@ -825,6 +814,30 @@ public class ToSqlTest extends FrontendTestBase {
         "having avg(tinyint_col) > 10 AND count(tinyint_col) > 5",
         "SELECT sum(id) FROM functional.alltypes GROUP BY tinyint_col " +
         "HAVING avg(tinyint_col) > 10 AND count(tinyint_col) > 5");
+
+    // CUBE, ROLLUP.
+    testToSql("select int_col, string_col, sum(id) from functional.alltypes " +
+        "group by rollup(int_col, string_col)",
+        "SELECT int_col, string_col, sum(id) FROM functional.alltypes " +
+        "GROUP BY ROLLUP(int_col, string_col)");
+    testToSql("select int_col, string_col, sum(id) from functional.alltypes " +
+        "group by int_col, string_col with rollup",
+        "SELECT int_col, string_col, sum(id) FROM functional.alltypes " +
+        "GROUP BY ROLLUP(int_col, string_col)");
+    testToSql("select int_col, string_col, sum(id) from functional.alltypes " +
+        "group by cube(int_col, string_col)",
+        "SELECT int_col, string_col, sum(id) FROM functional.alltypes " +
+        "GROUP BY CUBE(int_col, string_col)");
+    testToSql("select int_col, string_col, sum(id) from functional.alltypes " +
+        "group by int_col, string_col with cube",
+        "SELECT int_col, string_col, sum(id) FROM functional.alltypes " +
+        "GROUP BY CUBE(int_col, string_col)");
+
+    // GROUPING SETS.
+    testToSql("select int_col, string_col, sum(id) from functional.alltypes " +
+        "group by grouping sets((int_col, string_col), (int_col), ())",
+        "SELECT int_col, string_col, sum(id) FROM functional.alltypes " +
+        "GROUP BY GROUPING SETS((int_col, string_col), (int_col), ())");
   }
 
   // Test the toSql() output of the order by clause.
@@ -1063,7 +1076,6 @@ public class ToSqlTest extends FrontendTestBase {
 
   @Test
   public void TestUpdate() {
-    TestUtils.assumeKuduIsSupported();
     testToSql("update functional_kudu.dimtbl set name = '10' where name < '11'",
         "UPDATE functional_kudu.dimtbl SET name = '10' FROM functional_kudu.dimtbl " +
             "WHERE name < '11'");
@@ -1086,7 +1098,6 @@ public class ToSqlTest extends FrontendTestBase {
 
   @Test
   public void TestDelete() {
-    TestUtils.assumeKuduIsSupported();
     testToSql("delete functional_kudu.testtbl where zip = 10",
         "DELETE FROM functional_kudu.testtbl WHERE zip = 10");
     testToSql("delete from functional_kudu.testtbl where zip = 10",
@@ -1482,6 +1493,9 @@ public class ToSqlTest extends FrontendTestBase {
         "NOT TRUE, (NOT TRUE), NOT FALSE, (NOT FALSE)");
     testToSql("select ((true and (false or false) or true) and (false or true))",
         "SELECT ((TRUE AND (FALSE OR FALSE) OR TRUE) AND (FALSE OR TRUE))");
+    // CompoundVerticalBarExpr.
+    testToSql("select true || false, 'abc' || 'xyz'",
+        "SELECT TRUE OR FALSE, concat('abc', 'xyz')");
     // InPredicate.
     testToSql("select 5 in (4, 6, 7, 5), (5 in (4, 6, 7, 5))," +
         "5 not in (4, 6, 7, 5), (5 not In (4, 6, 7, 5))",
