@@ -27,9 +27,6 @@ import org.apache.impala.authorization.ranger.RangerAuthorizationFactory;
 import org.apache.impala.authorization.ranger.RangerCatalogdAuthorizationManager;
 import org.apache.impala.authorization.ranger.RangerImpalaPlugin;
 import org.apache.impala.authorization.ranger.RangerImpalaResourceBuilder;
-import org.apache.impala.authorization.sentry.SentryAuthorizationConfig;
-import org.apache.impala.authorization.sentry.SentryAuthorizationFactory;
-import org.apache.impala.authorization.sentry.SentryPolicyService;
 import org.apache.impala.catalog.Role;
 import org.apache.impala.catalog.ScalarFunction;
 import org.apache.impala.catalog.Type;
@@ -41,7 +38,6 @@ import org.apache.impala.thrift.TColumnValue;
 import org.apache.impala.thrift.TDescribeOutputStyle;
 import org.apache.impala.thrift.TDescribeResult;
 import org.apache.impala.thrift.TFunctionBinaryType;
-import org.apache.impala.thrift.TPrincipalType;
 import org.apache.impala.thrift.TPrivilege;
 import org.apache.impala.thrift.TPrivilegeLevel;
 import org.apache.impala.thrift.TPrivilegeScope;
@@ -50,15 +46,13 @@ import org.apache.impala.thrift.TTableName;
 import org.apache.ranger.plugin.util.GrantRevokeRequest;
 import org.apache.ranger.plugin.util.RangerRESTClient;
 import org.apache.ranger.plugin.util.RangerRESTUtils;
-import org.json.simple.JSONArray;
 import org.json.simple.JSONObject;
 import org.json.simple.parser.JSONParser;
+import org.json.simple.parser.ParseException;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
-import javax.ws.rs.core.Response;
 import javax.ws.rs.core.Response.Status.Family;
-import java.io.InputStreamReader;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
@@ -112,7 +106,6 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
   protected final AuthorizationFactory authzFactory_;
   protected final AuthorizationProvider authzProvider_;
   protected final AnalysisContext authzCtx_;
-  protected final SentryPolicyService sentryService_;
   protected final ImpaladTestCatalog authzCatalog_;
   protected final Frontend authzFrontend_;
   protected final RangerImpalaPlugin rangerImpalaPlugin_;
@@ -122,24 +115,10 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
       throws ImpalaException {
     authzProvider_ = authzProvider;
     switch (authzProvider) {
-      case SENTRY:
-        user_ = new User(System.getProperty("user.name"));
-        authzConfig_ = SentryAuthorizationConfig.createHadoopGroupAuthConfig(
-            "server1",
-            System.getenv("IMPALA_HOME") + "/fe/src/test/resources/sentry-site.xml");
-        authzFactory_ = createAuthorizationFactory(authzProvider);
-        authzCtx_ = createAnalysisCtx(authzFactory_, user_.getName());
-        authzCatalog_ = new ImpaladTestCatalog(authzFactory_);
-        authzFrontend_ = new Frontend(authzFactory_, authzCatalog_);
-        sentryService_ = new SentryPolicyService(
-            ((SentryAuthorizationConfig) authzConfig_).getSentryConfig());
-        rangerImpalaPlugin_ = null;
-        rangerRestClient_ = null;
-        break;
       case RANGER:
         user_ = new User("non_owner");
         authzConfig_ = new RangerAuthorizationConfig(RANGER_SERVICE_TYPE, RANGER_APP_ID,
-            SERVER_NAME);
+            SERVER_NAME, null, null, null);
         authzFactory_ = createAuthorizationFactory(authzProvider);
         authzCtx_ = createAnalysisCtx(authzFactory_, user_.getName());
         authzCatalog_ = new ImpaladTestCatalog(authzFactory_);
@@ -148,8 +127,8 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
             ((RangerAuthorizationChecker) authzFrontend_.getAuthzChecker())
                 .getRangerImpalaPlugin();
         assertEquals("test-cluster", rangerImpalaPlugin_.getClusterName());
-        sentryService_ = null;
-        rangerRestClient_ = new RangerRESTClient(RANGER_ADMIN_URL, null);
+        rangerRestClient_ = new RangerRESTClient(RANGER_ADMIN_URL, null,
+            rangerImpalaPlugin_.getConfig());
         rangerRestClient_.setBasicAuthInfo(RANGER_USER, RANGER_PASSWORD);
         break;
       default:
@@ -160,9 +139,7 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
 
   protected AuthorizationFactory createAuthorizationFactory(
       AuthorizationProvider authzProvider) {
-    return authzProvider == AuthorizationProvider.SENTRY ?
-        new SentryAuthorizationFactory(authzConfig_) :
-        new RangerAuthorizationFactory(authzConfig_);
+    return new RangerAuthorizationFactory(authzConfig_);
   }
 
   protected interface WithPrincipal {
@@ -171,68 +148,6 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
     String getName();
   }
 
-  protected abstract class WithSentryPrincipal implements WithPrincipal {
-    protected final String role_ = "authz_test_role";
-    protected final String sentry_user_ = user_.getName();
-
-    protected void createRole(TPrivilege[]... privileges) throws ImpalaException {
-      Role role = authzCatalog_.addRole(role_);
-      authzCatalog_.addRoleGrantGroup(role_, sentry_user_);
-      for (TPrivilege[] privs: privileges) {
-        for (TPrivilege privilege: privs) {
-          privilege.setPrincipal_id(role.getId());
-          privilege.setPrincipal_type(TPrincipalType.ROLE);
-          authzCatalog_.addRolePrivilege(role_, privilege);
-        }
-      }
-    }
-
-    protected void createUser(TPrivilege[]... privileges) throws ImpalaException {
-      org.apache.impala.catalog.User user = authzCatalog_.addUser(sentry_user_);
-      for (TPrivilege[] privs: privileges) {
-        for (TPrivilege privilege: privs) {
-          privilege.setPrincipal_id(user.getId());
-          privilege.setPrincipal_type(TPrincipalType.USER);
-          authzCatalog_.addUserPrivilege(sentry_user_, privilege);
-        }
-      }
-    }
-
-    protected void dropRole() throws ImpalaException {
-      authzCatalog_.removeRole(role_);
-    }
-
-    protected void dropUser() throws ImpalaException {
-      authzCatalog_.removeUser(sentry_user_);
-    }
-  }
-
-  public class WithSentryUser extends WithSentryPrincipal {
-    @Override
-    public void init(TPrivilege[]... privileges) throws ImpalaException {
-      createUser(privileges);
-    }
-
-    @Override
-    public void cleanUp() throws ImpalaException { dropUser(); }
-
-    @Override
-    public String getName() { return sentry_user_; }
-  }
-
-  public class WithSentryRole extends WithSentryPrincipal {
-    @Override
-    public void init(TPrivilege[]... privileges) throws ImpalaException {
-      createRole(privileges);
-    }
-
-    @Override
-    public void cleanUp() throws ImpalaException { dropRole(); }
-
-    @Override
-    public String getName() { return role_; }
-  }
-
   protected abstract class WithRanger implements WithPrincipal {
     private final List<GrantRevokeRequest> grants = new ArrayList<>();
     private final RangerCatalogdAuthorizationManager authzManager =
@@ -286,7 +201,7 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
           // to whether or not we test the query with the requesting user that is the
           // owner of the resource.
           getName(),
-          Collections.emptyList(),
+          Collections.emptyList(), Collections.emptyList(),
           rangerImpalaPlugin_.getClusterName(), "127.0.0.1", privileges);
     }
   }
@@ -299,7 +214,7 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
           // We provide the name of the grantee, which is a group in this case, according
           // to whether or not we test the query with the requesting user that is the
           // owner of the resource.
-          (as_owner_ ? OWNER_GROUPS : GROUPS),
+          (as_owner_ ? OWNER_GROUPS : GROUPS), Collections.emptyList(),
           // groups,
           rangerImpalaPlugin_.getClusterName(), "127.0.0.1", privileges);
     }
@@ -374,10 +289,6 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
   protected List<WithPrincipal> buildWithPrincipals() {
     List<WithPrincipal> withPrincipals = new ArrayList<>();
     switch (authzProvider_) {
-      case SENTRY:
-        withPrincipals.add(new WithSentryRole());
-        withPrincipals.add(new WithSentryUser());
-        break;
       case RANGER:
         withPrincipals.add(new WithRangerUser());
         withPrincipals.add(new WithRangerGroup());
@@ -635,7 +546,7 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
         "Principal: %s\nStatement: %s", withPrincipal.getName(), stmt));
   }
 
-  protected void createRangerPolicy(String policyName, String json) {
+  protected long createRangerPolicy(String policyName, String json) {
     ClientResponse response = rangerRestClient_
         .getResource("/service/public/v2/api/policy")
         .accept(RangerRESTUtils.REST_MIME_TYPE_JSON)
@@ -646,7 +557,17 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
           "Unable to create a Ranger policy: %s Response: %s",
           policyName, response.getEntity(String.class)));
     }
-    LOG.info("Created ranger policy {}: {}", policyName, json);
+    String content = response.getEntity(String.class);
+    JSONParser parser = new JSONParser();
+    long policyId = -1;
+    try {
+      Object obj = parser.parse(content);
+      policyId = (Long) ((JSONObject) obj).get("id");
+    } catch (ParseException e) {
+      LOG.error("Error parsing response content: {}", content);
+    }
+    LOG.info("Created ranger policy id={}, {}: {}", policyId, policyName, json);
+    return policyId;
   }
 
   protected void deleteRangerPolicy(String policyName) {
@@ -734,8 +655,8 @@ public abstract class AuthorizationTestBase extends FrontendTestBase {
   }
 
   protected static String rowFilterError(String object) {
-    return "Impala does not support row filtering yet. Row filtering is enabled on " +
-        "table: " + object;
+    return "Row filtering is disabled by --enable_row_filtering flag. Can't access " +
+        "table " + object + " that has row filtering policy.";
   }
 
   protected ScalarFunction addFunction(String db, String fnName, List<Type> argTypes,
