@@ -132,37 +132,41 @@ public class CatalogTest {
     Db functionalDb = catalog_.getDb("functional");
     assertNotNull(functionalDb);
     assertEquals(functionalDb.getName(), "functional");
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes_view", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes_view_sub", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypessmall", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypeserror", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypeserrornonulls", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesagg", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesaggnonulls", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesnopart", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesinsert", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "complex_view", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "testtbl", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "dimtbl", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "jointbl", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "liketbl", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "greptiny", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "rankingssmall", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "uservisitssmall", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "view_view", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional", "date_tbl", "test"));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes_view", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypes_view_sub", "test",
+      null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypessmall", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypeserror", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypeserrornonulls", "test",
+     null ));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesagg", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesaggnonulls", "test",
+      null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesnopart", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "alltypesinsert", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "complex_view", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "testtbl", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "dimtbl", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "jointbl", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "liketbl", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "greptiny", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "rankingssmall", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "uservisitssmall", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "view_view", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional", "date_tbl", "test", null));
     // IMP-163 - table with string partition column does not load if there are partitions
-    assertNotNull(catalog_.getOrLoadTable("functional", "StringPartitionKey", "test"));
+    assertNotNull(
+        catalog_.getOrLoadTable("functional", "StringPartitionKey", "test", null));
     // Test non-existent table
-    assertNull(catalog_.getOrLoadTable("functional", "nonexistenttable", "test"));
+    assertNull(catalog_.getOrLoadTable("functional", "nonexistenttable", "test", null));
 
     // functional_seq contains the same tables as functional
     Db testDb = catalog_.getDb("functional_seq");
     assertNotNull(testDb);
     assertEquals(testDb.getName(), "functional_seq");
-    assertNotNull(catalog_.getOrLoadTable("functional_seq", "alltypes", "test"));
-    assertNotNull(catalog_.getOrLoadTable("functional_seq", "testtbl", "test"));
+    assertNotNull(catalog_.getOrLoadTable("functional_seq", "alltypes", "test", null));
+    assertNotNull(catalog_.getOrLoadTable("functional_seq", "testtbl", "test", null));
 
     Db hbaseDb = catalog_.getDb("functional_hbase");
     assertNotNull(hbaseDb);
@@ -170,14 +174,16 @@ public class CatalogTest {
     // Loading succeeds for an HBase table that has binary columns and an implicit key
     // column mapping
     assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "alltypessmallbinary",
-        "test"));
-    assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "alltypessmall", "test"));
+        "test", null));
+    assertNotNull(
+        catalog_.getOrLoadTable(hbaseDb.getName(), "alltypessmall", "test", null));
     assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "hbasealltypeserror",
-        "test"));
+        "test", null));
     assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(),
-        "hbasealltypeserrornonulls", "test"));
-    assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "alltypesagg", "test"));
-    assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "stringids", "test"));
+        "hbasealltypeserrornonulls", "test", null));
+    assertNotNull(
+        catalog_.getOrLoadTable(hbaseDb.getName(), "alltypesagg", "test", null));
+    assertNotNull(catalog_.getOrLoadTable(hbaseDb.getName(), "stringids", "test", null));
 
     checkTableCols(functionalDb, "alltypes", 2,
         new String[]
@@ -327,8 +333,8 @@ public class CatalogTest {
         new Type[] {Type.DATE, Type.INT, Type.DATE});
 
     // case-insensitive lookup
-    assertEquals(catalog_.getOrLoadTable("functional", "alltypes", "test"),
-        catalog_.getOrLoadTable("functional", "AllTypes", "test"));
+    assertEquals(catalog_.getOrLoadTable("functional", "alltypes", "test", null),
+        catalog_.getOrLoadTable("functional", "AllTypes", "test", null));
   }
 
   // Count of listFiles (list status + blocks) calls
@@ -357,7 +363,7 @@ public class CatalogTest {
         /*dbWasAdded=*/new Reference<Boolean>());
 
     HdfsTable table = (HdfsTable)catalog_.getOrLoadTable("functional", "AllTypes",
-        "test");
+        "test", null);
     StorageStatistics opsCounts = stats.get(DFSOpsCountStatistics.NAME);
 
     // We expect:
@@ -398,14 +404,15 @@ public class CatalogTest {
     List<TPartitionKeyValue> partitionSpec = ImmutableList.of(
         new TPartitionKeyValue("year", "2010"),
         new TPartitionKeyValue("month", "10"));
-    catalog_.reloadPartition(table, partitionSpec, new Reference<>(false), "test");
+    catalog_.reloadPartition(table, partitionSpec, new Reference<>(false),
+        CatalogObject.ThriftObjectType.NONE, "test");
     assertEquals(0L, (long)opsCounts.getLong(GET_FILE_BLOCK_LOCS));
 
     // Loading or reloading an unpartitioned table with some files in it should not make
     // an RPC per file.
     stats.reset();
     HdfsTable unpartTable = (HdfsTable)catalog_.getOrLoadTable(
-        "functional", "alltypesaggmultifilesnopart", "test");
+        "functional", "alltypesaggmultifilesnopart", "test", null);
     assertEquals(0L, (long)opsCounts.getLong(GET_FILE_BLOCK_LOCS));
     stats.reset();
     catalog_.reloadTable(unpartTable, "test");
@@ -415,11 +422,13 @@ public class CatalogTest {
     // reload path. Since we can't modify HDFS itself via these tests, we
     // do the next best thing: modify the metadata to revise history as
     // though the partition used above were actually empty.
-    HdfsPartition hdfsPartition = table
-        .getPartitionFromThriftPartitionSpec(partitionSpec);
-    hdfsPartition.setFileDescriptors(new ArrayList<>());
+    HdfsPartition.Builder partBuilder = new HdfsPartition.Builder(
+        table.getPartitionFromThriftPartitionSpec(partitionSpec));
+    partBuilder.setFileDescriptors(new ArrayList<>());
+    table.updatePartition(partBuilder);
     stats.reset();
-    catalog_.reloadPartition(table, partitionSpec, new Reference<>(false), "test");
+    catalog_.reloadPartition(table, partitionSpec, new Reference<>(false),
+        CatalogObject.ThriftObjectType.NONE, "test");
 
     // Should not scan the directory file-by-file, should use a single
     // listLocatedStatus() to get the whole directory (partition)
@@ -432,8 +441,8 @@ public class CatalogTest {
   @Test
   public void TestPartitions() throws CatalogException {
     HdfsTable table =
-        (HdfsTable) catalog_.getOrLoadTable("functional", "AllTypes", "test");
-    checkAllTypesPartitioning(table, true);
+        (HdfsTable) catalog_.getOrLoadTable("functional", "AllTypes", "test", null);
+    checkAllTypesPartitioning(table);
   }
 
   /**
@@ -442,7 +451,7 @@ public class CatalogTest {
   @Test
   public void testGetSqlConstraints() throws Exception {
     FeFsTable t = (FeFsTable) catalog_.getOrLoadTable("functional", "parent_table",
-        "test");
+        "test", null);
     assertNotNull(t);
     assertTrue(t instanceof FeFsTable);
     List<SQLPrimaryKey> primaryKeys = t.getSqlConstraints().getPrimaryKeys();
@@ -457,9 +466,9 @@ public class CatalogTest {
     assertEquals("year", primaryKeys.get(1).getColumn_name());
 
     // Force load parent_table_2. Required for fetching foreign keys from child_table.
-    catalog_.getOrLoadTable("functional", "parent_table_2", "test");
+    catalog_.getOrLoadTable("functional", "parent_table_2", "test", null);
 
-    t = (FeFsTable) catalog_.getOrLoadTable("functional", "child_table", "test");
+    t = (FeFsTable) catalog_.getOrLoadTable("functional", "child_table", "test", null);
     assertNotNull(t);
     assertTrue(t instanceof FeFsTable);
     primaryKeys = t.getSqlConstraints().getPrimaryKeys();
@@ -484,7 +493,7 @@ public class CatalogTest {
 
     // Check tables without constraints.
     t = (FeFsTable) catalog_.getOrLoadTable("functional", "alltypes",
-        "test");
+        "test", null);
     assertNotNull(t);
     assertTrue(t instanceof FeFsTable);
     primaryKeys = t.getSqlConstraints().getPrimaryKeys();
@@ -495,8 +504,7 @@ public class CatalogTest {
     assertEquals(0, foreignKeys.size());
   }
 
-  public static void checkAllTypesPartitioning(FeFsTable table,
-      boolean checkFileDescriptors) {
+  public static void checkAllTypesPartitioning(FeFsTable table) {
     assertEquals(24, table.getPartitionIds().size());
     assertEquals(24, table.getPartitions().size());
     Collection<? extends FeFsPartition> partitions =
@@ -522,13 +530,7 @@ public class CatalogTest {
       assertTrue(key2 >= 1 && key2 <= 12);
 
       months.add(key1 * 100 + key2);
-
-      if (checkFileDescriptors) {
-        // TODO(todd): once LocalCatalog supports file descriptors,
-        // no need for this boolean anymore.
-        assertEquals(p.getFileDescriptors().size(), 1);
-      }
-
+      assertEquals(p.getFileDescriptors().size(), 1);
       uniqueSds.add(p.getInputFormatDescriptor());
     }
     assertEquals(months.size(), 24);
@@ -542,7 +544,7 @@ public class CatalogTest {
   public void testStats() throws CatalogException {
     // make sure the stats for functional.alltypesagg look correct
     HdfsTable table = (HdfsTable) catalog_.getOrLoadTable("functional", "AllTypesAgg",
-        "test");
+        "test", null);
 
     Column idCol = table.getColumn("id");
     assertEquals(idCol.getStats().getAvgSerializedSize(),
@@ -616,7 +618,7 @@ public class CatalogTest {
     // First load a table that has column stats.
     //catalog_.refreshTable("functional", "alltypesagg", false);
     HdfsTable table = (HdfsTable) catalog_.getOrLoadTable("functional", "alltypesagg",
-        "test");
+        "test", null);
 
     // Now attempt to update a column's stats with mismatched stats data and ensure
     // we get the expected results.
@@ -652,6 +654,8 @@ public class CatalogTest {
   private void assertStatsUnknown(Column column) {
     assertEquals(-1, column.getStats().getNumDistinctValues());
     assertEquals(-1, column.getStats().getNumNulls());
+    assertEquals(-1, column.getStats().getNumTrues());
+    assertEquals(-1, column.getStats().getNumFalses());
     double expectedSize = column.getType().isFixedLengthType() ?
         column.getType().getSlotSize() : -1;
     assertEquals(expectedSize, column.getStats().getAvgSerializedSize(), 0.0001);
@@ -701,7 +705,7 @@ public class CatalogTest {
   @Test
   public void testPullIncrementalStats() throws CatalogException {
     // Partitioned table with stats. Load the table prior to fetching.
-    catalog_.getOrLoadTable("functional", "alltypesagg", "test");
+    catalog_.getOrLoadTable("functional", "alltypesagg", "test", null);
     expectStatistics("functional", "alltypesagg", 11);
 
     // Partitioned table with stats. Invalidate the table prior to fetching.
@@ -730,7 +734,7 @@ public class CatalogTest {
   public void testInternalHBaseTable() throws CatalogException {
     // Cast will fail if table not an HBaseTable
    HBaseTable table = (HBaseTable)
-       catalog_.getOrLoadTable("functional_hbase", "internal_hbase_table", "test");
+       catalog_.getOrLoadTable("functional_hbase", "internal_hbase_table", "test", null);
     assertNotNull("functional_hbase.internal_hbase_table was not found", table);
   }
 
@@ -742,7 +746,7 @@ public class CatalogTest {
 
   @Test
   public void testCreateTableMetadata() throws CatalogException {
-    Table table = catalog_.getOrLoadTable("functional", "alltypes", "test");
+    Table table = catalog_.getOrLoadTable("functional", "alltypes", "test", null);
     // Tables are created via Impala so the metadata should have been populated properly.
     // alltypes is an external table.
     assertEquals(System.getProperty("user.name"), table.getMetaStoreTable().getOwner());
@@ -760,13 +764,13 @@ public class CatalogTest {
   public void testCreateTableMetadataHive3() throws CatalogException {
     Assume.assumeTrue(TestUtils.getHiveMajorVersion() > 2);
     // alltypesinsert is created using CREATE TABLE LIKE and is a MANAGED table
-    Table table = catalog_.getOrLoadTable("functional", "alltypesinsert", "test");
+    Table table = catalog_.getOrLoadTable("functional", "alltypesinsert", "test", null);
     assertEquals(System.getProperty("user.name"), table.getMetaStoreTable().getOwner());
     assertEquals(TableType.EXTERNAL_TABLE.toString(),
         table.getMetaStoreTable().getTableType());
     // ACID tables should be loaded as MANAGED tables
     table = catalog_.getOrLoadTable("functional", "insert_only_transactional_table",
-        "test");
+        "test", null);
     assertEquals(System.getProperty("user.name"), table.getMetaStoreTable().getOwner());
     assertEquals(TableType.MANAGED_TABLE.toString(),
         table.getMetaStoreTable().getTableType());
@@ -782,7 +786,7 @@ public class CatalogTest {
   public void testCreateTableMetadataHive2() throws CatalogException {
     Assume.assumeTrue(TestUtils.getHiveMajorVersion() <= 2);
     // alltypesinsert is created using CREATE TABLE LIKE and is a MANAGED table
-    Table table = catalog_.getOrLoadTable("functional", "alltypesinsert", "test");
+    Table table = catalog_.getOrLoadTable("functional", "alltypesinsert", "test", null);
     assertEquals(System.getProperty("user.name"), table.getMetaStoreTable().getOwner());
     assertEquals(TableType.MANAGED_TABLE.toString(),
         table.getMetaStoreTable().getTableType());
@@ -795,7 +799,7 @@ public class CatalogTest {
     Assume.assumeTrue(
         "Skipping this test since it is only supported when running against Hive-2",
         TestUtils.getHiveMajorVersion() == 2);
-    Table table = catalog_.getOrLoadTable("functional", "hive_index_tbl", "test");
+    Table table = catalog_.getOrLoadTable("functional", "hive_index_tbl", "test", null);
     assertTrue(table instanceof IncompleteTable);
     IncompleteTable incompleteTable = (IncompleteTable) table;
     assertTrue(incompleteTable.getCause() instanceof TableLoadingException);
@@ -806,7 +810,7 @@ public class CatalogTest {
   @Test
   public void testLoadingUnsupportedTableTypes() throws CatalogException {
     // Table with unsupported SerDe library.
-    Table table = catalog_.getOrLoadTable("functional", "bad_serde", "test");
+    Table table = catalog_.getOrLoadTable("functional", "bad_serde", "test", null);
     assertTrue(table instanceof IncompleteTable);
     IncompleteTable incompleteTable = (IncompleteTable) table;
     assertTrue(incompleteTable.getCause() instanceof TableLoadingException);
@@ -816,7 +820,8 @@ public class CatalogTest {
 
     // Impala does not yet support Hive's LazyBinaryColumnarSerDe which can be
     // used for RCFILE tables.
-    table = catalog_.getOrLoadTable("functional_rc", "rcfile_lazy_binary_serde", "test");
+    table = catalog_.getOrLoadTable("functional_rc", "rcfile_lazy_binary_serde", "test"
+        , null);
     assertTrue(table instanceof IncompleteTable);
     incompleteTable = (IncompleteTable) table;
     assertTrue(incompleteTable.getCause() instanceof TableLoadingException);
@@ -951,7 +956,7 @@ public class CatalogTest {
   }
 
   @Test
-  public void testSentryCatalog() throws CatalogException {
+  public void testAuthorizationCatalog() throws CatalogException {
     AuthorizationPolicy authPolicy = catalog_.getAuthPolicy();
 
     User user = catalog_.addUser("user1");
