@@ -27,7 +27,11 @@ import org.apache.hadoop.conf.Configuration;
 import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.FileUtil;
 import org.apache.hadoop.fs.Path;
+import org.apache.hadoop.hive.common.ValidReadTxnList;
+import org.apache.hadoop.hive.common.ValidWriteIdList;
+import org.apache.hadoop.hive.metastore.api.MetaException;
 import org.apache.impala.catalog.HdfsPartition.FileDescriptor;
+import org.apache.impala.compat.MetastoreShim;
 import org.apache.impala.thrift.TNetworkAddress;
 import org.apache.impala.util.ListMap;
 import org.junit.Test;
@@ -38,7 +42,8 @@ import com.google.common.collect.ImmutableList;
 public class FileMetadataLoaderTest {
 
   @Test
-  public void testRecursiveLoading() throws IOException {
+  public void testRecursiveLoading() throws IOException, CatalogException {
+    //TODO(IMPALA-9042): Remove "throws CatalogException"
     ListMap<TNetworkAddress> hostIndex = new ListMap<>();
     Path tablePath = new Path("hdfs://localhost:20500/test-warehouse/alltypes/");
     FileMetadataLoader fml = new FileMetadataLoader(tablePath, /* recursive=*/true,
@@ -75,7 +80,8 @@ public class FileMetadataLoaderTest {
   }
 
   @Test
-  public void testHudiParquetLoading() throws IOException {
+  public void testHudiParquetLoading() throws IOException, CatalogException {
+    //TODO(IMPALA-9042): Remove "throws CatalogException"
     ListMap<TNetworkAddress> hostIndex = new ListMap<>();
     Path tablePath = new Path("hdfs://localhost:20500/test-warehouse/hudi_parquet/");
     FileMetadataLoader fml = new FileMetadataLoader(tablePath, /* recursive=*/true,
@@ -106,7 +112,59 @@ public class FileMetadataLoaderTest {
   }
 
   @Test
-  public void testLoadMissingDirectory() throws IOException {
+  public void testIcebergLoading() throws IOException, CatalogException {
+    ListMap<TNetworkAddress> hostIndex = new ListMap<>();
+    Path table1Path = new Path(
+        "hdfs://localhost:20500/test-warehouse/iceberg_test/iceberg_partitioned");
+    FileMetadataLoader fml1 = new FileMetadataLoader(table1Path, /* recursive=*/true,
+        /* oldFds = */ Collections.emptyList(), hostIndex, null, null,
+        HdfsFileFormat.ICEBERG);
+    fml1.load();
+    assertEquals(25, fml1.getStats().loadedFiles);
+    assertEquals(25, fml1.getLoadedFds().size());
+
+    ArrayList<String> relPaths = new ArrayList<>(
+        Collections2.transform(fml1.getLoadedFds(), FileDescriptor::getRelativePath));
+    Collections.sort(relPaths);
+    assertEquals("data/event_time_hour=2020-01-01-08/action=view/" +
+        "00001-1-b975a171-0911-47c2-90c8-300f23c28772-00000.parquet", relPaths.get(0));
+
+    Path table2Path = new Path(
+        "hdfs://localhost:20500/test-warehouse/iceberg_test/iceberg_non_partitioned");
+    FileMetadataLoader fml2 = new FileMetadataLoader(table2Path, /* recursive=*/true,
+        /* oldFds = */ Collections.emptyList(), hostIndex, null, null,
+        HdfsFileFormat.ICEBERG);
+    fml2.load();
+    assertEquals(25, fml2.getStats().loadedFiles);
+    assertEquals(25, fml2.getLoadedFds().size());
+
+    relPaths = new ArrayList<>(
+        Collections2.transform(fml2.getLoadedFds(), FileDescriptor::getRelativePath));
+    Collections.sort(relPaths);
+    assertEquals("data/00001-1-5dbd44ad-18bc-40f2-9dd6-aeb2cc23457c-00000.parquet",
+        relPaths.get(0));
+  }
+
+  @Test
+  public void testAcidMinorCompactionLoading() throws IOException, CatalogException {
+    //TODO(IMPALA-9042): Remove "throws CatalogException"
+    ListMap<TNetworkAddress> hostIndex = new ListMap<>();
+    ValidWriteIdList writeIds = MetastoreShim.getValidWriteIdListFromString(
+        "functional_orc_def.complextypestbl_minor_compacted:10:10::");
+    Path tablePath = new Path("hdfs://localhost:20500/test-warehouse/managed/" +
+                              "complextypestbl_minor_compacted_orc_def/");
+    FileMetadataLoader fml = new FileMetadataLoader(tablePath, /* recursive=*/true,
+        /* oldFds = */ Collections.emptyList(), hostIndex, new ValidReadTxnList(""),
+        writeIds, HdfsFileFormat.ORC);
+    fml.load();
+    // Only load the compacted file.
+    assertEquals(1, fml.getStats().loadedFiles);
+    assertEquals(8, fml.getStats().filesSupersededByAcidState);
+  }
+
+  @Test
+  public void testLoadMissingDirectory() throws IOException, CatalogException {
+    //TODO(IMPALA-9042): Remove "throws CatalogException"
     for (boolean recursive : ImmutableList.of(false, true)) {
       ListMap<TNetworkAddress> hostIndex = new ListMap<>();
       Path tablePath = new Path("hdfs://localhost:20500/test-warehouse/does-not-exist/");
@@ -117,8 +175,9 @@ public class FileMetadataLoaderTest {
     }
   }
 
+  //TODO(IMPALA-9042): Remove 'throws CatalogException'
   @Test
-  public void testSkipHiddenDirectories() throws IOException {
+  public void testSkipHiddenDirectories() throws IOException, CatalogException {
     Path sourcePath = new Path("hdfs://localhost:20500/test-warehouse/alltypes/");
     Path tmpTestPath = new Path("hdfs://localhost:20500/tmp/test-filemetadata-loader");
     Configuration conf = new Configuration();
