@@ -28,12 +28,11 @@ import org.apache.impala.catalog.FeHBaseTable;
 import org.apache.impala.catalog.HBaseColumn;
 import org.apache.impala.catalog.Type;
 import org.apache.impala.common.ImpalaException;
-import org.apache.impala.common.RuntimeEnv;
 import org.apache.impala.datagenerator.HBaseTestDataRegionAssignment;
-import org.apache.impala.service.BackendConfig;
 import org.apache.impala.service.Frontend.PlanCtx;
 import org.apache.impala.testutil.TestUtils;
 import org.apache.impala.testutil.TestUtils.IgnoreValueFilter;
+import org.apache.impala.thrift.TEnabledRuntimeFilterTypes;
 import org.apache.impala.thrift.TExecRequest;
 import org.apache.impala.thrift.TExplainLevel;
 import org.apache.impala.thrift.TJoinDistributionMode;
@@ -41,7 +40,6 @@ import org.apache.impala.thrift.TQueryCtx;
 import org.apache.impala.thrift.TQueryOptions;
 import org.apache.impala.thrift.TRuntimeFilterMode;
 import org.junit.Assert;
-import org.junit.Assume;
 import org.junit.BeforeClass;
 import org.junit.Test;
 
@@ -57,6 +55,7 @@ public class PlannerTest extends PlannerTestBase {
     PlannerTestBase.setUp();
     // Rebalance the HBase tables. This is necessary because some tests rely on HBase
     // tables being arranged in a deterministic way. See IMPALA-7061 for details.
+
     HBaseTestDataRegionAssignment assignment = new HBaseTestDataRegionAssignment();
     assignment.performAssignment("functional_hbase.alltypessmall");
     assignment.performAssignment("functional_hbase.alltypesagg");
@@ -129,7 +128,7 @@ public class PlannerTest extends PlannerTestBase {
   }
 
   @Test
-  public void testConstantPropagataion() {
+  public void testConstantPropagation() {
     runPlannerTestFile("constant-propagation");
   }
 
@@ -179,12 +178,35 @@ public class PlannerTest extends PlannerTestBase {
     runPlannerTestFile("aggregation");
   }
 
+  @Test
+  public void testGroupingSets() {
+    runPlannerTestFile("grouping-sets");
+  }
+
   @Test
   public void testAnalyticFns() {
     runPlannerTestFile("analytic-fns",
         ImmutableSet.of(PlannerTestOption.VALIDATE_CARDINALITY));
   }
 
+  /**
+   * Tests for analytic functions with higher mt_dop to exercise partitioning logic that
+   * depends on estimates of # of instances.
+   */
+  @Test
+  public void testAnalyticFnsMtDop() {
+    TQueryOptions options = defaultQueryOptions();
+    options.setMt_dop(4);
+    runPlannerTestFile("analytic-fns-mt-dop", options,
+        ImmutableSet.of(PlannerTestOption.VALIDATE_CARDINALITY));
+  }
+
+  @Test
+  public void testAnalyticRankPushdown() {
+    runPlannerTestFile("analytic-rank-pushdown");
+  }
+
+
   @Test
   public void testHbase() {
     runPlannerTestFile("hbase");
@@ -289,8 +311,6 @@ public class PlannerTest extends PlannerTestBase {
   public void testInsertSortByZorder() {
     // Add a test table with a SORT BY ZORDER clause to test that the corresponding sort
     // nodes are added by the insert statements in insert-sort-by.test.
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(true);
-
     addTestDb("test_sort_by_zorder", "Test DB for SORT BY ZORDER clause.");
     addTestTable("create table test_sort_by_zorder.t (id int, int_col int, " +
         "bool_col boolean) partitioned by (year int, month int) " +
@@ -298,8 +318,19 @@ public class PlannerTest extends PlannerTestBase {
     addTestTable("create table test_sort_by_zorder.t_nopart (id int, int_col int, " +
         "bool_col boolean) sort by zorder (int_col, bool_col) location '/'");
     runPlannerTestFile("insert-sort-by-zorder", "test_sort_by_zorder");
+  }
 
-    BackendConfig.INSTANCE.setZOrderSortUnlocked(false);
+  @Test
+  public void testHdfsInsertWriterLimit() {
+    addTestDb("test_hdfs_insert_writer_limit",
+        "Test DB for MAX_FS_WRITERS query option.");
+    addTestTable( "create table test_hdfs_insert_writer_limit.partitioned_table "
+        + "(id int) partitioned by (year int, month int) location '/'");
+    addTestTable("create table test_hdfs_insert_writer_limit.unpartitioned_table"
+        + " (id int) location '/'");
+    runPlannerTestFile("insert-hdfs-writer-limit", "test_hdfs_insert_writer_limit",
+        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
@@ -357,8 +388,9 @@ public class PlannerTest extends PlannerTestBase {
     // The FK/PK detection result is included in EXTENDED or higher.
     TQueryOptions options = defaultQueryOptions();
     options.setDisable_hdfs_num_rows_estimate(false);
-    runPlannerTestFile("fk-pk-join-detection-hdfs-num-rows-est-enabled",
-        options, ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN));
+    runPlannerTestFile("fk-pk-join-detection-hdfs-num-rows-est-enabled", options,
+        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
@@ -366,6 +398,7 @@ public class PlannerTest extends PlannerTestBase {
     // The FK/PK detection result is included in EXTENDED or higher.
     TQueryOptions options = defaultQueryOptions();
     options.setDisable_hdfs_num_rows_estimate(true);
+    options.setMinmax_filter_threshold(0.0);
     runPlannerTestFile("fk-pk-join-detection",
         options, ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
             PlannerTestOption.VALIDATE_CARDINALITY));
@@ -419,11 +452,27 @@ public class PlannerTest extends PlannerTestBase {
     runPlannerTestFile("subquery-rewrite", options);
   }
 
+  /**
+   * Tests for the IMPALA-1270 optimization of automatically adding a distinct
+   * agg to semi joins.
+   */
+  @Test
+  public void testSemiJoinDistinct() {
+    runPlannerTestFile("semi-join-distinct");
+  }
+
   @Test
   public void testUnion() {
     runPlannerTestFile("union");
   }
 
+  @Test
+  public void testSetOperationRewrite() {
+    TQueryOptions options = defaultQueryOptions();
+    options.setMinmax_filter_threshold(0.0);
+    runPlannerTestFile("setoperation-rewrite", options);
+  }
+
   @Test
   public void testValues() {
     runPlannerTestFile("values");
@@ -453,7 +502,19 @@ public class PlannerTest extends PlannerTestBase {
   public void testPartitionKeyScans() {
     TQueryOptions options = new TQueryOptions();
     options.setOptimize_partition_key_scans(true);
-    runPlannerTestFile("partition-key-scans", options);
+    runPlannerTestFile("partition-key-scans", options, ImmutableSet.of(
+            PlannerTestOption.VALIDATE_CARDINALITY));
+  }
+
+  /**
+   * Test partition key scans with the option disabled - backend support is
+   * used to only return one row per file.
+   */
+  @Test
+  public void testPartitionKeyScansDefault() {
+    TQueryOptions options = new TQueryOptions();
+    runPlannerTestFile("partition-key-scans-default", options, ImmutableSet.of(
+            PlannerTestOption.VALIDATE_CARDINALITY));
   }
 
   @Test
@@ -489,22 +550,13 @@ public class PlannerTest extends PlannerTestBase {
 
   @Test
   public void testTpchNested() {
+    TQueryOptions options = new TQueryOptions();
     runPlannerTestFile("tpch-nested", "tpch_nested_parquet",
         ImmutableSet.of(PlannerTestOption.INCLUDE_RESOURCE_HEADER,
             PlannerTestOption.VALIDATE_RESOURCES,
             PlannerTestOption.VALIDATE_CARDINALITY));
   }
 
-  @Test
-  public void testTpcds() {
-    // Uses ss_sold_date_sk as the partition key of store_sales to allow static partition
-    // pruning. The original predicates were rephrased in terms of the ss_sold_date_sk
-    // partition key, with the query semantics identical to the original queries.
-    runPlannerTestFile("tpcds-all", "tpcds",
-        ImmutableSet.of(PlannerTestOption.INCLUDE_RESOURCE_HEADER,
-            PlannerTestOption.VALIDATE_RESOURCES));
-  }
-
   @Test
   public void testSmallQueryOptimization() {
     TQueryOptions options = new TQueryOptions();
@@ -549,11 +601,29 @@ public class PlannerTest extends PlannerTestBase {
     runPlannerTestFile("runtime-filter-propagation", options);
   }
 
+  @Test
+  public void testDisableRuntimeOverlapFilter() {
+    TQueryOptions options = new TQueryOptions();
+    options.setMinmax_filter_threshold(0.0);
+    runPlannerTestFile("disable-runtime-overlap-filter", options);
+
+    options.setMinmax_filter_threshold(1.0);
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.BLOOM);
+    runPlannerTestFile("disable-runtime-overlap-filter", options);
+  }
+
   @Test
   public void testRuntimeFilterQueryOptions() {
     runPlannerTestFile("runtime-filter-query-options");
   }
 
+  @Test
+  public void testBloomFilterAssignment() {
+    runPlannerTestFile("bloom-filter-assignment",
+        ImmutableSet.of(
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
+  }
+
   @Test
   public void testConjunctOrdering() {
     runPlannerTestFile("conjunct-ordering");
@@ -567,7 +637,8 @@ public class PlannerTest extends PlannerTestBase {
   @Test
   public void testParquetFiltering() {
     runPlannerTestFile("parquet-filtering",
-        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN));
+        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
@@ -576,39 +647,39 @@ public class PlannerTest extends PlannerTestBase {
     options.setParquet_dictionary_filtering(false);
     options.setParquet_read_statistics(false);
     runPlannerTestFile("parquet-filtering-disabled", options,
-        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN));
+        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
   public void testKudu() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
+    TQueryOptions options = defaultQueryOptions();
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.ALL);
     addTestDb("kudu_planner_test", "Test DB for Kudu Planner.");
     addTestTable("CREATE EXTERNAL TABLE kudu_planner_test.no_stats STORED AS KUDU " +
         "TBLPROPERTIES ('kudu.table_name' = 'impala::functional_kudu.alltypes');");
-    runPlannerTestFile("kudu");
+    runPlannerTestFile("kudu", options);
   }
 
   @Test
   public void testKuduUpsert() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
     runPlannerTestFile("kudu-upsert");
   }
 
   @Test
   public void testKuduUpdate() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
-    runPlannerTestFile("kudu-update");
+    TQueryOptions options = defaultQueryOptions();
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.ALL);
+    runPlannerTestFile("kudu-update", options);
   }
 
   @Test
   public void testKuduDelete() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
     runPlannerTestFile("kudu-delete");
   }
 
   @Test
   public void testKuduSelectivity() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
     TQueryOptions options = defaultQueryOptions();
     options.setExplain_level(TExplainLevel.VERBOSE);
     runPlannerTestFile("kudu-selectivity", options);
@@ -616,10 +687,11 @@ public class PlannerTest extends PlannerTestBase {
 
   @Test
   public void testKuduTpch() {
-    Assume.assumeTrue(RuntimeEnv.INSTANCE.isKuduSupported());
-    runPlannerTestFile("tpch-kudu", ImmutableSet.of(
-        PlannerTestOption.INCLUDE_RESOURCE_HEADER,
-        PlannerTestOption.VALIDATE_RESOURCES));
+    TQueryOptions options = defaultQueryOptions();
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.ALL);
+    runPlannerTestFile("tpch-kudu", options,
+        ImmutableSet.of(PlannerTestOption.INCLUDE_RESOURCE_HEADER,
+            PlannerTestOption.VALIDATE_RESOURCES));
   }
 
   @Test
@@ -629,38 +701,14 @@ public class PlannerTest extends PlannerTestBase {
 
   @Test
   public void testMtDopValidation() {
-    // Tests that queries supported with mt_dop > 0 produce a parallel plan, or
-    // throw a NotImplementedException otherwise (e.g. plan has a distributed join).
-    TQueryOptions options = defaultQueryOptions();
-    options.setMt_dop(3);
-    options.setDisable_hdfs_num_rows_estimate(false);
-    options.setExplain_level(TExplainLevel.EXTENDED);
-    try {
-      // Temporarily unset the test env such that unsupported queries with mt_dop > 0
-      // throw an exception. Those are otherwise allowed for testing parallel plans.
-      RuntimeEnv.INSTANCE.setEnableMtDopValidation(true);
-      runPlannerTestFile("mt-dop-validation-hdfs-num-rows-est-enabled", options);
-    } finally {
-      RuntimeEnv.INSTANCE.setEnableMtDopValidation(false);
-    }
-  }
-
-  @Test
-  public void testMtDopValidationWithHDFSNumRowsEstDisabled() {
-    // Tests that queries supported with mt_dop > 0 produce a parallel plan, or
-    // throw a NotImplementedException otherwise (e.g. plan has a distributed join).
+    // Tests that queries planned with mt_dop > 0 produce a parallel plan.
+    // Since IMPALA-9812 was fixed all plans are supported. Previously some plans
+    // were rejected.
     TQueryOptions options = defaultQueryOptions();
     options.setMt_dop(3);
     options.setDisable_hdfs_num_rows_estimate(true);
     options.setExplain_level(TExplainLevel.EXTENDED);
-    try {
-      // Temporarily unset the test env such that unsupported queries with mt_dop > 0
-      // throw an exception. Those are otherwise allowed for testing parallel plans.
-      RuntimeEnv.INSTANCE.setEnableMtDopValidation(true);
-      runPlannerTestFile("mt-dop-validation", options);
-    } finally {
-      RuntimeEnv.INSTANCE.setEnableMtDopValidation(false);
-    }
+    runPlannerTestFile("mt-dop-validation", options);
   }
 
   @Test
@@ -713,6 +761,8 @@ public class PlannerTest extends PlannerTestBase {
     // Tests the resource requirement computation from the planner.
     TQueryOptions options = defaultQueryOptions();
     options.setNum_scanner_threads(1); // Required so that output doesn't vary by machine
+    options.setMinmax_filter_threshold(0.0);
+    // Required so that output doesn't vary by whether parquet tables are used or not.
     options.setDisable_hdfs_num_rows_estimate(true);
     runPlannerTestFile("resource-requirements", options,
         ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
@@ -727,6 +777,7 @@ public class PlannerTest extends PlannerTestBase {
     TQueryOptions options = defaultQueryOptions();
     options.setExplain_level(TExplainLevel.EXTENDED);
     options.setNum_scanner_threads(1); // Required so that output doesn't vary by machine
+    options.setMinmax_filter_threshold(0.0);
     options.setDisable_hdfs_num_rows_estimate(true);
     runPlannerTestFile("spillable-buffer-sizing", options,
         ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
@@ -741,6 +792,8 @@ public class PlannerTest extends PlannerTestBase {
     TQueryOptions options = defaultQueryOptions();
     options.setExplain_level(TExplainLevel.EXTENDED);
     options.setNum_scanner_threads(1); // Required so that output doesn't vary by machine
+    // Required so that output doesn't vary by the format of the table used.
+    options.setMinmax_filter_threshold(0.0);
     options.setMax_row_size(8L * 1024L * 1024L);
     runPlannerTestFile("max-row-size", options,
         ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
@@ -751,7 +804,11 @@ public class PlannerTest extends PlannerTestBase {
   @Test
   public void testSortExprMaterialization() {
     addTestFunction("TestFn", Lists.newArrayList(Type.DOUBLE), false);
-    runPlannerTestFile("sort-expr-materialization",
+    // Avoid conversion of RANK()/ROW_NUMBER() predicates to Top-N limits, which
+    // would interfere with the purpose of this test.
+    TQueryOptions options = defaultQueryOptions();
+    options.setAnalytic_rank_pushdown_threshold(0);
+    runPlannerTestFile("sort-expr-materialization", options,
         ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN));
   }
 
@@ -759,7 +816,8 @@ public class PlannerTest extends PlannerTestBase {
   public void testTableSample() {
     TQueryOptions options = defaultQueryOptions();
     runPlannerTestFile("tablesample", options,
-        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN));
+        ImmutableSet.of(PlannerTestOption.EXTENDED_EXPLAIN,
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
@@ -819,7 +877,10 @@ public class PlannerTest extends PlannerTestBase {
     TQueryOptions options = defaultQueryOptions();
     options.setExplain_level(TExplainLevel.EXTENDED);
     options.setDisable_hdfs_num_rows_estimate(false);
-    runPlannerTestFile("min-max-runtime-filters-hdfs-num-rows-est-enabled", options);
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.MIN_MAX);
+    runPlannerTestFile("min-max-runtime-filters-hdfs-num-rows-est-enabled", options,
+        ImmutableSet.of(
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
   }
 
   @Test
@@ -827,6 +888,7 @@ public class PlannerTest extends PlannerTestBase {
     TQueryOptions options = defaultQueryOptions();
     options.setExplain_level(TExplainLevel.EXTENDED);
     options.setDisable_hdfs_num_rows_estimate(true);
+    options.setEnabled_runtime_filter_types(TEnabledRuntimeFilterTypes.MIN_MAX);
     runPlannerTestFile("min-max-runtime-filters", options);
   }
 
@@ -972,6 +1034,9 @@ public class PlannerTest extends PlannerTestBase {
     filter = TestUtils.ROW_SIZE_FILTER;
     assertEquals(" row-size= cardinality=10.3K",
         filter.transform(" row-size=10B cardinality=10.3K"));
+    filter = TestUtils.PARTITIONS_FILTER;
+    assertEquals(" partitions: 0/24 rows=",
+        filter.transform(" partitions: 0/24 rows=10.3K"));
   }
 
   @Test
@@ -1018,4 +1083,85 @@ public class PlannerTest extends PlannerTestBase {
     runPlannerTestFile("broadcast-bytes-limit-large", "tpch_parquet", options);
   }
 
+  /**
+   * Check that planner estimates reflect the preagg bytes limit.
+   */
+  @Test
+  public void testPreaggBytesLimit() {
+    TQueryOptions options = defaultQueryOptions();
+    options.setPreagg_bytes_limit(64 * 1024 * 1024);
+    options.setExplain_level(TExplainLevel.EXTENDED);
+    runPlannerTestFile("preagg-bytes-limit", "tpch_parquet", options);
+  }
+
+  /**
+   * Check conversion of predicates to conjunctive normal form.
+   */
+  @Test
+  public void testConvertToCNF() {
+    TQueryOptions options = defaultQueryOptions();
+    options.setMinmax_filter_threshold(0.0);
+    runPlannerTestFile("convert-to-cnf", "tpch_parquet", options);
+  }
+
+  /**
+   * Check that ACID table scans work as expected.
+   */
+  @Test
+  public void testAcidTableScans() {
+    runPlannerTestFile("acid-scans", "functional_orc_def",
+        ImmutableSet.of(
+            PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS));
+  }
+
+  /**
+   * Test limit pushdown into analytic sort in isolation.
+   */
+  @Test
+  public void testLimitPushdownAnalytic() {
+    // The partitioned top-n optimization interacts with limit pushdown. We run the
+    // basic limit pushdown tests with it disabled.
+    TQueryOptions options = defaultQueryOptions();
+    options.setAnalytic_rank_pushdown_threshold(0);
+    runPlannerTestFile("limit-pushdown-analytic", options);
+  }
+
+  /**
+   * Test limit pushdown into analytic sort with the partitioned top-n transformation
+   * also enabled.
+   */
+  @Test
+  public void testLimitPushdownPartitionedTopN() {
+    runPlannerTestFile("limit-pushdown-partitioned-top-n");
+  }
+
+  /**
+   * Test outer join simplification.
+   */
+  @Test
+  public void testSimplifyOuterJoins() {
+    TQueryOptions options = new TQueryOptions();
+    options.setEnable_outer_join_to_inner_transformation(true);
+    runPlannerTestFile("outer-to-inner-joins", options,
+        ImmutableSet.of(PlannerTestOption.VALIDATE_CARDINALITY));
+  }
+
+  /**
+   * Test simple limit optimization
+   */
+  @Test
+  public void testSimpleLimitOptimization() {
+    TQueryOptions options = new TQueryOptions();
+    options.setOptimize_simple_limit(true);
+    runPlannerTestFile("optimize-simple-limit", options);
+  }
+
+  /**
+   * Test the distribution method for a join
+   */
+  @Test
+  public void testDistributionMethod() {
+    runPlannerTestFile("tpcds-dist-method", "tpcds");
+  }
+
 }
