@@ -47,6 +47,7 @@ import org.apache.impala.testutil.TestFileParser.TestCase;
 import org.apache.impala.testutil.TestUtils;
 import org.apache.impala.testutil.TestUtils.ResultFilter;
 import org.apache.impala.thrift.ImpalaInternalServiceConstants;
+import org.apache.impala.thrift.QueryConstants;
 import org.apache.impala.thrift.TDescriptorTable;
 import org.apache.impala.thrift.TExecRequest;
 import org.apache.impala.thrift.TExplainLevel;
@@ -89,7 +90,7 @@ public class PlannerTestBase extends FrontendTestBase {
   private final static boolean GENERATE_OUTPUT_FILE = true;
   private final java.nio.file.Path testDir_ = Paths.get("functional-planner", "queries",
       "PlannerTest");
-  private static java.nio.file.Path outDir_;
+  protected static java.nio.file.Path outDir_;
   private static KuduClient kuduClient_;
 
   // Map from plan ID (TPlanNodeId) to the plan node with that ID.
@@ -108,9 +109,7 @@ public class PlannerTestBase extends FrontendTestBase {
     updateReq.setNum_executors(3);
     ExecutorMembershipSnapshot.update(updateReq);
 
-    if (RuntimeEnv.INSTANCE.isKuduSupported()) {
-      kuduClient_ = new KuduClient.KuduClientBuilder("127.0.0.1:7051").build();
-    }
+    kuduClient_ = new KuduClient.KuduClientBuilder("127.0.0.1:7051").build();
     String logDir = System.getenv("IMPALA_FE_TEST_LOGS_DIR");
     if (logDir == null) logDir = "/tmp";
     outDir_ = Paths.get(logDir, "PlannerTest");
@@ -494,7 +493,7 @@ public class PlannerTestBase extends FrontendTestBase {
     } else {
       // for distributed and parallel execution we want to run on all available nodes
       queryOptions.setNum_nodes(
-          ImpalaInternalServiceConstants.NUM_NODES_ALL);
+          QueryConstants.NUM_NODES_ALL);
     }
     if (section == Section.PARALLELPLANS
         && (!queryOptions.isSetMt_dop() || queryOptions.getMt_dop() == 0)) {
@@ -542,6 +541,10 @@ public class PlannerTestBase extends FrontendTestBase {
       if (!testOptions.contains(PlannerTestOption.VALIDATE_SCAN_FS)) {
         resultFilters.add(TestUtils.SCAN_NODE_SCHEME_FILTER);
       }
+      if (testOptions.contains(
+              PlannerTestOption.DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS)) {
+        resultFilters.add(TestUtils.PARTITIONS_FILTER);
+      }
 
       String planDiff = TestUtils.compareOutput(
           Lists.newArrayList(explainStr.split("\n")), expectedPlan, true, resultFilters);
@@ -832,7 +835,10 @@ public class PlannerTestBase extends FrontendTestBase {
     VALIDATE_SCAN_FS,
     // If set, disables the attempt to compute an estimated number of rows in an
     // hdfs table.
-    DISABLE_HDFS_NUM_ROWS_ESTIMATE
+    DISABLE_HDFS_NUM_ROWS_ESTIMATE,
+    // If set, make no attempt to validate the estimated number of rows for any
+    // partitions in an hdfs table.
+    DO_NOT_VALIDATE_ROWCOUNT_ESTIMATION_FOR_PARTITIONS
   }
 
   protected void runPlannerTestFile(String testFile, TQueryOptions options) {
@@ -861,7 +867,7 @@ public class PlannerTestBase extends FrontendTestBase {
         Collections.<PlannerTestOption>emptySet());
   }
 
-  private void runPlannerTestFile(String testFile, String dbName, TQueryOptions options,
+  protected void runPlannerTestFile(String testFile, String dbName, TQueryOptions options,
         Set<PlannerTestOption> testOptions) {
     String fileName = testDir_.resolve(testFile + ".test").toString();
     if (options == null) {
