@@ -45,7 +45,6 @@ CONFIG.update({
 
   # Location of Hive per-query log files of the form: hive_job_log_<hive_query_id>.txt
   'hive.querylog.location': '${IMPALA_CLUSTER_LOGS_DIR}/hive',
-  'hive.sentry.conf.url': 'file:///${IMPALA_HOME}/fe/src/test/resources/sentry-site.xml',
 
   # Change back to NOSASL when HIVE-4232 is fixed.
   # With NONE, Hive uses the plain SASL transport.
@@ -57,6 +56,12 @@ CONFIG.update({
   # TODO(todd): should we be enabling stats autogather?
   'hive.stats.autogather': 'false',
   'hive.support.concurrency': 'true',
+  # There are tests which issue Hive's replication command. The repl
+  # dump command will wait in case there are any open transactions.
+  # The default value of this configuration is 1h which will block
+  # the test for long time. Overriding this configuration helps with
+  # the runtime of the test.
+  'hive.repl.bootstrap.dump.open.txn.timeout': '120s'
 })
 
 if variant == 'changed_external_dir':
@@ -137,11 +142,14 @@ if hive_major_version >= 3:
    # are built around populating /test-warehouse, so use /test-warehouse/managed
    # to allow that logic to remain the same.
    'hive.metastore.warehouse.dir': '${WAREHOUSE_LOCATION_PREFIX}/test-warehouse/managed',
-   'hive.metastore.warehouse.external.dir': '${WAREHOUSE_LOCATION_PREFIX}/test-warehouse'
+   'hive.metastore.warehouse.external.dir': '${WAREHOUSE_LOCATION_PREFIX}/test-warehouse',
+
+   # Due to HIVE-23102 Hive will wait for at least this amount of time for compactions
+   # (the default value is 5 mins which is way too long). Setting it to 2 seconds.
+   'hive.compactor.wait.timeout': '2000'
   })
 else:
   CONFIG.update({
-   'hive.metastore.event.listeners': 'org.apache.sentry.binding.metastore.SentrySyncHMSNotificationsPostEventListener',
    # HMS-2 based environments have a different set of expected configurations for event processor
    'hive.metastore.alter.notifications.basic': 'false',
    'hive.metastore.notification.parameters.exclude.patterns': '',
@@ -152,11 +160,10 @@ else:
   })
 
 # Notifications-related configuration.
-# These are for enabling notification between Hive and Sentry as well as
-# metastore event processing in Impala (see IMPALA-7954)
+# These are for enabling notification for Hive as well as metastore event processing
+# in Impala (see IMPALA-7954)
 CONFIG.update({
  'hive.metastore.transactional.event.listeners': 'org.apache.hive.hcatalog.listener.DbNotificationListener,org.apache.kudu.hive.metastore.KuduMetastorePlugin',
- 'hcatalog.message.factory.impl.json': 'org.apache.sentry.binding.metastore.messaging.json.SentryJSONMessageFactory',
  'hive.metastore.dml.events': 'true',
 })
 
