@@ -1,4 +1,5 @@
 #!/usr/bin/env python
+# -*- coding: utf-8 -*-
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -16,7 +17,7 @@
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.
-from __future__ import print_function
+from __future__ import print_function, unicode_literals
 from compatibility import _xrange as xrange
 
 from bitarray import bitarray
@@ -47,7 +48,7 @@ from thrift.transport.TSocket import TSocket
 from thrift.transport.TTransport import TBufferedTransport, TTransportException
 from thrift.Thrift import TApplicationException, TException
 from shell_exceptions import (RPCException, QueryStateException, DisconnectedException,
-    QueryCancelledByShellException, MissingThriftMethodException)
+    QueryCancelledByShellException, MissingThriftMethodException, HttpError)
 
 
 # Helpers to extract and convert HS2's representation of values to the display version.
@@ -80,10 +81,29 @@ HS2_VALUE_CONVERTERS = {
     TTypeId.DATE_TYPE: (operator.attrgetter('stringVal'), None)
 }
 
+
+# Helper to decode utf8 encoded str to unicode type in Python 2. NOOP in Python 3.
+def utf8_decode_if_needed(val):
+  if sys.version_info.major < 3 and isinstance(val, str):
+    val = val.decode('utf-8', errors='replace')
+  return val
+
+
+# Helper to decode unicode to utf8 encoded str in Python 2. NOOP in Python 3.
+def utf8_encode_if_needed(val):
+  if sys.version_info.major < 3 and isinstance(val, unicode):
+    val = val.encode('utf-8', errors='replace')
+  return val
+
 # Regular expression that matches the progress line added to HS2 logs by
 # the Impala server.
 HS2_LOG_PROGRESS_REGEX = re.compile("Query.*Complete \([0-9]* out of [0-9]*\)\n")
 
+# Exception types to differentiate between the different RPCExceptions.
+# RPCException raised when TApplicationException is caught.
+RPC_EXCEPTION_TAPPLICATION = "TAPPLICATION_EXCEPTION"
+# RPCException raised when impala server sends a TStatusCode.ERROR_STATUS status code.
+RPC_EXCEPTION_SERVER = "SERVER_ERROR"
 
 class QueryOptionLevels:
   """These are the levels used when displaying query options.
@@ -108,12 +128,13 @@ class ImpalaClient(object):
   """Base class for shared functionality between HS2 and Beeswax. Includes stub methods
   for methods that are expected to be implemented in the subclasses.
   TODO: when beeswax support is removed, merge this with ImpalaHS2Client."""
-  def __init__(self, impalad, kerberos_host_fqdn, use_kerberos=False,
+  def __init__(self, impalad, fetch_size, kerberos_host_fqdn, use_kerberos=False,
                kerberos_service_name="impala", use_ssl=False, ca_cert=None, user=None,
                ldap_password=None, use_ldap=False, client_connect_timeout_ms=60000,
-               verbose=True, use_http_base_transport=False, http_path=None):
+               verbose=True, use_http_base_transport=False, http_path=None,
+               auth_cookie_names=None):
     self.connected = False
-    self.impalad_host = impalad[0].encode('ascii', 'ignore')
+    self.impalad_host = impalad[0]
     self.impalad_port = int(impalad[1])
     self.kerberos_host_fqdn = kerberos_host_fqdn
     self.imp_service = None
@@ -127,14 +148,18 @@ class ImpalaClient(object):
     self.client_connect_timeout_ms = int(client_connect_timeout_ms)
     self.default_query_options = {}
     self.query_option_levels = {}
-    self.fetch_batch_size = 1024
+    self.fetch_size = fetch_size
     self.use_http_base_transport = use_http_base_transport
     self.http_path = http_path
+    self.auth_cookie_names = auth_cookie_names
     # This is set from ImpalaShell's signal handler when a query is cancelled
     # from command line via CTRL+C. It is used to suppress error messages of
     # query cancellation.
     self.is_query_cancelled = False
     self.verbose = verbose
+    # This is set in connect(). It's used in constructing the retried query link after
+    # we parse the retried query id.
+    self.webserver_address = None
 
   def connect(self):
     """Creates a connection to an Impalad instance. Returns a tuple with the impala
@@ -197,7 +222,7 @@ class ImpalaClient(object):
   def _close_transport(self):
     """Closes transport if not closed and set self.connected to False. This is the last
     step of close_connection()."""
-    if self.transport and not self.transport.isOpen():
+    if self.transport and self.transport.isOpen():
       self.transport.close()
     self.connected = False
 
@@ -221,6 +246,10 @@ class ImpalaClient(object):
     'd74d8ce632c9d4d0:75c5a51100000000'"""
     raise NotImplementedError()
 
+  def get_query_link(self, query_id):
+    """Return the URL link to the debug page of the query"""
+    return "%s/query_plan?query_id=%s" % (self.webserver_address, query_id)
+
   def wait_to_finish(self, last_query_handle, periodic_callback=None):
     """Wait until the results can be fetched for 'last_query_handle' or until the
     query encounters an error or is cancelled. Raises an exception if the query
@@ -258,25 +287,10 @@ class ImpalaClient(object):
   def fetch(self, query_handle):
     """Returns an iterable of batches of result rows. Each batch is an iterable of rows.
     Each row is an iterable of strings in the format in which they should be displayed
-    Tries to ensure that the batches have a granularity of self.fetch_batch_size but
+    Tries to ensure that the batches have a granularity of self.fetch_size but
     does not guarantee it.
     """
-    result_rows = None
-    for rows in self._fetch_one_batch(query_handle):
-      if result_rows:
-        result_rows.extend(rows)
-      else:
-        result_rows = rows
-      if len(result_rows) > self.fetch_batch_size:
-        yield result_rows
-        result_rows = None
-
-    # Return the final batch of rows.
-    if result_rows:
-      yield result_rows
-
-  def _fetch_one_batch(self, query_handle):
-    """Returns an iterable of batches of result rows up to self.fetch_batch_size. Does
+    """Returns an iterable of batches of result rows up to self.fetch_size. Does
     not need to consolidate those batches into larger batches."""
     raise NotImplementedError()
 
@@ -304,12 +318,20 @@ class ImpalaClient(object):
 
   def get_runtime_profile(self, last_query_handle):
     """Get the runtime profile string from the server. Returns None if
-    an error was encountered."""
+    an error was encountered. If the query was retried, returns the profile of the failed
+    attempt as well; the tuple (profile, failed_profile) is returned where 'profile' is
+    the profile of the most recent query attempt and 'failed_profile' is the profile of
+    the original query attempt that failed. Currently, only the HS2 protocol supports
+    returning the failed profile."""
     raise NotImplementedError()
 
   def get_summary(self, last_query_handle):
     """Get the thrift TExecSummary from the server. Returns None if
-    an error was encountered."""
+    an error was encountered. If the query was retried, returns TExecSummary of the failed
+    attempt as well; the tuple (summary, failed_summary) is returned where 'summary' is
+    the TExecSummary of the most recent query attempt and 'failed_summary' is the
+    TExecSummary of the original query attempt that failed. Currently, only the HS2
+    protocol supports returning the failed summary"""
     raise NotImplementedError()
 
   def _get_warn_or_error_log(self, last_query_handle, warn):
@@ -330,6 +352,17 @@ class ImpalaClient(object):
     (e.g. warnings)."""
     return self._get_warn_or_error_log(last_query_handle, False)
 
+  def _append_retried_query_link(self, get_log_result):
+    """Append the retried query link if the original query has been retried"""
+    if self.webserver_address:
+      query_id_search = re.search("Query has been retried using query id: (.*)\n",
+                                  get_log_result)
+      if query_id_search and len(query_id_search.groups()) >= 1:
+        retried_query_id = query_id_search.group(1)
+        get_log_result += "Retried query link: %s" % \
+                          self.get_query_link(retried_query_id)
+    return get_log_result
+
   def _get_http_transport(self, connect_timeout_ms):
     """Creates a transport with HTTP as the base."""
     # Older python versions do not support SSLContext needed by ImpalaHttpClient. More
@@ -343,7 +376,7 @@ class ImpalaClient(object):
     # connect timeout does not achieve the desirable result as the subsequent open() could
     # block similary in case of problematic remote end points.
     # TODO: Investigate connection reuse in ImpalaHttpClient and revisit this.
-    if connect_timeout_ms > 0:
+    if connect_timeout_ms > 0 and self.verbose:
       print("Warning: --connect_timeout_ms is currently ignored with HTTP transport.",
             file=sys.stderr)
 
@@ -366,17 +399,22 @@ class ImpalaClient(object):
       else:
         ssl_ctx.check_hostname = False  # Mandated by the SSL lib for CERT_NONE mode.
         ssl_ctx.verify_mode = ssl.CERT_NONE
-      transport = ImpalaHttpClient(
-          "https://{0}/{1}".format(host_and_port, self.http_path), ssl_context=ssl_ctx)
+      url = "https://{0}/{1}".format(host_and_port, self.http_path)
+      transport = ImpalaHttpClient(url, ssl_context=ssl_ctx,
+                                   auth_cookie_names=self.auth_cookie_names)
     else:
-      transport = ImpalaHttpClient("http://{0}/{1}".format(host_and_port, self.http_path))
+      url = "http://{0}/{1}".format(host_and_port, self.http_path)
+      transport = ImpalaHttpClient(url, auth_cookie_names=self.auth_cookie_names)
 
     if self.use_ldap:
       # Set the BASIC auth header
-      auth = base64.encodestring(
-          "{0}:{1}".format(self.user, self.ldap_password)).strip('\n')
+      user_passwd = "{0}:{1}".format(self.user, self.ldap_password)
+      auth = base64.encodestring(user_passwd.encode()).decode().strip('\n')
       transport.setCustomHeaders({"Authorization": "Basic {0}".format(auth)})
 
+    # Without buffering Thrift would call socket.recv() each time it deserializes
+    # something (e.g. a member in a struct).
+    transport = TBufferedTransport(transport)
     transport.open()
     return transport
 
@@ -600,14 +638,33 @@ class ImpalaHS2Client(ImpalaClient):
     # If connected, this is the handle returned by the OpenSession RPC that needs
     # to be passed into most HS2 RPCs.
     self.session_handle = None
+    # Enable retries only for hs2-http protocol.
+    if self.use_http_base_transport:
+      # Maximum number of tries for idempotent rpcs.
+      self.max_tries = 4
+    else:
+      self.max_tries = 1
+    # Minimum sleep interval between retry attempts.
+    self.min_sleep_interval = 1
 
   def _get_thrift_client(self, protocol):
     return ImpalaHiveServer2Service.Client(protocol)
 
+  def _get_sleep_interval_for_retries(self, num_tries):
+    """Returns the sleep interval in seconds for the 'num_tries' retry attempt."""
+    assert num_tries > 0 and num_tries < self.max_tries
+    return self.min_sleep_interval * (num_tries - 1)
+
   def _open_session(self):
     open_session_req = TOpenSessionReq(TProtocolVersion.HIVE_CLI_SERVICE_PROTOCOL_V6,
         username=self.user)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.OpenSession(open_session_req))
+
+    def OpenSession():
+      return self.imp_service.OpenSession(open_session_req)
+    # OpenSession rpcs are idempotent and so ok to retry. If the client gets disconnected
+    # and the server successfully opened a session, the client will retry and rely on
+    # server to clean up the session.
+    resp = self._do_hs2_rpc(OpenSession, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
     assert (resp.serverProtocolVersion ==
             TProtocolVersion.HIVE_CLI_SERVICE_PROTOCOL_V6), resp.serverProtocolVersion
@@ -615,18 +672,45 @@ class ImpalaHS2Client(ImpalaClient):
     self.session_handle = resp.sessionHandle
 
     # List all of the query options and their levels.
-    set_all_handle = self.execute_query("set all", {})
-    self.default_query_options = {}
-    self.query_option_levels = {}
-    for rows in self.fetch(set_all_handle):
-      for name, value, level in rows:
-        self.default_query_options[name.upper()] = value
-        self.query_option_levels[name.upper()] = QueryOptionLevels.from_string(level)
-    try:
-      self.close_query(set_all_handle)
-    except Exception as e:
-      print('{0} {1}'.format(str(e), type(e)), file=sys.stderr)
-      raise
+    # Retrying "set all" should be idempotent
+    num_tries = 1
+    while num_tries <= self.max_tries:
+      raise_error = (num_tries == self.max_tries)
+      set_all_handle = None
+      if self.max_tries > 1:
+        retry_msg = 'Num remaining tries: {0}'.format(self.max_tries - num_tries)
+      else:
+        retry_msg = ''
+      try:
+        set_all_handle = self.execute_query("set all", {})
+        self.default_query_options = {}
+        self.query_option_levels = {}
+        for rows in self.fetch(set_all_handle):
+          for name, value, level in rows:
+            self.default_query_options[name.upper()] = value
+            self.query_option_levels[name.upper()] = QueryOptionLevels.from_string(level)
+        break
+      except (QueryCancelledByShellException, MissingThriftMethodException,
+        QueryStateException):
+        raise
+      except RPCException as r:
+        if (r.exception_type == RPC_EXCEPTION_TAPPLICATION or
+          r.exception_type == RPC_EXCEPTION_SERVER):
+          raise
+        print('Caught exception {0}, type={1} when listing query options. {2}'
+          .format(str(r), type(r), retry_msg), file=sys.stderr)
+        if raise_error:
+          raise
+      except Exception as e:
+        print('Caught exception {0}, type={1} when listing query options. {2}'
+          .format(str(e), type(e), retry_msg), file=sys.stderr)
+        if raise_error:
+          raise
+      finally:
+        if set_all_handle is not None:
+          self.close_query(set_all_handle)
+      time.sleep(self._get_sleep_interval_for_retries(num_tries))
+      num_tries += 1
 
   def close_connection(self):
     if self.session_handle is not None:
@@ -635,7 +719,12 @@ class ImpalaHS2Client(ImpalaClient):
       # server to clean up the session.
       try:
         req = TCloseSessionReq(self.session_handle)
-        resp = self._do_hs2_rpc(lambda: self.imp_service.CloseSession(req))
+
+        def CloseSession():
+          return self.imp_service.CloseSession(req)
+        # CloseSession rpcs don't need retries since we catch all exceptions and close
+        # transport.
+        resp = self._do_hs2_rpc(CloseSession)
         self._check_hs2_rpc_status(resp.status)
       except Exception as e:
         print("Warning: close session RPC failed: {0}, {1}".format(str(e), type(e)))
@@ -644,18 +733,26 @@ class ImpalaHS2Client(ImpalaClient):
 
   def _ping_impala_service(self):
     req = TPingImpalaHS2ServiceReq(self.session_handle)
-    try:
-      resp = self.imp_service.PingImpalaHS2Service(req)
-    except TApplicationException as t:
-      if t.type == TApplicationException.UNKNOWN_METHOD:
-        raise MissingThriftMethodException(t.message)
-      raise
+
+    def PingImpalaHS2Service():
+      return self.imp_service.PingImpalaHS2Service(req)
+    # PingImpalaHS2Service rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(PingImpalaHS2Service, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
+    self.webserver_address = resp.webserver_address
     return (resp.version, resp.webserver_address)
 
   def _create_query_req(self, query_str, set_query_options):
-    query = TExecuteStatementReq(sessionHandle=self.session_handle, statement=query_str,
-        confOverlay=set_query_options, runAsync=True)
+    conf_overlay = {}
+    if sys.version_info.major < 3:
+      key_value_pairs = set_query_options.iteritems()
+    else:
+      key_value_pairs = set_query_options.items()
+    for k, v in key_value_pairs:
+      conf_overlay[utf8_encode_if_needed(k)] = utf8_encode_if_needed(v)
+    query = TExecuteStatementReq(sessionHandle=self.session_handle,
+        statement=utf8_encode_if_needed(query_str),
+        confOverlay=conf_overlay, runAsync=True)
     return query
 
   def execute_query(self, query_str, set_query_options):
@@ -664,13 +761,24 @@ class ImpalaHS2Client(ImpalaClient):
     ImpalaClient method calls for the query."""
     query = self._create_query_req(query_str, set_query_options)
     self.is_query_cancelled = False
-    resp = self._do_hs2_rpc(lambda: self.imp_service.ExecuteStatement(query))
+
+    def ExecuteStatement():
+      return self.imp_service.ExecuteStatement(query)
+    # Read queries should be idempotent but most dml queries are not. Also retrying
+    # query execution from client could be expensive and so likely makes sense to do
+    # it if server is also aware of the retries.
+    resp = self._do_hs2_rpc(ExecuteStatement)
     if resp.status.statusCode != TStatusCode.SUCCESS_STATUS:
-      raise QueryStateException("ERROR: {0}".format(resp.status.errorMessage))
+      msg = utf8_decode_if_needed(resp.status.errorMessage)
+      raise QueryStateException("ERROR: {0}".format(msg))
     handle = resp.operationHandle
     if handle.hasResultSet:
       req = TGetResultSetMetadataReq(handle)
-      resp = self._do_hs2_rpc(lambda: self.imp_service.GetResultSetMetadata(req))
+
+      def GetResultSetMetadata():
+        return self.imp_service.GetResultSetMetadata(req)
+      # GetResultSetMetadata rpc is idempotent and should be safe to retry.
+      resp = self._do_hs2_rpc(GetResultSetMetadata, retry_on_error=True)
       self._check_hs2_rpc_status(resp.status)
       assert resp.schema is not None, resp
       # Attach the schema to the handle for convenience.
@@ -697,7 +805,7 @@ class ImpalaHS2Client(ImpalaClient):
 
     return "{low}:{high}".format(low=low_hex, high=high_hex)
 
-  def _fetch_one_batch(self, query_handle):
+  def fetch(self, query_handle):
     assert query_handle.hasResultSet
     prim_types = [column.typeDesc.types[0].primitiveEntry.type
                   for column in query_handle.schema.columns]
@@ -705,8 +813,13 @@ class ImpalaHS2Client(ImpalaClient):
                         for prim_type in prim_types]
     while True:
       req = TFetchResultsReq(query_handle, TFetchOrientation.FETCH_NEXT,
-          self.fetch_batch_size)
-      resp = self._do_hs2_rpc(lambda: self.imp_service.FetchResults(req))
+          self.fetch_size)
+
+      def FetchResults():
+        return self.imp_service.FetchResults(req)
+      # FetchResults rpc is not idempotent unless the client and server communicate and
+      # results are kept around for retry to be successful.
+      resp = self._do_hs2_rpc(FetchResults)
       self._check_hs2_rpc_status(resp.status)
 
       # Transpose the columns into a row-based format for more convenient processing
@@ -741,7 +854,11 @@ class ImpalaHS2Client(ImpalaClient):
 
   def close_dml(self, last_query_handle):
     req = TCloseImpalaOperationReq(last_query_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.CloseImpalaOperation(req))
+
+    def CloseImpalaOperation():
+      return self.imp_service.CloseImpalaOperation(req)
+    # CloseImpalaOperation rpc is not idempotent for dmls.
+    resp = self._do_hs2_rpc(CloseImpalaOperation)
     self._check_hs2_rpc_status(resp.status)
     if not resp.dml_result:
       raise RPCException("Impala DML operation did not return DML statistics.")
@@ -755,7 +872,11 @@ class ImpalaHS2Client(ImpalaClient):
     if last_query_handle.is_closed:
       return True
     req = TCloseImpalaOperationReq(last_query_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.CloseImpalaOperation(req))
+
+    def CloseImpalaOperation():
+      return self.imp_service.CloseImpalaOperation(req)
+    # CloseImpalaOperation rpc is idempotent for non dml queries and so safe to retry.
+    resp = self._do_hs2_rpc(CloseImpalaOperation, retry_on_error=True)
     last_query_handle.is_closed = True
     return self._is_hs2_nonerror_status(resp.status.statusCode)
 
@@ -765,28 +886,50 @@ class ImpalaHS2Client(ImpalaClient):
     if last_query_handle.is_closed:
       return True
     req = TCancelOperationReq(last_query_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.CancelOperation(req),
-        suppress_error_on_cancel=False)
+
+    def CancelOperation():
+      return self.imp_service.CancelOperation(req)
+    # CancelOperation rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(CancelOperation, retry_on_error=True)
     return self._is_hs2_nonerror_status(resp.status.statusCode)
 
   def get_query_state(self, last_query_handle):
-    resp = self._do_hs2_rpc(
-        lambda: self.imp_service.GetOperationStatus(
-          TGetOperationStatusReq(last_query_handle)))
+    req = TGetOperationStatusReq(last_query_handle)
+
+    def GetOperationStatus():
+      return self.imp_service.GetOperationStatus(req)
+    # GetOperationStatus rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(GetOperationStatus, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
     return resp.operationState
 
   def get_runtime_profile(self, last_query_handle):
-    req = TGetRuntimeProfileReq(last_query_handle, self.session_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.GetRuntimeProfile(req))
+    req = TGetRuntimeProfileReq(last_query_handle, self.session_handle,
+        include_query_attempts=True)
+
+    def GetRuntimeProfile():
+      return self.imp_service.GetRuntimeProfile(req)
+    # GetRuntimeProfile rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(GetRuntimeProfile, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
-    return resp.profile
+    failed_profile = None
+    if resp.failed_profiles and len(resp.failed_profiles) >= 1:
+      failed_profile = resp.failed_profiles[0]
+    return resp.profile, failed_profile
 
   def get_summary(self, last_query_handle):
-    req = TGetExecSummaryReq(last_query_handle, self.session_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.GetExecSummary(req))
+    req = TGetExecSummaryReq(last_query_handle, self.session_handle,
+        include_query_attempts=True)
+
+    def GetExecSummary():
+      return self.imp_service.GetExecSummary(req)
+    # GetExecSummary rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(GetExecSummary, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
-    return resp.summary
+    failed_summary = None
+    if resp.failed_summaries and len(resp.failed_summaries) >= 1:
+      failed_summary = resp.failed_summaries[0]
+    return resp.summary, failed_summary
 
   def get_column_names(self, last_query_handle):
     # The handle has the schema embedded in it.
@@ -804,38 +947,94 @@ class ImpalaHS2Client(ImpalaClient):
     if last_query_handle is None:
       return "Query could not be executed"
     req = TGetLogReq(last_query_handle)
-    resp = self._do_hs2_rpc(lambda: self.imp_service.GetLog(req))
+
+    def GetLog():
+      return self.imp_service.GetLog(req)
+    # GetLog rpc is idempotent and so safe to retry.
+    resp = self._do_hs2_rpc(GetLog, retry_on_error=True)
     self._check_hs2_rpc_status(resp.status)
 
-    log = resp.log
+    log = utf8_decode_if_needed(resp.log)
+
     # Strip progress message out of HS2 log.
     log = HS2_LOG_PROGRESS_REGEX.sub("", log)
     if log and log.strip():
+      log = self._append_retried_query_link(log)
       type_str = "WARNINGS" if warn is True else "ERROR"
       return "%s: %s" % (type_str, log)
     return ""
 
-  def _do_hs2_rpc(self, rpc, suppress_error_on_cancel=True):
+  def _do_hs2_rpc(self, rpc, suppress_error_on_cancel=True, retry_on_error=False):
     """Executes the provided 'rpc' callable and tranlates any exceptions in the
-    appropriate exception for the shell. Exceptions raised include:
+    appropriate exception for the shell. The input 'rpc' must be a python function
+    with the __name__ attribute and not a lambda function. Exceptions raised include:
     * DisconnectedException if the client cannot communicate with the server.
     * QueryCancelledByShellException if 'suppress_error_on_cancel' is true, the RPC
       fails and the query was cancelled from the shell (i.e. self.is_query_cancelled).
     * MissingThriftMethodException if the thrift method is not implemented on the server.
-    Does not validate any status embedded in the returned RPC message."""
+    Does not validate any status embedded in the returned RPC message.
+    If 'retry_on_error' is true, the rpc is retried if an exception is raised. The maximum
+    number of tries is determined by 'self.max_tries'. Retries, if enabled, are attempted
+    for all exceptions other than TApplicationException."""
     self._check_connected()
-    try:
-      return rpc()
-    except TTransportException as e:
-      # issue with the connection with the impalad
-      raise DisconnectedException("Error communicating with impalad: %s" % e)
-    except TApplicationException as t:
-      # Suppress the errors from cancelling a query that is in waiting_to_finish state
-      if suppress_error_on_cancel and self.is_query_cancelled:
-        raise QueryCancelledByShellException()
-      if t.type == TApplicationException.UNKNOWN_METHOD:
-        raise MissingThriftMethodException(t.message)
-      raise RPCException("Application Exception : {0}".format(t))
+    num_tries = 1
+    max_tries = num_tries
+    if retry_on_error:
+      max_tries = self.max_tries
+    while num_tries <= max_tries:
+      raise_error = (num_tries == max_tries)
+      # Generate a retry message, only if retries and supported.
+      will_retry = False
+      retry_secs = None
+      if retry_on_error and self.max_tries > 1:
+        retry_msg = 'Num remaining tries: {0}'.format(max_tries - num_tries)
+        if num_tries < max_tries:
+          will_retry = True
+      else:
+        retry_msg = ''
+      try:
+        return rpc()
+      except TTransportException as e:
+        # issue with the connection with the impalad
+        print('Caught exception {0}, type={1} in {2}. {3}'
+          .format(str(e), type(e), rpc.__name__, retry_msg), file=sys.stderr)
+        if raise_error:
+          raise DisconnectedException("Error communicating with impalad: %s" % e)
+      except TApplicationException as t:
+        # Suppress the errors from cancelling a query that is in waiting_to_finish state
+        if suppress_error_on_cancel and self.is_query_cancelled:
+          raise QueryCancelledByShellException()
+        if t.type == TApplicationException.UNKNOWN_METHOD:
+          raise MissingThriftMethodException(t.message)
+        raise RPCException("Application Exception : {0}".format(t),
+          RPC_EXCEPTION_TAPPLICATION)
+      except HttpError as h:
+        if will_retry:
+          retry_after = h.http_headers.get('Retry-After', None)
+          if retry_after:
+            try:
+              retry_secs = int(retry_after)
+            except ValueError:
+              retry_secs = None
+        if retry_secs:
+          print('Caught exception {0}, type={1} in {2}. {3}, retry after {4} secs'
+                .format(str(h), type(h), rpc.__name__, retry_msg, retry_secs),
+                file=sys.stderr)
+        else:
+          print('Caught exception {0}, type={1} in {2}. {3}'
+                .format(str(h), type(h), rpc.__name__, retry_msg), file=sys.stderr)
+        if raise_error:
+          raise
+      except Exception as e:
+        print('Caught exception {0}, type={1} in {2}. {3}'
+          .format(str(e), type(e), rpc.__name__, retry_msg), file=sys.stderr)
+        if raise_error:
+          raise
+      if retry_secs:
+        time.sleep(retry_secs)
+      else:
+        time.sleep(self._get_sleep_interval_for_retries(num_tries))
+      num_tries += 1
 
   def _check_hs2_rpc_status(self, status):
     """If the TCLIService.TStatus 'status' is an error status the raise an exception
@@ -848,7 +1047,8 @@ class ImpalaHS2Client(ImpalaClient):
       # Suppress the errors from cancelling a query that is in fetch state
       if self.is_query_cancelled:
         raise QueryCancelledByShellException()
-      raise RPCException("ERROR: {0}".format(status.errorMessage))
+      raise RPCException("ERROR: {0}".format(status.errorMessage),
+        RPC_EXCEPTION_SERVER)
     elif status.statusCode == TStatusCode.INVALID_HANDLE_STATUS:
       if self.is_query_cancelled:
         raise QueryCancelledByShellException()
@@ -884,12 +1084,11 @@ class ImpalaBeeswaxClient(ImpalaClient):
     return ImpalaService.Client(protocol)
 
   def _options_to_string_list(self, set_query_options):
-      if sys.version_info.major < 3:
-        key_value_pairs = set_query_options.iteritems()
-      else:
-        key_value_pairs = set_query_options.items()
-
-      return ["%s=%s" % (k, v) for (k, v) in key_value_pairs]
+    if sys.version_info.major < 3:
+      key_value_pairs = set_query_options.iteritems()
+    else:
+      key_value_pairs = set_query_options.items()
+    return [utf8_encode_if_needed("%s=%s" % (k, v)) for (k, v) in key_value_pairs]
 
   def _open_session(self):
     # Beeswax doesn't have a "session" concept independent of connections, so
@@ -930,12 +1129,13 @@ class ImpalaBeeswaxClient(ImpalaClient):
       raise
     except TTransportException as e:
       raise DisconnectedException("Error communicating with impalad: %s" % e)
+    self.webserver_address = resp.webserver_address
     return (resp.version, resp.webserver_address)
 
   def _create_query_req(self, query_str, set_query_options):
     query = BeeswaxService.Query()
     query.hadoop_user = self.user
-    query.query = query_str
+    query.query = utf8_encode_if_needed(query_str)
     query.configuration = self._options_to_string_list(set_query_options)
     return query
 
@@ -961,14 +1161,21 @@ class ImpalaBeeswaxClient(ImpalaClient):
       return self.ERROR_STATE
     return state
 
-  def _fetch_one_batch(self, query_handle):
+  def fetch(self, query_handle):
     while True:
       result, rpc_status = self._do_beeswax_rpc(
          lambda: self.imp_service.fetch(query_handle, False,
-                                        self.fetch_batch_size))
+                                        self.fetch_size))
       if rpc_status != RpcStatus.OK:
         raise RPCException()
-      yield [row.split('\t') for row in result.data]
+
+      def split_row_and_decode_if_needed(row):
+        # Decode before splitting as this can remove incidental tabs from
+        # multibyte characters.
+        return utf8_decode_if_needed(row).split('\t')
+
+      yield [split_row_and_decode_if_needed(row) for row in result.data]
+
       if not result.has_more:
         return
 
@@ -1003,15 +1210,15 @@ class ImpalaBeeswaxClient(ImpalaClient):
     profile, rpc_status = self._do_beeswax_rpc(
         lambda: self.imp_service.GetRuntimeProfile(last_query_handle))
     if rpc_status == RpcStatus.OK and profile:
-      return profile
-    return None
+      return profile, None
+    return None, None
 
   def get_summary(self, last_query_handle):
     summary, rpc_status = self._do_beeswax_rpc(
       lambda: self.imp_service.GetExecSummary(last_query_handle))
     if rpc_status == RpcStatus.OK and summary:
-      return summary
-    return None
+      return summary, None
+    return None, None
 
   def get_column_names(self, last_query_handle):
     # Note: the code originally ignored the RPC status. don't mess with it.
@@ -1037,6 +1244,8 @@ class ImpalaBeeswaxClient(ImpalaClient):
       type_str = "warn" if warn is True else "error"
       return "Failed to get %s log: %s" % (type_str, rpc_status)
     if log and log.strip():
+      log = utf8_decode_if_needed(log)
+      log = self._append_retried_query_link(log)
       type_str = "WARNINGS" if warn is True else "ERROR"
       return "%s: %s" % (type_str, log)
     return ""
@@ -1074,7 +1283,7 @@ class ImpalaBeeswaxClient(ImpalaClient):
       # Suppress the errors from cancelling a query that is in fetch state
       if suppress_error_on_cancel and self.is_query_cancelled:
         raise QueryCancelledByShellException()
-      raise RPCException("ERROR: %s" % b.message)
+      raise RPCException(utf8_encode_if_needed("ERROR: %s") % b.message)
     except TTransportException as e:
       # issue with the connection with the impalad
       raise DisconnectedException("Error communicating with impalad: %s" % e)
@@ -1120,4 +1329,3 @@ class ImpalaBeeswaxClient(ImpalaClient):
           raise RPCException("ERROR: %s" % e.message)
         if "QueryNotFoundException" in str(e):
           raise QueryStateException('Error: Stale query handle')
-
