@@ -1,4 +1,5 @@
 #!/usr/bin/env python
+# -*- coding: utf-8 -*-
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -18,7 +19,7 @@
 # under the License.
 #
 # Impala's shell
-from __future__ import print_function
+from __future__ import print_function, unicode_literals
 from compatibility import _xrange as xrange
 
 import cmd
@@ -36,12 +37,13 @@ import subprocess
 import sys
 import textwrap
 import time
+import traceback
 
 from impala_client import ImpalaHS2Client, ImpalaBeeswaxClient, QueryOptionLevels
 from impala_shell_config_defaults import impala_shell_defaults
 from option_parser import get_option_parser, get_config_from_file
-from shell_output import DelimitedOutputFormatter, OutputStream, PrettyOutputFormatter
-from shell_output import OverwritingStdErrOutputStream
+from shell_output import (DelimitedOutputFormatter, OutputStream, PrettyOutputFormatter,
+                          OverwritingStdErrOutputStream)
 from subprocess import call
 from shell_exceptions import (RPCException, DisconnectedException, QueryStateException,
     QueryCancelledByShellException, MissingThriftMethodException)
@@ -65,6 +67,7 @@ DEFAULT_BEESWAX_PORT = 21000
 DEFAULT_HS2_PORT = 21050
 DEFAULT_HS2_HTTP_PORT = 28000
 
+
 def strip_comments(sql):
   """sqlparse default implementation of strip comments has a bad performance when parsing
   very large SQL due to the grouping. This is because the default implementation tries to
@@ -77,6 +80,7 @@ def strip_comments(sql):
   stack.postprocess.append(sqlparse.filters.SerializerUnicode())
   return ''.join(stack.run(sql, 'utf-8')).strip()
 
+
 class CmdStatus:
   """Values indicate the execution status of a command to the cmd shell driver module
   SUCCESS and ERROR continue running the shell and ABORT exits the shell
@@ -93,23 +97,26 @@ class FatalShellException(Exception):
   error should be logged to stderr before raising this exception."""
   pass
 
-class ImpalaPrettyTable(prettytable.PrettyTable):
-  """Patched version of PrettyTable with different unicode handling - instead of throwing
-  exceptions when a character can't be converted to unicode, it is replaced with a
-  placeholder character."""
-  def _unicode(self, value):
-    if not isinstance(value, basestring):
-      value = str(value)
-    if not isinstance(value, unicode):
-      # If a value cannot be encoded, replace it with a placeholder.
-      value = unicode(value, self.encoding, "replace")
-    return value
 
 class QueryOptionDisplayModes:
   REGULAR_OPTIONS_ONLY = 1
   ALL_OPTIONS = 2
 
-class ImpalaShell(object, cmd.Cmd):
+
+class QueryAttemptDisplayModes:
+  """The display mode when runtime profiles or summaries are printed to the console.
+  If the query has not been retried, then the display mode does not change anything.
+  The format is always the same. If the query has been retried, then the ALL option will
+  print both the original and retried profiles/summaries. If the LATEST option is
+  specified, then only the retried profile/summary will be printed. If the ORIGINAL option
+  is specified, then only the original profile/summary will be printed."""
+  ALL = "all"
+  LATEST = "latest"
+  ORIGINAL = "original"
+
+
+# Py3 method resolution order requires 'object' to be last w/ multiple inheritance
+class ImpalaShell(cmd.Cmd, object):
   """ Simple Impala Shell.
 
   Implements the context manager interface to ensure client connections and sessions
@@ -222,6 +229,7 @@ class ImpalaShell(object, cmd.Cmd):
     self.ignore_query_failure = options.ignore_query_failure
 
     self.http_path = options.http_path
+    self.fetch_size = options.fetch_size
 
     # Due to a readline bug in centos/rhel7, importing it causes control characters to be
     # printed. This breaks any scripting against the shell in non-interactive mode. Since
@@ -450,7 +458,7 @@ class ImpalaShell(object, cmd.Cmd):
     # Strip any comments to make a statement such as the following be considered as
     # ending with a delimiter:
     # select 1 + 1; -- this is a comment
-    line = strip_comments(line).encode('utf-8').rstrip()
+    line = strip_comments(line).rstrip()
     if line.endswith(ImpalaShell.CMD_DELIM):
       try:
         # Look for an open quotation in the entire command, and not just the
@@ -504,6 +512,8 @@ class ImpalaShell(object, cmd.Cmd):
         a completed command.
       - The contents are passed to the appropriate method for execution.
       - partial_cmd is reset to an empty string.
+
+    Returns text type result (unicode in Python2, str in Python3)
     """
     if self.readline: current_history_len = self.readline.get_current_history_length()
     # Input is incomplete, store the contents and do nothing.
@@ -515,10 +525,10 @@ class ImpalaShell(object, cmd.Cmd):
 
       # partial_cmd is already populated, add the current input after a newline.
       if self.partial_cmd and cmd:
-        self.partial_cmd = "%s\n%s" % (self.partial_cmd, cmd)
+        self.partial_cmd = "{0}\n{1}".format(self.partial_cmd, cmd)
       else:
         # If the input string is empty or partial_cmd is empty.
-        self.partial_cmd = "%s%s" % (self.partial_cmd, cmd)
+        self.partial_cmd = "{0}{1}".format(self.partial_cmd, cmd)
       # Remove the most recent item from history if:
       #   -- The current state of user input in incomplete.
       #   -- The most recent user input is not an empty string
@@ -528,16 +538,21 @@ class ImpalaShell(object, cmd.Cmd):
       return str()
     elif self.partial_cmd:  # input ends with a delimiter and partial_cmd is not empty
       if cmd != ImpalaShell.CMD_DELIM:
-        completed_cmd = "%s\n%s" % (self.partial_cmd, cmd)
+        completed_cmd = "{0}\n{1}".format(self.partial_cmd, cmd)
       else:
-        completed_cmd = "%s%s" % (self.partial_cmd, cmd)
+        completed_cmd = "{0}{1}".format(self.partial_cmd, cmd)
       # Reset partial_cmd to an empty string
       self.partial_cmd = str()
       # Replace the most recent history item with the completed command.
       completed_cmd = sqlparse.format(completed_cmd)
       if self.readline and current_history_len > 0:
-        self.readline.replace_history_item(current_history_len - 1,
-            completed_cmd.encode('utf-8'))
+        # readline.replace_history_item(pos, line) requires 'line' in bytes in Python2,
+        # and str in Python3.
+        if sys.version_info.major == 2:
+          history_cmd = completed_cmd.encode('utf-8')
+        else:
+          history_cmd = completed_cmd
+        self.readline.replace_history_item(current_history_len - 1, history_cmd)
       # Revert the prompt to its earlier state
       self.prompt = self.cached_prompt
     else:  # Input has a delimiter and partial_cmd is empty
@@ -547,22 +562,24 @@ class ImpalaShell(object, cmd.Cmd):
   def _new_impala_client(self):
     protocol = options.protocol.lower()
     if protocol == 'hs2':
-      return ImpalaHS2Client(self.impalad, self.kerberos_host_fqdn, self.use_kerberos,
-                          self.kerberos_service_name, self.use_ssl,
-                          self.ca_cert, self.user, self.ldap_password,
-                          self.use_ldap, self.client_connect_timeout_ms, self.verbose,
-                          use_http_base_transport=False, http_path=self.http_path)
+      return ImpalaHS2Client(self.impalad, self.fetch_size, self.kerberos_host_fqdn,
+                          self.use_kerberos, self.kerberos_service_name, self.use_ssl,
+                          self.ca_cert, self.user, self.ldap_password, self.use_ldap,
+                          self.client_connect_timeout_ms, self.verbose,
+                          use_http_base_transport=False, http_path=self.http_path,
+                          auth_cookie_names=None)
     elif protocol == 'hs2-http':
-      return ImpalaHS2Client(self.impalad, self.kerberos_host_fqdn, self.use_kerberos,
-                          self.kerberos_service_name, self.use_ssl,
-                          self.ca_cert, self.user, self.ldap_password,
-                          self.use_ldap, self.client_connect_timeout_ms, self.verbose,
-                          use_http_base_transport=True, http_path=self.http_path)
+      return ImpalaHS2Client(self.impalad, self.fetch_size, self.kerberos_host_fqdn,
+                          self.use_kerberos, self.kerberos_service_name, self.use_ssl,
+                          self.ca_cert, self.user, self.ldap_password, self.use_ldap,
+                          self.client_connect_timeout_ms, self.verbose,
+                          use_http_base_transport=True, http_path=self.http_path,
+                          auth_cookie_names='impala.auth')
     elif protocol == 'beeswax':
-      return ImpalaBeeswaxClient(self.impalad, self.kerberos_host_fqdn, self.use_kerberos,
-                          self.kerberos_service_name, self.use_ssl,
-                          self.ca_cert, self.user, self.ldap_password,
-                          self.use_ldap, self.client_connect_timeout_ms, self.verbose)
+      return ImpalaBeeswaxClient(self.impalad, self.fetch_size, self.kerberos_host_fqdn,
+                          self.use_kerberos, self.kerberos_service_name, self.use_ssl,
+                          self.ca_cert, self.user, self.ldap_password, self.use_ldap,
+                          self.client_connect_timeout_ms, self.verbose)
     else:
       err_msg = "Invalid --protocol value {0}, must be beeswax, hs2 or hs2-http."
       print(err_msg.format(protocol), file=sys.stderr)
@@ -616,7 +633,12 @@ class ImpalaShell(object, cmd.Cmd):
         host=self.impalad[0], port=self.impalad[1], db=db)
 
   def precmd(self, args):
-    args = self.sanitise_input(args)
+    # In Python2, 'args' could in str type if it's the original input line, or in unicode
+    # type if it's a split query appended by us. See how we deal with 'parsed_cmds' below.
+    if sys.version_info.major == 2 and isinstance(args, str):
+      args = self.sanitise_input(args.decode('utf-8'))  # python2
+    else:
+      args = self.sanitise_input(args)  # python3
     if not args: return args
     # Split args using sqlparse. If there are multiple queries present in user input,
     # the length of the returned query list will be greater than one.
@@ -633,7 +655,7 @@ class ImpalaShell(object, cmd.Cmd):
       print("Connection lost, reconnecting...", file=sys.stderr)
       self._connect()
       self._validate_database(immediately=True)
-    return args.encode('utf-8')
+    return args
 
   def onecmd(self, line):
     """Overridden to ensure the variable replacement is processed in interactive
@@ -680,13 +702,22 @@ class ImpalaShell(object, cmd.Cmd):
     return status
 
   def do_summary(self, args):
+    split_args = args.split()
+    if len(split_args) > 1:
+      print("'summary' only accepts 0 or 1 arguments", file=sys.stderr)
+      return CmdStatus.ERROR
     if not self.last_query_handle:
       print("Could not retrieve summary: no previous query.", file=sys.stderr)
       return CmdStatus.ERROR
 
-    summary = None
+    display_mode = QueryAttemptDisplayModes.LATEST
+    if len(split_args) == 1:
+      display_mode = self.get_query_attempt_display_mode(split_args[0])
+      if display_mode is None:
+        return CmdStatus.ERROR
+
     try:
-      summary = self.imp_client.get_summary(self.last_query_handle)
+      summary, failed_summary = self.imp_client.get_summary(self.last_query_handle)
     except RPCException as e:
       import re
       error_pattern = re.compile("ERROR: Query id \d+:\d+ not found.")
@@ -698,6 +729,30 @@ class ImpalaShell(object, cmd.Cmd):
     if summary.nodes is None:
       print("Summary not available", file=sys.stderr)
       return CmdStatus.SUCCESS
+
+    if display_mode == QueryAttemptDisplayModes.ALL:
+      print("Query Summary:")
+      self.print_exec_summary(summary)
+      if failed_summary:
+        print("Failed Query Summary:")
+        self.print_exec_summary(failed_summary)
+    elif display_mode == QueryAttemptDisplayModes.LATEST:
+      self.print_exec_summary(summary)
+    elif display_mode == QueryAttemptDisplayModes.ORIGINAL:
+      self.print_exec_summary(failed_summary)
+    else:
+      raise FatalShellException("Invalid value for query summary display mode")
+
+  @staticmethod
+  def get_query_attempt_display_mode(arg_mode):
+    arg_mode = str(arg_mode).lower()
+    if arg_mode not in [QueryAttemptDisplayModes.ALL,
+        QueryAttemptDisplayModes.LATEST, QueryAttemptDisplayModes.ORIGINAL]:
+      print("Invalid value for query attempt display mode: \'" +
+          arg_mode + "\'. Valid values are [ALL | LATEST | ORIGINAL]")
+    return arg_mode
+
+  def print_exec_summary(self, summary):
     output = []
     table = self._default_summary_table()
     self.imp_client.build_summary_table(summary, 0, False, 0, False, output)
@@ -981,10 +1036,25 @@ class ImpalaShell(object, cmd.Cmd):
     if flush:
       file_descriptor.flush()
 
-  def print_runtime_profile(self, profile, status=False):
+  def print_runtime_profile(self, profile, failed_profile,
+        profile_display_mode=QueryAttemptDisplayModes.LATEST, status=False):
+    """Prints the given runtime profiles to the console. Optionally prints the failed
+    profile if the query was retried. The format the profiles are printed is controlled
+    by the option profile_display_mode, see QueryProfileDisplayModes docs above.
+    """
     if self.show_profiles or status:
-      if profile is not None:
-        print("Query Runtime Profile:\n" + profile)
+      if profile:
+        query_profile_prefix = "Query Runtime Profile:\n"
+        if profile_display_mode == QueryAttemptDisplayModes.ALL:
+          print(query_profile_prefix + profile)
+          if failed_profile:
+            print("Failed Query Runtime Profile(s):\n" + failed_profile)
+        elif profile_display_mode == QueryAttemptDisplayModes.LATEST:
+          print(query_profile_prefix + profile)
+        elif profile_display_mode == QueryAttemptDisplayModes.ORIGINAL:
+          print(query_profile_prefix + failed_profile if failed_profile else profile)
+        else:
+          raise FatalShellException("Invalid value for query profile display mode")
 
   def _parse_table_name_arg(self, arg):
     """ Parses an argument string and returns the result as a db name, table name combo.
@@ -1027,14 +1097,25 @@ class ImpalaShell(object, cmd.Cmd):
 
   def do_profile(self, args):
     """Prints the runtime profile of the last DML statement or SELECT query executed."""
-    if len(args) > 0:
-      print("'profile' does not accept any arguments", file=sys.stderr)
+    split_args = args.split()
+    if len(split_args) > 1:
+      print("'profile' only accepts 0 or 1 arguments", file=sys.stderr)
       return CmdStatus.ERROR
     elif self.last_query_handle is None:
       print('No previous query available to profile', file=sys.stderr)
       return CmdStatus.ERROR
-    profile = self.imp_client.get_runtime_profile(self.last_query_handle)
-    return self.print_runtime_profile(profile, True)
+
+    # Parse and validate the QueryProfileDisplayModes option.
+    profile_display_mode = QueryAttemptDisplayModes.LATEST
+    if len(split_args) == 1:
+      profile_display_mode = self.get_query_attempt_display_mode(split_args[0])
+      if profile_display_mode is None:
+        return CmdStatus.ERROR
+
+    profile, failed_profile = self.imp_client.get_runtime_profile(
+        self.last_query_handle)
+    return self.print_runtime_profile(profile, failed_profile, profile_display_mode,
+            True)
 
   def do_select(self, args):
     """Executes a SELECT... query, fetching all rows"""
@@ -1046,6 +1127,7 @@ class ImpalaShell(object, cmd.Cmd):
     Impala shell cannot get child query handle so it cannot
     query live progress for COMPUTE STATS query. Disable live
     progress/summary callback for COMPUTE STATS query."""
+
     query = self._build_query_string(self.last_leading_comment, self.orig_cmd, args)
     (prev_live_progress, prev_live_summary) = self.live_progress, self.live_summary
     (self.live_progress, self.live_summary) = False, False
@@ -1057,6 +1139,10 @@ class ImpalaShell(object, cmd.Cmd):
 
   def _format_outputstream(self):
     column_names = self.imp_client.get_column_names(self.last_query_handle)
+    if sys.version_info.major == 2:
+      column_names = [
+          col.decode("utf8", errors="replace") if isinstance(col, str) else col
+          for col in column_names]
     if self.write_delimited:
       formatter = DelimitedOutputFormatter(field_delim=self.output_delimiter)
       self.output_stream = OutputStream(formatter, filename=self.output_file)
@@ -1077,7 +1163,7 @@ class ImpalaShell(object, cmd.Cmd):
 
     checkpoint = time.time()
     if checkpoint - self.last_summary > self.PROGRESS_UPDATE_INTERVAL:
-      summary = self.imp_client.get_summary(self.last_query_handle)
+      summary, failed_summary = self.imp_client.get_summary(self.last_query_handle)
       if not summary:
         return
 
@@ -1087,6 +1173,18 @@ class ImpalaShell(object, cmd.Cmd):
         self.last_summary = time.time()
         return
 
+      data = ""
+      if summary.error_logs:
+        for error_line in summary.error_logs:
+          data += error_line + "\n"
+          if self.webserver_address:
+            query_id_search = re.search("Retrying query using query id: (.*)",
+                                        error_line)
+            if query_id_search and len(query_id_search.groups()) == 1:
+              retried_query_id = query_id_search.group(1)
+              data += "Retried query link: %s\n"\
+                      % self.imp_client.get_query_link(retried_query_id)
+
       if summary.progress:
         progress = summary.progress
 
@@ -1095,7 +1193,6 @@ class ImpalaShell(object, cmd.Cmd):
           self.last_summary = time.time()
           return
 
-        data = ""
         if self.live_progress and progress.total_scan_ranges > 0:
           val = ((summary.progress.num_completed_scan_ranges * 100) /
                  summary.progress.total_scan_ranges)
@@ -1109,7 +1206,7 @@ class ImpalaShell(object, cmd.Cmd):
           formatter = PrettyOutputFormatter(table)
           data += formatter.format(output) + "\n"
 
-        self.progress_stream.write(data)
+      self.progress_stream.write(data)
       self.last_summary = time.time()
 
   def _default_summary_table(self):
@@ -1145,8 +1242,7 @@ class ImpalaShell(object, cmd.Cmd):
       self.last_summary = time.time()
       if print_web_link:
         self._print_if_verbose(
-            "Query progress can be monitored at: %s/query_plan?query_id=%s" %
-            (self.webserver_address,
+            "Query progress can be monitored at: %s" % self.imp_client.get_query_link(
              self.imp_client.get_query_id_str(self.last_query_handle)))
 
       wait_to_finish = self.imp_client.wait_to_finish(self.last_query_handle,
@@ -1173,7 +1269,7 @@ class ImpalaShell(object, cmd.Cmd):
         for rows in rows_fetched:
           # IMPALA-4418: Break out of the loop to prevent printing an unnecessary empty line.
           if len(rows) == 0:
-            break
+            continue
           self.output_stream.write(rows)
           num_rows += len(rows)
 
@@ -1201,8 +1297,9 @@ class ImpalaShell(object, cmd.Cmd):
       if not is_dml:
         self.imp_client.close_query(self.last_query_handle)
       try:
-        profile = self.imp_client.get_runtime_profile(self.last_query_handle)
-        self.print_runtime_profile(profile)
+        profile, retried_profile = self.imp_client.get_runtime_profile(
+            self.last_query_handle)
+        self.print_runtime_profile(profile, retried_profile)
       except RPCException as e:
         if self.show_profiles: raise e
       return CmdStatus.SUCCESS
@@ -1211,11 +1308,26 @@ class ImpalaShell(object, cmd.Cmd):
     except RPCException as e:
       # could not complete the rpc successfully
       print(e, file=sys.stderr)
+    except UnicodeDecodeError as e:
+      # An error occoured possibly during the fetching.
+      # Depending of which protocol is at use it can come from different places.
+      # Possibly occours because we try to display binary data which contains
+      # undecodable elements.
+      if self.last_query_handle is not None:
+        self.imp_client.close_query(self.last_query_handle)
+      print('UnicodeDecodeError : %s \nPlease check for columns containing binary data '
+          'to find the possible source of the error.' % (e,), file=sys.stderr)
     except QueryStateException as e:
       # an exception occurred while executing the query
       if self.last_query_handle is not None:
         self.imp_client.close_query(self.last_query_handle)
-      print(e, file=sys.stderr)
+      msg = e.value
+      # Python2 will implicitly convert unicode to str when printing to stderr. It's done
+      # using the default 'ascii' encoding, which will fail for UTF-8 error messages.
+      # Here we use 'utf-8' to explicitly convert 'msg' to str if it's in unicode type.
+      if sys.version_info.major == 2 and isinstance(msg, unicode):
+        msg = msg.encode('utf-8')
+      print(msg, file=sys.stderr)
     except DisconnectedException as e:
       # the client has lost the connection
       print(e, file=sys.stderr)
@@ -1234,6 +1346,8 @@ class ImpalaShell(object, cmd.Cmd):
       # if the exception is unknown, there was possibly an issue with the connection
       # set the shell as disconnected
       print('Unknown Exception : %s' % (e,), file=sys.stderr)
+      # Print the stack trace for the exception.
+      traceback.print_exc()
       self.close_connection()
       self.prompt = ImpalaShell.DISCONNECTED_PROMPT
     return CmdStatus.ERROR
@@ -1245,10 +1359,9 @@ class ImpalaShell(object, cmd.Cmd):
     Should be called after the query has finished and before data is fetched.
     All data is left aligned.
     """
-    table = ImpalaPrettyTable()
+    table = prettytable.PrettyTable()
     for column in column_names:
-      # Column names may be encoded as utf-8
-      table.add_column(column.decode('utf-8', 'ignore'), [])
+      table.add_column(column, [])
     table.align = "l"
     return table
 
@@ -1260,17 +1373,15 @@ class ImpalaShell(object, cmd.Cmd):
   def do_with(self, args):
     """Executes a query with a WITH clause, fetching all rows"""
     query = self._build_query_string(self.last_leading_comment, self.orig_cmd, args)
-    # Use shlex to deal with escape quotes in string literals.
-    # Set posix=False to preserve the quotes.
-    tokens = shlex.split(strip_comments(query.lstrip()).encode('utf-8'),
-                         posix=False)
+    # Parse the query with sqlparse to identify if it is a DML query or not.
+    # Because the WITH clause may precede DML or SELECT queries, checking the
+    # first string token is insufficient.
+    parsed = sqlparse.parse(query)[0]
+    query_type = sqlparse.sql.Statement(parsed.tokens).get_type()
     try:
-      # Because the WITH clause may precede DML or SELECT queries,
-      # just checking the first token is insufficient.
-      is_dml = False
-      if filter(self.DML_REGEX.match, tokens): is_dml = True
-      return self._execute_stmt(query, is_dml=is_dml, print_web_link=True)
-    except ValueError as e:
+      is_dml = self.DML_REGEX.match(query_type.lower())
+      return self._execute_stmt(query, bool(is_dml), print_web_link=True)
+    except ValueError:
       return self._execute_stmt(query, print_web_link=True)
 
   def do_use(self, args):
@@ -1337,7 +1448,7 @@ class ImpalaShell(object, cmd.Cmd):
     if self.readline and self.readline.get_current_history_length() > 0:
       for index in xrange(1, self.readline.get_current_history_length() + 1):
         cmd = self.readline.get_history_item(index)
-        print('[%d]: %s' % (index, cmd), file=sys.stderr)
+        print('[%d]: %s' % (index, cmd.decode('utf-8', 'replace')), file=sys.stderr)
     else:
       print(READLINE_UNAVAILABLE_ERROR, file=sys.stderr)
 
@@ -1429,37 +1540,57 @@ class ImpalaShell(object, cmd.Cmd):
     command: select
     leading comment: /*first comment*/
     """
-    leading_comment, line = ImpalaShell.strip_leading_comment(line.strip())
-    line = line.encode('utf-8')
-    if leading_comment:
-      leading_comment = leading_comment.encode('utf-8')
+    if ImpalaShell._has_leading_comment(line):
+      leading_comment, line = ImpalaShell.strip_leading_comment(line.strip())
+    else:
+      leading_comment, line = None, line.strip()
+
     if line and line[0] == '@':
       line = 'rerun ' + line[1:]
     return super(ImpalaShell, self).parseline(line) + (leading_comment,)
 
+  @staticmethod
+  def _has_leading_comment(raw_line):
+    """
+    Helper function that returns Boolean true if a query starts with a comment.
+    This saves us from relying on sqlparse filtering, which can be slow.
+    """
+    line = raw_line.lstrip()
+    if line and (line.startswith('--') or line.startswith('/*')):
+      return True
+    else:
+      return False
+
   @staticmethod
   def strip_leading_comment(sql):
     """
     Filter a leading comment in the SQL statement. This function returns a tuple
     containing (leading comment, line without the leading comment).
     """
-    class StripLeadingCommentFilter:
+    class StripLeadingCommentFilter(object):
       def __init__(self):
         self.comment = None
 
       def _process(self, tlist):
+        """
+        Iterate through the list of tokens, appending each leading commment
+        to self.comment, and then popping that element off the list. When we
+        hit the first non-comment and non-whitespace token, then we're done --
+        the remainder after that point is the SQL statement.
+        """
         token = tlist.token_first()
-        if self._is_comment(token):
-          self.comment = ''
-          while token:
-            if self._is_comment(token) or self._is_whitespace(token):
-              tidx = tlist.token_index(token)
-              self.comment += token.value
-              tlist.tokens.pop(tidx)
-              tidx -= 1
-              token = tlist.token_next(tidx, False)
+        while token:
+          if self._is_comment(token) or self._is_whitespace(token):
+            if self.comment is None:
+              self.comment = token.value
             else:
-              break
+              self.comment += token.value
+            tlist.tokens.pop(0)
+
+            # skip_ws=False treats white space characters as tokens also
+            token = tlist.token_first(skip_ws=False)
+          else:
+            break
 
       def _is_comment(self, token):
         return isinstance(token, sqlparse.sql.Comment) or \
@@ -1470,8 +1601,8 @@ class ImpalaShell(object, cmd.Cmd):
         return token.ttype == sqlparse.tokens.Whitespace or \
                token.ttype == sqlparse.tokens.Newline
 
-      def process(self, stack, stmt):
-        [self.process(stack, sgroup) for sgroup in stmt.get_sublists()]
+      def process(self, stmt):
+        [self.process(sgroup) for sgroup in stmt.get_sublists()]
         self._process(stmt)
 
     stack = sqlparse.engine.FilterStack()
@@ -1490,6 +1621,9 @@ class ImpalaShell(object, cmd.Cmd):
     history_len = self.readline.get_current_history_length()
     # load the history and replace the shell's delimiter with EOL
     history_items = map(self.readline.get_history_item, xrange(1, history_len + 1))
+    if sys.version_info.major == 2:
+      src_delim = src_delim.encode('utf-8')
+      tgt_delim = tgt_delim.encode('utf-8')
     history_items = [item.replace(src_delim, tgt_delim) for item in history_items]
     # Clear the original history and replace it with the mutated history.
     self.readline.clear_history()
@@ -1533,7 +1667,7 @@ class ImpalaShell(object, cmd.Cmd):
     return True
 
 
-TIPS=[
+TIPS = [
   "Press TAB twice to see a list of available commands.",
   "After running a query, type SUMMARY to see a summary of where time was spent.",
   "The SET command shows the current value of all shell and query options.",
@@ -1559,16 +1693,18 @@ command."
 EXPLAIN command. You can change the level of detail in the EXPLAIN output by setting the \
 EXPLAIN_LEVEL query option.",
   "When you set a query option it lasts for the duration of the Impala shell session."
-  ]
+]
 
 HEADER_DIVIDER =\
   "***********************************************************************************"
 
+
 def _format_tip(tip):
   """Takes a tip string and splits it on word boundaries so that it fits neatly inside the
   shell header."""
   return '\n'.join([l for l in textwrap.wrap(tip, len(HEADER_DIVIDER))])
 
+
 WELCOME_STRING = """\
 ***********************************************************************************
 Welcome to the Impala shell.
@@ -1580,9 +1716,9 @@ Welcome to the Impala shell.
   % (VERSION_STRING, _format_tip(random.choice(TIPS)))
 
 
-def parse_query_text(query_text, utf8_encode_policy='strict'):
-  """Parse query file text to extract queries and encode into utf-8"""
-  query_list = [q.encode('utf-8', utf8_encode_policy) for q in sqlparse.split(query_text)]
+def parse_query_text(query_text):
+  """Parse query file text to extract queries"""
+  query_list = sqlparse.split(query_text)
   # Remove trailing comments in the input, if any. We do this because sqlparse splits the
   # input at query boundaries and associates the query only with preceding comments
   # (following comments are associated with the next query). This is a problem with
@@ -1600,6 +1736,7 @@ def parse_query_text(query_text, utf8_encode_policy='strict'):
     query_list.pop()
   return query_list
 
+
 def parse_variables(keyvals):
   """Parse variable assignments passed as arguments in the command line"""
   kv_pattern = r'(%s)=(.*)$' % (ImpalaShell.VALID_VAR_NAME_PATTERN,)
@@ -1618,9 +1755,17 @@ def parse_variables(keyvals):
 
 
 def replace_variables(set_variables, input_string):
-  """Replaces variable within the string with their corresponding values using the
-     given set_variables."""
+  """
+  Replaces variable within the input_string with their corresponding values
+  from the given set_variables.
+  """
   errors = False
+
+  # In the case of a byte sequence, re.findall() will choke under python 3
+  if sys.version_info.major > 2:
+    if not isinstance(input_string, str):
+      input_string = input_string.decode('utf-8')
+
   matches = set([v.upper() for v in re.findall(r'(?<!\\)\${([^}]+)}', input_string)])
   for name in matches:
     value = None
@@ -1660,21 +1805,21 @@ def get_var_name(name):
       return var_name
   return None
 
+
 def execute_queries_non_interactive_mode(options, query_options):
   """Run queries in non-interactive mode. Return True on success. Logs the
   error and returns False otherwise."""
   if options.query_file:
-    try:
-      # "-" here signifies input from STDIN
-      if options.query_file == "-":
-        query_file_handle = sys.stdin
-      else:
-        query_file_handle = open(options.query_file, 'r')
-    except Exception, e:
-      print("Could not open file '%s': %s" % (options.query_file, e), file=sys.stderr)
-      return False
-
-    query_text = query_file_handle.read()
+    # "-" here signifies input from STDIN
+    if options.query_file == "-":
+      query_text = sys.stdin.read()
+    else:
+      try:
+        with open(options.query_file, 'r') as query_file_handle:
+          query_text = query_file_handle.read()
+      except Exception as e:
+        print("Could not open file '%s': %s" % (options.query_file, e), file=sys.stderr)
+        return False
   elif options.query:
     query_text = options.query
   else:
@@ -1685,6 +1830,7 @@ def execute_queries_non_interactive_mode(options, query_options):
     return (shell.execute_query_list(shell.cmdqueue) and
             shell.execute_query_list(queries))
 
+
 def get_intro(options):
   """Get introduction message for start-up. The last character should not be a return."""
   if not options.verbose:
@@ -1695,6 +1841,11 @@ def get_intro(options):
   if not options.ssl and options.creds_ok_in_clear and options.use_ldap:
     intro += ("\n\nLDAP authentication is enabled, but the connection to Impala is "
               "not secured by TLS.\nALL PASSWORDS WILL BE SENT IN THE CLEAR TO IMPALA.")
+
+  if options.protocol == 'beeswax':
+    intro += ("\n\nWARNING: The beeswax protocol is deprecated and will be removed in a "
+              "future version of Impala.")
+
   return intro
 
 
@@ -1774,7 +1925,11 @@ def impala_shell_main():
     return
 
   if options.write_delimited:
-    delim = options.output_delimiter.decode('string-escape')
+    if isinstance(options.output_delimiter, str):
+      delim_sequence = bytearray(options.output_delimiter, 'utf-8')
+    else:
+      delim_sequence = options.output_delimiter
+    delim = delim_sequence.decode('unicode_escape')
     if len(delim) != 1:
       print("Illegal delimiter %s, the delimiter "
             "must be a 1-character string." % delim, file=sys.stderr)
@@ -1796,9 +1951,15 @@ def impala_shell_main():
           "mechanism (-l)", file=sys.stderr)
     raise FatalShellException()
 
+  start_msg = "Starting Impala Shell"
+
+  py_version_msg = "using Python {0}.{1}.{2}".format(
+    sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
+
   if options.use_kerberos:
     if options.verbose:
-      print("Starting Impala Shell using Kerberos authentication", file=sys.stderr)
+      kerb_msg = "with Kerberos authentication"
+      print("{0} {1} {2}".format(start_msg, kerb_msg, py_version_msg), file=sys.stderr)
       print("Using service name '%s'" % options.kerberos_service_name, file=sys.stderr)
     # Check if the user has a ticket in the credentials cache
     try:
@@ -1811,10 +1972,12 @@ def impala_shell_main():
       raise FatalShellException()
   elif options.use_ldap:
     if options.verbose:
-      print("Starting Impala Shell using LDAP-based authentication", file=sys.stderr)
+      ldap_msg = "with LDAP-based authentication"
+      print("{0} {1} {2}".format(start_msg, ldap_msg, py_version_msg), file=sys.stderr)
   else:
     if options.verbose:
-      print("Starting Impala Shell without Kerberos authentication", file=sys.stderr)
+      no_auth_msg = "with no authentication"
+      print("{0} {1} {2}".format(start_msg, no_auth_msg, py_version_msg), file=sys.stderr)
 
   options.ldap_password = None
   if options.use_ldap and options.ldap_password_cmd:
@@ -1858,9 +2021,11 @@ def impala_shell_main():
   # Non-interactive mode
   if options.query or options.query_file:
     # Impala shell will disable live_progress if non-interactive mode is detected.
-    options.live_progress = False
-    print("Warning: live_progress only applies to interactive shell sessions, "
-          "and is being skipped for now.", file=sys.stderr)
+    if options.live_progress:
+      if options.verbose:
+        print("Warning: live_progress only applies to interactive shell sessions, "
+              "and is being skipped for now.", file=sys.stderr)
+      options.live_progress = False
     if options.live_summary:
       print("Error: live_summary is available for interactive mode only.",
             file=sys.stderr)
