@@ -31,6 +31,18 @@ if [ "x${IMPALA_HOME}" == "x" ]; then
   exit 1
 fi
 
+# Detect whether IMPALA_HOME is a git repository. This is used below to allow extra
+# checks when building ext-py.
+pushd ${IMPALA_HOME}
+IS_GIT_CHECKOUT=false
+if git ls-files --error-unmatch > /dev/null 2>&1 ; then
+  IS_GIT_CHECKOUT=true
+  echo "IMPALA_HOME is a git repository"
+else
+  echo "IMPALA_HOME is not a git repository"
+fi;
+popd
+
 IMPALA_VERSION_INFO_FILE=${IMPALA_HOME}/bin/version.info
 
 if [ ! -f ${IMPALA_VERSION_INFO_FILE} ]; then
@@ -49,6 +61,9 @@ SHELL_HOME=${IMPALA_HOME}/shell
 BUILD_DIR=${SHELL_HOME}/build
 TARBALL_ROOT=${BUILD_DIR}/impala-shell-${VERSION}
 
+IMPALA_THRIFT_PY_VERSION="${IMPALA_THRIFT_VERSION}"
+THRIFT_GEN_PY_DIR="${SHELL_HOME}/gen-py"
+
 echo "Deleting all files in ${TARBALL_ROOT}/{gen-py,lib,ext-py}"
 rm -rf ${TARBALL_ROOT}/lib/* 2>&1 > /dev/null
 rm -rf ${TARBALL_ROOT}/gen-py/* 2>&1 > /dev/null
@@ -56,8 +71,8 @@ rm -rf ${TARBALL_ROOT}/ext-py/* 2>&1 > /dev/null
 mkdir -p ${TARBALL_ROOT}/lib
 mkdir -p ${TARBALL_ROOT}/ext-py
 
-rm -f ${SHELL_HOME}/gen-py/impala_build_version.py
-cat > ${SHELL_HOME}/gen-py/impala_build_version.py <<EOF
+rm -f ${THRIFT_GEN_PY_DIR}/impala_build_version.py
+cat > ${THRIFT_GEN_PY_DIR}/impala_build_version.py <<EOF
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -89,6 +104,16 @@ EOF
 # Building all eggs.
 echo "Building all external modules into eggs"
 for MODULE in ${SHELL_HOME}/ext-py/*; do
+  # Sometimes there are leftover module directories from version changes. If IMPALA_HOME
+  # is a git repository, then we can check if the module directory is tracked by git.
+  # If it is not tracked, skip building it. The downside of this check is that when
+  # adding a new directory, it won't build until added in git. This check does not apply
+  # when IMPALA_HOME is not a git repository (e.g. if building from a release tarball).
+  if ${IS_GIT_CHECKOUT} &&
+     ! git ls-files --error-unmatch ${MODULE} > /dev/null 2>&1 ; then
+    echo "WARNING: ${MODULE} is not tracked by the git repository, skipping..."
+    continue;
+  fi
   pushd ${MODULE} > /dev/null 2>&1
   echo "Cleaning up old build artifacts."
   rm -rf dist 2>&1 > /dev/null
@@ -96,7 +121,7 @@ for MODULE in ${SHELL_HOME}/ext-py/*; do
   echo "Creating an egg for ${MODULE}"
   if [[ "$MODULE" == *"/bitarray"* ]]; then
     # Need to use setuptools to build egg for bitarray module
-    python -c "import setuptools; execfile('setup.py')" -q bdist_egg clean
+    python -c "import setuptools; exec(open('setup.py').read())" -q bdist_egg
   else
     python setup.py -q bdist_egg clean
   fi
@@ -106,14 +131,18 @@ done
 
 # Copy all the shell files into the build dir
 # The location of python libs for thrift is different in rhel/centos/sles
-if [ -d ${THRIFT_HOME}/python/lib/python*/site-packages/thrift ]; then
-  cp -r ${THRIFT_HOME}/python/lib/python*/site-packages/thrift\
+
+THRIFT_PY_ROOT="${IMPALA_TOOLCHAIN_PACKAGES_HOME}/thrift-${IMPALA_THRIFT_PY_VERSION}"
+
+if [ -d ${THRIFT_PY_ROOT}/python/lib/python*/site-packages/thrift ]; then
+  cp -r ${THRIFT_PY_ROOT}/python/lib/python*/site-packages/thrift\
         ${TARBALL_ROOT}/lib
 else
-  cp -r ${THRIFT_HOME}/python/lib64/python*/site-packages/thrift\
+  cp -r ${THRIFT_PY_ROOT}/python/lib64/python*/site-packages/thrift\
         ${TARBALL_ROOT}/lib
 fi
-cp -r ${SHELL_HOME}/gen-py ${TARBALL_ROOT}
+
+cp -r ${THRIFT_GEN_PY_DIR} ${TARBALL_ROOT}
 cp ${SHELL_HOME}/option_parser.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/impala_shell_config_defaults.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/impala_client.py ${TARBALL_ROOT}/lib
@@ -122,6 +151,7 @@ cp ${SHELL_HOME}/ImpalaHttpClient.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/shell_exceptions.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/shell_output.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/pkg_resources.py ${TARBALL_ROOT}/lib
+cp ${SHELL_HOME}/cookie_util.py ${TARBALL_ROOT}/lib
 cp ${SHELL_HOME}/impala-shell ${TARBALL_ROOT}
 cp ${SHELL_HOME}/impala_shell.py ${TARBALL_ROOT}
 cp ${SHELL_HOME}/compatibility.py ${TARBALL_ROOT}
