@@ -108,6 +108,7 @@ import sys
 import tempfile
 from itertools import product
 from optparse import OptionParser
+from tests.common.environ import HIVE_MAJOR_VERSION
 from tests.util.test_file_parser import *
 from tests.common.test_dimensions import *
 
@@ -150,12 +151,27 @@ AVRO_SCHEMA_DIR = "avro_schemas"
 DEFAULT_FS=os.environ['DEFAULT_FS']
 IMPALA_SUPPORTED_INSERT_FORMATS = ['parquet', 'hbase', 'text', 'kudu']
 
+IMPALA_PARQUET_COMPRESSION_MAP = \
+  {
+    'uncompressed': 'NONE',
+    # UGLY: parquet/none always referred to the default compression, which is SNAPPY
+    # Maintain that for backwards compatibility.
+    'none': 'SNAPPY',
+    'snap': 'SNAPPY',
+    'gzip': 'GZIP',
+    'lz4': 'LZ4',
+    'zstd': 'ZSTD'
+  }
+
 COMPRESSION_TYPE = "SET mapred.output.compression.type=%s;"
 COMPRESSION_ENABLED = "SET hive.exec.compress.output=%s;"
 COMPRESSION_CODEC = "SET mapred.output.compression.codec=%s;"
+IMPALA_COMPRESSION_CODEC = "SET compression_codec=%s;"
 AVRO_COMPRESSION_CODEC = "SET avro.output.codec=%s;"
 SET_DYNAMIC_PARTITION_STATEMENT = "SET hive.exec.dynamic.partition=true;"
 SET_PARTITION_MODE_NONSTRICT_STATEMENT = "SET hive.exec.dynamic.partition.mode=nonstrict;"
+SET_MAX_DYNAMIC_PARTITIONS_STATEMENT = "SET hive.exec.max.dynamic.partitions=10000;\n"\
+    "SET hive.exec.max.dynamic.partitions.pernode=10000;"
 SET_HIVE_INPUT_FORMAT = "SET mapred.max.split.size=256000000;\n"\
                         "SET hive.input.format=org.apache.hadoop.hive.ql.io.%s;\n"
 SET_HIVE_HBASE_BULK_LOAD = "SET hive.hbase.bulk = true"
@@ -168,7 +184,6 @@ COMPRESSION_MAP = {'def': 'org.apache.hadoop.io.compress.DefaultCodec',
                    'gzip': 'org.apache.hadoop.io.compress.GzipCodec',
                    'bzip': 'org.apache.hadoop.io.compress.BZip2Codec',
                    'snap': 'org.apache.hadoop.io.compress.SnappyCodec',
-                   'lzo': 'com.hadoop.compression.lzo.LzopCodec',
                    'none': ''
                   }
 
@@ -185,9 +200,6 @@ FILE_FORMAT_MAP = {
   'orc': 'ORC',
   'parquet': 'PARQUET',
   'hudiparquet': 'HUDIPARQUET',
-  'text_lzo':
-    "\nINPUTFORMAT 'com.hadoop.mapred.DeprecatedLzoTextInputFormat'" +
-    "\nOUTPUTFORMAT 'org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat'",
   'avro': 'AVRO',
   'hbase': "'org.apache.hadoop.hive.hbase.HBaseStorageHandler'",
   'kudu': "KUDU",
@@ -221,7 +233,7 @@ WITH SERDEPROPERTIES (
   "{hbase_column_mapping}")
 {tbl_properties}{{hdfs_location}}"""
 
-KNOWN_EXPLORATION_STRATEGIES = ['core', 'pairwise', 'exhaustive', 'lzo']
+KNOWN_EXPLORATION_STRATEGIES = ['core', 'pairwise', 'exhaustive']
 
 def build_create_statement(table_template, table_name, db_name, db_suffix,
                            file_format, compression, hdfs_location,
@@ -229,8 +241,6 @@ def build_create_statement(table_template, table_name, db_name, db_suffix,
   create_stmt = ''
   if (force_reload):
     create_stmt += 'DROP TABLE IF EXISTS %s%s.%s;\n' % (db_name, db_suffix, table_name)
-  if compression == 'lzo':
-    file_format = '%s_%s' % (file_format, compression)
   # hbase / kudu tables are external, and not read from hdfs. We don't need an
   # hdfs_location.
   if file_format in ['hbase', 'kudu']:
@@ -312,7 +322,6 @@ def build_table_template(file_format, columns, partition_columns, row_format,
     # Kudu's test tables are managed.
     external = ""
 
-
   all_tblproperties = []
   for key, value in tblproperties.iteritems():
     all_tblproperties.append("'{0}' = '{1}'".format(key, value))
@@ -427,6 +436,11 @@ def build_codec_enabled_statement(codec):
   compression_enabled = 'false' if codec == 'none' else 'true'
   return COMPRESSION_ENABLED % compression_enabled
 
+
+def build_impala_parquet_codec_statement(codec):
+  parquet_codec = IMPALA_PARQUET_COMPRESSION_MAP[codec]
+  return IMPALA_COMPRESSION_CODEC % parquet_codec
+
 def build_insert_into_statement(insert, db_name, db_suffix, table_name, file_format,
                                 hdfs_path, for_impala=False):
   insert_hint = "/* +shuffle, clustered */" \
@@ -448,12 +462,13 @@ def build_insert_into_statement(insert, db_name, db_suffix, table_name, file_for
 
   statement = SET_PARTITION_MODE_NONSTRICT_STATEMENT + "\n"
   statement += SET_DYNAMIC_PARTITION_STATEMENT + "\n"
+  statement += SET_MAX_DYNAMIC_PARTITIONS_STATEMENT + "\n"
   statement += "set hive.auto.convert.join=true;\n"
 
   # For some reason (hive bug?) we need to have the CombineHiveInputFormat set
-  # for cases where we are compressing in bzip or lzo on certain tables that
+  # for cases where we are compressing in bzip on certain tables that
   # have multiple files.
-  if 'multi' in table_name and ('bzip' in db_suffix or 'lzo' in db_suffix):
+  if 'multi' in table_name and ('bzip' in db_suffix):
     statement += SET_HIVE_INPUT_FORMAT % "CombineHiveInputFormat"
   else:
     statement += SET_HIVE_INPUT_FORMAT % "HiveInputFormat"
@@ -467,15 +482,24 @@ def build_hbase_insert(db_name, db_suffix, table_name):
   return hbase_insert
 
 def build_insert(insert, db_name, db_suffix, file_format,
-                 codec, compression_type, table_name, hdfs_path, create_hive=False):
+                 codec, compression_type, table_name, hdfs_path, create_hive=False,
+                 for_impala=False):
   # HBASE inserts don't need the hive options to be set, and don't require and HDFS
   # file location, so they're handled separately.
   if file_format == 'hbase' and not create_hive:
     return build_hbase_insert(db_name, db_suffix, table_name)
-  output = build_codec_enabled_statement(codec) + "\n"
-  output += build_compression_codec_statement(codec, compression_type, file_format) + "\n"
+  output = ''
+  if not for_impala:
+    # If this is not for Impala, then generate the hive codec statements
+    output += build_codec_enabled_statement(codec) + "\n"
+    output += build_compression_codec_statement(codec, compression_type,
+                                                file_format) + "\n"
+  elif file_format == 'parquet':
+    # This is for Impala parquet, add the appropriate codec statement
+    output += build_impala_parquet_codec_statement(codec) + "\n"
   output += build_insert_into_statement(insert, db_name, db_suffix,
-                                        table_name, file_format, hdfs_path) + "\n"
+                                        table_name, file_format, hdfs_path,
+                                        for_impala) + "\n"
   return output
 
 def build_load_statement(load_template, db_name, db_suffix, table_name):
@@ -552,7 +576,7 @@ def eval_section(section_str):
 
 def generate_statements(output_name, test_vectors, sections,
                         schema_include_constraints, schema_exclude_constraints,
-                        schema_only_constraints):
+                        schema_only_constraints, convert_orc_to_full_acid):
   # TODO: This method has become very unwieldy. It has to be re-factored sooner than
   # later.
   # Parquet statements to be executed separately by Impala
@@ -625,6 +649,10 @@ def generate_statements(output_name, test_vectors, sections,
       else:
         create_kudu = None
 
+      if file_format == 'orc' and section["DEPENDENT_LOAD_ACID"]:
+        insert = None
+        insert_hive = eval_section(section["DEPENDENT_LOAD_ACID"])
+
       columns = eval_section(section['COLUMNS']).strip()
       partition_columns = section['PARTITION_COLUMNS'].strip()
       row_format = section['ROW_FORMAT'].strip()
@@ -641,6 +669,26 @@ def generate_statements(output_name, test_vectors, sections,
       force_reload = options.force_reload or (partition_columns and not alter) or \
           file_format == 'kudu'
 
+      # Empty tables (tables with no "LOAD" sections) are assumed to be used for insert
+      # testing. Since Impala currently only supports inserting into TEXT, PARQUET and
+      # HBASE we need to create these tables with a supported insert format.
+      create_file_format = file_format
+      create_codec = codec
+      if not (section['LOAD'] or section['DEPENDENT_LOAD']
+              or section['DEPENDENT_LOAD_HIVE'] or section['DEPENDENT_LOAD_ACID']):
+        create_codec = 'none'
+        create_file_format = file_format
+        if file_format not in IMPALA_SUPPORTED_INSERT_FORMATS:
+          create_file_format = 'text'
+
+      tblproperties = parse_table_properties(create_file_format, table_properties)
+      # ORC tables are full ACID by default.
+      if (convert_orc_to_full_acid and
+          HIVE_MAJOR_VERSION == 3 and
+          create_file_format == 'orc' and
+          'transactional' not in tblproperties):
+        tblproperties['transactional'] = 'true'
+
       hdfs_location = '{0}.{1}{2}'.format(db_name, table_name, db_suffix)
       # hdfs file names for functional datasets are stored
       # directly under /test-warehouse
@@ -648,30 +696,14 @@ def generate_statements(output_name, test_vectors, sections,
       # This needs to be done programmatically.
       if data_set == 'functional':
         hdfs_location = hdfs_location.split('.')[-1]
+      # Transactional tables need to be put under the 'managed' directory.
+      if is_transactional(tblproperties):
+        hdfs_location = os.path.join('managed', hdfs_location)
       data_path = os.path.join(options.hive_warehouse_dir, hdfs_location)
 
-      # Empty tables (tables with no "LOAD" sections) are assumed to be used for insert
-      # testing. Since Impala currently only supports inserting into TEXT, PARQUET and
-      # HBASE we need to create these tables with a supported insert format.
-      create_file_format = file_format
-      create_codec = codec
-      if not (section['LOAD'] or section['DEPENDENT_LOAD']
-              or section['DEPENDENT_LOAD_HIVE']):
-        create_codec = 'none'
-        create_file_format = file_format
-        if file_format not in IMPALA_SUPPORTED_INSERT_FORMATS:
-          create_file_format = 'text'
-
-      tblproperties = parse_table_properties(create_file_format, table_properties)
-
       output = impala_create
       if create_hive or file_format == 'hbase':
         output = hive_output
-      elif codec == 'lzo':
-        # Impala CREATE TABLE doesn't allow INPUTFORMAT.
-        output = hive_output
-      elif is_transactional(tblproperties):
-        output = hive_output
 
       # TODO: Currently, Kudu does not support partitioned tables via Impala.
       # If a CREATE_KUDU section was provided, assume it handles the partition columns
@@ -735,21 +767,7 @@ def generate_statements(output_name, test_vectors, sections,
       # moment, it assumes we're only using ALTER for partitioning the table.
       if alter and file_format not in ("hbase", "kudu"):
         use_db = 'USE {db_name};\n'.format(db_name=db)
-        if output == hive_output and codec == 'lzo':
-          # Hive ALTER TABLE ADD PARTITION doesn't handle null partitions, so
-          # we can't run the ALTER section in this case.
-          if options.force_reload:
-            # IMPALA-2278: Hive INSERT OVERWRITE won't clear out partition directories
-            # that weren't already added to the table. So, for force reload, manually
-            # delete the partition directories.
-            output.create.append(("DFS -rm -R {data_path};").format(
-              data_path=data_path))
-          else:
-            # If this is not a force reload use msck repair to add the partitions
-            # into the table.
-            output.create.append(use_db + 'msck repair table %s;' % (table_name))
-        else:
-          output.create.append(use_db + alter.format(table_name=table_name))
+        output.create.append(use_db + alter.format(table_name=table_name))
 
       # If the directory already exists in HDFS, assume that data files already exist
       # and skip loading the data. Otherwise, the data is generated using either an
@@ -769,8 +787,9 @@ def generate_statements(output_name, test_vectors, sections,
             hive_output.load.append(build_insert(insert_hive, db_name, db_suffix,
                 file_format, codec, compression_type, table_name, data_path))
           elif insert:
-            impala_load.load.append(build_insert_into_statement(insert, db_name,
-                db_suffix, table_name, file_format, data_path, for_impala=True))
+            impala_load.load.append(build_insert(insert, db_name, db_suffix,
+                file_format, codec, compression_type, table_name, data_path,
+                for_impala=True))
           else:
             print 'Empty parquet/kudu load for table %s. Skipping insert generation' \
               % table_name
@@ -807,7 +826,7 @@ def parse_schema_template_file(file_name):
   VALID_SECTION_NAMES = ['DATASET', 'BASE_TABLE_NAME', 'COLUMNS', 'PARTITION_COLUMNS',
                          'ROW_FORMAT', 'CREATE', 'CREATE_HIVE', 'CREATE_KUDU',
                          'DEPENDENT_LOAD', 'DEPENDENT_LOAD_KUDU', 'DEPENDENT_LOAD_HIVE',
-                         'LOAD', 'ALTER', 'HBASE_COLUMN_FAMILIES',
+                         'DEPENDENT_LOAD_ACID', 'LOAD', 'ALTER', 'HBASE_COLUMN_FAMILIES',
                          'TABLE_PROPERTIES', 'HBASE_REGION_SPLITS', 'HIVE_MAJOR_VERSION']
   return parse_test_file(file_name, VALID_SECTION_NAMES, skip_unknown_sections=False)
 
@@ -826,6 +845,10 @@ if __name__ == "__main__":
     test_vectors =\
         [TableFormatInfo.create_from_string(dataset, tf) for tf in table_formats]
 
+  # Hack to resolve IMPALA-9923.
+  # TODO: Try to remove it once we have HIVE-24145 in the dev environment.
+  convert_orc_to_full_acid = options.workload == 'functional-query'
+
   target_dataset = test_vectors[0].dataset
   print 'Target Dataset: ' + target_dataset
   dataset_load_dir = os.path.join(SQL_OUTPUT_DIR, target_dataset)
@@ -857,4 +880,5 @@ if __name__ == "__main__":
       parse_table_constraints(constraints_file)
   sections = parse_schema_template_file(schema_template_file)
   generate_statements('%s-%s' % (options.workload, options.exploration_strategy),
-      test_vectors, sections, include_constraints, exclude_constraints, only_constraints)
+      test_vectors, sections, include_constraints, exclude_constraints, only_constraints,
+      convert_orc_to_full_acid)
