@@ -22,23 +22,24 @@
 from collections import defaultdict
 import json
 import logging
+import os
 import re
 import requests
+import subprocess
+from datetime import datetime
 from time import sleep, time
 
 from tests.common.impala_connection import create_connection, create_ldap_connection
-from TCLIService import TCLIService
 from thrift.transport.TSocket import TSocket
 from thrift.transport.TTransport import TBufferedTransport
-from thrift.protocol import TBinaryProtocol, TCompactProtocol, TProtocol
-from thrift.TSerialization import deserialize
-from RuntimeProfile.ttypes import TRuntimeProfileTree
-import base64
-import zlib
 
 LOG = logging.getLogger('impala_service')
 LOG.setLevel(level=logging.DEBUG)
 
+WEBSERVER_USERNAME = os.environ.get('IMPALA_WEBSERVER_USERNAME', None)
+WEBSERVER_PASSWORD = os.environ.get('IMPALA_WEBSERVER_PASSWORD', None)
+
+
 # Base class for all Impala services
 # TODO: Refactor the retry/timeout logic into a common place.
 class BaseImpalaService(object):
@@ -51,6 +52,9 @@ class BaseImpalaService(object):
       self.webserver_interface = hostname
     self.webserver_port = webserver_port
     self.webserver_certificate_file = webserver_certificate_file
+    self.webserver_username_password = None
+    if WEBSERVER_USERNAME is not None and WEBSERVER_PASSWORD is not None:
+      self.webserver_username_password = (WEBSERVER_USERNAME, WEBSERVER_PASSWORD)
 
   def open_debug_webpage(self, page_name, timeout=10, interval=1):
     start_time = time()
@@ -62,7 +66,8 @@ class BaseImpalaService(object):
           protocol = "https"
         url = "%s://%s:%d/%s" % \
             (protocol, self.webserver_interface, int(self.webserver_port), page_name)
-        return requests.get(url, verify=self.webserver_certificate_file)
+        return requests.get(url, verify=self.webserver_certificate_file,
+            auth=self.webserver_username_password)
       except Exception as e:
         LOG.info("Debug webpage not yet available: %s", str(e))
       sleep(interval)
@@ -75,6 +80,13 @@ class BaseImpalaService(object):
     """Returns the json for the given Impala debug webpage, eg. '/queries'"""
     return json.loads(self.read_debug_webpage(page_name + "?json"))
 
+  def dump_debug_webpage_json(self, page_name, filename):
+    """Dumps the json for a given Impalad debug webpage to the specified file.
+       Prints the JSON with indenting to be somewhat human readable."""
+    debug_json = self.get_debug_webpage_json(page_name)
+    with open(filename, "w") as json_out:
+      json.dump(debug_json, json_out, indent=2)
+
   def get_metric_value(self, metric_name, default_value=None):
     """Returns the value of the the given metric name from the Impala debug webpage"""
     return self.get_metric_values([metric_name], [default_value])[0]
@@ -118,20 +130,81 @@ class BaseImpalaService(object):
             (metric_name, expected_value, value))
       LOG.info("Sleeping %ds before next retry." % interval)
       sleep(interval)
-    assert 0, 'Metric value %s did not reach value %s in %ss\nDumping impalad debug ' \
-              'pages:\nmemz: %s\nmetrics: %s\nqueries: %s\nsessions: %s\nthreadz: %s\n '\
-              'rpcz: %s' % \
-              (metric_name, expected_value, timeout,
-               json.dumps(self.read_debug_webpage('memz?json')),
-               json.dumps(self.read_debug_webpage('metrics?json')),
-               json.dumps(self.read_debug_webpage('queries?json')),
-               json.dumps(self.read_debug_webpage('sessions?json')),
-               json.dumps(self.read_debug_webpage('threadz?json')),
-               json.dumps(self.read_debug_webpage('rpcz?json')))
+
+    LOG.info("Metric {0} did not reach value {1} in {2}s. Failing...".format(metric_name,
+        expected_value, timeout))
+    self.__metric_timeout_assert(metric_name, expected_value, timeout)
+
+  def __request_minidump(self, pid):
+    """
+    Impala processes (impalad, catalogd, statestored) have a signal handler for
+    SIGUSR1 that dumps a minidump. This sends a SIGUSR1 to the specified pid
+    to trigger a minidump. Note that this will not work for processes that don't
+    implement a handler for SIGUSR1. This does not wait for the minidump to be
+    generated, so callers will need to consider this.
+    """
+    cmd = ["kill", "-SIGUSR1", pid]
+    subprocess.check_output(cmd)
+
+  def __metric_timeout_assert(self, metric_name, expected_value, timeout):
+    """
+    Helper function to dump diagnostic information for debugging a metric timeout and
+    then assert.
+    """
+    impala_home = os.environ["IMPALA_HOME"]
+    log_dir = os.environ["IMPALA_LOGS_DIR"]
+
+    # Create a diagnostic directory to contain all the relevent information
+    datetime_string = datetime.now().strftime("%Y%m%d_%H:%M:%S")
+    diag_dir = os.path.join(log_dir, "metric_timeout_diags_{0}".format(datetime_string))
+    if not os.path.exists(diag_dir):
+      os.makedirs(diag_dir)
+
+    # Read debug pages from the Web UI and dump them to files in the diagnostic
+    # directory
+    json_dir = os.path.join(diag_dir, "json")
+    if not os.path.exists(json_dir):
+      os.makedirs(json_dir)
+    json_diag_string = "Dumping debug webpages in JSON format...\n"
+    debug_pages = ["memz", "metrics", "queries", "sessions", "threadz", "rpcz"]
+    for debug_page in debug_pages:
+      json_filename = os.path.join(json_dir, "{0}.json".format(debug_page))
+      self.dump_debug_webpage_json(debug_page, json_filename)
+      json_filename_rewritten = json_filename.replace(impala_home, "$IMPALA_HOME")
+      json_diag_string += \
+          "Dumped {0} JSON to {1}\n".format(debug_page, json_filename_rewritten)
+
+    # Requests a minidump for each running impalad/catalogd. The minidump will be
+    # written to the processes's minidump_path. For simplicity, we leave it there,
+    # as it will be preserved along with everything else in the log directory.
+    impalad_pids = subprocess.check_output(["pgrep", "impalad"]).split("\n")[:-1]
+    catalogd_pids = subprocess.check_output(["pgrep", "-f", "catalogd"]).split("\n")[:-1]
+    minidump_diag_string = "Dumping minidumps for impalads/catalogds...\n"
+    for pid in impalad_pids:
+      self.__request_minidump(pid)
+      minidump_diag_string += "Dumped minidump for Impalad PID {0}\n".format(pid)
+    for pid in catalogd_pids:
+      self.__request_minidump(pid)
+      minidump_diag_string += "Dumped minidump for Catalogd PID {0}\n".format(pid)
+
+    # This is not a critical path, so sleep 30 seconds to be sure that the minidumps
+    # have been written.
+    sleep(30)
+
+    # Now, fire the assert with the information about the dumps. This provides
+    # information about where to find the json files and other diagnostics.
+    # This is in the logs directory, so it should be packed up along with everything
+    # else for automated jobs.
+    assert_string = \
+        "Metric {0} did not reach value {1} in {2}s.\n".format(metric_name,
+            expected_value, timeout)
+    assert_string += json_diag_string
+    assert_string += minidump_diag_string
+    assert 0, assert_string
 
   def get_catalog_object_dump(self, object_type, object_name):
     """ Gets the web-page for the given 'object_type' and 'object_name'."""
-    return self.read_debug_webpage('catalog_object?object_type=%s&object_name=%s' %\
+    return self.read_debug_webpage('catalog_object?object_type=%s&object_name=%s' %
         (object_type, object_name))
 
   def get_catalog_objects(self, excludes=['_impala_builtins']):
@@ -153,21 +226,22 @@ class BaseImpalaService(object):
     return objects
 
   def extract_catalog_object_version(self, thrift_txt):
-    """ Extracts and returns the version of the catalog object's 'thrift_txt' representation."""
+    """ Extracts and returns the version of the catalog object's 'thrift_txt'
+        representation."""
     result = re.search(r'catalog_version \(i64\) = (\d+)', thrift_txt)
     assert result, 'Unable to find catalog version in object: ' + thrift_txt
     return int(result.group(1))
 
+
 # Allows for interacting with an Impalad instance to perform operations such as creating
 # new connections or accessing the debug webpage.
 class ImpaladService(BaseImpalaService):
   def __init__(self, hostname, webserver_interface="", webserver_port=25000,
-      beeswax_port=21000, be_port=22000, krpc_port=27000, hs2_port=21050,
+      beeswax_port=21000, krpc_port=27000, hs2_port=21050,
       hs2_http_port=28000, webserver_certificate_file=""):
     super(ImpaladService, self).__init__(
         hostname, webserver_interface, webserver_port, webserver_certificate_file)
     self.beeswax_port = beeswax_port
-    self.be_port = be_port
     self.krpc_port = krpc_port
     self.hs2_port = hs2_port
     self.hs2_http_port = hs2_http_port
@@ -204,16 +278,25 @@ class ImpaladService(BaseImpalaService):
       groups[backend['executor_groups']].append(backend['krpc_address'])
     return groups
 
+  def get_queries_json(self):
+    """Return the full JSON from the /queries page."""
+    return json.loads(self.read_debug_webpage('queries?json', timeout=30, interval=1))
+
   def get_query_locations(self):
     # Returns a dictionary of the format <host_address, num_of_queries_running_there>
-    result = json.loads(self.read_debug_webpage('queries?json', timeout=30, interval=1))
+    result = self.get_queries_json()
     if result['query_locations'] is not None:
       return dict([(loc["location"], loc["count"]) for loc in result['query_locations']])
     return None
 
   def get_in_flight_queries(self, timeout=30, interval=1):
+    """Returns the number of in flight queries."""
+    return self.get_queries_json()['in_flight_queries']
+
+  def get_completed_queries(self, timeout=30, interval=1):
+    """Returns the number of completed queries."""
     result = json.loads(self.read_debug_webpage('queries?json', timeout, interval))
-    return result['in_flight_queries']
+    return result['completed_queries']
 
   def _get_pool_counter(self, pool_name, counter_name, timeout=30, interval=1):
     """Returns the value of the field 'counter_name' in pool 'pool_name' or 0 if the pool
@@ -275,7 +358,7 @@ class ImpaladService(BaseImpalaService):
         LOG.info("num_known_live_backends has reached value: %s" % value)
         return value
       else:
-        LOG.info("Waiting for num_known_live_backends=%s. Current value: %s" %\
+        LOG.info("Waiting for num_known_live_backends=%s. Current value: %s" %
             (expected_value, value))
       sleep(1)
     assert 0, 'num_known_live_backends did not reach expected value in time'
@@ -302,7 +385,8 @@ class ImpaladService(BaseImpalaService):
       try:
         query_state = client.get_state(query_handle)
       except Exception as e:
-        pass
+        LOG.error("Exception while getting state of query {0}\n{1}".format(
+            query_handle.get_handle().id, str(e)))
       if query_state == target_state:
         return
       sleep(interval)
@@ -323,8 +407,8 @@ class ImpaladService(BaseImpalaService):
         query_status = self.get_query_status(query_id)
         if query_status is None:
           assert False, "Could not find 'Query Status' section in profile of "\
-                    "query with id %s:\n%s" % (query_id)
-      except Exception as e:
+              "query with id %s:\n%s" % (query_id)
+      except Exception:
         pass
       if expected_content in query_status:
         return True
@@ -374,6 +458,7 @@ class ImpaladService(BaseImpalaService):
       LOG.info(e)
       return False
 
+
 # Allows for interacting with the StateStore service to perform operations such as
 # accessing the debug webpage.
 class StateStoredService(BaseImpalaService):
@@ -409,3 +494,10 @@ class CatalogdService(BaseImpalaService):
         LOG.info('Catalogd version not yet available.')
       sleep(interval)
     assert False, 'Catalog version not ready in expected time.'
+
+
+class AdmissiondService(BaseImpalaService):
+  def __init__(self, hostname, webserver_interface, webserver_port,
+      webserver_certificate_file):
+    super(AdmissiondService, self).__init__(
+        hostname, webserver_interface, webserver_port, webserver_certificate_file)
