@@ -27,17 +27,18 @@ import pwd
 import pytest
 import re
 import requests
-import shutil
 import socket
 import subprocess
-import tempfile
 import time
 from functools import wraps
 from getpass import getuser
 from random import choice
 from subprocess import check_call
 from tests.common.base_test_suite import BaseTestSuite
-from tests.common.environ import HIVE_MAJOR_VERSION, MANAGED_WAREHOUSE_DIR
+from tests.common.environ import (
+    HIVE_MAJOR_VERSION,
+    MANAGED_WAREHOUSE_DIR,
+    EXTERNAL_WAREHOUSE_DIR)
 from tests.common.errors import Timeout
 from tests.common.impala_connection import create_connection
 from tests.common.impala_service import ImpaladService
@@ -63,6 +64,7 @@ from tests.util.filesystem_utils import (
     IS_S3,
     IS_ABFS,
     IS_ADLS,
+    IS_GCS,
     IS_HDFS,
     S3_BUCKET_NAME,
     S3GUARD_ENABLED,
@@ -197,6 +199,7 @@ class ImpalaTestSuite(BaseTestSuite):
     #     ABFS: uses the HDFS CLI
     #     ADLS: uses a mixture of azure-data-lake-store-python and the HDFS CLI (TODO:
     #           this should completely switch to the HDFS CLI once we test it)
+    #     GCS:  uses the HDFS CLI
     #
     # 'hdfs_client' is a HDFS-specific client library, and it only works when running on
     # HDFS. When using 'hdfs_client', the test must be skipped on everything other than
@@ -217,6 +220,9 @@ class ImpalaTestSuite(BaseTestSuite):
       cls.filesystem_client = HadoopFsCommandLineClient("ABFS")
     elif IS_ADLS:
       cls.filesystem_client = ADLSClient(ADLS_STORE_NAME)
+    elif IS_GCS:
+      # GCS is implemented via HDFS command line client
+      cls.filesystem_client = HadoopFsCommandLineClient("GCS")
 
     # Override the shell history path so that commands run by any tests
     # don't write any history into the developer's file.
@@ -435,7 +441,8 @@ class ImpalaTestSuite(BaseTestSuite):
         "IMPALA_HOME",
         "INTERNAL_LISTEN_HOST",
         "INTERNAL_LISTEN_IP",
-        "MANAGED_WAREHOUSE_DIR"])
+        "MANAGED_WAREHOUSE_DIR",
+        "EXTERNAL_WAREHOUSE_DIR"])
     repl.update({
         '$SECONDARY_FILESYSTEM': os.environ.get("SECONDARY_FILESYSTEM", ""),
         '$USER': getuser()})
@@ -487,7 +494,7 @@ class ImpalaTestSuite(BaseTestSuite):
       actual values are easily compared.
     """
     replace_filenames_with_placeholder = True
-    for section_name in ('RESULTS', 'DBAPI_RESULTS', 'ERRORS'):
+    for section_name in ('RESULTS', 'ERRORS'):
       if section_name in test_section:
         if "$NAMENODE" in test_section[section_name]:
           replace_filenames_with_placeholder = False
@@ -500,18 +507,14 @@ class ImpalaTestSuite(BaseTestSuite):
                                               INTERNAL_LISTEN_HOST) \
                                      .replace('$INTERNAL_LISTEN_IP', INTERNAL_LISTEN_IP) \
                                      .replace('$MANAGED_WAREHOUSE_DIR',
-                                              MANAGED_WAREHOUSE_DIR)
+                                              MANAGED_WAREHOUSE_DIR) \
+                                     .replace('$EXTERNAL_WAREHOUSE_DIR',
+                                              EXTERNAL_WAREHOUSE_DIR)
         if use_db:
           test_section[section_name] = test_section[section_name].replace('$DATABASE', use_db)
     result_section, type_section = 'RESULTS', 'TYPES'
     if vector.get_value('protocol').startswith('hs2'):
       # hs2 or hs2-http
-      if 'DBAPI_RESULTS' in test_section:
-        assert 'RESULTS' in test_section,\
-            "Base RESULTS section must always be included alongside DBAPI_RESULTS"
-        # In some cases Impyla (the HS2 dbapi client) is expected to return different
-        # results, so use the dbapi-specific section if present.
-        result_section = 'DBAPI_RESULTS'
       if 'HS2_TYPES' in test_section:
         assert 'TYPES' in test_section,\
             "Base TYPES section must always be included alongside HS2_TYPES"
@@ -644,9 +647,6 @@ class ImpalaTestSuite(BaseTestSuite):
             '-- QUERY or HIVE_QUERY section.\n%s') %\
             (test_file_name, pprint.pformat(test_section))
 
-      if 'SETUP' in test_section:
-        self.execute_test_case_setup(test_section['SETUP'], table_format_info)
-
       # TODO: support running query tests against different scale factors
       query = QueryTestSectionReader.build_query(
           self.__do_replacements(query_section, use_db=use_db, extra=test_file_vars))
@@ -767,31 +767,9 @@ class ImpalaTestSuite(BaseTestSuite):
             return lineage
     return ""
 
-  def execute_test_case_setup(self, setup_section, table_format):
-    """
-    Executes a test case 'SETUP' section
-
-    The test case 'SETUP' section is mainly used for insert tests. These tests need to
-    have some actions performed before each test case to ensure the target tables are
-    empty. The current supported setup actions:
-    RESET <table name> - Drop and recreate the table
-    DROP PARTITIONS <table name> - Drop all partitions from the table
-    """
-    setup_section = QueryTestSectionReader.build_query(setup_section)
-    for row in setup_section.split('\n'):
-      row = row.lstrip()
-      if row.startswith('RESET'):
-        db_name, table_name = QueryTestSectionReader.get_table_name_components(\
-          table_format, row.split('RESET')[1])
-        self.__reset_table(db_name, table_name)
-        self.client.execute("invalidate metadata " + db_name + "." + table_name)
-      elif row.startswith('DROP PARTITIONS'):
-        db_name, table_name = QueryTestSectionReader.get_table_name_components(\
-          table_format, row.split('DROP PARTITIONS')[1])
-        self.__drop_partitions(db_name, table_name)
-        self.client.execute("invalidate metadata " + db_name + "." + table_name)
-      else:
-        assert False, 'Unsupported setup command: %s' % row
+  @staticmethod
+  def get_db_name_from_format(table_format, scale_factor=''):
+    return QueryTestSectionReader.get_db_name(table_format, scale_factor)
 
   @classmethod
   def change_database(cls, impala_client, table_format=None,
@@ -861,6 +839,17 @@ class ImpalaTestSuite(BaseTestSuite):
   def execute_query(self, query, query_options=None):
     return self.__execute_query(self.client, query, query_options)
 
+  def exec_and_time(self, query, query_options=None, impalad=0):
+    """Executes a given query on the given impalad and returns the time taken in
+    millisecondsas seen by the client."""
+    client = self.create_client_for_nth_impalad(impalad)
+    if query_options is not None:
+      client.set_configuration(query_options)
+    start_time = int(round(time.time() * 1000))
+    client.execute(query)
+    end_time = int(round(time.time() * 1000))
+    return end_time - start_time
+
   def execute_query_using_client(self, client, query, vector):
     self.change_database(client, vector.get_value('table_format'))
     query_options = vector.get_value('exec_option')
@@ -921,25 +910,12 @@ class ImpalaTestSuite(BaseTestSuite):
       assert False, 'Test file not found: %s' % file_name
     return parse_query_test_file(test_file_path, valid_section_names, encoding=encoding)
 
-  def __drop_partitions(self, db_name, table_name):
-    """Drops all partitions in the given table"""
-    for partition in self.hive_client.get_partition_names(db_name, table_name, -1):
-      assert self.hive_client.drop_partition_by_name(db_name, table_name, \
-          partition, True), 'Could not drop partition: %s' % partition
-
   @classmethod
   def __execute_query(cls, impalad_client, query, query_options=None, user=None):
     """Executes the given query against the specified Impalad"""
     if query_options is not None: impalad_client.set_configuration(query_options)
     return impalad_client.execute(query, user=user)
 
-  def __reset_table(self, db_name, table_name):
-    """Resets a table (drops and recreates the table)"""
-    table = self.hive_client.get_table(db_name, table_name)
-    assert table is not None
-    self.hive_client.drop_table(db_name, table_name, True)
-    self.hive_client.create_table(table)
-
   def clone_table(self, src_tbl, dst_tbl, recover_partitions, vector):
     src_loc = self._get_table_location(src_tbl, vector)
     self.client.execute("create external table {0} like {1} location '{2}'"\
@@ -951,7 +927,7 @@ class ImpalaTestSuite(BaseTestSuite):
     """Returns True if 'a' and 'b' are within 'diff_perc' percent of each other,
     False otherwise. 'diff_perc' must be a float in [0,1]."""
     if a == b: return True # Avoid division by 0
-    assert abs(a - b) / float(max(a,b)) <= diff_perc
+    assert abs(a - b) / float(max(abs(a), abs(b))) <= diff_perc
 
   def _get_table_location(self, table_name, vector):
     """ Returns the HDFS location of the table """
@@ -970,54 +946,33 @@ class ImpalaTestSuite(BaseTestSuite):
     Run a statement in Hive, returning stdout if successful and throwing
     RuntimeError(stderr) if not.
     """
-    # When HiveServer2 is configured to use "local" mode (i.e., MR jobs are run
-    # in-process rather than on YARN), Hadoop's LocalDistributedCacheManager has a
-    # race, wherein it tires to localize jars into
-    # /tmp/hadoop-$USER/mapred/local/<millis>. Two simultaneous Hive queries
-    # against HS2 can conflict here. Weirdly LocalJobRunner handles a similar issue
-    # (with the staging directory) by appending a random number. To overcome this,
-    # in the case that HS2 is on the local machine (which we conflate with also
-    # running MR jobs locally), we move the temporary directory into a unique
-    # directory via configuration. This workaround can be removed when
-    # https://issues.apache.org/jira/browse/MAPREDUCE-6441 is resolved.
-    # A similar workaround is used in bin/load-data.py.
-    tmpdir = None
-    beeline_opts = []
-    if pytest.config.option.hive_server2.startswith("localhost:"):
-      tmpdir = tempfile.mkdtemp(prefix="impala-tests-")
-      beeline_opts += ['--hiveconf', 'mapreduce.cluster.local.dir={0}'.format(tmpdir)]
-    try:
-      # Remove HADOOP_CLASSPATH from environment. Beeline doesn't need it,
-      # and doing so avoids Hadoop 3's classpath de-duplication code from
-      # placing $HADOOP_CONF_DIR too late in the classpath to get the right
-      # log4j configuration file picked up. Some log4j configuration files
-      # in Hadoop's jars send logging to stdout, confusing Impala's test
-      # framework.
-      env = os.environ.copy()
-      env.pop("HADOOP_CLASSPATH", None)
-      call = subprocess.Popen(
-          ['beeline',
-           '--outputformat=csv2',
-           '-u', 'jdbc:hive2://' + pytest.config.option.hive_server2,
-           '-n', username or getuser(),
-           '-e', stmt] + beeline_opts,
-          stdout=subprocess.PIPE,
-          stderr=subprocess.PIPE,
-          # Beeline in Hive 2.1 will read from stdin even when "-e"
-          # is specified; explicitly make sure there's nothing to
-          # read to avoid hanging, especially when running interactively
-          # with py.test.
-          stdin=file("/dev/null"),
-          env=env)
-      (stdout, stderr) = call.communicate()
-      call.wait()
-      if call.returncode != 0:
-        raise RuntimeError(stderr)
-      return stdout
-    finally:
-      # IMPALA-8315: removing the directory may race with Hive's own cleanup and cause
-      # errors.
-      if tmpdir is not None: shutil.rmtree(tmpdir, ignore_errors=True)
+    # Remove HADOOP_CLASSPATH from environment. Beeline doesn't need it,
+    # and doing so avoids Hadoop 3's classpath de-duplication code from
+    # placing $HADOOP_CONF_DIR too late in the classpath to get the right
+    # log4j configuration file picked up. Some log4j configuration files
+    # in Hadoop's jars send logging to stdout, confusing Impala's test
+    # framework.
+    env = os.environ.copy()
+    env.pop("HADOOP_CLASSPATH", None)
+    call = subprocess.Popen(
+        ['beeline',
+         '--outputformat=csv2',
+         '-u', 'jdbc:hive2://' + pytest.config.option.hive_server2,
+         '-n', username or getuser(),
+         '-e', stmt],
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE,
+        # Beeline in Hive 2.1 will read from stdin even when "-e"
+        # is specified; explicitly make sure there's nothing to
+        # read to avoid hanging, especially when running interactively
+        # with py.test.
+        stdin=file("/dev/null"),
+        env=env)
+    (stdout, stderr) = call.communicate()
+    call.wait()
+    if call.returncode != 0:
+      raise RuntimeError(stderr)
+    return stdout
 
   def hive_partition_names(self, table_name):
     """Find the names of the partitions of a table, as Hive sees them.
@@ -1040,10 +995,10 @@ class ImpalaTestSuite(BaseTestSuite):
       tf_dimensions = ImpalaTestDimension('table_format', *table_formats)
     else:
       tf_dimensions = load_table_info_dimension(cls.get_workload(), exploration_strategy)
-    # If 'skip_hbase' is specified or the filesystem is isilon, s3 or local, we don't
-    # need the hbase dimension.
+    # If 'skip_hbase' is specified or the filesystem is isilon, s3, GCS(gs) or local,
+    # we don't need the hbase dimension.
     if pytest.config.option.skip_hbase or TARGET_FILESYSTEM.lower() \
-        in ['s3', 'isilon', 'local', 'abfs', 'adls']:
+        in ['s3', 'isilon', 'local', 'abfs', 'adls', 'gs']:
       for tf_dimension in tf_dimensions:
         if tf_dimension.value.file_format == "hbase":
           tf_dimensions.remove(tf_dimension)
@@ -1102,6 +1057,27 @@ class ImpalaTestSuite(BaseTestSuite):
                     actual_state))
     return actual_state
 
+  def wait_for_progress(self, handle, expected_progress, timeout, client=None):
+    """Waits for the given query handle to reach expected progress rate"""
+    if client is None: client = self.client
+    start_time = time.time()
+    summary = client.get_exec_summary(handle)
+    while time.time() - start_time < timeout and \
+        self.__get_query_progress_rate(summary.progress) <= expected_progress:
+      summary = client.get_exec_summary(handle)
+      time.sleep(0.5)
+    actual_progress = self.__get_query_progress_rate(summary.progress)
+    if actual_progress <= expected_progress:
+      raise Timeout("query {0} did not reach the expected progress {1}, "
+                    "current progress {2}".format(handle.get_handle().id,
+                    expected_progress, actual_progress))
+    return actual_progress
+
+  def __get_query_progress_rate(self, progress):
+    if progress is None:
+      return 0
+    return float(progress.num_completed_scan_ranges) / progress.total_scan_ranges
+
   def wait_for_db_to_appear(self, db_name, timeout_s):
     """Wait until the database with 'db_name' is present in the impalad's local catalog.
     Fail after timeout_s if the doesn't appear."""
