@@ -240,8 +240,8 @@ def verify_query_result_is_subset(expected_results, actual_results):
   """Check whether the results in expected_results are a subset of the results in
   actual_results. This uses set semantics, i.e. any duplicates are ignored."""
   expected_literals, expected_non_literals = expected_results.separate_rows()
-  expected_literal_strings = set([str(row) for row in expected_literals])
-  actual_literal_strings = set([str(row) for row in actual_results.rows])
+  expected_literal_strings = set([unicode(row) for row in expected_literals])
+  actual_literal_strings = set([unicode(row) for row in actual_results.rows])
   # Expected literal strings must all be present in the actual strings.
   assert expected_literal_strings <= actual_literal_strings
   # Expected patterns must be present in the actual strings.
@@ -251,18 +251,18 @@ def verify_query_result_is_subset(expected_results, actual_results):
       if actual_row == expected_row:
         matched = True
         break
-    assert matched, "Could not find expected row {0} in actual rows:\n{1}".format(
-        str(expected_row), str(actual_results))
+    assert matched, u"Could not find expected row {0} in actual rows:\n{1}".format(
+        unicode(expected_row), unicode(actual_results))
 
 def verify_query_result_is_superset(expected_results, actual_results):
   """Check whether the results in expected_results are a superset of the results in
   actual_results. This uses set semantics, i.e. any duplicates are ignored."""
   expected_literals, expected_non_literals = expected_results.separate_rows()
-  expected_literal_strings = set([str(row) for row in expected_literals])
+  expected_literal_strings = set([unicode(row) for row in expected_literals])
   # Check that all actual rows are present in either expected_literal_strings or
   # expected_non_literals.
   for actual_row in actual_results.rows:
-    if str(actual_row) in expected_literal_strings:
+    if unicode(actual_row) in expected_literal_strings:
       # Matched to a literal string
       continue
     matched = False
@@ -270,8 +270,8 @@ def verify_query_result_is_superset(expected_results, actual_results):
       if actual_row == expected_row:
         matched = True
         break
-    assert matched, "Could not find actual row {0} in expected rows:\n{1}".format(
-        str(actual_row), str(expected_results))
+    assert matched, u"Could not find actual row {0} in expected rows:\n{1}".format(
+        unicode(actual_row), unicode(expected_results))
 
 def verify_query_result_is_equal(expected_results, actual_results):
   assert_args_not_none(expected_results, actual_results)
@@ -279,8 +279,8 @@ def verify_query_result_is_equal(expected_results, actual_results):
 
 def verify_query_result_is_not_in(expected_results, actual_results):
   assert_args_not_none(expected_results, actual_results)
-  expected_set = set(map(str, expected_results.rows))
-  actual_set = set(map(str, actual_results.rows))
+  expected_set = set(map(unicode, expected_results.rows))
+  actual_set = set(map(unicode, actual_results.rows))
   assert expected_set.isdisjoint(actual_set)
 
 # Global dictionary that maps the verification type to appropriate verifier.
@@ -357,6 +357,15 @@ def verify_raw_results(test_section, exec_result, file_format, result_section,
   expected_results = None
   if result_section in test_section:
     expected_results = remove_comments(test_section[result_section])
+    if isinstance(expected_results, str):
+      # Always convert 'str' to 'unicode' since pytest will fail to report assertion
+      # failures when any 'str' values contain non-ascii bytes (IMPALA-10419).
+      try:
+        expected_results = expected_results.decode('utf-8')
+      except UnicodeDecodeError as e:
+        LOG.info("Illegal UTF-8 characters in expected results: {0}\n{1}".format(
+            expected_results, e))
+        assert False
   else:
     assert 'ERRORS' not in test_section, "'ERRORS' section must have accompanying 'RESULTS' section"
     LOG.info("No results found. Skipping verification")
@@ -442,15 +451,19 @@ def verify_raw_results(test_section, exec_result, file_format, result_section,
   if verifier and verifier.upper() == 'VERIFY_IS_EQUAL_SORTED':
     order_matters = False
   expected_results_list = []
+  is_raw_string = 'RAW_STRING' in test_section
   if 'MULTI_LINE' in test_section:
-    expected_results_list = map(lambda s: s.replace('\n', '\\n'),
-        re.findall(r'\[(.*?)\]', expected_results, flags=re.DOTALL))
+    expected_results_list = re.findall(r'\[(.*?)\]', expected_results, flags=re.DOTALL)
+    if not is_raw_string:
+      # Needs escaping
+      expected_results_list = map(lambda s: s.replace('\n', '\\n'), expected_results_list)
   else:
     expected_results_list = split_section_lines(expected_results)
   expected = QueryTestResult(expected_results_list, expected_types,
       actual_labels, order_matters)
-  actual = QueryTestResult(parse_result_rows(exec_result), actual_types,
-      actual_labels, order_matters)
+  actual = QueryTestResult(
+      parse_result_rows(exec_result, escape_strings=(not is_raw_string)),
+      actual_types, actual_labels, order_matters)
   assert verifier in VERIFIER_MAP.keys(), "Unknown verifier: " + verifier
   try:
     VERIFIER_MAP[verifier](expected, actual)
@@ -470,7 +483,8 @@ def create_query_result(exec_result, order_matters=False):
   return QueryTestResult(data, exec_result.column_types, exec_result.column_labels,
                          order_matters)
 
-def parse_result_rows(exec_result):
+
+def parse_result_rows(exec_result, escape_strings=True):
   """
   Parses a query result set and transforms it to the format used by the query test files
   """
@@ -490,9 +504,18 @@ def parse_result_rows(exec_result):
     new_cols = list()
     for i in xrange(len(cols)):
       if col_types[i] in ['STRING', 'CHAR', 'VARCHAR']:
-        col = cols[i].encode('unicode_escape')
-        # Escape single quotes to match .test file format.
-        col = col.replace("'", "''")
+        col = cols[i]
+        if isinstance(col, str):
+          try:
+            col = col.decode('utf-8')
+          except UnicodeDecodeError as e:
+            LOG.info("Illegal UTF-8 characters in actual results: {0}\n{1}".format(
+                col, e))
+            assert False
+        if escape_strings:
+          col = col.encode('unicode_escape').decode('utf-8')
+          # Escape single quotes to match .test file format.
+          col = col.replace("'", "''")
         new_cols.append("'%s'" % col)
       else:
         new_cols.append(cols[i])
@@ -505,7 +528,7 @@ def parse_result_rows(exec_result):
 # Currently, the only implemented function is SUM and only integers are supported.
 AGGREGATION_PREFIX_PATTERN = 'aggregation\('
 AGGREGATION_PREFIX = re.compile(AGGREGATION_PREFIX_PATTERN)
-AGGREGATION_SYNTAX_MATCH_PATTERN = 'aggregation\((\w+)[ ]*,[ ]*([^)]+)\):[ ]*(\d+)'
+AGGREGATION_SYNTAX_MATCH_PATTERN = 'aggregation\((\w+)[ ]*,[ ]*([^)]+)\)([:><])[ ]*(\d+)'
 
 def try_compile_aggregation(row_string):
   """
@@ -514,12 +537,13 @@ def try_compile_aggregation(row_string):
   aggregation. Otherwise, it returns None.
   """
   if row_string and AGGREGATION_PREFIX.match(row_string):
-    function, field, value = re.findall(AGGREGATION_SYNTAX_MATCH_PATTERN, row_string)[0]
+    function, field, op, value = \
+        re.findall(AGGREGATION_SYNTAX_MATCH_PATTERN, row_string)[0]
     # Validate function
     assert(function == 'SUM')
     # Validate value is integer
     expected_value = int(value)
-    return (function, field, expected_value)
+    return (function, field, op, expected_value)
   return None
 
 def compute_aggregation(function, field, runtime_profile):
@@ -615,19 +639,75 @@ def verify_runtime_profile(expected, actual, update_section=False):
   # Compute the aggregations and check against values
   for i in xrange(len(expected_aggregations)):
     if (expected_aggregations[i] is None): continue
-    function, field, expected_value = expected_aggregations[i]
+    function, field, op, expected_value = expected_aggregations[i]
     actual_value = compute_aggregation(function, field, actual)
     if update_section:
-      updated_aggregations.append("aggregation(%s, %s): %d"
-                                  % (function, field, actual_value))
+      updated_aggregations.append("aggregation(%s, %s)%s %d"
+                                  % (function, field, op, actual_value))
     else:
-        assert actual_value == expected_value, ("Aggregation of %s over %s did not match "
-            "expected results.\nEXPECTED VALUE:\n%d\n\nACTUAL VALUE:\n%d"
-            "\n\nPROFILE:\n%s\n"
-            % (function, field, expected_value, actual_value, actual))
+        if op == ':' and actual_value != expected_value:
+          assert actual_value == expected_value, ("Aggregation of %s over %s did not "
+              "match expected results.\nEXPECTED VALUE:\n%d\n\n\nACTUAL VALUE:\n%d\n\n"
+              "OP:\n%s\n\n"
+              "\n\nPROFILE:\n%s\n"
+              % (function, field, expected_value, actual_value, op, actual))
+        elif op == '>' and actual_value <= expected_value:
+          assert actual_value > expected_value, ("Aggregation of %s over %s did not "
+              "match expected results.\nEXPECTED VALUE:\n%d\n\n\nACTUAL VALUE:\n%d\n\n"
+              "OP:\n%s\n\n"
+              "\n\nPROFILE:\n%s\n"
+              % (function, field, expected_value, actual_value, op, actual))
+        elif op == '<' and actual_value >= expected_value:
+          assert actual_value < expected_value, ("Aggregation of %s over %s did not "
+              "match expected results.\nEXPECTED VALUE:\n%d\n\n\nACTUAL VALUE:\n%d\n\n"
+              "OP:\n%s\n\n"
+              "\n\nPROFILE:\n%s\n"
+              % (function, field, expected_value, actual_value, op, actual))
+
   return updated_aggregations
 
 
+def extract_event_sequence(runtime_profile):
+  """ Returns a list containing the names of the events on the event sequence in the
+  provided runtime profile"""
+  # The lines corresponding to the events in the event sequence.
+  events = []
+
+  # The number of leading whitespace in the lines containing the events, used to
+  # detect the line after the last event.
+  indent_len = None
+
+  # Set to true when encountering the header before the first event. This means the
+  # following lines contain the events.
+  found_events_start = False
+
+  for line in runtime_profile.splitlines():
+    if found_events_start:
+      leading_whitespace = len(line) - len(line.lstrip())
+      if indent_len is None:
+        # This was the first event. We store the indentation of the events.
+        indent_len = leading_whitespace
+      elif leading_whitespace < indent_len:
+        # We've reached the line after the events, stop the iteration.
+        break
+
+      # If we reach here we are processing a line containing an event.
+      events.append(__extract_event_name(line))
+
+    elif 'Fragment Instance Lifecycle Event Timeline' in line:
+      found_events_start = True
+
+  return events
+
+
+def __extract_event_name(line):
+  # A typical event sequence line from which we extract the name is:
+  # "- Prepare Finished: 1.778ms (1.778ms)"
+  start = line.index('-') + 2  # There is a space after the dash.
+  end = line.index(':')
+  return line[start:end]
+
+
 def verify_lineage(expected, actual, lineage_skip_json_keys=DEFAULT_LINEAGE_SKIP_KEYS):
   """Compares the lineage JSON objects expected and actual."""
   def recursive_sort(obj):
