@@ -35,7 +35,7 @@ from TCLIService import TCLIService
 from beeswaxd.BeeswaxService import QueryState
 from tests.beeswax.impala_beeswax import ImpalaBeeswaxException
 from tests.common.custom_cluster_test_suite import CustomClusterTestSuite
-from tests.common.skip import SkipIfNotHdfsMinicluster
+from tests.common.skip import SkipIfNotHdfsMinicluster, SkipIfGCS
 from tests.hs2.hs2_test_suite import HS2TestSuite, needs_session
 
 LOG = logging.getLogger(__name__)
@@ -163,110 +163,6 @@ class TestRestart(CustomClusterTestSuite):
     thread.join()
     self.wait_for_state(query_handle[0], QueryState.FINISHED, 30000)
 
-  @pytest.mark.execute_serially
-  @CustomClusterTestSuite.with_args(
-    statestored_args="--statestore_update_frequency_ms=5000")
-  def test_restart_catalogd(self):
-    self.execute_query_expect_success(self.client, "drop table if exists join_aa")
-    self.execute_query_expect_success(self.client, "create table join_aa(id int)")
-    # Make the catalog object version grow large enough
-    self.execute_query_expect_success(self.client, "invalidate metadata")
-
-    # No need to care whether the dll is executed successfully, it is just to make
-    # the local catalog catche of impalad out of sync
-    for i in range(0, 10):
-      try:
-        query = "alter table join_aa add columns (age" + str(i) + " int)"
-        self.execute_query_async(query)
-      except Exception, e:
-        LOG.info(str(e))
-      if i == 5:
-        self.cluster.catalogd.restart()
-
-    self.execute_query_expect_success(self.client,
-        "alter table join_aa add columns (name string)")
-    self.execute_query_expect_success(self.client, "select name from join_aa")
-    self.execute_query_expect_success(self.client, "drop table join_aa")
-
-  @pytest.mark.execute_serially
-  @CustomClusterTestSuite.with_args(
-    statestored_args="--statestore_update_frequency_ms=5000")
-  def test_restart_catalogd_sync_ddl(self):
-    self.execute_query_expect_success(self.client, "drop table if exists join_aa")
-    self.execute_query_expect_success(self.client, "create table join_aa(id int)")
-    # Make the catalog object version grow large enough
-    self.execute_query_expect_success(self.client, "invalidate metadata")
-    query_options = {"sync_ddl": "true"}
-
-    # No need to care whether the dll is executed successfully, it is just to make
-    # the local catalog catche of impalad out of sync
-    for i in range(0, 10):
-      try:
-        query = "alter table join_aa add columns (age" + str(i) + " int)"
-        self.execute_query_async(query, query_options)
-      except Exception, e:
-        LOG.info(str(e))
-      if i == 5:
-        self.cluster.catalogd.restart()
-
-    self.execute_query_expect_success(self.client,
-        "alter table join_aa add columns (name string)", query_options)
-    self.execute_query_expect_success(self.client, "select name from join_aa")
-    self.execute_query_expect_success(self.client, "drop table join_aa")
-
-  UPDATE_FREQUENCY_S = 10
-
-  @pytest.mark.execute_serially
-  @CustomClusterTestSuite.with_args(
-    statestored_args="--statestore_update_frequency_ms={frequency_ms}"
-    .format(frequency_ms=(UPDATE_FREQUENCY_S * 1000)))
-  def test_restart_catalogd_twice(self):
-    self.execute_query_expect_success(self.client, "drop table if exists join_aa")
-    self.cluster.catalogd.restart()
-    query = "create table join_aa(id int)"
-    query_handle = []
-
-    def execute_query_async():
-      query_handle.append(self.execute_query(query))
-
-    thread = threading.Thread(target=execute_query_async)
-    thread.start()
-    sleep(self.UPDATE_FREQUENCY_S - 5)
-    self.cluster.catalogd.restart()
-    thread.join()
-    self.execute_query_expect_success(self.client,
-        "alter table join_aa add columns (name string)")
-    self.execute_query_expect_success(self.client, "select name from join_aa")
-    self.execute_query_expect_success(self.client, "drop table join_aa")
-
-  @pytest.mark.execute_serially
-  @CustomClusterTestSuite.with_args(
-      impalad_args="--use_local_catalog=true",
-      catalogd_args="--catalog_topic_mode=minimal",
-      statestored_args="--statestore_update_frequency_ms=5000")
-  def test_restart_catalogd_with_local_catalog(self):
-    self.execute_query_expect_success(self.client, "drop table if exists join_aa")
-    self.execute_query_expect_success(self.client, "create table join_aa(id int)")
-    # Make the catalog object version grow large enough
-    self.execute_query_expect_success(self.client, "invalidate metadata")
-
-    # No need to care whether the dll is executed successfully, it is just to make
-    # the local catalog catche of impalad out of sync
-    for i in range(0, 10):
-      try:
-        query = "alter table join_aa add columns (age" + str(i) + " int)"
-        self.execute_query_async(query)
-      except Exception, e:
-        LOG.info(str(e))
-      if i == 5:
-        self.cluster.catalogd.restart()
-
-    self.execute_query_expect_success(self.client,
-        "alter table join_aa add columns (name string)")
-    self.execute_query_expect_success(self.client, "select name from join_aa")
-    self.execute_query_expect_success(self.client, "select age0 from join_aa")
-    self.execute_query_expect_success(self.client, "drop table join_aa")
-
   SUBSCRIBER_TIMEOUT_S = 2
   CANCELLATION_GRACE_PERIOD_S = 5
 
@@ -360,6 +256,7 @@ class TestGracefulShutdown(CustomClusterTestSuite, HS2TestSuite):
   def get_workload(cls):
     return 'functional-query'
 
+  @SkipIfGCS.jira(reason="IMPALA-10562")
   @pytest.mark.execute_serially
   @CustomClusterTestSuite.with_args(
       impalad_args="--shutdown_grace_period_s={grace_period} \
@@ -381,12 +278,11 @@ class TestGracefulShutdown(CustomClusterTestSuite, HS2TestSuite):
     assert ("This may be because the port specified is wrong.") not in str(ex)
 
     # Test that pointing to the wrong thrift service (the HS2 port) fails gracefully-ish.
-    thrift_ports = [21051, 22001]  # HS2 port, old backend port.
-    for port in thrift_ports:
-      ex = self.execute_query_expect_failure(self.client,
-          ":shutdown('localhost:{0}')".format(port))
-      assert ("failed with error 'RemoteShutdown() RPC failed") in str(ex)
-      assert ("This may be because the port specified is wrong.") in str(ex)
+    thrift_port = 21051  # HS2 port.
+    ex = self.execute_query_expect_failure(self.client,
+        ":shutdown('localhost:{0}')".format(thrift_port))
+    assert ("failed with error 'RemoteShutdown() RPC failed") in str(ex)
+    assert ("This may be because the port specified is wrong.") in str(ex)
 
     # Test RPC error handling with debug action.
     ex = self.execute_query_expect_failure(self.client, ":shutdown('localhost:27001')",
@@ -478,7 +374,7 @@ class TestGracefulShutdown(CustomClusterTestSuite, HS2TestSuite):
     # the shutdown grace period expires. This demonstrates that queries don't get
     # cancelled if the cluster membership changes while they're waiting for admission.
     before_shutdown_admission_handle = self.execute_query_async(QUERY,
-        {'debug_action': 'CRS_BEFORE_ADMISSION:SLEEP@30000'})
+        {'debug_action': 'AC_BEFORE_ADMISSION:SLEEP@30000'})
 
     # Shut down and wait for the shutdown state to propagate through statestore.
     result = self.execute_query_expect_success(self.client, SHUTDOWN_EXEC2)
@@ -681,8 +577,10 @@ class TestGracefulShutdown(CustomClusterTestSuite, HS2TestSuite):
   @pytest.mark.execute_serially
   @CustomClusterTestSuite.with_args(
       impalad_args="--shutdown_grace_period_s={grace_period} \
+          --shutdown_deadline_s={deadline} \
           --hostname={hostname}".format(grace_period=IDLE_SHUTDOWN_GRACE_PERIOD_S,
-            hostname=socket.gethostname()), cluster_size=1)
+            deadline=EXEC_SHUTDOWN_DEADLINE_S, hostname=socket.gethostname()),
+      cluster_size=1)
   def test_shutdown_signal(self):
     """Test that an idle impalad shuts down in a timely manner after the shutdown grace
     period elapses."""
@@ -701,7 +599,8 @@ class TestGracefulShutdown(CustomClusterTestSuite, HS2TestSuite):
     # Make sure signal was received and the grace period and deadline are as expected.
     self.assert_impalad_log_contains('INFO',
       "Shutdown signal received. Current Shutdown Status: shutdown grace period left: "
-      "{0}s000ms, deadline left: 8760h".format(self.IDLE_SHUTDOWN_GRACE_PERIOD_S))
+      "{0}s000ms, deadline left: {1}s000ms".format(self.IDLE_SHUTDOWN_GRACE_PERIOD_S,
+        self.EXEC_SHUTDOWN_DEADLINE_S))
 
   @pytest.mark.execute_serially
   @CustomClusterTestSuite.with_args(cluster_size=1)
