@@ -19,6 +19,7 @@
 
 import os
 import pytest
+import re
 
 from testdata.common import widetable
 from tests.common.impala_cluster import ImpalaCluster
@@ -87,6 +88,7 @@ class TestInsertQueries(ImpalaTestSuite):
     table_name = unique_database + ".insert_largestring"
 
     self.client.set_configuration_option("mem_limit", "4gb")
+    self.client.set_configuration_option("max_row_size", "257mb")
     file_format = vector.get_value('table_format').file_format
     if file_format == "parquet":
       stored_as = file_format
@@ -113,6 +115,9 @@ class TestInsertQueries(ImpalaTestSuite):
 
     # IMPALA-7648: test that we gracefully fail when there is not enough memory
     # to fit the scanned string in memory.
+    # IMPALA-9856: Disable result spooling for this query since it is intended to test
+    # for OOM.
+    self.client.set_configuration_option("spool_query_results", "0")
     self.client.set_configuration_option("mem_limit", "50M")
     try:
       self.client.execute("select s from {0}".format(table_name))
@@ -125,17 +130,19 @@ class TestInsertQueries(ImpalaTestSuite):
   def setup_class(cls):
     super(TestInsertQueries, cls).setup_class()
 
-  @pytest.mark.execute_serially
+  @UniqueDatabase.parametrize(sync_ddl=True)
   # Erasure coding doesn't respect memory limit
   @SkipIfEC.fix_later
   # ABFS partition names cannot end in periods
   @SkipIfABFS.file_or_folder_name_ends_with_period
-  def test_insert(self, vector):
+  def test_insert(self, vector, unique_database):
     if (vector.get_value('table_format').file_format == 'parquet'):
       vector.get_value('exec_option')['COMPRESSION_CODEC'] = \
           vector.get_value('compression_codec')
-    self.run_test_case('QueryTest/insert', vector,
-        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1)
+    self.run_test_case('QueryTest/insert', vector, unique_database,
+        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1,
+        test_file_vars={'$ORIGINAL_DB': ImpalaTestSuite
+        .get_db_name_from_format(vector.get_value('table_format'))})
 
   @SkipIfHive2.acid
   @UniqueDatabase.parametrize(sync_ddl=True)
@@ -160,14 +167,17 @@ class TestInsertQueries(ImpalaTestSuite):
     self.run_test_case('QueryTest/acid-insert-fail', vector, unique_database,
         multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1)
 
+  @UniqueDatabase.parametrize(sync_ddl=True)
   @pytest.mark.execute_serially
   @SkipIfNotHdfsMinicluster.tuned_for_minicluster
-  def test_insert_mem_limit(self, vector):
+  def test_insert_mem_limit(self, vector, unique_database):
     if (vector.get_value('table_format').file_format == 'parquet'):
       vector.get_value('exec_option')['COMPRESSION_CODEC'] = \
           vector.get_value('compression_codec')
-    self.run_test_case('QueryTest/insert-mem-limit', vector,
-        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1)
+    self.run_test_case('QueryTest/insert-mem-limit', vector, unique_database,
+        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1,
+        test_file_vars={'$ORIGINAL_DB': ImpalaTestSuite
+        .get_db_name_from_format(vector.get_value('table_format'))})
     # IMPALA-7023: These queries can linger and use up memory, causing subsequent
     # tests to hit memory limits. Wait for some time to allow the query to
     # be reclaimed.
@@ -176,19 +186,22 @@ class TestInsertQueries(ImpalaTestSuite):
     for v in verifiers:
       v.wait_for_metric("impala-server.num-fragments-in-flight", 0, timeout=180)
 
-  @pytest.mark.execute_serially
-  @SkipIfS3.eventually_consistent
-  def test_insert_overwrite(self, vector):
-    self.run_test_case('QueryTest/insert_overwrite', vector,
-        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1)
+  @UniqueDatabase.parametrize(sync_ddl=True)
+  def test_insert_overwrite(self, vector, unique_database):
+    self.run_test_case('QueryTest/insert_overwrite', vector, unique_database,
+        multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1,
+        test_file_vars={'$ORIGINAL_DB': ImpalaTestSuite
+        .get_db_name_from_format(vector.get_value('table_format'))})
 
-  @pytest.mark.execute_serially
-  def test_insert_bad_expr(self, vector):
+  @UniqueDatabase.parametrize(sync_ddl=True)
+  def test_insert_bad_expr(self, vector, unique_database):
     # The test currently relies on codegen being disabled to trigger an error in
     # the output expression of the table sink.
     if vector.get_value('exec_option')['disable_codegen']:
-      self.run_test_case('QueryTest/insert_bad_expr', vector,
-          multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1)
+      self.run_test_case('QueryTest/insert_bad_expr', vector, unique_database,
+          multiple_impalad=vector.get_value('exec_option')['sync_ddl'] == 1,
+          test_file_vars={'$ORIGINAL_DB': ImpalaTestSuite
+          .get_db_name_from_format(vector.get_value('table_format'))})
 
   @UniqueDatabase.parametrize(sync_ddl=True)
   def test_insert_random_partition(self, vector, unique_database):
@@ -306,9 +319,10 @@ class TestInsertNullQueries(ImpalaTestSuite):
   def setup_class(cls):
     super(TestInsertNullQueries, cls).setup_class()
 
-  @pytest.mark.execute_serially
-  def test_insert_null(self, vector):
-    self.run_test_case('QueryTest/insert_null', vector)
+  def test_insert_null(self, vector, unique_database):
+    self.run_test_case('QueryTest/insert_null', vector, unique_database,
+        test_file_vars={'$ORIGINAL_DB': ImpalaTestSuite
+        .get_db_name_from_format(vector.get_value('table_format'))})
 
 
 class TestInsertFileExtension(ImpalaTestSuite):
@@ -339,3 +353,79 @@ class TestInsertFileExtension(ImpalaTestSuite):
     for path in self.filesystem_client.ls("test-warehouse/{0}.db/{1}".format(
         unique_database, table_name)):
       if not path.startswith('_'): assert path.endswith(file_extension)
+
+
+class TestInsertHdfsWriterLimit(ImpalaTestSuite):
+  """Test to make sure writer fragment instances are distributed evenly when using max
+      hdfs_writers query option."""
+  @classmethod
+  def get_workload(self):
+    return 'functional-query'
+
+  @classmethod
+  def add_test_dimensions(cls):
+    super(TestInsertHdfsWriterLimit, cls).add_test_dimensions()
+    cls.ImpalaTestMatrix.add_constraint(lambda v:
+        (v.get_value('table_format').file_format == 'parquet'))
+
+  @UniqueDatabase.parametrize(sync_ddl=True)
+  @SkipIfNotHdfsMinicluster.tuned_for_minicluster
+  def test_insert_writer_limit(self, unique_database):
+    # Root internal (non-leaf) fragment.
+    query = "create table {0}.test1 as select int_col from " \
+            "functional_parquet.alltypes".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=2, mt_dop=0,
+                                           expected_num_instances_per_host=[1, 2, 2])
+    # Root coordinator fragment.
+    query = "create table {0}.test2 as select int_col from " \
+            "functional_parquet.alltypes limit 100000".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=2, mt_dop=0,
+                                           expected_num_instances_per_host=[1, 1, 2])
+    # Root scan fragment. Instance count within limit.
+    query = "create table {0}.test3 as select int_col from " \
+            "functional_parquet.alltypes".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=4, mt_dop=0,
+                                           expected_num_instances_per_host=[1, 1, 1])
+
+  @UniqueDatabase.parametrize(sync_ddl=True)
+  @SkipIfNotHdfsMinicluster.tuned_for_minicluster
+  def test_mt_dop_writer_limit(self, unique_database):
+    # Root internal (non-leaf) fragment.
+    query = "create table {0}.test1 as select int_col from " \
+            "functional_parquet.alltypes".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=11, mt_dop=10,
+                                           expected_num_instances_per_host=[11, 12, 12])
+    # Root coordinator fragment.
+    query = "create table {0}.test2 as select int_col from " \
+            "functional_parquet.alltypes limit 100000".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=2, mt_dop=10,
+                                           expected_num_instances_per_host=[8, 8, 9])
+    # Root scan fragment. Instance count within limit.
+    query = "create table {0}.test3 as select int_col from " \
+            "functional_parquet.alltypes".format(unique_database)
+    self.__run_insert_and_verify_instances(query, max_fs_writers=30, mt_dop=10,
+                                           expected_num_instances_per_host=[8, 8, 8])
+
+  def __run_insert_and_verify_instances(self, query, max_fs_writers, mt_dop,
+                                        expected_num_instances_per_host):
+    self.client.set_configuration_option("max_fs_writers", max_fs_writers)
+    self.client.set_configuration_option("mt_dop", mt_dop)
+    # Test depends on both planner and scheduler to see the same state of the cluster
+    # having 3 executors, so to reduce flakiness we make sure all 3 executors are up
+    # and running.
+    self.impalad_test_service.wait_for_metric_value("cluster-membership.backends.total",
+                                                    3)
+    result = self.client.execute(query)
+    assert 'HDFS WRITER' in result.exec_summary[0]['operator'], result.runtime_profile
+    assert int(result.exec_summary[0]['num_instances']) <= int(
+      max_fs_writers), result.runtime_profile
+    regex = r'Per Host Number of Fragment Instances' \
+            r':.*?\((.*?)\).*?\((.*?)\).*?\((.*?)\).*?\n'
+    matches = re.findall(regex, result.runtime_profile)
+    assert len(matches) == 1 and len(matches[0]) == 3, result.runtime_profile
+    num_instances_per_host = [int(i) for i in matches[0]]
+    num_instances_per_host.sort()
+    expected_num_instances_per_host.sort()
+    assert num_instances_per_host == expected_num_instances_per_host, \
+      result.runtime_profile
+    self.client.clear_configuration()
