@@ -29,7 +29,7 @@ from tests.common.environ import impalad_basedir
 from tests.common.impala_test_suite import ImpalaTestSuite
 from tests.common.parametrize import UniqueDatabase
 from tests.common.skip import (SkipIfEC, SkipIfIsilon, SkipIfLocal, SkipIfS3, SkipIfABFS,
-    SkipIfADLS)
+                               SkipIfADLS, SkipIfGCS)
 from tests.common.test_dimensions import create_exec_option_dimension
 from tests.common.test_result_verifier import verify_query_result_is_equal
 from tests.common.test_vector import ImpalaTestDimension
@@ -38,6 +38,7 @@ from tests.util.get_parquet_metadata import (decode_stats_value,
     get_parquet_metadata_from_hdfs_folder)
 
 PARQUET_CODECS = ['none', 'snappy', 'gzip', 'zstd', 'lz4']
+IMPALA_HOME = os.environ['IMPALA_HOME']
 
 
 class RoundFloat():
@@ -116,7 +117,6 @@ class TestInsertParquetQueries(ImpalaTestSuite):
 
   @SkipIfEC.oom
   @SkipIfLocal.multiple_impalad
-  @SkipIfS3.eventually_consistent
   @UniqueDatabase.parametrize(sync_ddl=True)
   def test_insert_parquet(self, vector, unique_database):
     vector.get_value('exec_option')['PARQUET_FILE_SIZE'] = \
@@ -265,7 +265,8 @@ class TestHdfsParquetTableWriter(ImpalaTestSuite):
       for f in files:
         if not f.endswith('parq'):
           continue
-        check_call([os.path.join(impalad_basedir, 'util/parquet-reader'), '--file',
+        check_call([os.path.join(IMPALA_HOME, "bin/run-binary.sh"),
+                    os.path.join(impalad_basedir, 'util/parquet-reader'), '--file',
                     os.path.join(tmpdir.strpath, str(f))])
 
   def test_sorting_columns(self, vector, unique_database, tmpdir):
@@ -532,10 +533,34 @@ class TestHdfsParquetTableWriter(ImpalaTestSuite):
     self._ctas_and_check_int64_timestamps(vector, unique_database, tmpdir, "micros")
     self._ctas_and_check_int64_timestamps(vector, unique_database, tmpdir, "nanos")
 
+  # Skip test for non-HDFS environment as it uses Hive statement.
+  # Hive statement is being used as Impala's result are converted
+  # by python to string. In both HS2 and beewax, it only handles float
+  # precision uptil 16 decimal digits and test needs 17.
+  # IMPALA-9365 describes why HS2 is not started on non-HDFS test env.
+  @SkipIfS3.hive
+  @SkipIfGCS.hive
+  @SkipIfABFS.hive
+  @SkipIfADLS.hive
+  @SkipIfIsilon.hive
+  @SkipIfLocal.hive
+  def test_double_precision(self, vector, unique_database):
+    # IMPALA-10654: Test inserting double into Parquet table retains the precision.
+    src_tbl = "{0}.{1}".format(unique_database, "i10654_parquet")
+    create_tbl_stmt = """create table {0} (dbl1 double)
+        stored as parquet""".format(src_tbl)
+    self.execute_query_expect_success(self.client, create_tbl_stmt)
+    insert_tbl_stmt = """insert into table {0}
+        values (-0.43149576573887316)""".format(src_tbl)
+    self.execute_query_expect_success(self.client, insert_tbl_stmt)
+    select_stmt = """select * from {0}""".format(src_tbl)
+    result = self.run_stmt_in_hive(select_stmt)
+    assert result.split('\n')[1] == '-0.43149576573887316'
 
 @SkipIfIsilon.hive
 @SkipIfLocal.hive
 @SkipIfS3.hive
+@SkipIfGCS.hive
 @SkipIfABFS.hive
 @SkipIfADLS.hive
 # TODO: Should we move this to test_parquet_stats.py?
