@@ -26,9 +26,11 @@ from tests.common.impala_test_suite import ImpalaTestSuite
 from tests.common.skip import (
     SkipIfIsilon,
     SkipIfS3,
+    SkipIfGCS,
     SkipIfABFS,
     SkipIfADLS,
     SkipIfEC,
+    SkipIfHive2,
     SkipIfLocal,
     SkipIfNotHdfsMinicluster
     )
@@ -143,20 +145,11 @@ class TestNestedTypesNoMtDop(ImpalaTestSuite):
     self.run_test_case('QueryTest/nested-types-tpch-mem-limit', vector,
                        use_db='tpch_nested' + db_suffix)
 
-  @SkipIfNotHdfsMinicluster.tuned_for_minicluster
-  def test_tpch_mem_limit_single_node(self, vector):
-    """Queries over the larger nested TPCH dataset with memory limits tuned for
-    a 3-node HDFS minicluster with num_nodes=1."""
-    new_vector = deepcopy(vector)
-    new_vector.get_value('exec_option')['num_nodes'] = 1
-    if vector.get_value('table_format').file_format == 'orc':
-      # IMPALA-8336: lower memory limit for ORC
-      new_vector.get_value('exec_option')['mem_limit'] = '20M'
-    else:
-      new_vector.get_value('exec_option')['mem_limit'] = '28M'
+  def test_tpch_errors(self, vector):
+    """Queries that test error handling on the TPC-H nested data set."""
     db_suffix = vector.get_value('table_format').db_suffix()
-    self.run_test_case('QueryTest/nested-types-tpch-mem-limit-single-node',
-                       new_vector, use_db='tpch_nested' + db_suffix)
+    self.run_test_case('QueryTest/nested-types-tpch-errors',
+                       vector, use_db='tpch_nested' + db_suffix)
 
   @SkipIfEC.fix_later
   def test_parquet_stats(self, vector):
@@ -168,6 +161,7 @@ class TestNestedTypesNoMtDop(ImpalaTestSuite):
 
   @SkipIfIsilon.hive
   @SkipIfS3.hive
+  @SkipIfGCS.hive
   @SkipIfABFS.hive
   @SkipIfADLS.hive
   @SkipIfLocal.hive
@@ -207,13 +201,61 @@ class TestNestedTypesNoMtDop(ImpalaTestSuite):
         )
         STORED AS {1}""".format(db_table, file_format))
     # Add multiple partitions pointing to the complextypes_tbl data.
+    if file_format == 'parquet':
+      base_table = "functional%s.complextypestbl" % db_suffix
+    else:
+      assert file_format == 'orc'
+      base_table = "functional%s.complextypestbl_non_transactional" % db_suffix
     for partition in [1, 2]:
       self.client.execute("ALTER TABLE {0} ADD PARTITION(part={1}) LOCATION '{2}'".format(
           db_table, partition,
-          self._get_table_location("functional%s.complextypestbl" % db_suffix, vector)))
+          self._get_table_location(base_table, vector)))
     self.run_test_case('QueryTest/nested-types-basic-partitioned', vector,
         unique_database)
 
+  # Skip this test on non-HDFS filesystems, because the test contains Hive
+  # queries that hang in some cases due to IMPALA-9365.
+  @SkipIfABFS.hive
+  @SkipIfADLS.hive
+  @SkipIfIsilon.hive
+  @SkipIfLocal.hive
+  @SkipIfS3.hive
+  @SkipIfGCS.hive
+  @SkipIfHive2.acid
+  def test_partitioned_table_acid(self, vector, unique_database):
+    """IMPALA-6370: Test that a partitioned table with nested types can be scanned."""
+    table = "complextypes_partitioned"
+    db_table = "{0}.{1}".format(unique_database, table)
+    table_format_info = vector.get_value('table_format')  # type: TableFormatInfo
+    file_format = table_format_info.file_format
+    if file_format != "orc":
+      pytest.skip('Full ACID tables are only supported in ORC format.')
+
+    self.client.execute("""
+        CREATE TABLE {0} (
+          id BIGINT,
+          int_array ARRAY<INT>,
+          int_array_array ARRAY<ARRAY<INT>>,
+          int_map MAP<STRING,INT>,
+          int_map_array ARRAY<MAP<STRING,INT>>,
+          nested_struct STRUCT<
+              a:INT,
+              b:ARRAY<INT>,
+              c:STRUCT<d:ARRAY<ARRAY<STRUCT<e:INT,f:STRING>>>>,
+              g:MAP<STRING,STRUCT<h:STRUCT<i:ARRAY<DOUBLE>>>>>
+        )
+        PARTITIONED BY (
+          part int
+        )
+        STORED AS ORC
+        TBLPROPERTIES('transactional'='true')""".format(db_table))
+    # Add multiple partitions with the complextypestbl data.
+    base_table = "functional_orc_def.complextypestbl"
+    for partition in [1, 2]:
+      self.run_stmt_in_hive("INSERT INTO TABLE {0} PARTITION(part={1}) "
+          "SELECT * FROM {2}".format(db_table, partition, base_table))
+    self.run_test_case('QueryTest/nested-types-basic-partitioned', vector,
+        unique_database)
 
 class TestParquetArrayEncodings(ImpalaTestSuite):
   TESTFILE_DIR = os.path.join(os.environ['IMPALA_HOME'],
@@ -329,7 +371,6 @@ class TestParquetArrayEncodings(ImpalaTestSuite):
   # .array = 34
   # .array = 35
   # .array = 36
-  @SkipIfS3.eventually_consistent
   def test_avro_primitive_in_list(self, vector, unique_database):
     self.__test_primitive_in_list(unique_database, "AvroPrimitiveInList",
       "AvroPrimitiveInList.parquet", vector)
@@ -437,7 +478,6 @@ class TestParquetArrayEncodings(ImpalaTestSuite):
   # ..intListsColumn_tuple_tuple = 6
   # ..intListsColumn_tuple_tuple = 7
   # ..intListsColumn_tuple_tuple = 8
-  @SkipIfS3.eventually_consistent
   def test_thrift_array_of_arrays(self, vector, unique_database):
     self.__test_array_of_arrays(unique_database, "ThriftArrayOfArrays",
       "bad-thrift.parquet", vector, 1)
@@ -682,6 +722,7 @@ class TestMaxNestingDepth(ImpalaTestSuite):
 
   @SkipIfIsilon.hive
   @SkipIfS3.hive
+  @SkipIfGCS.hive
   @SkipIfABFS.hive
   @SkipIfADLS.hive
   @SkipIfLocal.hive
