@@ -20,7 +20,7 @@ from datetime import datetime
 from tests.beeswax.impala_beeswax import ImpalaBeeswaxException
 from tests.common.impala_test_suite import ImpalaTestSuite
 from tests.common.skip import (SkipIfS3, SkipIfABFS, SkipIfADLS, SkipIfIsilon,
-                               SkipIfLocal, SkipIfNotHdfsMinicluster)
+                               SkipIfGCS, SkipIfLocal, SkipIfNotHdfsMinicluster)
 from tests.util.filesystem_utils import IS_EC
 from time import sleep
 from RuntimeProfile.ttypes import TRuntimeProfileFormat
@@ -92,6 +92,7 @@ class TestObservability(ImpalaTestSuite):
     self.hs2_client.close_query(handle)
 
   @SkipIfS3.hbase
+  @SkipIfGCS.hbase
   @SkipIfLocal.hbase
   @SkipIfIsilon.hbase
   @SkipIfABFS.hbase
@@ -183,7 +184,7 @@ class TestObservability(ImpalaTestSuite):
     expected_str = ("Query Options (set by configuration and planner): "
         "MEM_LIMIT=8589934592,"
         "NUM_NODES=1,NUM_SCANNER_THREADS=1,"
-        "RUNTIME_FILTER_MODE=0,MT_DOP=0,{erasure_coding}TIMEZONE={timezone},"
+        "RUNTIME_FILTER_MODE=OFF,MT_DOP=0,{erasure_coding}TIMEZONE={timezone},"
         "CLIENT_IDENTIFIER="
         "query_test/test_observability.py::TestObservability::()::test_query_options"
         "\n")
@@ -196,7 +197,7 @@ class TestObservability(ImpalaTestSuite):
     """Test that the exec summary is populated correctly in every query state"""
     query = "select count(*) from functional.alltypes"
     handle = self.execute_query_async(query,
-        {"debug_action": "CRS_BEFORE_ADMISSION:SLEEP@1000"})
+        {"debug_action": "AC_BEFORE_ADMISSION:SLEEP@1000"})
     # If ExecuteStatement() has completed and the query is paused in the admission control
     # phase, then the coordinator has not started yet and exec_summary should be empty.
     exec_summary = self.client.get_exec_summary(handle)
@@ -218,7 +219,7 @@ class TestObservability(ImpalaTestSuite):
     query state"""
     query = "select count(*) from functional.alltypes"
     handle = self.execute_query_async(query,
-        {"debug_action": "CRS_BEFORE_ADMISSION:SLEEP@1000"})
+        {"debug_action": "AC_BEFORE_ADMISSION:SLEEP@1000"})
 
     # If ExecuteStatement() has completed and the query is paused in the admission control
     # phase, then the coordinator has not started yet and exec_summary should be empty.
@@ -241,10 +242,12 @@ class TestObservability(ImpalaTestSuite):
     """IMPALA-6081: Test that the expected number of fragment instances and their exec
     nodes appear in the runtime profile, even when fragments may be quickly cancelled when
     all results are already returned."""
+    query_opts = {'report_skew_limit': -1}
     results = self.execute_query("""
         with l as (select * from tpch.lineitem UNION ALL select * from tpch.lineitem)
         select STRAIGHT_JOIN count(*) from (select * from tpch.lineitem a LIMIT 1) a
-        join (select * from l LIMIT 2000000) b on a.l_orderkey = -b.l_orderkey;""")
+        join (select * from l LIMIT 2000000) b on a.l_orderkey = -b.l_orderkey;""",
+        query_opts)
     # There are 3 scan nodes and each appears in the profile n+1 times (for n fragment
     # instances + the averaged fragment). n depends on how data is loaded and scheduler's
     # decision.
@@ -314,10 +317,10 @@ class TestObservability(ImpalaTestSuite):
         r'CatalogFetch.RPCs.Requests',
         r'CatalogFetch.RPCs.Time',
         r'CatalogFetch.StorageLoad.Time',
-        r'CatalogFetch.TableNames.Hits',
-        r'CatalogFetch.TableNames.Misses',
-        r'CatalogFetch.TableNames.Requests',
-        r'CatalogFetch.TableNames.Time',
+        r'CatalogFetch.TableList.Hits',
+        r'CatalogFetch.TableList.Misses',
+        r'CatalogFetch.TableList.Requests',
+        r'CatalogFetch.TableList.Time',
         r'CatalogFetch.Tables.Hits',
         r'CatalogFetch.Tables.Misses',
         r'CatalogFetch.Tables.Requests',
@@ -404,7 +407,6 @@ class TestObservability(ImpalaTestSuite):
   def __verify_profile_event_sequence(self, event_regexes, runtime_profile):
     """Check that 'event_regexes' appear in a consecutive series of lines in
        'runtime_profile'"""
-    lines = runtime_profile.splitlines()
     event_regex_index = 0
 
     # Check that the strings appear in the above order with no gaps in the profile.
@@ -501,9 +503,13 @@ class TestObservability(ImpalaTestSuite):
         if key in line:
           # Match byte count within parentheses
           m = re.search("\(([0-9]+)\)", line)
-          assert m, "Cannot match pattern for key %s in line '%s'" % (key, line)
-          # Only keep first (query-level) counter
-          if counters[key] == 0:
+
+          # If a match was not found, then the value of the key should be 0
+          if not m:
+            assert key + ": 0" in line, "Invalid format for key %s" % key
+            assert counters[key] != 0, "Query level counter for key %s cannot be 0" % key
+          elif counters[key] == 0:
+            # Only keep first (query-level) counter
             counters[key] = int(m.group(1))
 
     # All counters have values
@@ -666,6 +672,7 @@ class TestObservability(ImpalaTestSuite):
         cluster_properties)
 
   @SkipIfS3.hbase
+  @SkipIfGCS.hbase
   @SkipIfLocal.hbase
   @SkipIfIsilon.hbase
   @SkipIfABFS.hbase
@@ -727,6 +734,33 @@ class TestObservability(ImpalaTestSuite):
     assert result.success
     self.__verify_hashtable_stats_profile(result.runtime_profile)
 
+  @SkipIfNotHdfsMinicluster.tuned_for_minicluster
+  def test_skew_reporting_in_runtime_profile(self):
+    """Test that the skew summary and skew details are reported in runtime profile
+    correctly"""
+    query = """select ca_state, count(*) from tpcds_parquet.store_sales,
+            tpcds_parquet.customer, tpcds_parquet.customer_address
+            where ss_customer_sk = c_customer_sk and
+            c_current_addr_sk = ca_address_sk
+            group by ca_state
+            order by ca_state
+            """
+    "Set up the skew threshold to 0.0"
+    query_opts = {'report_skew_limit': 0.0}
+    results = self.execute_query(query, query_opts)
+    assert results.success
+
+    "When the skew summary is seen, look for the details"
+    skews_found = 'skew\(s\) found at:.*HASH_JOIN.*HASH_JOIN.*HDFS_SCAN_NODE'
+    if len(re.findall(skews_found, results.runtime_profile, re.M)) == 1:
+
+      "Expect to see skew details twice at the hash join nodes."
+      probe_rows_at_hj = 'HASH_JOIN_NODE.*\n.*Skew details: ProbeRows'
+      assert len(re.findall(probe_rows_at_hj, results.runtime_profile, re.M)) == 2
+
+      "Expect to see skew details once at the scan node."
+      probe_rows_at_hdfs_scan = 'HDFS_SCAN_NODE.*\n.*Skew details: RowsRead'
+      assert len(re.findall(probe_rows_at_hdfs_scan, results.runtime_profile, re.M)) == 1
 
 class TestQueryStates(ImpalaTestSuite):
   """Test that the 'Query State' and 'Impala Query State' are set correctly in the
@@ -740,7 +774,7 @@ class TestQueryStates(ImpalaTestSuite):
     """Tests that the query profile shows expected query states."""
     query = "select count(*) from functional.alltypes where bool_col = sleep(10)"
     handle = self.execute_query_async(query,
-        {"debug_action": "CRS_BEFORE_ADMISSION:SLEEP@1000"})
+        {"debug_action": "AC_BEFORE_ADMISSION:SLEEP@1000"})
     # If ExecuteStatement() has completed and the query is paused in the admission control
     # phase, then the query must be in COMPILED state.
     profile = self.client.get_runtime_profile(handle)
@@ -770,7 +804,7 @@ class TestQueryStates(ImpalaTestSuite):
       return self.__is_line_in_profile("Query State: FINISHED", profile) and \
              self.__is_line_in_profile("Impala Query State: FINISHED", profile)
 
-    self.assert_eventually(30, 1, assert_finished,
+    self.assert_eventually(300, 1, assert_finished,
       lambda: self.client.get_runtime_profile(handle))
 
     try:
