@@ -36,9 +36,12 @@ from tests.common.impala_test_suite import ImpalaTestSuite, LOG
 from tests.common.skip import (
     SkipIf,
     SkipIfS3,
+    SkipIfGCS,
     SkipIfABFS,
     SkipIfADLS,
     SkipIfEC,
+    SkipIfHive2,
+    SkipIfHive3,
     SkipIfIsilon,
     SkipIfLocal,
     SkipIfNotHdfsMinicluster)
@@ -53,7 +56,7 @@ from tests.common.test_result_verifier import (
     QueryTestResult,
     parse_result_rows)
 from tests.common.test_vector import ImpalaTestDimension
-from tests.util.filesystem_utils import WAREHOUSE, get_fs_path
+from tests.util.filesystem_utils import IS_HDFS, WAREHOUSE, get_fs_path
 from tests.util.hdfs_util import NAMENODE
 from tests.util.get_parquet_metadata import get_parquet_metadata
 from tests.util.parse_util import get_bytes_summary_stats_counter
@@ -68,10 +71,7 @@ DEBUG_ACTION_DIMS = [None,
 # Trigger injected soft limit failures when scanner threads check memory limit.
 DEBUG_ACTION_DIMS.append('HDFS_SCANNER_THREAD_CHECK_SOFT_MEM_LIMIT:FAIL@0.5')
 
-# Map from the test dimension file_format string to the SQL "STORED AS"
-# argument.
-STORED_AS_ARGS = {'text': 'textfile', 'parquet': 'parquet', 'avro': 'avro', 'orc': 'orc',
-    'seq': 'sequencefile', 'rc': 'rcfile'}
+MT_DOP_VALUES = [0, 1, 4]
 
 class TestScannersAllTableFormats(ImpalaTestSuite):
   BATCH_SIZES = [0, 1, 16]
@@ -133,6 +133,7 @@ class TestScannersAllTableFormatsWithLimit(ImpalaTestSuite):
   @classmethod
   def add_test_dimensions(cls):
     super(TestScannersAllTableFormatsWithLimit, cls).add_test_dimensions()
+    cls.ImpalaTestMatrix.add_dimension(ImpalaTestDimension('mt_dop', *MT_DOP_VALUES))
 
   def test_limit(self, vector):
     vector.get_value('exec_option')['abort_on_error'] = 1
@@ -174,6 +175,7 @@ class TestScannersMixedTableFormats(ImpalaTestSuite):
         ImpalaTestDimension('batch_size', *TestScannersAllTableFormats.BATCH_SIZES))
     cls.ImpalaTestMatrix.add_dimension(
         ImpalaTestDimension('debug_action', *DEBUG_ACTION_DIMS))
+    cls.ImpalaTestMatrix.add_dimension(ImpalaTestDimension('mt_dop', *MT_DOP_VALUES))
 
   def test_mixed_format(self, vector):
     new_vector = deepcopy(vector)
@@ -197,26 +199,37 @@ class TestUnmatchedSchema(ImpalaTestSuite):
     cls.ImpalaTestMatrix.add_constraint(
         lambda v: v.get_value('table_format').file_format != 'avro')
 
-  def _create_test_table(self, vector):
+  def _create_test_table(self, vector, unique_database):
     """
     Creates the test table
 
     Cannot be done in a setup method because we need access to the current test vector
     """
-    self._drop_test_table(vector)
-    self.execute_query_using_client(self.client,
-        "create external table jointbl_test like jointbl", vector)
+    file_format = vector.get_value('table_format').file_format
+    if file_format == 'orc':
+      # TODO: Enable this test on non-HDFS filesystems once IMPALA-9365 is resolved.
+      if not IS_HDFS: pytest.skip()
+      self.run_stmt_in_hive(
+          "create table {0}.jointbl_test like functional.jointbl "
+          "stored as orc".format(unique_database))
+      self.run_stmt_in_hive(
+          'insert into {0}.jointbl_test '
+          'select * from functional_orc_def.jointbl'.format(unique_database))
+      self.execute_query_using_client(self.client,
+          'invalidate metadata {0}.jointbl_test'.format(unique_database),
+          vector)
+    else:
+      self.execute_query_using_client(self.client,
+          "create external table {0}.jointbl_test like jointbl".format(
+              unique_database), vector)
 
-    # Update the location of the new table to point the same location as the old table
-    location = self._get_table_location('jointbl', vector)
-    self.execute_query_using_client(self.client,
-        "alter table jointbl_test set location '%s'" % location, vector)
+      # Update the location of the new table to point the same location as the old table
+      location = self._get_table_location('jointbl', vector)
+      self.execute_query_using_client(self.client,
+          "alter table {0}.jointbl_test set location '{1}'".format(
+              unique_database, location), vector)
 
-  def _drop_test_table(self, vector):
-    self.execute_query_using_client(self.client,
-        "drop table if exists jointbl_test", vector)
-
-  def test_unmatched_schema(self, vector):
+  def test_unmatched_schema(self, vector, unique_database):
     if vector.get_value('table_format').file_format == 'kudu':
       pytest.xfail("IMPALA-2890: Missing Kudu DDL support")
 
@@ -225,11 +238,11 @@ class TestUnmatchedSchema(ImpalaTestSuite):
     # different, as hbase collapses duplicates.
     if table_format.file_format == 'hbase':
       pytest.skip()
-    self._create_test_table(vector)
-    self.run_test_case('QueryTest/test-unmatched-schema', vector)
-    self._drop_test_table(vector)
+    self._create_test_table(vector, unique_database)
+    self.run_test_case('QueryTest/test-unmatched-schema', vector, use_db=unique_database)
 
 
+# Tests that scanners can read a single-column, single-row, 10MB table
 class TestWideRow(ImpalaTestSuite):
   @classmethod
   def get_workload(cls):
@@ -244,8 +257,7 @@ class TestWideRow(ImpalaTestSuite):
     cls.ImpalaTestMatrix.add_constraint(
       lambda v: v.get_value('table_format').file_format != 'hbase')
 
-  def test_single_wide_row(self, vector):
-    """Tests that scanners can read a single-column, single-row, 10MB table"""
+  def test_wide_row(self, vector):
     if vector.get_value('table_format').file_format == 'kudu':
       pytest.xfail("KUDU-666: Kudu support for large values")
 
@@ -255,48 +267,16 @@ class TestWideRow(ImpalaTestSuite):
     # We need > 10 MB of memory because we're creating extra buffers:
     # - 10 MB table / 5 MB scan range = 2 scan ranges, each of which may allocate ~20MB
     # - Sync reads will allocate ~5MB of space
-    # The 100MB value used here was determined empirically by raising the limit until the
+    # - Result spooling require 32 MB initial reservation (2 page of 16 MB each) to fit
+    #   10 MB row.
+    # The 132MB value used here was determined empirically by raising the limit until the
     # query succeeded for all file formats -- I don't know exactly why we need this much.
     # TODO: figure out exact breakdown of memory usage (IMPALA-681)
-    new_vector.get_value('exec_option')['mem_limit'] = 100 * 1024 * 1024
+    new_vector.get_value('exec_option')['mem_limit'] = 132 * 1024 * 1024
+    # Specify that the query should able to handle 10 MB MAX_ROW_SIZE.
+    new_vector.get_value('exec_option')['max_row_size'] = 10 * 1024 * 1024
     self.run_test_case('QueryTest/wide-row', new_vector)
 
-  @SkipIfABFS.hive
-  @SkipIfADLS.hive
-  @SkipIfIsilon.hive
-  @SkipIfLocal.hive
-  @SkipIfS3.hive
-  def test_multi_wide_rows(self, vector, unique_database):
-    """Tests that scanners can read multi rows of a wide table"""
-    if vector.get_value('table_format').file_format == 'kudu':
-      pytest.xfail("Kudu table can have a maximum of 300 columns")
-    format = STORED_AS_ARGS[vector.get_value('table_format').file_format]
-
-    create_tbl_stmt = 'create table %s.wide_tbl(col0 bigint' % unique_database
-    for i in range(1, 2000):
-      create_tbl_stmt += ',col%d bigint' % i
-    create_tbl_stmt += ') stored as %s' % format
-    self.client.execute(create_tbl_stmt)
-
-    insert_stmt = 'insert into %s.wide_tbl ' % unique_database +\
-                  'select id' + (',id' * 1999) +\
-                  ' from functional.alltypes order by id limit 1000'
-    if format in ('textfile', 'parquet', 'kudu'):
-      self.client.execute(insert_stmt)
-    else:
-      self.run_stmt_in_hive(insert_stmt)
-      self.client.execute('refresh %s.wide_tbl' % unique_database)
-
-    result = self.client.execute(
-        "select * from %s.wide_tbl where col0 = 1" % unique_database)
-    assert len(result.data) == 1
-    assert type(result.data[0]) == str
-    cols = result.data[0].split('\t')
-    assert len(cols) == 2000
-    for i in range(2000):
-      assert cols[i] == '1'
-
-
 class TestWideTable(ImpalaTestSuite):
   # TODO: expand this to more rows when we have the capability
   NUM_COLS = [250, 500, 1000]
@@ -361,6 +341,26 @@ class TestHudiParquet(ImpalaTestSuite):
     self.run_test_case('QueryTest/hudi-parquet', vector)
 
 
+class TestIceberg(ImpalaTestSuite):
+  @classmethod
+  def get_workload(cls):
+    return 'functional-query'
+
+  @classmethod
+  def add_test_dimensions(cls):
+    super(TestIceberg, cls).add_test_dimensions()
+    cls.ImpalaTestMatrix.add_dimension(
+        create_exec_option_dimension(debug_action_options=DEBUG_ACTION_DIMS))
+    cls.ImpalaTestMatrix.add_constraint(
+      lambda v: v.get_value('table_format').file_format == 'parquet')
+
+  def test_iceberg_query(self, vector):
+    self.run_test_case('QueryTest/iceberg-query', vector)
+
+  def test_iceberg_profile(self, vector, unique_database):
+    self.run_test_case('QueryTest/iceberg-profile', vector, use_db=unique_database)
+
+
 class TestParquet(ImpalaTestSuite):
   @classmethod
   def get_workload(cls):
@@ -403,6 +403,19 @@ class TestParquet(ImpalaTestSuite):
     self.run_test_case('QueryTest/out-of-range-timestamp-abort-on-error',
         vector, unique_database)
 
+  def test_dateless_timestamp_parquet(self, vector, unique_database):
+    """Test scanning parquet files which still includes dateless timestamps."""
+    tbl_name = "timestamp_table"
+    create_sql = "create table %s.%s (t timestamp) stored as parquet" % (
+        unique_database, tbl_name)
+    create_table_and_copy_files(self.client, create_sql, unique_database, tbl_name,
+        ["/testdata/data/dateless_timestamps.parq"])
+
+    new_vector = deepcopy(vector)
+    del new_vector.get_value('exec_option')['abort_on_error']
+    self.run_test_case('QueryTest/dateless_timestamp_parquet', new_vector,
+        use_db=unique_database)
+
   def test_date_out_of_range_parquet(self, vector, unique_database):
     """Test scanning parquet files with an out of range date."""
     create_table_from_parquet(self.client, unique_database, "out_of_range_date")
@@ -454,6 +467,7 @@ class TestParquet(ImpalaTestSuite):
   @SkipIfIsilon.hive
   @SkipIfLocal.hive
   @SkipIfS3.hive
+  @SkipIfGCS.hive
   def test_multi_compression_types(self, vector, unique_database):
     """IMPALA-5448: Tests that parquet splits with multi compression types are counted
     correctly. Cases tested:
@@ -572,6 +586,7 @@ class TestParquet(ImpalaTestSuite):
         vector, unique_database)
 
   @SkipIfS3.hdfs_block_size
+  @SkipIfGCS.hdfs_block_size
   @SkipIfABFS.hdfs_block_size
   @SkipIfADLS.hdfs_block_size
   @SkipIfIsilon.hdfs_block_size
@@ -629,6 +644,47 @@ class TestParquet(ImpalaTestSuite):
     assert total == num_scanners_with_no_reads
 
   @SkipIfS3.hdfs_block_size
+  @SkipIfGCS.hdfs_block_size
+  @SkipIfABFS.hdfs_block_size
+  @SkipIfADLS.hdfs_block_size
+  @SkipIfIsilon.hdfs_block_size
+  @SkipIfLocal.multiple_impalad
+  @SkipIfEC.fix_later
+  def test_multiple_blocks_mt_dop(self, vector):
+    """Sanity check for MT scan nodes to make sure all blocks from the same file are read.
+    2 scan ranges per node should be created to read 'lineitem_sixblocks' because
+    there are 6 blocks and 3 scan nodes. We set mt_dop to 2, so ideally every instance
+    should read a single range, but since they share a queue its not deterministic and
+    instead we verify sum of ranges read on a backend is 2."""
+    query = 'select count(l_orderkey) from functional_parquet.lineitem_sixblocks'
+    try:
+      self.client.set_configuration_option('mt_dop', '2')
+      result = self.client.execute(query)
+      TOTAL_ROWS = 40000
+      ranges_complete_list = re.findall(r'ScanRangesComplete: ([0-9]*)',
+        result.runtime_profile)
+      num_rows_read_list = re.findall(r'RowsRead: [0-9.K]* \(([0-9]*)\)',
+        result.runtime_profile)
+      # The extra fragment is the "Averaged Fragment"
+      assert len(num_rows_read_list) == 7
+      assert len(ranges_complete_list) == 7
+
+      total_rows_read = 0
+      # Skip the Averaged Fragment; it comes first in the runtime profile.
+      for num_row_read in num_rows_read_list[1:]:
+        total_rows_read += int(num_row_read)
+      assert total_rows_read == TOTAL_ROWS
+
+      # Again skip the Averaged Fragment; it comes first in the runtime profile.
+      # With mt_dop 2, every backend will have 2 instances which are printed consecutively
+      # in the profile.
+      for i in range(1, len(ranges_complete_list), 2):
+        assert int(ranges_complete_list[i]) + int(ranges_complete_list[i + 1]) == 2
+    finally:
+      self.client.clear_configuration()
+
+  @SkipIfS3.hdfs_block_size
+  @SkipIfGCS.hdfs_block_size
   @SkipIfABFS.hdfs_block_size
   @SkipIfADLS.hdfs_block_size
   @SkipIfIsilon.hdfs_block_size
@@ -645,6 +701,7 @@ class TestParquet(ImpalaTestSuite):
     self._multiple_blocks_helper(table_name, 40000, ranges_per_node=2)
 
   @SkipIfS3.hdfs_block_size
+  @SkipIfGCS.hdfs_block_size
   @SkipIfABFS.hdfs_block_size
   @SkipIfADLS.hdfs_block_size
   @SkipIfIsilon.hdfs_block_size
@@ -687,9 +744,9 @@ class TestParquet(ImpalaTestSuite):
 
     # This will fail if the number of impalads != 3
     # The fourth fragment is the "Averaged Fragment"
-    assert len(num_row_groups_list) == 4
-    assert len(scan_ranges_complete_list) == 4
-    assert len(num_rows_read_list) == 4
+    assert len(num_row_groups_list) == 4, result.runtime_profile
+    assert len(scan_ranges_complete_list) == 4, result.runtime_profile
+    assert len(num_rows_read_list) == 4, result.runtime_profile
 
     total_num_row_groups = 0
     # Skip the Averaged Fragment; it comes first in the runtime profile.
@@ -788,6 +845,8 @@ class TestParquet(ImpalaTestSuite):
 
     create_table_from_parquet(self.client, unique_database, 'decimal_stored_as_int32')
     create_table_from_parquet(self.client, unique_database, 'decimal_stored_as_int64')
+    create_table_from_parquet(self.client, unique_database, 'decimal_padded_fixed_len_byte_array')
+    create_table_from_parquet(self.client, unique_database, 'decimal_padded_fixed_len_byte_array2')
 
     self.run_test_case('QueryTest/parquet-decimal-formats', vector, unique_database)
 
@@ -807,6 +866,12 @@ class TestParquet(ImpalaTestSuite):
         "select * from {0}.{1}".format(unique_database, TABLE_NAME))
     assert(len(result.data) == 33)
 
+  def test_rle_dictionary_encoding(self, vector, unique_database):
+    """IMPALA-6434: Add support to decode RLE_DICTIONARY encoded pages."""
+    TABLE_NAME = "alltypes_tiny_rle_dictionary"
+    create_table_from_parquet(self.client, unique_database, TABLE_NAME)
+    self.run_test_case("QueryTest/parquet-rle-dictionary", vector, unique_database)
+
   def test_type_widening(self, vector, unique_database):
     """IMPALA-6373: Test that Impala can read parquet file with column types smaller than
        the schema with larger types"""
@@ -896,9 +961,11 @@ class TestParquet(ImpalaTestSuite):
        when reading [un]compressed Parquet files, and that the counter
        Parquet[Un]compressedPageSize is not updated."""
     # lineitem_sixblocks is not compressed so ParquetCompressedPageSize should be empty,
-    # but ParquetUncompressedPageSize should have been updated
-    result = self.client.execute("select * from functional_parquet.lineitem_sixblocks"
-                                 " limit 10")
+    # but ParquetUncompressedPageSize should have been updated. Query needs an order by
+    # so that all rows are read. Only access a couple of columns to reduce query runtime.
+    result = self.client.execute("select l_orderkey"
+                                 " from functional_parquet.lineitem_sixblocks"
+                                 " order by l_orderkey limit 10")
 
     compressed_page_size_summaries = get_bytes_summary_stats_counter(
         "ParquetCompressedPageSize", result.runtime_profile)
@@ -919,8 +986,9 @@ class TestParquet(ImpalaTestSuite):
 
     # alltypestiny is compressed so both ParquetCompressedPageSize and
     # ParquetUncompressedPageSize should have been updated
-    result = self.client.execute("select * from functional_parquet.alltypestiny"
-                                 " limit 10")
+    # Query needs an order by so that all rows are read.
+    result = self.client.execute("select int_col from functional_parquet.alltypestiny"
+                                 " order by int_col limit 10")
 
     for summary_name in ("ParquetCompressedPageSize", "ParquetUncompressedPageSize"):
       page_size_summaries = get_bytes_summary_stats_counter(
@@ -929,45 +997,60 @@ class TestParquet(ImpalaTestSuite):
       for summary in page_size_summaries:
         assert not self._is_summary_stats_counter_empty(summary)
 
+  @SkipIfNotHdfsMinicluster.tuned_for_minicluster
   def test_bytes_read_per_column(self, vector):
     """IMPALA-6964: Test that the counter Parquet[Un]compressedBytesReadPerColumn is
        updated when reading [un]compressed Parquet files, and that the counter
        Parquet[Un]CompressedBytesReadPerColumn is not updated."""
     # lineitem_sixblocks is not compressed so ParquetCompressedBytesReadPerColumn should
     # be empty, but ParquetUncompressedBytesReadPerColumn should have been updated
-    result = self.client.execute("select * from functional_parquet.lineitem_sixblocks"
-                                 " limit 10")
+    # Query needs an order by so that all rows are read. Only access a couple of
+    # columns to reduce query runtime.
+    result = self.client.execute("select l_orderkey, l_partkey "
+                                 "from functional_parquet.lineitem_sixblocks "
+                                 " order by l_orderkey limit 10")
 
     compressed_bytes_read_per_col_summaries = get_bytes_summary_stats_counter(
         "ParquetCompressedBytesReadPerColumn", result.runtime_profile)
 
-    assert len(compressed_bytes_read_per_col_summaries) > 0
+    # One aggregated counter and three per-instance counters. Agg counter is first.
+    assert len(compressed_bytes_read_per_col_summaries) == 4
     for summary in compressed_bytes_read_per_col_summaries:
       assert self._is_summary_stats_counter_empty(summary)
 
     uncompressed_bytes_read_per_col_summaries = get_bytes_summary_stats_counter(
         "ParquetUncompressedBytesReadPerColumn", result.runtime_profile)
 
-    assert len(uncompressed_bytes_read_per_col_summaries) > 0
-    for summary in uncompressed_bytes_read_per_col_summaries:
-      assert not self._is_summary_stats_counter_empty(summary)
-      # There are 16 columns in lineitem_sixblocks so there should be 16 samples
-      assert summary.total_num_values == 16
+    # One aggregated counter and three per-instance counters. Agg counter is first.
+    assert len(uncompressed_bytes_read_per_col_summaries) == 4
+    for i, summary in enumerate(uncompressed_bytes_read_per_col_summaries):
+      assert not self._is_summary_stats_counter_empty(summary), summary
+      # There are 2 columns read from in lineitem_sixblocks so there should be 2 samples
+      # per instance a 6 total.
+      if i == 0:
+          assert summary.total_num_values == 6
+      else:
+          assert summary.total_num_values == 2
 
     # alltypestiny is compressed so both ParquetCompressedBytesReadPerColumn and
     # ParquetUncompressedBytesReadPerColumn should have been updated
+    # Query needs an order by so that all rows are read.
     result = self.client.execute("select * from functional_parquet.alltypestiny"
-                                 " limit 10")
+                                 " order by int_col limit 10")
 
     for summary_name in ("ParquetCompressedBytesReadPerColumn",
                          "ParquetUncompressedBytesReadPerColumn"):
       bytes_read_per_col_summaries = get_bytes_summary_stats_counter(summary_name,
           result.runtime_profile)
-      assert len(bytes_read_per_col_summaries) > 0
-      for summary in bytes_read_per_col_summaries:
+      # One aggregated counter and three per-instance counters. Agg counter is first.
+      assert len(bytes_read_per_col_summaries) == 4
+      for i, summary in enumerate(bytes_read_per_col_summaries):
         assert not self._is_summary_stats_counter_empty(summary)
-        # There are 11 columns in alltypestiny so there should be 11 samples
-        assert summary.total_num_values == 11
+        # There are 11 columns in alltypestiny so there should be 11 samples per instance.
+        if i == 0:
+          assert summary.total_num_values == 33
+        else:
+          assert summary.total_num_values == 11
 
 
 # We use various scan range lengths to exercise corner cases in the HDFS scanner more
@@ -989,6 +1072,9 @@ class TestScanRangeLengths(ImpalaTestSuite):
     super(TestScanRangeLengths, cls).add_test_dimensions()
     cls.ImpalaTestMatrix.add_dimension(
         ImpalaTestDimension('max_scan_range_length', *MAX_SCAN_RANGE_LENGTHS))
+    # Test doesn't need to be run for non-HDFS table formats.
+    cls.ImpalaTestMatrix.add_constraint(
+        lambda v: not v.get_value('table_format').file_format in ('kudu', 'hbase'))
 
   def test_scan_ranges(self, vector):
     vector.get_value('exec_option')['max_scan_range_length'] =\
@@ -1161,8 +1247,11 @@ class TestTextSplitDelimiters(ImpalaTestSuite):
     max_scan_range_length = DEFAULT_IO_BUFFER_SIZE * 2
     expected_result = data.split("\r\n")
 
+    new_vector = deepcopy(vector)
+    new_vector.get_value('exec_option')['max_row_size'] = 9 * 1024 * 1024
+
     self._create_and_query_test_table(
-      vector, unique_database, data, max_scan_range_length, expected_result)
+      new_vector, unique_database, data, max_scan_range_length, expected_result)
 
   def _create_and_query_test_table(self, vector, unique_database, data,
         max_scan_range_length, expected_result):
@@ -1205,9 +1294,21 @@ class TestTextScanRangeLengths(ImpalaTestSuite):
     self.run_test_case('QueryTest/hdfs-text-scan-with-header', new_vector,
                        test_file_vars={'$UNIQUE_DB': unique_database})
 
+  def test_dateless_timestamp_text(self, vector, unique_database):
+    """Test scanning text files which still includes dateless timestamps."""
+    tbl_name = "timestamp_text_table"
+    create_sql = "create table %s.%s (t timestamp) stored as textfile" % (
+        unique_database, tbl_name)
+    create_table_and_copy_files(self.client, create_sql, unique_database, tbl_name,
+        ["/testdata/data/dateless_timestamps.txt"])
+
+    new_vector = deepcopy(vector)
+    del new_vector.get_value('exec_option')['abort_on_error']
+    self.run_test_case('QueryTest/dateless_timestamp_text', new_vector, unique_database)
 
 # Missing Coverage: No coverage for truncated files errors or scans.
 @SkipIfS3.hive
+@SkipIfGCS.hive
 @SkipIfABFS.hive
 @SkipIfADLS.hive
 @SkipIfIsilon.hive
@@ -1291,6 +1392,7 @@ class TestOrc(ImpalaTestSuite):
       lambda v: v.get_value('table_format').file_format == 'orc')
 
   @SkipIfS3.hdfs_block_size
+  @SkipIfGCS.hdfs_block_size
   @SkipIfABFS.hdfs_block_size
   @SkipIfADLS.hdfs_block_size
   @SkipIfEC.fix_later
@@ -1330,6 +1432,7 @@ class TestOrc(ImpalaTestSuite):
     # lineitem_sixblocks.orc occupies 6 blocks.
     check_call(['hdfs', 'dfs', '-Ddfs.block.size=156672', '-copyFromLocal', '-d', '-f',
         os.environ['IMPALA_HOME'] + "/testdata/LineItemMultiBlock/" + file, tbl_loc])
+    self.client.execute("refresh %s.%s" % (db, tbl))
 
   def _misaligned_orc_stripes_helper(
           self, table_name, rows_in_table, num_scanners_with_no_reads=0):
@@ -1364,7 +1467,9 @@ class TestOrc(ImpalaTestSuite):
   @SkipIfIsilon.hive
   @SkipIfLocal.hive
   @SkipIfS3.hive
-  def test_type_conversions(self, vector, unique_database):
+  @SkipIfGCS.hive
+  @SkipIfHive3.non_acid
+  def test_type_conversions_hive2(self, vector, unique_database):
     # Create "illtypes" tables whose columns can't match the underlining ORC file's.
     # Create an "safetypes" table likes above but ORC columns can still fit into it.
     # Reuse the data files of alltypestiny and date_tbl in funtional_orc_def.
@@ -1403,6 +1508,68 @@ class TestOrc(ImpalaTestSuite):
 
     self.run_test_case('DataErrorsTest/orc-type-checks', vector, unique_database)
 
+  # Skip this test on non-HDFS filesystems, because orc-type-check.test contains Hive
+  # queries that hang in some cases (IMPALA-9345). It would be possible to separate
+  # the tests that use Hive and run most tests on S3, but I think that running these on
+  # S3 doesn't add too much coverage.
+  @SkipIfABFS.hive
+  @SkipIfADLS.hive
+  @SkipIfIsilon.hive
+  @SkipIfLocal.hive
+  @SkipIfS3.hive
+  @SkipIfGCS.hive
+  @SkipIfHive2.acid
+  def test_type_conversions_hive3(self, vector, unique_database):
+    # Create "illtypes" tables whose columns can't match the underlining ORC file's.
+    # Create an "safetypes" table likes above but ORC columns can still fit into it.
+    # Reuse the data files of alltypestiny and date_tbl in funtional_orc_def.
+    def create_plain_orc_table(fq_tbl_src, fq_tbl_dest):
+      self.run_stmt_in_hive(
+          "create table %s like %s stored as orc" % (fq_tbl_dest, fq_tbl_src))
+      self.run_stmt_in_hive("insert into %s select * from %s" % (fq_tbl_dest, fq_tbl_src))
+      self.client.execute("invalidate metadata %s" % fq_tbl_dest)
+    tmp_alltypes = unique_database + ".alltypes"
+    create_plain_orc_table("functional.alltypestiny", tmp_alltypes)
+    tbl_loc = self._get_table_location(tmp_alltypes, vector)
+    self.client.execute("""create table %s.illtypes (c1 boolean, c2 float,
+        c3 boolean, c4 tinyint, c5 smallint, c6 int, c7 boolean, c8 string, c9 int,
+        c10 float, c11 bigint) partitioned by (year int, month int) stored as ORC
+        location '%s'""" % (unique_database, tbl_loc))
+    self.client.execute("""create table %s.illtypes_ts_to_date (c1 boolean,
+        c2 float, c3 boolean, c4 tinyint, c5 smallint, c6 int, c7 boolean, c8 string,
+        c9 int, c10 float, c11 date) partitioned by (year int, month int) stored as ORC
+        location '%s'""" % (unique_database, tbl_loc))
+    self.client.execute("""create table %s.safetypes (c1 bigint, c2 boolean,
+        c3 smallint, c4 int, c5 bigint, c6 bigint, c7 double, c8 double, c9 char(3),
+        c10 varchar(3), c11 timestamp) partitioned by (year int, month int) stored as ORC
+        location '%s'""" % (unique_database, tbl_loc))
+    tmp_date_tbl = unique_database + ".date_tbl"
+    create_plain_orc_table("functional.date_tbl", tmp_date_tbl)
+    date_tbl_loc = self._get_table_location(tmp_date_tbl, vector)
+    self.client.execute("""create table %s.illtypes_date_tbl (c1 boolean,
+        c2 timestamp) partitioned by (date_part date) stored as ORC location '%s'"""
+        % (unique_database, date_tbl_loc))
+    self.client.execute("alter table %s.illtypes recover partitions" % unique_database)
+    self.client.execute("alter table %s.illtypes_ts_to_date recover partitions"
+        % unique_database)
+    self.client.execute("alter table %s.safetypes recover partitions" % unique_database)
+    self.client.execute("alter table %s.illtypes_date_tbl recover partitions"
+        % unique_database)
+
+    # Create a decimal table whose precisions don't match the underlining orc files.
+    # Reuse the data files of functional_orc_def.decimal_tbl.
+    tmp_decimal_tbl = unique_database + ".decimal_tbl"
+    create_plain_orc_table("functional.decimal_tbl", tmp_decimal_tbl)
+    decimal_loc = self._get_table_location(tmp_decimal_tbl, vector)
+    self.client.execute("""create table %s.mismatch_decimals (d1 decimal(8,0),
+        d2 decimal(8,0), d3 decimal(19,10), d4 decimal(20,20), d5 decimal(2,0))
+        partitioned by (d6 decimal(9,0)) stored as orc location '%s'"""
+        % (unique_database, decimal_loc))
+    self.client.execute("alter table %s.mismatch_decimals recover partitions"
+        % unique_database)
+
+    self.run_test_case('DataErrorsTest/orc-type-checks', vector, unique_database)
+
   def test_orc_timestamp_out_of_range(self, vector, unique_database):
       """Test the validation of out-of-range timestamps."""
       test_files = ["testdata/data/out_of_range_timestamp.orc"]
@@ -1432,7 +1599,7 @@ class TestOrc(ImpalaTestSuite):
     self._run_invalid_schema_test(unique_database, "corrupt_schema",
         "Encountered parse error during schema selection")
     self._run_invalid_schema_test(unique_database, "corrupt_root_type",
-        "Root of the file type returned by the ORC lib is not STRUCT: boolean.")
+        "Root of the selected type returned by the ORC lib is not STRUCT: boolean.")
 
 
   def test_date_out_of_range_orc(self, vector, unique_database):
@@ -1460,31 +1627,6 @@ class TestOrc(ImpalaTestSuite):
 
     self.run_test_case('QueryTest/hive2-pre-gregorian-date-orc', vector, unique_database)
 
-  @SkipIfABFS.hive
-  @SkipIfADLS.hive
-  @SkipIfIsilon.hive
-  @SkipIfLocal.hive
-  @SkipIfS3.hive
-  def test_missing_field_orc(self, unique_database):
-    """Test scanning orc files with missing fields in file meta."""
-    orc_tbl_name = unique_database + ".missing_field_orc"
-    self.client.execute("create table %s (f0 int) stored as orc" % orc_tbl_name)
-    self.run_stmt_in_hive("insert into table %s select 1" % orc_tbl_name)
-    self.client.execute("refresh %s" % orc_tbl_name)
-    self.client.execute("alter table %s add columns(f1 int)" % orc_tbl_name)
-    result = self.client.execute("select f1 from %s " % orc_tbl_name)
-    assert result.data == ['NULL']
-
-    orc_tbl_name = unique_database + ".lineitem_orc_ext"
-    test_file = "/test-warehouse/tpch.lineitem_orc_def"
-    create_sql = "create external table %s like tpch_orc_def.lineitem " \
-                 "location '%s'" % (orc_tbl_name, test_file)
-    self.client.execute(create_sql)
-    self.client.execute("alter table %s add columns (new_col int)" % orc_tbl_name)
-    result = self.execute_query("select count(*) from %s where new_col is null"
-                                % orc_tbl_name)
-    assert len(result.data) == 1
-    assert "6001215" in result.data
 
 class TestScannerReservation(ImpalaTestSuite):
   @classmethod
