@@ -17,6 +17,7 @@
 
 from copy import copy
 import os
+import re
 import pytest
 import tempfile
 from subprocess import call, check_call
@@ -355,7 +356,7 @@ class TestUdfExecution(TestUdfBase):
     # admission control and a simple scan can run.
     vector = copy(vector)
     vector.get_value('exec_option')['mem_limit'] = '1mb'
-    vector.get_value('exec_option')['buffer_pool_limit'] = '32kb'
+    vector.get_value('exec_option')['buffer_pool_limit'] = '4.04mb'
     try:
       self.run_test_case('QueryTest/udf-mem-limit', vector, use_db=unique_database)
       assert False, "Query was expected to fail"
@@ -516,7 +517,7 @@ class TestUdfTargeted(TestUdfBase):
     old_udf = os.path.join(
         os.environ['IMPALA_HOME'], 'testdata/udfs/impala-hive-udfs.jar')
     new_udf = os.path.join(
-        os.environ['IMPALA_HOME'], 'tests/test-hive-udfs/target/test-hive-udfs-1.0.jar')
+        os.environ['IMPALA_HOME'], 'testdata/udfs/impala-hive-udfs-modified.jar')
     udf_dst = get_fs_path(
         '/test-warehouse/{0}.db/impala-hive-udfs.jar'.format(unique_database))
 
@@ -549,7 +550,7 @@ class TestUdfTargeted(TestUdfBase):
     old_udf = os.path.join(
         os.environ['IMPALA_HOME'], 'testdata/udfs/impala-hive-udfs.jar')
     new_udf = os.path.join(
-        os.environ['IMPALA_HOME'], 'tests/test-hive-udfs/target/test-hive-udfs-1.0.jar')
+        os.environ['IMPALA_HOME'], 'testdata/udfs/impala-hive-udfs-modified.jar')
     udf_dst = get_fs_path(
         '/test-warehouse/{0}.db/impala-hive-udfs.jar'.format(unique_database))
     old_function_name = "udf_update_test_create1"
@@ -614,3 +615,20 @@ class TestUdfTargeted(TestUdfBase):
     results = self.client.fetch(query, handle, -1)
     assert results.success
     assert len(results.data) == 9999
+
+  def test_udf_profile(self, vector, unique_database):
+    """Test to validate that explain plans and runtime profiles contain information about
+    any custom UDFs used in an Impala query."""
+    self.client.execute(
+        "create function {0}.hive_substring(string, int) returns string location '{1}' "
+        "symbol='org.apache.hadoop.hive.ql.udf.UDFSubstr'".format(
+            unique_database, get_fs_path('/test-warehouse/hive-exec.jar')))
+    profile = self.execute_query_expect_success(self.client,
+        "select {0}.hive_substring(string_col, 1), {0}.hive_substring(string_col, 2) "
+        "from functional.alltypes limit 10".format(unique_database)).runtime_profile
+
+    assert re.search("output exprs.*hive_substring.*/\* JAVA UDF \*/", profile)
+    # Ensure that hive_substring only shows up once in the list of UDFs.
+    assert re.search(
+        "User Defined Functions \(UDFs\): {0}\.hive_substring\s*[\r\n]".format(
+            unique_database), profile)
