@@ -1,4 +1,5 @@
-# encoding=utf-8
+#!/usr/bin/env impala-python
+# -*- coding: utf-8 -*-
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -18,6 +19,7 @@
 # under the License.
 
 import errno
+import getpass
 import os
 import pytest
 import re
@@ -35,7 +37,8 @@ from tests.common.skip import SkipIf
 from tests.common.test_dimensions import create_client_protocol_dimension
 from time import sleep, time
 from util import (get_impalad_host_port, assert_var_substitution, run_impala_shell_cmd,
-                  ImpalaShell, IMPALA_SHELL_EXECUTABLE)
+                  ImpalaShell, IMPALA_SHELL_EXECUTABLE, SHELL_IS_PYTHON_2,
+                  build_shell_env, wait_for_query_state)
 from contextlib import closing
 
 
@@ -102,6 +105,16 @@ def populated_table(empty_table, request):
   return fq_table_name
 
 
+@pytest.yield_fixture
+def tmp_file():
+  """
+  Test fixture which manages a temporary file
+  """
+  _, tmp_file = tempfile.mkstemp()
+  yield tmp_file
+  os.remove(tmp_file)
+
+
 class TestImpalaShell(ImpalaTestSuite):
   """A set of sanity tests for the Impala shell commandline parameters.
 
@@ -134,8 +147,6 @@ class TestImpalaShell(ImpalaTestSuite):
 
   def test_multiple_queries_with_escaped_backslash(self, vector):
     """Regression test for string containing an escaped backslash.
-
-    This relies on the patch at thirdparty/patches/sqlparse/0001-....patch.
     """
     run_impala_shell_cmd(vector, ['-q', r'''select '\\'; select '\'';''', '-B'])
 
@@ -157,7 +168,7 @@ class TestImpalaShell(ImpalaTestSuite):
 
   def test_unsecure_message(self, vector):
     results = run_impala_shell_cmd(vector, [], wait_until_connected=False)
-    assert "Starting Impala Shell without Kerberos authentication" in results.stderr
+    assert "with no authentication" in results.stderr
 
   def test_print_header(self, vector, populated_table):
     args = ['--print_header', '-B', '--output_delim=,', '-q',
@@ -186,7 +197,7 @@ class TestImpalaShell(ImpalaTestSuite):
     results = run_impala_shell_cmd(vector, args, expect_success=False)
     # Check that impala is using the right service name.
     assert "Using service name 'impala'" in results.stderr
-    assert "Starting Impala Shell using Kerberos authentication" in results.stderr
+    assert "with Kerberos authentication" in results.stderr
     # Check that Impala warns the user if klist does not exist on the system, or if
     # no kerberos tickets are initialized.
     try:
@@ -255,7 +266,7 @@ class TestImpalaShell(ImpalaTestSuite):
   def test_removed_query_option(self, vector):
     """Test that removed query options produce warning."""
     result = run_impala_shell_cmd(vector, ['-q', 'set disable_cached_reads=true'],
-        expect_success=True)
+                                  expect_success=True)
     assert "Ignoring removed query option: 'disable_cached_reads'" in result.stderr,\
         result.stderr
 
@@ -364,7 +375,7 @@ class TestImpalaShell(ImpalaTestSuite):
     sleep(5)
     assert 1 == impalad_service.get_num_in_flight_queries()
     assert p.get_result().rc == 0
-    assert 0 == impalad_service.get_num_in_flight_queries()
+    assert impalad_service.wait_for_num_in_flight_queries(0)
 
   def test_cancellation(self, vector):
     """Test cancellation (Ctrl+C event). Run a query that sleeps 10ms per row so will run
@@ -395,46 +406,25 @@ class TestImpalaShell(ImpalaTestSuite):
   def test_query_cancellation_during_wait_to_finish(self, vector):
     """IMPALA-1144: Test cancellation (CTRL+C) while the query is in the
     wait_to_finish state"""
+
     # A select where wait_to_finish takes several seconds
     stmt = "select * from tpch.customer c1, tpch.customer c2, " + \
            "tpch.customer c3 order by c1.c_name"
     # Kill happens in wait_to_finish state
     self.run_and_verify_query_cancellation_test(vector, stmt, "RUNNING")
 
-  def wait_for_query_state(self, vector, stmt, state, max_retry=15):
-    """Checks the in flight queries on Impala debug page. Polls the state of
-    the query statement from parameter every second until the query gets to
-    a state given via parameter or a maximum retry count is reached.
-    Restriction: Only works if there is only one in flight query."""
-    impalad_service = ImpaladService(get_impalad_host_port(vector).split(':')[0])
-    if not impalad_service.wait_for_num_in_flight_queries(1):
-      raise Exception("No in flight query found")
-
-    retry_count = 0
-    while retry_count <= max_retry:
-      query_info = impalad_service.get_in_flight_queries()[0]
-      if query_info['stmt'] != stmt:
-        exc_text = "The found in flight query is not the one under test: " + \
-            query_info['stmt']
-        raise Exception(exc_text)
-      if query_info['state'] == state:
-        return
-      retry_count += 1
-      sleep(1.0)
-    raise Exception("Query didn't reach desired state: " + state)
-
   def run_and_verify_query_cancellation_test(self, vector, stmt, cancel_at_state):
     """Starts the execution of the received query, waits until the query
     execution in fact starts and then cancels it. Expects the query
     cancellation to succeed."""
     p = ImpalaShell(vector, ['-q', stmt])
 
-    self.wait_for_query_state(vector, stmt, cancel_at_state)
+    wait_for_query_state(vector, stmt, cancel_at_state)
 
     os.kill(p.pid(), signal.SIGINT)
     result = p.get_result()
     assert "Cancelling Query" in result.stderr
-    assert "Invalid query handle" not in result.stderr
+    assert "Invalid or unknown query handle" not in result.stderr
 
   def test_get_log_once(self, vector, empty_table):
     """Test that get_log() is always called exactly once."""
@@ -474,7 +464,9 @@ class TestImpalaShell(ImpalaTestSuite):
   def test_international_characters_prettyprint_tabs(self, vector):
     """IMPALA-2717: ensure we can handle international characters in pretty-printed
     output when pretty-printing falls back to delimited output."""
+
     args = ['-q', "select '{0}\\t'".format(RUSSIAN_CHARS.encode('utf-8'))]
+
     result = run_impala_shell_cmd(vector, args)
     protocol = vector.get_value('protocol')
     if protocol == 'beeswax':
@@ -486,6 +478,26 @@ class TestImpalaShell(ImpalaTestSuite):
     assert 'UnicodeDecodeError' not in result.stderr
     assert RUSSIAN_CHARS.encode('utf-8') in result.stdout
 
+  def test_utf8_decoding_error_handling(self, vector):
+    """IMPALA-10145,IMPALA-10299: Regression tests for elegantly handling malformed utf-8
+    characters."""
+    result = run_impala_shell_cmd(vector, ['-B', '-q', "select substr('引擎', 1, 4)"])
+    assert 'UnicodeDecodeError' not in result.stderr
+    # Thrift changes its internal strings representation from bytes to unicodes since
+    # 0.10.0. The results differ when impala-shell uses different versions of Thrift.
+    # The UTF-8 encoded bytes of "引擎" are \xe5\xbc\x95\xe6\x93\x8e. The substr result
+    # gets the first 4 bytes. In thrift-0.9.3-p8, it will be raw bytes, i.e. "引\xe6".
+    # In thrift-0.11.0-p4, it will be decoded to utf-8 strings. The last byte can't be
+    # decoded correctly so it will be replaced to \xef\xbf\xbd, i.e. U+FFFD. The result
+    # is "引\xef\xbf\xbd". To make this test robust in all branches, here we just check
+    # the existense of "引".
+    assert '引' in result.stdout
+    result = run_impala_shell_cmd(vector, ['-B', '-q', "select unhex('aa')"])
+    assert 'UnicodeDecodeError' not in result.stderr
+    # Same as above, the result using thrift <0.10.0 is '\xaa'. The result using
+    # thrift >=0.10.0 is '\xef\xbf\xbd'.
+    assert '\xef\xbf\xbd' in result.stdout or '\xaa' in result.stdout
+
   def test_global_config_file(self, vector):
     """Test global and user configuration files."""
     args = []
@@ -530,6 +542,7 @@ class TestImpalaShell(ImpalaTestSuite):
     assert 'WARNING:' not in result.stderr, \
       "A valid config file should not trigger any warning: {0}".format(result.stderr)
     assert 'Query: select 1' in result.stderr
+
     # override option in config file through command line
     args = ['--config_file={0}/good_impalarc'.format(QUERY_FILE_PATH), '--query=select 2']
     result = run_impala_shell_cmd(vector, args)
@@ -557,6 +570,7 @@ class TestImpalaShell(ImpalaTestSuite):
     # specified config file does not exist
     args = ['--config_file=%s/does_not_exist' % QUERY_FILE_PATH]
     run_impala_shell_cmd(vector, args, expect_success=False)
+
     # bad formatting of config file
     args = ['--config_file=%s/bad_impalarc' % QUERY_FILE_PATH]
     run_impala_shell_cmd(vector, args, expect_success=False)
@@ -633,14 +647,23 @@ class TestImpalaShell(ImpalaTestSuite):
 
   def test_ldap_password_from_shell(self, vector):
     args = ['-l', '--auth_creds_ok_in_clear']
-    result = run_impala_shell_cmd(vector, args + ['--ldap_password_cmd=cmddoesntexist'],
-                                  expect_success=False)
-    assert ("Error retrieving LDAP password (command was: 'cmddoesntexist', exception "
-            "was: '[Errno 2] No such file or directory')") in result.stderr
-    result = run_impala_shell_cmd(
-        vector, args + ['--ldap_password_cmd=cat filedoesntexist'], expect_success=False)
-    assert ("Error retrieving LDAP password (command was 'cat filedoesntexist', error "
-            "was: 'cat: filedoesntexist: No such file or directory')") in result.stderr
+
+    result_1 = run_impala_shell_cmd(vector, args + ['--ldap_password_cmd=cmddoesntexist'],
+                                    expect_success=False)
+
+    assert "Error retrieving LDAP password" in result_1.stderr
+    assert "command was: 'cmddoesntexist'" in result_1.stderr
+    # On GCE instances, the error thrown in subprocess is "[Errno 13] Permission denied".
+    assert "No such file or directory" in result_1.stderr \
+        or "Permission denied" in result_1.stderr
+
+    result_2 = run_impala_shell_cmd(vector,
+                                    args + ['--ldap_password_cmd=cat filedoesntexist'],
+                                    expect_success=False)
+
+    assert "Error retrieving LDAP password" in result_2.stderr
+    assert "command was 'cat filedoesntexist'" in result_2.stderr
+    assert "No such file or directory" in result_2.stderr
 
     # TODO: Without an Impala daemon with LDAP authentication enabled, we can't test the
     # positive case where the password is correct.
@@ -742,7 +765,6 @@ class TestImpalaShell(ImpalaTestSuite):
         (expected_rows_modified, expected_row_errors)
     assert expected_output in results.stderr, results.stderr
 
-  @SkipIf.kudu_not_supported
   def test_kudu_dml_reporting(self, vector, unique_database):
     db = unique_database
     run_impala_shell_cmd(vector, [
@@ -789,7 +811,8 @@ class TestImpalaShell(ImpalaTestSuite):
     with open(os.devnull, 'w') as devnull:
       try:
         connection = None
-        impala_shell = Popen(shell_cmd + args, stdout=devnull, stderr=devnull)
+        impala_shell = Popen(shell_cmd + args, stdout=devnull, stderr=devnull,
+                             env=build_shell_env())
         # IMPALA-9547: retry accept(). This is required in Python < 3.5 because some
         # EINTR return calls from syscalls are not automatically retried. See PEP475.
         while True:
@@ -991,5 +1014,90 @@ class TestImpalaShell(ImpalaTestSuite):
     assert "0\tNULL\tNULL" in result.stdout, result.stdout
     assert "1\t1\t10.1" in result.stdout, result.stdout
     assert "2\t2\t20.2" in result.stdout, result.stdout
-    assert "3\t3\t30.3" in result.stdout, result.stdout
+
+    if (vector.get_value("protocol") in ('hs2', 'hs2-http')) and not SHELL_IS_PYTHON_2:
+      # The HS2 client returns binary values for float/double types, and these must
+      # be converted to strings for display. However, due to differences between the
+      # way that python2 and python3 represent floating point values, the output
+      # from the shell will differ with regard to which version of python the
+      # shell is running under.
+      assert "3\t3\t30.299999999999997" in result.stdout, result.stdout
+    else:
+      # python 2, or python 3 with beeswax protocol
+      assert "3\t3\t30.3" in result.stdout, result.stdout
+
     assert "4\t4\t40.4" in result.stdout, result.stdout
+
+  def test_fetch_size(self, vector):
+    """Test the --fetch_size option with and without result spooling enabled."""
+    query = "select * from functional.alltypes limit 1024"
+    query_with_result_spooling = "set spool_query_results=true; " + query
+    for query in [query, query_with_result_spooling]:
+      result = run_impala_shell_cmd(vector, ['-q', query, '-B', '--fetch_size', '512'])
+      result_rows = result.stdout.strip().split('\n')
+      assert len(result_rows) == 1024
+
+  def test_result_spooling_timeout(self, vector):
+    """Regression test for IMPALA-9953. Validates that if a fetch timeout occurs in the
+    middle of reading rows from Impala that all rows are still printed by the Impala
+    shell."""
+    # This query was stolen from __test_fetch_timeout in test_fetch_timeout.py. The query
+    # has a large delay between RowBatch production. So a fetch timeout will occur while
+    # fetching rows.
+    query_options = "set num_nodes=1; \
+                     set fetch_rows_timeout_ms=1; \
+                     set batch_size=1; \
+                     set spool_query_results=true;"
+    query = "select bool_col, avg(id) from functional.alltypes group by bool_col"
+    result = run_impala_shell_cmd(vector, ['-q', query_options + query, '-B'])
+    result_rows = result.stdout.strip().split('\n')
+    assert len(result_rows) == 2
+
+  def test_quiet_mode(self, vector):
+    """Checks that extraneous output isn't included when --quiet is set."""
+    args = ['--quiet', '-q', 'select 1']
+    result = run_impala_shell_cmd(vector, args)
+    expected_result = """+---+\n| 1 |\n+---+\n| 1 |\n+---+\n"""
+    assert result.stdout == expected_result
+    assert result.stderr == ""
+
+  def test_user_flag(self, vector):
+    """Check that the --user flag is respected. This test assumes we are running against
+    an unsecured cluster and can impersonate any user when connecting."""
+    base_args = ['--quiet', '-B', '-q', 'select user(), effective_user()']
+
+    # Test that default user is the user that impala-shell is run under.
+    result = run_impala_shell_cmd(vector, base_args)
+    expected_result = """{0}\t{0}\n""".format(getpass.getuser())
+    assert result.stdout == expected_result
+    assert result.stderr == ""
+
+    # Test that we can set an arbitrary user.
+    result = run_impala_shell_cmd(vector, base_args + ['--user=mickey mouse'])
+    expected_result = """mickey mouse\tmickey mouse\n"""
+    assert result.stdout == expected_result
+    assert result.stderr == ""
+
+    # Test that empty user results in "anonymous" - see IMPALA-10027.
+    result = run_impala_shell_cmd(vector, base_args + ['--user='])
+    expected_result = """anonymous\tanonymous\n"""
+    assert result.stdout == expected_result
+    assert result.stderr == ""
+
+  def test_output_file(self, vector, tmp_file):
+    """Test that writing output to a file using '--output_file' produces the same output
+    as is written to stdout."""
+    row_count = 6000  # Should be > 2048 to tickle IMPALA-10447.
+    query = "select * from tpcds.item order by i_item_sk limit %d" % row_count
+    # Run the query normally and keep the stdout.
+    output = run_impala_shell_cmd(vector, ['-q', query, '-B', '--output_delimiter=;'])
+    assert "Fetched %d row(s)" % row_count in output.stderr
+    rows_from_stdout = output.stdout.strip().split('\n')
+    # Run the query with output sent to a file using '--output_file'.
+    result = run_impala_shell_cmd(vector, ['-q', query, '-B', '--output_delimiter=;',
+                                           '--output_file=%s' % tmp_file])
+    assert "Fetched %d row(s)" % row_count in result.stderr
+    # Check that the output from the file is the same as that written to stdout.
+    with open(tmp_file, "r") as f:
+      rows_from_file = [line.rstrip() for line in f]
+      assert rows_from_stdout == rows_from_file
