@@ -1,5 +1,5 @@
 #!/usr/bin/env impala-python
-# encoding=utf-8
+# -*- coding: utf-8 -*-
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -29,7 +29,6 @@ import socket
 import sys
 import threading
 from time import sleep
-from contextlib import closing
 
 # This import is the actual ImpalaShell class from impala_shell.py.
 # We rename it to ImpalaShellClass here because we later import another
@@ -42,8 +41,9 @@ from tests.common.impala_service import ImpaladService
 from tests.common.impala_test_suite import ImpalaTestSuite
 from tests.common.skip import SkipIfLocal
 from tests.common.test_dimensions import create_client_protocol_dimension
+from tests.shell.util import get_unused_port
 from util import (assert_var_substitution, ImpalaShell, get_impalad_port, get_shell_cmd,
-                  get_open_sessions_metric, IMPALA_SHELL_EXECUTABLE)
+                  get_open_sessions_metric, IMPALA_SHELL_EXECUTABLE, spawn_shell)
 import SimpleHTTPServer
 import SocketServer
 
@@ -55,6 +55,7 @@ PROMPT_REGEX = r'\[[^:]+:2(1|8)0[0-9][0-9]\]'
 
 LOG = logging.getLogger('test_shell_interactive')
 
+
 @pytest.fixture
 def tmp_history_file(request):
   """
@@ -75,18 +76,83 @@ def tmp_history_file(request):
   return tmp.name
 
 
-class UnavailableRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
-  """An HTTP server that always returns 503"""
+class RequestHandler503(SimpleHTTPServer.SimpleHTTPRequestHandler):
+  """A custom http handler that checks for duplicate 'Host' headers from the most
+  recent http request, and always returns a 503 http code."""
+
+  def __init__(self, request, client_address, server):
+    SimpleHTTPServer.SimpleHTTPRequestHandler.__init__(self, request, client_address,
+                                                       server)
+
+  def should_send_body_text(self):
+    # in RequestHandler503 we do not send any body text
+    return False
+
   def do_POST(self):
+    # The unfortunately named self.headers here is an instance of mimetools.Message that
+    # contains the request headers.
+    request_headers = self.headers.headers
+
+    # Ensure that only one 'Host' header is contained in the request before responding.
+    host_hdr_count = sum([header.startswith('Host:') for header in request_headers])
+    assert host_hdr_count == 1, "duplicate 'Host:' headers in %s" % request_headers
+
+    # Respond with 503.
     self.send_response(code=httplib.SERVICE_UNAVAILABLE, message="Service Unavailable")
+    if self.should_send_body_text():
+      # Optionally send ody text with 503 message.
+      self.end_headers()
+      self.wfile.write("EXTRA")
 
 
-def get_unused_port():
-  """ Find an unused port http://stackoverflow.com/questions/1365265 """
-  with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
-    s.bind(('', 0))
-    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-    return s.getsockname()[1]
+class RequestHandler503Extra(RequestHandler503):
+  """"Override RequestHandler503 so as to send body text with the 503 message."""
+
+  def __init__(self, request, client_address, server):
+    RequestHandler503.__init__(self, request, client_address, server)
+
+  def should_send_body_text(self):
+    # in RequestHandler503Extra we will send body text
+    return True
+
+
+class TestHTTPServer503(object):
+  def __init__(self, clazz):
+    self.HOST = "localhost"
+    self.PORT = get_unused_port()
+    self.httpd = SocketServer.TCPServer((self.HOST, self.PORT), clazz)
+
+    self.http_server_thread = threading.Thread(target=self.httpd.serve_forever)
+    self.http_server_thread.start()
+
+
+def shutdown_server(server):
+  """Helper method to shutdown a http server."""
+  if server.httpd is not None:
+    server.httpd.shutdown()
+  if server.http_server_thread is not None:
+    server.http_server_thread.join()
+
+
+@pytest.yield_fixture
+def http_503_server():
+  """A fixture that creates an http server that returns a 503 http code."""
+  server = TestHTTPServer503(RequestHandler503)
+  yield server
+
+  # Cleanup after test.
+  shutdown_server(server)
+
+
+@pytest.yield_fixture
+def http_503_server_extra():
+  """A fixture that creates an http server that returns a 503 http code with extra
+  body text."""
+  server = TestHTTPServer503(RequestHandler503Extra)
+  yield server
+
+  # Cleanup after test.
+  shutdown_server(server)
 
 
 class TestImpalaShellInteractive(ImpalaTestSuite):
@@ -116,14 +182,13 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     try:
       actual = impala_service.wait_for_metric_value(metric_name, expected)
     except AssertionError:
-      LOG.exception("Error: " % err)
+      LOG.exception("Error: %s" % err)
       raise
     assert actual == expected, err
 
   def test_local_shell_options(self, vector):
     """Test that setting the local shell options works"""
-    shell_cmd = get_shell_cmd(vector)
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+    proc = spawn_shell(get_shell_cmd(vector))
     proc.expect(":{0}] default>".format(get_impalad_port(vector)))
     self._expect_with_cmd(proc, "set", vector,
         ("LIVE_PROGRESS: True", "LIVE_SUMMARY: False"))
@@ -156,7 +221,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     p.send_cmd("select * from nation")
     result = p.get_result()
     assert "+----------------+" not in result.stdout
-    assert "21\tVIETNAM\t2" in result.stdout
+    assert "21\tVIETNAM\t2" in result.stdout, result.stdout
 
   @pytest.mark.execute_serially
   def test_change_delimiter(self, vector):
@@ -241,45 +306,53 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
 
   @pytest.mark.execute_serially
   def test_cancellation_mid_command(self, vector):
-    """The test starts with sending in a multi-line input without a command delimiter.
-       When the impala-shell is waiting for more input, the test sends a SIGINT signal (to
-       simulate pressing Ctrl-C) followed by a final query terminated with semicolon.
-       The expected behavior for the impala shell is to discard everything before the
-       SIGINT signal was sent and execute the final query only."""
+    """Test that keyboard interrupt cancels multiline query strings"""
     shell_cmd = get_shell_cmd(vector)
-    queries = [
-        "line 1\n", "line 2\n", "line 3\n\n", "line 4 and", " 5\n",
-        "line 6\n", "line 7\n", "line 8\n", "line 9\n", "line 10"]
-    # Check when the last line before Ctrl-C doesn't end with newline.
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
-    for query in queries:
-      child_proc.send(query)
+    multiline_query = ["select column_1\n", "from table_1\n", "where ..."]
+    # Test keyboard interrupt mid line
+    child_proc = spawn_shell(shell_cmd)
+    child_proc.expect(PROMPT_REGEX)
+    for query_line in multiline_query:
+      child_proc.send(query_line)
     child_proc.sendintr()
-    child_proc.send('select "test without newline";\n')
-    child_proc.expect("test without newline")
+    child_proc.expect("\^C")
+    child_proc.expect(PROMPT_REGEX)
     child_proc.sendline('quit;')
     child_proc.wait()
-    # Check when the last line before Ctrl-C ends with newline.
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
-    for query in queries:
-      child_proc.send(query)
-    # Sending in a newline so it will end with one
+    # Test keyboard interrupt in new line
+    child_proc = spawn_shell(shell_cmd)
+    child_proc.expect(PROMPT_REGEX)
+    for query_line in multiline_query:
+      child_proc.send(query_line)
     child_proc.send("\n")
-    # checking if it realy is a new line
-    child_proc.expect(" > ")
+    child_proc.expect(">")
     child_proc.sendintr()
-    child_proc.send('select "test with newline";\n')
-    child_proc.expect("test with newline")
+    child_proc.expect("> \^C")
+    child_proc.expect(PROMPT_REGEX)
     child_proc.sendline('quit;')
     child_proc.wait()
 
   def test_unicode_input(self, vector):
-    "Test queries containing non-ascii input"
+    """Test queries containing non-ascii input"""
     # test a unicode query spanning multiple lines
-    unicode_text = u'\ufffd'
-    args = "select '%s'\n;" % unicode_text.encode('utf-8')
+    unicode_bytes = u'\ufffd'.encode('utf-8')
+    args = "select '{0}'\n;".format(unicode_bytes)
     result = run_impala_shell_interactive(vector, args)
-    assert "Fetched 1 row(s)" in result.stderr
+    assert "Fetched 1 row(s)" in result.stderr, result.stderr
+    # IMPALA-10415: Multi queries in a line containing unicode chars.
+    args = "select '{0}'; select '{0}'".format(unicode_bytes)
+    result = run_impala_shell_interactive(vector, args)
+    assert "Fetched 1 row(s)" in result.stderr, result.stderr
+    # IMPALA-10415: Multiline query with history enabled and unicode chars.
+    # readline gets its input from tty, so using stdin does not work.
+    shell_cmd = get_shell_cmd(vector)
+    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+    child_proc.expect(PROMPT_REGEX)
+    child_proc.sendline("select '{0}'\n;".format(unicode_bytes))
+    child_proc.expect("Fetched 1 row\(s\) in [0-9]+\.?[0-9]*s")
+    child_proc.expect(PROMPT_REGEX)
+    child_proc.sendline('quit;')
+    child_proc.wait()
 
   def test_welcome_string(self, vector):
     """Test that the shell's welcome message is only printed once
@@ -333,7 +406,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
       hostname = socket.getfqdn()
       initial_impala_service = ImpaladService(hostname)
       target_impala_service = ImpaladService(hostname, webserver_port=25001,
-          beeswax_port=21001, be_port=22001, hs2_port=21051, hs2_http_port=28001)
+          beeswax_port=21001, hs2_port=21051, hs2_http_port=28001)
       protocol = vector.get_value("protocol").lower()
       if protocol == "hs2":
         target_port = 21051
@@ -378,6 +451,8 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     webpage to confirm that they've been closed.
     TODO: Add every statement type.
     """
+    # IMPALA-10312: deflake the test by increasing timeout above the default
+    TIMEOUT_S = 30
     # Disconnect existing clients so there are no open sessions.
     self.close_impala_clients()
 
@@ -395,14 +470,14 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
       p.send_cmd('create database if not exists %s' % TMP_DB)
       p.send_cmd('use %s' % TMP_DB)
       impalad.wait_for_metric_value(NUM_QUERIES, start_num_queries + 2)
-      assert impalad.wait_for_num_in_flight_queries(0), MSG % 'use'
+      assert impalad.wait_for_num_in_flight_queries(0, TIMEOUT_S), MSG % 'use'
       p.send_cmd('create table %s(i int)' % TMP_TBL)
       p.send_cmd('alter table %s add columns (j int)' % TMP_TBL)
       impalad.wait_for_metric_value(NUM_QUERIES, start_num_queries + 4)
-      assert impalad.wait_for_num_in_flight_queries(0), MSG % 'alter'
+      assert impalad.wait_for_num_in_flight_queries(0, TIMEOUT_S), MSG % 'alter'
       p.send_cmd('drop table %s' % TMP_TBL)
       impalad.wait_for_metric_value(NUM_QUERIES, start_num_queries + 5)
-      assert impalad.wait_for_num_in_flight_queries(0), MSG % 'drop'
+      assert impalad.wait_for_num_in_flight_queries(0, TIMEOUT_S), MSG % 'drop'
     finally:
       # get_result() must be called to exit the shell.
       p.get_result()
@@ -420,8 +495,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     Additionally, also test that comments are preserved.
     """
     # readline gets its input from tty, so using stdin does not work.
-    shell_cmd = get_shell_cmd(vector)
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+    child_proc = spawn_shell(get_shell_cmd(vector))
     # List of (input query, expected text in output).
     # The expected output is usually the same as the input with a number prefix, except
     # where the shell strips newlines before a semicolon.
@@ -451,25 +525,28 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     occur."""
     # readline gets its input from tty, so using stdin does not work.
     shell_cmd = get_shell_cmd(vector)
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
-    # set up history
+    child_proc = spawn_shell(shell_cmd)
+
+    # initialize history
     child_proc.expect(PROMPT_REGEX)
     child_proc.sendline("select 1;")
     child_proc.expect("Fetched 1 row\(s\) in [0-9]+\.?[0-9]*s")
     child_proc.expect(PROMPT_REGEX)
     child_proc.sendline("quit;")
     child_proc.wait()
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
-    child_proc.expect(PROMPT_REGEX)
 
-    # send SIGINT then quit to save history
+    # create a new shell and send SIGINT
+    child_proc = spawn_shell(shell_cmd)
+    child_proc.expect(PROMPT_REGEX)
     child_proc.sendintr()
+    child_proc.expect("\^C")
     child_proc.sendline("select 2;")
     child_proc.expect("Fetched 1 row\(s\) in [0-9]+\.?[0-9]*s")
+    child_proc.expect(PROMPT_REGEX)
     child_proc.sendline("quit;")
     child_proc.wait()
 
-    # check history in a new instance
+    # check history in a new shell instance
     p = ImpalaShell(vector)
     p.send_cmd('history')
     result = p.get_result().stderr.splitlines()
@@ -487,7 +564,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     """
     with NamedTemporaryFile() as new_hist:
       shell_cmd = get_shell_cmd(vector) + ["--history_file=%s" % new_hist.name]
-      child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+      child_proc = spawn_shell(shell_cmd)
       child_proc.expect(":{0}] default>".format(get_impalad_port(vector)))
       self._expect_with_cmd(child_proc, "select 'hi'", vector, ('hi'))
       child_proc.sendline('exit;')
@@ -497,8 +574,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
 
   def test_rerun(self, vector, tmp_history_file):
     """Smoke test for the 'rerun' command"""
-    shell_cmd = get_shell_cmd(vector)
-    child_proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+    child_proc = spawn_shell(get_shell_cmd(vector))
     child_proc.expect(":{0}] default>".format(get_impalad_port(vector)))
     self._expect_with_cmd(child_proc, "@1", vector, ("Command index out of range"))
     self._expect_with_cmd(child_proc, "rerun -1", vector,
@@ -586,9 +662,9 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     # Positive tests
     # set live_summary and live_progress as True with config file
     rcfile_path = os.path.join(QUERY_FILE_PATH, 'good_impalarc3')
-    args = ['--config_file=%s' % rcfile_path]
+    cmd_line_args = ['--config_file=%s' % rcfile_path]
     cmds = "set all;"
-    result = run_impala_shell_interactive(vector, cmds, shell_args=args)
+    result = run_impala_shell_interactive(vector, cmds, shell_args=cmd_line_args)
     assert 'WARNING:' not in result.stderr, \
       "A valid config file should not trigger any warning: {0}".format(result.stderr)
     assert "\tLIVE_SUMMARY: True" in result.stdout
@@ -730,6 +806,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
       result = run_impala_shell_interactive(vector,
         "sOuRcE shell_case_sensitive.cmds; SeLeCt 'second command'")
       print result.stderr
+
       assert "Query: uSe FUNCTIONAL" in result.stderr
       assert "Query: ShOw TABLES" in result.stderr
       assert "alltypes" in result.stdout
@@ -759,32 +836,25 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     result = run_impala_shell_interactive(vector, '/* comment */\n'
                                           'select * from {0};'.format(table))
     assert 'Fetched 1 row(s)' in result.stderr
-
     result = run_impala_shell_interactive(vector, '/* comment1 */\n'
                                           '-- comment2\n'
                                           'select * from {0};'.format(table))
     assert 'Fetched 1 row(s)' in result.stderr
-
     result = run_impala_shell_interactive(vector, '/* comment1\n'
                                           'comment2 */ select * from {0};'.format(table))
     assert 'Fetched 1 row(s)' in result.stderr
-
     result = run_impala_shell_interactive(vector, '/* select * from {0} */ '
                                           'select * from {0};'.format(table))
     assert 'Fetched 1 row(s)' in result.stderr
-
     result = run_impala_shell_interactive(vector, '/* comment */ help use')
     assert 'Executes a USE... query' in result.stdout
-
     result = run_impala_shell_interactive(vector, '-- comment\n'
                                           ' help use;')
     assert 'Executes a USE... query' in result.stdout
-
     result = run_impala_shell_interactive(vector, '/* comment1 */\n'
                                           '-- comment2\n'
                                           'desc {0};'.format(table))
     assert 'Fetched 1 row(s)' in result.stderr
-
     result = run_impala_shell_interactive(vector, '/* comment1 */\n'
                                           '-- comment2\n'
                                           'help use;')
@@ -830,6 +900,31 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
 
   def test_fix_infinite_loop(self, vector):
     # IMPALA-6337: Fix infinite loop.
+
+    # In case of TL;DR:
+    # - see IMPALA-9362 for details
+    # - see tests/shell/util.py for explanation of IMPALA_SHELL_EXECUTABLE
+    if os.getenv("IMPALA_HOME") not in IMPALA_SHELL_EXECUTABLE:
+      # The fix for IMPALA-6337 involved patching our internal verison of
+      # sqlparse 0.1.19 in ${IMPALA_HOME}/shell/ext-py. However, when we
+      # create the the stand-alone python package of the impala-shell for PyPI,
+      # we don't include the bundled 3rd party libs -- we expect users to
+      # install 3rd upstream libraries from PyPI.
+      #
+      # We could try to bundle sqlparse with the PyPI package, but there we
+      # run into the issue that the our bundled version is not python 3
+      # compatible. The real fix for this would be to upgrade to sqlparse 0.3.0,
+      # but that's not without complications. See IMPALA-9362 for details.
+      #
+      # For the time being, what this means is that IMPALA-6337 is fixed for
+      # people who are running the shell locally from any host/node that's part
+      # of a cluster where Impala is installed, but if they are running a
+      # standalone version of the shell on a client outside of a cluster, then
+      # they will still be relying on the upstream version of sqlparse 0.1.19,
+      # and so they may still be affected by the IMPALA-6337.
+      #
+      pytest.skip("Test will fail if shell is not part of dev environment.")
+
     result = run_impala_shell_interactive(vector, "select 1 + 1; \"\n;\";")
     assert '| 2     |' in result.stdout
     result = run_impala_shell_interactive(vector, "select '1234'\";\n;\n\";")
@@ -860,7 +955,7 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
 
   def test_shell_prompt(self, vector):
     shell_cmd = get_shell_cmd(vector)
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:])
+    proc = spawn_shell(shell_cmd)
     proc.expect(":{0}] default>".format(get_impalad_port(vector)))
     self._expect_with_cmd(proc, "use foo", vector, (), 'default')
     self._expect_with_cmd(proc, "use functional", vector, (), 'functional')
@@ -868,20 +963,20 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     self._expect_with_cmd(proc, 'use `tpch`', vector, (), 'tpch')
     self._expect_with_cmd(proc, 'use ` tpch `', vector, (), 'tpch')
 
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:] + ['-d', 'functional'])
+    proc = spawn_shell(shell_cmd + ['-d', 'functional'])
     proc.expect(":{0}] functional>".format(get_impalad_port(vector)))
     self._expect_with_cmd(proc, "use foo", vector, (), 'functional')
     self._expect_with_cmd(proc, "use tpch", vector, (), 'tpch')
     self._expect_with_cmd(proc, "use foo", vector, (), 'tpch')
 
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:] + ['-d', ' functional '])
+    proc = spawn_shell(shell_cmd + ['-d', ' functional '])
     proc.expect(":{0}] functional>".format(get_impalad_port(vector)))
 
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:] + ['-d', '` functional `'])
+    proc = spawn_shell(shell_cmd + ['-d', '` functional `'])
     proc.expect(":{0}] functional>".format(get_impalad_port(vector)))
 
     # Start an Impala shell with an invalid DB.
-    proc = pexpect.spawn(shell_cmd[0], shell_cmd[1:] + ['-d', 'foo'])
+    proc = spawn_shell(shell_cmd + ['-d', 'foo'])
     proc.expect(":{0}] default>".format(get_impalad_port(vector)))
     self._expect_with_cmd(proc, "use foo", vector, (), 'default')
     self._expect_with_cmd(proc, "use functional", vector, (), 'functional')
@@ -946,6 +1041,15 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     assert "ERROR: ParseException: Unmatched string literal" in result.stderr,\
            result.stderr
 
+  def test_utf8_error_message(self, vector):
+    """Test UTF-8 error messages are shown correctly"""
+    shell = ImpalaShell(vector)
+    query = "select cast(now() as string format 'yyyy年MM月dd日')"
+    shell.send_cmd(query)
+    result = shell.get_result()
+    assert "ERROR: Bad date/time conversion format: yyyy年MM月dd日" in result.stderr,\
+           result.stderr
+
   def test_timezone_validation(self, vector):
     """Test that query option TIMEZONE is validated when executing a query.
 
@@ -966,39 +1070,52 @@ class TestImpalaShellInteractive(ImpalaTestSuite):
     for keyword in ["insert", "upsert", "update", "delete", "\\'insert\\'",
                     "\\'upsert\\'", "\\'update\\'", "\\'delete\\'"]:
       p = ImpalaShell(vector)
-      p.send_cmd("with foo as "
-                 "(select * from functional.alltypestiny where string_col='%s') "
-                 "select * from foo limit 1" % keyword)
+      cmd = ("with foo as "
+             "(select * from functional.alltypestiny where string_col='%s') "
+             "select * from foo limit 1" % keyword)
+      p.send_cmd(cmd)
       result = p.get_result()
       assert "Fetched 0 row" in result.stderr
 
-  def test_http_codes(self, vector):
-    """Check that the shell prints a good message when using hs2-http protocol
-    and the http server returns a 503 error."""
+  def test_quotes_in_with_clause(self, vector):
+    # IMPALA-10051: This test verifies that the fix prevents ValueErrors caused by
+    # shlex library when quotes and whitespace characters are mixed.
+    p = ImpalaShell(vector)
+    cmd = ("with foo as "
+           "(select *, regexp_replace(string_col,\"[a-zA-Z]\",\"+ \") "
+           "from functional.alltypestiny) "
+           "select * from foo limit 1")
+    p.send_cmd(cmd)
+    result = p.get_result()
+    assert "Fetched 1 row" in result.stderr
+
+  def test_http_interactions(self, vector, http_503_server):
+    """Test interactions with the http server when using hs2-http protocol.
+    Check that the shell prints a good message when the server returns a 503 error."""
     protocol = vector.get_value("protocol")
     if protocol != 'hs2-http':
       pytest.skip()
 
-    # Start an http server that always returns 503.
-    HOST = "localhost"
-    PORT = get_unused_port()
-    httpd = None
-    http_server_thread = None
-    try:
-      httpd = SocketServer.TCPServer((HOST, PORT), UnavailableRequestHandler)
-      http_server_thread = threading.Thread(target=httpd.serve_forever)
-      http_server_thread.start()
+    # Check that we get a message about the 503 error when we try to connect.
+    shell_args = ["--protocol={0}".format(protocol),
+                  "-i{0}:{1}".format(http_503_server.HOST, http_503_server.PORT)]
+    shell_proc = spawn_shell([IMPALA_SHELL_EXECUTABLE] + shell_args)
+    shell_proc.expect("HTTP code 503", timeout=10)
 
-      # Check that we get a message about the 503 error when we try to connect.
-      shell_args = ["--protocol={0}".format(protocol), "-i{0}:{1}".format(HOST, PORT)]
-      shell_proc = pexpect.spawn(IMPALA_SHELL_EXECUTABLE, shell_args)
-      shell_proc.expect("HTTP code 503", timeout=10)
-    finally:
-      # Clean up.
-      if httpd is not None:
-        httpd.shutdown()
-      if http_server_thread is not None:
-        http_server_thread.join()
+  def test_http_interactions_extra(self, vector, http_503_server_extra):
+    """Test interactions with the http server when using hs2-http protocol.
+    Check that the shell prints a good message when the server returns a 503 error,
+    including the body text from the message."""
+    protocol = vector.get_value("protocol")
+    if protocol != 'hs2-http':
+      pytest.skip()
+
+    # Check that we get a message about the 503 error when we try to connect.
+    shell_args = ["--protocol={0}".format(protocol),
+                  "-i{0}:{1}".format(http_503_server_extra.HOST,
+                                     http_503_server_extra.PORT)]
+    shell_proc = spawn_shell([IMPALA_SHELL_EXECUTABLE] + shell_args)
+    shell_proc.expect("HTTP code 503: Service Unavailable \[EXTRA\]", timeout=10)
 
 
 def run_impala_shell_interactive(vector, input_lines, shell_args=None,
@@ -1014,5 +1131,6 @@ def run_impala_shell_interactive(vector, input_lines, shell_args=None,
   p = ImpalaShell(vector, args=shell_args, env=my_env,
       wait_until_connected=wait_until_connected)
   for line in input_lines:
+    LOG.info("Running %s" % line)
     p.send_cmd(line)
   return p.get_result()
