@@ -18,39 +18,18 @@
 import pytest
 import random
 import time
-import traceback
 
 from multiprocessing import Value
-from multiprocessing.pool import ThreadPool
 
 from tests.common.impala_test_suite import ImpalaTestSuite
 from tests.common.parametrize import UniqueDatabase
-from tests.common.skip import SkipIfHive2, SkipIfS3
+from tests.common.skip import SkipIfHive2, SkipIfS3, SkipIfGCS
+from tests.stress.stress_util import Task, run_tasks
 
 NUM_OVERWRITES = 2
 NUM_INSERTS_PER_OVERWRITE = 4
 
 
-class Task:
-  """Helper class for parallel execution."""
-  def __init__(self, func, *args, **kwargs):
-    self.func = func
-    self.args = args
-    self.kwargs = kwargs
-
-  def run(self):
-    try:
-      return self.func(*self.args, **self.kwargs)
-    except Exception:
-      traceback.print_exc()
-      raise
-
-
-def run_tasks(tasks):
-  pool = ThreadPool(processes=len(tasks))
-  pool.map_async(Task.run, tasks).get(600)
-
-
 class TestAcidStress(ImpalaTestSuite):
   @classmethod
   def get_workload(self):
@@ -181,6 +160,7 @@ class TestAcidInsertsBasic(TestAcidStress):
 
   @SkipIfHive2.acid
   @SkipIfS3.hive
+  @SkipIfGCS.hive
   @pytest.mark.execute_serially
   @pytest.mark.stress
   def test_read_hive_inserts(self, unique_database):
@@ -248,6 +228,7 @@ class TestConcurrentAcidInserts(TestAcidStress):
     finally:
       impalad_client.close()
 
+  @SkipIfGCS.jira(reason="IMPALA-10563")
   @SkipIfHive2.acid
   @pytest.mark.execute_serially
   @pytest.mark.stress
@@ -343,6 +324,7 @@ class TestFailingAcidInserts(TestAcidStress):
                 for i in xrange(0, num_checkers)]
     run_tasks(writers + checkers)
 
+  @SkipIfGCS.jira(reason="IMPALA-10563")
   @SkipIfHive2.acid
   @pytest.mark.execute_serially
   @pytest.mark.stress
