@@ -31,7 +31,6 @@ import requests
 class TestWebPage(ImpalaTestSuite):
 
   ROOT_URL = "http://localhost:{0}/"
-  GET_JAVA_LOGLEVEL_URL = "http://localhost:{0}/get_java_loglevel"
   SET_JAVA_LOGLEVEL_URL = "http://localhost:{0}/set_java_loglevel"
   RESET_JAVA_LOGLEVEL_URL = "http://localhost:{0}/reset_java_loglevel"
   SET_GLOG_LOGLEVEL_URL = "http://localhost:{0}/set_glog_level"
@@ -50,6 +49,7 @@ class TestWebPage(ImpalaTestSuite):
   RESET_RESOURCE_POOL_STATS_URL = "http://localhost:{0}/resource_pool_reset"
   BACKENDS_URL = "http://localhost:{0}/backends"
   PROMETHEUS_METRICS_URL = "http://localhost:{0}/metrics_prometheus"
+  QUERIES_URL = "http://localhost:{0}/queries"
   HEALTHZ_URL = "http://localhost:{0}/healthz"
 
   # log4j changes do not apply to the statestore since it doesn't
@@ -201,49 +201,31 @@ class TestWebPage(ImpalaTestSuite):
     malformed inputs. This however does not test that the log level changes are actually
     in effect."""
     # Check that the log_level end points are accessible.
-    self.get_and_check_status_jvm(self.GET_JAVA_LOGLEVEL_URL)
     self.get_and_check_status_jvm(self.SET_JAVA_LOGLEVEL_URL)
     self.get_and_check_status_jvm(self.RESET_JAVA_LOGLEVEL_URL)
     self.get_and_check_status(self.SET_GLOG_LOGLEVEL_URL)
     self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL)
-    # Try getting log level of a class.
-    get_loglevel_url = (self.GET_JAVA_LOGLEVEL_URL + "?class" +
-        "=org.apache.impala.catalog.HdfsTable")
-    self.get_and_check_status_jvm(get_loglevel_url, "DEBUG")
 
     # Set the log level of a class to TRACE and confirm the setting is in place
     set_loglevel_url = (self.SET_JAVA_LOGLEVEL_URL + "?class" +
         "=org.apache.impala.catalog.HdfsTable&level=trace")
-    self.get_and_check_status_jvm(set_loglevel_url, "Effective log level: TRACE")
+    self.get_and_check_status_jvm(
+      set_loglevel_url,
+      "org.apache.impala.catalog.HdfsTable : TRACE")
 
-    get_loglevel_url = (self.GET_JAVA_LOGLEVEL_URL + "?class" +
-        "=org.apache.impala.catalog.HdfsTable")
-    self.get_and_check_status_jvm(get_loglevel_url, "TRACE")
-    # Check the log level of a different class and confirm it is still DEBUG
-    get_loglevel_url = (self.GET_JAVA_LOGLEVEL_URL + "?class" +
-        "=org.apache.impala.catalog.HdfsPartition")
-    self.get_and_check_status_jvm(get_loglevel_url, "DEBUG")
-
-    # Reset Java logging levels and check the logging level of the class again
+    # Reset Java logging levels
     self.get_and_check_status_jvm(self.RESET_JAVA_LOGLEVEL_URL, "Java log levels reset.")
-    get_loglevel_url = (self.GET_JAVA_LOGLEVEL_URL + "?class" +
-        "=org.apache.impala.catalog.HdfsTable")
-    self.get_and_check_status_jvm(get_loglevel_url, "DEBUG")
 
     # Set a new glog level and make sure the setting has been applied.
     set_glog_url = (self.SET_GLOG_LOGLEVEL_URL + "?glog=3")
-    self.get_and_check_status(set_glog_url, "v set to 3")
+    self.get_and_check_status(set_glog_url, "val(3)")
 
     # Try resetting the glog logging defaults again.
-    self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL, "v set to ")
-
-    # Try to get the log level of an empty class input
-    get_loglevel_url = (self.GET_JAVA_LOGLEVEL_URL + "?class=")
-    self.get_and_check_status_jvm(get_loglevel_url)
+    self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL, "Current backend log level: ")
 
     # Same as above, for set log level request
     set_loglevel_url = (self.SET_JAVA_LOGLEVEL_URL + "?class=")
-    self.get_and_check_status_jvm(get_loglevel_url)
+    self.get_and_check_status_jvm(set_loglevel_url)
 
     # Empty input for setting a glog level request
     set_glog_url = (self.SET_GLOG_LOGLEVEL_URL + "?glog=")
@@ -254,7 +236,9 @@ class TestWebPage(ImpalaTestSuite):
     # Level.toLevel() method.
     set_loglevel_url = (self.SET_JAVA_LOGLEVEL_URL + "?class" +
         "=org.apache.impala.catalog.HdfsTable&level=foo&")
-    self.get_and_check_status_jvm(set_loglevel_url, "Effective log level: DEBUG")
+    self.get_and_check_status_jvm(
+      set_loglevel_url,
+      "org.apache.impala.catalog.HdfsTable : DEBUG")
 
     # Try setting an invalid glog level.
     set_glog_url = self.SET_GLOG_LOGLEVEL_URL + "?glog=foo"
@@ -274,11 +258,11 @@ class TestWebPage(ImpalaTestSuite):
     self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL)
     # Set log level to 3.
     set_glog_url = (self.SET_GLOG_LOGLEVEL_URL + "?glog=3")
-    self.get_and_check_status(set_glog_url, "v set to 3")
+    self.get_and_check_status(set_glog_url, "val(3)")
     # Check that Impala doesn't crash when running a query that aggregates.
     self.client.execute("select avg(int_col) from functional.alltypessmall")
     # Reset log level.
-    self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL, "v set to ")
+    self.get_and_check_status(self.RESET_GLOG_LOGLEVEL_URL, "Current backend log level: ")
 
   def test_catalog(self, cluster_properties, unique_database):
     """Tests the /catalog and /catalog_object endpoints."""
@@ -315,6 +299,10 @@ class TestWebPage(ImpalaTestSuite):
     self.__test_catalog_object(unique_database, "foo_kudu", cluster_properties)
     self.__test_catalog_object(unique_database, "foo_part_parquet", cluster_properties)
     self.__test_catalog_object(unique_database, "foo", cluster_properties)
+    self.__test_json_db_object(unique_database)
+    self.__test_json_table_object(unique_database, "foo")
+    self.__test_json_table_object(unique_database, "foo_part")
+    self.__test_json_table_object(unique_database, "foo_part_parquet")
     self.__test_table_metrics(unique_database, "foo_part", "total-file-size-bytes")
     self.__test_table_metrics(unique_database, "foo_part", "num-files")
     self.__test_table_metrics(unique_database, "foo_part", "alter-duration")
@@ -351,6 +339,52 @@ class TestWebPage(ImpalaTestSuite):
       self.get_and_check_status(obj_url, impalad_expected_str,
           ports_to_test=self.IMPALAD_TEST_PORT)
 
+  def __test_json_db_object(self, db_name):
+    """Tests the /catalog_object?json endpoint of catalogd for the given db."""
+    obj_url = self.CATALOG_OBJECT_URL + \
+              "?json&object_type=DATABASE&object_name={0}".format(db_name)
+    responses = self.get_and_check_status(obj_url, ports_to_test=self.CATALOG_TEST_PORT)
+    obj = json.loads(json.loads(responses[0].text)["json_string"])
+    assert obj["type"] == 2, "type should be DATABASE"
+    assert "catalog_version" in obj, "TCatalogObject should have catalog_version"
+    db_obj = obj["db"]
+    assert db_obj["db_name"] == db_name
+    assert "metastore_db" in db_obj, "Loaded database should have metastore_db"
+
+  def __test_json_table_object(self, db_name, tbl_name):
+    """Tests the /catalog_object?json endpoint of catalogd for the given db/table. Runs
+    against an unloaded as well as a loaded table."""
+    obj_url = self.CATALOG_OBJECT_URL + \
+              "?json&object_type=TABLE&object_name={0}.{1}".format(db_name, tbl_name)
+    self.client.execute("invalidate metadata %s.%s" % (db_name, tbl_name))
+    responses = self.get_and_check_status(obj_url, ports_to_test=self.CATALOG_TEST_PORT)
+    obj = json.loads(json.loads(responses[0].text)["json_string"])
+    assert obj["type"] == 3, "type should be TABLE"
+    assert "catalog_version" in obj, "TCatalogObject should have catalog_version"
+    tbl_obj = obj["table"]
+    assert tbl_obj["db_name"] == db_name
+    assert tbl_obj["tbl_name"] == tbl_name
+    assert "hdfs_table" not in tbl_obj, "Unloaded table should not have hdfs_table"
+
+    self.client.execute("refresh %s.%s" % (db_name, tbl_name))
+    responses = self.get_and_check_status(obj_url, ports_to_test=self.CATALOG_TEST_PORT)
+    obj = json.loads(json.loads(responses[0].text)["json_string"])
+    assert obj["type"] == 3, "type should be TABLE"
+    assert "catalog_version" in obj, "TCatalogObject should have catalog_version"
+    tbl_obj = obj["table"]
+    assert tbl_obj["db_name"] == db_name
+    assert tbl_obj["tbl_name"] == tbl_name
+    assert "columns" in tbl_obj, "Loaded TTable should have columns"
+    assert tbl_obj["table_type"] == 0, "table_type should be HDFS_TABLE"
+    assert "metastore_table" in tbl_obj
+    hdfs_tbl_obj = tbl_obj["hdfs_table"]
+    assert "hdfsBaseDir" in hdfs_tbl_obj
+    assert "colNames" in hdfs_tbl_obj
+    assert "nullPartitionKeyValue" in hdfs_tbl_obj
+    assert "nullColumnValue" in hdfs_tbl_obj
+    assert "partitions" in hdfs_tbl_obj
+    assert "prototype_partition" in hdfs_tbl_obj
+
   def check_endpoint_is_disabled(self, url, string_to_search="", ports_to_test=None):
     """Helper method that verifies the given url does not exist."""
     if ports_to_test is None:
@@ -417,6 +451,23 @@ class TestWebPage(ImpalaTestSuite):
       if tblinfo["name"] == tbl_fname:
         assert tblinfo["num_files"] == int(numfiles)
 
+  def test_query_stmt(self):
+    """Create a long select query then check if it is truncated in the response json."""
+    # The imput query is a select + 450 "x " long, which is long enough to get truncated.
+    query = "select \"{0}\"".format("x " * 450)
+    # The expected result query should be 253 long and contains the first 250
+    # chars + "..."
+    expected_result = "select \"{0}...".format("x " * 121)
+    check_if_contains = False
+    self.execute_query(query)
+    response_json = self.__run_query_and_get_debug_page(query, self.QUERIES_URL)
+    # Search the json for the expected value.
+    for json_part in response_json['completed_queries']:
+      if expected_result in json_part['stmt']:
+        check_if_contains = True
+        break
+    assert check_if_contains, "No matching statement found in the jsons."
+
   def __run_query_and_get_debug_page(self, query, page_url, query_options=None,
                                      expected_state=None):
     """Runs a query to obtain the content of the debug page pointed to by page_url, then
@@ -621,12 +672,7 @@ class TestWebPage(ImpalaTestSuite):
     # check that metrics exist
     assert 'max_query_mem_limit' in pool_config
     assert 'min_query_mem_limit' in pool_config
-    assert 'max_running_queries_multiple' in pool_config
-    assert 'max_memory_multiple' in pool_config
     assert 'clamp_mem_limit_query_option' in pool_config
-    assert 'max_running_queries_derived' in pool_config
-    assert 'max_queued_queries_derived' in pool_config
-    assert 'max_memory_derived' in pool_config
 
   def __fetch_resource_pools_json(self, pool_name=None):
     """Helper method used to fetch the resource pool json from the admission debug page.
@@ -665,8 +711,8 @@ class TestWebPage(ImpalaTestSuite):
 
     # The 'address' column is the backend port of the impalad.
     assert len(backend_row['address']) > 0
-    be_ports = ('22000', '22001', '22002')
-    assert backend_row['address'].endswith(be_ports)
+    krpc_ports = ('27000', '27001', '27002')
+    assert backend_row['address'].endswith(krpc_ports)
 
     # The 'krpc_address' is the krpc address of the impalad.
     assert len(backend_row['krpc_address']) > 0
